{
        Program Name : App
        Package Name : LB
        Purpose      : Global application interface
        Institution  : Conroy & Conroy Co.
        Date Written : 31-January-2003
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2003 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  06-Mar-2006   RBJ         Cleaned up memory leaks.       *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          The classes in this unit are used to provide an OO application
	interface.
}

unit App ;

interface

{$I EDEFINES.INC}

uses { Layer 1... }
     _App, { TApplication_Manager }
     _DebugMa, { TCOM_Debug_Manager }
     Heaps, { PSHeap }
     _ObjectM ; { TCOM_Object_Manager }

{$IFDEF DELPHI}
function Static_Application_Manager_COM_Wrapper( COM : TCOM_Application_Manager ;
    Old : PSApplication_Manager ) : PSApplication_Manager ;
function Static_Debug_Manager_Wrapper( Value : TCOM_Debug_Manager ) : PSDebug_Manager ;
function Static_Object_Manager_Wrapper( Value : TCOM_Object_Manager ) : PSObject_Manager ;
function COM_Unit_Debugger_Wrapper( Value : PSUnit_Debugger ) : TCOM_Unit_Debugger ;
{$ENDIF}

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

procedure Terminate ; { Optional call to shut down application manager }

implementation

uses { Borland... }
{$IFDEF Delphi}
     SysUtils, { Exception }
     Dialogs, { ShowMessage }
{$ELSE}
     Objects, { TCollection }
{$ENDIF}

     { Layer 0... }
     _AEL, { PSApplication_Event_Logger }
     _APM, { PSApplication_Performance_Monitor }
     _DebugIn, { PSDebug_Interface }
     _FacMan, { PSFacility_Manager }
     _Common, { MT_Heap }

     { Layer 1... }
     AEL, { TStandard_Application_Event_Logger }
     APM, { TSStandard_Application_Performance_Monitor }
     Collect, { TCollection }
     CommonUt, { Boolean_To_String }
     DebugInt, { TText_Debugger }
     MStores, { PSManaged_Store }
     _Streams, { TStream }
{$IFDEF Delphi}
     VCL_Std, { TVCL_Debugger }
{$ENDIF}
     O_S ; { OS }

type TSStandard_Application_Manager = object( TSApplication_Manager )
                                          private { Instance data... }
                                              _Event_Logger : PSApplication_Event_Logger ;
                                              _Debug_Manager : PSDebug_Manager ;
                                              _Object_Manager : PSObject_Manager ;
                                              _Halting : boolean ;
                                              _Heap : PSManaged_Store ;
                                              _In_Shut_Down : boolean ;
                                              _Performance_Monitor : PSApplication_Performance_Monitor ;
                                              _Program_Name : string ;
                                              _Version : string ;

                                          public { Constructors and destructors... }
                                              constructor Init ;

                                              destructor Done ; virtual;

                                          public { API... }
                                              { Status... }
                                              function Get_In_Shut_Down : boolean ;
                                                  virtual ;

                                              procedure Set_In_Shut_Down( Value : boolean ) ;
                                                  virtual ;

                                              function Get_Halting : boolean ;
                                                  virtual ;

                                              procedure Set_Halting( Value : boolean ) ;
                                                  virtual ;

                                              { Debugging... }
                                              function Get_Debug_Manager : PSDebug_Manager ;
                                                  virtual ;

                                              procedure Set_Debug_Manager( Value : PSDebug_Manager ) ;
                                                  virtual ;

                                              function Application_Debugger : PSDebug_Interface ;
                                                  virtual ;

                                              function Debugger : PSDebug_Interface ;
                                                  virtual ;

                                              function Object_Manager : PSObject_Manager ;
                                                  virtual ;

                                              { Idle processing... }
                                              procedure Do_Idle ; virtual ;

                                              procedure Add_Idle_Delegation( Value : PSIdle_Consumer ) ;
                                                  virtual ;

                                              procedure Remove_Idle_Delegation( Value : PSIdle_Consumer ) ;
                                                  virtual ;

                                              { Event logging... }
                                              function Get_Event_Logger : PSApplication_Event_Logger ;
                                                  virtual ;

                                              procedure Set_Event_Logger( EV : PSApplication_Event_Logger ) ;
                                                  virtual ;

                                              { Event logging... }
                                              function Get_Performance_Monitor : PSApplication_Performance_Monitor ;
                                                  virtual ;

                                              procedure Set_Performance_Monitor( PM : PSApplication_Performance_Monitor ) ;
                                                  virtual ;

                                              { Other... }
                                              procedure Fatal_Exception( const S : string ) ;
                                                  virtual ;

                                              procedure Popup_Message( S : string ) ;
                                                  virtual ;

                                              function Get_Heap : PSManaged_Store ;
                                                  virtual ;

                                              procedure Set_Heap( Value : PSManaged_Store ) ;
                                                  virtual ;

                                              function Get_Name : string ;
                                                  virtual ;

                                              procedure Set_Name( const Value : string ) ;
                                                  virtual ;

                                              function Get_Serial_Number( Pool : longint ) : longint ;
                                                  virtual ;

                                              { Full version string including "V" }
                                              function Get_Version : string ;
                                                  virtual ;

                                              procedure Set_Version( const Value : string ) ;
                                                   virtual ;
                                      end ;
     PSStandard_Application_Manager = ^TSStandard_Application_Manager ;

type TSApplication_Debugger = object( TSText_Debugger )
                                  private
                                      _OS : PSDebug_Interface ;
                                      _AM : PSDebug_Interface ;
                                      _VC : PSDebug_Interface ;
                                      _Debuggers : array[ 0..2 ] of PSDebug_Interface ;

                                  public
                                      constructor Init ;

                                      function Count : longint ; virtual ;

                                      function Child( Ordinal : longint ) : PSDebug_Interface ;
                                          virtual ;
                               end ;
     PSApplication_Debugger = ^TSApplication_Debugger ;

type TSApplication_Manager_Debugger = object( TSText_Debugger )
                                         public
                                             function Count : longint ; virtual ;

                                             function Child( Ordinal : longint ) : PSDebug_Interface ;
                                                 virtual ;
                                      end ;
     PSApplication_Manager_Debugger = ^TSApplication_Manager_Debugger ;

     TSUnits_Debugger = object( TSText_Debugger )
                          public
                              function Count : longint ; virtual ;

                              function Child( Ordinal : longint ) : PSDebug_Interface ;
                                  virtual ;
                       end ;
     PSUnits_Debugger = ^TSUnits_Debugger ;

type TSApplication_Object_Manager = object( TSObject_Manager )
                                        public { Constructors and destructors... }
                                            constructor Init ;
                                            destructor Done ; virtual ;

    				        private { Instance data }
                                            _Objects : TCollection ;
                                            _Facilities : TCollection ;
                                            _Contexts : TCollection ;

                                        public { API... }
                                            procedure Add_Object( O : pointer ;
                                                Facility : integer ;
                                                Context : pointer ) ;
                                                virtual ;

                                            function Find_Object( O : pointer ;
                                                Facility : integer ) : boolean ;
                                                virtual ;

                                            function Get_Context( Index : integer ) : pointer ;
                                                virtual ;

                                            function Get_Facility( Index : integer ) : integer ;
                                                virtual ;

                                            function Get_Object( Index : integer ) : pointer ;
                                                virtual ;

                                            procedure Remove_Object( O : pointer ;
                                                Facility : longint ) ;
                                                virtual ;

                                            function Object_Count : longint ;
                                                virtual ;

                                            function Construct( Stream : PSStream ) : pointer ;
                                                virtual ;

                                            procedure Register_Facility( Facility_Manager : PSFacility_Manager ) ;
                                                virtual ;

                                            procedure Remove_Facility( Facility_Manager : PSFacility_Manager ) ;
                                                virtual ;

                                            function Get_Facility_Manager( Index : longint ) : PSFacility_Manager ;
                                                virtual ;
                                    end ;
     PSApplication_Object_Manager = ^TSApplication_Object_Manager ;

{$IFDEF DELPHI}
     TStandard_Application_Manager = class( TApplication_Manager )
                                         private { Instance data... }
                                              _Event_Logger : TApplication_Event_Logger ;
                                              _Debug_Manager : TDebug_Manager ;
                                              _Object_Manager : TObject_Manager ;
                                              _Performance_Monitor : TApplication_Performance_Monitor ;
                                              _Halting : boolean ;
                                              _Heap : TManaged_Store ;

                                         public { API... }
                                             { Status... }
                                             function Get_In_Shut_Down : boolean ;
                                                 override ;

                                             procedure Set_In_Shut_Down( Value : boolean ) ;
                                                 override ;

                                             function Get_Halting : boolean ;
                                                 override ;

                                             procedure Set_Halting( Value : boolean ) ;
                                                 override ;

                                             { Debugging... }
                                             function Get_Debug_Manager : TDebug_Manager ;
                                                 override ;

                                             procedure Set_Debug_Manager( Value : TDebug_Manager ) ;
                                                 override ;

                                             property Debug_Manager : TDebug_Manager
                                                 read Get_Debug_Manager
                                                 write Set_Debug_Manager ;

                                             function Application_Debugger : TDebug_Interface ;
                                                 override ;

                                             function Debugger : TDebug_Interface ;
                                                 override ;

                                             function Object_Manager : TObject_Manager ;
                                                 override ;

                                             { Idle processing... }
                                             procedure Do_Idle ; override ;

                                             procedure Add_Idle_Delegation( Value : TIdle_Consumer ) ;
                                                 override ;

                                             procedure Remove_Idle_Delegation( Value : TIdle_Consumer ) ;
                                                 override ;

                                             { Event logging... }
                                             function Get_Event_Logger : TApplication_Event_Logger ;
                                                 override ;

                                             procedure Set_Event_Logger( EV : TApplication_Event_Logger ) ;
                                                 override ;

                                             { Performance... }
                                             function Get_Performance_Monitor : TApplication_Performance_Monitor ;
                                                 override ;

                                             procedure Set_Performance_Monitor( PM : TApplication_Performance_Monitor ) ;
                                                 override ;

                                             { Other... }
                                             procedure Fatal_Exception( const S : string ) ;
                                                 override ;

                                             procedure Popup_Message( const S : string ) ;
                                                 override ;

                                             function Get_Heap : TManaged_Store ;
                                                 override ;

                                             procedure Set_Heap( Value : TManaged_Store ) ;
                                                 override ;

                                             function Get_Name : string ;
                                                 override ;

                                             procedure Set_Name( const Value : string ) ;
                                                 override ;

                                             function Get_Serial_Number( Pool : longint ) : longint ;
                                                 override ;

                                             function Get_Clipboard( const Path : string ) : string ;
                                                 override ;

                                             procedure Set_Clipboard( const Path, Value : string ) ;
                                                 override ;

                                             { Full version string including "V" }
                                             function Get_Version : string ;
                                                 override ;

                                             procedure Set_Version( const Value : string ) ;
                                                 override ;
                                     end ; { TStandard_Application_Manager }

type TApplication_Debugger = class( TText_Debugger )
                                private
                                    _OS : TDebug_Interface ;
                                    _AM : TDebug_Interface ;
                                    _VC : TDebug_Interface ;
                                    _Debuggers : array[ 0..2 ] of TDebug_Interface ;

                                public
                                    constructor Create ;

                                    function Count : longint ; override ;

                                    function Child( Ordinal : longint ) : TCOM_Debug_Interface ;
                                        override ;
                             end ;

type TApplication_Manager_Debugger = class( TText_Debugger )
                                         public
                                             function Count : longint ; override ;

                                             function Child( Ordinal : longint ) : TCOM_Debug_Interface ;
                                                 override ;
                                      end ;

type TUnits_Debugger = class( TText_Debugger )
                                         public
                                             function Count : longint ; override ;

                                             function Child( Ordinal : longint ) : TCOM_Debug_Interface ;
                                                 override ;
                                      end ;

type TApplication_Object_Manager = class( TObject_Manager )
                                       public { Constructors and destructors... }
                                           constructor Create ;
                                           destructor Destroy ; override ;

      				        private { Instance data }
                                           _Objects : TCollection ;
                                           _Facilities : TCollection ;
                                           _Contexts : TCollection ;

                                       public { API... }
                                           procedure Add_Object( O : pointer ;
                                               Facility : integer ;
                                               Context : pointer ) ;
                                               override ;

                                           function Debugger : TDebug_Interface ;
                                               override ;

                                           function Find_Object( O : pointer ;
                                               Facility : integer ) : boolean ;
                                               override ;

                                           function Get_Context( Index : integer ) : pointer ;
                                               override ;

                                           function Get_Facility( Index : integer ) : integer ;
                                               override ;

                                           function Get_Object( Index : integer ) : pointer ;
                                               override ;

                                           procedure Remove_Object( O : pointer ;
                                               Facility : longint ) ;
                                               override ;

                                           function Object_Count : longint ;
                                               override ;

                                           function Construct( Stream : TStream ) : pointer ;
                                               override ;

                                           procedure Register_Facility( Facility_Manager : TFacility_Manager ) ;
                                               override ;

                                           procedure Remove_Facility( Facility_Manager : TFacility_Manager ) ;
                                               override ;

                                           function Get_Facility_Manager( Index : longint ) : TFacility_Manager ;
                                               override ;
                                   end ; { TApplication_Object_Manager }
{$ENDIF}

constructor TSApplication_Object_Manager.Init ;

begin
    inherited Init ;

    _Objects.Init( 0, 4 ) ;
    _Facilities.Init( 0, 4 ) ;
    _Contexts.Init( 0, 4 ) ;
end ;


destructor TSApplication_Object_Manager.Done ;

begin
    _Objects.Done ;
    _Facilities.Done ;
    _Contexts.Done ;

    inherited Done ;
end ;


{ API... }

procedure TSApplication_Object_Manager.Add_Object( O : pointer ;
    Facility : integer ; Context : pointer ) ;

begin
    if( _Objects.IndexOf( O ) <> -1 ) then
    begin
        Application_Manager^.Fatal_Exception( 'Adding existing object or failure to delete old object' ) ;
        exit ;
    end ;
    _Objects.Insert( O ) ;
    _Facilities.Insert( pointer( Facility ) ) ;
    _Contexts.Insert( Context ) ;
end ;


function TSApplication_Object_Manager.Find_Object( O : pointer ;
    Facility : integer ) : boolean ;

var Index : integer ;

begin
    Find_Object := False ;
    Index := _Objects.IndexOf( O ) ;
    if( Index = -1 ) then
    begin
        exit ;
    end ;
    if( ( Facility <> -1 ) and ( Facility <> Get_Facility( Index ) ) ) then
    begin
        Application_Manager^.Fatal_Exception( 'Found object has facility mismatch' ) ;
        exit ;
    end ;
    Find_Object := True ;
end ;


function TSApplication_Object_Manager.Get_Context( Index : integer ) : pointer ;

begin
{$IFDEF Debug}
    if( ( Index < 0 ) or ( Index >= _Objects.Count ) ) then
    begin
        Application_Manager^.Fatal_Exception( 'Facility out of range' ) ;
        Get_Context := nil ;
        exit ;
    end ;
{$ENDIF}
    Get_Context := _Contexts.At( Index ) ;
end ;


function TSApplication_Object_Manager.Get_Facility( Index : integer ) : integer ;

begin
{$IFDEF Debug}
    if( ( Index < 0 ) or ( Index >= _Objects.Count ) ) then
    begin
        Application_Manager^.Fatal_Exception( 'Facility out of range' ) ;
        Get_Facility := -1 ;
        exit ;
    end ;
{$ENDIF}
    Get_Facility := integer( _Facilities.At( Index ) ) ;
end ;


function TSApplication_Object_Manager.Get_Object( Index : integer ) : pointer ;

begin
{$IFDEF Debug}
    if( ( Index < 0 ) or ( Index >= _Objects.Count ) ) then
    begin
        Application_Manager^.Fatal_Exception( 'Facility out of range' ) ;
        Get_Object := nil ;
        exit ;
    end ;
{$ENDIF}
    Get_Object := _Objects.At( Index ) ;
end ;


procedure TSApplication_Object_Manager.Remove_Object( O : pointer ;
    Facility : longint ) ;

var Index : integer ;

begin
    Index := _Objects.IndexOf( O ) ;
    if( Index = -1 ) then
    begin
        Application_Manager^.Fatal_Exception( 'Removing non-existant object' ) ;
        exit ;
    end ;
    if( ( Facility <> -1 ) and ( Get_Facility( Index ) <> Facility ) ) then
    begin
        Application_Manager^.Fatal_Exception( 'Facility mismatch' ) ;
        exit ;
    end ;
    _Objects.AtDelete( Index ) ;
    _Facilities.AtDelete( Index ) ;
    _Contexts.AtDelete( Index ) ;
end ;


function TSApplication_Object_Manager.Object_Count : longint ;

begin
    Object_Count := _Objects.Count ;
end ;


function TSApplication_Object_Manager.Construct( Stream : PSStream ) : pointer ;

begin
    Construct := nil ; {TODO}
end ;


procedure TSApplication_Object_Manager.Register_Facility( Facility_Manager : PSFacility_Manager ) ;

begin
    {TODO}
end ;


procedure TSApplication_Object_Manager.Remove_Facility( Facility_Manager : PSFacility_Manager ) ;

begin
    {TODO}
end ;


function TSApplication_Object_Manager.Get_Facility_Manager( Index : longint ) : PSFacility_Manager ;

begin
    Get_Facility_Manager := nil ; {TODO}
end ;


{$IFDEF DELPHI}
{ TApplication_Object_Manager... }

{ Constructors and destructors... }

constructor TApplication_Object_Manager.Create ;

begin
    inherited Create ;

    _Objects.Init( 0, 4 ) ;
    _Facilities.Init( 0, 4 ) ;
    _Contexts.Init( 0, 4 ) ;
end ;


destructor TApplication_Object_Manager.Destroy ;

begin
    _Objects.Done ;
    _Facilities.Done ;
    _Contexts.Done ;

    inherited Destroy ;
end ;


{ API... }

procedure TApplication_Object_Manager.Add_Object( O : pointer ;
    Facility : integer ; Context : pointer ) ;

begin
    if( _Objects.IndexOf( O ) <> -1 ) then
    begin
        Application_Manager.Fatal_Exception( 'Adding existing object or failure to delete old object' ) ;
        exit ;
    end ;
    _Objects.Insert( O ) ;
    _Facilities.Insert( pointer( Facility ) ) ;
    _Contexts.Insert( Context ) ;
end ;


function TApplication_Object_Manager.Debugger : TDebug_Interface ;

begin
    Debugger := nil ;
end ;


function TApplication_Object_Manager.Find_Object( O : pointer ;
    Facility : integer ) : boolean ;

var Index : integer ;

begin
    Find_Object := False ;
    Index := _Objects.IndexOf( O ) ;
    if( Index = -1 ) then
    begin
        exit ;
    end ;
    if( ( Facility <> -1 ) and ( Facility <> Get_Facility( Index ) ) ) then
    begin
        Application_Manager.Fatal_Exception( 'Found object has facility mismatch' ) ;
        exit ;
    end ;
    Find_Object := True ;
end ;


function TApplication_Object_Manager.Get_Context( Index : integer ) : pointer ;

begin
  {$IFDEF Debug}
    if( ( Index < 0 ) or ( Index >= _Objects.Count ) ) then
    begin
        Application_Manager.Fatal_Exception( 'Facility out of range' ) ;
        Get_Context := nil ;
        exit ;
    end ;
  {$ENDIF}
    Get_Context := _Contexts.At( Index ) ;
end ;


function TApplication_Object_Manager.Get_Facility( Index : integer ) : integer ;

begin
  {$IFDEF Debug}
    if( ( Index < 0 ) or ( Index >= _Objects.Count ) ) then
    begin
        Application_Manager.Fatal_Exception( 'Facility out of range' ) ;
        Get_Facility := -1 ;
        exit ;
    end ;
  {$ENDIF}
    Get_Facility := integer( _Facilities.At( Index ) ) ;
end ;


function TApplication_Object_Manager.Get_Object( Index : integer ) : pointer ;

begin
  {$IFDEF Debug}
    if( ( Index < 0 ) or ( Index >= _Objects.Count ) ) then
    begin
        Application_Manager.Fatal_Exception( 'Facility out of range' ) ;
        Get_Object := nil ;
        exit ;
    end ;
  {$ENDIF}
    Get_Object := _Objects.At( Index ) ;
end ;


procedure TApplication_Object_Manager.Remove_Object( O : pointer ;
    Facility : longint ) ;

var Index : integer ;

begin
    Index := _Objects.IndexOf( O ) ;
    if( Index = -1 ) then
    begin
        Application_Manager.Fatal_Exception( 'Removing non-existant object' ) ;
        exit ;
    end ;
    if( ( Facility <> -1 ) and ( Get_Facility( Index ) <> Facility ) ) then
    begin
        Application_Manager.Fatal_Exception( 'Facility mismatch' ) ;
        exit ;
    end ;
    _Objects.AtDelete( Index ) ;
    _Facilities.AtDelete( Index ) ;
    _Contexts.AtDelete( Index ) ;
end ;


function TApplication_Object_Manager.Object_Count : longint ;

begin
    Object_Count := _Objects.Count ;
end ;


function TApplication_Object_Manager.Construct( Stream : TStream ) : pointer ;

begin
    Construct := nil ; { TODO: Implement }
end ;


procedure TApplication_Object_Manager.Register_Facility( Facility_Manager : TFacility_Manager ) ;

begin
    {TODO: Implement}
end ;


procedure TApplication_Object_Manager.Remove_Facility( Facility_Manager : TFacility_Manager ) ;

begin
    {TODO: Implement}
end ;


function TApplication_Object_Manager.Get_Facility_Manager( Index : longint ) : TFacility_Manager ;

begin
    Get_Facility_Manager := nil ; {TODO: Implement}
end ;



{ TUnits_Debugger methods... }

function TUnits_Debugger.Count : longint ;

begin
    Count := Application_Manager.Debug_Manager.Unit_Debugger_Count ;
end ;


function TUnits_Debugger.Child( Ordinal : longint ) : TDebug_Interface ;

begin
    Child := Delphi_Debug_Wrapper( Application_Manager.Debug_Manager.Get_Unit_Debugger( Ordinal ).Get_Instance ) ;
end ;


{ TStandard_Application_Manager methods... }

{ API... }

function TStandard_Application_Manager.Get_In_Shut_Down : boolean ;

begin
    Get_In_Shut_Down := Application_Manager.Get_In_Shut_Down ;
end ;


procedure TStandard_Application_Manager.Set_In_Shut_Down( Value : boolean ) ;

begin
    Application_Manager.Set_In_Shut_Down( Value ) ;
end ;


function TStandard_Application_Manager.Get_Halting : boolean ;

begin
    Get_Halting := _Halting ;
end ;


procedure TStandard_Application_Manager.Set_Halting( Value : boolean ) ;

begin
    _Halting := Value ;
end ;


function TStandard_Application_Manager.Get_Debug_Manager : TDebug_Manager ;

begin
    Get_Debug_Manager := _Debug_Manager ;
end ;


procedure TStandard_Application_Manager.Set_Debug_Manager( Value : TDebug_Manager ) ;

begin
    if( Value <> nil ) then
    begin
        Value.Attach ;
    end ;
    if( _Debug_Manager <> nil ) then
    begin
        _Debug_Manager.Detach ;
    end ;
    _Debug_Manager := Value ;
end ;


function TStandard_Application_Manager.Application_Debugger : TDebug_Interface ;

{$IFDEF Debug}
var Res : TApplication_Debugger ;
{$ENDIF}

begin
{$IFDEF Debug}
    Res := TApplication_Debugger.Create ;
    Res.Set_Title( PChar( Name + ' ' + Version ) ) ;
    Application_Debugger := Res ;
{$ELSE}
    Application_Debugger := nil ;
{$ENDIF}
end ;


function TStandard_Application_Manager.Debugger : TCOM_Debug_Interface ;

var Res : TCOM_Debug_Interface ;

begin
{$IFDEF Debug}
    Res := TApplication_Manager_Debugger.Create ;
    Res.Set_Title( 'TStandard_Application_Manager' ) ;
{$ELSE}
    Res := nil ; { No application debugger by default }
{$ENDIF}
    Debugger := Res ;
end ;


function TStandard_Application_Manager.Object_Manager : TObject_Manager ;

begin
    if( _Object_Manager = nil ) then
    begin
        _Object_Manager := TApplication_Object_Manager.Create ;
    end ;
    Object_Manager := _Object_Manager ;
end ;


procedure TStandard_Application_Manager.Do_Idle ;

begin
end ;


procedure TStandard_Application_Manager.Add_Idle_Delegation( Value : TIdle_Consumer ) ;

begin
end ;


procedure TStandard_Application_Manager.Remove_Idle_Delegation( Value : TIdle_Consumer ) ;

begin
end ;


function TStandard_Application_Manager.Get_Event_Logger : TApplication_Event_Logger ;

begin
    if( _Event_Logger = nil ) then
    begin
        _Event_Logger := TStandard_Application_Event_Logger.Create ;
        _Event_Logger.Attach ;
    end ;
    Get_Event_Logger := _Event_Logger ;
end ;


procedure TStandard_Application_Manager.Set_Event_Logger( EV : TApplication_Event_Logger ) ;

begin
    if( EV <> nil ) then
    begin
        EV.Attach ;
    end ;
    if( _Event_Logger <> nil ) then
    begin
        _Event_Logger.Detach ;
    end ;
    _Event_Logger := EV ;
end ;


{ Performance... }

function TStandard_Application_Manager.Get_Performance_Monitor : TApplication_Performance_Monitor ;

begin
    if( _Performance_Monitor = nil ) then
    begin
        _Performance_Monitor := TStandard_Application_Performance_Monitor.Create ;
        _Performance_Monitor.Attach ;
    end ;
    Get_Performance_Monitor := _Performance_Monitor ;
end ;


procedure TStandard_Application_Manager.Set_Performance_Monitor( PM : TApplication_Performance_Monitor ) ;

begin
    if( PM <> nil ) then
    begin
        PM.Attach ;
    end ;
    if( _Performance_Monitor <> nil ) then
    begin
        _Performance_Monitor.Detach ;
    end ;
    _Performance_Monitor := PM ;
end ;


{ Other... }

procedure TStandard_Application_Manager.Fatal_Exception( const S : string ) ;

begin
    raise Exception.Create( S ) ;
end ;


procedure TStandard_Application_Manager.Popup_Message( const S : string ) ;

begin
    ShowMessage( S ) ;
end ;


function TStandard_Application_Manager.Get_Heap : TManaged_Store ;

begin
    if( _Heap = nil ) then
    begin
        _Heap := THeap.Create ;
        _Heap.Attach ;
    end ;
    Get_Heap := _Heap ;
end ;


procedure TStandard_Application_Manager.Set_Heap( Value : TManaged_Store ) ;

begin
    if( Value <> nil ) then
    begin
        Value.Attach ;
    end ;
    if( _Heap <> nil ) then
    begin
        _Heap.Detach ;
    end ;
    _Heap := Value ;
end ;


function TStandard_Application_Manager.Get_Name : string ;

begin
    Get_Name := Application_Manager^.Name ;
end ;


procedure TStandard_Application_Manager.Set_Name( const Value : string ) ;

begin
    Application_Manager^.Name := Value ;
end ;


function TStandard_Application_Manager.Get_Serial_Number( Pool : longint ) : longint ;

begin
    Get_Serial_Number := Application_Manager^.Get_Serial_Number( Pool ) ;
end ;


function TStandard_Application_Manager.Get_Clipboard( const Path : string ) : string ;

begin
    Get_Clipboard := Application_Manager^.Get_Clipboard( PChar( Path ) ) ;
end ;


procedure TStandard_Application_Manager.Set_Clipboard( const Path, Value : string ) ;

begin
    Application_Manager^.Set_Clipboard( PChar( Path ), PChar( Value ) ) ;
end ;


function TStandard_Application_Manager.Get_Version : string ;

begin
    Version := Application_Manager^.Version ;
end ;


procedure TStandard_Application_Manager.Set_Version( const Value : string ) ;

begin
    Application_Manager^.Version := Value ;
end ;
{$ENDIF}



{ TSStandard_Application_Manager methods... }

{ Constructors and destructors... }

constructor TSStandard_Application_Manager.Init ;

begin
{$IFNDEF DELPHI}
    inherited Init ; { No inherited Init in Delphi }
{$ENDIF}

    _Event_Logger := nil ;
    _Debug_Manager := nil ;
    _Object_Manager := nil ;
    _In_Shut_Down := False ;
{$IFDEF DELPHI}
    Initialize( _Program_Name ) ;
    Initialize( _Version ) ;
{$ENDIF}
    _Program_Name := '' ;
    _Version := '' ;
end ;


destructor TSStandard_Application_Manager.Done ;

begin
{$IFDEF DELPHI}
    Finalize( _Program_Name ) ;
    Finalize( _Version ) ;
{$ENDIF}

    if( _Event_Logger <> nil ) then
    begin
        dispose( _Event_Logger, Done ) ;
        _Event_Logger := nil ;
    end ;

    if( _Debug_Manager <> nil ) then
    begin
        _Debug_Manager^.Detach ;
        _Debug_Manager := nil ;
    end ;

    if( _Heap <> nil ) then
    begin
        dispose( _Heap, Done ) ;
        _Heap := nil ;
    end ;

    if( _Performance_Monitor <> nil ) then
    begin
        dispose( _Performance_Monitor, Done ) ;
        _Performance_Monitor := nil ;
    end ;

    if( _Object_Manager <> nil ) then
    begin
        dispose( _Object_Manager, Done ) ;
        _Object_Manager := nil ;
    end ;

    inherited Done ;
end ;


{ API... }

function TSStandard_Application_Manager.Get_In_Shut_Down : boolean ;

begin
    Get_In_Shut_Down := _In_Shut_Down ;
end ;


procedure TSStandard_Application_Manager.Set_In_Shut_Down( Value : boolean ) ;

begin
    _In_Shut_Down := Value ;
end ;


function TSStandard_Application_Manager.Get_Halting : boolean ;

begin
    Get_Halting := _Halting ;
end ;


procedure TSStandard_Application_Manager.Set_Halting( Value : boolean ) ;

begin
    _Halting := Value ;
end ;


function TSStandard_Application_Manager.Get_Debug_Manager : PSDebug_Manager ;

begin
    Get_Debug_Manager := _Debug_Manager ;
end ;


procedure TSStandard_Application_Manager.Set_Debug_Manager( Value : PSDebug_Manager ) ;

begin
    if( Value <> nil ) then
    begin
        Value^.Attach ;
    end ;
    if( _Debug_Manager <> nil ) then
    begin
        _Debug_Manager^.Detach ;
    end ;
    _Debug_Manager := Value ;
end ;


function TSStandard_Application_Manager.Application_Debugger : PSDebug_Interface ;

{$IFDEF Debug}
var Res : PSDebug_Interface ;
{$ENDIF}

begin
{$IFDEF Debug}
    Res := new( PSApplication_Debugger, Init ) ;
    Res.Set_Title( PChar( Name + ' ' + Version ) ) ;
    Application_Debugger := Res ;
{$ELSE}
    Application_Debugger := nil ;
{$ENDIF}
end ;


function TSStandard_Application_Manager.Debugger : PSDebug_Interface ;

var Res : PSDebug_Interface ;

begin
{$IFDEF Debug}
    Res := new( PSApplication_Manager_Debugger, Init ) ;
    Res.Set_Title( 'TSStandard_Application_Manager' ) ;
{$ELSE}
    Res := nil ; { No application debugger by default }
{$ENDIF}
    Debugger := Res ;
end ;


function TSStandard_Application_Manager.Object_Manager : PSObject_Manager ;

begin
    if( _Object_Manager = nil ) then
    begin
        _Object_Manager := new( PSApplication_Object_Manager, Init ) ;
    end ;
    Object_Manager := _Object_Manager ;
end ;


procedure TSStandard_Application_Manager.Do_Idle ;

begin
end ;


procedure TSStandard_Application_Manager.Add_Idle_Delegation( Value : PSIdle_Consumer ) ;

begin
end ;


procedure TSStandard_Application_Manager.Remove_Idle_Delegation( Value : PSIdle_Consumer ) ;

begin
end ;


function TSStandard_Application_Manager.Get_Event_Logger : PSApplication_Event_Logger ;

begin
    if( _Event_Logger = nil ) then
    begin
        _Event_Logger := new( PSApplication_Event_Logger, Init ) ;
        _Event_Logger^.Set_Memory_Type( MT_Heap ) ;
        _Event_Logger^.Attach ;
    end ;
    Get_Event_Logger := _Event_Logger ;
end ;


procedure TSStandard_Application_Manager.Set_Event_Logger( EV : PSApplication_Event_Logger ) ;

begin
    if( _Event_Logger <> nil ) then
    begin
        _Event_Logger^.Detach ;
    end ;
    _Event_Logger := EV ;
end ;


{ Performance... }

function TSStandard_Application_Manager.Get_Performance_Monitor : PSApplication_Performance_Monitor ;

begin
    if( _Performance_Monitor = nil ) then
    begin
        _Performance_Monitor := new( PSStandard_Application_Performance_Monitor, Init ) ;
        _Performance_Monitor.Set_Memory_Type( MT_Heap ) ;
        _Performance_Monitor^.Attach ;
    end ;
    Get_Performance_Monitor := _Performance_Monitor ;
end ;


procedure TSStandard_Application_Manager.Set_Performance_Monitor( PM : PSApplication_Performance_Monitor ) ;

begin
    if( _Performance_Monitor <> nil ) then
    begin
        _Performance_Monitor^.Detach ;
    end ;
    _Performance_Monitor := PM ;
end ;


{ Other... }

procedure TSStandard_Application_Manager.Fatal_Exception( const S : string ) ;

begin
{$IFDEF Delphi}
    raise Exception.Create( S ) ;
{$ELSE}
    writeln( S ) ;
    halt ;
{$ENDIF}
end ;


procedure TSStandard_Application_Manager.Popup_Message( S : string ) ;

{$IFNDEF Delphi}
var S1 : string ;
{$ENDIF}

begin
{$IFDEF Delphi}
    ShowMessage( S ) ;
{$ELSE}
    write( S, '... Press ENTER to continue' ) ;
    readln( S1 ) ;
{$ENDIF}
end ;


function TSStandard_Application_Manager.Get_Heap : PSManaged_Store ;

begin
    if( _Heap = nil ) then
    begin
        _Heap := new( PSHeap, Init ) ;
        _Heap^.Attach ;
    end ;
    Get_Heap := _Heap ;
end ;


procedure TSStandard_Application_Manager.Set_Heap( Value : PSManaged_Store ) ;

begin
    if( Value <> nil ) then
    begin
        Value^.Attach ;
    end ;
    if( _Heap <> nil ) then
    begin
        _Heap^.Detach ;
    end ;
    _Heap := Value ;
end ;


function TSStandard_Application_Manager.Get_Name : string ;

begin
    Get_Name := _Program_Name ;
end ;


procedure TSStandard_Application_Manager.Set_Name( const Value : string ) ;

begin
    _Program_Name := Value ;
end ;

                           
var Next_Serial_Number : integer ; { For pool 0 }
    Serial_Number_Pool : PInteger_Collection ;

function TSStandard_Application_Manager.Get_Serial_Number( Pool : longint ) : longint ;

var SN : longint ;

begin
    if( Pool <= 0 ) then { Default pool - needed to prevent infinite loop }
    begin
        Get_Serial_Number := Next_Serial_Number ;
        inc( Next_Serial_Number ) ;
    end else
    begin
        dec( Pool ) ;
        if( Serial_Number_Pool = nil ) then
        begin
            Serial_Number_Pool := new( PInteger_Collection, Init( 0, 4 ) ) ;
        end ;
        while( Pool >= Serial_Number_Pool^.Count ) do
        begin
            Serial_Number_Pool^.Insert( 0 ) ;
        end ;
        SN := Serial_Number_Pool^.At( Pool ) ;
        Get_Serial_Number := SN ;
        Serial_Number_Pool^.AtPut( Pool, SN )
    end ;
end ;


function TSStandard_Application_Manager.Get_Version : string ;

begin
    Get_Version := _Version ;
end ;


procedure TSStandard_Application_Manager.Set_Version( const Value : string ) ;

begin
    _Version := Value ;
end ;


{$IFDEF Delphi}
type TUnit_List_Debugger = class( TText_Debugger )
                               function Count : longint ; override ;

                               function Child( Ordinal : longint ) : TDebug_Interface ;
                                   override ;
                           end ;

function TUnit_List_Debugger.Count : longint ;

begin
    Count := Application_Manager.Debug_Manager.Unit_Debugger_Count ;
end ;


function TUnit_List_Debugger.Child( Ordinal : longint ) : TDebug_Interface ;

begin
    Child := Delphi_Debug_Wrapper( Application_Manager.Debug_Manager.Get_Unit_Debugger( Ordinal ).Get_Instance ) ;
end ;



{ TApplication_Debugger methods... }

constructor TApplication_Debugger.Create ;

begin
    inherited Create ;

    _OS := Delphi_Debug_Wrapper( OS.Debugger ) ;
    if( _OS <> nil ) then
    begin
        _OS.Set_Title( PChar( OS.Name ) ) ;
    end ;
    _AM := Delphi_Debug_Wrapper( Application_Manager.Debugger ) ;
    if( _AM <> nil ) then
    begin
        _AM.Set_Title( 'Application' ) ;
    end ;
    _VC := VCL_Debugger ;
    if( _VC <> nil ) then
    begin
        _VC.Set_Title( 'VCL' ) ;
    end ;
end ;


function TApplication_Debugger.Count : longint ;

var C : integer ;

begin
    C := 0 ;
    if( _AM <> nil ) then
    begin
        _Debuggers[ C ] := _AM ;
        inc( C ) ;
    end ;
    if( _OS <> nil ) then
    begin
        _Debuggers[ C ] := _OS ;
        inc( C ) ;
    end ;
    if( _VC <> nil ) then
    begin
        _Debuggers[ C ] := _VC ;
        inc( C ) ;
    end ;
    Count := C ;
end ;


function TApplication_Debugger.Child( Ordinal : longint ) : TDebug_Interface ;

begin
    Child := _Debuggers[ Ordinal ] ;
end ;


{ TApplication_Manager_Debugger methods... }

function TApplication_Manager_Debugger.Count : longint ;

begin
    Count := 0;{4 ;}
end ;


function TApplication_Manager_Debugger.Child( Ordinal : longint ) : TCOM_Debug_Interface ;

var Res : TCOM_Debug_Interface ;

begin
    Res := nil ;
(*
    case Ordinal of
        0 : begin
                Res := Create_Text_Debugger ;
                if( Application_Manager.Debug_Manager = nil ) then
                begin
                    Res.Set_Title( 'Debug_Manager = nil' ) ;
                end else
                begin
                    Res.Set_Title( 'Debug_Manager' ) ;
                end ;
            end ;
        1 : begin
                if( Application_Manager.Event_Logger = nil ) then
                begin
                    Res := Create_Text_Debugger ;
                    Res.Set_Title( '_Event_Logger = nil' ) ;
                end else
                begin
                    Res := Application_Manager.Event_Logger.Debugger ;
                    if( Res <> nil ) then
                    begin
                        Res.Set_Title( '_Event_Logger' ) ;
                    end else
                    begin
                        Res := Create_Text_Debugger ;
                        Res.Set_Title( '_Event_Logger = nil' ) ;
                    end ;
                end ;
            end ;
        2 : begin
                Res := Create_Text_Debugger ;
                Res.Set_Title( PChar( '_In_Shut_Down = ' + Boolean_To_String( Application_Manager._In_Shut_Down ) ) ) ;
            end ;
        3 : begin
                if(
                    ( Application_Manager.Debug_Manager <> nil )
                    and
                    ( Application_Manager.Debug_Manager.Unit_Debugger_Count > 0 )
                  ) then
                begin
{$IFDEF Delphi}
                    Res := TUnits_Debugger.Create ;
{$ELSE}
                    Res := new( PUnits_Debugger, Init ) ;
{$ENDIF}
                end else
                begin
                    Res := Create_Text_Debugger ;
                end ;
                Res.Set_Title( 'Units' ) ;
            end ;
    end ;
*)
    Child := Res ;
end ; { TApplication_Manager_Debugger.Child }
{$ENDIF}


{ TSApplication_Manager_Debugger methods... }

function TSApplication_Manager_Debugger.Count : longint ;

begin
    Count := 6 ;
end ;


function TSApplication_Manager_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

var Res : PSDebug_Interface ;
    S : string[ 255 ] ;

begin
    Res := nil ;
    case Ordinal of
        0 : begin
                Res := Create_Text_Debugger ;
                if( Application_Manager^.Get_Debug_Manager = nil ) then
                begin
                    Res^.Set_Title( 'Debug_Manager = nil' ) ;
                end else
                begin
                    Res^.Set_Title( 'Debug_Manager' ) ;
                end ;
            end ;
        1 : begin
                if( Application_Manager^.Get_Event_Logger = nil ) then
                begin
                    Res := Create_Text_Debugger ;
                    Res^.Set_Title( '_Event_Logger = nil' ) ;
                end else
                begin
                    Res := Application_Manager^.Get_Event_Logger^.Debugger ;
                    if( Res <> nil ) then
                    begin
                        Res^.Set_Title( '_Event_Logger' ) ;
                    end else
                    begin
                        Res := Create_Text_Debugger ;
                        Res^.Set_Title( '_Event_Logger = nil' ) ;
                    end ;
                end ;
            end ;
        2 : begin
                Res := Create_Text_Debugger ;
                S := '_In_Shut_Down = ' +
		    Boolean_To_String( PSStandard_Application_Manager( Application_Manager )^._In_Shut_Down ) ;
                Res^.Set_Title( PChar( @S[ 1 ] ) ) ;
            end ;
        3 : begin
                if(
                    ( Application_Manager^.Get_Debug_Manager <> nil )
                    and
                    ( Application_Manager^.Get_Debug_Manager^.Unit_Debugger_Count > 0 )
                  ) then
                begin
                    Res := new( PSUnits_Debugger, Init ) ;
                end else
                begin
                    Res := Create_Text_Debugger ;
                end ;
                Res^.Set_Title( 'Units' ) ;
            end ;
        4 : begin
                Res := Create_Text_Debugger ;
                S := '_Name = ' + PSStandard_Application_Manager( Application_Manager )^._Program_Name ;
                Res^.Set_Title( @S[ 1 ] ) ;
            end ;
        5 : begin
                Res := Create_Text_Debugger ;
                S := '_Version = ' + PSStandard_Application_Manager( Application_Manager )^._Version ;
                Res^.Set_Title( @S[ 1 ] ) ;
            end ;
    end ;
    Child := Res ;
end ; { TSApplication_Manager_Debugger.Child }



{ TSApplication_Debugger methods... }

constructor TSApplication_Debugger.Init ;

var S : string[ 255 ] ;

begin
    inherited Init ;

    _OS := OS^.Debugger ;
    if( _OS <> nil ) then
    begin
        S := OS^.Name ;
        _OS^.Set_Title( PChar( @S[ 1 ] ) ) ;
    end ;
    _AM := Application_Manager^.Debugger ;
    if( _AM <> nil ) then
    begin
        _AM^.Set_Title( 'Application' ) ;
    end ;
{$IFDEF Delphi}
    _VC := Static_Debug_Wrapper( VCL_Debugger ) ;
    if( _VC <> nil ) then
    begin
        _VC^.Set_Title( 'VCL' ) ;
    end ;
{$ELSE}
    _VC := nil ;
{$ENDIF}
end ; { TSApplication_Debugger.Init }


function TSApplication_Debugger.Count : longint ;

var C : integer ;

begin
    C := 0 ;
    if( _AM <> nil ) then
    begin
        _Debuggers[ C ] := _AM ;
        inc( C ) ;
    end ;
    if( _OS <> nil ) then
    begin
        _Debuggers[ C ] := _OS ;
        inc( C ) ;
    end ;
    if( _VC <> nil ) then
    begin
        _Debuggers[ C ] := _VC ;
        inc( C ) ;
    end ;
    Count := C ;
end ;


function TSApplication_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

begin
    Child := _Debuggers[ Ordinal ] ;
end ;


{ TSUnits_Debugger methods... }

function TSUnits_Debugger.Count : longint ;

begin
    Count := Application_Manager^.Get_Debug_Manager^.Unit_Debugger_Count ;
end ;


function TSUnits_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

begin
    Child := Application_Manager^.Get_Debug_Manager^.Get_Unit_Debugger( Ordinal )^.Get_Instance ;
end ;



{$IFDEF DELPHI}
type TSApplication_Manager_COM_Wrapper = object( TSStandard_Application_Manager )
                                            public { Constructors and destructors... }
                                                constructor Init ;

                                            private
                                                COM : TCOM_Application_Manager ;

                                                Cached_Object_Manager : PSObject_Manager ;
                                                Cached_Debug_Manager : PSDebug_Manager ;
                                                Cached_Event_Logger : PSApplication_Event_LOgger ;
                                                Temp_Get_Name : string ;
                                                Temp_Get_Version : string ;

                                            public { API... }
                                                { Status... }
                                                function Get_In_Shut_Down : boolean ;
                                                    virtual ;

                                                procedure Set_In_Shut_Down( Value : boolean ) ;
                                                    virtual ;

                                                function Get_Halting : boolean ;
                                                    virtual ;

                                                procedure Set_Halting( Value : boolean ) ;
                                                    virtual ;

                                                { Debugging... }
                                                function Get_Debug_Manager : PSDebug_Manager ;
                                                    virtual ;

                                                function Object_Manager : PSObject_Manager ;
                                                    virtual ;

                                                { Idle processing... }
                                                procedure Do_Idle ; virtual ;

                                                { Other... }
                                                procedure Fatal_Exception( const S : string ) ;
                                                    virtual ;

                                                procedure Popup_Message( S : string ) ;
                                                    virtual ;

                                                { Event logging... }
                                                function Get_Event_Logger : PSApplication_Event_Logger ;
                                                    virtual ;

                                                procedure Set_Event_Logger( EV : PSApplication_Event_Logger ) ;
                                                    virtual ;

                                                function Debugger : PSDebug_Interface ;
                                                    virtual ;

                                                procedure Set_Debug_Manager( Value : PSDebug_Manager ) ;
                                                    virtual ;

                                                function Application_Debugger : PSDebug_Interface ;
                                                    virtual ;

                                                procedure Add_Idle_Delegation( Value : PSIdle_Consumer ) ;
                                                    virtual ;

                                                procedure Remove_Idle_Delegation( Value : PSIdle_Consumer ) ;
                                                    virtual ;

                                                function Get_Heap : PSManaged_Store ;
                                                    virtual ;

                                                procedure Set_Heap( Value : PSManaged_Store ) ;
                                                    virtual ;

                                                function Get_Name : string ;
                                                    virtual ;

                                                procedure Set_Name( const Value : string ) ;
                                                    virtual ;

                                                function Get_Serial_Number( Pool : longint ) : longint ;
                                                    virtual ;

                                                { Full version string including "V" }
                                                function Get_Version : string ;
                                                    virtual ;

                                                procedure Set_Version( const Value : string ) ;
                                                    virtual ;
                                         end ; { TSApplication_Manager_COM_Wrapper }
     PSApplication_Manager_COM_Wrapper = ^TSApplication_Manager_COM_Wrapper ;

{ Constructors and destructors... }

constructor TSApplication_Manager_COM_Wrapper.Init ;

begin
    inherited Init ;

    Cached_Object_Manager := nil ;
    Cached_Debug_Manager := nil ;
end ;


{ API... }

function TSApplication_Manager_COM_Wrapper.Get_In_Shut_Down : boolean ;

begin
    Get_In_Shut_Down := COM.In_Shut_Down ;
end ;


procedure TSApplication_Manager_COM_Wrapper.Set_In_Shut_Down( Value : boolean ) ;

begin
    COM.In_Shut_Down := Value ;
end ;


function TSApplication_Manager_COM_Wrapper.Get_Halting : boolean ;

begin
    Get_Halting := COM.Halting ;
end ;


procedure TSApplication_Manager_COM_Wrapper.Set_Halting( Value : boolean ) ;

begin
    COM.Halting := Value ;
end ;


function TSApplication_Manager_COM_Wrapper.Get_Debug_Manager : PSDebug_Manager ;

begin
    if( Cached_Debug_Manager = nil ) then
    begin
        Cached_Debug_Manager :=
            Static_Debug_Manager_Wrapper( COM.Get_Debug_Manager ) ;
    end ;
    Get_Debug_Manager := Cached_Debug_Manager ;
end ;


function TSApplication_Manager_COM_Wrapper.Object_Manager : PSObject_Manager ;

begin
    if( Cached_Object_Manager = nil ) then
    begin
        Cached_Object_Manager := Static_Object_Manager_Wrapper( COM.Object_Manager ) ;
    end ;
    Object_Manager := Cached_Object_Manager ;
end ;


procedure TSApplication_Manager_COM_Wrapper.Do_Idle ;

begin
    COM.Do_Idle ;
end ;


function TSApplication_Manager_COM_Wrapper.Get_Event_Logger : PSApplication_Event_Logger ;

var CEL : TCOM_Application_Event_Logger ;

begin
    if( Cached_Event_Logger = nil ) then
    begin
        CEL := COM.Get_Event_Logger ;
        if( CEL = nil ) then
        begin
            Cached_Event_Logger := nil ;
        end else
        begin
            Cached_Event_Logger :=
                Static_COM_Event_Logger_Wrapper( CEL ) ;
        end ;
    end ;
    Result := Cached_Event_Logger ;
end ;


procedure TSApplication_Manager_COM_Wrapper.Set_Event_Logger( EV : PSApplication_Event_Logger ) ;

begin
end ;


function TSApplication_Manager_COM_Wrapper.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ;
end ;


procedure TSApplication_Manager_COM_Wrapper.Fatal_Exception( const S : string ) ;

begin
    COM.Fatal_Exception( PChar( S ) ) ;
end ;


procedure TSApplication_Manager_COM_Wrapper.Popup_Message( S : string ) ;

begin
    COM.Popup_Message( PChar( S ) ) ;
end ;


procedure TSApplication_Manager_COM_Wrapper.Set_Debug_Manager( Value : PSDebug_Manager ) ;

begin
end ;


function TSApplication_Manager_COM_Wrapper.Application_Debugger : PSDebug_Interface ;

begin
    Application_Debugger := nil ;
end ;


procedure TSApplication_Manager_COM_Wrapper.Add_Idle_Delegation( Value : PSIdle_Consumer ) ;

begin
end ;


procedure TSApplication_Manager_COM_Wrapper.Remove_Idle_Delegation( Value : PSIdle_Consumer ) ;

begin
end ;


function TSApplication_Manager_COM_Wrapper.Get_Heap : PSManaged_Store ;

begin
    Result := nil ;
end ;


procedure TSApplication_Manager_COM_Wrapper.Set_Heap( Value : PSManaged_Store ) ;

begin
end ;


function TSApplication_Manager_COM_Wrapper.Get_Name : string ;

begin
    Temp_Get_Name := COM.Name ;
    Get_Name := Temp_Get_Name ;
end ;


procedure TSApplication_Manager_COM_Wrapper.Set_Name( const Value : string ) ;

begin
    COM.Name := PChar( Value ) ;
end ;


function TSApplication_Manager_COM_Wrapper.Get_Serial_Number( Pool : longint ) : longint ;

begin
    Get_Serial_Number := COM.Get_Serial_Number( Pool ) ;
end ;


function TSApplication_Manager_COM_Wrapper.Get_Version : string ;

begin
    Temp_Get_Version := COM.Version ;
    Get_Version := Temp_Get_Version ;
end ;


procedure TSApplication_Manager_COM_Wrapper.Set_Version( const Value : string ) ;

begin
    COM.Version := PChar( Value ) ;
end ;



function Static_Application_Manager_COM_Wrapper( COM : TCOM_Application_Manager ;
    Old : PSApplication_Manager ) : PSApplication_Manager ;

begin
    Result := new( PSApplication_Manager_COM_Wrapper, Init ) ;
    PSApplication_Manager_COM_Wrapper( Result ).COM := COM ;
    if( Old <> nil ) then
    begin
        Old.Free ;
    end ;
end ;


type TStatic_Debug_Manager_Wrapper = object( TSDebug_Manager )
                                         private
                                             COM : TCOM_Debug_Manager ;

                                         public
                                             function Get_Unit_Debugger( Index : longint ) : PSUnit_Debugger ;
                                                 virtual ;

                                             { Register a unit debugger }
                                             procedure Register_Unit_Debugger( D : PSUnit_Debugger ) ;
                                                 virtual ;

                                             { Remove a registered unit debugger }
                                             procedure Remove_Unit_Debugger( D : PSUnit_Debugger ) ;
                                                 virtual ;

                                             { Returns count of registered unit debuggers }
                                             function Unit_Debugger_Count : longint ;
                                                 virtual ;
                                     end ;
     PStatic_Debug_Manager_Wrapper = ^TStatic_Debug_Manager_Wrapper ;

function TStatic_Debug_Manager_Wrapper.Get_Unit_Debugger( Index : longint ) : PSUnit_Debugger ;

begin
    Result := nil ;
end ;


procedure TStatic_Debug_Manager_Wrapper.Register_Unit_Debugger( D : PSUnit_Debugger ) ;

begin
    COM.Register_Unit_Debugger( COM_Unit_Debugger_Wrapper( D ) ) ;
end ;


procedure TStatic_Debug_Manager_Wrapper.Remove_Unit_Debugger( D : PSUnit_Debugger ) ;

begin
end ;


function TStatic_Debug_Manager_Wrapper.Unit_Debugger_Count : longint ;

begin
    Result := COM.Unit_Debugger_Count ;
end ;


function Static_Debug_Manager_Wrapper( Value : TCOM_Debug_Manager ) : PSDebug_Manager ;

begin
    Result := new( PStatic_Debug_Manager_Wrapper, Init ) ;
    PStatic_Debug_Manager_Wrapper( Result )^.COM := Value ;
end ;


type TStatic_Object_Manager_Wrapper = object( TSObject_Manager )
                                          private
                                              COM : TCOM_Object_Manager ;

                                          public
                                            procedure Add_Object( O : pointer ;
                                                Facility : integer ;
                                                Context : pointer ) ; virtual ;

                                            function Debugger : PSDebug_Interface ;
                                                virtual ;

                                            function Find_Object( O : pointer ;
                                                Facility : integer ) : boolean ;
                                                virtual ;

                                            function Get_Context( Index : integer ) : pointer ;
                                                virtual ;

                                            function Get_Facility( Index : integer ) : integer ;
                                                virtual ;

                                            function Get_Object( Index : integer ) : pointer ;
                                                virtual ;

                                            procedure Remove_Object( O : pointer ;
                                                Facility : longint ) ; virtual ;

                                            function Object_Count : longint ;
                                                virtual ;

                                            procedure Register_Facility( Facility_Manager : PSFacility_Manager ) ;
                                                virtual ;

                                            procedure Remove_Facility( Facility_Manager : PSFacility_Manager ) ;
                                                virtual ;

                                            function Get_Facility_Manager( Index : longint ) : PSFacility_Manager ;
                                                virtual ;

                                            function Construct( Stream : PSStream ) : pointer ;
                                                virtual ;
                                      end ; { TStatic_Object_Manager_Wrapper }
     PStatic_Object_Manager_Wrapper = ^TStatic_Object_Manager_Wrapper ;

procedure TStatic_Object_Manager_Wrapper.Add_Object( O : pointer ;
    Facility : integer ; Context : pointer ) ;

begin
    COM.Add_Object( O, Facility, Context ) ;
end ;


function TStatic_Object_Manager_Wrapper.Debugger : PSDebug_Interface ;

begin
    Result := nil ;
end ;


function TStatic_Object_Manager_Wrapper.Find_Object( O : pointer ;
    Facility : integer ) : boolean ;

begin
    Result := COM.Find_Object( O, Facility ) ;
end ;


function TStatic_Object_Manager_Wrapper.Get_Context( Index : integer ) : pointer ;

begin
    Result := COM.Get_Context( Index ) ;
end ;


function TStatic_Object_Manager_Wrapper.Get_Facility( Index : integer ) : integer ;

begin
    Result := COM.Get_Facility( Index ) ;
end ;


function TStatic_Object_Manager_Wrapper.Get_Object( Index : integer ) : pointer ;

begin
    Result := COM.Get_Object( Index ) ;
end ;


procedure TStatic_Object_Manager_Wrapper.Remove_Object( O : pointer ;
    Facility : longint ) ;

begin
    COM.Remove_Object( O, Facility ) ;
end ;


function TStatic_Object_Manager_Wrapper.Object_Count : longint ;

begin
    Result := COM.Object_Count ;
end ;


procedure TStatic_Object_Manager_Wrapper.Register_Facility( Facility_Manager : PSFacility_Manager ) ;

begin
end ;


procedure TStatic_Object_Manager_Wrapper.Remove_Facility( Facility_Manager : PSFacility_Manager ) ;

begin
end ;


function TStatic_Object_Manager_Wrapper.Get_Facility_Manager( Index : longint ) : PSFacility_Manager ;

begin
    Result := nil ;
end ;


function TStatic_Object_Manager_Wrapper.Construct( Stream : PSStream ) : pointer ;

begin
    Result := nil ;
end ;


function Static_Object_Manager_Wrapper( Value : TCOM_Object_Manager ) : PSObject_Manager ;

begin
    Result := new( PStatic_Object_Manager_Wrapper, Init ) ;
    PStatic_Object_Manager_Wrapper( Result )^.COM := Value ;
end ;



type TCOM_Unit_Debugger_Wrapper = class( TCOM_Unit_Debugger )
                                      private
                                          UD : PSUnit_Debugger ;
                                          Temp_Get_Name : string ;

                                      public
                                          function Get_Name : PChar ;
                                              override ;

                                          function Get_Instance : TDebug_Interface ;
                                              override ;
                                  end ;

function TCOM_Unit_Debugger_Wrapper.Get_Name : PChar ;

begin
    Temp_Get_Name := UD.Get_Name ;
    Result := PChar( Temp_Get_Name ) ;
end ;


function TCOM_Unit_Debugger_Wrapper.Get_Instance : TDebug_Interface ;

begin
    Result := Delphi_Debug_Wrapper( UD.Get_Instance ) ;
end ;



function COM_Unit_Debugger_Wrapper( Value : PSUnit_Debugger ) : TCOM_Unit_Debugger ;

begin
    Result := TCOM_Unit_Debugger_Wrapper.Create ;
    TCOM_Unit_Debugger_Wrapper( Result ).UD := Value ;
end ;
{$ENDIF}


{$IFDEF Test}
type Test_Unit_Debugger = object( TSUnit_Debugger )
                              public
                                  Last_Routine : string ;

                                  function Get_Name : string ;
                                      virtual ;

                                  function Get_Instance : PSDebug_Interface ;
                                      virtual ;
                          end ;
     PTest_Unit_Debugger = ^Test_Unit_Debugger ;

function Test_Unit_Debugger.Get_Name : string ;

begin
    Last_Routine := 'Get_Name' ;
    Get_Name := '' ;
end ;


function Test_Unit_Debugger.Get_Instance : PSDebug_Interface ;

begin
    Last_Routine := 'Get_Instance' ;
    Get_Instance := nil ;
end ;


procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


procedure Test_Unit_Debugger_Wrapper ;

var S : string ;
    Test_Static : PTest_Unit_Debugger ;
    Test : TCOM_Unit_Debugger ;

begin
    Test_Static := new( PTest_Unit_Debugger, Init ) ;
    Test := COM_Unit_Debugger_Wrapper( Test_Static ) ;

    S := Test.Get_Name ;
    if( S <> '' ) then
    begin
        Error ;
    end ;
    if( Test_Static.Last_Routine <> 'Get_Name' ) then
    begin
        Error ;
    end ;
    if( Test.Get_Instance <> nil ) then
    begin
        Error ;
    end ;
    if( Test_Static.Last_Routine <> 'Get_Instance' ) then
    begin
        Error ;
    end ;

    Test.Free ;
    dispose( Test_Static, Done ) ;
end ;


type Test_COM_Object_Manager = class( TCOM_Object_Manager )
                                   public
                                       Last_Routine : string ;

                                       procedure Add_Object( O : pointer ;
                                           Facility : integer ;
                                           Context : pointer ) ;
                                           override ;

                                       function Find_Object( O : pointer ;
                                           Facility : integer ) : boolean ;
                                           override ;

                                       function Get_Context( Index : integer ) : pointer ;
                                           override ;

                                       function Get_Facility( Index : integer ) : integer ;
                                           override ;

                                       function Get_Object( Index : integer ) : pointer ;
                                           override ;

                                       procedure Remove_Object( O : pointer ;
                                           Facility : longint ) ;
                                           override ;

                                       function Object_Count : longint ;
                                           override ;

                                       function Construct( Stream : TCOM_Stream ) : pointer ;
                                           override ;

                                       procedure Register_Facility( Facility_Manager : TCOM_Facility_Manager ) ;
                                           override ;

                                       procedure Remove_Facility( Facility_Manager : TCOM_Facility_Manager ) ;
                                           override ;

                                       function Get_Facility_Manager( Index : longint ) : TCOM_Facility_Manager ;
                                           override ;
                               end ;

procedure Test_COM_Object_Manager.Add_Object( O : pointer ; Facility : integer ; Context : pointer ) ;

begin
    Last_Routine := 'Add_Object' ;
end ;


function Test_COM_Object_Manager.Find_Object( O : pointer ;
    Facility : integer ) : boolean ;

begin
    Last_Routine := 'Find_Object' ;
    Find_Object := False ;
end ;


function Test_COM_Object_Manager.Get_Context( Index : integer ) : pointer ;

begin
    Last_Routine := 'Get_Context' ;
    Get_Context := nil ;
end ;


function Test_COM_Object_Manager.Get_Facility( Index : integer ) : integer ;

begin
    Last_Routine := 'Get_Facility' ;
    Get_Facility := 0 ;
end ;


function Test_COM_Object_Manager.Get_Object( Index : integer ) : pointer ;

begin
    Last_Routine := 'Get_Object' ;
    Get_Object := nil ;
end ;


procedure Test_COM_Object_Manager.Remove_Object( O : pointer ;
    Facility : longint ) ;

begin
    Last_Routine := 'Remove_Object' ;
end ;


function Test_COM_Object_Manager.Object_Count : longint ;

begin
    Last_Routine := 'Object_Count' ;
    Object_Count := 0 ;
end ;


function Test_COM_Object_Manager.Construct( Stream : TCOM_Stream ) : pointer ;

begin
    Last_Routine := 'Construct' ;
    Construct := nil ;
end ;


procedure Test_COM_Object_Manager.Register_Facility( Facility_Manager : TCOM_Facility_Manager ) ;

begin
    Last_Routine := 'Register_Facility' ;
end ;


procedure Test_COM_Object_Manager.Remove_Facility( Facility_Manager : TCOM_Facility_Manager ) ;

begin
    Last_Routine := 'Remove_Facility' ;
end ;


function Test_COM_Object_Manager.Get_Facility_Manager( Index : longint ) : TCOM_Facility_Manager ;

begin
    Last_Routine := 'Get_Facility_Manager' ;
    Get_Facility_Manager := nil ;
end ;



procedure Test_Object_Manager_Wrapper ;

var Test_Delphi : Test_COM_Object_Manager ;
    Test : PSObject_Manager ;

begin
    Test_Delphi := Test_COM_Object_Manager.Create ;
    Test := Static_Object_Manager_Wrapper( Test_Delphi ) ;

    Test.Add_Object( nil, -1, nil ) ;
    if( Test_Delphi.Last_Routine <> 'Add_Object' ) then
    begin
        Error ;
    end ;
    if( Test.Find_Object( nil, -1 ) ) then
    begin
        Error ;
    end ;
    if( Test_Delphi.Last_Routine <> 'Find_Object' ) then
    begin
        Error ;
    end ;
    if( Test.Get_Context( -1 ) <> nil ) then
    begin
        Error ;
    end ;
    if( Test_Delphi.Last_Routine <> 'Get_Context' ) then
    begin
        Error ;
    end ;
    if( Test.Get_Facility( -1 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_Delphi.Last_Routine <> 'Get_Facility' ) then
    begin
        Error ;
    end ;
    if( Test.Get_Object( -1 ) <> nil ) then
    begin
        Error ;
    end ;
    if( Test_Delphi.Last_Routine <> 'Get_Object' ) then
    begin
        Error ;
    end ;
    Test.Remove_Object( nil, -1 ) ;
    if( Test_Delphi.Last_Routine <> 'Remove_Object' ) then
    begin
        Error ;
    end ;
    if( Test.Object_Count <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_Delphi.Last_Routine <> 'Object_Count' ) then
    begin
        Error ;
    end ;
    if( Test.Construct( nil ) <> nil ) then
    begin
        Error ;
    end ;
    Test.Register_Facility( nil ) ;
    Test.Remove_Facility( nil ) ;
    if( Test.Get_Facility_Manager( -1 ) <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Debugger <> nil ) then
    begin
        Error ;
    end ;

    Test_Delphi.Create ;
    dispose( Test, Done ) ;
end ;



type Test_COM_Application_Manager = class( TCOM_Application_Manager )
                                        public
                                            Last_Routine : string ;
                                            Fake_Event_Logger : boolean ;

                                            function Get_In_Shut_Down : boolean ;
                                                override ;

                                            procedure Set_In_Shut_Down( Value : boolean ) ;
                                                override ;

                                            function Get_Halting : boolean ;
                                                override ;

                                            procedure Set_Halting( Value : boolean ) ;
                                                override ;

{
                                            function Get_State_Manager : TCOM_Application_State_Manager ;
                                                override ;

                                            procedure Set_State_Manager( Value : TCOM_Application_State_Manager ) ;
                                                override ;
}

                                            function Get_Serial_Number( Pool : longint ) : longint ;
                                                override ;

                                            function Get_Debug_Manager : TCOM_Debug_Manager ;
                                                override ;

                                            procedure Set_Debug_Manager( Value : TCOM_Debug_Manager ) ;
                                                override ;

                                            function Application_Debugger : PSDebug_Interface ;
                                                override ;

                                            function Object_Manager : TCOM_Object_Manager ;
                                                override ;

                                            procedure Do_Idle ;
                                                override ;

                                            procedure Add_Idle_Delegation( Value : TIdle_Consumer ) ;
                                                override ;

                                            procedure Remove_Idle_Delegation( Value : TIdle_Consumer ) ;
                                                override ;

                                            function Get_Event_Logger : TCOM_Application_Event_Logger ;
                                                override ;

                                            procedure Set_Event_Logger( EV : TCOM_Application_Event_Logger ) ;
                                                override ;

                                            procedure Fatal_Exception( S : PChar ) ;
                                                override ;

                                            procedure Popup_Message( S : PChar ) ;
                                                override ;

                                            function Get_Heap : TCOM_Managed_Store ;
                                                override ;

                                            procedure Set_Heap( Value : TCOM_Managed_Store ) ;
                                                override ;

                                            function Get_Name : PChar ;
                                                override ;

                                            procedure Set_Name( Value : Pchar ) ;
                                                override ;

                                            function Get_Version : PChar ;
                                                override ;

                                            procedure Set_Version( Value : PChar ) ;
                                                override ;

                                            function Get_Clipboard( Path : PChar ;
                                                var Len : longint ) : PChar ;
                                                override ;

                                            procedure Set_Clipboard( Path, Value : PChar ;
                                                var Len : longint ) ;
                                                override ;

                                            { Event logging... }
                                            function Get_Performance_Monitor : TCOM_Application_Performance_Monitor ;
                                                override ;

                                            procedure Set_Performance_Monitor( EV : TCOM_Application_Performance_Monitor ) ;
                                                override ;
                                    end ; { Test_COM_Application_Manager }

function Test_COM_Application_Manager.Get_In_Shut_Down : boolean ;

begin
    Last_Routine := 'Get_In_Shut_Down' ;
    Result := False ;
end ;


procedure Test_COM_Application_Manager.Set_In_Shut_Down( Value : boolean ) ;

begin
    Last_Routine := 'Set_In_Shut_Down' ;
end ;


function Test_COM_Application_Manager.Get_Halting : boolean ;

begin
    Last_Routine := 'Get_Halting' ;
    Result := False ;
end ;


procedure Test_COM_Application_Manager.Set_Halting( Value : boolean ) ;

begin
    Last_Routine := 'Set_Halting' ;
end ;


{
function Test_COM_Application_Manager.Get_State_Manager : TCOM_Application_State_Manager ;

begin
    Last_Routine := 'Get_State_Manager' ;
    Result := nil ;
end ;


procedure Test_COM_Application_Manager.Set_State_Manager( Value : TCOM_Application_State_Manager ) ;

begin
    Last_Routine := 'Set_State_Manager' ;
end ;
}


function Test_COM_Application_Manager.Get_Debug_Manager : TCOM_Debug_Manager ;

begin
    Last_Routine := 'Get_Debug_Manager' ;
    Result := nil ;
end ;


procedure Test_COM_Application_Manager.Set_Debug_Manager( Value : TCOM_Debug_Manager ) ;

begin
    Last_Routine := 'Set_Debug_Manager' ;
end ;


function Test_COM_Application_Manager.Application_Debugger : PSDebug_Interface ;

begin
    Last_Routine := 'Application_Debugger' ;
    Result := nil ;
end ;


function Test_COM_Application_Manager.Object_Manager : TCOM_Object_Manager ;

begin
    Last_Routine := 'Object_Manager' ;
    Result := nil ;
end ;


procedure Test_COM_Application_Manager.Do_Idle ;

begin
    Last_Routine := 'Do_Idle' ;
end ;


procedure Test_COM_Application_Manager.Add_Idle_Delegation( Value : TIdle_Consumer ) ;

begin
    Last_Routine := 'Add_Idle_Delegation' ;
end ;


procedure Test_COM_Application_Manager.Remove_Idle_Delegation( Value : TIdle_Consumer ) ;

begin
    Last_Routine := 'Remove_Idle_Delegation' ;
end ;


function Test_COM_Application_Manager.Get_Event_Logger : TCOM_Application_Event_Logger ;

begin
    Last_Routine := 'Get_Event_Logger' ;
    if( Fake_Event_Logger ) then
    begin
        Result := pointer( $123456 ) ;
    end else
    begin
        Result := nil ;
    end ;
end ;


procedure Test_COM_Application_Manager.Set_Event_Logger( EV : TCOM_Application_Event_Logger ) ;

begin
    Last_Routine := 'Set_Event_Logger' ;
end ;


procedure Test_COM_Application_Manager.Fatal_Exception( S : PChar ) ;

begin
    Last_Routine := 'Fatal_Exception' ;
end ;


procedure Test_COM_Application_Manager.Popup_Message( S : PChar ) ;

begin
    Last_Routine := 'Popup_Message' ;
end ;


function Test_COM_Application_Manager.Get_Heap : TCOM_Managed_Store ;

begin
    Last_Routine := 'Get_Heap' ;
    Result := nil ;
end ;


procedure Test_COM_Application_Manager.Set_Heap( Value : TCOM_Managed_Store ) ;

begin
    Last_Routine := 'Set_Heap' ;
end ;


function Test_COM_Application_Manager.Get_Name : PChar ;

begin
    Last_Routine := 'Get_Name' ;
    Result := nil ;
end ;


procedure Test_COM_Application_Manager.Set_Name( Value : Pchar ) ;

begin
    Last_Routine := 'Set_Name' ;
end ;


function Test_COM_Application_Manager.Get_Serial_Number( Pool : longint ) : longint ;

begin
    Last_Routine := 'Get_Serial_Number' ;
    Result := 0 ;
end ;


function Test_COM_Application_Manager.Get_Version : PChar ;

begin
    Last_Routine := 'Get_Version' ;
    Result := nil ;
end ;


procedure Test_COM_Application_Manager.Set_Version( Value : PChar ) ;

begin
    Last_Routine := 'Set_Version' ;
end ;


function Test_COM_Application_Manager.Get_Clipboard( Path : PChar ;
    var Len : longint ) : PChar ;

begin
    Last_Routine := 'Get_Clipboard' ;
    Result := nil ;
end ;


procedure Test_COM_Application_Manager.Set_Clipboard( Path, Value : PChar ;
    var Len : longint ) ;

begin
    Last_Routine := 'Set_Clipboard' ;
end ;


function Test_COM_Application_Manager.Get_Performance_Monitor : TCOM_Application_Performance_Monitor ;

begin
    Last_Routine := 'Get_Performance_Monitor' ;
    Result := nil ;
end ;


procedure Test_COM_Application_Manager.Set_Performance_Monitor( EV : TCOM_Application_Performance_Monitor ) ;

begin
    Last_Routine := 'Set_Performance_Monitor' ;
end ;


procedure Test_Static_Application_Manager_COM_Wrapper ;

var Test_COM : Test_COM_Application_Manager ;
    Test : PSApplication_Manager ;

begin
    Test_COM := Test_COM_Application_Manager.Create ;
    Test := Static_Application_Manager_COM_Wrapper( Test_COM, nil ) ;

    if( Test.In_Shut_Down ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Get_In_Shut_Down' ) then
    begin
        Error ;
    end ;
    Test.In_Shut_Down := True ;
    if( Test_COM.Last_Routine <> 'Set_In_Shut_Down' ) then
    begin
        Error ;
    end ;
    if( Test.Halting ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Get_Halting' ) then
    begin
        Error ;
    end ;
    Test.Halting := True ;
    if( Test_COM.Last_Routine <> 'Set_Halting' ) then
    begin
        Error ;
    end ;
{
    if( Test.State_Manager = nil ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Get_State_Manager' ) then
    begin
        Error ;
    end ;
}
    if( Test.Debug_Manager = nil ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Get_Debug_Manager' ) then
    begin
        Error ;
    end ;
    if( Test.Debug_Manager = nil ) then
    begin
        Error ;
    end ;
    Test.Debug_Manager := nil ;
    if( Test.Application_Debugger <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Debugger <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Object_Manager = nil ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Object_Manager' ) then
    begin
        Error ;
    end ;
    if( Test.Object_Manager = nil ) then
    begin
        Error ;
    end ;
    Test.Do_Idle ;
    if( Test_COM.Last_Routine <> 'Do_Idle' ) then
    begin
        Error ;
    end ;
    Test.Add_Idle_Delegation( nil ) ;
    Test.Remove_Idle_Delegation( nil ) ;
    if( Test.Event_Logger <> nil ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Get_Event_Logger' ) then
    begin
        Error ;
    end ;
    Test_COM.Fake_Event_Logger := True ;
    if( Test.Event_Logger = nil ) then
    begin
        Error ;
    end ;
    if( Test.Event_Logger = nil ) then
    begin
        Error ;
    end ;
    Test.Event_Logger := nil ;
    Test.Fatal_Exception( '' ) ;
    if( Test_COM.Last_Routine <> 'Fatal_Exception' ) then
    begin
        Error ;
    end ;
    Test.Popup_Message( '' ) ;
    if( Test_COM.Last_Routine <> 'Popup_Message' ) then
    begin
        Error ;
    end ;
    if( Test.Heap <> nil ) then
    begin
        Error ;
    end ;
    Test.Heap := nil ;
    if( Test.Name <> '' ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Get_Name' ) then
    begin
        Error ;
    end ;
    Test.Name := '' ;
    if( Test_COM.Last_Routine <> 'Set_Name' ) then
    begin
        Error ;
    end ;
    if( Test.Version <> '' ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Get_Version' ) then
    begin
        Error ;
    end ;
    Test.Version := '' ;
    if( Test_COM.Last_Routine <> 'Set_Version' ) then
    begin
        Error ;
    end ;

    Test_COM.Free ;
{$IFNDEF Delphi}
    Test :=
{$ENDIF}
        Static_Application_Manager_COM_Wrapper( nil, Test ) ;
end ;


type Test_COM_Debug_Manager = class( TCOM_Debug_Manager )
                                  public
                                      Last_Routine : string ;

                                      function Get_Unit_Debugger( Index : longint ) : TCOM_Unit_Debugger ;
                                          override ;

                                      procedure Register_Unit_Debugger( D : TCOM_Unit_Debugger ) ;
                                          override ;

                                      procedure Remove_Unit_Debugger( D : TCOM_Unit_Debugger ) ;
                                          override ;

                                      function Unit_Debugger_Count : longint ;
                                          override ;
                              end ;

function Test_COM_Debug_Manager.Get_Unit_Debugger( Index : longint ) : TCOM_Unit_Debugger ;

begin
    Last_Routine := 'Get_Unit_Debugger' ;
    Result := nil ;
end ;


procedure Test_COM_Debug_Manager.Register_Unit_Debugger( D : TCOM_Unit_Debugger ) ;

begin
    Last_Routine := 'Register_Unit_Debugger' ;
end ;


procedure Test_COM_Debug_Manager.Remove_Unit_Debugger( D : TCOM_Unit_Debugger ) ;

begin
    Last_Routine := 'Remove_Unit_Debugger' ;
end ;


function Test_COM_Debug_Manager.Unit_Debugger_Count : longint ;

begin
    Last_Routine := 'Unit_Debugger_Count' ;
    Result := 0 ;
end ;


procedure Test_Static_Debug_Manager_Wrapper ;

var Test_COM : Test_COM_Debug_Manager ;
    Test : PSDebug_Manager ;

begin
    Test_COM := Test_COM_Debug_Manager.Create ;
    Test := Static_Debug_Manager_Wrapper( Test_COM ) ;

    if( Test.Get_Unit_Debugger( 0 ) <> nil ) then
    begin
        Error ;
    end ;
    Test.Register_Unit_Debugger( nil ) ;
    if( Test_COM.Last_Routine <> 'Register_Unit_Debugger' ) then
    begin
        Error ;
    end ;
    Test.Remove_Unit_Debugger( nil ) ;
    if( Test.Unit_Debugger_Count <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Unit_Debugger_Count' ) then
    begin
        Error ;
    end ;

    dispose( Test, Done ) ;
    Test_COM.Free ;
end ;


procedure Test_Unit ;

begin
    Test_Unit_Debugger_Wrapper ;
    Test_Object_Manager_Wrapper ;
    Test_Static_Application_Manager_COM_Wrapper ;
    Test_Static_Debug_Manager_Wrapper ;
end ;
{$ENDIF}


procedure Terminate ;

begin
    if( Application_Manager <> nil ) then
    begin
        dispose( Application_Manager, Done ) ;
        Application_Manager := nil ;
    end ;
{$IFDEF DELPHI}
    if( Delphi_Application_Manager <> nil ) then
    begin
        Delphi_Application_Manager.Free ;
        Delphi_Application_Manager := nil ;
    end ;
{$ENDIF}
end ;


{$IFDEF Delphi}
initialization
{$ELSE}
begin
{$ENDIF}
    Application_Manager := new( PSStandard_Application_Manager, Init ) ;
{$IFDEF Delphi}
    Delphi_Application_Manager := TStandard_Application_Manager.Create ;
    Next_Serial_Number := 0 ;
    Serial_Number_Pool := nil ;

finalization
    if( Application_Manager <> nil ) then
    begin
        Application_Manager^.Halting := True ;
    end ;
    if( Delphi_Application_Manager <> nil ) then
    begin
        Delphi_Application_Manager.Halting := True ;
    end ;
  {$IFDEF CC_Auto_Free_Application_Manager}
        Terminate ;
  {$ENDIF}
{$ENDIF}
end.
