{
        Program Name : ArrayDis
        Package Name : LB
        Purpose      : Display array contents in a window
        Institution  : Conroy & Conroy Co.
        Date Written : 27-March-2001
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2001 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit contains the TArray_Display object which is used to
        display the contents of an array in a Delphi form.
}

unit ArrayDis ;

interface

uses // Borland...
     Windows, // HDC

     Controls, // TWinCOntrol
     Graphics, // TColor
     
     // C&C...
     ArrayInt ; // TArray_Interface

type TArray_Display = class
                          private // Instance data...
                              _Control : TWinControl ;
                              __Array : TCOM_Array_Interface ;
                              _Home_Index : int64 ; // Top-left index
                              _Font : integer ; // Handle of font to use
                              _Foreground, _Background : TColor ;
                              _Foreground_RGB, _Background_RGB : TColor ;
                              _Number_Size : integer ;
                              _Unsigned : boolean ;
                              _Ascii_Ebcdic : Integer ; {0=ASCII, 1=EBCDIC}
                              _Base : integer ;
                              _Max_Data : integer ; // Bytes per line
                              _LPP : integer ; // Lines per page
                              Serial_Number : integer ;

                          public // Constructors and destructors...
                              constructor Create( C : TWinControl ;
                                  AI : TCOM_Array_Interface ; Font : integer ;
                                  Background, Foreground : TColor ) ;

                          protected // Property handlers...
                              function Get_Number_Size : integer ;
                              procedure Set_Array( Value : TCOM_Array_Interface ) ;
                              procedure Set_Number_Size( Value : integer ) ;

                          public // API...
                              property _Array : TCOM_Array_Interface
                                           read __Array
                                           write Set_Array ;

                              property Ascii_Ebcdic : integer
                                  read _Ascii_Ebcdic
                                  write _Ascii_Ebcdic ;

                              property Base : integer
                                  read _Base
                                  write _Base ;
                                  
                              procedure Display ;

                              property Home_Index : int64
                                  read _Home_Index
                                  write _Home_Index ;

                              property LPP : integer
                                  read _LPP ;

                              property Max_Data : integer
                                  read _Max_Data
                                  write _Max_Data ;

                              property Number_Size : integer
                                  read Get_Number_Size
                                  write Set_Number_Size ;

                              property Unsigned : boolean
                                  read _Unsigned
                                  write _Unsigned ;
                      end ;

implementation

uses // C&C...
     CVT, // CVTB
     Misc, // Rad
     Num1s, // Num1
     Radix50s, // Rad
     Standard ; // Get_Serial_Number

// Unit variables...

var EBCDIC : array[ 0..255 ] of Char ;



// TArray_Display methods...

// Constructors and destructors...

constructor TArray_Display.Create( C : TWinControl ; AI : TCOM_Array_Interface ;
    Font : integer ; Background, Foreground : TColor ) ;

var B : TBitmap ;
    Brush : TBrush ;
    Rect : TRect ;

begin
    inherited Create ;
    _Control := C ;
    __Array := AI ;
    if( _Array <> nil ) then
    begin
        _Home_Index := _Array.Low_Bound( 0 ) ;
    end ;
    _Font := Font ;
    _Foreground := Foreground ;
    _Background := Background ;
    B := TBitmap.Create ;
    Brush := TBrush.Create ;
    B.PixelFormat := pf32bit ;
    Rect.Left := 0 ;
    Rect.Top := 0 ;
    Rect.Right := 1 ;
    Rect.Bottom := 1 ;
    Brush.Color := _Foreground ;
    fillrect( B.Canvas.Handle, Rect, Brush.Handle ) ;
    _Foreground_RGB := B.Canvas.Pixels[ 0, 0 ] ;
    Brush.Color := _Background ;
    fillrect( B.Canvas.Handle, Rect, Brush.Handle ) ;
    _Background_RGB := B.Canvas.Pixels[ 0, 0 ] ;
    B.Free ;
    Brush.Free ;
    Base := 10 ;
    Unsigned := True ;
    Serial_Number := Get_Serial_Number ;
end ; // TArray_Display.Create


// Property handlers...

function TArray_Display.Get_Number_Size : integer ;

begin
    Result := _Number_Size + 1 ;
end ;


procedure TArray_Display.Set_Array( Value : TCOM_Array_Interface ) ;

begin
    __Array := Value ;
    if( _Array <> nil ) then
    begin
        _Home_Index := _Array.Low_Bound( 0 ) ;
    end ;
end ;


procedure TArray_Display.Set_Number_Size( Value : integer ) ;

begin
    _Number_Size := Value - 1 ;
end ;


// API...

procedure TArray_Display.Display ;

var A_A, B_B, D_D : string ;
    C_C : string ; // String for writing out to a file
    A, A1, B, C1, C2, C3, D, Siz : Integer ;
    Address : string ;
    Address_Size : integer ;
    Brush : TBrush ;
    C : smallint ;
    ch : char ;
    CR : double ;
    DC : HDC ;
    Eeol : string[ 8 ] ;
    Font_Handle : integer ;
    LogFont : TTextMetric ;
    Rev : Boolean ; { True if showing reverse video }
    Font_Height, Font_Width : integer ;
    Text_String : string ;
    X, Y : integer ;

    procedure Write( const S : string ) ;

    var Rect : TRect ;

    begin
        Rect.Left := X ;
        Rect.Right := X + length( S ) * Font_Width ;
        Rect.Top := Y ;
        Rect.Bottom := Y + Font_Height ;
        FillRect( DC, Rect, Brush.Handle ) ;
        TextOut( DC, X, Y, PChar( S ), length( S ) ) ;
        X := Rect.Right ;
    end ;


    procedure Reverse ;

    begin
        // If showing on white or black background, use the highest contrast
        if( _Foreground_RGB = clBlack ) then
        begin
            SetTextColor( DC, clWhite ) ;
        end else
        begin
            if( _Foreground_RGB = clWhite ) then
            begin
                SetTextColor( DC, clBlack ) ;
            end else
            begin
                SetTextColor( DC, _Background_RGB ) ; // ...otherwise just switch foreground/background
            end ;
        end ;
        Brush.Color := _Foreground ;
    end ;


    procedure Normal ;

    begin
        SetTextColor( DC, _Foreground_RGB ) ;
        Brush.Color := _Background ;
    end ;


begin // TArray_Display.Display
    if( _Array = nil ) then
    begin
        DC := GetDC( _Control.Handle ) ;
        ReleaseDC( _Control.Handle, DC ) ;
        exit ;
    end ;

    // Setup...
    DC := GetDC( _Control.Handle ) ;
    Font_Handle := SelectObject( DC, _Font ) ;
    GetTextMetrics( DC, LogFont ) ;
    Font_Height := LogFont.tmHeight ; ;
    Font_Width := LogFont.tmMaxCharWidth ;
    Brush := TBrush.Create ;
    Brush.Color := _Background ;
    SetTextColor( DC, _Foreground ) ;
    SetBkMode( DC, TRANSPARENT ) ;
    Y := 0 ;
    Rev := False ;
    Eeol := '' ;
    case _Number_Size of
        3 : Siz := 19 ;
        1 : Siz := 9 ;
        else Siz := 4 ;
    end ;
    if( Base > 15 ) then
    begin
        Siz := Siz div 2 ;
    end ;

    if( Base = 50 ) then
    begin
        Address_Size := length( num1( _Array.High_Bound( 0 ) ) ) ;
    end else
    begin
        Address_Size := length( CVTB( 10, Base, num1( _Array.High_Bound( 0 ) ) ) ) ;
    end ;
(*
    if( Length( Out_File_Name ) > 0 then for A := 1 to 256 do
    begin
        Write( Out_File, Chr( _Array.Bytes[ _Home_Index + A ] ) ) ;
    end ;
*)

    // Now draw the data...
    D := 0 ;
    _LPP := ( _Control.ClientHeight div Font_Height ) ;
    for A := 0 to _LPP do { For each row }
    begin
        // Early out when we run out of data...
        if( _Home_Index + D > _Array.High_Bound( 0 ) ) then
        begin
            break ;
        end ;

        // Setup for row...
        X := 0 ;
        Text_String := '' ;
        A_A := '' ;
        C_C := '' ;

        // Determine number of bytes on line
        B := _Control.ClientWidth - ( Address_Size + 2 ) * Font_Width ; // Line width minus address
        B := B div Font_Width div ( Siz + _Number_Size + 2 ) ; // Six digits, 1 space, and _Number_Size + 1 characters per value
        if( B < 1 ) then
        begin
            B := 1 ; // Always at least one value per line
        end ;
        setlength( Text_String, B * ( _Number_Size + 1 ) ) ;

        if( Base <> 50 ) then
        begin
            // Do address...
            Address := CVTB( 10, Base, Num1( _Home_Index + D ) ) ;
            while( length( Address ) < Address_Size ) do
            begin
                Address := '0' + Address ;
            end ;
            Write( Address + ': ' ) ;

            // Now do values...
            B := 1 ;
            while( B <= length( Text_String ) ) do { For each byte in the row }
            begin
                C := _Array.Bytes[ _Home_Index + D ] ; { Get current byte }
                Text_String[ B ] := char( C ) ; { Build up ASCII portion }
                if( _Number_Size = 1 ) then { Word size }
                begin
                    C := C or Swap( ord( _Array.Bytes[ _Home_Index + D + 1 ] ) ) ; {Add in high byte}
                    Text_String[ B + 1 ] := char( _Array.Bytes[ _Home_Index + D + 1 ] ) ;
                end ;
                if( _Number_Size = 3 ) then { Longword size }
                begin
                    Text_String[ B + 1 ] := char( _Array.Bytes[ _Home_Index + D + 1 ] ) ;
                    Text_String[ B + 2 ] := char( _Array.Bytes[ _Home_Index + D + 2 ] ) ;
                    Text_String[ B + 3 ] := char( _Array.Bytes[ _Home_Index + D + 3 ] ) ;
                end ;
{$WARNINGS OFF}
                if( _Number_Size = 0 ) then if( not Unsigned ) then if( ( C and 128 ) = 128 )
                    then C := C or Swap( 255 ) ; { Sign extend if( signed byte }
{$WARNINGS ON}
                Cr := C ; {Convert to FPF}
                if( _Number_Size = 3 ) then
                    Cr := _Array.Bytes[ _Home_Index + D ] + _Array.Bytes[ _Home_Index + D + 1 ] * 256.0 +
                        _Array.Bytes[ _Home_Index + D + 2 ] * 65536.0 +
                        _Array.Bytes[ _Home_Index + D + 3 ] * 16777216.0 ;
                if( Unsigned or ( Base <> 10 ) ) then {Unsigned value in base 10}
                    if( C < 0 ) then {if( negative value}
                        if( _Number_Size = 1 ) then
                            Cr := ( C and 32767 ) + 32768.0 else
                            Cr := ( C and 127 ) + 128.0 ; {Convert to unsigned value}
                if( _Number_Size = 3 ) then if( not Unsigned ) then
                    if( Cr > 2147483647.0 ) then Cr := Cr - 4294967296.0 ;
                Str( Cr : Siz : 0, B_B ) ; {Convert from real to string}
                for A1 := 1 to Siz do if( B_B[ A1 ] = ' ' ) then B_B[ A1 ] := '0' ;
                {left-pad with zeroes}
                A1 := Pos( '-', B_B ) ; {Find location of minus}
                if( A1 > 0 ) then
                begin
                    B_B[ A1 ] := '0' ;
                    B_B[ 1 ] := '-' ;
                end else
                begin
                    B_B[ 1 ] := ' ' ;
                end ;
                {if a minus sign, then put it in the proper position}
                if( Base = 10 ) then
                begin
                    C_C := C_C + B_B ;
                end else { Not base 10 }
                begin
                    D_D := Cvtb( 10, Base, Num1( Cr ) ) ;
                    B_B := Copy( D_D, 1, Siz ) ;
                    while( Length( B_B ) < Siz ) do B_B := '0' + B_B ;
                    while( Length( D_D ) < Siz ) do D_D := '0' + D_D ;
                    C_C := C_C + D_D ;
                end ;
                B_B := B_B + ' ' ;
                Write( B_B ) ;
                C_C := C_C + ' ' ;
                B := B + _Number_Size + 1 ;
                D := D + _Number_Size + 1 ;
            end ; { while( B <= length( Text_String ) ) }
        end else { Radix-50 }
        begin
            B_B := '' ;
            C_C := '' ;
            B := 1 ;
            for C3 := 0 to 7 do // For each word in row
            begin
                C2 := 0 ;
                C := _Array.Bytes[ _Home_Index + D ] ;
                C1 := _Array.Bytes[ _Home_Index + D + 1 ] ;
                if( D < 255 ) then
                begin
                    C2 := _Array.Bytes[ _Home_Index + D + 2 ] ;
                end ;
                B_B := B_B + Rad( Swap( C1 ) + C ) ;
                C_C := C_C + Rad( Swap( C2 ) + C1 ) ;
                Text_String[ B ] := char( C ) ;
                Text_String[ B + 1 ] := char( C1 ) ;
                D := D + 2 ;
                B := B + 2 ;
            end ;
            C_C := B_B + '        ' + C_C + '        ' ;
            Write( C_C ) ;
        end ; { if( Base <> 50 }

        { Now process text area... }
        for B := 1 to length( Text_String ) do
        begin
            C := ord( Text_String[ B ] ) and 127 ;
            if( ASCII_EBCDIC = 0 ) then
            begin
                if( ( C < 32 ) or ( C = 127 ) ) then
                begin
                    if( not Rev ) then
                    begin
                        Reverse ;
                        Rev := True ;
                    end ;
                    Ch := chr( C + 64 ) ;
                    Write( Ch ) ;
                    C_C := C_C + '.' ;
                end else
                begin
                    if( Rev ) then
                    begin
                        Rev := False ;
                        Normal ;
                    end ;
                    Ch := chr( C ) ;
                    Write( Ch ) ;
                    C_C := C_C + Chr( C ) ;
                end
            end else
            begin
                Ch := Ebcdic[ C ] ;
                Write( Ch ) ;
                X := X + Font_Width ;
                C_C := C_C + Ebcdic[ C ] ;
            end
        end ;
        if( Rev ) then
        begin
            Rev := False ;
            Normal ;
        end ;
//        if( Print_Display ) then Writeln( Display_File, C_C, ' +', D - 17 )
        Y := Y + Font_Height ;
    end ; { for A }
    SelectObject( DC, Font_Handle ) ;
    ReleaseDC( _Control.Handle, DC ) ;
    Brush.Free ;
    Max_Data := D - length( Text_String ) ;
    if( Max_Data = 0 ) then
    begin
        Max_Data := 1 ;
    end ;
end ; // TArray_Display.Display


procedure Unit_Init ;

var A : integer ;
    In_Line : string ;

begin
    { Build EBCDIC array }
    fillchar( EBCDIC, sizeof( EBCDIC ), 32 ) ;
    In_Line := '[.<(+!&         ]$*);^-\         ,%_>?         :#@' + Chr( 39 ) +
        '=" ABCDEFGHI       JKLMNOPQR       STUVWXYZ                       ' +
        'ABCDEFGHI       JKLMNOPQR     \  STUVWXYZ      0123456789' ;
    for A := 1 to 174 do EBCDIC[ A + 73 ] := In_Line[ A ] ;
end ;


initialization
    Unit_Init ;
end.
