{$N+}
{$E+}
{
        Program Name : TArray_Interface
        Package Name : LB
        Purpose      : Abstract virtual array interface class
        Institution  : Conroy & Conroy Co.
        Date Written : 27-March-2001
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2001 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This class is used to pass array information.
}

unit ArrayInt ;

interface

{$I EDEFINES.INC}

uses { Layer 0... }
     COMInter, { TCommon_COM_Interface }
     Common, { TBase_Delphi_Object }
     Typedefs ; { int64 }

type TSArray_Interface = object( TBase_Object )
                             public { API... }
                                 { TArray_Interface version 0 interface: }

                                 { Return a byte from the given index. }
                                 function Get_Byte( Index : int64 ) : byte ;
                                     virtual ;

                                 { Set a byte at a specific index. }
                                 procedure Set_Byte( Index : int64 ;
                                     Value : byte ) ; virtual ;

                                 { Returns true if the data is read-only. }
                                 function Read_Only : boolean ; virtual ;

                                 { Returns low bound of specified subscript. }
                                 function Low_Bound( Subscript : integer ) : int64 ;
                                     virtual ;

                                 { Returns high bound of specified subscript. }
                                 function High_Bound( Subscript : integer ) : int64 ;
                                     virtual ;

                                 { Returns number of subscripts. }
                                 function Subscripts : integer ; virtual ;

                                 { Version 1 interface: }
                                 { Nothing yet defined }

{$IFDEF DELPHI}
                             public { Properties }
                                 property Bytes[ Index : int64 ] : byte
                                              read Get_Byte
                                              write Set_Byte ;
{$ENDIF}
                         end ;

{$IFDEF DELPHI}
type TArray_Interface = class( TBase_Delphi_Object )
                            public { API... }
                                { TArray_Interface version 0 interface: }

                                { Return a byte from the given index. }
                                function Get_Byte( Index : int64 ) : byte ;
                                    virtual ; abstract ;

                                { Set a byte at a specific index. }
                                procedure Set_Byte( Index : int64 ;
                                    Value : byte ) ; virtual ; abstract ;

                                { Returns true if the data is read-only. }
                                function Read_Only : boolean ;
                                    virtual ; abstract ;

                                { Returns low bound of specified subscript. }
                                function Low_Bound( Subscript : integer ) : int64 ;
                                    virtual ; abstract ;

                                { Returns high bound of specified subscript. }
                                function High_Bound( Subscript : integer ) : int64 ;
                                    virtual ; abstract ;

                                { Returns number of subscripts. }
                                function Subscripts : integer ;
                                    virtual ; abstract ;

                                { Version 1 interface: }
                                { Nothing yet defined }

                            public { Properties }
                                property Bytes[ Index : int64 ] : byte
                                             read Get_Byte
                                             write Set_Byte ;
                        end ;

type TCOM_Array_Interface = class( TCommon_COM_Interface )
                                public { API... }
                                    { TArray_Interface version 0 interface: }

                                    { Return a byte from the given index. }
                                    function Get_Byte( Index : int64 ) : byte ;
                                        virtual ; stdcall ; abstract ;

                                    { Set a byte at a specific index. }
                                    procedure Set_Byte( Index : int64 ;
                                        Value : byte ) ;
                                        virtual ; stdcall ; abstract ;

                                    { Returns true if the data is read-only. }
                                    function Read_Only : boolean ;
                                        virtual ; stdcall ; abstract ;

                                    { Returns low bound of specified subscript. }
                                    function Low_Bound( Subscript : integer ) : int64 ;
                                        virtual ; stdcall ; abstract ;

                                    { Returns high bound of specified subscript. }
                                    function High_Bound( Subscript : integer ) : int64 ;
                                        virtual ; stdcall ; abstract ;

                                    { Returns number of subscripts. }
                                    function Subscripts : integer ;
                                        virtual ; stdcall ; abstract ;

                                    { Version 1 interface: }
                                    { Nothing yet defined }

                                public { Properties }
                                    property Bytes[ Index : int64 ] : byte
                                                 read Get_Byte
                                                 write Set_Byte ;
                            end ;
{$ENDIF}

implementation

uses { Layer 0... }
     _Common ; { Abstract_Error }

{ TSArray_Interface methods... }

{ API... }

function TSArray_Interface.Get_Byte( Index : int64 ) : byte ;

begin
    Get_Byte := 0 ;
    Abstract_Error ;
end ;


procedure TSArray_Interface.Set_Byte( Index : int64 ; Value : byte ) ;

begin
    Abstract_Error ;
end ;


function TSArray_Interface.Read_Only : boolean ;

begin
    Read_Only := True ;
    Abstract_Error ;
end ;


function TSArray_Interface.Low_Bound( Subscript : integer ) : int64 ;

begin
    Low_Bound := Zero64 ;
    Abstract_Error ;
end ;


function TSArray_Interface.High_Bound( Subscript : integer ) : int64 ;

begin
    High_Bound := Zero64 ;
    Abstract_Error ;
end ;


function TSArray_Interface.Subscripts : integer ;

begin
    Subscripts := 0 ;
    Abstract_Error ;
end ;



end.
