{
        Program Name : Array_Display_Main
        Package Name : LB
        Purpose      : Array display GUI
        Institution  : Conroy & Conroy Co.
        Date Written : 24-Jan-2002
        Written By   : Alan Conroy
        Version      : 1.0A

        Copyright (C) 2002 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  13-Nov-2002   EAC         Fix compilation errors.        *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          Displays arrays on a Windows control.
}

unit Array_Display_Main ;

interface

uses // Borland...
     Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
     ExtCtrls, StdCtrls,
     Menus,

     // C&C...
     ArrayInt, // TArray_Interface
     ArrayDis,
     Paint_Panel ; // TPaint_Panel

type TPChar_Array = class( TCOM_Array_Interface )
                            private // Instance data...
                                _P : PChar ;
                                _Length : integer ;
                                _RO : boolean ;

                            public // Constructors and destructors...
                                constructor Create( P : PChar ; Len : integer ;
                                    RO : boolean ) ;

                            public // API...
                                function Get_Byte( Index : int64 ) : byte ;
                                    override ;
                                procedure Set_Byte( Index : int64 ;
                                    Value : byte ) ;
                                    override ;

                                function Read_Only : boolean ;
                                    override ;

                                function Low_Bound( Subscript : integer ) : int64 ;
                                    override ;
                                function High_Bound( Subscript : integer ) : int64 ;
                                    override ;
                                function Subscripts : integer ; override ;
                    end ;

     TArray_Display_Form = class(TForm)
                MainMenu1: TMainMenu;
                Size1: TMenuItem;
                N11: TMenuItem;
                N21: TMenuItem;
                N41: TMenuItem;
                N81: TMenuItem;
                Base1: TMenuItem;
                Binary1: TMenuItem;
                Octal1: TMenuItem;
                Hexadecimal1: TMenuItem;
                Hexadecimal2: TMenuItem;
                Radix501: TMenuItem;
                procedure FormCreate(Sender: TObject);
                procedure FormDestroy(Sender: TObject);
                procedure FormResize(Sender: TObject);
                procedure FormPaint(Sender: TObject);
                procedure FormKeyPress(Sender: TObject; var Key: Char);
                procedure Size_Menu_Click(Sender: TObject);
                procedure Binary1Click(Sender: TObject);
                procedure Octal1Click(Sender: TObject);
                procedure Hexadecimal1Click(Sender: TObject);
                procedure Hexadecimal2Click(Sender: TObject);
                procedure Radix501Click(Sender: TObject);

              private // Instance data...
                _Panel: TPaint_Panel ;
                _Display : TArray_Display ;
                _SB : TScrollbar ;
                SB_Page_Size : integer ;
                _Caption : string ;
                __Array : TCOM_Array_Interface ;

              protected
                  procedure Set_Array( Value : TCOM_Array_Interface ) ;

              protected
                  procedure CB_Scroll( Sender : TObject ;
                      ScrollCode : TScrollCode ; var ScrollPos : Integer ) ;
                  procedure CB_Paint( Sender : TObject ) ;

              public
                  Serial_Number : longint ;

                  property _Array : TCOM_Array_Interface
                               read __Array
                               write Set_Array ;
                  property Display : TArray_Display
                               read _Display ;
                  procedure Set_Caption( Value : string ) ;
          end ; // TArray_Display_Form

implementation

uses { C&C... }
     Standard ; // Get_Serial_Number

{$R *.DFM}

// TPChar_Array methods...

// Constructors and destructors...

constructor TPChar_Array.Create( P : PChar ; Len : integer ; RO : boolean ) ;

begin
    inherited Create ;

    _P := P ;
    _Length := Len ;
    _RO := RO ;
end ;


// API...

function TPChar_Array.Get_Byte( Index : int64 ) : byte ;

begin
    if( ( Index < 0 ) or ( Index >= _Length ) ) then
    begin
        Result := 0 ;
    end else
    begin
        try
            Result := ord( _P[ Index ] ) ;
        except
            Result := 0 ;
        end ;
    end ;
end ;


procedure TPChar_Array.Set_Byte( Index : int64 ; Value : byte ) ;

begin
    if( ( Index >= 0 ) and ( Index < _Length ) ) then
    begin
        try
            _P[ Index ] := char( Value ) ;
        except
        end ;
    end ;
end ;


function TPChar_Array.Read_Only : boolean ;

begin
    Result := _RO ;
end ;


function TPChar_Array.Low_Bound( Subscript : integer ) : int64 ;

begin
    Result := 0 ;
end ;


function TPChar_Array.High_Bound( Subscript : integer ) : int64 ;

begin
    Result := _Length - 1 ; ;
end ;


function TPChar_Array.Subscripts : integer ;

begin
    Result := 1 ;
end ;


// TArray_Display_Form methods...

procedure TArray_Display_Form.FormCreate( Sender : TObject ) ;

begin
    _SB := TScrollbar.Create( self ) ;
    _SB.Parent := self ;
    _SB.Align := alRight ;
    _SB.Kind := sbVertical ;
    _SB.OnScroll := CB_Scroll ;
    _SB.Width := GetSystemMetrics( SM_CXVSCROLL ) ;
    _Panel := TPaint_Panel.Create( self ) ;
    _Panel.Parent := self ;
    _Panel.Align := alClient ;
    _Panel.Font.Name := 'Courier New' ;
    _Panel.On_Paint := CB_Paint ;
    __Array := TPChar_Array.Create( nil, 0, False ) ;
    _Display := TArray_Display.Create( _Panel, nil, _Panel.Font.Handle,
        _Panel.Color, _Panel.Font.Color ) ;
    Serial_Number := Get_Serial_Number ;
end ;


procedure TArray_Display_Form.FormDestroy( Sender : TObject ) ;

begin
    _Array.Detach ;
    _Display.Free ;
end ;


procedure TArray_Display_Form.FormResize( Sender : TObject ) ;

var M : integer ;

begin
    if( ( _Display <> nil ) and ( _Display.Max_Data > 0 ) ) then
    begin
        SB_Page_Size := _Display.LPP ; // Lines per page
        M := trunc( ( _Array.High_Bound( 0 ) / _Display.Max_Data ) * SB_Page_Size ) - SB_Page_Size ; // Total lines (Total pages * lines per page)
        if( M < 0 ) then
        begin
            M := 0 ;
        end ;
        _SB.Max := M ;
        _SB.LargeChange := SB_Page_Size ;
    end ;
end ;


procedure TArray_Display_Form.Hexadecimal1Click(Sender: TObject);

begin
    _Display.Base := 10 ;
    _Panel.Repaint ;
end ;


procedure TArray_Display_Form.Hexadecimal2Click(Sender: TObject);

begin
    _Display.Base := 16 ;
    _Panel.Repaint ;
end ;


procedure TArray_Display_Form.Octal1Click(Sender: TObject);

begin
    _Display.Base := 8 ;
    _Panel.Repaint ;
end ;


procedure TArray_Display_Form.Radix501Click(Sender: TObject);

begin
    _Display.Base := 50 ;
    _Panel.Repaint ;
end ;


procedure TArray_Display_Form.Size_Menu_Click( Sender : TObject ) ;

begin
    _Display.Number_Size := TMenuItem( Sender ).Tag ;
    _Panel.RePaint ;
end ;


procedure TArray_Display_Form.Set_Array( Value : TCOM_Array_Interface ) ;

begin
    __Array := Value ;
    _Display._Array := Value ;
end ;


// Utility routines...

procedure TArray_Display_Form.Binary1Click(Sender: TObject);

begin
    _Display.Base := 2 ;
    _Panel.Repaint ;
end ;


procedure TArray_Display_Form.CB_Paint( Sender : TObject ) ;

begin
    inherited ;

    _Display.Display ;
end ;



procedure TArray_Display_Form.CB_Scroll( Sender : TObject ; ScrollCode : TScrollCode ;
    var ScrollPos : Integer ) ;

begin
    if( SB_Page_Size = 0 ) then
    begin
        FormResize( nil ) ;
    end ;
    _Display.Home_Index := ScrollPos * _Display.Max_Data div SB_Page_Size ;
    _Panel.Repaint ;
end ;


procedure TArray_Display_Form.FormPaint( Sender : TObject ) ;

begin
    if( self = Application.MainForm ) then
    begin
        Visible := False ; // Hides us on task bar (also need to leave main form alone so DLL doesn't shut down)
    end ;
    _Panel.RePaint ;
    FormResize( nil ) ;
end ;


procedure TArray_Display_Form.FormKeyPress( Sender : TObject ; var Key : Char ) ;

var S : string ;

begin
    case upcase( Key ) of
        'A' : begin // ASCII
                  _Display.ASCII_EBCDIC := 0 ;
                  _Panel.Repaint ;
              end ;
        'B' : begin // Byte display
                  _Display.Number_Size := 0 ;
                  _Panel.Repaint ;
              end ;
        'C' : ; // Change base
        'D' : begin // Decimal (base 10)
                  _Display.Base := 10 ;
                  _Panel.Repaint ;
              end ;
        'E' : begin // ENCDIC
                  _Display.ASCII_EBCDIC := 1 ;
                  _Panel.Repaint ;
              end ;
        'H' : begin // Hexadecimal (base 16)
                  _Display.Base := 16 ;
                  _Panel.Repaint ;
              end ;
        'L' : begin // Longword display
                  _Display.Number_Size := 3 ;
                  _Panel.Repaint ;
              end ;
        'O' : begin // Octal (base 8)
                  _Display.Base := 8 ;
                  _Panel.Repaint ;
              end ;
        'R' : begin // Radix-50 (base 40)
                  _Display.Base := 50 ;
                  _Panel.Repaint ;
              end ;
        'U' : begin // Toggle signed
                  _Display.Unsigned := not _Display.Unsigned ;
                  _Panel.Repaint ;
              end ;
        'W' : begin // Word display
                  _Display.Number_Size := 1 ;
                  _Panel.Repaint ;
              end ;
        'Z' : begin // Reset to default settings
                  _Display.Number_Size := 0 ;
                  _Display.Unsigned := True ;
                  _Display.Base := 10 ;
                  _Panel.Repaint ;
              end ;
    end ;
    if( _Display.Unsigned ) then
    begin
        S := 'Unsigned ' ;
    end else
    begin
        S := 'Signed ' ;
    end ;
    case _Display.Number_Size of
        0 : S := S + 'byte' ;
        1 : S := S + 'word' ;
        3 : S := S + 'longword' ;
    end ;
    case _Display.Base of
        2 : S := S + ', Binary' ;
        8 : S := S + ', Octal' ;
        10 : S := S + ', Decimal' ;
        16 : S := S + ', Hexadecimal' ;
        50 : S := S + ', Radix-50' ;
        else S := S + ', Base ' + inttostr( _Display.Base ) ;
    end ;
    if( length( _Caption ) = 0 ) then
    begin
        Caption := S ;
    end else
    begin
        Caption := _Caption + ': ' + S ;
    end ;
end ;


// API...

procedure TArray_Display_Form.Set_Caption( Value : string ) ;

begin
    _Caption := Value ;
end ;



end.
