{
        Program Name : TCommon_COM_Interface
        Package Name : LB
        Purpose      : Abstract virtual common COM interface class
        Institution  : Conroy & Conroy Co.
        Date Written : 21-March-2002
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2002 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TCommon_COM_Interface is a base class definition for COM objects
        used in the Application Foundation.
}

unit COMInter ;

{$I EDEFINES.INC}

interface

uses { C&C... }
     _DebugIn, { TDebug_Interface }
     _UEHDefs ; { TUEC }

{$IFDEF Delphi}
type TCommon_COM_Interface = class
				 private { Instance data... }
                                     Reference_Count : longint ;

                                 public { API... }
                                     { Initialize the object.  Returns result of
                                       initialization attempt. }
                                     function Initialize : TUEC ;
                                         virtual ; stdcall ;

                                     { Terminate the use of the object.  Returns
                                       result of termination.  This generally
                                       should not be used - instead, use the
                                       Detach method. }
                                     function Terminate : TUEC ;
                                         virtual ; stdcall ;

                                     { Increments the reference count for the
                                       object. }
                                     procedure Attach ;
                                         virtual ; stdcall ;

                                     { Decrements the reference count for the
                                       object.  When count reaches 0, the object
                                       is destroyed. }
                                     procedure Detach ;
                                         virtual ; stdcall ;

                                     { Version of the COM interface for this
                                       object. }
                                     function Interface_Version : longint ;
                                         virtual ; stdcall ;

                                     { Facility ID for the facility represented
                                       by this object.  Returns -1 if no
                                       facility assigned. }
                                     function Facility_ID : longint ;
                                         virtual ; stdcall ;

                                     { Facility version.  Result only has
                                       meaning if Facility_ID doesn't return
                                       -1. }
                                     function Facility_Version : longint ;
                                         virtual ; stdcall ;

                                     { Returns a debugging interface for the
                                       object.  Returns nil if not supported. }
                                     function Debugger : TDebug_Interface ;
                                         virtual ; stdcall ;

                                     { Returns a pointer to an object with
                                       extended common COM methods for object.
                                       Always returns nil for now. }
                                     function Extension : pointer ;
                                         virtual ; stdcall ;
                             end ; { TCommon_COM_Interface }
     PCommon_COM_Interface = class( TCommon_COM_Interface ) ;
{$ENDIF}

implementation

{$IFDEF DELPHI}
{ TCommon_COM_Interface methods... }

{ API... }

function TCommon_COM_Interface.Initialize ;

begin
    Reference_Count := 0 ;
end ;


function TCommon_COM_Interface.Terminate : TUEC ;

begin
    Free ;
end ;


procedure TCommon_COM_Interface.Attach ;

begin
    inc( Reference_Count ) ;
end ;


procedure TCommon_COM_Interface.Detach ;

begin
    dec( Reference_Count ) ;
    if( Reference_Count < 1 ) then
    begin
        Terminate ;
    end ;
end ;


function TCommon_COM_Interface.Interface_Version : longint ;

begin
    Interface_Version := 0 ;
end ;


function TCommon_COM_Interface.Facility_ID : longint ;

begin
    Facility_ID := -1 ;
end ;


function TCommon_COM_Interface.Facility_Version : longint ;

begin
    Facility_Version := 0 ;
end ;


function TCommon_COM_Interface.Debugger : TDebug_Interface ;

begin
    Debugger := nil ;
end ;


function TCommon_COM_Interface.Extension : pointer ;

begin
    Extension := nil ;
end ;
{$ENDIF}



end.
