{$V-}
{
        Program Name : Compatib
        Package Name : LB Layer 1
        Purpose      : Provides compatibility between different Pascal versions
        Institution  : Conroy & Conroy Co.
        Date Written : 03-Jul-1996
        Written By   : Alan Conroy
        Version      : 1.0A

        Copyright (C) 1996-2002 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  21-Aug-2001   EAC         Added GetEnv.                  *
        *  23-Aug-2002   EAC         Redo FileExists.               *
        *  13-Nov-2002   EAC         Added strlen.                  *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit provides definitions to reduce the differences between
	the various versions of Borland Pascal/Delphi.
}

unit Compatib ;

{$I EDEFINES.INC}

interface

{$IFDEF BP}
uses WinDOS,
  {$IFDEF WINDOWS}
     WinProcs, WinTypes,
  {$ELSE}
     { IMPORTANT: DOS MUST follow WINDOS, or the FindNext fails }
     DOS,
  {$ENDIF}
     Strings ;

  {$IFDEF WINDOWS}
type Registers = TRegisters ;
  {$ENDIF}

function FileGetAttr( const S : string ) : integer ;

function FileGetTime( const S : string ) : longint ;

procedure File_Split( Nam : string ; var S1, S2, S3 : string ) ;

function FindFirst( S : string ; Flags : integer ;
    var File_Stuff : TSearchRec ) : integer ;

function FindNext( var File_Stuff : TSearchRec ) : integer ;

procedure Intr( I : byte ; R : TRegisters ) ;

procedure ReallocMem( var P : pointer ; Old_Size, New_Size : longint ) ;

procedure SetLength( var X : string ; Y : integer ) ;

const HInstance = 0 ;

{$ELSE}

uses { Borland... }
     Classes,   { TList }
     SysUtils ;

type DateTime = record
                    Year, Month, Day, Hour, Min, Sec : word ;
                end ;

     TRegisters = record
                      case integer of
                          0 : ( AX, BX, CX, DX, BP, SI, DI, DS, ES, Flags : word ) ;
                          1 : ( _AX, _BX, _CX, _DX, _BP, _SI, _DI, _DS, _ES, _Flags : word ) ;
                          2 : ( AL, AH, BL, BH, CL, CH, DL, DH : byte ) ;
                  end ;

const faDirectory = SysUtils.faDirectory ;

function FileGetAttr( const S : string ) : integer ;

function FileGetTime( const S : string ) : integer ;

procedure File_Split( Nam : string ; var S1, S2, S3 : string ) ;

function FindFirst( S : string ; Flags : integer ; var File_Stuff : TSearchRec ) : integer ;

function FindNext( var File_Stuff : TSearchRec ) : integer ;

procedure GetCurDir( X : PChar ; Y : integer ) ;

procedure GetDate( var Year, Month, Day, DayofWeek : word ) ;

procedure GetTime( var Hour, Minute, Second, Sec100 : word ) ;

function MaxAvail : longint ;

function MemAvail : longint ;

procedure PackTime( var DT : DateTime ; var Time : longint ) ;

procedure ReallocMem( var P : pointer ; Old_Size, New_Size : longint ) ;
{$ENDIF}


{ Functions for both Borland Pascal and Delphi... }

function FileExists( const S : string ) : boolean ;

function strlen( P : PChar ) : integer ;

function strpas( P : PChar ) : string ;

function Uppercase( const X : string ) : string ;

{$IFDEF WINDOWS}
procedure Exec( const Path, Parameters : string ) ;
{$ENDIF}

{$IFDEF Delphi}
procedure Test_Unit ;
{$ENDIF}

implementation

{$IFDEF DELPHI}
uses Windows ;
{$ELSE}
uses CommonUt ; { Set_Length }
{$ENDIF}

{$IFNDEF DELPHI}
procedure Intr( I : byte ; R : TRegisters ) ;

begin
  {$IFDEF WINDOWS}
    WinDOS.Intr( I, R ) ;
  {$ELSE}
    DOS.Intr( I, Registers( R ) ) ;
  {$ENDIF}
end ;


procedure ReallocMem( var P : pointer ; Old_Size, New_Size : longint ) ;

var T : pointer ;

begin
    if( Old_Size = New_Size ) then
    begin
        exit ;
    end ;
    if( New_Size = 0 ) then
    begin
	T := nil ;
    end else
    begin
	getmem( T, New_Size ) ;
    end ;
    if( P <> nil ) then
    begin
	move( P^, T^, Old_Size ) ;
	freemem( P, Old_Size ) ;
    end ;
    P := T ;
end ;


procedure SetLength( var X : String ; Y : integer ) ;

begin
    Set_Length( X, Y ) ;
end ;


function strlen( P : PChar ) : integer ;

begin
    strlen := strings.strlen( P ) ;
end ;


function strpas( P : PChar ) : string ;

begin
    strpas := strings.strpas( P ) ;
end ;


function Uppercase( const X : string ) : string ;

begin
    Uppercase := Edit( X, 32 ) ;
end ;


function FileGetAttr( const S : string ) : integer ;

var F : file ;
    X : word ;

begin
    assign( F, S ) ;
    GetFAttr( F, X ) ;
    FileGetAttr := X ;
end ;


function FileGetTime( const S : string ) : longint ;

var F : file ;
    X : longint ;

begin
    assign( F, S ) ;
    reset( F ) ;
    GetFTime( F, X ) ;
    close( F ) ;
    FileGetTime := X ;
end ;


procedure File_Split( Nam : string ; var S1, S2, S3 : string ) ;

begin
  {$IFDEF WINDOWS}
    Nam[ length( Nam ) ] := #0 ;
    FileSplit( @Nam[ 1 ], @S1[ 1 ], @S2[ 1 ], @S3[ 1 ] ) ;
    S1[ 0 ] := chr( strpos( @S1[ 1 ], #0 ) - @S1[ 1 ] ) ;
    S2[ 0 ] := chr( strpos( @S2[ 1 ], #0 ) - @S2[ 1 ] ) ;
    S3[ 0 ] := chr( strpos( @S3[ 1 ], #0 ) - @S3[ 1 ] ) ;
  {$ELSE}
    FSplit( Nam, S1, S2, S3 ) ;
  {$ENDIF}
end ;


  {$IFNDEF WINDOWS}
procedure Convert_SearchRec( var File_Stuff : SearchRec ) ;
{ Convert DOS SearchRec record to Windows TSearchRec record }

var L : integer ;

begin
    L := length( File_Stuff.Name ) ;
    move( File_Stuff.Name[ 1 ], File_Stuff.Name[ 0 ], L ) ;
    File_Stuff.Name[ L ] := #0 ;
end ;
  {$ENDIF}


function FindFirst( S : string ; Flags : integer ; var File_Stuff : TSearchRec ) : integer ;

var R : integer ;

begin
  {$IFDEF WINDOWS}
    S := S + #0 ;
    WinDOS.FindFirst( @S[ 1 ], Flags, File_Stuff ) ;
  {$ELSE}
    DOS.FindFirst( S, Flags, SearchRec( File_Stuff ) ) ;
  {$ENDIF}
    R := DOSError ;
  {$IFNDEF WINDOWS}
    if( R = 0 ) then
    begin
	Convert_SearchRec( SearchRec( File_Stuff ) ) ;
    end ;
  {$ENDIF}
    FindFirst := R ;
end ;


function FindNext( var File_Stuff : TSearchRec ) : integer ;

var R : integer ;

begin
  {$IFDEF WINDOWS}
    WinDOS.FindNext( File_Stuff ) ;
  {$ELSE}
    SearchRec( File_Stuff ).Name := strpas( File_Stuff.Name ) ;
    DOS.FindNext( SearchRec( File_Stuff ) ) ;
  {$ENDIF}
    R := DOSError ;
  {$IFNDEF WINDOWS}
    if( R = 0 ) then
    begin
	Convert_SearchRec( SearchRec( File_Stuff ) ) ;
    end ;
  {$ENDIF}
    FindNext := R ;
end ;
{$ELSE}


{ DELPHI code... }

function FileGetAttr( const S : string ) : integer ;

begin
    FileGetAttr := SysUtils.FileGetAttr( S ) ;
end ;


function FileGetTime( const S : string ) : integer ;

var H : integer ;

begin
    H := FileOpen( S, fmOpenRead ) ;
    FileGetTime := SysUtils.FileGetDate( H ) ;
    FileClose( H ) ;
end ;


procedure File_Split( Nam : string ; var S1, S2, S3 : string ) ;

var X : integer ;

begin
    S1 := ExtractFileDir( Nam ) ;
    if( ( length( S1 ) > 0 ) and ( copy( S1, length( S1 ), 1 ) <> '\' ) ) then
    begin
        S1 := S1 + '\' ; { ExtractFileDir trims trailing "\" }
    end ;
    S2 := ExtractFileName( Nam ) ;
    X := pos( '.', S2 + '.' ) ;
    S2 := copy( S2, 1, X - 1 ) ;
    S3 := ExtractFileExt( Nam ) ;
end ;


function FindFirst( S : string ; Flags : integer ;
    var File_Stuff : TSearchRec ) : integer ;

begin
    Result := SysUtils.FindFirst( S, Flags, File_Stuff ) ;
end ;


function FindNext( var File_Stuff : TSearchRec ) : integer ;

begin
    Result := SysUtils.FindNext( File_Stuff ) ;
end ;


procedure GetCurDir( X : PChar ; Y : integer ) ;

begin
    GetCurrentDirectory( 127, X ) ;
end ;


procedure GetDate( var Year, Month, Day, DayofWeek : word ) ;

begin
    DecodeDate( SysUtils.Date, Year, Month, Day ) ;
    DayofWeek := SysUtils.DayOfWeek( SysUtils.Date ) ;
end ;


procedure GetTime( var Hour, Minute, Second, Sec100 : word ) ;

begin
    DecodeTime( SysUtils.Time, Hour, Minute, Second, Sec100 ) ;
end ;


function MaxAvail : longint ;

begin
    MaxAvail := MemAvail ;
end ;


function MemAvail : longint ;

begin
    MemAvail := $7FFFFFFF ;
end ;


procedure PackTime( var DT : DateTime ; var Time : longint ) ;

var X : array[ 0..3 ] of byte ;

begin
    X[ 0 ] := 0 ;
    X[ 1 ] := DT.Sec ;
    X[ 2 ] := DT.Min ;
    X[ 3 ] := DT.Hour ;
    Time := longint( X ) ;
end ;


procedure ReallocMem( var P : pointer ; Old_Size, New_Size : longint ) ;

begin
    System.ReallocMem( P, New_Size ) ;
end ;


function strlen( P : PChar ) : integer ;

begin
    strlen := SysUtils.strlen( P ) ;
end ;


function strpas( P : PChar ) : string ;

begin
    strpas := string( P ) ;
end ;


function Uppercase( const X : string ) : string ;

begin
    Uppercase := SysUtils.Uppercase( X ) ;
end ;
{$ENDIF}


function FileExists( const S : string ) : boolean ;

var Dummy : integer ;
    F : file ;
    Saved : integer ;

begin
    assign( F, S ) ;
    Saved := FileMode ;
    FileMode := 0 ; { Open in read-only }
    {$I-}
    reset( F ) ;
    {$I+}
    FileMode := Saved ;
    Dummy := IOResult ;
    if( Dummy = 0 ) then { If opened }
    begin
	close( F ) ;
    end ;
    if( ( Dummy = 2 ) or ( Dummy = 3 ) ) then { File/Path doesn't exist }
    begin
	FileExists := False ;
    end else
    begin
	FileExists := True ;
    end ;
end ; { FileExists }


{$IFDEF WINDOWS}
procedure Exec( const Path, Parameters : string ) ;

begin
    O_S^.Spawn( Path, Parameters ) ;
end ;
{$ENDIF}


{$IFDEF Delphi}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


procedure Test_File_Split ;

var S1, S2, S3 : string ;

begin
    File_Split( 'File.Ext', S1, S2, S3 ) ;
    if( S1 <> '' ) then
    begin
        Error ;
    end ;
    if( S2 <> 'File' ) then
    begin
        Error ;
    end ;
    if( S3 <> '.Ext' ) then
    begin
        Error ;
    end ;
    File_Split( 'C:\File.Ext', S1, S2, S3 ) ;
    if( S1 <> 'C:\' ) then
    begin
        Error ;
    end ;
    if( S2 <> 'File' ) then
    begin
        Error ;
    end ;
    if( S3 <> '.Ext' ) then
    begin
        Error ;
    end ;
    File_Split( 'C:File.Ext', S1, S2, S3 ) ;
    if( S1 <> 'C:\' ) then
    begin
        Error ;
    end ;
    if( S2 <> 'File' ) then
    begin
        Error ;
    end ;
    if( S3 <> '.Ext' ) then
    begin
        Error ;
    end ;
end ;


procedure Test_ReallocMem ;

var P : pointer ;

begin
    getmem( P, 10 ) ;
    Reallocmem( P, 10, 32 ) ;
end ;


procedure Test_PackTime ;

var DT : DateTime ;
    Time : longint ;

begin
    DT.Sec := 1 ;
    DT.Min := 2 ;
    DT.Hour := 3 ;
    PackTime( DT, Time ) ;
    if( Time <> 50462976 ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Strlen ;

begin
    if( strlen( 'ABCD' ) <> 4 ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Uppercase ;

begin
    if( Uppercase( 'abC!' ) <> 'ABC!' ) then
    begin
        Error ;
    end ;
end ;


procedure Test_SetLength ;

var S : string ;

begin
    S := '1234' ;
    SetLength( S, 2 ) ;
    if( S <> '12' ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Unit ;

begin
    Test_File_Split ;
    Test_ReallocMem ;
    Test_PackTime ;
    Test_Strlen ;
    Test_Uppercase ;
    Test_SetLength ;
end ;
{$ENDIF}



end.
