{
        Program Name : Command
        Package Name : LB
        Purpose      : Command object
        Institution  : Conroy & Conroy Co.
        Date Written : 4-August-2004
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2004 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit Commands ;

interface

{$I EDEFINES.INC}

uses { Borland... }
{$IFDEF DELPHI}
     Classes, { TList }
{$ENDIF}
     { C&C... }
     Behavior, { TSBehavior }
     Common, { TObject }
     Events ; { TEvent }

type TSCommand = object( TBase_Object )
                     public { Constructors and destructors... }
                         constructor Init ;
                         destructor Done ; virtual ;

                     private { Instance data... }
                         _Enabled : boolean ;
                         _Behavior : PSBehavior ;

                     public { API... }
                         function Get_Behavior : PSBehavior ; virtual ;
                         function Get_Enabled : boolean ; virtual ;
                         procedure Set_Behavior( Value : PSBehavior ) ; virtual ;
                         procedure Set_Enabled( Value : boolean ) ; virtual ;

                         procedure Execute( Event : PSEvent ) ; virtual ;
                 end ;
     PSCommand = ^TSCommand ;

{$IFDEF DELPHI}
type TCommand = class( Tbase_Delphi_Object )
                    public { Constructors and destructors... }
                        constructor Create ;
                        destructor Destroy ; override ;

                    private { Instance data... }
                        _Enabled : boolean ;
                        _Behavior : TBehavior ;
                        _List : TList ;

                    protected { Property handlers... }
                        function Get_Behavior : TBehavior ; virtual ;
                        function Get_Enabled : boolean ; virtual ;
                        procedure Set_Behavior( Value : TBehavior ) ; virtual ;
                        procedure Set_Enabled( Value : boolean ) ; virtual ;

                    public { API... }
                        procedure Add( Component : TComponent ) ;
                            virtual ;

                        procedure Clear ; virtual ;

                        procedure Execute( Event : TEvent ) ; virtual ;

                        procedure Remove( Component : TComponent ) ;
                            virtual ;

                        property Behavior : TBehavior
                            read Get_Behavior
                            write Set_Behavior ;

                        property Enabled : boolean
                            read Get_Enabled
                            write Set_Enabled ;
                end ;
{$ENDIF}

implementation

{$IFDEF DELPHI}
uses { Borland... }
     Controls ; { TControl }
{$ENDIF}


{ TSCommand methods... }

{ Constructors and destructors... }

constructor TSCommand.Init ;

begin
    inherited Init ;

    _Behavior := nil ;
    _Enabled := True ;
end ;


destructor TSCommand.Done ;

begin
    Set_Behavior( nil ) ;

    inherited Done ;
end ;


{ API... }

function TSCommand.Get_Behavior : PSBehavior ;

begin
    Get_Behavior := _Behavior ;
end ;


function TSCommand.Get_Enabled : boolean ;

begin
    Get_Enabled := _Enabled ;
end ;


procedure TSCommand.Set_Behavior( Value : PSBehavior ) ;

begin
    if( Value <> nil ) then
    begin
        Value^.Attach ;
    end ;
    if( _Behavior <> nil ) then
    begin
        _Behavior^.Detach ;
    end ;
    _Behavior := Value ;
end ;


procedure TSCommand.Set_Enabled( Value : boolean ) ;

begin
    _Enabled := Value ;
end ;


procedure TSCommand.Execute( Event : PSEvent ) ;

begin
    if( _Enabled and ( _Behavior <> nil ) ) then
    begin
        _Behavior^.Execute( Event ) ;
    end ;
end ;


{$IFDEF Delphi}
{ TCommand methods... }

{ Constructors and destructors... }

constructor TCommand.Create ;

begin
    inherited Create ;

    _Enabled := True ;
end ;


destructor TCommand.Destroy ;

begin
    Set_Behavior( nil ) ;
    Clear ;
    _List.Free ;
    _List := nil ;

    inherited Destroy ;
end ;


{ API... }

function TCommand.Get_Behavior : TBehavior ;

begin
    Get_Behavior := _Behavior ;
end ;


function TCommand.Get_Enabled : boolean ;

begin
    Get_Enabled := _Enabled ;
end ;


procedure TCommand.Set_Behavior( Value : TBehavior ) ;

begin
    if( Value <> nil ) then
    begin
        Value.Attach ;
    end ;
    if( _Behavior <> nil ) then
    begin
        _Behavior.Detach ;
    end ;
    _Behavior := Value ;
end ;


procedure TCommand.Set_Enabled( Value : boolean ) ;

var Loop : integer ;

begin
    _Enabled := Value ;

    if( _List <> nil ) then
    begin
        for Loop := 0 to _List.Count - 1 do
        begin
            if( TComponent( _List[ Loop ] ) is TControl ) then
            begin
                TControl( _List[ Loop ] ).Enabled := Value ;
            end ;
        end ;
    end ;
end ;


{ API... }

procedure TCommand.Add( Component : TComponent ) ;

begin
    if( Component <> nil ) then
    begin
        if( _List = nil ) then
        begin
            _List := TList.Create ;
        end ;
        _List.Add( Component ) ;
    end ;
end ;


procedure TCommand.Clear ;

begin
    if( _List <> nil ) then
    begin
        _List.Clear ;
    end ;
end ;


procedure TCommand.Execute( Event : TEvent ) ;

begin
    if( Enabled and ( _Behavior <> nil ) ) then
    begin
        _Behavior.Execute( Event ) ;
    end ;
end ;


procedure TCommand.Remove( Component : TComponent ) ;

begin
    if( _List <> nil ) then
    begin
        _List.Remove( Component ) ;
    end ;
end ;
{$ENDIF}



end.

