{$N+}
{
        Program Name : CommonUt
        Package Name : Subroutine Library - Layer 1
        Purpose      : Common Utility functions and classes
        Institution  : Conroy & Conroy Co.
        Date Written : 28-Nov-88
        Written By   : Alan Conroy
        Version      : 2.1E

        Copyright (c) 1988-2005 by Alan Conroy.  Released to the Public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit contains common utility function and classes.
}

{$I EDEFINES.INC}

unit CommonUt ;

interface

uses
{$IFDEF DELPHI}
     { Borland... }
     Classes,
{$ENDIF}
     { Layer 0... }
     _DebugIn, { PSDebug_Interface }
     _Types ; { Small_String }

{$IFDEF DELPHI}
type TInteger_List = class( TList )
                         public { Constructors and destructors... }
                             constructor Create ;
                             destructor Destroy ; override ;

                         private { Instance data... }
                             Serial_Number : integer ;

                         protected { Property handlers }
                             function Get_Index( Index : integer ) : integer ;
                                 virtual ;
                             procedure Set_Index( Index : integer ; Value : integer ) ;
                                 virtual ;

                         public { API... }
                             function Add( Value : integer ) : integer ;
                                 virtual ;
                             procedure Assign( Source : TInteger_List ) ;
                             function Debugger : TCOM_Debug_Interface ;
                             procedure Deserialize( const Image : string ) ;
                             procedure Remove( Value : integer ) ;
                             procedure Increment( Index : integer ) ;
                             function IndexOf( Value : integer ) : integer ;
                             procedure Insert( Index, Value : integer ) ;
                             function Serialize : string ;
                             procedure Sort( Mask : integer ) ;

                             property Items[ Index : integer ] : integer
                                 read Get_Index
                                 write Set_Index ; default ;
  {$IFDEF Test}
                             class procedure Test_Class ;
  {$ENDIF}
                     end ;
{$ENDIF}

function Boolean_To_String( X : boolean ) : string ;
function Edit( const Original : string ; Flags : Integer ) : string ;
{ Slow edit function }
procedure FEdit( var Dst : string ; const Src : string ; Flags : longint ) ;
function Pointer_To_String( P : pointer ) : string ;
{$IFDEF DELPHI}
procedure Set_Length( var X : string ; L : integer ) ; overload ;
procedure Set_Length( var X : Small_String ; L : integer ) ; overload ;
{$ELSE}
procedure Set_Length( var X : string ; L : integer ) ;
procedure SetLength( X : PChar ; L : integer ) ;
{$ENDIF}

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses

{$IFDEF DELPHI}
     { Borland... }
     SysUtils, { inttostr }

     { Layer 0... }
     _App, { Application_Manager }

     { Layer 1... }
     DebugInt, { TText_Debugger }
{$ENDIF}
     ASCIIDef, { HT }
     CVT, { CVTB }
     Num1s ; { Num1 }

{$IFDEF DELPHI}
type TInteger_List_Debugger = class( TText_Debugger )
                                  public { Instance data }
                                      List : TInteger_List ;

                                  public { API... }
                                      function Count : longint ;
                                          override ;

                                      function Child( Ordinal : longint ) : PDebug_Interface ;
                                          override ;
                              end ;

function TInteger_List_Debugger.Count : longint ;

begin
    Count := List.Count ;
end ;


function TInteger_List_Debugger.Child( Ordinal : longint ) : PDebug_Interface ;

var Res : PText_Debugger ;

begin
    Res := TText_Debugger.Create ;
    Res.Set_Title( PChar( inttostr( Ordinal ) + ': ' + inttostr( List[ Ordinal ] ) ) ) ;
    Child := Res ;
end ;



const Integer_List_Facility = 88 ;

{ TInteger_List methods... }

{ Constructors and destructors... }

constructor TInteger_List.Create ;

begin
    inherited Create ;

    if( Application_Manager <> nil ) then
    begin
        Serial_Number := Application_Manager^.Get_Serial_Number( 0 ) ;
        Application_Manager^.Object_Manager^.Add_Object( self, Integer_List_Facility, pointer( Serial_Number ) ) ;
    end ;
end ;


destructor TInteger_List.Destroy ;

begin
    if( Application_Manager <> nil ) then
    begin
        Application_Manager^.Object_Manager^.Remove_Object( self, Integer_List_Facility ) ;
    end ;

    inherited Destroy ;
end ;


{ Property handlers... }

function TInteger_List.Get_Index( Index : integer ) : integer ;

begin
{$IFDEF Development}
    Result := integer( inherited Items[ Index ] ) ;
{$ELSE}
    Result := integer( List[ Index ] ) ; 
{$ENDIF}
end ;


procedure TInteger_List.Set_Index( Index : integer ; Value : integer ) ;

begin
{$IFDEF Development}
    inherited Items[ Index ] := pointer( Value ) ;
{$ELSE}
    List[ Index ] := pointer( Value ) ; 
{$ENDIF}
end ;


function TInteger_List.Add( Value : integer ) : integer ;

begin
    if( Count = Capacity ) then { Need to expand }
    begin
        if( Count < 1024 ) then
        begin
            Capacity := Capacity * 2 ;
        end else
        begin
            Capacity := Capacity + 1024 ;
        end ;
    end ;
    inherited Add( pointer( Value ) ) ;
    Result := Count - 1 ;
end ;


procedure TInteger_List.Assign( Source : TInteger_List ) ;

begin
    Count := Source.Count ;
    System.move( Source.List[ 0 ], List[ 0 ], Count * sizeof( integer ) ) ;
end ;


function TInteger_List.Debugger : TCOM_Debug_Interface ;

var Res : TInteger_List_Debugger ;

begin
    Res := TInteger_List_Debugger.Create ;
    Res.List := self ;
    Debugger := Res ;
end ;


procedure TInteger_List.Deserialize( const Image : string ) ;

var Dummy : integer ;
    Temp : string ;

begin
    Clear ;
    if( Image = '0' ) then
    begin
        exit ;
    end ;
    Dummy := pos( '|', Image ) ;
    Capacity := strtoint( copy( Image, 1, Dummy - 1 ) ) ;
    Temp := copy( Image, Dummy + 1, length( Image ) ) ;
    while( length( Temp ) > 0 ) do
    begin
        Dummy := pos( '|', Temp ) ;
        if( Dummy = 0 ) then
        begin
            Dummy := length( Temp ) + 1 ;
        end ;
        Add( strtoint( copy( Temp, 1, Dummy - 1 ) ) ) ;
        Temp := copy( Image, Dummy + 1, length( Image ) ) ;
    end ;
end ;


procedure TInteger_List.Remove( Value : integer ) ;

begin
    inherited Remove( pointer( Value ) ) ;
end ;


procedure TInteger_List.Increment( Index : integer ) ;

begin
    Items[ Index ] := Items[ Index ] + 1 ;
end ;


function TInteger_List.IndexOf( Value : integer ) : integer ;

begin
    Result := inherited IndexOf( pointer( Value ) ) ;
end ;


procedure TInteger_List.Insert( Index, Value : integer ) ;

begin
    inherited Insert( Index, pointer( Value ) ) ;
end ;


function TInteger_List.Serialize : string ;

var Loop : integer ;

begin
    Result := inttostr( Count ) ;
    for Loop := 0 to Count - 1 do
    begin
        Result := Result + '|' + inttostr( Items[ Loop ] ) ;
    end ;
end ;


procedure TInteger_List.Sort( Mask : integer ) ;

var Loop, Loop1, I, I1, M, M1 : integer ;

begin
    for Loop := 0 to Count - 1 do
    begin
        I := Items[ Loop ] ;
        M := I and Mask ;
        for Loop1 := Loop to Count - 1 do
        begin
            I1 := Items[ Loop1 ] ;
            M1 := I1 and Mask ;
            if( M > M1 ) then
            begin
                Items[ Loop ] := I1 ;
                Items[ Loop1 ] := I ;
                I := I1 ;
                M := M1 ;
            end ;
        end ;
    end ;
end ;


  {$IFDEF Test}
class procedure TInteger_List.Test_Class ;

var {Intermediate : integer ;}
    Loop : integer ;
    Test, Test1 : TInteger_List ;

begin
    { Set up memory manager }
{    Start_Memory_Test ;}

    { Create it... }
    Test := TInteger_List.Create ;
    Test1 := TInteger_List.Create ;
{    Intermediate := Allocation_Count ;}

    { Test Add and IndexOf... }
    for Loop := 0 to 10 do
    begin
        Test.Add( Loop ) ;
        if( Test[ Loop ] <> Loop ) then
        begin
            raise Exception.Create( 'TInteger_List.Add (or default array) failed' ) ;
        end ;
        if( Test.IndexOf( Loop ) <> Loop ) then
        begin
            raise Exception.Create( 'TInteger_List.IndexOf failed' ) ;
        end ;
    end ;
    if( Test.IndexOf( 11 ) <> -1 ) then
    begin
        raise Exception.Create( 'TInteger_List.IndexOf of non-existant value failed' ) ;
    end ;

    { Test Insert... }
    Test.Insert( 0, 11 ) ;
    if( ( Test[ 0 ] <> 11 ) or ( Test.Count <> 12 ) ) then
    begin
        raise Exception.Create( 'TInteger_List.Insert failed' ) ;
    end ;
    for Loop := 1 to 11 do
    begin
        if( Test[ Loop ] <> Loop - 1 ) then
        begin
            raise Exception.Create( 'TInteger_List.Insert failed' ) ;
        end ;
    end ;
    Test.Insert( 12, -1 ) ;
    if( Test[ 12 ] <> -1 ) then
    begin
        raise Exception.Create( 'TInteger_List.Insert failed' ) ;
    end ;

    { Test Assign... }
    Test1.Assign( Test ) ;
    if( Test1.Count <> Test.Count ) then
    begin
        raise Exception.Create( 'TInteger_List.Assign failed' ) ;
    end ;
    for Loop := 0 to Test.Count - 1 do
    begin
        if( Test[ Loop ] <> Test1[ Loop ] ) then
        begin
            raise Exception.Create( 'TInteger_List.Assign failed' ) ;
        end ;
    end ;

    Test.Clear ;
    Test1.Clear ;

    { Check memory usage... }
{
    if( Allocation_Count <> Intermediate ) then
    begin
        raise Exception.Create( 'Memory loss' ) ;
    end ;
}

    Test.Free ;
    Test1.Free ;

    { Check memory usage... }
{
    if( Allocation_Count <> 0 ) then
    begin
        raise Exception.Create( 'Memory loss' ) ;
    end ;
    End_Memory_Test ;
}
end ;
  {$ENDIF}

{$ENDIF}


function Boolean_To_String( X : boolean ) : string ;

begin
    if( X ) then
    begin
        Boolean_To_String := 'true' ;
    end else
    begin
        Boolean_To_String := 'false' ;
    end ;
end ;


procedure FEdit( var Dst : string ; const Src : string ; Flags : longint ) ;
(*
          This function will perform various editing functions on
        strings.  The format for function usage is: FEDIT(dst,src,flgs)
        where flgs has the following meanings:

              AND     1 = Trim Most Significant Bit.
              AND     2 = Remove all white space (spaces and tabs).
              AND     4 = Remove all nulls, linefeeds, formfeeds, carriage
                          returns, and deletes.
              AND     8 = Remove leading spaces/tabs.
              AND    16 = Reduce multiple white space (spaces and tabs) to one space.
              AND    32 = Convert lower to upper case.
              AND    64 = Convert '[' to '(' and ']' to ')'.
              AND   128 = Remove all trailing spaces.
              AND   256 = Leave characters within quotes (" and/or ') unmodified.
              AND   512 = Convert upper case to lower case.
              AND  1024 = Convert '[' to '{' and ']' to '}'.
              AND  2048 = Convert '(' to '[' and ')' to ']'.
              AND  4096 = Convert '(' to '{' and ')' to '}'.
              AND  8192 = Convert '{' to '(' and '}' to ')'.
              AND 16384 = Convert '{' to '[' and '}' to ']'.

              The following only apply to DELPHI...
              AND 32768 = Treat high-ASCII as Windows latin characters and
                          treat ASCII 160 as a space for flag 16.
              AND 65536 = Treat high-ASCII as IBM latin characters.
              AND $20000 = Convert accented characters to non-accented.
*)

{$IFDEF DELPHI}
var AH : char ; { Current character }
    AL : char ; { Original character (high bit trimmed) }
    Quotes : boolean ; { True if within quotes }
    ESI : integer ; { Position within source string }
    Leading : boolean ; { True if leading spaces }
    OK : boolean ; { if current character is OK for destination }
{$ENDIF}

var Destination : string ;
    Quote_Type : char ;
    Last : integer ;
    Space : boolean ;

label L0 ;

begin
    if( Length( Src ) = 0 ) then { No edits on zero-length strings }
    begin
        Dst := '' ;
        Exit ;
    end ;
    if ( Flags And 544 ) = 544 then
    begin
        Flags := Flags And Not 512 ;
    end ;
    { Disallow lower-case command if upper-case command is also present }
    if ( Flags And 1088 ) = 1088 then
    begin
        Flags := Flags And Not 1024 ;
    end ;
    { Disallow [] to {}{ if [] to () }
    if ( Flags And 6144 ) = 6144 then
    begin
        Flags := Flags And Not 4096 ;
    end ;
    { Disallow () to {}{ if () to [] }
    if ( Flags And 24576 ) = 24576 then
    begin
        Flags := Flags And Not 16384 ;
    end ;
    { Disallow {}{ to [] if {}{ to () }
    if( ( Flags and ( 32768 + 65536 ) ) = 32768 + 65536 ) then { Incompatible }
    begin
        Flags := Flags and not 32768 ;
        Flags := Flags and not 65536 ;
    end ;
{$IFDEF DELPHI}
    Destination := '' ;
    Space := False ; { No spaces }
    Last := 0 ;
    Quotes := False ; { No quotes yet }
    Quote_Type := ' ' ; { Avoid compiler warning }
    ESI := 0 ;
    Leading := True ;
L0:
    inc( ESI ) ; { Increment source string pointer }
    OK := True ; { This byte is OK - so far }
    AH := Src[ ESI ] ; { Current character }
    if( not Quotes ) then
    begin
        if( ( Flags and 1 ) = 1 ) then { Trim parity }
        begin
            AH := chr( ord( AH ) and 127 ) ;
        end ;
        AL := AH ; { Save original value }
        if( ( AH <> ' ' ) and ( AH <> HT ) ) then
        begin
            Leading := False ;
        end ;
        if( ( Flags and 2 ) = 2 ) then { Remove all spaces/tabs }
        begin
            if( ( AH = ' ' ) or ( AH = HT ) ) then
            begin
                OK := False ;
            end ;
        end ;
        if( ( Flags and 4 ) = 4 ) then { Ignore special values? }
        begin
            if( ( AH = NUL ) or ( AH = LF ) or ( AH = FF ) or ( AH = CR ) or ( AH = DEL ) ) then
            begin
                OK := False ;
            end ;
        end ;
        if( ( Flags and 8 ) = 8 ) then { Ignore leading spaces }
        begin
            if( Leading ) then
            begin
                OK := False ;
            end ;
        end ;
        if( ( Flags and 16 ) = 16 ) then { Reduce tabs/spaces to single space }
        begin
            if(
                ( AH = ' ' )
                or
                ( AH = HT )
                or
                ( ( AH = #160 ) and ( ( Flags and 32768 ) <> 0 ) ) { Non-breaking space }
              ) then
            begin
                if( Space ) then
                begin
                    OK := False ;
                end else
                begin
                    Space := True ;
                    AH := ' ' ;
                end ;
            end ;
        end ;
        if( ( Flags and $20000 ) <> 0 ) then
        begin
            if( ( Flags and 32768 ) <> 0 ) then { Windows }
            begin
                case AH of
                    #138 : AH := 'S' ;
                    #142 : AH := 'Z' ;
                    #154 : AH := 's' ;
                    #158 : AH := 'z' ;
                    #159 : AH := 'Y' ;
                    #192..#197 : AH := 'A' ;
                    #200..#203 : AH := 'E' ;
                    #204..#207 : AH := 'I' ;
                    #210..#214 : AH := 'O' ;
                    #217..#220 : AH := 'U' ;
                    #221 : AH := 'Y' ;
                    #224..#229 : AH := 'a' ;
                    #232..#235 : AH := 'e' ;
                    #236..#239 : AH := 'i' ;
                    #240 : AH := 'o' ;
                    {  is not considered accented - it is a different character }
                    #242..#246 : AH := 'o' ;
                    #249..#252 : AH := 'u' ;
                    #253 : AH := 'y' ;
                    #255 : AH := 'y' ;
                end ;
            end ;
            if( ( Flags and 65536 ) <> 0 ) then { DOS }
            begin
                case AH of
                    #129 : AH := 'u' ;
                    #130 : AH := 'e' ;
                    #131..#134 : AH := 'a' ;
                    #136..#138 : AH := 'e' ;
                    #139..#141 : AH := 'i' ;
                    #142..#143 : AH := 'A' ;
                    #144 : AH := 'E' ;
                    #147..#149 : AH := 'o' ;
                    #150..#151 : AH := 'u' ;
                    #152 : AH := 'y' ;
                    #153 : AH := 'O' ;
                    #154 : AH := 'U' ;
                    #160 : AH := 'a' ;
                    #161 : AH := 'i' ;
                    #162 : AH := 'o' ;
                    #163 : AH := 'u' ;
                    {  is not considered accented - it is a different character }
                end ;
            end ;
        end ; { if( ( Flags and $20000 ) <> 0 ) }
        if( ( Flags and 32 ) = 32 ) then { Lower to upper case }
        begin
            if( ( AH >= 'a' ) and ( AH <= 'z' ) ) then
            begin
                AH := char( ord( AH ) - 32 ) ;
            end ;
            if( ( Flags and 32768 ) <> 0 ) then
            begin
                if( ( AH >= #224 ) and ( AH <= #254 ) and ( AH <> #247 ) ) then
                begin
                    AH := char( ord( AH ) - 32 ) ;
                end ;
                if( AH = #255 ) then
                begin
                    AH := #159 ;
                end ;
                if( AH = #154 ) then
                begin
                    AH := #138 ;
                end ;
                if( AH = #156 ) then
                begin
                    AH := #140 ;
                end ;
            end ;
            if( ( Flags and 65536 ) <> 0 ) then
            begin
                if( AH = #164 ) then
                begin
                    AH := #165 ;
                end ;
                if( AH = #135 ) then
                begin
                    AH := #128 ;
                end ;
                if( AH = #129 ) then
                begin
                    AH := #154 ;
                end ;
                if( AH = #132 ) then
                begin
                    AH := #142 ;
                end ;
                if( AH = #134 ) then
                begin
                    AH := #143 ;
                end ;
                if( AH = #130 ) then
                begin
                    AH := #144 ;
                end ;
                if( AH = #145 ) then
                begin
                    AH := #146 ;
                end ;
                if( AH = #148 ) then
                begin
                    AH := #153 ;
                end ;
            end ;
        end ;
        if( ( Flags and 512 ) = 512 ) then { Upper to lower case }
        begin
            if( ( AH >= 'A' ) and ( AH <= 'Z' ) ) then
            begin
                AH := char( ord( AH ) + 32 ) ;
            end ;
            if( ( Flags and 32768 ) <> 0 ) then
            begin
                if( ( AH >= #192 ) and ( AH <= #222 ) and ( AH <> #215 ) ) then
                begin
                    AH := char( ord( AH ) + 32 ) ;
                end ;
                if( AH = #159 ) then
                begin
                    AH := #255 ;
                end ;
                if( AH = #138 ) then
                begin
                    AH := #154 ;
                end ;
                if( AH = #140 ) then
                begin
                    AH := #156 ;
                end ;
            end ;
            if( ( Flags and 65536 ) <> 0 ) then
            begin
                if( AH = #165 ) then
                begin
                    AH := #164 ;
                end ;
                if( AH = #128 ) then
                begin
                    AH := #135 ;
                end ;
                if( AH = #154 ) then
                begin
                    AH := #129 ;
                end ;
                if( AH = #142 ) then
                begin
                    AH := #132 ;
                end ;
                if( AH = #143 ) then
                begin
                    AH := #134 ;
                end ;
                if( AH = #144 ) then
                begin
                    AH := #130 ;
                end ;
                if( AH = #146 ) then
                begin
                    AH := #145 ;
                end ;
                if( AH = #153 ) then
                begin
                    AH := #148 ;
                end ;
            end ;
        end ;
        if( ( Flags and 64 ) = 64 ) then { Convert [] to () }
        begin
            if( AL = '[' ) then
            begin
                AH := '(' ;
            end else
            if( AL = ']' ) then
            begin
                AH := ')' ;
            end ;
        end ;
        if( ( Flags and 1024 ) = 1024 ) then { Convert [] to braces }
        begin
            if( ( AL = '[' ) or ( AL = ']' ) ) then
            begin
                AH := char( ord( AL ) + 32 ) ;
            end ;
        end ;
        if( ( Flags and 2048 ) = 2048 ) then { Convert () to [] }
        begin
            if( AL = '(' ) then
            begin
                AH := '[' ;
            end else
            if( AL = ')' ) then
            begin
                AH := ']' ;
            end ;
        end ;
        if( ( Flags and 4096 ) = 4096 ) then { Convert () to braces }
        begin
            if( AL = '(' ) then
            begin
                AH := '{' ;
            end else
            if( AL = ')' ) then
            begin
                AH := '}' ;
            end ;
        end ;
        if( ( Flags and 8192 ) = 8192 ) then { Convert braces to () }
        begin
            if( AL = '{' ) then
            begin
                AH := '(' ;
            end else
            if( AL = '}' ) then
            begin
                AH := ')' ;
            end ;
        end ;
        if( ( Flags and 16384 ) = 16384 ) then { Convert braces to [] }
        begin
            if( ( AL = '{' ) or ( AL = '}' ) ) then
            begin
                AH := char( ord( AL ) - 32 ) ;
            end ;
        end ;
    end ;
    if( OK ) then
    begin
        Destination := Destination + AH ; { Build result }
        if( ( AH = ' ' ) or ( AH = HT ) ) then
        begin
            Space := True ;
        end ;
    end ;
    if( ( AH <> ' ' ) and ( AH <> HT ) ) then
    begin
        Space := False ;
        if( ( Flags and 256 ) = 256 ) then { Allow no alter within quotes }
        begin
            if( ( AH = '"' ) or ( AH = #39 ) ) then
            begin
                if( not Quotes ) then
                begin
                    Quote_Type := AH ;
                end ;
                if( Quote_Type = AH ) then
                begin
                    Quotes := not Quotes ;
                end ;
            end ;
        end ;
    end ;
    if( ( AH <> ' ' ) and ( AH <> HT ) ) then
    begin
        Last := length( Destination ) ; { Last non-space character }
    end ;
    if( ESI < length( Src ) ) then
    begin
        goto L0 ;
    end ;
    if( ( Flags and 128 ) = 128 ) then { Trim following spaces }
    begin
        if( not Quotes ) then { Not within quotes }
        begin
            SetLength( Destination, Last ) ; { Last non-space character position }
        end ;
    end ;
    Dst := Destination ;
{$ELSE}
    Inline(
        {0000} $1E /               { PUSH DS ; Save DS and BP }
        {0001} $55 /               { PUSH BP }
        {0002} $C5 / $B6 / Src /   { LDS SI, [BP].Src ; Address of source string }
        {0006} $8A / $0C /         { MOV CL, [SI] ; Length of source string }
        {0008} $C4 / $Be / Dst /   { LES DI, [BP].Dst ; Address of destination string }
        {000C} $B3 / $00 /         { MOV BL, 0 ; No quotes yet }
        {000E} $B5 / $01 /         { MOV CH, 1 ; Leading spaces=true }
        {0010} $C6 / $86 / Space / $00 / { MOV [BP].Space, BYTE PTR 0 ; No spaces }
        {0015} $C6 / $86 / Last / $00 / { MOV [BP].Last, BYTE PTR 0; }
        {001A}                     { $0: }
        {001A} $46 /               { INC SI ; Increment source string pointer }
        {001B} $8B / $96 / Flags / { MOV DX, [BP].Flags ; Get flags }
        {001F} $B7 / $01 /         { MOV BH, 1 ; This byte is OK - so far }
        {0021} $8A / $04 /         { MOV AL, [SI] ; Get current character }
        {0023} $8A / $E0 /         { MOV AH, AL }
        {0025} $24 / $7F /         { AND AL, 127 ; Strip high bit for easy comparisons }
        {0027} $80 / $FB / $00 /   { CMP BL, 0 ; Quotes? }
        {002A} $74 / $03 /         { JE $0A ; no }
        {002C} $E9 / $0107 /       { R JMP $1 ; yes - so skip editing code }
        {002F}                     { $0A: }
        {002F} $3C / $20 /         { CMP AL, 32 ; space? }
        {0031} $74 / $06 /         { JE $2 ; yes }
        {0033} $3C / $09 /         { CMP AL, 9 ; tab? }
        {0035} $74 / $02 /         { JE $2 ; yes }
        {0037} $B5 / $00 /         { MOV CH, 0 ; end of leading spaces }
        {0039}                     { $2: }
        {0039} $F7 / $C2 / > $0001 / { TEST DX, 1 ; Trim parity? }
        {003D} $74 / $02 /         { JZ $3 ; no }
        {003F} $8A / $E0 /         { MOV AH, AL ; yes - so do it }
        {0041}                     { $3: }
        {0041} $F7 / $C2 / > $0002 / { TEST DX, 2 ; Remove all spaces/tabs? }
        {0045} $74 / $0C /         { JZ $4 ; no }
        {0047} $3C / $20 /         { CMP AL, 32 ; A space? }
        {0049} $75 / $02 /         { JNE $3A }
        {004B} $B7 / $00 /         { MOV BH, 0 ; Not OK }
        {004D}                     { $3A: }
        {004D} $3C / $09 /         { CMP AL, 9 ; A tab? }
        {004F} $75 / $02 /         { JNE $4 }
        {0051} $B7 / $00 /         { MOV BH, 0 }
        {0053}                     { $4: }
        {0053} $F7 / $C2 / > $0004 / { TEST DX, 4 ; Ignore special values? }
        {0057} $74 / $1E /         { JZ $5 ; no }
        {0059} $3C / $00 /         { CMP AL, 0 }
        {005B} $75 / $02 /         { JNE $4A }
        {005D} $B7 / $00 /         { MOV BH, 0 }
        {005F}                     { $4A: }
        {005F} $3C / $0A /         { CMP AL, 10 }
        {0061} $75 / $02 /         { JNE $4B }
        {0063} $B7 / $00 /         { MOV BH, 0 }
        {0065}                     { $4B: }
        {0065} $3C / $0C /         { CMP AL, 12 }
        {0067} $75 / $02 /         { JNE $4C }
        {0069} $B7 / $00 /         { MOV BH, 0 }
        {006B}                     { $4C: }
        {006B} $3C / $0D /         { CMP AL, 13 }
        {006D} $75 / $02 /         { JNE $4D }
        {006F} $B7 / $00 /         { MOV BH, 0 }
        {0071}                     { $4D: }
        {0071} $3C / $7F /         { CMP AL, 127 }
        {0073} $75 / $02 /         { JNE $5 }
        {0075} $B7 / $00 /         { MOV BH, 0 }
        {0077}                     { $5: }
        {0077} $F7 / $C2 / > $0008 / { TEST DX, 8 ; Ignore leading spaces? }
        {007B} $74 / $07 /         { JZ $6 ; no }
        {007D} $80 / $Fd / $01 /   { CMP CH, 1 ; Currently leading spaces? }
        {0080} $75 / $02 /         { JNE $6 ; no }
        {0082} $B7 / $00 /         { MOV BH, 0 }
        {0084}                     { $6: }
        {0084} $F7 / $C2 / > $0010 / { TEST DX, 16 ; Reduce tabs/spaces to single space? }
        {0088} $74 / $18 /         { JZ $7 ; no }
        {008A} $3C / $20 /         { CMP AL, 32 ; A space? }
        {008C} $74 / $04 /         { JE $6B ; yes }
        {008E} $3C / $09 /         { CMP AL, 9 ; A tab? }
        {0090} $75 / $10 /         { JNE $7 ; no }
        {0092}                     { $6B: }
        {0092} $80 / $Be / Space / $01 / { CMP [BP].Space, BYTE PTR 1 ; Already spaces? }
        {0097} $75 / $02 /         { JNE $6A ; no }
        {0099} $B7 / $00 /         { MOV BH, 0 }
        {009B}                     { $6A: }
        {009B} $B4 / $20 /         { MOV AH, ' ' ; Change tab to space }
        {009D} $C6 / $86 / Space / $01 / { MOV [BP].Space, BYTE PTR 1 }
        {00A2}                     { $7: }
        {00A2} $F7 / $C2 / > $0020 / { TEST DX, 32 ; Lower to upper case? }
        {00A6} $74 / $0D /         { JZ $8 }
        {00A8} $80 / $Fc / $61 /   { CMP AH, 'a' ; Character<'a' }
        {00AB} $7C / $08 /         { JL $8 ; yes }
        {00AD} $80 / $Fc / $7A /   { CMP AH, 'z' ; Character>'z' }
        {00B0} $7F / $03 /         { JG $8 ; yes }
        {00B2} $80 / $Ec / $20 /   { SUB AH, 32 ; Convert to uppercase }
        {00B5}                     { $8: }
        {00B5} $F7 / $C2 / $0200 / { TEST DX, 512 ; Upper to lower case? }
        {00B9} $74 / $0B /         { JZ $9 ; no }
        {00BB} $3C / $41 /         { CMP AL, 'A' ; Character<'A' }
        {00BD} $7C / $07 /         { JL $9 ; yes }
        {00BF} $3C / $5A /         { CMP AL, 'Z' ; Character>'Z' }
        {00C1} $7F / $03 /         { JG $9 ; yes }
        {00C3} $80 / $C4 / $20 /   { ADD AH, 32 ; Convert to lowercase }
        {00C6}                     { $9: }
        {00C6} $F7 / $C2 / > $0040 / { TEST DX, 64 ; Convert [] to () ? }
        {00CA} $74 / $0C /         { JZ $11 ; no }
        {00CC} $3C / $5B /         { CMP AL, '[' }
        {00CE} $75 / $02 /         { JNE $10 }
        {00D0} $B4 / $28 /         { MOV AH, '(' }
        {00D2}                     { $10: }
        {00D2} $3C / $5D /         { CMP AL, ']' }
        {00D4} $75 / $02 /         { JNE $11 }
        {00D6} $B4 / $29 /         { MOV AH, ')' }
        {00D8}                     { $11: }
        {00D8} $F7 / $C2 / $0400 / { TEST DX, 1024 ; Convert [] to braces? }
        {00DC} $74 / $0E /         { JZ $13 ; no }
        {00DE} $3C / $5B /         { CMP AL, '[' }
        {00E0} $75 / $03 /         { JNE $12 }
        {00E2} $80 / $C4 / $20 /   { ADD AH, 32 }
        {00E5}                     { $12: }
        {00E5} $3C / $5D /         { CMP AL, ']' }
        {00E7} $75 / $03 /         { JNE $13 }
        {00E9} $80 / $C4 / $20 /   { ADD AH, 32 }
        {00EC}                     { $13: }
        {00EC} $F7 / $C2 / $0800 / { TEST DX, 2048 ; Convert () to [] ? }
        {00F0} $74 / $0C /         { JZ $15 }
        {00F2} $3C / $28 /         { CMP AL, '(' }
        {00F4} $75 / $02 /         { JNE $14 }
        {00F6} $B4 / $5B /         { MOV AH, '[' }
        {00F8}                     { $14: }
        {00F8} $3C / $29 /         { CMP AL, ')' }
        {00FA} $75 / $02 /         { JNE $15 }
        {00FC} $B4 / $5D /         { MOV AH, ']' }
        {00FE}                     { $15: }
        {00FE} $F7 / $C2 / $1000 / { TEST DX, 4096 ; Convert () to braces? }
        {0102} $74 / $0C /         { JZ $17 ; no }
        {0104} $3C / $28 /         { CMP AL, '(' }
        {0106} $75 / $02 /         { JNE $16 }
        {0108} $B4 / $7B /         { MOV AH, '{' }
        {010A}                     { $16: }
        {010A} $3C / $29 /         { CMP AL, ')' }
        {010C} $75 / $02 /         { JNE $17 }
        {010E} $B4 / $7D /         { MOV AH, '}{' }
        {0110}                     { $17: }
        {0110} $F7 / $C2 / $2000 / { TEST DX, 8192 ; Convert braces to () ? }
        {0114} $74 / $0C /         { JZ $19 ; no }
        {0116} $3C / $7B /         { CMP AL, '{' }
        {0118} $75 / $02 /         { JNE $18 }
        {011A} $B4 / $28 /         { MOV AH, '(' }
        {011C}                     { $18: }
        {011C} $3C / $7D /         { CMP AL, '}{' }
        {011E} $75 / $02 /         { JNE $19 }
        {0120} $B4 / $29 /         { MOV AH, ')' }
        {0122}                     { $19: }
        {0122} $F7 / $C2 / $4000 / { TEST DX, 16384 ; Convert braces to [] ? }
        {0126} $74 / $0E /         { JZ $1 ; no }
        {0128} $3C / $7B /         { CMP AL, '{' }
        {012A} $75 / $03 /         { JNE $20 }
        {012C} $80 / $Ec / $20 /   { SUB AH, 32 }
        {012F}                     { $20: }
        {012F} $3C / $7D /         { CMP AL, '}{' }
        {0131} $75 / $03 /         { JNE $1 }
        {0133} $80 / $Ec / $20 /   { SUB AH, 32 }
        {0136}                     { $1: }
        {0136} $80 / $Ff / $00 /   { CMP BH, 0 ; OK? }
        {0139} $74 / $19 /         { JE $21 ; no }
        {013B} $47 /               { INC DI }
        {013C} $26 / $88 / $25 /   { MOV ES:[DI], AH ; Build result }
        {013F} $3C / $20 /         { CMP AL, 32 ; A space? }
        {0141} $75 / $08 /         { JNE $1A ; no }
        {0143} $C6 / $86 / Space / $01 / { MOV [BP].Space, BYTE PTR 1 }
        {0148} $Eb / $0A / $90 /   { JMP $21 }
        {014B}                     { $1A: }
        {014B} $3C / $09 /         { CMP AL, 9 ; A tab? }
        {014D} $75 / $05 /         { JNE $21 ; no }
        {014F} $C6 / $86 / Space / $01 / { MOV [BP].Space, BYTE PTR 1 }
        {0154}                     { $21: }
        {0154} $3C / $20 /         { CMP AL, 32 ; A space? }
        {0156} $74 / $2B /         { JE $22 ; yup }
        {0158} $3C / $09 /         { CMP AL, 9 ; A tab? }
        {015A} $74 / $27 /         { JE $22 ; yup }
        {015C} $C6 / $86 / Space / $00 / { MOV [BP].Space, BYTE PTR 0 ; end of spaces }
        {0161} $F7 / $C2 / $0100 / { TEST DX, 256 ; Allow alter within quotes? }
        {0165} $74 / $1C /         { JZ $22 ; yes }
        {0167} $3C / $22 /         { CMP AL, 34 ; A " ? }
        {0169} $74 / $04 /         { JE $21A ; yes }
        {016B} $3C / $27 /         { CMP AL, 39 ; A ' ? }
        {016D} $75 / $14 /         { JNE $22 ; no }
        {016F}                     { $21A: }
        {016F} $80 / $FB / $00 /   { CMP BL, 0 ; Within quotes? }
        {0172} $75 / $04 /         { JNE $21B ; yes }
        {0174} $88 / $86 / Quote_Type / { MOV [BP].Quote_Type, AL ; New quote type}
        {0178}                     { $21B: }
        {0178} $38 / $86 / Quote_Type / { CMP [BP].Quote_Type, AL ; Equal quote types? }
        {017C} $75 / $05 /         { JNE $22 ; no }
        {0183} $F6 / $D3 /         { NOT BL }
        {0185} $80 / $E3 / $01 /   { AND BL, 1 ; begin/end of quoted segment }
        {0188}                     { $22: }
        {0188} $3C / $20 /         { CMP AL, 32 ; Space }
        {018A} $74 / $10 /         { JE $23 ; yes }
        {018C} $3C / $09 /         { CMP AL, 9 ; Tab? }
        {018E} $74 / $0C /         { JE $23 ; yes }
        {0190} $8B / $C7 /         { MOV AX, DI }
        {0192} $C4 / $96 / Dst /   { LES DX, [BP].Dst }
        {0196} $2B / $C2 /         { SUB AX, DX ; Get current position in destination }
        {0198} $89 / $86 / Last /  { MOV [BP].Last, AX ; Last non-space character }
        {019C}                     { $23: }
        {019C} $8B / $D6 /         { MOV DX, SI ; Get source address }
        {019E} $53 /               { PUSH BX ; Save BX }
        {019F} $51 /               { PUSH CX ; Save CX }
        {01A0} $B5 / $00 /         { MOV CH, 0 }
        {01A2} $C5 / $9E / Src /   { LDS BX, [BP].Src ; Get source base address }
        {01A6} $2B / $D3 /         { SUB DX, BX ; Get current offset in source }
        {01A8} $3B / $D1 /         { CMP DX, CX ; Are we done? }
        {01AA} $59 /               { POP CX }
        {01AB} $5B /               { POP BX }
        {01AC} $7D / $03 /         { JGE $23A ; yes }
        {01AE} $E9 / $Fe6E /       { R JMP $0 ; no - loop }
        {01B1}                     { $23A: }
        {01B1} $8B / $96 / Flags / { MOV DX, [BP].Flags }
        {01B5} $81 / $E2 / > $0080 / { AND DX, 128 ; Trim following spaces? }
        {01B9} $74 / $0F /         { JZ $24 ; no }
        {01C1} $80 / $FB / $00 /   { CMP BL, 0 ; Within quotes? }
        {01C4} $75 / $0A /         { JNE $24 ; yes }
        {01C6} $8B / $Be / Last /  { MOV DI, [BP].Last ; Get last non-space character position }
        {01CA} $C5 / $9E / Dst /   { LDS BX, [BP].Dst }
        {01CE} $03 / $FB /         { ADD DI, BX ; Convert to address }
        {01D0}                     { $24: }
        {01D0} $C5 / $9E / Dst /   { LDS BX, [BP].Dst }
        {01D4} $8B / $C7 /         { MOV AX, DI ; Get destination address}
        {01D6} $2B / $C3 /         { SUB AX, BX ; Convert destination address to offset }
        {01D8} $88 / $07 /         { MOV [BX], AL ; Set destination string length }
        {01DA} $5D /               { POP BP ; Restore DS and BP }
        {01DB} $1F                 { POP DS }
    )
{$ENDIF}
end ;


function Edit( const Original : string ; Flags : Integer ) : string ;

var Edit_Value : string ;

begin
    Fedit( Edit_Value, Original, Flags ) ; { Do edit }
    Edit := Edit_Value ;
end ;


function Pointer_To_String( P : pointer ) : string ;

begin
    if( P = nil ) then
    begin
        Pointer_To_String := 'nil' ;
    end else
    begin
{$IFDEF DELPHI}
        Pointer_To_String := Num1( cardinal( P ) ) + ' ($' + CVTB( 10, 16, Num1( cardinal( P ) ) ) + ')' ;
{$ELSE}
        Pointer_To_String := Num1( integer( P ) ) + ' ($' +
            CVTB( 10, 16, Num1( integer( P ) and $7FFFFFFF ) ) + ')' ;
{$ENDIF}
    end ;
end ;


{$IFDEF DELPHI}
procedure Set_Length( var X : Small_String ; L : integer ) ; overload ;
{$ELSE}
procedure Set_Length( var X : string ; L : integer ) ;
{$ENDIF}

begin
    X[ 0 ] := chr( L ) ;
end ;


{$IFDEF DELPHI}
procedure Set_Length( var X : string ; L : integer ) ; overload ;

begin
    SetLength( X, L ) ;
end ;
{$ENDIF}

{$IFNDEF DELPHI}
procedure SetLength( X : PChar ; L : integer ) ;

begin
    X[ L ] := #0 ;
end ;
{$ENDIF}


{$IFDEF Test}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


  {$IFDEF DELPHI}
procedure Test_TInteger_List ;

var Test : TInteger_List ;

begin
    Test := TInteger_List.Create ;
    Test.Test_Class ;
end ;
  {$ENDIF}

procedure Test_Boolean_To_String ;

begin
    if( Boolean_To_String( False ) <> 'false' ) then
    begin
        Error ;
    end ;
    if( Boolean_To_String( True ) <> 'true' ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Edit ;

begin
    if( Edit( '', 1 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 1 ) <> '  Hi there "Hi", []{}       ()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 2 ) <> 'Hithere"Hi",[]{}()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 4 ) <> '  Hi there "Hi", []{}       ()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()'#0, 4 ) <> '  Hi there "Hi", []{}       ()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 8 ) <> 'Hi there "Hi", []{}       ()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 16 ) <> ' Hi there "Hi", []{} ()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 32 ) <> '  HI THERE "HI", []{}       ()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 64 ) <> '  Hi there "Hi", (){}       ()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 128 ) <> '  Hi there "Hi", []{}       ()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 256 ) <> '  Hi there "Hi", []{}       ()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 512 ) <> '  hi there "hi", []{}       ()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 1024 ) <> '  Hi there "Hi", {}{}       ()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 2048 ) <> '  Hi there "Hi", []{}       []' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 4096 ) <> '  Hi there "Hi", []{}       {}' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 8192 ) <> '  Hi there "Hi", []()       ()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', 16384 ) <> '  Hi there "Hi", [][]       ()' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "Hi", []{}       ()', -1 ) <> 'HITHERE"Hi",()()[]' ) then
    begin
        Error ;
    end ;
    if( Edit( '  Hi there "H'#39'i, []{}       ', -1 ) <> 'HITHERE"H'#39'i, []{}       ' ) then
    begin
        Error ;
    end ;
    if( Edit( #138#142#154#158#159#192#193#194#195#196#197#221#240#253#255, $20000 or 32768 ) <> 'SZszYAAAAAAYoyy' ) then
    begin
        Error ;
    end ;
    if( Edit( #129#130#144#152#153#154#160#161#162#163#131#132#133#134, $20000 or 65536 ) <> 'ueEyOUaiouaaaa' ) then
    begin
        Error ;
    end ;
    if( Edit( #232#255#154#156, 32 or 32768 ) <> #200#159#138#140 ) then
    begin
        Error ;
    end ;
    if( Edit( #164#135#129#132#134#130#145#148, 32 or 65536 ) <> #165#128#154#142#143#144#146#153 ) then
    begin
        Error ;
    end ;
    if( Edit( #200#159#138#140, 512 or 32768 ) <> #232#255#154#156 ) then
    begin
        Error ;
    end ;
    if( Edit( #165#128#154#142#143#144#146#153, 512 or 65536 ) <> #164#135#129#132#134#130#145#148 ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Pointer_To_String ;

begin
    if( Pointer_To_String( nil ) <> 'nil' ) then
    begin
        Error ;
    end ;
    if( Pointer_To_String( pointer( $12345 ) ) <> '74565 ($12345)' ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Set_Length ;

var S : string ;
    SS : Small_String ;

begin
    S := '1234' ;
    SS := '1234' ;
    Set_Length( S, 2 ) ;
    if( S <> '12' ) then
    begin
        Error ;
    end ;
    Set_Length( SS, 2 ) ;
    if( SS <> '12' ) then
    begin
        Error ;
    end ;
end ;


procedure Test_SetLength ;

var S : string ;

begin
    S := '1234' ;
    SetLength( S, 2 ) ;
    if( S <> '12' ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Unit ;

begin
  {$IFDEF DELPHI}
    Test_TInteger_List ;
  {$ENDIF}
    Test_Boolean_To_String ;
    Test_Edit ;
    Test_Pointer_To_String ;
    Test_Set_Length ;
    Test_SetLength ;
end ;
{$ENDIF}



end.
