{
        Program Name : DB_B3
        Package Name : LB
        Purpose      : Database implemented as a B3 tree
        Institution  : Conroy & Conroy Co.
        Date Written : 11-April-2002
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2002 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit DB_B3 ;

interface

uses Database ; { TDatabase }

const DB_B3Err_Facility = 0 ; { ~~~ }
      DB_B3Err_Success = 0 ;
      DB_B3Err_RecordNotFound = 1 ;

type PB3_Record = ^TB3_Record ;
     TB3_Database = object( TISAM_DataBase )
                          private { Instance data... }
                              Root : PB3_Record ;
                              Last_Record_Index : integer ;

                          public { Constructors and destructors... }
                              constructor Init ;
                              destructor Done ; virtual ;

                          public { API... }
                              function Add( Rec : PDatabase_Record ) : PDatabase_Record ;
                                  virtual ;

                              procedure Delete( Index : longint ;
                                  const Key : string ) ; virtual ;

                              function Get( Index : longint ;
                                  const Key : string ) : PDatabase_Record ;
                                  virtual ;

		      	      function Get_GE( Index : longint ;
                                  const Key : string ) : PDatabase_Record ;
                                  virtual ;

			      function Get_LE( Index : longint ;
                                  const Key : string ) : PDatabase_Record ;
			    	 virtual ;

                              function Next( Index : longint ;
                                  Rec : PDatabase_Record ) : PDatabase_Record ;
                                  virtual ;

                              function Previous( Index : longint ;
                                  Rec : PDatabase_Record ) : PDatabase_Record ;
                                  virtual ;

                              procedure Zero ; virtual ;
                      end ; { TB3_Database }
     PB3_Database = ^TB3_Database ;
     TB3_Record = object( TDatabase_Record )
		      private { Instance data... }
{                          Parent : PB3_Record ; { Our parent key }
{                          Data : boolean ; { True if this is data, false if key }
     		  end ;
     TB3_Key = object( TB3_Record )
                   private { Instance data... }
                       Active : integer ; { Number of active pointers }
                       Pointers : array[ 0..2 ] of PB3_Record ;
                       Keys : array[ 0..1 ] of string ;

                   public
                       constructor Init ;
               end ;
     PB3_Key = ^TB3_Key ;

implementation


{ TB3_Key methods... }

constructor TB3_Key.Init ;

begin
    fillchar( Keys, sizeof( Keys ), 0 ) ;
end ;



{ TB3_Database methods... }

{ Constructors and destructors... }

constructor TB3_Database.Init ;

begin
    inherited Init ;

    Root := nil ;
end ;


destructor TB3_Database.Done ;

begin
    Zero ;

    inherited Done ;
end ;


function TB3_Database.Next( Index : longint ;
    Rec : PDatabase_Record ) : PDatabase_Record ;

begin
    Next := nil ;
end ;


function TB3_Database.Previous( Index : longint ;
    Rec : PDatabase_Record ) : PDatabase_Record ;

begin
    Previous := nil ;
end ;


{ API... }

function TB3_Database.Add( Rec : PDatabase_Record ) : PDatabase_Record ;

begin
    Add := nil ;
    Set_Error( 0, 0 ) ;
end ;


procedure TB3_Database.Delete( Index : longint ; const Key : string ) ;

{var Rec : PDatabase_Record ;  }

begin
{~~~    Rec := Get( Index, Key ) ;}
end ;


function TB3_Database.Get( Index : longint ;
    const Key : string ) : PDatabase_Record ;

var Active : integer ;
    Current : PB3_Key ;
    Loop : integer ;
    This_Key : string ;

begin
    Get := nil ; { Assume failure }
    Last_Record_Index := -1 ;
    if( Root <> nil ) then
    begin
        Current := PB3_Key( Root ) ;
        while( Current <> nil ) do
        begin
            Active := Current^.Active - 1 ;
            if( Active < 1 ) then
            begin
                inc( Active ) ;
            end ;
            for Loop := 0 to Active - 1 do
            begin
(*
                This_Key := Keys[ Loop ] ;
                if( This_Key <= Key ) then
                begin
                    Current := Current^.Pointers[ Loop ] ;
                    break ;
                end ; { if( This_Key <= Key ) }
*)
            end ; { for Loop := 0 to Active - 1 }
            if( This_Key > Key ) then
            begin
                Current := PB3_Key( Current^.Pointers[ Current^.Active - 1 ] ) ;
            end ;
        end ; { while( Current <> nil ) }
    end ; { if( Root <> nil ) }
    Set_Error( 0, DB_B3Err_RecordNotFound ) ;
end ; { TB3_Database.Get }


function TB3_Database.Get_GE( Index : longint ;
    const Key : string ) : PDatabase_Record ;

begin
    Get_GE := nil ;
end ;


function TB3_Database.Get_LE( Index : longint ;
    const Key : string ) : PDatabase_Record ;

begin
    Get_LE := nil ;
end ;


procedure TB3_Database.Zero ;

begin
end ;



end.
