unit EditWin ;

{
        Module Name  : EditWin
        Package Name : LB
        Purpose      : Editor window
        Institution  : Conroy & Conroy Co.
        Date Written : 10-Nov-1992
        Written By   : Alan Conroy
        Version      : 1.3

        Copyright (C) 1992-2005 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TEditor_Window is a Delphi TForm descendant which provides a complete
        editor for Delphi applications.
}

interface

uses // Borland...
     Windows,
     Classes,
     ComCtrls,      // TStatusBar
     Controls,
     Dialogs,
     Forms,         // TForm
     Graphics,
     Menus,         // TMainMenu
     Messages,
     STDCtrls,

     // C&C...
     Editor_Panel ; // TEditor_Panel

type
  TEditor_Window = class( TForm )
    Main_Menu: TMainMenu;
    Edit1: TMenuItem;
    IncludeFile1: TMenuItem;
    N1: TMenuItem;
    Delete_Menu: TMenuItem;
    N2: TMenuItem;
    Paste_Menu: TMenuItem;
    Copy_Menu: TMenuItem;
    Cut_Menu: TMenuItem;
    N3: TMenuItem;
    Undo1: TMenuItem;
    Search1: TMenuItem;
    GotoLineNumber1: TMenuItem;
    N4: TMenuItem;
    FindNext1: TMenuItem;
    Replace1: TMenuItem;
    Find1: TMenuItem;
    Include_Dialog: TOpenDialog;

    procedure FormClose( Sender : TObject ; var Action : TCloseAction ) ;
    procedure FormCreate( Sender: TObject);
    procedure FormResize( Sender: TObject);
    procedure FormDestroy( Sender: TObject);
    procedure FormActivate( Sender: TObject);
    procedure Find1Click(Sender: TObject);
    procedure Replace1Click(Sender: TObject);
    procedure FindNext1Click(Sender: TObject);
    procedure GotoLineNumber1Click(Sender: TObject);
    procedure Undo1Click( Sender : TObject ) ;
    procedure Paste_MenuClick(Sender: TObject);
    procedure Delete_MenuClick(Sender: TObject);
    procedure Cut_MenuClick(Sender: TObject);
    procedure IncludeFile1Click(Sender: TObject);
    procedure Copy_MenuClick(Sender: TObject);

  protected // Private declarations...

    Panel : TEditor_Panel ;
    Horizontal_Scrollbar, Vertical_Scrollbar : TScrollbar ;

    procedure CB_KeyDown( Sender : TObject ; var Key : word ;
        Shift : TShiftState ) ;
    procedure CB_Focus( Sender : TObject ) ;
    procedure VScroll( Sender : TObject ; Scrollcode : TScrollCode ;
        var Scrollpos : integer ) ;
    procedure HScroll( Sender : TObject ; Scrollcode : TScrollCode ;
        var Scrollpos : integer ) ;
    procedure Selection_State_Change( Sender : TObject ) ;

  protected // Property handlers...
      function Get_Column_Number : integer ;
      function Get_Line_Number : integer ;
      function Get_Top_Line : integer ;
      procedure Set_Column_Number( Value : integer ) ;
      procedure Set_Line_Number( Value : integer ) ;
      procedure Set_Top_Line( Value : integer ) ;

  public // Public declarations...
    On_Kill : TNotifyEvent ;
    On_Selection_Change : TNotifyEvent ;
    On_State_Change : TNotifyEvent ;
    OnUpdateMenu : TNotifyEvent ;

    Active_State : boolean ;
    New_File : boolean ;
    StatusBar : TStatusBar ;
    File_Name : string ;

    function Active_Selection : boolean ;
    procedure Bring_Into_View ;
    property Column_Number : integer
             read Get_Column_Number
             write Set_Column_Number ;
    function Copy_Text : string ;
    function Current_Column : integer ;
    function Current_Row : integer ;
    procedure Delete_Text ;
    function Get_Line( X : integer ) : string ;
    procedure Get_Text( T : PChar ) ;
    procedure Include_File ; overload ;
    procedure Include_File( X : string ) ; overload ;
    procedure Insert_Text( T : string ) ;
    property Line_Number : integer
             read Get_Line_Number
             write Set_Line_Number ;
    function Max_Line : integer ;
    function Modified : boolean ;
    procedure Notify( X : integer ) ;
    procedure Open_File ; overload ;
    procedure Open_File( const S : string ) ; overload ;
    procedure Print ;
    procedure Save_File ;
    function Search( X : string ) : boolean ;
    procedure Select_Line( X : integer ) ;
    procedure Set_Focus( var Message : tWMSetFocus ) ; message wm_SetFocus ;
    procedure Set_Preferences ;
    function Text_Size : integer ;
    property Top_Line : integer
             read Get_Top_Line
             write Set_Top_Line ;
    procedure Undo ;
  end ;

implementation

uses // Borland
     Buttons, // bkCancel
     Clipbrd,  // Clipboard
     Printers, // assignprn
     SysUtils, // inttostr

     // Conroy & Conroy
     ASCIIDef,  // CRLF
     SMU,       // Create_Window_Handle
     Standard, // Clear_IOResult
     CommonUt,  // Edit
     UEditor, SSearch, SReplace, Linenumb, SaveEdit, Editpref ;

{$R *.DFM}

var Print_Dialog : TPrintDialog = nil ;

// TEditor_Window methods...

// Component delegation...

procedure TEditor_Window.Find1Click( Sender : TObject ) ;

var Saved_Row, Saved_Column : integer ;
    X : string ;

begin
    if( Search_Dialog.ShowModal = mrOK ) then
    begin
        Saved_Row := Line_Number ;
        Saved_Column := Column_Number ;
        if( Search_Dialog.Search_Entire.Checked ) then
        begin
            Line_Number := 1 ;
            Column_Number := 1 ;
        end ;
        X := Search_Dialog.Edit1.Text ;
        if( not Search_Dialog.Case_Sensitive.Checked ) then
        begin
            X := Edit( X, 32 ) ;
        end ;
        if( not Search( X ) ) then // Text not found
        begin
            // Restore our original position
            if( Search_Dialog.Search_Entire.Checked ) then
            begin
                Line_Number := Saved_Row ;
                Column_Number := Saved_Column ;
            end ;

            // Let the user know...
            X := 'Search string "' + Search_Dialog.Edit1.Text + '" not found' ;
            MessageBox( 0, PChar( X ), 'Search', MB_ICONINFORMATION or MB_OK ) ;
        end ;
    end ;
end ; // TEditor_Window.Find1Click


procedure TEditor_Window.Replace1Click( Sender : TObject ) ;

var Found : boolean ;
    Saved_Row, Saved_Column : integer ;
    X : string ;

begin
    if( Replace_Dialog.ShowModal = mrOK ) then
    begin
        Found := False ;
        Saved_Row := Line_Number ;
        Saved_Column := Column_Number ;
        if( Replace_Dialog.Search_Entire.Checked ) then
        begin
            Line_Number := 1 ;
            Column_Number := 1 ;
        end ;
        X := Replace_Dialog.Edit1.Text ;
        if( not Replace_Dialog.Case_Sensitive.Checked ) then
        begin
            X := Edit( X, 32 ) ;
        end ;
        while( true ) do
        begin
            if( not Search( X ) ) then // Text not found
            begin
                // Restore our original position
                if( Replace_Dialog.Search_Entire.Checked ) then
                begin
                    Line_Number := Saved_Row ;
                    Column_Number := Saved_Column ;
                end ;

                // Let the user know...
                if( not Found ) then
                begin
                    X := 'Search string "' + Replace_Dialog.Edit1.Text + '" not found' ;
                    MessageBox( 0, PChar( X ), 'Replace', MB_ICONINFORMATION or MB_OK ) ;
                end ;
                exit ;
            end else
            begin
                Found := True ;
                if( Replace_Dialog.Replace_Prompt.Checked ) then
                begin
                    Panel.Repaint ;
                    case Application.MessageBox( 'Replace this instance', 'Replace', mb_YesNoCancel ) of
                        IDCANCEL : exit ;
                        IDNO : continue ;
                    end ;
                end ;
                Delete_Text ;
                Insert_Text( Replace_Dialog.Edit2.Text ) ;
            end ; // if( not Search( X ) )
        end ; // while( true )
    end ; // if( Replace_Dialog.ShowModal = mrOK )
end ; // TEditor_Window.Replace1Click


procedure TEditor_Window.FindNext1Click( Sender : TObject ) ;

var X : string ;

begin
    X := Search_Dialog.Edit1.Text ;
    if( not Search_Dialog.Case_Sensitive.Checked ) then
    begin
        X := Edit( X, 32 ) ;
    end ;
    if( not Search( X ) ) then
    begin
        X := 'Search string "' + Search_Dialog.Edit1.Text + '" not found' ;
        MessageBox( 0, PChar( X ), 'Search', MB_ICONINFORMATION or MB_OK ) ;
    end ;
end ;


procedure TEditor_Window.GotoLineNumber1Click( Sender : TObject ) ;

begin
    Line_Number_Dialog.Line_Number_Edit.Text := inttostr( Line_Number ) ;
    if( Line_Number_Dialog.ShowModal = mrOK ) then
    begin
        Line_Number := Line_Number_Dialog.Line_Number ;
    end ;
end ;


procedure TEditor_Window.Undo1Click( Sender : TObject ) ;

begin
    Undo ;
end ;


procedure TEditor_Window.Paste_MenuClick( Sender : TObject ) ;

begin
    Insert_Text( Clipboard.AsText ) ;
end ;


procedure TEditor_Window.Delete_MenuClick( Sender : TObject ) ;

begin
    Delete_Text ;
end ;


procedure TEditor_Window.Cut_MenuClick( Sender : TObject ) ;

begin
    Copy_MenuClick( Sender ) ;
    Delete_Text ;
end ;


procedure TEditor_Window.Copy_MenuClick( Sender : TObject ) ;

begin
    Clipboard.AsText := Copy_Text ;
    if( assigned( OnUpdateMenu ) ) then
    begin
        OnUpdateMenu( self ) ;
    end ;
//    UpdateMenuItems( Application.MainForm ) ;
end ;


// Property handlers...

function TEditor_Window.Get_Column_Number : integer ;

begin
    Result := Panel.Editor^.Cursor_Column ;
end ;


function TEditor_Window.Get_Line_Number : integer ;

begin
    Get_Line_Number := Panel.Editor^.Screen_Line + Panel.Editor^.Cursor_Row - 1 ;
end ;


function TEditor_Window.Get_Top_Line : integer ;

begin
    Result := Panel.Editor^.Screen_Line ;
end ;


procedure TEditor_Window.Set_Column_Number( Value : integer ) ;

begin
    Panel.Editor^.Set_Cursor_Column( Value ) ;
end ;


procedure TEditor_Window.Set_Line_Number( Value : integer ) ;

begin
    if( Value > Panel.Editor^.File_Root^.Last_Line ) then
    begin
        Value := Panel.Editor^.File_Root^.Last_Line ;
    end ;
    Panel.Position_Cursor( Value, 1 ) ;
    Panel.Bring_Into_View ;
end ;


procedure TEditor_Window.Set_Top_Line( Value : integer ) ;

begin
    if( Value > Panel.Editor^.File_Root^.Last_Line ) then
    begin
        Value := Panel.Editor^.File_Root^.Last_Line ;
    end ;
    Panel.Editor^.Screen_Line := Value ;
end ;


// Public API...

function TEditor_Window.Active_Selection : boolean ;

begin
    Active_Selection := Panel.Editor^.Selection ;
end ;


procedure TEditor_Window.Bring_Into_View ;

begin
    Panel.Bring_Into_View ;
end ;


function TEditor_Window.Copy_Text : string ;

var Dummy : integer ; // Loop control
    First_Line, Last_Line : longint ;
    First_Column, Last_Column : integer ;

begin
    Panel.Editor^.Get_Selection_Range( First_Line, Last_Line,
        First_Column, Last_Column ) ; // Get range
    if( First_Line = Last_Line ) then // Selection within same line
    begin
        Copy_Text := copy( Panel.Editor^.Fetch_Line( First_Line ), First_Column,
            Last_Column - First_Column + 1 ) ;
        exit ;
    end ;

    // Do copy
    Copy_Text := copy( Panel.Editor^.Fetch_Line( First_Line ), First_Column, 255 ) ;
    if( First_Line <> Last_Line ) then
    begin
        Result := Result + CRLF ;
    end ;
    for Dummy := First_Line + 1 to Last_Line - 1 do
    begin
        Result := Result + Panel.Editor^.Fetch_Line( Dummy ) + CRLF ;
    end ;
    Result := Result + copy( Panel.Editor^.Fetch_Line( Last_Line ), 1, Last_Column ) ;
end ; // TEditor_Window.Copy_Text


function TEditor_Window.Current_Column : integer ;

begin
    Result := Panel.Editor.Cursor_Column ;
end ;


function TEditor_Window.Current_Row : integer ;

begin
    Result := Panel.Editor.Screen_Line + Panel.Editor.Cursor_Row - 1 ;
end ;


procedure TEditor_Window.Delete_Text ;

begin
    Panel.Delete_Text ;
end ;


function TEditor_Window.Get_Line( X : integer ) : string ;

begin
    Result := Panel.Editor.Fetch_Line( X ) ;
end ;


procedure TEditor_Window.Get_Text( T : PChar ) ;

begin
    Panel.Editor^.Get_Text( T ) ;
end ; { TEditor_Window.Get_Text }


procedure TEditor_Window.Include_File ;

begin
    if( Include_Dialog.Execute ) then
    begin
        Include_File( Include_Dialog.Filename ) ;
    end ;
end ;


procedure TEditor_Window.Include_File( X : string ) ;

var F : TextFile ;
    T1_Pointer : pLine_Buffer ;

begin
    T1_Pointer := new( pLine_Buffer, Init ) ;
    assignFile( F, X ) ;
    reset( F ) ;
    while( not eof( F ) ) do
    begin
        readln( F, X ) ;
        T1_Pointer.Append( X ) ;
    end ;
    closeFile( F ) ;
    T1_Pointer.Append( '' ) ;
    Panel.Insert_Buffer( T1_Pointer ) ;
    dispose( T1_Pointer, Done ) ;
end ;


procedure TEditor_Window.Insert_Text( T : string ) ;

var Dummy, Start : integer ;
    T1_Pointer : pLine_Buffer ;

begin
    Panel.Editor^.Do_Update ; { Update current line }
    T1_Pointer := new( pLine_Buffer, Init ) ;
    Dummy := 1 ;
    while( Dummy <= length( T ) ) do
    begin
        Start := Dummy ;
        while( ( Dummy <= length( T ) ) and ( T[ Dummy ] <> #13 ) ) do
        begin
            inc( Dummy ) ;
        end ;
        T1_Pointer.Append( copy( T, Start, Dummy - Start ) ) ;
        Dummy := Dummy + 2 ; { Skip past CRLF }
    end ;
    if( Dummy = length( T ) + 1 ) then { Line ended in CRLF }
    begin
        T1_Pointer.Append( '' ) ;
    end ;
    if( Panel.Editor^.File_Root^.Last_Line <> 0 ) then
    { Pasting into existing text}
    begin
        Panel.Editor^.Paste( T1_Pointer ) ; { Append to existing buffer }
        dispose( T1_Pointer, Done ) ;
    end else
    begin
        dispose( Panel.Editor^.File_Root, Done ) ; { Becomes new buffer }
        Panel.Editor^.Set_Buffer( T1_Pointer ) ;
    end ;

    Panel.Editor^.Redraw( 1 ) ;
end ; { TEditor_Window.Insert_Text }


function TEditor_Window.Max_Line : integer ;

begin
    Result := Panel.Editor^.File_Root^.Last_Line ;
end ;


function TEditor_Window.Modified : boolean ;

begin
    try
        Result := ( Panel.Editor^.Any_Changes or Panel.Editor^.Changes ) ;
    except
        Result := False ;
    end ;
end ;


procedure TEditor_Window.Notify( X : integer ) ;

var Y : TObject ;

begin
    if( Panel.Editor^.Something_Selected ) then
    begin
        Y := self ;
    end else
    begin
        Y := nil ;
    end ;
    Selection_State_Change( Y ) ;
    if( ( Panel.Undo_List = nil ) or ( Panel.Undo_List.Count = 0 ) ) then
    begin
        Undo1.Enabled := False ;
    end else
    begin
        Undo1.Enabled := True ;
    end ;
    if( ( X = edn_LineCount ) or ( X = edn_Scroll ) or ( X = 0 ) ) then // Line count or position changed
    begin
        // Handle vertical scrollbar...
        Vertical_ScrollBar.Max := Panel.Editor^.File_Root^.Last_Line + 1 ;
        Vertical_ScrollBar.Position := Panel.Editor^.Screen_Line ;
        if(
            ( Panel.Editor^.Screen_Line > 1 )
            or
            ( Panel.Editor^.File_Root^.Last_Line > Panel.Editor^.Screen_Window^.Rows )
          ) then
        begin
            Vertical_ScrollBar.Visible := True ;
        end else
        begin
            Vertical_ScrollBar.Visible := False ;
        end ;

        // Handle horizontal scrollbar
        Horizontal_Scrollbar.Position := Panel.Editor^.Screen_Column ;
    end ;
    if( assigned( On_State_Change ) ) then
    begin
        On_State_Change( self ) ;
    end ;
end ; // TEditor_Window.Notify


procedure TEditor_Window.Open_File ;

begin
    Open_File( File_Name ) ;
end ;


procedure TEditor_Window.Open_File( const S : string ) ;

begin
    if( Panel.Open( S ) ) then
    begin
        File_Name := S ;
    end ;
    Notify( 0 ) ; // Setup scrollbars, status bar, etc.
    New_File := False ;
end ;


procedure TEditor_Window.Print ;

var Dummy : integer ;
    F : TFont ;
    Handle : integer ;
    LPP : integer ;
    M : integer ;
    P : textfile ;

begin
    LPP := 66 ;
    if( Print_Dialog = nil ) then
    begin
        Print_Dialog := TPrintDialog.Create( Application ) ;
    end ;
    Print_Dialog.MaxPage := ( ( Max_Line - 1 ) div LPP ) + 1 ;
    if( Print_Dialog.Execute ) then
    begin
        AssignPrn( P ) ;
        Rewrite( P ) ;
        Handle := SelectObject( Panel.Editor^.Screen_Window^.ID, Font.Handle ) ;
        SelectObject( Panel.Editor^.Screen_Window^.ID, Handle ) ;
        F := TFont.Create ;
        F.Handle := Handle ;
        Printer.Canvas.Font := F ;
        M := Print_Dialog.ToPage * LPP ;
        if( M > Max_Line ) then
        begin
            M := Max_Line ;
        end ;
        for Dummy := ( Print_Dialog.FromPage - 1 ) * LPP + 1 to M do
        begin
            Writeln( P, Get_Line( Dummy ) ) ;
        end ;
        {$I-}
        CloseFile( P ) ;
        {$I+}
        Clear_IOResult ;
    end ;
end ;


procedure TEditor_Window.Save_File ;

var Dummy : integer ;
    F : TextFile ;
    S : String ;

begin
    Panel.Editor^.Do_Update ;
    Panel.Editor^.Any_Changes := False ;
    Panel.Editor^.Changes := False ;
    Panel.Undo_List.Zero ;
    assignFile( F, File_Name ) ;
    rewrite( F ) ;
    for Dummy := 1 to Panel.Editor^.File_Root.Last_Line do
    begin
        S := Edit( Panel.Editor^.Fetch_Line( Dummy ), 128 ) ;
        writeln( F, S ) ;
    end ;
    {$I-}
    closeFile( F ) ;
    {$I+}
    Clear_IOResult ;
    Notify( 0 ) ;
end ;


function TEditor_Window.Text_Size : integer ;

begin
    Text_Size := Panel.Editor^.Text_Size ;
end ; { TEditor_Window.Text_Size }


procedure TEditor_Window.FormClose( Sender : TObject ; var Action : TCloseAction ) ;

begin
    if( Panel.Editor^.Any_Changes ) then
    begin
        Save_Modified_Dialog.Label1.Caption := 'Save changes to ' + File_Name + '?' ;
        case Save_Modified_Dialog.ShowModal of
            mrCancel :
                begin
                    Action := caNone ;
                    exit ;
                end ;
            mrYes :
                begin
                    Save_File ;
                end ;
        end ;
    end ;
    Action := caFree ;
    Active_State := False ;
end ;


procedure TEditor_Window.FormCreate( Sender : TObject ) ;

begin
    // Setup...
    ControlStyle := ControlStyle + [ csOpaque ] ;
    Panel := nil ;
    File_Name := 'Noname' ;

    // Create vertical scrollbar...
    Vertical_ScrollBar := TScrollBar.Create( self ) ;
    Vertical_ScrollBar.Parent := self ;
    Vertical_ScrollBar.Visible := False ;
    Vertical_ScrollBar.Align := alRight ;
    Vertical_ScrollBar.Enabled := True ;
    Vertical_ScrollBar.Kind := sbVertical ;
    Vertical_ScrollBar.TabStop := False ;
    Vertical_ScrollBar.Min := 1 ;
    Vertical_ScrollBar.OnScroll := VScroll ;

    // Set up horizontal scrollbar
    Horizontal_ScrollBar := TScrollBar.Create( self ) ;
    Horizontal_ScrollBar.Parent := self ;
    Horizontal_ScrollBar.Visible := True ;
    Horizontal_ScrollBar.Align := alBottom ;
    Horizontal_ScrollBar.Enabled := True ;
    Horizontal_ScrollBar.Kind := sbHorizontal ;
    Horizontal_ScrollBar.Min := 1 ;
    Horizontal_ScrollBar.Max := 255 ;
    Horizontal_ScrollBar.TabStop := False ;
    Horizontal_ScrollBar.OnScroll := HScroll ;

    // Create panel...
    Panel := TEditor_Panel.Create( self ) ;
    Panel.Parent := self ;
    Panel.Align := alClient ;
    Panel._OnKeyDown := CB_KeyDown ;
    Panel.On_Focus := CB_Focus ;
    Panel.BevelInner := bvNone ;
    Panel.BevelOuter := bvNone ;
    Panel.TabStop := True ;
    Set_Preferences ;
    Panel.Editor^.Screen_Window := Create_Window_Wrapper( Handle ) ;
    Panel.Editor^.Screen_Window^.Flags :=
        Panel.Editor^.Screen_Window^.Flags and ( not 2 ) or 4 ;
    FormResize( self ) ;

    // Misc other setup...
    Panel.Editor^.Notification_Routine1 := Notify ;
    Active_State := True ;
    New_File := True ;
end ; // TEditor_Window.FormCreate


procedure TEditor_Window.FormResize( Sender : TObject ) ;

var SI : TScrollInfo ;

begin
    Panel.Editor^.Calculate_Rows( Panel.Height ) ;
    Horizontal_Scrollbar.LargeChange := Panel.Editor^.Screen_Window^.Columns ;
    Vertical_Scrollbar.LargeChange := Panel.Editor^.Screen_Window^.Rows ;
    SI.cbSize := sizeof( SI ) ;
    SI.fMask := SIF_PAGE ;
    SI.nPage := Panel.Editor^.Screen_Window^.Rows ;
    SetScrollInfo( Vertical_Scrollbar.Handle, SB_CTL, SI, True ) ;
    SI.nPage := Panel.Editor^.Screen_Window^.Columns ;
    SetScrollInfo( Horizontal_Scrollbar.Handle, SB_CTL, SI, True ) ;
    Panel.Editor^.Redraw( 1 ) ;
end ;


procedure TEditor_Window.Set_Preferences ;

begin
    Panel.Editor^.Options := 0 ;
    if( Editor_Preferences.Auto_Indent.Checked ) then
    begin
        Panel.Editor^.Options := Panel.Editor^.Options or eds_AutoIndent ;
    end ;
    if( Editor_Preferences.Bounded_Caret.Checked ) then
    begin
        Panel.Editor^.Options := Panel.Editor^.Options or eds_CursorBound ;
    end ;
    if( Editor_Preferences.Sub_Window.Checked ) then
    begin
        Panel.Editor^.Options := Panel.Editor^.Options or eds_SubWindow ;
    end ;
    if( not Editor_Preferences.Tab_Processing.Checked ) then
    begin
        Panel.Editor^.Options := Panel.Editor^.Options or eds_NoTabs ;
    end ;
    if( not Editor_Preferences.Wrap_Caret.Checked ) then
    begin
        Panel.Editor^.Options := Panel.Editor^.Options or eds_NoCursorWrap ;
    end ;
end ;


function TEditor_Window.Search( X : string ) : boolean ;

var Dummy : integer ;

begin
    Search := Panel.Search( X, Search_Dialog.Case_Sensitive.Checked ) ;
    if( Result ) then
    begin
        Panel.Editor^.Unselect ;
        Panel.Editor^.Select ;
        Dummy := length( X ) ;
        while( Dummy > 0 ) do
        begin
            Panel.Cursor_Right ;
            dec( Dummy ) ;
        end ;
        Panel.Editor^.Freeze_Selection ;
        Set_Cursor( Panel.Editor^.Cursor_Row,
            Panel.Editor^.Cursor_Column - Panel.Editor^.Screen_Column + 1 ) ;
        if( Panel.Editor^.Cursor_Row > Panel.ClientHeight div Panel.Editor^.Text_Height ) then // Last line may not be fully visible
        begin
            // Scroll one line
            if( Panel.Editor^.Screen_Line < Panel.Editor^.File_Root^.Last_Line ) then
            begin
                inc( Panel.Editor^.Screen_Line ) ;
                dec( Panel.Editor^.Cursor_Row ) ;
                Panel.Invalidate ;
            end ;
        end ;
    end ;
end ; // Code_Window.Search


procedure TEditor_Window.FormDestroy( Sender : TObject ) ;

begin
    if( assigned( On_Kill ) ) then
    begin
        On_Kill( self ) ;
    end ;
end ;


procedure TEditor_Window.FormActivate( Sender : TObject ) ;

begin
    Panel.Editor^.Notification_Routine1 := Notify ;
    Notify( edn_Modify ) ;
end ;


procedure TEditor_Window.CB_Focus( Sender : TObject ) ;

begin
    Notify( 0 ) ;
end ;


procedure TEditor_Window.CB_KeyDown( Sender : TObject ; var Key : word ;
    Shift : TShiftState ) ;

begin
    if( assigned( OnKeyDown ) ) then
    begin
        OnKeyDown( Sender, Key, Shift ) ;
    end ;
end ;


procedure TEditor_Window.VScroll( Sender : TObject ; Scrollcode : TScrollCode ;
    var Scrollpos : integer ) ;

begin
    case Scrollcode of
        scLineUp:
            begin
                if( Panel.Editor^.Screen_Line > 1 ) then
                begin
                    dec( Panel.Editor^.Screen_Line ) ;
                    inc( Panel.Editor^.Cursor_Row ) ;
                    Panel.Invalidate ;
                end ;
            end ;
        scLineDown:
            begin
                if( Panel.Editor^.Screen_Line < Panel.Editor^.File_Root^.Last_Line ) then
                begin
                    inc( Panel.Editor^.Screen_Line ) ;
                    dec( Panel.Editor^.Cursor_Row ) ;
                    Panel.Invalidate ;
                end ;
            end ;
        scPageUp:
            begin
                Panel.Editor^.Screen_Line := Panel.Editor^.Screen_Line -
                    Panel.Editor^.Screen_Window^.Rows ;
                Panel.Editor^.Cursor_Row := Panel.Editor^.Cursor_Row +
                    Panel.Editor^.Screen_Window^.Rows ;
                if( Panel.Editor^.Screen_Line < 1 ) then
                begin
                    Panel.Editor^.Cursor_Row := Panel.Editor^.Cursor_Row +
                        Panel.Editor^.Screen_Line - 1 ;
                    Panel.Editor^.Screen_Line := 1 ;
                end ;
                Panel.Invalidate ;
            end ;
        scPageDown:
            begin
                Panel.Editor^.Screen_Line := Panel.Editor^.Screen_Line +
                    Panel.Editor^.Screen_Window^.Rows ;
                Panel.Editor^.Cursor_Row := Panel.Editor^.Cursor_Row -
                    Panel.Editor^.Screen_Window^.Rows ;
                if( Panel.Editor^.Screen_Line > Panel.Editor^.File_Root^.Last_Line ) then
                begin
                    Panel.Editor^.Cursor_Row := Panel.Editor^.Cursor_Row -
                        ( Panel.Editor^.File_Root^.Last_Line - Panel.Editor^.Screen_Line ) ;
                    Panel.Editor^.Screen_Line :=
                        Panel.Editor^.File_Root^.Last_Line ;
                end ;
                Panel.Invalidate ;
            end ;
        scTop:
            begin
                Panel.Editor^.Top ;
            end ;
        scBottom:
            begin
                Panel.Editor^.Bottom ;
            end ;
        scPosition:
            begin
                Panel.Editor^.Cursor_Row := Panel.Editor^.Cursor_Row -
                    ( ScrollPos - Panel.Editor^.Screen_Line ) ;
                Panel.Editor^.Screen_Line := ScrollPos ;
                Panel.Invalidate ;
            end ;
        else exit ;
    end ;
end ;


procedure TEditor_Window.HScroll( Sender : TObject ; Scrollcode : TScrollCode ;
    var Scrollpos : integer ) ;

begin
    case Scrollcode of
        scLineUp: // Left
            begin
                Panel.Editor^.Scroll_Columns( -1 ) ;
            end ;
        scLineDown: // Right
            begin
                Panel.Editor^.Scroll_Columns( 1 ) ;
            end ;
        scPageUp: // Left
            begin
                Panel.Editor^.Scroll_Columns( -Panel.Editor^.Screen_Window^.Columns ) ;
            end ;
        scPageDown: // Right
            begin
                Panel.Editor^.Scroll_Columns( Panel.Editor^.Screen_Window^.Columns ) ;
            end ;
        scTop: // Left
            begin
                Panel.Editor^.Screen_Column := 1 ;
                Panel.Invalidate ;
            end ;
        scBottom: // Right
            begin
                Panel.Editor^.Screen_Column := 255 - Panel.Editor^.Screen_Window^.Columns ;
                Panel.Invalidate ;
            end ;
        scPosition:
            begin
                Panel.Editor^.Screen_Column := ScrollPos ;
                Panel.Invalidate ;
            end ;
        else exit ;
    end ;
end ;


procedure TEditor_Window.Select_Line( X : integer ) ;

begin
    Panel.Editor^.Unselect ;
    Panel.Editor^.Select_Text( X, 1, X, 255 ) ;
    Panel.Editor^.Position_Cursor( X, 1 ) ;
end ;


procedure TEditor_Window.Set_Focus( var Message : tWMSetFocus ) ;

begin
    inherited ;

    Notify( 0 ) ;
    Panel.SetFocus ;
    BringToFront ;
end ;


procedure TEditor_Window.Selection_State_Change( Sender : TObject ) ;

begin
    // Change menu...
    Cut_Menu.Enabled := Sender <> nil ;
    Copy_Menu.Enabled := Sender <> nil ;
    Delete_Menu.Enabled := Sender <> nil ;

    // Synchronize buttons...
    if( assigned( On_Selection_Change ) ) then
    begin
        On_Selection_Change( self ) ;
    end ;
//    Cut_Button.Enabled := Cut_Menu.Enabled ;
//    Erase_Button.Enabled := Delete_Menu.Enabled ;
end ;


procedure TEditor_Window.Undo ;

begin
    Panel.Undo ;

    // Check to see if everything was undone
    if( ( Panel.Undo_List = nil ) or ( Panel.Undo_List.Count = 0 ) ) then
    begin
        Undo1.Enabled := False ;
        Panel.Editor^.Any_Changes := False ;
        Panel.Editor^.Changes := False ;
        Notify( 0 ) ; // Clear "Modified"
    end ;
end ;


procedure TEditor_Window.IncludeFile1Click( Sender : TObject ) ;

begin
    Include_File ;
end ;



end.
