{
        Program Name : HTML
        Package Name : LB
        Purpose      : HTML support
        Institution  : Conroy & Conroy Co.
        Date Written : 3-April-2003
        Written By   : Alan Conroy
        Version      : 1.0F

        Copyright (C) 2003-2006 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  6-Jan-2004    EAC         Added Max_X to Process_HTML.   *
        * 25-Jun-2004                Fix problem with translated    *
        *                            HTML text output.              *
        * 12-Jan-2004    EAC         Added XML support.             *
        *  2-Feb-2005    EAC         Added HTML_To_Text.            *
        *  7-Mar-2005    EAC         Added strikethrough support.   *
        *  7-Mar-2006    EAC         Fixed infinite loop.           *
        * 16-Mar-2006    NLR         Fixed infinite loop.           *
        * 10-Dec-2006    EAC         Fixed FONT tag parsing.        *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          HTML support routines.
}

unit HTML ;

interface

{$I EDEFINES.INC}

uses { Borland... }
     Windows, { HDC }
{$IFDEF DELPHI}
     Classes, {  TStringList }
     Graphics, { TFont }
{$ENDIF}

     { C&C... }
{$IFDEF DELPHI}
     ComInter, { TCommon_COM_Interface }
{$ENDIF}
     Parse, { TParser }
     UStrings ; { TASCII_String }

function HTML_ERT( X : integer ) : string ; { Return text of HTML error code }
function From_HTML( X : string ; Options : longint ) : string ;
{ Convert HTML to ASCII text.  Options is processing flags: and 1 = keep unknown entity references in string }

function To_HTML( const X : string ) : string ; { Convert ASCII text to HTML }
function HTML_To_Text( X : string ) : string ; { Convert HTML to plain text }

{$IFDEF Delphi}
function Process_HTML_Text( DC : HDC ; Font : TFont ; const S : string ;
     Draw : boolean ; X, Y, Max_X : integer ; var Max_Height : integer ;
     Single_Line : boolean ) : integer ;
{$ENDIF}

type TSHTML_Parser = object( TSParser )
                        public { Constructors and destructors... }
                            constructor Init ;
                            destructor Done ; virtual ;

{$IFDEF DELPHI}
                        protected { Internal utility routines... }
{$ENDIF}
                            function Get_Token : PString ; virtual ;
                            function _Get_Section( S : string ;
                                XML : boolean ) : PString ;

                        public { API... }
                            Preserve_Whitespace : boolean ;
                            Source : TASCII_String ;

                            function EOF : boolean ;
                            function Get_Section( const S : string ) : PString ;
                                virtual ;
                    end ;
     PSHTML_Parser = ^TSHTML_Parser ;

type TSXML_Parser = object( TSHTML_Parser )
                        public { API... }
                            function Get_Section( const S : string ) : PString ;
                                virtual ;
                            function Get_Section_String( const S : string ) : string ;
                            function Get_Token_String : string ;
                    end ;
     PSXML_Parser = ^TSXML_Parser ;

{$IFDEF DELPHI}
type THTML_Parser = class( TParser )
                        public { Constructors and destructors... }
                            constructor Create ;

                        protected { Internal utility routines... }
                            function Get_Source : string ;
                            procedure Set_Source( const S : string ) ;
                            function Get_Preserve_Whitespace : boolean ;
                            procedure Set_Preserve_Whitespace( Value : boolean ) ;

                            function _Get_Section( const S : string ;
                                XML : boolean ) : string ; virtual ;
                                
                        public { API... }
                            function EOF : boolean ;

                            property Preserve_Whitespace : boolean
                                read Get_Preserve_Whitespace
                                write Set_Preserve_Whitespace ;

                            property Source : string
                                read Get_Source
                                write Set_Source ;

                            function Get_Section( const S : string ) : string ;
                                virtual ;
                    end ;

     THTML_Tag = class
                     private
                         Names, Values : TStringList ;

                     public
                         Name : String ;

                         constructor Create( S : string ) ;
                         destructor Destroy ; override ;

                         function Attribute_Name( Index : integer ) : string ;
                         function Attribute_Value( Index : integer ) : String ;
                         function Count : integer ;
                 end ;

type TXML_Parser = class( TCommon_COM_Interface )
                        public { Constructors and destructors... }
                            constructor Create ;
                            destructor Destroy ; override ;

                        private { Instance data }
                            _Parser : THTML_Parser ;
                            Temp : string ;

                        public { API... }
                            function Get_Current_Line : longint ;
                                virtual ; stdcall ;
                            function Get_Non_Terminals : PChar ;
                                virtual ; stdcall ;
                            procedure Set_Non_Terminals( S : PChar ) ;
                                virtual ; stdcall ;
                            function Get_Preserve_Whitespace : boolean ;
                                virtual ; stdcall ;
                            procedure Set_Preserve_Whitespace( Value : boolean ) ;
                                virtual ; stdcall ;
                            function Get_Source : PChar ; virtual ; stdcall ;
                            procedure Set_Source( S : PChar ) ;
                                virtual ; stdcall ;
                            function Get_Token_Line : integer ;
                                virtual ; stdcall ;
                            procedure Set_Token_Line( V : integer ) ;
                                virtual ; stdcall ;

                            procedure Add_Special_Token( S : PChar ) ;
                                virtual ; stdcall ;
                            function EOF : boolean ; virtual ; stdcall ;
                            function Get_Section( S : PChar ) : PChar ;
                                virtual ; stdcall ;
                            function Get_Token : PChar ; virtual ; stdcall ;
                            function Previous_Token_Line : Integer ;
                                virtual ; stdcall ;
                            procedure Push_Token_Line ; virtual ; stdcall ;
                            function Read_Next_Line : boolean ;
                                virtual ; stdcall ;
                            function Grab_Line : PChar ; virtual ; stdcall ;
                            function Last_Error : longint ; virtual ; stdcall ;
                            procedure Put_Token( S : PChar ) ;
                                virtual ; stdcall ;
                            function Token : PChar ; virtual ; stdcall ;
                            function Token_EOL : boolean ; virtual ; stdcall ;
                    end ; { TXML_Parser }

     TXML_Tag = class( TCommon_COM_Interface )
                     public { Constructors and destructors }
                         constructor Create( const S : string ) ;
                         destructor Destroy ; override ;

                     private { Instance data... }
                         _Tag : THTML_Tag ;
                         Temp : string ;

                     public { API... }
                         function Attribute_Name( Index : integer ) : Pchar ;
                             virtual ; stdcall ;
                         function Attribute_Value( Index : integer ) : PChar ;
                             virtual ; stdcall ;
                         function Count : integer ; virtual ; stdcall ;
                         function Get_Name : PChar ; virtual ; stdcall ;
                         procedure Set_Name( P : PChar ) ; virtual ; stdcall ;
                 end ;
{$ENDIF}

{$IFDEF Test}
procedure Unit_Test ;
{$ENDIF}

function Extract_Tag_ID( const S : string ) : string ;

implementation

uses { Borland... }
{$IFDEF Delphi}
     SysUtils, { MemAlloc }
{$ENDIF}

     { C&C... }
     ASCIIDef, { CR }
     Instrs, { Instr }
     Num1s, { Num1 }
     CommonUt ; { Edit }

function HTML_ERT( X : integer ) : string ;

begin
    case X of
        204 : HTML_ERT := 'No Content' ;
        301 : HTML_ERT := 'Moved Permanently' ;
        302 : HTML_ERT := 'Moved Temporarily' ;
        400 : HTML_ERT := 'Bad Request' ;
        401 : HTML_ERT := 'Unauthorized' ;
        402 : HTML_ERT := 'Payment Required' ;
        403 : HTML_ERT := 'Forbidden' ;
        404 : HTML_ERT := 'Not Found' ;
        405 : HTML_ERT := 'Method Not Allowed' ;
        407 : HTML_ERT := 'Proxy Authentication Required' ;
        408 : HTML_ERT := 'Request Timeout' ;
        409 : HTML_ERT := 'Conflict' ;
        410 : HTML_ERT := 'Gone' ;
        502 : HTML_ERT := 'Bad Gateway' ;
        503 : HTML_ERT := 'Service Unavailable' ;
        504 : HTML_ERT := 'Gateway Timeout' ;
        else HTML_ERT := 'Unknown error number ' + Num1( X ) ; ;
    end ;
end ;


{$IFDEF VER140}
type tagSize = TSize ;
{$ENDIF}

{$IFDEF Delphi}
procedure Write_Text( Handle : HDC ; Font : HFONT ; Color : TColor ;
    X, Y, Max_X : integer ; Text : string ) ;

    procedure Write_Segment( X, Y : integer ; const Text : string ) ;

    var Rect : TRect ;

    begin
        if( length( Text ) > 0 ) then
        begin
            SetBKMode( Handle, TRANSPARENT ) ; { Something resets this, so change it back }
//            Windows.TextOut( Handle, X, Y, PChar( Text ), length( Text ) ) ;
            Rect.Top := 0 ;
            Rect.Bottom := 32767 ;
            Rect.Left := X ;
            Rect.Right := X + Max_X ;
            if( Rect.Right > 32767 ) then
            begin
                Rect.Right := 32767 ;
            end ;
            ExtTextOut( Handle, X, Y, ETO_CLIPPED, @Rect, PChar( Text ), length( Text ), nil ) ;
        end ;
    end ; { Write_Text.Write_Segment }

var Dummy : integer ;
    S : string ;
    Saved : Thandle ;
    Size : tSize ;

begin
    { Setup... }
    Saved := SelectObject( Handle, Font ) ;
    try
        SetTextColor( Handle, ColorToRGB( Color ) ) ;

        { Handle up to tab... }
        Dummy := pos( #9, Text + #9 ) ;
        S := copy( Text, 1, Dummy - 1 ) ;
        Write_Segment( X, Y, S ) ;

        { Write text following tab... }
        if( Dummy < length( Text ) ) then
        begin
            S := copy( Text, Dummy + 1, length( Text ) ) ;
            GetTextExtentPoint32( Handle, PChar( S ), length( S ), Size ) ;
            X := Max_X - Size.cx ;
            Write_Segment( X, Y, S ) ;
        end ;
    finally
        { Cleanup }
        SelectObject( Handle, Saved ) ;
    end ;
end ; { Write_Text }


type TFont_Info = class
                      public { API... }
                          Color : TColor ;
                          Face : string ;
                          Size : integer ;
                  end ;

function HTML_To_Text( X : string ) : string ; { Convert HTML to plain text }

var Dummy, Dummy1 : integer ;

begin
    Dummy := pos( '<', X ) ;
    while( Dummy > 0 ) do
    begin
        Dummy1 := pos( '>', X + '>' ) ;
        X := copy( X, 1, Dummy - 1 ) + copy( X, Dummy1 + 1, length( X ) ) ;
        Dummy := pos( '<', X ) ;
    end ;
    Result := From_HTML( X, 0 ) ;
end ; { HTML_To_Text }


function Max_Position( DC : HDC ; Temp : string ; Max_X : integer ) : integer ;

var Dummy : integer ;
    Size : tSize ;

begin
    if( Max_X <= 0 ) then
    begin
        Result := 0 ;
        exit ;
    end ;
    Dummy := length( Temp ) ;
    Temp := Edit( Temp, 128 ) ; { Remove trailing spaces }
    if( ( length( Temp ) = 0 ) and ( Dummy > 0 ) ) then // Must have been all space(s)
    begin
        Result := Dummy ;
        exit ;
    end ;
    GetTextExtentPoint( DC, PChar( Temp ), length( Temp ), Size ) ;
    while( Size.cx > Max_X ) do
    begin
        while( ( length( Temp ) > 0 ) and ( Temp[ length( Temp ) ] <> ' ' ) ) do
        begin
            setlength( Temp, length( Temp ) - 1 ) ;
        end ;
        Temp := Edit( Temp, 128 ) ; { Remove trailing spaces }
        GetTextExtentPoint( DC, PChar( Temp ), length( Temp ), Size ) ;
        if( Temp = '' ) then
        begin
            break ;
        end ;
    end ;
    Result := length( Temp ) ;
end ;


const Font_Scaling = 3 ;

function _Process_HTML_Text( DC : HDC ; Font : TFont ; S : string ;
     Draw : boolean ; X, Y, Max_X : integer ;
     var Max_Height : integer ; List : TInteger_List ;
     Single_Line : boolean ) : integer ;

var Big_Count : integer ;
    Bold_Count : integer ;
    Color : TColor ;
    This_Font : HFONT ;
    Fonts : TList ;
    LogFont : TLogFont ;
    P : TFont_Info ;
    Size : tSize ;
    Strikeout_Count : integer ;
    Superscript_Count : integer ;
    Tag : string ;
    Temp : string ;
    TM : TTextMetric ;
    Underline_Count : integer ;
    Work : string ;

    procedure Create_New_Font ;

    var Temp : HFONT ;

    begin
        Temp := This_Font ;
        This_Font := CreateFontIndirect( LogFont ) ;
        selectobject( DC, This_Font ) ;
        if( Temp <> 0 ) then
        begin
            deleteobject( Temp ) ;
        end ;
        GetTextMetrics( DC, TM ) ;
    end ;

var Adjust : integer ; // Amount to adjust Y offset by to align text baselines
    Dummy, Dummy1 : integer ;
    Height : integer ; { Current height }
    Index, Loop, M : integer ;
    Line : integer ; { Current line (0 based) }
    Original_Max_X, Original_X : integer ;
    Pen : TPen ;
    Saved : THandle ;
    Total_Height : integer ;
    X_Center, Y_Center : integer ;
    X_Offset, Y_Offset : integer ;
    Weight : integer ;
    Width : integer ;

    procedure New_Line ;

    begin
        // Move to next line...
        X := Original_X ;
        Y := Y + Height - 1 ; // EAC, 10-Oct-2006: Added "- 1" because Copyright notice appears to have an extra pixel between lines
        inc( Line ) ;
        if( List <> nil ) then
        begin
            if( Draw ) then
            begin
                Height := 0 ;
                if( List.Count > Line ) then
                begin
                    Height := List[ Line ] ;
                end ;
            end else
            begin
                List.Add( Height ) ;
                Height := 0 ;
            end ;
        end ;
        Max_X := Original_Max_X ;
        Total_Height := Total_Height + Height ;//Max_Height ;
        Max_Height := 0 ;
        if( Result > Width ) then
        begin
            Width := Result ;
        end ;
        Result := 0 ;
    end ;

label Again ;

begin
    { Setup... }
    Max_Height := 0 ;
    Height := 0 ;
    Total_Height := 0 ;
    Width := 0 ;
    Result := 0 ;
    Big_Count := 0 ;
    Bold_Count := 0 ;
    Strikeout_Count := 0 ;
    Superscript_Count := 0 ;
    Underline_Count := 0 ;
    This_Font := 0 ;
    Line := 0 ;
    Original_X := X ;
    Original_Max_X := Max_X ;
    Fonts := TList.Create ;
    Color := Font.Color ;
    fillchar( LogFont, sizeof( LogFont ), 0 ) ;
    GetObject( Font.Handle, sizeof( LogFont ), @LogFont ) ;
    Create_New_Font ;

    { Process text... }
    try
        if( ( List <> nil ) and ( List.COunt > 0 ) and Draw ) then
        begin
            Height := List[ 0 ] ;
        end ;
        while( length( S ) > 0 ) do
        begin
            if( copy( S, 1, 1 ) = '<' ) then { Tag }
            begin
                Dummy := pos( '>', S + '>' ) ;
                Tag := copy( S, 1, Dummy ) ;
                if( copy( Tag, length( Tag ), 1 ) = '>' ) then
                begin
                    Tag := copy( Tag, 1, length( Tag ) - 1 ) ; { Trim trailing ">" }
                end ;
                S := copy( S, Dummy + 1, length( S ) ) ;
                Tag := Edit( Tag, 16 or 32 or 256 ) ;
                if( Tag = '<B' ) then
                begin
                    Bold_Count := Bold_Count + 1 ;
                    LogFont.lfWeight := FW_BOLD ;
                    Create_New_Font ;
                end else
                if( Tag = '</B' ) then
                begin
                    Bold_Count := Bold_Count - 1 ;
                    if( Bold_Count = 0 ) then
                    begin
                        LogFont.lfWeight := FW_NORMAL ;
                        Create_New_Font ;
                    end ;
                end else
                if( Tag = '<BIG' ) then
                begin
                    Big_Count := Big_Count + 1 ;
                    LogFont.lfHeight := LogFont.lfHeight * 2 ;
                    Create_New_Font ;
                end else
                if( Tag = '</BIG' ) then
                begin
                    Big_Count := Big_Count - 1 ;
                    if( Big_Count = 0 ) then
                    begin
                        LogFont.lfHeight := LogFont.lfHeight div 2 ;
                        Create_New_Font ;
                    end ;
                end else
                if( Tag = '<BR' ) then
                begin
                    New_Line ;
                end else
                if( copy( Tag, 1, 5 ) = '<FONT' ) then
                begin
                    P := TFont_Info.Create ;
                    if( Fonts.Count > 0 ) then
                    begin
                        P.Color := TFont_Info( Fonts[ Fonts.Count - 1 ] ).Color ;
                        P.Size := TFont_Info( Fonts[ Fonts.Count - 1 ] ).Size ;
                    end else
                    begin
                        P.Color := Color ;
                        P.Size := Font.Size div Font_Scaling ;
                    end ;
                    Fonts.Add( P ) ;
                    Color := P.Color ;
                    Tag := copy( Tag, 6, length( Tag ) ) ; // Trim "FONT"
                    while( length( Tag ) > 0 ) do
                    begin
                        Tag := Edit( Tag, 8 or 128 ) ; // Trim leading/trailing spaces
                        Work := Parse_Parameter( ' ', Tag ) ;
                        if( copy( Work, 1, 6 ) = 'FACE="' ) then
                        begin
                            Work := copy( Work, 7, length( Work ) ) ;
                            if( copy( Work, length( Work ), 1 ) = '"' ) then
                            begin
                                Work := copy( Work, 1, length( Work ) - 1 ) ;
                            end ;
                            if( length( Work ) <= sizeof( LogFont.lfFaceName ) ) then
                            begin
                                fillchar(LogFont.lfFaceName, sizeof(LogFont.lfFaceName), 0) ;
                                move( Work[ 1 ], LogFont.lfFaceName, length( Work ) ) ;
                                P.Face := Work ;
                                Create_New_Font ;
                            end ;
                        end else
                        if( copy( Work, 1, 5 ) = 'SIZE=' ) then
                        begin
                            { Strip SIZE }
                            Work := copy( Work, 6, length( Work ) ) ;

                            { Strip quotes... }
                            if( copy( Work, 1, 1 ) = '"' ) then
                            begin
                                Work := copy( Work, 2, length( Work ) - 1 ) ;
                            end ;
                            if( copy( Work, length( Work ), 1 ) = '"' ) then
                            begin
                                Work := copy( Work, 1, length( Work ) - 1 ) ;
                            end ;

                            if( copy( Work, 1, 1 ) = '-' ) then
                            begin
                                Work := copy( Work, 2, length( Work ) - 1 ) ;
                                val( Work, Index, Dummy ) ;
                                if( Dummy = 0 ) then
                                begin
                                    P.Size := P.Size - Index ;
                                end ;
                            end else
                            if( copy( Work, 1, 1 ) = '+' ) then
                            begin
                                Work := copy( Work, 2, length( Work ) - 1 ) ;
                                val( Work, Index, Dummy ) ;
                                if( Dummy = 0 ) then
                                begin
                                    P.Size := P.Size + Index ;
                                end ;
                            end else
                            begin
                                val( Work, Index, Dummy ) ;
                                if( Dummy = 0 ) then
                                begin
                                    P.Size := Index ;
                                end ;
                            end ;
                            LogFont.lfHeight := ( P.Size * Font_Scaling ) + 8 ;
                            Create_New_Font ;
                        end else
                        if( copy( Work, 1, 6 ) = 'COLOR=' ) then
                        begin
                            Work := copy( Work, 7, length( Work ) ) ;
                            if( Work = 'BLACK' ) then
                            begin
                                P.Color := clBlack ;
                            end else
                            if( Work = 'WHITE' ) then
                            begin
                                P.Color := clWhite ;
                            end ;
                            Color := P.Color ;
                        end ;
                    end ; { while( length( Tag ) > 0 ) }
                end else
                if( Tag = '</FONT' ) then
                begin
                    if( Fonts.Count > 0 ) then { Something to unnest }
                    begin
                        TFont_Info( Fonts[ Fonts.Count - 1 ] ).Free ;
                        Fonts[ Fonts.Count - 1 ] := nil ;
                        Fonts.Delete( Fonts.Count - 1 ) ;
                        if( Fonts.Count > 0 ) then { We are nested }
                        begin
                            Color := TFont_Info( Fonts[ Fonts.Count - 1 ] ).Color ;
                            if( TFont_Info( Fonts[ Fonts.Count - 1 ] ).Face <> '' ) then
                            begin
                                if( length( Work ) <= sizeof( LogFont.lfFaceName ) ) then
                                begin
                                    fillchar(LogFont.lfFaceName, sizeof(LogFont.lfFaceName), 0) ;
                                    move( Work[ 1 ], LogFont.lfFaceName, length( Work ) ) ;
                                    Create_New_Font ;
                                end ;
                            end ;
                        end else
                        begin
                            GetObject( Font.Handle, sizeof( LogFont ), @LogFont ) ;
                            Create_New_Font ;
                        end ;
                    end ;
                end else
                if( ( Tag = '<S' ) or ( Tag = '<STRIKE' ) ) then
                begin
                    Strikeout_Count := Strikeout_Count + 1 ;
                    LogFont.lfStrikeOut := 1 ;
                    Create_New_Font ;
                end else
                if( ( Tag = '</S' ) or ( Tag = '</STRIKE' ) ) then
                begin
                    Strikeout_Count := Strikeout_Count - 1 ;
                    if( Strikeout_Count = 0 ) then
                    begin
                        LogFont.lfStrikeOut := 0 ;
                        Create_New_Font ;
                    end ;
                end else
                if( copy( Tag, 1, 8 ) = '<STDIMG ' ) then
                begin
                    Tag := copy( Tag, 9, length( Tag ) ) ;
                    if( Draw ) then
                    begin
                        { Setup... }
                        Pen := TPen.Create ;
                        Pen.Color := Color ;
                        Saved := selectobject( DC, Pen.Handle ) ;

                        // Determine maximum segment length...
                        M := TM.tmAscent - TM.tmExternalLeading ;
                        if( TM.tmExternalLeading = 0 ) then
                        begin
                            M := M - 1 ;
                        end ;
                        if( M > TM.tmMaxCharWidth - 2 ) then
                        begin
                            M := TM.tmMaxCharWidth - 2 ;
                        end ;

                        // Determine x and Y offset and center points...
                        X_Offset := ( TM.tmMaxCharWidth - M ) div 2 ;
                        Y_Offset := ( TM.tmAscent - M ) div 2 ;
                        if( TM.tmExternalLeading = 0 ) then
                        begin
                            Y_Offset := Y_Offset + 1 ;
                        end ;
                        X_Center := X_Offset + M div 2 ;
                        Y_Center := Y_Offset + M div 2 ;

                        // Determine number of lines for weight...
                        Weight := M div 2 ; // Maximum weight
                        Weight := Weight * LogFont.lfWeight div 1000 ;
                        if( Weight < 1 ) then
                        begin
                            Weight := 1 ;
                        end ;
                        X_Center := X_Center - Weight div 2 ;
                        Y_Center := Y_Center - Weight div 2 ;

                        // Process specific control type
                        if( Tag = 'COLLAPSE' ) then
                        begin
                            Y_Center := Y + Y_Center ;
                            X_Offset := X + X_Offset ;
                            for Loop := 0 to Weight - 1 do
                            begin
                                Windows.MoveToEx( DC, X_Offset, Y_Center + Loop, nil ) ;
                                Windows.LineTo( DC, X_Offset + M, Y_Center + Loop ) ;
                            end ;
                        end else
                        if( Tag = 'EXPAND' ) then
                        begin
                            X_Center := X + X_Center ;
                            Y_Center := Y + Y_Center ;
                            X_Offset := X + X_Offset ;
                            Y_Offset := Y + Y_Offset ;
                            for Loop := 0 to Weight - 1 do
                            begin
                                Windows.MoveToEx( DC, X_Center + Loop, Y_Offset, nil ) ;
                                Windows.LineTo( DC, X_Center + Loop, Y_Offset + M ) ;
                                Windows.MoveToEx( DC, X_Offset, Y_Center + Loop, nil ) ;
                                Windows.LineTo( DC, X_Offset + M, Y_Center + Loop ) ;
                            end ;
                        end ;

                        // Cleanup...
                        selectobject( DC, Saved ) ;
                        Pen.Free ;

                        // Adjust position...
                        X := X + TM.tmMaxCharWidth ;
                        Max_X := Max_X - Size.cx ;
                        if( Max_X <= 0 ) then
                        begin
                            Result := Result + TM.tmMaxCharWidth ;
                            break ;
                        end ;
                    end ;
                    Result := Result + TM.tmMaxCharWidth ;
                end else
                if( Tag = '<SUP' ) then
                begin
                    Superscript_Count := Superscript_Count + 1 ;
                end else
                if( Tag = '</SUP' ) then
                begin
                    Superscript_Count := Superscript_Count - 1 ;
                end else
                if( Tag = '<U' ) then
                begin
                    Underline_Count := Underline_Count + 1 ;
                    LogFont.lfUnderline := 1 ;
                    Create_New_Font ;
                end else
                if( Tag = '</U' ) then
                begin
                    Underline_Count := Underline_Count - 1 ;
                    if( Underline_Count = 0 ) then
                    begin
                        LogFont.lfUnderline := 0 ;
                        Create_New_Font ;
                    end ;
                end ;
                continue ;
            end ; { if( copy( S, 1, 1 ) = '<' ) }

            { Normal text... }
            Dummy := pos( '<', S + '<' ) ;
            Dummy1 := pos( ' ', S + ' ' ) ;
            if( Dummy1 = 1 ) then
            begin
                Dummy1 := instr( 2, S + ' ', ' ' ) ;
            end ;
            if( Dummy > Dummy1 ) then
            begin
                Dummy := Dummy1 ;
            end ;
            { At this point, Dummy is the end of some text to process.  This may
              be the literal end of the text, or a space in the text, or the
              start of a tag. }
Again:
            //NLR Oct-16-2006 Changed options in following code to 1 - allowing processing of '&' tab text.
            Temp := From_HTML( copy( S, 1, Dummy - 1 ), 1{0} ) ;
            GetTextExtentPoint( DC, PChar( Temp ), length( Temp ), Size ) ;
            if( Superscript_Count > 0 ) then
            begin
                Size.cy := Size.cy + Size.cy div 3 ;
            end ;
            if( Size.cy > Max_Height ) then
            begin
                Max_Height := Size.cy ;
            end ;
            Dummy1 := Max_Position( DC, Temp, Max_X ) ;
            if( ( Dummy1 > 0 ) and ( Dummy1 < length( Temp ) ) ) then // Some, but not all, of the text fits
            begin
                Dummy1 := Dummy ;
                while( ( Dummy1 > 0 ) and ( S[ Dummy1 ] <> ' ' ) ) do
                begin
                    dec( Dummy1 ) ;
                end ;
                if( Dummy1 > 1 ) then // Succeeded in trimming a word
                begin
                    Dummy := Dummy1 ;
                    goto Again ;
                end ;
            end else // EAC, 27-Nov-2006, 1714: Added "( not Single_Line )" to next line
            if( ( Dummy1 = 0 ) and ( not Single_Line ) ) then // None of the text fits on multi-line
            begin
                if( X = Original_X ) then // First item on line
                begin
                    Dummy1 := pos( ' ', Temp + ' ' ) ; // Find end of first word
                    while(
                           ( Dummy1 <= length( Temp ) )
                           and
                           ( Temp[ Dummy1 ] = ' ' )
                         ) do
                    begin
                        inc( Dummy1 ) ; // Skip past all spaces
                    end ;
                    Dummy := Dummy1 ;
                    Temp := copy( Temp, 1, Dummy - 1 ) ;
                end else
                begin
                    New_Line ;
                    Dummy := 1 ;
                    Temp := '' ;
                    S := Edit( S, 8 ) ; // Strip leading spaces on new line
                end ; // if( X = 0 )
(*
            end else
            begin
                if( Dummy1 <= length( Temp ) ) then
                begin
                    Dummy := Dummy1 + 1 ;
                end ;
                Temp := copy( Temp, 1, Dummy - 1 ) ;
*)
            end ; // if( Dummy1 = 0 )

            GetTextExtentPoint( DC, PChar( Temp ), length( Temp ), Size ) ;
            if( length( Temp ) > 0 ) then
            begin
                if( Superscript_Count > 0 ) then
                begin
                    Size.cy := Size.cy + Size.cy div 3 ;
                end ;

                { Normalize baseline... }
//                Size.cy := Size.cy - TM.tmDescent ;
                Adjust := 0 ;
                if( Height <> 0 ) then // Line height already determined
                begin
                    // Adjust the top of the text by the difference between the text height and the line height
                    Adjust := ( Height - Size.cy ) ;
(*
                    if( Size.cy > Height ) then
                    begin
                        Y := Y - ( Size.cy - Height ) ;
                    end else
                    if( Size.cy < Height ) then
                    begin
                        Y := Y + ( Height - Size.cy ) ;
                    end ;
*)
                end ;
                if( Size.cy {+ TM.tmDescent} > Height ) then
                begin
                    Height := Size.cy {+ TM.tmDescent} ;
                end ;

                if( Draw ) then
                begin
                    { Draw the text }
                    Write_Text( DC, This_Font, Color, X, Y + Adjust, Max_X, Temp ) ;
                end ; { if( Draw ) }
                X := X + Size.cx ;
                Max_X := Max_X - Size.cx ;
                if( ( Max_X <= 0 ) and Single_Line ) then
                begin
                    Result := Result + Size.cx ;
                    break ;
                end ;
            end ; { if( length( Temp ) > 0 ) }
            S := copy( S, Dummy, length( S ) ) ;
            Result := Result + Size.cx ;
        end ; { while( length( S ) > 0 ) }
        if( ( List <> nil ) and ( not Draw ) ) then
        begin
            List.Add( Height ) ;
        end ;

        // Restore original font
        selectobject( DC, Font.Handle ) ;
        SetTextColor( DC, ColorToRGB( Font.Color ) ) ;
    finally
        if( Result < Width ) then
        begin
            Result := Width ;
        end ;
        Max_Height := Total_Height + Max_Height ;

        // Clean up...
        for Dummy := 0 to Fonts.Count - 1 do
        begin
            TFont_Info( Fonts[ Dummy ] ).Free ;
            Fonts[ Dummy ] := nil ;
        end ;
        Fonts.Free ;
        deleteobject( This_Font ) ;
    end ;
end ; { _Process_HTML_Text }


function Process_HTML_Text( DC : HDC ; Font : TFont ; const S : string ;
     Draw : boolean ; X, Y, Max_X : integer ;
     var Max_Height : integer ; Single_Line : boolean ) : integer ;

var List : TInteger_List ;
    Temp : integer ;

begin
    List := TInteger_List.Create ;
    try
        Result := _Process_HTML_Text( DC, Font, S, False, X, Y, Max_X,
            Max_Height, List, Single_Line ) ;
        // Now add together all the line heights
        Max_Height := 0 ;
        for Temp := 0 to List.Count - 1 do
        begin
            Max_Height := Max_Height + List[ Temp ] ;
        end ;
        if( Draw ) then
        begin
            _Process_HTML_Text( DC, Font, S, True, X, Y, Max_X, Temp, List, Single_Line ) ;
        end ;
    finally
        List.Free ;
    end ;
end ;
{$ENDIF}


var Entities : array[ 0..255 ] of string[ 6 ] ;

function From_HTML( X : string ; Options : longint ) : string ;

var Dummy, Index, Loop : integer ;
    Found : boolean ;
    Res, LTemp, Temp, Work : string ;

begin
    Res := '' ;
    while( length( X ) > 0 ) do
    begin
        Dummy := pos( '&', X + '&' ) ;
        Res := Res + copy( X, 1, Dummy - 1 ) ;
        X := copy( X, Dummy, length( X ) ) ;
        Dummy := pos( ';', X ) ;
        if( Dummy = 0 ) then { No terminator - bad entity reference }
        begin
            if( ( Options and 1 ) <> 0 ) then
            begin
                Res := Res + X ;
            end ;
            X := '' ;
        end else
        if( Dummy < 4 ) then { "&;" or single-character entity }
        begin
            if( ( Options and 1 ) <> 0 ) then
            begin
                Res := Res + copy( X, 1, Dummy ) ;
            end ;
            X := copy( X, Dummy + 1, length( X ) ) ;
        end else
        begin
            Temp := copy( X, 2, Dummy - 2 ) ;
            X := copy( X, Dummy + 1, length( X ) ) ;
            if( ( copy( Temp, 1, 1 ) = '#' ) and ( length( Temp ) > 1 ) ) then
            begin
                Work := copy( Temp, 2, length( Temp ) ) ;
                val( copy( Temp, 2, length( Temp ) ), Index, Dummy ) ;
                if( Dummy = 0 ) then { Valid number }
                begin
                    Res := Res + char( Index ) ;
                    continue ;
                end ;
            end ;
            Found := False ;
            for Loop := 0 to 255 do
            begin
                if( Entities[ Loop ] = Temp ) then
                begin
                    Res := Res + char( Loop ) ;
                    Found := True ;
                    break ;
                end ;
            end ;
            if( not Found ) then
            begin
                LTemp := lowercase( Temp ) ;
                for Loop := 0 to 255 do
                begin
                    if( Entities[ Loop ] = LTemp ) then
                    begin
                        Res := Res + char( Loop ) ;
                        Found := True ;
                        break ;
                    end ;
                end ;
            end ;
            if( not Found ) then
            begin
                if(
                    ( LTemp = 'spades' )
                    or
                    ( LTemp = 'clubs' )
                    or
                    ( LTemp = 'hearts' )
                    or
                    ( LTemp = 'diams' )
                    or
                    ( LTemp = 'oline' )
                    or
                    ( LTemp = 'larr' )
                    or
                    ( LTemp = 'rarr' )
                    or
                    ( LTemp = 'uarr' )
                    or
                    ( LTemp = 'darr' )
                  ) then
                begin
                    Found := True ;
                    // Cannot translate to a string
                end ;
            end ;
            if( ( not Found ) and ( ( Options and 1 ) <> 0 ) ) then
            begin
                Res := Res + '&' + Temp + ';' ;
            end ;
        end ; { if( Dummy = 0 )  }
    end ; { while( length( X ) > 0 )  }
    From_HTML := Res ;
end ; { From_HTML }


function To_HTML( const X : string ) : string ;

var C : byte ;
    Loop : integer ;
    S : string ;

begin
    S := '' ;
    for Loop := 1 to length( X ) do
    begin
        C := ord( X[ Loop ] ) ;
        if( length( Entities[ C ] ) > 1 ) then
        begin
            S := S + '&' + Entities[ C ] + ';' ;
        end else
        begin
            S := S + Entities[ C ] ;
        end ;
    end ;
    To_HTML := S ;
end ;


procedure Initialize_Unit ;

var Loop : integer ;

begin
    // Note: Doesn't currently support certain entities which are UNICODE characters
    for Loop := 0 to 255 do
    begin
        Entities[ Loop ] := char( Loop ) ;
    end ;
    Entities[ 9 ] := '#TAB' ;
    Entities[ 10 ] := '#RS' ;
    Entities[ 13 ] := '#RE' ;
    Entities[ 38 ] := 'amp' ;
    Entities[ 60 ] := 'lt' ;
    Entities[ 62 ] := 'gt' ;
    Entities[ 126 ] := 'tilde' ;
    Entities[ 130 ] := 'sbquo' ;
    Entities[ 132 ] := 'dbquo' ;
    Entities[ 134 ] := 'dagger' ;
    Entities[ 135 ] := 'Dagger' ;
    Entities[ 137 ] := 'permil' ;
    Entities[ 139 ] := 'lsaquo' ;
    Entities[ 145 ] := 'lsquo' ;
    Entities[ 146 ] := 'rsquo' ;
    Entities[ 147 ] := 'ldquo' ;
    Entities[ 148 ] := 'rdquo' ;
    Entities[ 150 ] := 'ndash' ;
    Entities[ 151 ] := 'mdash' ;
    Entities[ 153 ] := 'trade' ;
    Entities[ 155 ] := 'rsaquo' ;
    Entities[ 160 ] := 'nbsp' ;
    Entities[ 161 ] := 'iexcl' ;
    Entities[ 162 ] := 'cent' ;
    Entities[ 163 ] := 'pound' ;
    Entities[ 164 ] := 'curren' ;
    Entities[ 165 ] := 'yen' ;
    Entities[ 166 ] := 'brvbar' ;
    Entities[ 167 ] := 'sect' ;
    Entities[ 168 ] := 'uml' ;
    Entities[ 169 ] := 'copy' ;
    Entities[ 170 ] := 'ordf' ;
    Entities[ 171 ] := 'laquo' ;
    Entities[ 172 ] := 'not' ;
    Entities[ 173 ] := 'shy' ;
    Entities[ 174 ] := 'reg' ;
    Entities[ 175 ] := 'macr' ;
    Entities[ 176 ] := 'deg' ;
    Entities[ 177 ] := 'plusmn' ;
    Entities[ 178 ] := 'sup2' ;
    Entities[ 179 ] := 'sup3' ;
    Entities[ 180 ] := 'acute' ;
    Entities[ 181 ] := 'micro' ;
    Entities[ 182 ] := 'para' ;
    Entities[ 183 ] := 'middot' ;
    Entities[ 184 ] := 'cedil' ;
    Entities[ 185 ] := 'supl' ;
    Entities[ 186 ] := 'ordm' ;
    Entities[ 187 ] := 'raquo' ;
    Entities[ 188 ] := 'frac14' ;
    Entities[ 189 ] := 'frac12' ;
    Entities[ 190 ] := 'frac34' ;
    Entities[ 191 ] := 'iquest' ;
    Entities[ 215 ] := 'times' ;
    Entities[ 247 ] := 'divide' ;
    Entities[ ord( '' ) ] := 'quot' ;
    Entities[ ord( '' ) ] := 'AElig' ;
    Entities[ ord( '' ) ] := 'Aacute' ;
    Entities[ ord( '' ) ] := 'Acirc' ;
    Entities[ ord( '' ) ] := 'Agrave' ;
    Entities[ ord( '' ) ] := 'Aring' ;
    Entities[ ord( '' ) ] := 'Atilde' ;
    Entities[ ord( '' ) ] := 'Auml' ;
    Entities[ ord( '' ) ] := 'Ccedil' ;
    Entities[ ord( '' ) ] := 'ETH' ;
    Entities[ ord( '' ) ] := 'Eacute' ;
    Entities[ ord( '' ) ] := 'Ecirc' ;
    Entities[ ord( '' ) ] := 'Egrave' ;
    Entities[ ord( '' ) ] := 'Euml' ;
    Entities[ ord( '' ) ] := 'Iacute' ;
    Entities[ ord( '' ) ] := 'Icirc' ;
    Entities[ ord( '' ) ] := 'Igrave' ;
    Entities[ ord( '' ) ] := 'Iuml' ;
    Entities[ ord( '' ) ] := 'Ntilde' ;
    Entities[ ord( '' ) ] := 'Oacute' ;
    Entities[ ord( '' ) ] := 'Ocirc' ;
    Entities[ ord( '' ) ] := 'Ograve' ;
    Entities[ ord( '' ) ] := 'Oslash' ;
    Entities[ ord( '' ) ] := 'Otilde' ;
    Entities[ ord( '' ) ] := 'Ouml' ;
    Entities[ ord( '' ) ] := 'Uacute' ;
    Entities[ ord( '' ) ] := 'Ucirc' ;
    Entities[ ord( '' ) ] := 'Ugrave' ;
    Entities[ ord( '' ) ] := 'Uuml' ;
    Entities[ ord( '' ) ] := 'Yacute' ;
    Entities[ ord( '' ) ] := 'aacute' ;
    Entities[ ord( '' ) ] := 'acirc' ;
    Entities[ ord( '' ) ] := 'aelig' ;
    Entities[ ord( '' ) ] := 'agrave' ;
    Entities[ ord( '' ) ] := 'aring' ;
    Entities[ ord( '' ) ] := 'atilde' ;
    Entities[ ord( '' ) ] := 'auml' ;
    Entities[ ord( '' ) ] := 'ccedil' ;
    Entities[ ord( '' ) ] := 'eacute' ;
    Entities[ ord( '' ) ] := 'ecirc' ;
    Entities[ ord( '' ) ] := 'egrave' ;
    Entities[ ord( '' ) ] := 'emul' ;
    Entities[ ord( '' ) ] := 'iacute' ;
    Entities[ ord( '' ) ] := 'icirc' ;
    Entities[ ord( '' ) ] := 'igrave' ;
    Entities[ ord( '' ) ] := 'iuml' ;
    Entities[ ord( '' ) ] := 'ntilde' ;
    Entities[ ord( '' ) ] := 'oacute' ;
    Entities[ ord( '' ) ] := 'ocirc' ;
    Entities[ ord( '' ) ] := 'ograve' ;
    Entities[ ord( '' ) ] := 'oslash' ;
    Entities[ ord( '' ) ] := 'otilde' ;
    Entities[ ord( '' ) ] := 'ouml' ;
    Entities[ ord( '' ) ] := 'szlig' ;
    Entities[ ord( '' ) ] := 'thorn' ;
    Entities[ ord( '' ) ] := 'THORN' ;
    Entities[ ord( '' ) ] := 'uacute' ;
    Entities[ ord( '' ) ] := 'ucirc' ;
    Entities[ ord( '' ) ] := 'ugrave' ;
    Entities[ ord( '' ) ] := 'uuml' ;
    Entities[ ord( '' ) ] := 'yacute' ;
    Entities[ ord( '' ) ] := 'yuml' ;
    Entities[ ord( '' ) ] := 'cent' ;
end ; // Initialize_Unit


{ TSHTML_Parser methods... }

{ Constructors and destructors... }

constructor TSHTML_Parser.Init ;

begin
    inherited Init ;

    Source.Init ;
    Preserve_Whitespace := False ;
end ;


destructor TSHTML_Parser.Done ;

begin
    Source.Done ;

    inherited Done ;
end ;


{ Internal utility routines... }

function TSHTML_Parser.EOF : boolean ;

begin
    EOF := ( Source.Length = 0 ) ;
end ;


function TSHTML_Parser.Get_Token : PString ;

var Dummy : integer ;
    Res : PASCII_String ;
    Work : PString ;

begin
    Dummy := Source.Instr( 1, Temp_String( '<' ) ) ;
    if( Dummy = 0 ) then
    begin
        Dummy := Source.Length + 1 ;
    end ;
    if( Dummy = 1 ) then
    begin
        Dummy := Source.Instr( 1, Temp_String( '>' ) ) ;
        if( Dummy = 0 ) then
        begin
            Dummy := Source.Length + 1 ;
        end else
        begin
            Dummy := Dummy + 1 ;
        end ;
    end ;

    { Extract token... }
    Res := new( PASCII_String, Init ) ;
    Work := Source.Copy( 1, Dummy - 1 ) ;
    Res.Assign( Work ) ;
    dispose( Work, Done ) ;
    Work := Source.copy( Dummy, Source.Length ) ;
    Source.Assign( Work ) ;
    dispose( Work, Done ) ;

    { Process EOL characters... }
    Dummy := Res.Instr( 1, Temp_String( CR ) ) ;
    while( Dummy > 0 ) do
    begin
        Res.At_Put( Dummy, 32 ) ;
        Dummy := Res.Instr( 1, Temp_String( CR ) ) ;
    end ;
    Dummy := Res.Instr( 1, Temp_String( LF ) ) ;
    while( Dummy > 0 ) do
    begin
        Res.At_Put( Dummy, 32 ) ;
        Dummy := Res.Instr( 1, Temp_String( LF ) ) ;
    end ;
    if( not Preserve_Whitespace ) then
    begin
        Work := Res.Edit( 16 ) ; { Multiple whitespace to single space }
        dispose( Res, Done ) ;
        Get_Token := Work ;
    end else
    begin
        Get_Token := Res ;
    end ;
end ; { TSHTML_Parser.Get_Token }


{ API... }

function TSHTML_Parser._Get_Section( S : string ; XML : boolean ) : PString ;

var Dummy : integer ;
    Ending, Starting : string ;
    Nest_Count : integer ;
    Res : PASCII_String ;
    Temp : PString ;
    Work : string ;

begin
    { Normalize passed section name... }
    if( copy( S, 1, 1 ) = '<' ) then
    begin
        S := copy( S, 2, length( S ) ) ;
    end ;
    if( copy( S, length( S ), 1 ) = '>' ) then
    begin
        S := copy( S, 1, length( S ) - 1 ) ;
    end ;
    Dummy := pos( '=', S + '=' ) ;
    S := copy( S, 1, Dummy - 1 ) ;

    { Build ending tag to search for... }
    S := Edit( S, 512 ) ;
    if( XML ) then
    begin
        Ending := '<' + S + '/>' ;
    end else
    begin
        Ending := '</' + S + '>' ;
    end ;
    Starting := '<' + S ;

    { Setup }
    Nest_Count := 0 ;
    Res := new( PASCII_String, Init ) ;

    { Process section... }
    Temp := Token ;
    while( Temp.Length > 0 ) do
    begin
        Work := Edit( Temp.Text, 512 ) ;
        if( Work = Ending ) then
        begin
            dec( Nest_Count ) ;
            if( Nest_Count < 0 ) then
            begin
                _Get_Section := Res ;
                dispose( Temp, Done ) ;
                exit ;
            end ;
        end ;
        Res.Append( Temp ) ;
        dispose( Temp, Done ) ;
        if( copy( Work, 1, 1 ) = '<' ) then
        begin
            Dummy := pos( '=', Work + '=' ) ;
            Work := copy( Work, 1, Dummy - 1 ) ;
            if( copy( Work, length( Work ), 1 ) = '>' ) then
            begin
                Work := copy( Work, 1, length( Work ) - 1 ) ;
            end ;
            if( Work = Starting ) then
            begin
                inc( Nest_Count ) ;
            end ;
        end ;

        Temp := Token ;
    end ;
    _Get_Section := Res ; { Must have hit EOF before terminating tag }
    dispose( Temp, Done ) ;
end ; { TSHTML_Parser._Get_Section }


function TSHTML_Parser.Get_Section( const S : string ) : PString ;

begin
    Result := _Get_Section( S, False ) ;
end ;


{ TSXML_Parser methods... }

{ API... }

function TSXML_Parser.Get_Section( const S : string ) : PString ;

var Dummy : integer ;
    Work : string ;

begin
    Work := string( S ) ;
    Dummy := pos( ' ', Work + ' ' ) ;
    Work := copy( Work, 1, Dummy - 1 ) ;
    if( copy( Work, length( Work ), 1 ) <> '>' ) then
    begin
        Work := Work + '>' ;
    end ;
    Result := _Get_Section( Work, True ) ;
end ;


function TSXML_Parser.Get_Section_String( const S : string ) : string ;

var Work : PString ;

begin
    Work := Get_Section( S ) ;
    Get_Section_String := Work^.Text ;
    dispose( Work, Done ) ;
end ;


function TSXML_Parser.Get_Token_String : string ;

var Work : PString ;

begin
    Work := Token ;
    Get_Token_String := Work^.Text ;
    dispose( Work, Done ) ;
end ;


{$IFDEF DELPHI}
// THTML_Parser methods...

{ Constructors and destructors... }

constructor THTML_Parser.Create ;

begin
    inherited Create ;

    dispose( _Parser, Done ) ;
    _Parser := new( PSHTML_Parser, Init ) ;
end ;


function THTML_Parser.Get_Source : string ;

begin
    Result := PSHTML_Parser( _Parser )^.Source.Text ;
end ;


procedure THTML_Parser.Set_Source( const S : string ) ;

begin
    PSHTML_Parser( _Parser )^.Source.Set_String( S ) ;
end ;


function THTML_Parser.Get_Preserve_Whitespace : boolean ;

begin
    Result := PSHTML_Parser( _Parser )^.Preserve_Whitespace ;
end ;


procedure THTML_Parser.Set_Preserve_Whitespace( Value : boolean ) ;

begin
    PSHTML_Parser( _Parser )^.Preserve_Whitespace := Value ;
end ;


{ Internal utility routines... }

function THTML_Parser._Get_Section( const S : string ; XML : boolean ) : string ;

var Temp : PString ;

begin
    Temp := PSHTML_Parser( _Parser )^._Get_Section( S, XML ) ;
    Result := Temp.Text ;
    dispose( Temp, Done ) ;
end ;


{ API.. }

function THTML_Parser.EOF : boolean ;

begin
    EOF := PSHTML_Parser( _Parser )^.EOF ;
end ;


function THTML_Parser.Get_Section( const S : string ) : string ;

var Temp : PString ;

begin
    Temp := PSHTML_Parser( _Parser )^.Get_Section( S ) ;
    Result := Temp.Text ;
    dispose( Temp, Done ) ;
end ;



// THTML_Tag methods...

constructor THTML_Tag.Create( S : string ) ;

var N, V : string ;

begin
    inherited Create ;

    Names := TStringList.Create ;
    Values := TStringList.Create ;

    // Parse tag...
    if( copy( S, 1, 1 ) = '<' ) then
    begin
        S := copy( S, 2, length( S ) ) ;
    end ;
    if( copy( S, length( S ), 1 ) = '>' ) then
    begin
        S := copy( S, 1, length( S ) - 1 ) ;
    end ;
    Name := Parse_Parameter( ' ', S ) ;

    while( length( S ) > 0 ) do
    begin
        V := Parse_Parameter( ' ', S ) ;
        N := Parse_Parameter( '=', V ) ;
        Trim_Quotes( V ) ;
        Names.Add( N ) ;
        Values.Add( V ) ;
    end ;
end ;


destructor THTML_Tag.Destroy ;

begin
    Names.Free ;
    Names := nil ;
    Values.Free ;
    Values := nil ;

    inherited Destroy ;
end ;


function THTML_Tag.Attribute_Name( Index : integer ) : string ;

begin
    Result := Names[ Index ] ;
end ;


function THTML_Tag.Attribute_Value( Index : integer ) : String ;

begin
    Result := Values[ Index ] ;
end ;


function THTML_Tag.Count : integer ;

begin
    Result := Names.Count ;
end ;



{ TXML_Parser methods... }

{ Constructors and destructors... }

constructor TXML_Parser.Create ;

begin
    inherited Create ;

    _Parser := THTML_Parser.Create ;
end ;


destructor TXML_Parser.Destroy ;

begin
    _Parser.Free ;
    _Parser := nil ;

    inherited Destroy ;
end ;


{ API... }

function TXML_Parser.Get_Current_Line : longint ;

begin
    Result := _Parser.Get_Current_Line ;
end ;


function TXML_Parser.Get_Non_Terminals : PChar ;

begin
    Temp := _Parser.Non_Terminals ;
    Result := PChar( Temp ) ;
end ;


procedure TXML_Parser.Set_Non_Terminals( S : PChar ) ;

begin
    _Parser.Non_Terminals := string( S ) ;
end ;


function TXML_Parser.Get_Preserve_Whitespace : boolean ;

begin
    Result := _Parser.Preserve_Whitespace ;
end ;


procedure TXML_Parser.Set_Preserve_Whitespace( Value : boolean ) ;

begin
    _Parser.Preserve_Whitespace := Value ;
end ;


function TXML_Parser.Get_Source : PChar ;

begin
    Temp := _Parser.Get_Source ;
    Result := PChar( Temp ) ;
end ;


procedure TXML_Parser.Set_Source( S : PChar ) ;

begin
    _Parser.Set_Source( S ) ;
end ;


function TXML_Parser.Get_Token_Line : integer ;

begin
    Result := _Parser.Token_Line ;
end ;


procedure TXML_Parser.Set_Token_Line( V : integer ) ;

begin
    _Parser.Token_Line := V ;
end ;


procedure TXML_Parser.Add_Special_Token( S : PChar ) ;

begin
    _Parser.Add_Special_Token( string( S ) ) ;
end ;


function TXML_Parser.EOF : boolean ;

begin
    Result := _Parser.EOF ;
end ;


function TXML_Parser.Get_Section( S : PChar ) : PChar ;

var Dummy : integer ;
    Work : string ;

begin
    Work := string( S ) ;
    Dummy := pos( ' ', Work + ' ' ) ;
    Work := copy( Work, 1, Dummy - 1 ) ;
    if( copy( Work, length( Work ), 1 ) <> '>' ) then
    begin
        Work := Work + '>' ;
    end ;
    Temp := _Parser._Get_Section( Work, True ) ;
    Result := PChar( Temp ) ;
end ;


function TXML_Parser.Get_Token : PChar ;

begin
    Temp := _Parser.Get_Token ;
    Result := PChar( Temp ) ;
end ;


function TXML_Parser.Previous_Token_Line : Integer ;

begin
    Result := _Parser.Previous_Token_Line ;
end ;


procedure TXML_Parser.Push_Token_Line ;

begin
    _Parser.Push_Token_Line ;
end ;


function TXML_Parser.Read_Next_Line : boolean ;

begin
    Result := _Parser.Read_Next_Line ;
end ;


function TXML_Parser.Grab_Line : PChar ;

begin
    Temp := _Parser.Grab_Line ;
    Result := PChar( Temp ) ;
end ;


function TXML_Parser.Last_Error : longint ;

begin
    Result := _Parser.Last_Error ;
end ;


procedure TXML_Parser.Put_Token( S : PChar ) ;

begin
    _Parser.Put_Token( string( S ) ) ;
end ;


function TXML_Parser.Token : PChar ;

begin
    Temp := _Parser.Token ;
    Result := PChar( Temp ) ;
end ;


function TXML_Parser.Token_EOL : boolean ;

begin
    Result := _Parser.Token_EOL ;
end ;



{ TXML_Tag methods... }

{ Constructors and destructors... }

constructor TXML_Tag.Create( const S : string ) ;

begin
    inherited Create ;

    _Tag := THTML_Tag.Create( S ) ;
end ;


destructor TXML_Tag.Destroy ;

begin
    _Tag.Free ;
    _Tag := nil ;

    inherited Destroy ;
end ;


{ API... }

function TXML_Tag.Attribute_Name( Index : integer ) : Pchar ;

begin
    Temp := _Tag.Attribute_Name( Index ) ;
    Result := PChar( Temp ) ;
end ;


function TXML_Tag.Attribute_Value( Index : integer ) : PChar ;

begin
    Temp := _Tag.Attribute_Value( Index ) ;
    Result := PChar( Temp ) ;
end ;


function TXML_Tag.Count : integer ;

begin
    Result := _Tag.Count ;
end ;


function TXML_Tag.Get_Name : PChar ;

begin
    Result := PChar( _Tag.Name ) ;
end ;


procedure TXML_Tag.Set_Name( P : PChar ) ;

begin
    _Tag.Name := string( P ) ;
end ;

{$ENDIF}


function Extract_Tag_ID( const S : string ) : string ;

var Tag : THTML_Tag ;

begin
    Tag := THTML_Tag.Create( S ) ;
    Result := Tag.Name ;
    Tag.Free ;
end ;


{$IFDEF Test}
procedure Unit_Test ;

begin
    {~~~}
end ;
{$ENDIF}


begin
    Initialize_Unit ;
end.
