{
              Program Name : LIB
              Package Name : LB
              Purpose      : Librarian routines
              Institution  : Conroy & Conroy Co.
              Date Written : 15-Dec-1988
              Written By   : Alan Conroy
              Version      : 1.0

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *            F I L E   D E S C R I P T I O N S          *
        *                                                       *
        *********************************************************

}

unit Lib ;

interface

{$I EDEFINES.INC}

{$IFDEF DELPHI}
uses SysUtils, _UEHDefs ;
{$ELSE}
uses _UEHDefs ; { TUEC}
{$ENDIF}

const LibErr_Facility = 14 ;
      LibErr_Success = 0 ;
      LibErr_Invalid_Library_File_Format = 1 ;
      LibErr_Module_Not_Found = 2 ;
      LibErr_Module_Already_Exists = 3 ;

type Block = array[ 0..255 ] of Byte ;
     TModule_Info = record
                        Name : string[ 255 ] ;
                        Size : longint ;
                        Day, Month, Year, Hour, Minute : integer ;
                        Flags : longint ;
                    end ;
     TLibrary = object
                    public { Constructors and destructors... }
                        constructor Init ;
                        destructor Done ;

                    private { Instance data... }
                        _Error : tUEC ;
                        Libr : file of Block ;
                        Lib_Buffer : Block ;
                        Opened : boolean ;
                        Day : integer ; { Day of last change }
                        Month : integer ; { Month of last change }
                        Year : integer ; { Year of last change }
                        Hour : integer ; { Hour of last change }
                        Minute : integer ; { Minute of last change }

                    public
                        Creation_Hour, Creation_Minute : Integer ;
                        Creation_Day, Creation_Month, Creation_Year : Integer ;
                        Deleted_Space : longint ;
                        Modified_Hour, Modified_Minute : Integer ;
                        Modified_Day, Modified_Month, Modified_Year : Integer ;
                        Version : string ;

                    private { Internal methods... }
                        function Get_Value( Offset : Integer ) : Longint ;
                        procedure Update ;
                        procedure Stuff( Offset : Integer ; Value : Longint ) ;

                    public { API... }
                        procedure Create( Command : String ;
                            Language_Id, Typ : Integer) ;
                        procedure Error( var UEC : tUEC ) ;
                        procedure Extract( Module, Filename : string ;
                            Start_Block : longint ) ;
                        procedure Open( X : string ) ;
                        procedure Delete( Command : string ) ;
                        procedure Insert( Command, Filename : string ;
                            Vector : Boolean ; Start_Block : Integer ) ;
                        procedure Get_Module_Info( X : integer ;
                            var Info : TModule_Info ) ;
                        function Size : longint ;
                end ;

    {Library file format:
     Block    Description
     -----    -----------
       1      header block
       2+     name and data blocks

     Header block format:
         Byte(s)    Description
         -------    -----------
         0          =0 : Good, <>0 : Bad
         1          Version (10 = V1.0)
         2          0 (Adventure Builder), 1 (Sirius), 2 (Graphics images)
         3-5        First name block
         6-8        Number of deleted blocks
         9          File type (2=Library)
         10         Day of creation
         11         Month of creation
         12         Year of creation
         13         Hour of creation
         14         Minute of creation
         15         Day of last update
         16         Month of last update
         17         Year of last update
         18         Hour of last update
         19         Minute of last update
         20         Library type (0 = OLB, 1 = ILB)

     Name block format:
         Byte(s)    Description
         -------    -----------
           0-15     Module name
          16-18     Back Link
          19-21     Forward Link
          22-24     Module start block
          25-27     Module size (in blocks)
          28        Module insertion day
          29        Module insertion month
          30        Module insertion year
          31        Module insertion hour
          32        Module insertion minute
          33        Flags:
                        AND 1 = No vector table
                        AND 2-128 = reserved
          34-255    reserved}

procedure Update ; { Update library header with current date/time for change }
procedure Insertion ;
function Delete( Command : String ) : Integer ;
function Insert( Command : String ; Vector : Boolean ; Start_Block : Integer ) :
    Integer ;
function Get_Value( Offset : Integer ) : Longint ;
procedure Get_Time ;
procedure Get_Date ;
function Create_Library( Command : String ; Language_Id, Typ : Integer ) :
    Integer ;

var Object_File, Output, Libr : file of Block ;
    Lib_Buffer : Block ;
    Hour, Minute : Integer ;
    Day, Month, Year : Integer ;

implementation

uses { Conroy & Conroy Co. }
     ErrMastr, { Register_Error_Interface }
     Num1s, { Num1 }
     Files, { FileName }
     O_S, { OS }
     UEHDefs ; { TStandard_Error_Interface }

{ tLib_EI definition and methods... }

{$IFDEF DELPHI}
type tLib_EI = class( TStandard_Error_Interface )
                            public
                                function Get_Facility : longint ; override ; stdcall ;
                                function Get_Name : PChar ; override ; stdcall ;
                                function Translate_Error( Code : longint ) : PChar ; override ; stdcall ;
                                procedure Done ; override ; stdcall ;
                                function Facility_Version : longint ;
                                    override ; stdcall ;
                                function Version : longint ;
                                    override ; stdcall ;
                  end ;
     pLib_EI = tLib_EI ;
{$ELSE}
type tLib_EI = object( TSStandard_Error_Interface )
                            public
                                function Get_Facility : longint ; virtual ;
                                function Get_Name : PChar ; virtual ;
                                function Translate_Error( Code : longint ) : PChar ; virtual ;
                                function Facility_Version : longint ;
                                    virtual ;
                                function Version : longint ;
                                    virtual ;
                  end ;
     pLib_EI = ^tLib_EI ;
{$ENDIF}

function tLib_EI.Get_Facility : longint ;

begin
    Get_Facility := LibErr_Facility ;
end ;


const Lib_Facility_Name : Pchar = 'Lib' ;

function tLib_EI.Get_Name : PChar ;

begin
    Get_Name := @Lib_Facility_Name ;
end ;


var Lib_Error_Text : array[ 0..127 ] of char ;

function tLib_EI.Translate_Error( Code : longint ) : PChar ;

var S : string ;

begin
    case Code of
        LibErr_Success : S := 'Success' ;
        LibErr_Invalid_Library_File_Format : S := 'Invalid library file format' ;
        LibErr_Module_Not_Found : S := 'Module not found' ;
        LibErr_Module_Already_Exists : S := 'Module already exists' ;
    end ;
    S := S + #0 ;
    move( S[ 1 ], Lib_Error_Text[ 0 ], length( S ) ) ;
    Translate_Error := @Lib_Error_Text ;
end ;


function tLib_EI.Facility_Version : longint ;

begin
    Facility_Version := 0 ;
end ;


function tLib_EI.Version : longint ;

begin
    Version := 0 ;
end ;


{$IFDEF DELPHI}
procedure tLib_EI.Done ;

begin
end ;
{$ENDIF}



{ Unit routines... }

function Cvtbl( X : Byte ) : Longint ; {Convert byte to long integer}

begin
    Cvtbl := X
end ;



{ TLibrary methods... }

{ Constructors and destructors... }
constructor TLibrary.Init ;

begin
    _Error.Code := 0 ;
    _Error.Facility := LibErr_Facility ;
    Opened := False ;
{$IFDEF DELPHI}
    Initialize( Version ) ;
{$ELSE}
    Version := '' ;
{$ENDIF}
end ;


destructor TLibrary.Done ;

begin
    if( Opened ) then
    begin
        close( Libr ) ;
    end ;
{$IFDEF DELPHI}
    Finalize( Version ) ;
{$ENDIF}
end ;


{ Internal routines... }

function TLibrary.Get_Value( Offset : Integer ) : Longint ;
{Get and unpack value from buffer}

begin
    Get_Value := Lib_Buffer[ Offset ] + Cvtbl( Lib_Buffer[ Offset + 1 ] ) * 256 +
        Lib_Buffer[ Offset + 2 ] * 65536
end ;


procedure TLibrary.Stuff( Offset : Integer ; Value : Longint ) ;
{Stuff a 3-byte value into the buffer}

begin
    Lib_Buffer[ Offset + 2 ] := Trunc( Value Div 65536 ) ;
    Value := Value - Lib_Buffer[ Offset + 2 ] * 65536 ;
    Lib_Buffer[ Offset + 1 ] := Trunc( Value Div 256 ) ;
    Value := Value - Cvtbl( Lib_Buffer[ Offset + 1 ] ) * 256 ;
    Lib_Buffer[ Offset ] := Trunc( Value )
end ;


procedure TLibrary.Update ;

begin
    Seek( Libr, 0 ) ;
    Read( Libr, Lib_Buffer ) ; { Read header }
    Seek( Libr, 0 ) ; { Reposition }
    Get_Date ;
    Get_Time ;
    Lib_Buffer[ 15 ] := Day ; { Day of last change }
    Lib_Buffer[ 16 ] := Month ; { Month of last change }
    Lib_Buffer[ 17 ] := Year ; { Year of last change }
    Lib_Buffer[ 18 ] := Hour ; { Hour of last change }
    Lib_Buffer[ 19 ] := Minute ; { Minute of last change }
    Modified_Day := Day ;
    Modified_Month := Month ;
    Modified_Year := Year ;
    Modified_Hour := Hour ;
    Modified_Minute := Minute ;
    Write( Libr, Lib_Buffer ) { Write updated header }
end ; { TLibrary.Update }


{ API... }

procedure TLibrary.Create( Command : String ; Language_Id, Typ : Integer ) ;

    Procedure Get_Date ;

    var Y, M, D : word ;

    Begin
        OS^.Get_Date( Y, M, D ) ;
        Month := D ;
        Day := D ;
        Year := Y - 1900 ;
        while( Year > 99 ) do
        begin
            Year := Year - 100 ;
        end ;
    End ;


    Procedure Get_Time ;

    var H, M, S, MS : word ;

    Begin
        OS^.Get_Time( H, M, S, MS ) ;
        Hour := H ;
        Minute := M ;
    End ;


begin
    if( Opened ) then
    begin
        Opened := False ;
        close( Libr ) ;
    end ;
    Assign( Libr, Command ) ;
    {$I-}
    Rewrite( Libr ) ;
    {$I+}
    _Error.Code := Ioresult ;
    _Error.Facility := LibErr_Facility ;
    If( _Error.Code <> 0 ) then
    Begin
        Exit ;
    End ;
    Lib_Buffer[ 0 ] := 0 ; {Anything but 0 indicates a bad file}
    Lib_Buffer[ 1 ] := 10 ; {Version 1.0}
    Lib_Buffer[ 2 ] := Language_Id ;
    Lib_Buffer[ 3 ] := 0 ; {First name block}
    Lib_Buffer[ 4 ] := 0 ;
    Lib_Buffer[ 5 ] := 0 ;
    Lib_Buffer[ 6 ] := 0 ; {Unused blocks count}
    Lib_Buffer[ 7 ] := 0 ;
    Lib_Buffer[ 8 ] := 0 ;
    Lib_Buffer[ 9 ] := 2 ; {Library file}
    Get_Date ;
    Get_Time ;
    Lib_Buffer[ 10 ] := Day ; {Day of creation}
    Lib_Buffer[ 11 ] := Month ; {Month of creation}
    Lib_Buffer[ 12 ] := Year ; {Year of creation}
    Lib_Buffer[ 13 ] := Hour ; {Hour of creation}
    Lib_Buffer[ 14 ] := Minute ; {Minute of creation}
    Lib_Buffer[ 15 ] := Day ; {Day of last change}
    Lib_Buffer[ 16 ] := Month ; {Month of last change}
    Lib_Buffer[ 17 ] := Year ; {Year of last change}
    Lib_Buffer[ 18 ] := Hour ; {Hour of last change}
    Lib_Buffer[ 19 ] := Minute ; {Minute of last change}
    Lib_Buffer[ 20 ] := Typ ; { Library file type }
    {$I-}
    Write( Libr, Lib_Buffer ) ;
    {$I+}
    _Error.Code := IOResult ;
    _Error.Facility := LibErr_Facility ;
    Close( Libr ) ;
end ; { TLibrary.Create }


procedure TLibrary.Error( var UEC : tUEC ) ;

begin
    UEC := _Error ;
    _Error.Code := 0 ;
    _Error.Facility := LibErr_Facility ;
end ;


procedure TLibrary.Extract( Module, Filename : string ;
    Start_Block : longint ) ;

var A : integer ;
    Inp : string[ 255 ] ;
    L, Link : longint ;
    Output : file of block ;
    Vectors : boolean ;

begin
    _Error.Code := 0 ;
    _Error.Facility := LiBErr_Facility ;
    if( not Opened ) then
    begin
        _Error.Facility := 1 ;
        _Error.Code := 103 ;
        exit ;
    end ;

    Seek( Libr, 0 ) ;
    Read( Libr, Lib_Buffer ) ;
    Link := Get_Value( 3 ) ; { Get first name link }
    while( Link <> 0 ) do
    begin
        Seek( Libr, Link ) ;
        Read( Libr, Lib_Buffer ) ;
        for A := 0 to 15 do
        begin
            Inp[ A ] := Chr( Lib_Buffer[ A ] ) ;
        end ;
        if( Inp > Module ) then { End early if module not found }
        begin
            _Error.Facility := LibErr_Facility ;
            _Error.Code := LibErr_Module_Not_Found ; { Module not found }
            exit ;
        end ;
        if( Inp = Module ) then { Found module }
        begin
            assign( Output, Filename ) ;
            {$I+}
            rewrite( Output ) ;
            {$I+}
            _Error.Code := IOResult ;
            if( _Error.Code <> 0 ) then
            begin
                _Error.Code := LibErr_Facility ;
                exit ;
            end ;

            L := Get_Value( 25 ) ; { Get size of module }
            if( ( Lib_Buffer[ 33 ] and 1 ) = 0 )
                Then Vectors := True
                Else Vectors := False ;
            for A := 1 to Trunc( L ) do
            begin
                if( ( A = 1 ) or ( A > Start_Block ) or Vectors )
                    Then
                        Read( Libr, Lib_Buffer )
                    Else
                        Fillchar( Lib_Buffer, 256, 0 ) ;
                Write( Output, Lib_Buffer )
            end ;
            close( Output ) ;
            exit ;
        end ;
        Link := Get_Value( 19 ) ; { Follow link to next module }
    end ;
    _Error.Facility := LibErr_Facility ;
    _Error.Code := LibErr_Module_Not_Found ; { Module not found }
end ; { TLibrary.Extract }


procedure TLibrary.Open( X : string ) ;

var Dummy : integer ;

begin
    _Error.Facility := LibErr_Facility ;
    _Error.Code := 0 ;
    assign( Libr, FileName( X ) ) ;
    {$I-}
    reset( Libr ) ;
    {$I+}
    _Error.Facility := 2 ;
    _Error.Code := IOResult ;
    if( _Error.Code = 0 ) then
    begin
        if( filesize( Libr ) < 1 ) then
        begin
            close( Libr ) ;
            _Error.Facility := LibErr_Facility ;
            _Error.Code := LibErr_Invalid_Library_File_Format ; { Invalid file format }
            exit ;
        end ;
        read( Libr, Lib_Buffer ) ;
        if( ( Lib_Buffer[ 0 ] <> 0 ) or ( Lib_Buffer[ 9 ] <> 2 ) ) then
        begin
            close( Libr ) ;
            _Error.Facility := LibErr_Facility ;
            _Error.Code := LibErr_Invalid_Library_File_Format ; { Invalid file format }
            exit ;
        end ;
        Opened := True ;
        Creation_Day := Lib_Buffer[ 10 ] ; { Day of creation }
        Creation_Month := Lib_Buffer[ 11 ] ; { Month of creation }
        Creation_Year := Lib_Buffer[ 12 ] ; { Year of creation }
        Creation_Hour := Lib_Buffer[ 13 ] ; { Hour of creation }
        Creation_Minute := Lib_Buffer[ 14 ] ; { Minute of creation }
        Modified_Day := Lib_Buffer[ 15 ] ; { Day of last change }
        Modified_Month := Lib_Buffer[ 16 ] ; { Month of last change }
        Modified_Year := Lib_Buffer[ 17 ] ; { Year of last change }
        Modified_Hour := Lib_Buffer[ 18 ] ; { Hour of last change }
        Modified_Minute := Lib_Buffer[ 19 ] ; { Minute of last change }
        Dummy := Lib_Buffer[ 1 ] ;
        Version := num1( Dummy div 10 ) + '.' +
            num1( Dummy - ( ( Dummy div 10 ) * 10 ) ) ;
        Deleted_Space := Get_Value( 6 ) ;
    end ; { if( _Error = 0 ) }
end ; { TLibrary.Open }


procedure TLibrary.Delete( Command : string ) ;

var A : Integer ;
    Inp : string[ 255 ] ;
    _Back_Link, _Forward_Link, L, L1, Link : Longint ;

begin
    _Error.Facility := LibErr_Facility ;
    _Error.Code := 0 ;
    if( not Opened ) then
    begin
        _Error.Facility := 2 ;
        _Error.Code := 103 ;
        exit ;
    end ;

    Seek( Libr, 0 ) ;
    Read( Libr, Lib_Buffer ) ; { Get library header }
    Link := Get_Value( 3 ) ; { Get first name link }
    while( Link <> 0 ) do
    begin
        Seek( Libr, Link ) ;
        {$I-}
        Read( Libr, Lib_Buffer ) ;
        {$I+}
        _Error.Facility := 2 ;
        _Error.Code := IOResult ;
        if( _Error.Code <> 0 ) then
        begin
            exit ;
        end ;
        For A := 0 To 15 Do
        begin
            Inp[ A ] := Chr( Lib_Buffer[ A ] ) ;
        end ;
        _Forward_Link := Get_Value( 19 ) ;
        If( Inp = Command ) then {Found module}
        Begin
            _Back_Link := Get_Value( 16 ) ;
            L1 := Get_Value( 25 ) ; {Get module size}
            If( _Back_Link > 0 ) Then {A previous record}
            Begin
                Seek( Libr, _Back_Link ) ; {Get previous record}
                {$I-}
                Read( Libr, Lib_Buffer ) ;
                {$I+}
                _Error.Facility := 2 ;
                _Error.Code := IOResult ;
                if( _Error.Code <> 0 ) then
                begin
                    exit ;
                end ;
                Stuff( 19, _Forward_Link ) ;
                Seek( Libr, _Back_Link ) ;
                {$I-}
                Write( Libr, Lib_Buffer ) ; {Patch forward link}
                {$I+}
                _Error.Facility := 2 ;
                _Error.Code := IOResult ;
                if( _Error.Code <> 0 ) then
                begin
                    exit ;
                end ;
            End ;
            If( _Forward_Link > 0 ) then {If a following record}
            Begin
                Seek( Libr, _Forward_Link ) ; {Get next record}
                {$I-}
                Read( Libr, Lib_Buffer ) ;
                {$I+}
                _Error.Code := IOResult ;
                _Error.Facility := 2 ;
                if( _Error.Code <> 0 ) then
                begin
                    exit ;
                end ;
                Stuff( 16, _Back_Link ) ;
                Seek( Libr, _Forward_Link ) ;
                {$I-}
                Write( Libr, Lib_Buffer ) ; { Patch back link }
                {$I+}
                _Error.Facility := 2 ;
                _Error.Code := IOResult ;
                if( _Error.Code <> 0 ) then
                begin
                    exit ;
                end ;
            End ;
            Seek( Libr, 0 ) ;
            Read( Libr, Lib_Buffer ) ; {Get header record}
            L := Get_Value( 6 ) + L1 + 10 ; {Add up deleted block count}
            Stuff( 6, L ) ;
            If( _Back_Link = 0 ) then
            begin
                Stuff( 3, _Forward_Link ) ;
                { If deleted module was first module }
            end ;
            Seek( Libr, 0 ) ;
            {$I-}
            Write( Libr, Lib_Buffer ) ; {Update header}
            {$I+}
            _Error.Facility := 2 ;
            _Error.Code := IOResult ;
            if( _Error.Code <> 0 ) then
            begin
                exit ;
            end ;
            Update ;
            Exit ;
        End ;
        If( Inp > Command ) then
        begin
            Link := 0 ;
        end else
        begin
            Link := _Forward_Link ;
        end ;
    End ;
    _Error.Facility := LibErr_Facility ;
    _Error.Code := LibErr_Module_Not_Found ; { Module not found }
end ; { TLibrary.Delete }


procedure TLibrary.Insert( Command, Filename : String ; Vector : Boolean ;
    Start_Block : Integer ) ;

Var A, Dummy : Integer ;
    Inp : String[ 255 ] ;
    _Back_Link, _Forward_Link, _L, L1, Last_Link, Link : Longint ;
    Object_File : file of Block ;

begin
    _Error.Facility := LibErr_Facility ;
    _Error.Code := 0 ;
    if( not Opened ) then
    begin
        _Error.Facility := 2 ;
        _Error.Code := 103 ;
        exit ;
    end ;

    assign( Object_File, Filename ) ;
    {$I-}
    reset( Object_File ) ;
    {$I+}
    _Error.Code := IOResult ;
    if( _Error.Code <> 0 ) then
    begin
        _Error.Facility := 2 ;
        exit ;
    end ;

    Seek( Libr, 0 ) ;
    Read( Libr, Lib_Buffer ) ; {Read library header}
    Link := Get_Value( 3 ) ; {Get first name link}
    _L := Filesize( Libr ) ; {Get current library size}
    _Back_Link := 0 ;
    _Forward_Link := 0 ;
    If( Link = 0 ) then {First module}
    Begin
        _Back_Link := 0 ;
        _Forward_Link := 0 ;
        Stuff( 3, _L ) ;
        Seek( Libr, 0 ) ;
        Write( Libr, Lib_Buffer ) ; { Update library header }
    End Else
    While( Link <> 0 ) do { Until end of list }
    Begin
        Seek( Libr, Link ) ;
        Read( Libr, Lib_Buffer ) ;
        For A := 0 To 15 Do Inp[ A ] := Chr( Lib_Buffer[ A ] ) ;
        If( Command = Inp ) then
        Begin
            _Error.Facility := LibErr_Facility ;
            _Error.Code := LibErr_Module_Already_Exists ; { Module already exists }
            close( Object_File ) ;
            Exit ;
        End ;
        _Back_Link := Link ;
        _Forward_Link := Link ;
        If( Inp > Command ) then { Insert new module just before this one }
        Begin
            If(
                ( Lib_Buffer[ 16 ] = 0 )
                and
                ( Lib_Buffer[ 17 ] = 0 )
                and
                ( Lib_Buffer[ 18 ] = 0 )
              ) then { No back link }
            Begin
                Stuff( 16, _L ) ;
                Seek( Libr, Link ) ;
                Write( Libr, Lib_Buffer ) ;
                _Forward_Link := Link ;
                _Back_Link := 0 ;
                Link := 0 ;
                Seek( Libr, 0 ) ;
                Read( Libr, Lib_Buffer ) ;
                Seek( Libr, 0 ) ;
                Stuff( 3, _L ) ;
                Write( Libr, Lib_Buffer ) ;
            End Else {Must insert this module between two others}
            Begin
                L1 := Get_Value( 16 ) ; {Get back link}
                Stuff( 16, _L ) ; {Set back link to new record}
                Seek( Libr, Link ) ;
                Write( Libr, Lib_Buffer ) ; {Update record}
                Seek( Libr, L1 ) ; {Get previous module}
                Read( Libr, Lib_Buffer ) ;
                Stuff( 19, _L ) ; {Set forward link}
                Seek( Libr, L1 ) ;
                Write( Libr, Lib_Buffer ) ; {Update previous record}
                _Forward_Link := Link ;
                _Back_Link := L1 ; {Set up links for new record}
                Link := 0 ; {Terminate loop}
            End ; { Follow back link }
        End Else
        Begin { Inp < Command }
            If(
                ( Lib_Buffer[ 19 ] = 0 )
                and
                ( Lib_Buffer[ 20 ] = 0 )
                and
                ( Lib_Buffer[ 21 ] = 0 )
              ) then { No forward link }
            Begin
                Stuff( 19, _L ) ;
                Seek( Libr, Link ) ;
                Write( Libr, Lib_Buffer ) ;
                _Back_Link := Link ;
                _Forward_Link := 0 ;
                Link := 0
            End Else
            Begin
                Last_Link := Link ; {Save current record number}
                L1 := Get_Value( 19 ) ; {Get forward link}
                If L1 = Last_Link Then
                Begin
                    _Forward_Link := Last_Link ;
                    _Back_Link := Link ;
                    Insertion ;
                    Link := 0
                End Else Link := L1
            End
            {Follow forward link}
        End
    end ;
    Seek( Libr, _L ) ;
    Inp := Command ;
    for A := 0 to 15 do Lib_Buffer[ A ] := Ord( Inp[ A ] ) ;
    Stuff( 16, _Back_Link ) ;
    Stuff( 19, _Forward_Link ) ;
    Stuff( 22, _L + 1 ) ;
    L1 := Filesize( Object_File ) ;
    Stuff( 25, L1 ) ;
    Get_Date ;
    Get_Time ;
    Lib_Buffer[ 28 ] := Day ; {Day of last change}
    Lib_Buffer[ 29 ] := Month ; {Month of last change}
    Lib_Buffer[ 30 ] := Year ; {Year of last change}
    Lib_Buffer[ 31 ] := Hour ; {Hour of last change}
    Lib_Buffer[ 32 ] := Minute ; {Minute of last change}
    Lib_Buffer[ 33 ] := 0 ;
    if not Vector
	then
	    Lib_Buffer[ 33 ] := Lib_Buffer[ 33 ] Or 1 ;
    Write( Libr, Lib_Buffer ) ; {Write out name block}
    Reset( Object_File ) ;
    Dummy := 0 ;
    while Not Eof( Object_File ) Do
    begin
        Read( Object_File, Lib_Buffer ) ;
        If Vector Or ( Dummy = 0 ) Or ( Dummy >= Start_Block ) Then
            Write( Libr, Lib_Buffer ) ; {Write unless vector and /NV}
        Dummy := Dummy + 1 ;
    end ;
    close( Object_File ) ;
    
    Update ;
end ; { TLibrary.Insert }


procedure TLibrary.Get_Module_Info( X : integer ; var Info : TModule_Info ) ;

var Dummy : integer ;
    L : longint ;

begin
    _Error.Facility := LibErr_Facility ;
    _Error.Code := 0 ;
    if( not Opened ) then
    begin
        _Error.Facility := 2 ;
        _Error.Code := 103 ;
        exit ;
    end ;

    seek( Libr, 0 ) ;
    {$I-}
    read( Libr, Lib_Buffer ) ;
    {$I+}
    _Error.Facility := 2 ;
    _Error.Code := IOResult ;
    if( _Error.Code <> 0 ) then
    begin
        exit ;
    end ;
    L := Get_Value( 3 ) ; { First name block }
    while( X >= 0 ) do
    begin
        if( L = 0 ) then
        begin
            _Error.Facility := LibErr_Facility ;
            _Error.Code := LibErr_Module_Not_Found ;
            exit ;
        end ;
        seek( Libr, L ) ;
        {$I-}
        read( Libr, Lib_Buffer ) ;
        {$I+}
        _Error.Code := IOResult ;
        if( _Error.Code <> 0 ) then
        begin
            _Error.Facility := 2 ;
            exit ;
        end ;
        L := Get_Value( 19 ) ;
        dec( X ) ;
    end ; { while( X >= 0 ) }

    { Now return information about the module }
    for Dummy := 0 To 15 Do
    begin
        Info.Name[ Dummy ] := Chr( Lib_Buffer[ Dummy ] ) ;
    end ;
    Info.Size := Get_Value( 25 ) * 256 ;
    Info.Day := Lib_Buffer[ 28 ] ;
    Info.Month := Lib_Buffer[ 29 ] ;
    Info.Year := Lib_Buffer[ 30 ] ;
    Info.Hour := Lib_Buffer[ 31 ] ;
    Info.Minute := Lib_Buffer[ 32 ] ;
    Info.Flags := Lib_Buffer[ 33 ] ;
end ; { TLibrary.Get_Module }


function TLibrary.Size : longint ;

begin
    _Error.Facility := LibErr_Facility ;
    _Error.Code := 0 ;
    Size := 0 ;
    if( not Opened ) then
    begin
        _Error.Facility := 2 ;
        _Error.Code := 103 ;
        exit ;
    end ;

    Size := filesize( Libr ) * 256 ;
end ;


var L, Back_Link, Forward_Link : Longint ; { File link }

Procedure Stuff( Offset : Integer ; Value : Longint ) ;
{Stuff a 3-byte value into the buffer}

Begin
    Lib_Buffer[ Offset + 2 ] := Trunc( Value Div 65536 ) ;
    Value := Value - Lib_Buffer[ Offset + 2 ] * 65536 ;
    Lib_Buffer[ Offset + 1 ] := Trunc( Value Div 256 ) ;
    Value := Value - Cvtbl( Lib_Buffer[ Offset + 1 ] ) * 256 ;
    Lib_Buffer[ Offset ] := Trunc( Value )
End ;


Function Get_Value( Offset : Integer ) : Longint ;
{Get and unpack value from buffer}

Begin
    Get_Value := Lib_Buffer[ Offset ] + Cvtbl( Lib_Buffer[ Offset + 1 ] ) * 256 +
        Lib_Buffer[ Offset + 2 ] * 65536
End ;


Procedure Get_Date ;

var Y, M, D : word ;

Begin
    OS^.Get_Date( Y, M, D ) ;
    Month := M ;
    Day := D ;
    Year := Y - 1900 ;
    While( Year > 99 ) do
    begin
        Year := Year - 100 ;
    end ;
End ;


Procedure Get_Time ;

var H, M, S, MS : word ;

Begin
    OS^.Get_Time( H, M, S, MS ) ;
    Hour := H ;
    Minute := M ;
End ;


Procedure Update ; { Update library header with current date/time for change }

Begin
    Seek( Libr, 0 ) ;
    Read( Libr, Lib_Buffer ) ; {Read header}
    Seek( Libr, 0 ) ; {Reposition}
    Get_Date ;
    Get_Time ;
    Lib_Buffer[ 15 ] := Day ; {Day of last change}
    Lib_Buffer[ 16 ] := Month ; {Month of last change}
    Lib_Buffer[ 17 ] := Year ; {Year of last change}
    Lib_Buffer[ 18 ] := Hour ; {Hour of last change}
    Lib_Buffer[ 19 ] := Minute ; {Minute of last change}
    Write( Libr, Lib_Buffer ) {Write updated header}
End ;


Procedure Insertion ;
{Change links so as to insert this module (name).  L is the name record #}

Begin
    Seek( Libr, Back_Link ) ; {Get previous logical record}
    Read( Libr, Lib_Buffer ) ;
    Stuff( 19, L ) ; {Set forward link}
    Seek( Libr, Back_Link ) ;
    Write( Libr, Lib_Buffer ) ; {Update record}
    Seek( Libr, Forward_Link ) ; {Get next logical record}
    Read( Libr, Lib_Buffer ) ;
    Stuff( 16, L ) ; {Set back link}
    Seek( Libr, Forward_Link ) ; {Update record}
    Write( Libr, Lib_Buffer )
End ;


Function Delete( Command : String ) : Integer ;
{Delete module specified in Command}

Var A : Integer ;
    Inp : String[ 255 ] ;
    Back_Link, Forward_Link, L, L1, Link : Longint ;

Begin
    Seek( Libr, 0 ) ;
    Read( Libr, Lib_Buffer ) ; {Get library header}
    Link := Get_Value( 3 ) ; {Get first name link}
    while( Link <> 0 ) do
    begin
        Seek( Libr, Link ) ;
        Read( Libr, Lib_Buffer ) ;
        For A := 0 To 15 Do Inp[ A ] := Chr( Lib_Buffer[ A ] ) ;
        Forward_Link := Get_Value( 19 ) ;
        If Inp = Command Then {Found module}
        Begin
            Back_Link := Get_Value( 16 ) ;
            L1 := Get_Value( 25 ) ; {Get module size}
            If Back_Link > 0 Then {A previous record}
            Begin
                Seek( Libr, Back_Link ) ; {Get previous record}
                Read( Libr, Lib_Buffer ) ;
                Stuff( 19, Forward_Link ) ;
                Seek( Libr, Back_Link ) ;
                Write( Libr, Lib_Buffer ) {Patch forward link}
            End ;
            If Forward_Link > 0 Then {If a following record}
            Begin
                Seek( Libr, Forward_Link ) ; {Get next record}
                Read( Libr, Lib_Buffer ) ;
                Stuff( 16, Back_Link ) ;
                Seek( Libr, Forward_Link ) ;
                Write( Libr, Lib_Buffer ) {Patch back link}
            End ;
            Seek( Libr, 0 ) ;
            Read( Libr, Lib_Buffer ) ; {Get header record}
            L := Get_Value( 6 ) + L1 + 10 ; {Add up deleted block count}
            Stuff( 6, L ) ;
            If Back_Link = 0 Then Stuff( 3, Forward_Link ) ;
            {If deleted module was first module}
            Seek( Libr, 0 ) ;
            Write( Libr, Lib_Buffer ) ; {Update header}
            Update ;
            Delete := 0 ;
            Exit
        End ;
        If Inp > Command Then Link := 0 Else Link := Forward_Link
    End ;
    Delete := 1 ; {Module not found}
End ;


Function Insert( Command : String ; Vector : Boolean ; Start_Block : Integer ) :
    Integer ;
{This function inserts a module with the name in Command, from the Object file.
 If Vector is false, we don't include vector table.}

Var A, Dummy : Integer ;
    Inp : String[ 255 ] ;
    Back_Link, Forward_Link, L, L1, Last_Link, Link : Longint ;

Begin
    Seek( Libr, 0 ) ;
    Read( Libr, Lib_Buffer ) ; {Read library header}
    Link := Get_Value( 3 ) ; {Get first name link}
    L := Filesize( Libr ) ; {Get current library size}
    Back_Link := 0 ;
    Forward_Link := 0 ;
    If( Link = 0 ) then {First module}
    Begin
        Back_Link := 0 ;
        Forward_Link := 0 ;
        Stuff( 3, L ) ;
        Seek( Libr, 0 ) ;
        Write( Libr, Lib_Buffer ) ; { Update library header }
    End Else
    While( Link <> 0 ) do {Until end of list}
    Begin
        Seek( Libr, Link ) ;
        Read( Libr, Lib_Buffer ) ;
        For A := 0 To 15 Do Inp[ A ] := Chr( Lib_Buffer[ A ] ) ;
        If( Command = Inp ) then
        Begin
            Insert := 2 ;
            Exit ;
        End ;
        Back_Link := Link ;
        Forward_Link := Link ;
        If( Inp > Command ) then { Insert new module just before this one }
        Begin
            If(
                ( Lib_Buffer[ 16 ] = 0 )
                and
                ( Lib_Buffer[ 17 ] = 0 )
                and
                ( Lib_Buffer[ 18 ] = 0 )
              ) then { No back link }
            Begin
                Stuff( 16, L ) ;
                Seek( Libr, Link ) ;
                Write( Libr, Lib_Buffer ) ;
                Forward_Link := Link ;
                Back_Link := 0 ;
                Link := 0 ;
                Seek( Libr, 0 ) ;
                Read( Libr, Lib_Buffer ) ;
                Seek( Libr, 0 ) ;
                Stuff( 3, L ) ;
                Write( Libr, Lib_Buffer ) ;
            End Else {Must insert this module between two others}
            Begin
                L1 := Get_Value( 16 ) ; {Get back link}
                Stuff( 16, L ) ; {Set back link to new record}
                Seek( Libr, Link ) ;
                Write( Libr, Lib_Buffer ) ; {Update record}
                Seek( Libr, L1 ) ; {Get previous module}
                Read( Libr, Lib_Buffer ) ;
                Stuff( 19, L ) ; {Set forward link}
                Seek( Libr, L1 ) ;
                Write( Libr, Lib_Buffer ) ; {Update previous record}
                Forward_Link := Link ;
                Back_Link := L1 ; {Set up links for new record}
                Link := 0 ; {Terminate loop}
            End ; { Follow back link }
        End Else
        Begin { Inp < Command }
            If(
                ( Lib_Buffer[ 19 ] = 0 )
                and
                ( Lib_Buffer[ 20 ] = 0 )
                and
                ( Lib_Buffer[ 21 ] = 0 )
              ) then { No forward link }
            Begin
                Stuff( 19, L ) ;
                Seek( Libr, Link ) ;
                Write( Libr, Lib_Buffer ) ;
                Back_Link := Link ;
                Forward_Link := 0 ;
                Link := 0
            End Else
            Begin
                Last_Link := Link ; {Save current record number}
                L1 := Get_Value( 19 ) ; {Get forward link}
                If L1 = Last_Link Then
                Begin
                    Forward_Link := Last_Link ;
                    Back_Link := Link ;
                    Insertion ;
                    Link := 0
                End Else Link := L1
            End
            {Follow forward link}
        End
    End ;
    Seek( Libr, L ) ;
    Inp := Command ;
    For A := 0 To 15 Do Lib_Buffer[ A ] := Ord( Inp[ A ] ) ;
    Stuff( 16, Back_Link ) ;
    Stuff( 19, Forward_Link ) ;
    Stuff( 22, L + 1 ) ;
    L1 := Filesize( Object_File ) ;
    Stuff( 25, L1 ) ;
    Get_Date ;
    Get_Time ;
    Lib_Buffer[ 28 ] := Day ; {Day of last change}
    Lib_Buffer[ 29 ] := Month ; {Month of last change}
    Lib_Buffer[ 30 ] := Year ; {Year of last change}
    Lib_Buffer[ 31 ] := Hour ; {Hour of last change}
    Lib_Buffer[ 32 ] := Minute ; {Minute of last change}
    Lib_Buffer[ 33 ] := 0 ;
    If Not Vector
	Then
	    Lib_Buffer[ 33 ] := Lib_Buffer[ 33 ] Or 1 ;
    Write( Libr, Lib_Buffer ) ; {Write out name block}
    Reset( Object_File ) ;
    Dummy := 0 ;
    While Not Eof( Object_File ) Do
    Begin
        Read( Object_File, Lib_Buffer ) ;
        If Vector Or ( Dummy = 0 ) Or ( Dummy >= Start_Block ) Then
            Write( Libr, Lib_Buffer ) ; {Write unless vector and /NV}
        Dummy := Dummy + 1 ;
    End ;
    Insert := 0 ;
End ;


Function Create_Library( Command : String ; Language_Id, Typ : Integer ) :
    Integer ;

var A : Integer ;

begin
    Assign( Libr, Command ) ;
    {$I-}
    Rewrite( Libr ) ;
    {$I+}
    A := Ioresult ;
    If A <> 0 Then
    Begin
	Create_Library := A ;
        Exit ;
    End ;
    Create_Library := 0 ;
    Lib_Buffer[ 0 ] := 0 ; {Anything but 0 indicates a bad file}
    Lib_Buffer[ 1 ] := 10 ; {Version 1.0}
    Lib_Buffer[ 2 ] := Language_Id ;
    Lib_Buffer[ 3 ] := 0 ; {First name block}
    Lib_Buffer[ 4 ] := 0 ;
    Lib_Buffer[ 5 ] := 0 ;
    Lib_Buffer[ 6 ] := 0 ; {Unused blocks count}
    Lib_Buffer[ 7 ] := 0 ;
    Lib_Buffer[ 8 ] := 0 ;
    Lib_Buffer[ 9 ] := 2 ; {Library file}
    Get_Date ;
    Get_Time ;
    Lib_Buffer[ 10 ] := Day ; {Day of creation}
    Lib_Buffer[ 11 ] := Month ; {Month of creation}
    Lib_Buffer[ 12 ] := Year ; {Year of creation}
    Lib_Buffer[ 13 ] := Hour ; {Hour of creation}
    Lib_Buffer[ 14 ] := Minute ; {Minute of creation}
    Lib_Buffer[ 15 ] := Day ; {Day of last change}
    Lib_Buffer[ 16 ] := Month ; {Month of last change}
    Lib_Buffer[ 17 ] := Year ; {Year of last change}
    Lib_Buffer[ 18 ] := Hour ; {Hour of last change}
    Lib_Buffer[ 19 ] := Minute ; {Minute of last change}
    Lib_Buffer[ 20 ] := Typ ; {Library file type}
    {$I-}
    Write( Libr, Lib_Buffer ) ;
    {$I+}
    A := Ioresult ;
    If A <> 0 then
    begin
	Create_Library := A ;
	Exit ;
    end ;
    Close( Libr ) ;
end ;


var Lib_EI : pLib_EI ;

procedure Unit_Initialization ;

begin
{$IFDEF DELPHI}
    Lib_EI := tLib_EI.Create ;
    Register_Error_Interface( PSError_Interface( Lib_EI ) ) ;
{$ELSE}
    new( Lib_EI, Init ) ;
    Register_Error_Interface( @Lib_EI ) ;
{$ENDIF}
end ;


procedure Unit_Deallocation ;

begin
{$IFDEF DELPHI}
    Unregister_Error_Interface( Lib_EI.Get_Facility ) ;
    Lib_EI.Free ;
{$ELSE}
    Unregister_Error_Interface( Lib_EI^.Get_Facility ) ;
    dispose( Lib_EI ) ;
{$ENDIF}
end ;


{$IFDEF DELPHI}
initialization
{$ELSE}
begin
{$ENDIF}
    Unit_Initialization ;
{$IFDEF DELPHI}

finalization
    Unit_Deallocation ;
{$ENDIF}
end.
