{
	      Program Name : Lines
	      Package Name : LB
	      Purpose      : 2D graphics support
	      Institution  : Conroy & Conroy
	      Date Written : 21-Jul-1999
	      Written By   : Alan Conroy
	      Version      : 1.0

              Copyright (C) 1999 by Alan Conroy.  Released to the public domain.

	      TO THE GLORY OF GOD THROUGH JESUS CHRIST

	*********************************************************
	*                                                       *
	*        M O D I F I C A T I O N   H I S T O R Y        *
	*                                                       *
	*     DATE      BY             REASON                   *
	*                                                       *
	*********************************************************

	*********************************************************
	*                                                       *
	*           P R O G R A M   P U R P O S E               *
	*                                                       *
	*********************************************************

	  2-D graphics primitives.
}

unit Lines ;

interface

uses // Borland...
     Windows,   // TPoint

     Classes,   // TList
     Graphics,  // TCanvas

     // C&C...
     TypeDefs ;    // PInteger_Array

type TTexture = class
                    public // Instance data
                        Height, Width : integer ; // Image dimensions
                        _Array : PInteger_Array ; // Image

                    public // API...
                        destructor Destroy ; override ;
                        procedure Clear ;
                        procedure Load_From_Bitmap( Bitmap : TBitmap ) ;
                end ;

     TXY_List = class
                    private // Instance data
                        X_List : TList ;
                        Y_List : TList ;

                    protected // Property handlers
                        function Get_X( Index : integer ) : TPoint ;

                        procedure Set_X( Index : integer ; Value : TPoint ) ;

                    public // API
                        constructor Create ; virtual ;
                        destructor Destroy ; override ;

                        function Count : integer ;
                        procedure Clear ; virtual ;
                        procedure Generate( A, B : TPoint ) ;

                        property X[ Index : integer ] : TPoint
                                     read Get_X
                                     write Set_X ;
                                     default ;
                end ;

     TPolygon = class( TXY_List )
                    private // Instance data
                        Bitmap_X_List : TList ;
                        Bitmap_Y_List : TList ;
                        Vertex_X_List : TList ;
                        Vertex_Y_List : TList ;
			Lines_Left_X : TList ;
			Lines_Y : TList ;
			Lines_Right_X : TList ;
                        Lines_Bitmap_X_Right : TList ;
                        Lines_Bitmap_Y_Right : TList ;
                        Lines_Bitmap_X_Left : TList ;
                        Lines_Bitmap_Y_Left : TList ;

                    protected // Property handlers
                        function Get_X( Index : integer ) : TPoint ;

                        procedure Set_X( Index : integer ; Value : TPoint ) ;

                    public // API
                        constructor Create ; override ;
                        destructor Destroy ; override ;

                        procedure Add( X : TPoint ) ;
                        procedure Clear ; override ;
                        procedure Generate ;
                        procedure Generate_Texture_Line( Bitmap, Other : integer ;
                            B_X : boolean ; A, B : TPoint ) ;
                        procedure Generate_With_Texture( Bitmap : TTexture ) ;
			function Line_Count : integer ;
			function Line_Left( Index : integer ) : TPoint ;
			function Line_Right( Index : integer ) : TPoint ;
                        function Line_Left_Texture( Index : integer ) : TPoint ;
                        function Line_Right_Texture( Index : integer ) : TPoint ;
			procedure MinMax ;
                        function Vertex_Count : integer ;

                        property Vertex[ Index : integer ] : TPoint
                                     read Get_X
                                     write Set_X ;
                end ;

procedure Draw_Line( C : TCanvas ; A, B : TPoint ) ;
procedure Polygon( C : TCanvas ; Points : Array of TPoint ) ;
procedure Filled_Polygon( C : TCanvas ; Points : Array of TPoint ) ;
procedure Draw_Polygon( C : TCanvas ; Bitmap : TTexture ; Points : Array of TPoint ) ;

implementation

uses // Borland...
     SysUtils ; // Allocmem

procedure Draw_Line( C : TCanvas ; A, B : TPoint ) ;

var i : integer ;
    List : TXY_List ;

begin
    List := TXY_List.Create ;
    List.Generate( A, B ) ;
    for i := 0 to List.Count - 1 do
    begin
        C.Pixels[ List[ i ].X, List[ i ].Y ] := clBlack ;
    end ;
    List.Free ;
end ;


procedure Polygon( C : TCanvas ; Points : Array of TPoint ) ;

var Loop : integer ;
    P : TPolygon ;

begin
    P := TPolygon.Create ;
    for Loop := low( Points ) to high( Points ) do
    begin
        P.Add( Points[ Loop ] ) ;
    end ;
    P.Generate ;
    for Loop := 0 to P.Count - 1 do
    begin
        C.Pixels[ P[ Loop ].X, P[ Loop ].Y ] := clBlack ;
    end ;
    P.Free ;
end ;


procedure Filled_Polygon( C : TCanvas ; Points : Array of TPoint ) ;

var Loop : integer ;
    P : TPolygon ;
    Pt : TPoint ;

begin
    P := TPolygon.Create ;
    for Loop := low( Points ) to high( Points ) do
    begin
        P.Add( Points[ Loop ] ) ;
    end ;
    P.Generate ;
    P.MinMax ;
    for Loop := 0 to P.Line_Count - 1 do
    begin
        C.PenPos := P.Line_Left( Loop ) ;
        Pt := P.Line_Right( Loop ) ;
        C.LineTo( Pt.X, Pt.Y ) ;
    end ;
    P.Free ;
end ;


procedure Draw_Polygon( C : TCanvas ; Bitmap : TTexture ; Points : Array of TPoint ) ;

var B_X, B_Y : integer ; // Current texture bitmap positions
    B_Left, B_Right : TPoint ; // Endpoints of current texture path
    Delta_X, Delta_Y : integer ; // Deltas to add to texture positions
    Line_Length : integer ;
    Loop : integer ;
    P : TPolygon ;
    Pixel : integer ;
    Pt_Left, Pt_Right: TPoint ;

begin
    P := TPolygon.Create ;
    for Loop := low( Points ) to high( Points ) do
    begin
        P.Add( Points[ Loop ] ) ;
    end ;
    P.Generate_With_Texture( Bitmap ) ;
    P.MinMax ;
    for Loop := 0 to P.Line_Count - 1 do // For each line
    begin
        // Get line left/right end points...
        Pt_Right := P.Line_Right( Loop ) ;
        Pt_Left := P.Line_Left( Loop ) ;

        // Get texture path end points...
        B_Left := P.Line_Left_Texture( Loop ) ;
        B_Right := P.Line_Right_Texture( Loop ) ;

        // Determine X and Y deltas through the texture...
        Delta_X := abs( B_Left.X - B_Right.X ) ;
        Delta_Y := abs( B_Left.Y - B_Right.Y ) ;
        Line_Length := abs( Pt_Right.X - Pt_Left.X + 1 ) ;
        Delta_X := ( Delta_X shl 16 ) div Line_Length ;
        Delta_Y := ( Delta_Y shl 16 ) div Line_Length ;
        if( B_Left.X > B_Right.X ) then
        begin
            Delta_X := -Delta_X ;
        end ;
        if( B_Left.Y > B_Right.Y ) then
        begin
            Delta_Y := -Delta_Y ;
        end ;
        B_X := B_Left.X shl 16 ;
        B_Y := B_Left.Y shl 16 ;
                                                
        // Draw the line
        for Pixel := Pt_Left.X to Pt_Right.X do
        begin
            C.Pixels[ Pixel, Pt_Left.Y ] :=
                Bitmap._Array[ B_X shr 16 + ( B_Y shr 16 ) * Bitmap.Width ] ;
            B_X := B_X + Delta_X ;
            B_Y := B_Y + Delta_Y ;
        end ;
    end ; // for Loop := 0 to P.Line_Count - 1 do
    P.Free ;
end ; // Draw_Polygon


// TTexture methods...

// API...
destructor TTexture.Destroy ;

begin
    Clear ;

    inherited Destroy ;
end ;


procedure TTexture.Clear ;

begin
    if( _Array <> nil ) then
    begin
        freemem( _Array ) ;
        Height := 0 ;
        Width := 0 ;
    end ;
end ;


procedure TTexture.Load_From_Bitmap( Bitmap : TBitmap ) ;

var I, X, Y : integer ;

begin
    Clear ;
    Height := Bitmap.Height ;
    Width := Bitmap.Width ;
    _Array := Allocmem( Height * Width * sizeof( integer ) ) ;
    I := 0 ;
    for Y := 0 to Height - 1 do
    begin
        for X := 0 to Width - 1 do
        begin
            _Array[ I ] := integer( Bitmap.Canvas.Pixels[ X, Y ] ) ;
            inc( I ) ;
        end ;
    end ;
end ;



// TXY_List methods...

// Property handlers...

function TXY_List.Get_X( Index : integer ) : TPoint ;

begin
    Result.X := integer( X_List[ Index ] ) ;
    Result.Y := integer( Y_List[ Index ] ) ;
end ;


procedure TXY_List.Set_X( Index : integer ; Value : TPoint ) ;

begin
    X_List[ Index ] := pointer( Value.X ) ;
    Y_List[ Index ] := pointer( Value.Y ) ;
end ;


// API...

constructor TXY_List.Create ;

begin
    inherited Create ;

    X_List := TList.Create ;
    Y_List := TList.Create ;
end ;


destructor TXY_List.Destroy ;

begin
    X_List.Free ;
    Y_List.Free ;
    inherited Destroy ;
end ;


function TXY_List.Count : integer ;

begin
    Result := X_List.Count ;
end ;


procedure TXY_List.Clear ;

begin
    // Clear out old points...
    X_List.Clear ;
    Y_List.Clear ;
end ;


procedure TXY_List.Generate( A, B : TPoint ) ;

var C : TPoint ;
    New_X, New_Y : integer ;
    Count : integer ; // Number of steps in dependant access for each step in the independant
    i : integer ; // Loop count
    X_Delta, Y_Delta : integer ;

begin
    // Setup...
    X_Delta := abs( A.X - B.X ) ;
    Y_Delta := abs( A.Y - B.Y ) ;
    if( X_Delta > Y_Delta ) then
    begin
        Count := ( Y_Delta shl 16 ) div X_Delta  ;
        X_List.Capacity := X_List.Capacity + X_Delta + 1 ;
        Y_List.Capacity := X_List.Capacity ;
        if( A.X > B.X ) then // If inverted, swap the points
        begin
            C := B ;
            B := A ;
            A := C ;
        end ;
        New_Y := A.Y shl 16 ;
        for i := A.X to B.X do
        begin
            X_List.Add( pointer( i ) ) ;
            Y_List.Add( pointer( New_Y shr 16 ) ) ;
            New_Y := New_Y + Count ;
        end ;
    end else
    begin
        Count := ( X_Delta shl 16 ) div Y_Delta ;
        X_List.Capacity := X_List.Capacity + Y_Delta + 1 ;
        Y_List.Capacity := X_List.Capacity ;
        if( A.Y > B.Y ) then // If inverted, swap the points
        begin
            C := B ;
            B := A ;
            A := C ;
        end ;
        New_X := A.X shl 16 ;
        for i := A.Y to B.Y do
        begin
            Y_List.Add( pointer( i ) ) ;
            X_List.Add( pointer( New_X shr 16 ) ) ;
            New_X := New_X + Count ;
        end ;
    end ;
end ;



// TPolygon methods...

// Property handlers

function TPolygon.Get_X( Index : integer ) : TPoint ;

begin
    Result.X := integer( Vertex_X_List[ Index ] ) ;
    Result.Y := integer( Vertex_Y_List[ Index ] ) ;
end ;


procedure TPolygon.Set_X( Index : integer ; Value : TPoint ) ;

begin
    Vertex_X_List[ Index ] := pointer( Value.X ) ;
    Vertex_Y_List[ Index ] := pointer( Value.Y ) ;
end ;


// API

constructor TPolygon.Create ;

begin
    inherited Create ;

    Vertex_X_List := TList.Create ;
    Vertex_Y_List := TList.Create ;
    Bitmap_X_List := TList.Create ;
    Bitmap_Y_List := TList.Create ;
end ;


destructor TPolygon.Destroy ;

begin
    Vertex_X_List.Free ;
    Vertex_Y_List.Free ;
    Bitmap_X_List.Free ;
    Bitmap_Y_List.Free ;

    inherited Destroy ;
end ;


procedure TPolygon.Add( X : TPoint ) ;

begin
    Vertex_X_List.Add( pointer( X.X ) ) ;
    Vertex_Y_List.Add( pointer( X.Y ) ) ;
end ;


procedure TPolygon.Clear ;

begin
    Vertex_X_List.Clear ;
    Vertex_Y_List.Clear ;
end ;


procedure TPolygon.Generate ;

var Loop : integer ;

begin
    X_List.Clear ;
    Y_List.Clear ;

    if( Vertex_Count < 2 ) then
    begin
        exit ;
    end ;

    for Loop := 0 to Vertex_Count - 2 do
    begin
        inherited Generate( Vertex[ Loop ], Vertex[ Loop + 1 ] ) ;
    end ;
    if( Vertex_Count > 2 ) then
    begin
        inherited Generate( Vertex[ Vertex_Count - 1 ], Vertex[ 0 ] ) ;
    end ;
end ;


{ Generate a line of points with matching texture X,Y positions. Bitmap is the
  size of the side of the texture we are traversing.  Other is a constant.  B_X
  is true if we are traversing the X-axis of the bitmap.  If Bitmap is negative,
  we are traversing from the lower right, otherwise from the upper left.  If
  B_X is true, Bitmap is the Width otherwise it is the Height of the texture.
  Other is the maximum of the other dimension (0, Height, or Width).  A and B
  are the end points for this line in the polygon. }
procedure TPolygon.Generate_Texture_Line( Bitmap, Other : integer ; B_X : boolean ;
    A, B : TPoint ) ;

var C : TPoint ;
    New_X, New_Y : integer ;
    Count : integer ; // Number of steps in dependant access for each step in the independant
    Bit : integer ; // Current bitmap position
    B_Count : integer ; // Number of steps across passed bitmap range
    i : integer ; // Loop count
    X_Delta, Y_Delta : integer ;

begin
    // Setup...
    X_Delta := abs( A.X - B.X ) ;
    Y_Delta := abs( A.Y - B.Y ) ;
    if( Bitmap >= 0 ) then
    begin
        Bit := 0 ;
    end else
    begin
        Bit := ( -Bitmap ) shl 16 ;
    end ;

    // Calculate the points
    if( X_Delta > Y_Delta ) then
    begin
        B_Count := ( abs( Bitmap ) shl 16 ) div X_Delta ;
        if( Bitmap < 0 ) then // Counting backwards
        begin
            B_Count := -B_Count ;
        end ;
        Count := ( Y_Delta shl 16 ) div X_Delta  ;
        X_List.Capacity := X_List.Capacity + X_Delta + 1 ;
        Y_List.Capacity := X_List.Capacity ;
        Bitmap_X_List.Capacity := X_List.Capacity ;
        Bitmap_Y_List.Capacity := X_List.Capacity ;
        if( A.X > B.X ) then // If inverted, swap the points
        begin
            C := B ;
            B := A ;
            A := C ;
            if( Bit = 0 ) then
            begin
                Bit := ( abs( Bitmap ) - 1 ) shl 16 ;
            end else
            begin
                Bit := 0 ;
            end ;
            B_Count := -B_Count ;
        end ;
        if( A.Y > B.Y ) then
        begin
            Count := -Count ;
        end ;
        New_Y := A.Y shl 16 ;
        for i := A.X to B.X do
        begin
            X_List.Add( pointer( i ) ) ;
            Y_List.Add( pointer( New_Y shr 16 ) ) ;
            New_Y := New_Y + Count ;
            if( B_X ) then // Stepping across the X-axis of the bitmap
            begin
                Bitmap_X_List.Add( pointer( Bit shr 16 ) ) ;
                Bitmap_Y_List.Add( pointer( Other ) ) ;
            end else
            begin
                Bitmap_X_List.Add( pointer( Other ) ) ;
                Bitmap_Y_List.Add( pointer( Bit shr 16 ) ) ;
            end ;
            Bit := Bit + B_Count ;
        end ; // for i := A.X to B.X
    end else
    begin
        B_Count := ( abs( Bitmap ) shl 16 ) div Y_Delta ;
        if( Bitmap < 0 ) then // Counting backwards
        begin
            B_Count := -B_Count ;
        end ;
        Count := ( X_Delta shl 16 ) div Y_Delta ;
        X_List.Capacity := X_List.Capacity + Y_Delta + 1 ;
        Y_List.Capacity := X_List.Capacity ;
        Bitmap_X_List.Capacity := X_List.Capacity ;
        Bitmap_Y_List.Capacity := X_List.Capacity ;
        if( A.Y > B.Y ) then // If inverted, swap the points
        begin
            C := B ;
            B := A ;
            A := C ;
            if( Bit = 0 ) then
            begin
                Bit := ( abs( Bitmap ) - 1 ) shl 16 ;
            end else
            begin
                Bit := 0 ;
            end ;
            B_Count := -B_Count ;
        end ;
        if( A.X > B.X ) then
        begin
            Count := -Count ;
        end ;
        New_X := A.X shl 16 ;
        for i := A.Y to B.Y do
        begin
            Y_List.Add( pointer( i ) ) ;
            X_List.Add( pointer( New_X shr 16 ) ) ;
            New_X := New_X + Count ;
            if( B_X ) then // Stepping across the X-axis of the bitmap
            begin
                Bitmap_X_List.Add( pointer( Bit shr 16 ) ) ;
                Bitmap_Y_List.Add( pointer( Other ) ) ;
            end else
            begin
                Bitmap_X_List.Add( pointer( Other ) ) ;
                Bitmap_Y_List.Add( pointer( Bit shr 16 ) ) ;
            end ;
            Bit := Bit + B_Count ;
        end ; // for i := A.Y to B.Y
    end ; // if( X_Delta > Y_Delta )
end ; // TPolygon.Generate_Texture_Line



procedure TPolygon.Generate_With_Texture( Bitmap : TTexture ) ;

var Loop : integer ;

begin
    X_List.Clear ;
    Y_List.Clear ;

    if( Vertex_Count < 2 ) then
    begin
        exit ;
    end ;

    // Do upper-left to upper-right (top line)
    Generate_Texture_Line( Bitmap.Width, 0, True, Vertex[ 0 ], Vertex[ 1 ] ) ;
    if( Vertex_Count > 2 ) then // Not just a line
    begin
        // Do right-hand side
        Generate_Texture_Line( Bitmap.Height, Bitmap.Width - 1, False, Vertex[ 1 ], Vertex[ 2 ] ) ;
        if( Vertex_Count > 3 ) then // More sides than a triangle
        begin
            if( Vertex_Count > 4 ) then // More sides than a rectangle
            begin
                for Loop := 4 to Vertex_Count - 2 do
                begin
                    Generate_Texture_Line( Bitmap.Height, Bitmap.Width - 1, False, Vertex[ Loop ], Vertex[ Loop + 1 ] ) ;
                end ; // for Loop := 0 to Vertex_Count - 2
            end else
            begin
                // Do bottom of rectangle
                Generate_Texture_Line( -Bitmap.Width, Bitmap.Height - 1, True, Vertex[ 2 ], Vertex[ 3 ] ) ;
                // Do left-hand side of rectangle
                Generate_Texture_Line( -Bitmap.Height, 0, False, Vertex[ 3 ], Vertex[ 0 ] ) ;
            end ;
        end else
        begin
            // Do "left" side of triangle
            Generate_Texture_Line( -Bitmap.Height, 0, False, Vertex[ 2 ], Vertex[ 0 ] ) ;
        end ; // if( Vertex_Count > 2 )
    end ; // if( Vertex_Count > 2 )
end ; // TPolygon.Generate_With_Texture


function TPolygon.Line_Count : integer ;

begin
    Result := Lines_Left_X.Count ;
end ;


function TPolygon.Line_Left( Index : integer ) : TPoint ;

begin
    Result.X := integer( Lines_Left_X.List[ Index ] ) ;
    Result.Y := integer( Lines_Y.List[ Index ] ) ;
end ;


function TPolygon.Line_Right( Index : integer ) : TPoint ;

begin
    Result.X := integer( Lines_Right_X.List[ Index ] ) ;
    Result.Y := integer( Lines_Y.List[ Index ] ) ;
end ;


function TPolygon.Line_Left_Texture( Index : integer ) : TPoint ;

begin
    Result.X := integer( Lines_Bitmap_X_Left.List[ Index ] ) ;
    Result.Y := integer( Lines_Bitmap_Y_Left.List[ Index ] ) ;
end ;


function TPolygon.Line_Right_Texture( Index : integer ) : TPoint ;

begin
    Result.X := integer( Lines_Bitmap_X_Right.List[ Index ] ) ;
    Result.Y := integer( Lines_Bitmap_Y_Right.List[ Index ] ) ;
end ;


procedure TPolygon.MinMax ;

var Loop, Max_Y, Min_Y, This_Y : integer ;
    P : TPoint ;

begin
    // Initialize the matched X,Y lists
    if( Lines_Left_X = nil ) then
    begin
	Lines_Left_X := TList.Create ;
	Lines_Y := TList.Create ;
	Lines_Right_X := TList.Create ;
        Lines_Bitmap_X_Right := TList.Create ;
        Lines_Bitmap_Y_Right := TList.Create ;
        Lines_Bitmap_X_Left := TList.Create ;
        Lines_Bitmap_Y_Left := TList.Create ;
    end else
    begin
	Lines_Left_X.Clear ;
	Lines_Y.Clear ;
	Lines_Right_X.Clear ;
        Lines_Bitmap_X_Right.Clear ;
        Lines_Bitmap_Y_Right.Clear ;
        Lines_Bitmap_X_Left.Clear ;
        Lines_Bitmap_Y_Left.Clear ;
    end ;

    // Determine minimum and maximum Y values...
    Min_Y := 32767 ;
    Max_Y := 0 ;
    for Loop := 0 to Vertex_Count - 1 do
    begin
	This_Y := Vertex[ Loop ].Y ;
	if( This_Y < Min_Y ) then
	begin
	    Min_Y := This_Y ;
	end ;
	if( This_Y > Max_Y ) then
	begin
	    Max_Y := This_Y ;
	end ;
    end ;

    // Initialize line lists...
    Lines_Left_X.Capacity := Max_Y - Min_Y + 1 ;
    Lines_Y.Capacity := Lines_Left_X.Capacity ;
    Lines_Right_X.Capacity := Lines_Left_X.Capacity ;
    Lines_Bitmap_X_Right.Capacity := Lines_Left_X.Capacity ;
    Lines_Bitmap_Y_Right.Capacity := Lines_Left_X.Capacity ;
    Lines_Bitmap_X_Left.Capacity := Lines_Left_X.Capacity ;
    Lines_Bitmap_Y_Left.Capacity := Lines_Left_X.Capacity ;
    for Loop := 0 to Max_Y - Min_Y do
    begin
	Lines_Left_X.Add( pointer( 32767 ) ) ;
	Lines_Y.Add( pointer( Loop + Min_Y ) ) ;
	Lines_Right_X.Add( nil ) ;
        Lines_Bitmap_X_Left.Add( nil ) ;
        Lines_Bitmap_Y_Left.Add( nil ) ;
        Lines_Bitmap_X_Right.Add( nil ) ;
        Lines_Bitmap_Y_Right.Add( nil ) ;
    end ;

    // Now build list
    for Loop := 0 to Count - 1 do
    begin
	P := X[ Loop ] ;
	This_Y := P.Y - Min_Y ;
	if( integer( Lines_Left_X.List[ This_Y ] ) > P.X ) then
	begin
	    Lines_Left_X.List[ This_Y ] := pointer( P.X ) ;
            Lines_Bitmap_X_Left.List[ This_Y ] := Bitmap_X_List.List[ Loop ] ;
            Lines_Bitmap_Y_Left.List[ This_Y ] := Bitmap_Y_List.List[ Loop ] ;
	end ;
	if( integer( Lines_Right_X.List[ This_Y ] ) < P.X ) then
	begin
	    Lines_Right_X.List[ This_Y ] := pointer( P.X ) ;
            Lines_Bitmap_X_Right.List[ This_Y ] := Bitmap_X_List.List[ Loop ] ;
            Lines_Bitmap_Y_Right.List[ This_Y ] := Bitmap_Y_List.List[ Loop ] ;
	end ;
    end ;
end ; // TPolygon.MinMax


function TPolygon.Vertex_Count : integer ;

begin
    Result := Vertex_X_List.Count ;
end ;



end.
