{$N+}
{$E+}
unit MemStore ;
{
        Program Name : MemStore
        Package Name : Subroutine Library
        Purpose      : In-memory TStore.
        Institution  : Conroy & Conroy Co.
        Date Written : 10-June-2002
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2002 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  TMemory_Store is a contiguous heap store (THeap is a non-contiguous
        heap store).
}

interface

{$I EDEFINES.INC}

{$IFDEF DELPHI}
uses Classes,
{$ELSE}
uses Objects,
{$ENDIF}
     Compatib,
     _DebugIn, { PSDebug_Interface }
     MStores, { PSManaged_Store }
     _Stores, { TSStore }
     _UEHDefs ; { TUEC }

type TMemory_Store = object( TSStore )
                         private { Instance data... }
                             _Data : TStore_Address ;
                             _Fixed : boolean ;
                             _Heap : PSManaged_Store ;
                             _Len : TStore_Address ;
                             Serial_Number : longint ;

                         public { Constructors and destructors... }
                             constructor Init( H : PSManaged_Store ) ;

                             destructor Done ; virtual ;

                         public { API... }
                             function Read_Data( var Data ; Address, _Size : TStore_Address ;
                                 var UEC : TUEC ) : TStore_Address ; virtual ;

                             function Write_Data( var Data ; Address, _Size : TStore_Address ;
                                 var UEC : TUEC ) : TStore_Address ; virtual ;

                             function Max_Storage : TStore_Size ; virtual ;

                             function Min_Storage : TStore_Address ; virtual ;

                             function Extend( Amount : TStore_Address ) : TStore_Address ;
                                 virtual ;

                             function Debugger : PSDebug_Interface ; virtual ;

                             function Contiguous_Store : boolean ; virtual ;

                             function Get_Name : string ; virtual ;

                             procedure Set_Max_Storage( Value : TStore_Address ;
                                 var Res : TUEC ) ; virtual ;

                             function Extended_Size : TStore_Address ; virtual ;

                             procedure Fix_Size( Size : TStore_Address ) ;
                                 virtual ;
                     end ;
     PMemory_Store = ^TMemory_Store ;

implementation

uses { C&C... }
{$IFDEF DELPHI}
     SysUtils, { uppercacse }
{$ENDIF}
     _App, { Application_Manager }
     Collect, { PString_Collection }
     Files, { PFile_System }
     Standard, { Get_Serial_Number }
     Num1s ; { Num1 }

var Memory_Store_List : PCollection ;

{ TMemory_Store_Server... }

type TMemory_Store_Server = object( TStore_Server )
                                public { API... }
                                   function Get_Store_ID( Name : string ) : string ;
                                       virtual ;

                                   function Get_Store( var Name : string ) : PSStore ;
                                       virtual ;
			    end ;
     PMemory_Store_Server = ^TMemory_Store_Server ;

{ API... }

function TMemory_Store_Server.Get_Store_ID( Name : string ) : string ;

var Dummy : integer ;

begin
    Dummy := pos( ':', Name ) ;
    if( Dummy = 0 ) then
    begin
        Get_Store_ID := '' ;
        exit ;
    end ;
    Name := uppercase( copy( Name, 1, Dummy - 1 ) ) ;
    for Dummy := 0 to Memory_Store_List.Count - 1 do
    begin
        if( Name = PMemory_Store( Memory_Store_List.At( Dummy ) )^.Get_Name ) then
        begin
            Get_Store_ID := Name + ':' ;
            exit ;
        end ;
    end ;
end ;


function TMemory_Store_Server.Get_Store( var Name : string ) : PSStore ;

var Dummy : integer ;
    _Name : string ;

begin
    Get_Store := nil ;
    Dummy := pos( ':', Name ) ;
    if( Dummy = 0 ) then
    begin
        exit ;
    end ;
    _Name := uppercase( copy( Name, 1, Dummy - 1 ) ) ;
    for Dummy := 0 to Memory_Store_List.Count - 1 do
    begin
        if( _Name = PMemory_Store( Memory_Store_List.At( Dummy ) )^.Get_Name ) then
        begin
            Get_Store := PMemory_Store( Memory_Store_List.At( Dummy ) ) ;
            Name := _Name + ':' ;
            exit ;
        end ;
    end ;
end ;



{ TMemory_Store methods... }

{ Constructors and destructors... }

constructor TMemory_Store.Init( H : PSManaged_Store ) ;

begin
    inherited Init ;

    _Heap := H ;
    _Len := 0 ;
    _Data := 0 ;
    _Fixed := False ;
    Serial_Number := Get_Serial_Number ;
    Memory_Store_List.Insert( @self ) ;
end ;


destructor TMemory_Store.Done ;

begin
    Memory_Store_List.Delete( @self ) ;
    inherited Done ;
end ;


{ API... }

function TMemory_Store.Read_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : TUEC ) : TStore_Address ;

begin
    if( Address + _Size > _Len ) then { Reading past end }
    begin
        _Size := _Len - Address ;
    end ;
    Read_Data := _Heap^.Read_Data( Data, _Data + Address, _Size, UEC ) ;
end ;


function TMemory_Store.Write_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : TUEC ) : TStore_Address ;

begin
    if( Address + _Size > _Len ) then
    begin
        Application_Manager^.Fatal_Exception( 'Store overwrite' ) ;
    end ;
    Write_Data := _Heap^.Write_Data( Data, _Data + Address, _Size, UEC ) ;
end ;


function TMemory_Store.Max_Storage : TStore_Size ;

begin
    Max_Storage := _Len ;
end ;


function TMemory_Store.Min_Storage : TStore_Address ;

begin
    Min_Storage := _Heap^.Min_Storage ;
end ;


function TMemory_Store.Extend( Amount : TStore_Address ) : TStore_Address ;

var Dummy : TStore_Address ;

begin
    if( Amount = 0 ) then
    begin
        Extend := 0 ;
        exit ;
    end ;
    Dummy := _Heap^.Reallocate( _Data, _Len, _Len + Amount ) ;
    if( Dummy <> 0 ) then
    begin
        _Data := Dummy ;
        _Len := _Len + Amount ;
	Extend := Amount ;
    end else
    begin
        Extend := 0 ;
    end ;
end ;


function TMemory_Store.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ; { Not yet supported }
end ;


function TMemory_Store.Contiguous_Store : boolean ;

begin
    Contiguous_Store := True ;
end ;


function TMemory_Store.Get_Name : string ;

begin
    Get_Name := 'MEMORY' + Num1( Serial_Number ) ;
end ;


procedure TMemory_Store.Set_Max_Storage( Value : TStore_Address ;
    var Res : TUEC ) ;

var Dummy : TStore_Address ;

begin
    fillchar( Res, sizeof( Res ), 0 ) ;
    Dummy := _Heap^.Reallocate( _Data, _Len, Value ) ;
    if( Dummy <> 0 ) then
    begin
        _Data := Dummy ;
        _Len := Value ;
    end ;
end ;


function TMemory_Store.Extended_Size : TStore_Address ;

begin
    Extended_Size := _Len + _Heap^.Extended_Size ;
end ;


procedure TMemory_Store.Fix_Size( Size : TStore_Address ) ;

var Dummy : TStore_Address ;

begin
    Dummy := _Heap^.Reallocate( _Data, _Len, Size ) ;
    if( Dummy <> 0 ) then
    begin
        _Data := Dummy ;
        _Len := Size ;
        _Fixed := True ;
    end ;
end ;



var Memory_Store_Server : PMemory_Store_Server ;

begin
    Memory_Store_List := new( PCollection, Init( 0, 4 ) ) ;
    Memory_Store_Server := new( PMemory_Store_Server, Init ) ;
    Register_Store_Server( Memory_Store_Server ) ;
end.
