/*
        Program Name : Operatio
        Package Name : LB
        Purpose      : Operation logger
        Institution  : Conroy & Conroy Co.
        Date Written :
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (c) 2002 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

*/

#include <alloc.h>

#include <owl\owlpch.h>

#include "c:\lb\operatio.h"

TOperation_Logger::TOperation_Logger( int max )
{
    Root = NULL ; // Root of operation list
    Count = 0 ; // Count of logged operations
    Transaction_Number = 0 ;
    Transaction_Type = 0 ;
    Original = TRUE ;
    Set_Maximum( max ) ; // Maximum logged operations
}


TOperation_Logger::~TOperation_Logger()
{
    Zero() ;
}


void TOperation_Logger::Check_Extras()
// Delete any originals which overflow the maximum
{
    ROperation *Temp ;
    long int Transaction ;

    if( Root == NULL ) // Nothing logged
    {
	return ;
    } ;

    if( Root->Transaction == Transaction_Number )
    {
	return ; // Done if one big transaction
    } ;

    if( Count > Max_Operations )
    {
	Original = FALSE ;
	while( Count > Max_Operations )
	{
	    Count = Count - 1 ;
	    Temp = Root ;
	    Root = Root->Next ;
	    Transaction = Temp->Transaction ;
	    if( Temp->Data != NULL )
	    {
		free( Temp->Data ) ;
	    } ;
	    free( Temp ) ;
	} ;
	// Make sure all operations within a given transaction go away
	while( Root->Transaction == Transaction )
	{
	    Temp = Root ;
	    if( Temp->Data != NULL )
	    {
		free( Temp->Data ) ;
	    } ;
	    Root = Root->Next ;
	    free( Temp ) ;
	} ;
    } ;
}


void TOperation_Logger::Set_Maximum( int max )
{
    // Update maximum operations
    Max_Operations = max ;

    // Delete any extra operations
    Check_Extras() ;
}


void TOperation_Logger::Zero() // Erase all storage
{
    ROperation *Temp ;

    Original = TRUE ;
    while( Root != NULL )
    {
	Temp = Root ;
	Root = Root->Next ;
	if( Temp->Data != NULL )
	{
	    free( Temp->Data ) ;
	} ;
	free( Temp ) ;
    } ;
    Count = 0 ; // Nothing left
    Transaction_Number = 0 ;
} // TOperation_Logger::Zero


void TOperation_Logger::New_Transaction( int Type ) // Start a new transaction
{
    Transaction_Number = Transaction_Number + 1 ;
    Transaction_Type = Type ;
}


void TOperation_Logger::Unwind( int Code )
// Unwind the last transaction on the list
{
    long int Transaction ;
    ROperation *Temp ; // Current operation
    ROperation *First ; // First operation of this transaction
    ROperation *Previous ; // Previous operation
    ROperation *Last ; // Last operation of previous transaction

    // Find beginning of last transaction
    if( Root == NULL )
    {
	return ; // Nothing to unwind
    } ;
    Temp = Root ;
    Transaction = Root->Transaction ;
    Last = NULL ;
    First = Root ;
    Previous = NULL ;
    while( Temp != NULL ) // Until EOL
    {
	if( Transaction != Temp->Transaction ) // New transaction
	{
	    Transaction = Temp->Transaction ;
	    First = Temp ;
	    Last = Previous ;
	} ;
	Previous = Temp ;
	Temp = Temp->Next ;
    } ;

    // At this point, Last is the last operation of the previous transaction
    // (NULL if no previous transaction), First is the first operation of
    // the current transaction.
    if( Last != NULL )
    {
	Last->Next = NULL ; // Terminate the list
    }
    else
    {
        Root = NULL ;  // No more items in list
    } ;
    while( First != NULL )
    {
	// Find the last transaction on the current operation
	Temp = First ;
	Previous = Last ;
	while( Temp->Next != NULL )
	{
	    Previous = Temp ;
	    Temp = Temp->Next ;
	} ;
	if( Previous != NULL )
	{
	    Previous->Next = NULL ;
	} ;
	Temp->ID( Code, Temp->Data ) ;
	if( Temp->Data != NULL )
	{
	    free( Temp->Data ) ;
	} ;
	if( Temp == First ) // Processed the first operation in the transaction
	{
	    First = NULL ; // Terminate the loop
	} ;
	free( Temp ) ;
	Count = Count - 1 ; // One less operation
    } ;
} // TOperation_Logger::Unwind


int TOperation_Logger::Query_Type() // Return type of next undo record
{
    ROperation *Temp ; // Current operation

    // Find the last logged record
    if( Root == NULL )
    {
	return 0 ; // Nothing to unwind
    } ;
    Temp = Root ;
    while( Temp->Next != NULL ) // Until EOL
    {
	Temp = Temp->Next ;
    } ;
    return Temp->Type ;
}


void TOperation_Logger::New_Operation( void (*id)(int,void*), void *data )
// Store new operation on list
{
    ROperation *Temp, *Operation ;

    // Create new operation record
    Operation = (ROperation*) malloc( sizeof( ROperation ) ) ;
    Operation->Next = NULL ;
    Operation->Transaction = Transaction_Number ;
    Operation->Type = Transaction_Type ;
    Operation->ID = id ;
    Operation->Data = data ;

    // Find end of list and insert operation
    if( Root == NULL )
    {
	Root = Operation ;
    }
    else
    {
	Temp = Root ;
	while( Temp->Next != NULL )
	{
	    Temp = Temp->Next ;
	} ;
	Temp->Next = Operation ;
    } ;

    // Increment count of operations
    Count = Count + 1 ;
    Check_Extras() ;
}
