{$E+}
{$N+}
{
        Program Name : OS_DOS
        Package Name : LB
        Purpose      : DOS Operating system interface
        Institution  : Conroy & Conroy Co.
        Date Written : 6-Nov-2002
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2002 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit OS_DOS ;

interface

{$I EDEFINES.INC}

uses _Common, { TObject }
     _DevMan, { PSDevice_Manager }
     _OS, { TOperating_System }
     _Types, { int64 }
     _UEHDefs ; { TUEC }

type TOS_DOS = object( TSOperating_System )
                   public { Constructors and destructors... }
                       constructor Init ;
                       destructor Done ; virtual ;

                   private
                       Internal_Clipboard : string ;
                       _Device_Manager : PSDevice_Manager ;

                   public { API... }
	               function Name : string ; virtual ;
	               function Version : string ; virtual ;

                       function Application_Path : string ; virtual ;
                       procedure Beep( Typ : integer ) ; virtual ;
                       function Command_Line : string ; virtual ;
                       function Get_Char( Wait : boolean ;
                           var Valid : boolean ) : char ;
                           virtual ;
                       procedure Get_Date( var Year, Month, Day : word ) ;
                           virtual ;
                       function Get_Symbol( const X : string ) : string ;
                           virtual ;
                       procedure Get_Time( var Hour, Minute, Second, Hundreths : word ) ;
                           virtual ;
                       procedure Last_Error( var UEC : TUEC ) ;
                           virtual ;
	               function Error_Text( const UEC : TUEC ;
                           const Language : string ) : string ; virtual ;
                       function Pending_Input : boolean ; virtual ;
                       function PID : longint ; virtual ;
                       procedure Print( T : string ) ; virtual ;
	          	 procedure Spawn( const Path, Parameters : string ) ; virtual ;
                       function Switch_Delimiters( Index : integer ) : string ;
                           virtual ;
                       function System_Path : string ;
	          	 function Timestamp : int64 ; virtual ;
                       function Get_Clipboard( Path : string ) : string ;
                           virtual ;
                       procedure Set_Clipboard( Path, Value : string ) ;
                           virtual ;
                       function Max_File_Size : int64 ; virtual ;
                       function Device_Manager : PSDevice_Manager ;
                           virtual ;
               end ; { TOS_DOS }
     POS_DOS = ^TOS_DOS ;

implementation

uses ASCIIDef,
     Compatib, { GetCurDir }
     WinDOS, { GetDate }
{$IFNDEF WINDOWS}
     DOS, { GetDate }
{$ENDIF}
     Files, { Clear_IOResult }
     Standard, { ERT }
     TSRs ; { pPSP }

var _PID : longint ;

{$IFNDEF WINDOWS}
type TRegisters = Registers ;
{$ENDIF}


{ TOS_Device_Manager... }

type TOS_Device_Manager = object( TSDevice_Manager )
                              public { API... }
                                  function Disk_Clusters( Drive : string ) : int64 ;
                                      virtual ;
                                  { Returns number of clusters on specified disk
                                    drive.  Returns -1 if an error or unknown. }

                                  function Disk_Sector_Size( Drive : string ) : longint ;
                                      virtual ;
                                  { Returns size of disk sectors, in bytes.
                                    Returns -1 if an error or unknown. }

                                  function File_System( Device : string ) : string ;
                                      virtual ;

                                  function File_System_Flags( Device : string ) : longint ;
                                      virtual ;

                                  function Fixed_Device( Device : string ) : boolean ;
                                      virtual ;
                                  { Returns true if the passed device is fixed
                                    (non-removable) }

                                  function Max_Filename_Length( Device : string ) : longint ;
                                      virtual ;

                                  function Media_ID( Device : string ) : longint ;
                                      virtual ;
                                  { Return type of media in/on device.  -1 if
                                    error or unknown. }

                                  function Valid_Device( Device : string ) : boolean ;
                                      virtual ;
                                  { Returns true if the passed device is valid. }
                          end ;
     POS_Device_Manager = ^TOS_Device_Manager ;


{ API... }

function TOS_Device_Manager.Disk_Clusters( Drive : string ) : int64 ;

var Reg : TRegisters ;

begin
    if(
        ( length( Drive ) = 0 )
        or
        ( ( length( Drive ) > 1 ) and ( Drive[ 2 ] <> ':' ) )
      ) then
    begin
        Disk_Clusters := -1 ;
        exit ;
    end ;
    Reg.Ax := $1C00 ;
    Reg.Dx := Ord( Drive[ 1 ] ) - 64 ;
    Msdos( Reg ) ; { Get drive information }
    if( Reg.AL = 255 ) then
    begin
        Disk_Clusters := -1 ;
        exit ;
    end ;
    Disk_Clusters := Reg.DX ;
end ;


function TOS_Device_Manager.Disk_Sector_Size( Drive : string ) : longint ;

var Reg : Tregisters ;

begin
    if(
        ( length( Drive ) = 0 )
        or
        ( ( length( Drive ) > 1 ) and ( Drive[ 2 ] <> ':' ) )
      ) then
    begin
        Disk_Sector_Size := -1 ;
        exit ;
    end ;
    Reg.Ax := $1C00 ;
    Reg.Dx := Ord( Drive[ 1 ] ) - 64 ;
    Msdos( Reg ) ; { Get drive information }
    if( Reg.AL = 255 ) then
    begin
        Disk_Sector_Size := -1 ;
        exit ;
    end ;
    Disk_Sector_Size := Reg.CX ;
end ;


function TOS_Device_Manager.File_System( Device : string ) : string ;

var Reg : Tregisters ;
    S : string[ 255 ] ;

begin
    if(
        ( length( Device ) = 0 )
        or
        ( ( length( Device ) > 1 ) and ( Device[ 2 ] <> ':' ) )
      ) then
    begin
        File_System := '' ;
        exit ;
    end ;
    Reg.Ax := $71A0 ;
    Reg.ES := seg( S ) ;
    Reg.DI := ofs( S ) + 1 ;
    Reg.Cx := 254 ;
    Reg.DS := seg( Device ) ;
    Reg.Dx := ofs( Device ) + 1 ;
    Msdos( Reg ) ; { Get fs information }
    if( ( Flags and 64 ) <> 0 ) then
    begin
        File_System := '' ;
        exit ;
    end ;
    S[ 0 ] := strlen( PChar( @S[ 1 ] ) ) ;
    File_System := S ;
end ;


function TOS_Device_Manager.File_System_Flags( Device : string ) : longint ;

var Reg : Tregisters ;
    S : string[ 255 ] ;

begin
    File_System_Flags := 0 ;
    if(
        ( length( Device ) = 0 )
        or
        ( ( length( Device ) > 1 ) and ( Device[ 2 ] <> ':' ) )
      ) then
    begin
        exit ;
    end ;
    Reg.Ax := $71A0 ;
    Reg.ES := seg( S ) ;
    Reg.DI := ofs( S ) + 1 ;
    Reg.Cx := 254 ;
    Reg.DS := seg( Device ) ;
    Reg.Dx := ofs( Device ) + 1 ;
    Msdos( Reg ) ; { Get fs information }
    if( ( Flags and 64 ) <> 0 ) then
    begin
        exit ;
    end ;
    File_System_Flags := Reg.Bx ;
end ;


function TOS_Device_Manager.Fixed_Device( Device : string ) : boolean ;

var Reg : Tregisters ;

begin
    if(
        ( length( Device ) = 0 )
        or
        ( ( length( Device ) > 1 ) and ( Device[ 2 ] <> ':' ) )
      ) then
    begin
        Fixed_Device := False ;
        exit ;
    end ;
    Reg.Ax := $1C00 ;
    Reg.Dx := Ord( Device[ 1 ] ) - 64 ;
    Msdos( Reg ) ; { Get drive information }
    if( Reg.AL = 255 ) then
    begin
        Fixed_Device := False ;
        exit ;
    end ;
    Fixed_Device := ( Mem[ Reg.Ds : Reg.Bx ] = $F8 ) ;
end ;


function TOS_Device_Manager.Max_Filename_Length( Device : string ) : longint ;

var Reg : Tregisters ;
    S : string[ 255 ] ;

begin
    if(
        ( length( Device ) = 0 )
        or
        ( ( length( Device ) > 1 ) and ( Device[ 2 ] <> ':' ) )
      ) then
    begin
        Max_Filename_Length := -1 ;
        exit ;
    end ;
    Reg.Ax := $71A0 ;
    Reg.ES := seg( S ) ;
    Reg.DI := ofs( S ) + 1 ;
    Reg.Cx := 254 ;
    Reg.DS := seg( Device ) ;
    Reg.Dx := ofs( Device ) + 1 ;
    Msdos( Reg ) ; { Get fs information }
    if( ( Flags and 64 ) <> 0 ) then
    begin
        Max_Filename_Length := -1 ;
        exit ;
    end ;
    Max_Filename_Length := Reg.Cx ;
end ;


function TOS_Device_Manager.Media_ID( Device : string ) : longint ;

var Reg : TRegisters ;

begin
    if(
        ( length( Device ) = 0 )
        or
        ( ( length( Device ) > 1 ) and ( Device[ 2 ] <> ':' ) )
      ) then
    begin
        Media_ID := -1 ;
        exit ;
    end ;
    Reg.Ax := $1C00 ;
    Reg.Dx := Ord( Device[ 1 ] ) - 64 ;
    Msdos( Reg ) ; { Get drive information }
    if( Reg.AL <> 255 ) then
    begin
        Media_ID := -1 ;
        exit ;
    end ;
    Media_ID := Mem[ Reg.Ds : Reg.Bx ] ;
end ;


function TOS_Device_Manager.Valid_Device( Device : string ) : boolean ;

var Reg : TRegisters ;

begin
    if(
        ( length( Device ) = 0 )
        or
        ( ( length( Device ) > 1 ) and ( Device[ 2 ] <> ':' ) )
      ) then
    begin
        Valid_Device := False ;
        exit ;
    end ;
    Reg.Ax := $1C00 ;
    Reg.Dx := Ord( Device[ 1 ] ) - 64 ;
    Msdos( Reg ) ; { Get drive information }
    Valid_Device := ( Reg.AL <> 255 ) ;
end ;



{ TOS_DOS methods... }

{ Constructors and destructors... }

constructor TOS_DOS.Init ;

begin
    inherited Init ;

{$IFDEF Delphi}
    Initialize( Internal_Clipboard ) ;
{$ENDIF}
    _Device_Manager := nil ;
    Internal_Clipboard := '' ;
end ;


destructor TSOS_DOS.Done ;

begin
    _Device_Manager.Free ;
    _Device_Manager := nil ;

    inherited Done ;
end ;


{ API... }

function TOS_DOS.Name : string ; { O/S name }

begin
    Name := 'MSDOS' ;
end ;


function TOS_DOS.Version : string ; { O/S version }

var S1, S2 : string ;

begin
    str( lo( DOSVersion ), S1 ) ;
    str( hi( DOSVersion ), S2 ) ;
    Version := S1 + '.' + S2 ;
end ;


function TOS_DOS.Application_Path : string ;

var Dummy : integer ;
    S : string[ 255 ] ;

begin
    S := Paramstr( 0 ) ;
    Dummy := length( S ) ;
    while( S[ Dummy ] <> '\' ) do
    begin
        dec( Dummy ) ;
    end ;
    S := copy( S, 1, Dummy ) ;
    if( length( S ) = 0 ) then
    begin
        GetCurDir( @S[ 1 ], 0 ) ;
        S[ 0 ] := chr( strlen( PChar( @S[ 1 ] ) ) ) ;
    end ;
    Application_Path := S ;
end ;


procedure TOS_DOS.Beep( Typ : integer ) ;

var R : TRegisters ;

begin
    R.Ax := $0200 ; { Do it this way so it beeps even if output is redirected }
    R.Dx := 7 ;
    Msdos( R ) ;
end ;


function TOS_DOS.Command_Line : string ;

var P : Pointer ;

begin
    P := PTR( PrefixSeg, 0 ) ;
    Command_Line := pPSP( P )^.Command_Tail ;
end ;


function TOS_DOS.Get_Char( Wait : boolean ; var Valid : boolean ) : char ;

var C : Char ;
    Dummy : integer ;
    Reg : Tregisters ;

label Loop ;

begin
Loop:
    with Reg do
    begin
        Ax := $0600 ;
        Dx := $00FF ;
        Msdos( Reg ) ; { Get character }
        Get_Char := Chr( Lo( Ax ) ) ; { ...and return it }
        Valid := ( ( Flags and 64 ) = 0 ) ;
        if( not Valid ) then { No input }
        begin
            if( Wait ) then { Wait for input }
            begin
                goto Loop ;
            end else
            begin
                Get_Char := NUL ; { Return nul }
            end ;
        end ;
    end ;
end ;


procedure TOS_DOS.Get_Date( var Year, Month, Day : word ) ;

var DOW : word ;

begin
    GetDate( Year, Month, Day, DOW ) ;
end ;


function TOS_DOS.Get_Symbol( const X : string ) : string ;

{$IFDEF WINDOWS}
Type Env_Array = array[ 1..255 ] of Byte ;
     Env_Pointer = ^Env_Array ;

var A : integer ;
    Ep : Env_Pointer ;
    Env : string ;
    Z : array[ 0..255 ] of char ;
{$ENDIF}

begin
{$IFDEF WINDOWS}
    StrPCopy( Z, X ) ;
    Ep := Env_Pointer( Getenvvar( Z ) ) ;
    Env := '' ;
    if( Ep <> nil ) then {If variable is defined}
    begin
        A := 1 ;
        while( Ep^[ A ] <> 0 ) do {Until end of variable}
        begin
            Env := Env + chr( Ep^[ A ] ) ;
            inc( A ) ;
        end ;
    end ;
    Get_Symbol := Env ;
{$ELSE}
    Get_Symbol := GetEnv( X ) ;
{$ENDIF}
end ;


procedure TOS_DOS.Get_Time( var Hour, Minute, Second, Hundreths : word ) ;

begin
    GetTime( Hour, Minute, Second, Hundreths ) ;
end ;


procedure TOS_DOS.Last_Error( var UEC : TUEC ) ;

var Reg : TRegisters ;

begin
    UEC.Facility := 1 ; { MSDOS }
    if( Lo( Dosversion ) < 3 ) then
    begin
        UEC.Code := 31 ; { No extended error function until version 3 }
    end else
    begin
        Reg.Ax := $5900 ;
        Reg.Bx := 0 ;
        Msdos( Reg ) ;
        UEC.Code := Reg.Ax ;
    end ;
end ;


function TOS_DOS.Error_Text( const UEC : TUEC ; const Language : string ) : string ;

begin
    if( UEC.Facility = 1 ) then
    begin
        Error_Text := ERT( UEC.Code ) ;
    end else
    begin
        Error_Text := '' ;
    end ;
end ;


function TOS_DOS.PID : longint ;

begin
    PID := _PID ;
end ;


function TOS_DOS.Pending_Input : boolean ;

var R : TRegisters ;

begin
    R.Ax := $0B00 ; { Check input status }
    Msdos( R ) ;
    if( Lo( R.Ax ) <> 0 ) then
    begin
        Pending_Input := True ;
        exit ; { Exit if input waiting }
    end ;
    Pending_Input := False ;
end ;


procedure TOS_DOS.Print( T : string ) ;

var R : TRegisters ;
    Loop : integer ;

begin
    R.Ax := $0500 ; { Printer output }
    for Loop := 1 to length( T ) do
    begin
        R.Dx := ord( T[ Loop ] ) ;
        MSDOS( R ) ;
    end ;
end ;


procedure TOS_DOS.Spawn( const Path, Parameters : string ) ;

begin
    Swapvectors ;
    Exec( Path, Parameters ) ;
    Swapvectors ;
end ;


function TOS_DOS.Switch_Delimiters( Index : integer ) : string ;

var Reg : TRegisters ;

begin
    Reg.Ax := $3700 ;
    Msdos( Reg ) ; { Get switchar }
    Switch_Delimiters := Chr( Reg.Dx ) ;
end ;


function TOS_DOS.Timestamp : int64 ;

begin
    Timestamp := 0 ;
end ;


function TOS_DOS.System_Path : string ;

var D : byte ;
    Reg : TRegisters ;

begin
    System_Path := '' ;
    if( CheckDOS and ( lo( DOSVersion ) >= 4 ) ) then
    begin
        Reg.Ax := $5200 ;
        Reg.Bx := 0 ;
        Reg.Es := 0 ;
        Msdos( Reg ) ; { Get list of lists }
        D := Mem[ Reg.Es : Reg.Bx + $43 ]
        System_Path := chr( D + 64 ) + ':\' ;
    end ;
end ;


function TOS_DOS.Get_Clipboard( Path : string ) : string ;

var F : file ;
    S : string ;

begin
    if( Path = '' ) then
    begin
        Get_Clipboard := Internal_Clipboard ;
        exit ;
    end ;
    Path := Edit( Path, -1 ) ;
    S := Get_Symbol( 'Clipboard' ) ;
    if( length( S ) = 0 ) then
    begin
        S := Get_Symbol( 'BootDisk' ) ;
        if( length( S ) = 0 ) then
        begin
            S := System_Path ;
        end ;
    end ;
    if( length( S ) = 0 ) then
    begin
        S := 'C:\Clipboard\' ;
    end else
    if( copy( S, length( S ), 1 ) <> '\'  ) then
    begin
        S := S + '\' ;
    end ;
    S := S + Path ;
    assign( F, S ) ;
    {$I-}
    reset( F, 1 ) ;
    {$I+}
    if( IOResult = 0 ) then
    begin
        setlength( S, filesize( F ) ) ;
        {$I-}
        blockread( F, S[ 1 ], length( S ) ) ;
        {$I+}
        if( IOResult = 0 ) then
        begin
            Get_Clipboard := S ;
        end ;
        {$I-}
        close( F ) ;
        {$I+}
        Clear_IOResult ;
        exit ;
    end ;

    Get_Clipboard := '' ;
end ;


procedure TOS_DOS.Set_Clipboard( Path, Value : string ) ;

var F : text ;
    S : string ;

begin
    if( Path = '' ) then
    begin
        Internal_Clipboard := Value ;
        exit ;
    end ;
    S := Get_Symbol( 'Clipboard' ) ;
    if( length( S ) = 0 ) then
    begin
        S := Get_Symbol( 'BootDisk' ) ;
        if( length( S ) = 0 ) then
        begin
            S := System_Path ;
        end ;
    end ;
    if( length( S ) = 0 ) then
    begin
        S := 'C:\Clipboard\' ;
    end else
    if( copy( S, length( S ), 1 ) <> '\'  ) then
    begin
        S := S + '\' ;
    end ;
    S := S + Path ;
    assign( F, S ) ;
    {$I-}
    rewrite( F ) ;
    {$I+}
    if( IOResult = 0 ) then
    begin
        setlength( S, filesize( F ) ) ;
        {$I-}
        write( F, S ) ;
        {$I+}
        Clear_IOResult ;
        {$I-}
        close( F ) ;
        {$I+}
        Clear_IOResult ;
    end ;
end ;


function TOS_DOS.Max_File_Size : int64 ;

begin
    Max_File_Size := $3FFFFFFF ;
end ;


function TOS_DOS.Device_Manager : PSDevice_Manager ;

begin

end ;


procedure Unit_Setup ;

var Reg : Tregisters ;

begin
    { We must get the PID at startup because, for TSRS, the get-PSP function
      returns the foreground PID, not our PID.  At startup, get-PSP returns our
      PID, so we save it here so it is available later. }
    Reg.AX := $6200 ;
    MSDOS( Reg ) ;
    _PID := Reg.BX ;
end ;


begin
    Unit_Setup ;
end.
