{
        Program Name : Outlines
        Package Name : LB
        Purpose      : Outline object
        Institution  : Conroy & Conroy Co.
        Date Written : 11-November-2002
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2002 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        Certified by EAC on 25-November-2002.  Level 1.
        Added Serialize and Deserialize.

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit contains TSOutline, a class for maintaining outlines.
}

unit Outlines ;

interface

uses Collect, { TCollection }
     Common, { TBase_Object }
     UStrings ; { PString }

type PSOutline = ^TSOutline ;
     PSOutline_Node = ^TSOutline_Node ;

     TSOutline = object( TBase_Object )
                    public { Constructors and destructors... }
                        constructor Init ;
                        destructor Done ; virtual ;

                    private { Instance data }
                        Children : TCollection ;

                    private { Internal utility routines... }
                        procedure Delete_Node( Node : PSOutline_Node ) ;

                    public { API... }
                        function Add_Child( Node : PSOutline_Node ;
                            const Text : string ) : PSOutline_Node ;
                            virtual ;
                        { Add a child node to specified node.  If Node is nil,
                          add a top-level child.  The child node is always the
                          last at its level. }

                        function Add_Node( Node, Child : PSOutline_Node ) : PSOutline_Node ;
                            virtual ;
                        { Add a child node to specified node.  If Node is nil,
                          add a top-level child.  The child node is always the
                          last at its level. }

                        procedure Add_Outline( Source : PSOutline ) ;
                        { Append the passed outline to the end of this outline. }

                        function Add_Sibling( Node : PSOutline_Node ;
                            const Text : string ) : PSOutline_Node ;
                            virtual ;
                        { Add a node as a sibling of the specified node.  If
                          Node is nil, add a top-level child.  The child node is
                          always the last at its level. }

                        procedure Clear ; virtual ;
                        { Clears all data in the outline. }

                        function Create_Node : PSOutline_Node ; virtual ;
                        { Creates an empty node.  This allows descendants to
                          create outlines with TSOutline_Node descendants. }

                        function Count : integer ; virtual ;
                        { Count of top-level children }

                        procedure Delete( Node : PSOutline_Node ; Index : integer ) ;
                            virtual ;
                        { Removes the Indexth child of specified node.  If node
                          is nil, removes the Indexth top-level child.  If the
                          index is out of range, it does nothing. }

                        procedure String_Deserialize( X : PString ) ; virtual ;
                        { Fill outline from stream }

                        function Find( Text : string ; const Delimiter : string ;
                            Wildcards : boolean ) : PSOutline_Node ; virtual ;
                        { Using the specified path and delimiter, find a
                          matching node.  If Wildcards is true, the Text may
                          contain "*" and "?" for wildcard matching, otherwise
                          all characters are literal.  If none found, returns
                          nil. }

                        function Get_Child( Node : PSOutline_Node ; Index : integer ) : PSOutline_Node ;
                            virtual ;
                        { Returns the Indexth child of specified node.  If node
                          is nil, returns the Indexth top-level child.  If the
                          index is out of range, it returns nil. }

                        function Get_Next( Node : PSOutline_Node ) : PSOutline_Node ;
                            virtual ;
                        { Returns the next sibling of specified node.  If node
                          is nil, returns the first top-level child.  If there
                          are no nodes, or Node is the last node at its level,
                          the function returns nil. }

                        function Has_Children : boolean ;
                            virtual ;
                        { Returns whether or not there are child components under
                          this level.  If this is a terminal (leaf)
                          node, the method returns false. }

                        function Insert_Child( Node : PSOutline_Node ;
                            Index : integer ;
                            const Text : string ) : PSOutline_Node ; virtual ;
                        { Inserts a new node as a child of the specified node,
                          at position Index in the list (0 is the first).  If
                          node is nil, the child is created as a top-level
                          object. }

                        function Get_Previous( Node : PSOutline_Node ) : PSOutline_Node ;
                            virtual ;
                        { Returns the previous sibling of specified node.  If
                          node is nil, returns the last top-level child.  If
                          there are no nodes, or Node is the first node at its,
                          level the function returns nil. }

                        function Node_With( Text : string ;
                            Wildcards, Search_Children : boolean ) : PSOutline_Node ;
                            virtual ;
                        { Return child node with specified text.  If Wildcards
                          is true, match on * and ?.  Returns nil if no match. }

                        function String_Serialize : PString ; virtual ;
                        { Serialize outline to stream }
                end ; { TSOutline }

     TSOutline_Node = object( TBase_Object )
                         public { Constructors and destructors... }
                             constructor Init ;
                             destructor Done ; virtual ;

                         private { Instance data... }
                             Children : PSOutline ;
                             Parent : PSOutline ;

                         public { API... }
                             Data : pointer ;
                             Text : string ;

                             function Add_Child( const _Text : string ) : PSOutline_Node ;
                                 virtual ;
                             { Add a top-level child.  The child node is
                               always the last. }

                             function Add_Node( Node : PSOutline_Node ) : PSOutline_Node ;
                                 virtual ;
                             { Add node as a top-level child.  The node is
                               always the last. }

                             procedure Clear ; virtual ;
                             { Deletes all children. }

                             function Count : integer ; virtual ;
                             { Count of top-level children }

                             function Create_Outline : PSOutline ; virtual ;
                             { Create an outline }

                             procedure Delete( Index : integer ) ;
                                 virtual ;
                             { Removes the Indexth child. }

                             procedure String_Deserialize( X : PString ) ; virtual ;
                             { Fill outline from stream }

                             function Find( const _Text, Delimiter : string ;
    				 Wildcards : boolean ) : PSOutline_Node ;
                                 virtual ;
                             { Finds node with path in Text, delimited by
                               Delimiter.  If Wildcards is true, match on ? and
                               *. }

                             function Get_Child( Index : integer ) : PSOutline_Node ;
                                 virtual ;
                             { Returns the Indexth top-level child.  If the
                               index is out of range, it returns nil. }

                              function Has_Children : boolean ;
                                  virtual ;
                              { Returns whether or not there are child components under
                                this level.  If this is a terminal (leaf)
                                node, the method returns false. }

                             function Insert_Child( Index : integer ;
                                 const _Text : string ) : PSOutline_Node ; virtual ;
                             { Inserts a new node as a top-level child at
                               position Index in the list (0 is the first). }

                             function Node_With( const _Text : string ;
                                 Wildcards, Search_Children : boolean ) : PSOutline_Node ;
                                 virtual ;
                             { Return child node with specified text.  If
                               Wildcards is true, match on * and ?.  Returns nil
                               if no match. }

                             function String_Serialize : PString ; virtual ;
                             { Serialize outline to stream }
                     end ; { TSOutline_Node }

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses HTML, { THTML_Parser }
     Parse, { Parse_Parameter }
     CommonUt, { Edit }
     XLate ; { WC_Match }

{$I EDEFINES.INC}

{ TSOutline_Node methods... }

{ Constructors and destructors... }

constructor TSOutline_Node.Init ;

begin
    inherited Init ;

    Children := Create_Outline ;
    Parent := nil ;
    Data := nil ;
{$IFDEF DELPHI}
    Initialize( Text ) ;
{$ENDIF}
end ;


destructor TSOutline_Node.Done ;

begin
    dispose( Children, Done ) ;
    if( Parent <> nil ) then
    begin
        Parent.Delete_Node( @self ) ;
    end ;

{$IFDEF DELPHI}
    Finalize( Text ) ;
{$ENDIF}

    inherited Done ;
end ;


{ API... }

function TSOutline_Node.Add_Child( const _Text : string ) : PSOutline_Node ;

begin
    Add_Child := Children.Add_Child( nil, _Text ) ;
end ;


function TSOutline_Node.Add_Node( Node : PSOutline_Node ) : PSOutline_Node ;

begin
    Add_Node := Children.Add_Node( nil, Node ) ;
end ;


procedure TSOutline_Node.Clear ;

begin
    Children.Clear ;
end ;


function TSOutline_Node.Count : integer ;

begin
    Count := Children.Count ;
end ;


function TSOutline_Node.Create_Outline : PSOutline ;

begin
    Create_Outline := new( PSOutline, Init ) ;
end ;


procedure TSOutline_Node.Delete( Index : integer ) ;

begin
    { This method has not been validated }
    Children.Delete( nil, Index ) ;
end ;


procedure TSOutline_Node.String_DeSerialize( X : PString ) ;

var Node : PSOutline_Node ;
    Parser : TSHTML_Parser ;
    S : string ;
    St : PString ;

begin
    Parser.Init ;
    Parser.Source.Assign( X ) ;
    St := Parser.Token ;
    S := St.Text ;
    dispose( St, Done ) ;
    while( length( S ) > 0 ) do
    begin
        if( copy( S, 1, 1 ) = '<' ) then
        begin
            S := Edit( S, 32 ) ;
            if( S = '<NODE>' ) then
            begin
                St := Parser.Get_Section( 'node' ) ;
                Node := Add_Child( '' ) ;
                Node^.String_Deserialize( St ) ;
                dispose( St, Done ) ;
            end else
            if( copy( S, 1, 2 ) <> '<#' ) then
            begin
                St := Parser.Get_Section( S ) ;
                dispose( St, Done ) ;
            end ;
        end else
        begin
            if( S <> ' ' ) then
            begin
                Text := Edit( S, 8 or 128 ) ;
            end ;
        end ;
        St := Parser.Token ;
        S := St.Text ;
        dispose( St, Done ) ;
    end ;
    Parser.Done ;
end ;


function TSOutline_Node.Find( const _Text, Delimiter : string ;
    Wildcards : boolean ) : PSOutline_Node ;

begin
    Find := Children.Find( _Text, Delimiter, Wildcards ) ;
end ;


function TSOutline_Node.Get_Child( Index : integer ) : PSOutline_Node ;

begin
    Get_Child := Children.Get_Child( nil, Index ) ;
end ;


function TSOutline_Node.Has_Children : boolean ;

begin
    Result := Count <> 0 ;
end ;


function TSOutline_Node.Insert_Child( Index : integer ;
    const _Text : string ) : PSOutline_Node ;

begin
    Insert_Child := Children.Insert_Child( nil, Index, _Text ) ;
end ;


function TSOutline_Node.Node_With( const _Text : string ;
    Wildcards, Search_Children : boolean ) : PSOutline_Node ;

begin
    Node_With := Children.Node_With( _Text, Wildcards, Search_Children ) ;
end ;


function TSOutline_Node.String_Serialize : PString ;

var Loop : integer ;
    Node : PSOutline_Node ;
    Res, S : PString ;

begin
    Res := new( PASCII_String, Init ) ;
    for Loop := 0 to Count - 1 do
    begin
        Node := Get_Child( Loop ) ;
        Res.Append_String( '<node>' + Text ) ;
        S := Node.String_Serialize ;
        Res.Append( S ) ;
        S.Free ;
        Res.Append_String( '<\node>' ) ;
    end ;
    String_Serialize := Res ;
end ;


{ TSOutline methods... }

{ Constructors and destructors... }

constructor TSOutline.Init ;

begin
    inherited Init ;

    Children.Init( 0, 4 ) ;
end ;


destructor TSOutline.Done ;

begin
    Clear ;
    Children.Done ;

    inherited Done ;
end ;


{ Internal utility routines... }

procedure TSOutline.Delete_Node( Node : PSOutline_Node ) ;

begin
    Children.Delete( Node ) ;
end ;


{ API... }

function TSOutline.Add_Child( Node : PSOutline_Node ;
    const Text : string ) : PSOutline_Node ;

var N : PSOutline_Node ;

begin
    if( Node = nil ) then
    begin
        N := Create_Node ;
        N^.Text := Text ;
        N^.Parent := @self ;
        Children.Insert( N ) ;
        Add_Child := N ;
    end else
    begin
        Add_Child := Node^.Add_Child( Text ) ;
    end ;
end ;


function TSOutline.Add_Node( Node, Child : PSOutline_Node ) : PSOutline_Node ;

begin 
    if( Node = nil ) then
    begin
        Children.Insert( Child ) ;
        if( Child^.Parent <> nil ) then
        begin
            Child^.Parent^.Delete_Node( Child ) ;
        end ;
        Child^.Parent := @self ;
        Add_Node := Child ;
    end else
    begin
        Add_Node := Node^.Add_Node( Child ) ;
    end ;
end ;


procedure TSOutline.Add_Outline( Source : PSOutline ) ;

var Loop : integer ;

begin
    for Loop := 0 to Source^.Children.Count - 1 do
    begin
        PSOutline_Node( Source^.Children.At( Loop ) )^.Parent := @self ;
        Children.Insert( Source^.Children.At( Loop ) ) ;
    end ;
    Source^.Children.DeleteAll ;
end ;


function TSOutline.Add_Sibling( Node : PSOutline_Node ;
    const Text : string ) : PSOutline_Node ;

var N : PSOutline_Node ;

begin
    if( ( Node = nil ) or ( Node^.Parent = nil ) or ( Node^.Parent = @self ) ) then
    begin
        N := Create_Node ;
        N^.Text := Text ;
        N^.Parent := @self ;
        Children.Insert( N ) ;
        Add_Sibling := N ;
    end else
    begin
        Add_Sibling := Node^.Parent^.Add_Child( nil, Text ) ;
    end ;
end ;


procedure TSOutline.Clear ;

var Loop : integer ;
    Node : PSOutline_Node ;

begin
    for Loop := Children.Count - 1 downto 0 do
    begin
        Node := PSOutline_Node( Children.At( Loop ) ) ;
        dispose( Node, Done ) ;
    end ;
    Children.DeleteAll ;
end ;


function TSOutline.Count : integer ;

begin
    Count := Children.Count ;
end ;


function TSOutline.Create_Node : PSOutline_Node ;

var N : PSOutline_Node ;

begin
    N := new( PSOutline_Node, Init ) ;
    Create_Node := N ;
end ;


procedure TSOutline.Delete( Node : PSOutline_Node ; Index : integer ) ;

begin
    { This method has not been validated }
    if( Node = nil ) then
    begin
        if( ( Index >= 0 ) and ( Index < Children.Count ) ) then
        begin
            Children.AtDelete( Index ) ;
        end ;
    end else
    begin
        Node^.Delete( Index ) ;
    end ;
end ;


procedure TSOutline.String_Deserialize( X : PString ) ;

var Node : PSOutline_Node ;
    Parser : TSHTML_Parser ;
    S, Work : PString ;

begin
    Parser.Init ;
    Parser.Source.Assign( X ) ;
    S := Parser.Token ;
    while( S.Length > 0 ) do
    begin
        Work := S.Edit( 32 ) ; { Uppercase }
        if( Work.At( 1 ) = ord( '<' ) ) then
        begin
            if( Work.Text = '<NODE>' ) then
            begin
                dispose( Work, Done ) ;
                Work := Parser.Get_Section( 'node' ) ;
                Node := Add_Child( nil, '' ) ;
                Node^.String_Deserialize( Work ) ;
            end else
            if( Work.At( 2 ) <> ord( '#' ) ) then
            begin
                dispose( Work, Done ) ;
                Work := Parser.Get_Section( S.Text ) ;
            end ;
        end ;
        dispose( Work, Done ) ;
        dispose( S, Done ) ;
        S := Parser.Token ;
    end ;
    dispose( S, Done ) ;
    Parser.Done ;
end ;


function TSOutline.Find( Text : string ; const Delimiter : string ;
    Wildcards : boolean ) : PSOutline_Node ;

var This_Text : string ;
    Node : PSOutline_Node ;

begin
    This_Text := Parse_Parameter( Delimiter, Text ) ;
    Node := Node_With( This_Text, Wildcards, False ) ;
    if( ( Node <> nil ) and ( length( Text ) > 0 ) ) then
    begin
	Node := Node.Find( Text, Delimiter, Wildcards ) ;
    end ;
    Find := Node ;
end ;


function TSOutline.Get_Child( Node : PSOutline_Node ;
    Index : integer ) : PSOutline_Node ;

begin
    if( Node = nil ) then
    begin
        if( ( Index >= Children.Count ) or ( Index < 0 ) ) then
        begin
            Get_Child := nil ;
        end else
        begin
            Get_Child := PSOutline_Node( Children.At( Index ) ) ;
        end ;
    end else
    begin
        Get_Child := Node^.Get_Child( Index ) ;
    end ;
end ;


function TSOutline.Get_Next( Node : PSOutline_Node ) : PSOutline_Node ;

var Index : integer ;

begin
    if( Node = nil ) then
    begin
        if( Children.Count = 0 ) then
        begin
            Get_Next := nil ;
        end else
        begin
            Get_Next := PSOutline_Node( Children.At( 0 ) ) ;
        end ;
        exit ;
    end ;
    if( ( Node.Parent = nil ) or ( Node.Parent = @self ) ) then
    begin
        Index := 0 ;
        while( ( Index < Children.Count ) and ( Children.At( Index ) <> Node ) ) do
        begin
            inc( Index ) ;
        end ;
        { Index is now the index of the passed Node. }

        inc( Index ) ; { Index of next node. }
        if( Index >= Children.Count ) then
        begin
            Get_Next := nil ;
        end else
        begin
            Get_Next := PSOutline_Node( Children.At( Index ) ) ;
        end ;
    end else
    begin
        Get_Next := Node^.Parent^.Get_Next( Node ) ;
    end ;
end ;


function TSOutline.Has_Children : boolean ;

begin
    Result := Count <> 0 ;
end ;


function TSOutline.Insert_Child( Node : PSOutline_Node ; Index : integer ;
    const Text : string ) : PSOutline_Node ;

var N : PSOutline_Node ;

begin
    if( Node = nil ) then
    begin
        if( Index >= Children.Count ) then
        begin
            Index := Children.Count ;
        end ;
        N := Create_Node ;
        N^.Text := Text ;
        N^.Parent := @self ;
        Children.AtInsert( Index, N ) ;
        Insert_Child := N ;
    end else
    begin
        Insert_Child := Node^.Insert_Child( Index, Text ) ;
    end ;
end ;


function TSOutline.Get_Previous( Node : PSOutline_Node ) : PSOutline_Node ;

var Index : integer ;

begin
    if( Node = nil ) then
    begin
        if( Children.Count = 0 ) then
        begin
            Get_Previous := nil ;
        end else
        begin
            Get_Previous := PSOutline_Node( Children.At( Children.Count - 1 ) ) ;
        end ;
        exit ;
    end ;
    if( ( Node^.Parent = nil ) or ( Node^.Parent = @self ) ) then
    begin
        Index := Children.Count ;
        while( ( Index > 0 ) and ( Children.At( Index ) <> Node ) ) do
        begin
            dec( Index ) ;
        end ;
        { Index is now the index of the passed Node. }

        dec( Index ) ; { Index of previous node. }
        if( Index < 0 ) then
        begin
            Get_Previous := nil ; { Node not found - shouldn't get here }{EAC: not tested}
        end else
        begin
            Get_Previous := PSOutline_Node( Children.At( Index ) ) ;
        end ;
    end else
    begin
        Get_Previous := Node^.Parent^.Get_Previous( Node ) ;
    end ;
end ; { TSOutline.Get_Previous }



function TSOutline.Node_With( Text : string ;
    Wildcards, Search_Children : boolean ) : PSOutline_Node ;

var Loop : integer ;
    Node : PSOutline_Node ;

begin
    { Setup... }
    Node_With := nil ; { Assume failure }
    Text := Edit( Text, 32 ) ;

    for Loop := 0 to Children.Count - 1 do
    begin
        if( Wildcards ) then
        begin
            if( WC_Match( Text, Edit( PSOutline_Node( Children.At( Loop ) ).Text, 32 ) ) ) then
            begin
                Node_With := PSOutline_Node( Children.At( Loop ) ) ;
                exit ;
            end ;
        end else
        begin
            if( Edit( PSOutline_Node( Children.At( Loop ) ).Text, 32 ) = Text ) then
            begin
                Node_With := PSOutline_Node( Children.At( Loop ) ) ;
                exit ;
            end ;
        end ;
    end ;

    if( Search_Children ) then
    begin
        for Loop := 0 to Children.Count - 1 do
        begin
            Node := PSOutline_Node( Children.At( Loop ) ).Node_With( Text, Wildcards, True ) ;
            if( Node <> nil ) then
            begin
                Node_With := Node ;
                exit ;
            end ;
        end ;
    end ;
end ; { TSOutline.Node_With }


function TSOutline.String_Serialize : PString ;

var Loop : integer ;
    Res : PASCII_String ;
    S : PString ;
    Node : PSOutline_Node ;

begin
    Res := new( PASCII_String, Init ) ;
    for Loop := 0 to Count - 1 do
    begin
        Node := Get_Child( nil, Loop ) ;
        S := Node.String_Serialize ;
        Res.Append( S ) ;
        dispose( S, Done ) ;
    end ;
    String_Serialize := Res ;
end ;


{$IFDEF Test}
procedure Test_Unit ;

    procedure Error( S : string ) ;

    begin
        writeln( S ) ;
        halt ;
    end ;

var Outline : TSOutline ;
    Loop : integer ;
    Nodes : array[ 0..9 ] of PSOutline_Node ;
    Node : PSOutline_Node ;

begin
    { Setup... }
    Outline.Init ;

    { Add nodes... }
    Nodes[ 0 ] := Outline.Add_Child( nil, 'Node 0' ) ;
    Nodes[ 1 ] := Outline.Add_Child( nil, 'Node 1' ) ;
    Nodes[ 2 ] := Outline.Add_Child( nil, 'Node 2' ) ;
    if( Outline.Count <> 3 ) then
    begin
        Error( 'Node creation failure' ) ;
    end ;
    if( Outline.Get_Next( nil ) <> Nodes[ 0 ] ) then
    begin
        Error( 'Get_Next( nil ) failure' ) ;
    end ;
    if( Outline.Get_Previous( nil ) <> Nodes[ 2 ] ) then
    begin
        Error( 'Get_Previous( nil ) failure' ) ;
    end ;

    { Delete node at end... }
    dispose( Nodes[ 2 ], Done ) ;
    if( Outline.Count <> 2 ) then
    begin
        Error( 'Node deletion failure' ) ;
    end ;

    { Delete node in middle... }
    Nodes[ 2 ] := Outline.Add_Child( nil, 'Node 2' ) ;
    Nodes[ 3 ] := Outline.Add_Child( nil, 'Node 3' ) ;
    dispose( Nodes[ 2 ], Done ) ;
    if( Outline.Count <> 3 ) then
    begin
        Error( 'Node deletion failure' ) ;
    end ;

    { Add children nodes... }
    Nodes[ 4 ] := Outline.Add_Child( Nodes[ 0 ], 'Node 0.0' ) ;
    Nodes[ 5 ] := Outline.Add_Child( Nodes[ 0 ], 'Node 0.1' ) ;
    if( Nodes[ 0 ].Count <> 2 ) then
    begin
        Error( 'Child node creation failure' ) ;
    end ;

    { Add siblings... }
    Nodes[ 6 ] := Outline.Add_Sibling( Nodes[ 5 ], 'Node 0.2' ) ;
    if( Nodes[ 0 ].Count <> 3 ) then
    begin
        Error( 'Sibling node creation failure' ) ;
    end ;
    Nodes[ 7 ] := Outline.Add_Sibling( nil, 'Node 4' ) ;
    if( Outline.Count <> 4 ) then
    begin
        Error( 'Sibling node creation failure' ) ;
    end ;

    { Insert... }
    Nodes[ 2 ] := Outline.Insert_Child( nil, 0, 'Node -1' ) ;
    if( Outline.Count <> 5 ) then
    begin
        Error( 'Insert_Child failed' ) ;
    end ;

    Outline.Insert_Child( Nodes[ 0 ], 0, 'Node 0.-1' ) ;
    if( Nodes[ 0 ].Count <> 4 ) then
    begin
        Error( 'Child node insertion failure' ) ;
    end ;

    { Get tests... }
    Node := nil ;
    for Loop := 0 to Outline.Count do
    begin
        Node := Outline.Get_Child( nil, Loop )
    end ;
    if( Node <> nil ) then
    begin
        Error( 'Get_Child out of range failed' ) ;
    end ;
    for Loop := 0 to Nodes[ 0 ].Count do
    begin
        Node := Outline.Get_Child( Nodes[ 0 ], Loop )
    end ;
    if( Node <> nil ) then
    begin
        Error( 'Get_Child out of range failed' ) ;
    end ;
    if( Outline.Get_Next( Nodes[ 0 ] ) <> Nodes[ 1 ] ) then
    begin
        Error( 'Get_Next failure' ) ;
    end ;
    if( Outline.Get_Next( Nodes[ 6 ] ) <> nil ) then
    begin
        Error( 'Get_Next failure' ) ;
    end ;
    if( Outline.Get_Next( Nodes[ 7 ] ) <> nil ) then
    begin
        Error( 'Get_Next failure' ) ;
    end ;
    if( Outline.Get_Previous( Nodes[ 1 ] ) <> Nodes[ 0 ] ) then
    begin
        Error( 'Get_Previous failure' ) ;
    end ;
    if( Outline.Get_Previous( Nodes[ 6 ] ) <> Nodes[ 5 ] ) then
    begin
        Error( 'Get_Previous failure' ) ;
    end ;
    if( Outline.Get_Previous( Nodes[ 7 ] ) <> Nodes[ 3 ] ) then
    begin
        Error( 'Get_Previous failure' ) ;
    end ;

    { Test clear... }
    Nodes[ 0 ].Clear ;
    if( Nodes[ 0 ].Count <> 0 ) then
    begin
        Error( 'Clear failed' ) ;
    end ;
    Outline.Clear ;
    if( Outline.Count <> 0 ) then
    begin
        Error( 'Clear failed' ) ;
    end ;
    if( Outline.Get_Previous( nil ) <> nil ) then
    begin
        Error( 'Get_Previous( nil ) failure' ) ;
    end ;
    if( Outline.Get_Next( nil ) <> nil ) then
    begin
        Error( 'Get_Next( nil ) failure' ) ;
    end ;

    { Insert... }
    Outline.Insert_Child( nil, 10, 'Node 0' ) ;
    if( Outline.Count <> 1 ) then
    begin
        Error( 'Insert_Child failed' ) ;
    end ;
    Nodes[ 8 ] := new( PSOutline_Node, Init ) ;
    Nodes[ 8 ].Text := 'Node 8' ;
    Outline.Add_Node( nil, Nodes[ 8 ] ) ;
    if( Nodes[ 8 ].Parent = nil ) then
    begin
        Error( 'Add_Node failed' ) ;
    end ;
    Nodes[ 9 ] := new( PSOutline_Node, Init ) ;
    Nodes[ 9 ].Text := 'Node 9' ;
    Outline.Add_Node( Nodes[ 8 ], Nodes[ 9 ] ) ;
    if( Nodes[ 9 ].Parent = nil ) then
    begin
        Error( 'Add_Node failed' ) ;
    end ;
    Outline.Add_Node( nil, Nodes[ 9 ] ) ;
    if( Outline.Find( 'Node 8', '\', False ) = nil ) then
    begin
        Error( 'Find failed' ) ;
    end ;
    if( Outline.Find( 'Node 8', '\', True ) = nil ) then
    begin
        Error( 'Find failed' ) ;
    end ;
    if( Outline.Find( 'N*', '\', True ) = nil ) then
    begin
        Error( 'Find failed' ) ;
    end ;
    if( Outline.Find( 'Node 8888', '\', False ) <> nil ) then
    begin
        Error( 'Find failed' ) ;
    end ;

    {~~~Test Add_Outline}
    {~~~Test delete}
    {~~~Test Has_Children}

    { Cleanup... }
    Outline.Done ;
end ;
{$ENDIF}


end.

