{
        Program Name : RGB
        Package Name : LB
        Purpose      : Convert wavelength to TRGB value
        Institution  : Conroy & Conroy Co.
        Date Written : 20-Aug-2000
        Written By   : Alan Conroy
        Version      : 1.0

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          Converts a wavelength of light (in Angstroms) to TRGB.

          Original routine converted from Perl source by Teemu Makinen who
        graciously gave me permission to copy it and use as desired.  I
        modified the low-end to look more violet since his code made the
        colors look dark blue all the way to the lowest visible light
        wavelength.  The resulting RGB values more closely resemble pictures
        I've seen of white light being split by prisms.

}

unit RGB ;

interface

type TRGB = record { 16-bit RGB }
                R : word ;
                G : word ;
                B : word ;
            end ;

{ Returns RGB value for a wavelength of light (in Angstroms, or nm x 10). }
procedure Wavelength_To_RGB( W : integer ; var RGB : TRGB ) ;

implementation


{ Returns RGB value for a wavelength of light (in Angstroms, or nm x 10). }
procedure Wavelength_To_RGB( W : integer ; var RGB : TRGB ) ;

const RGBs : array[ 0..71, 0..3 ] of word =
	(       { Wavelength, R, G, B }
		( 3850, 0, 0, 0 ), { Black }
                ( 3900, 25000, 0, 65535 ), { Violet }
		( 3950, 24000, 0, 65535 ),
		( 4000, 23000, 0, 65535 ),
		( 4050, 22000, 0, 65535 ),
		( 4100, 21000, 0, 65535 ),
		( 4150, 20000, 0, 65535 ),
		( 4200, 19000, 0, 65535 ),
		( 4250, 14000, 0, 65535 ),
		( 4300, 9200, 0, 65535 ),
		( 4350, 4100, 0, 65535 ),
		( 4400, 400, 0, 65535 ),
		( 4450, 0, 45, 65535 ), { Blue }
		( 4500, 0, 699, 65535 ),
		( 4550, 0, 1727, 65535 ),
		( 4600, 0, 2960, 65535 ),
		( 4650, 0, 5413, 65535 ),
		( 4700, 0, 8092, 65535 ),
		( 4750, 0, 13283, 65535 ),
		( 4800, 0, 21765, 65535 ),
		( 4850, 0, 34439, 65535 ),
		( 4900, 0, 55780, 65535 ),
		( 4950, 0, 65535, 47646 ),
		( 5000, 0, 65535, 28717 ),
		( 5050, 0, 65535, 16850 ),
		( 5100, 0, 65535, 9384 ),
		( 5150, 0, 65535, 4666 ),
		( 5200, 0, 65535, 1915 ),
		( 5250, 0, 65535, 282 ),
		( 5300, 2691, 65535, 0 ), { Green }
		( 5350, 6463, 65535, 0 ),
		( 5400, 10468, 65535, 0 ),
		( 5450, 14797, 65535, 0 ),
		( 5500, 19975, 65535, 0 ),
		( 5550, 26465, 65535, 0 ),
		( 5600, 34333, 65535, 0 ),
		( 5650, 43850, 65535, 0 ),
		( 5700, 55305, 65535, 0 ),
		( 5750, 65535, 60534, 0 ),
		( 5800, 65535, 46814, 0 ),
		( 5850, 65535, 37062, 0 ),
		( 5900, 65535, 29227, 0 ),
		( 5950, 65535, 22444, 0 ),
		( 6000, 65535, 17155, 0 ),
		( 6050, 65535, 13126, 0 ),
		( 6100, 65535, 9839, 0 ),
		( 6150, 65535, 7170, 0 ),
		( 6200, 65535, 5153, 0 ),
		( 6250, 65535, 3619, 0 ),
		( 6300, 65535, 2378, 0 ),
		( 6350, 65535, 1367, 0 ),
		( 6400, 65535, 588, 0 ),
		( 6450, 65535, 16, 0 ),
		( 6500, 65535, 0, 46 ), { Red }
		( 6550, 65535, 0, 105 ),
		( 6600, 65535, 0, 168 ),
		( 6650, 65535, 0, 218 ),
		( 6700, 65535, 0, 229 ),
		( 6750, 65535, 0, 189 ),
		( 6800, 65535, 0, 138 ),
		( 6850, 65535, 0, 135 ),
		( 6900, 65535, 0, 170 ),
		( 6950, 65535, 0, 194 ),
		( 7000, 65535, 0, 193 ),
		( 7050, 65535, 0, 191 ),
		( 7100, 65535, 0, 214 ),
		( 7150, 65535, 0, 259 ),
		( 7200, 65535, 0, 307 ),
		( 7250, 65535, 0, 363 ),
		( 7300, 65535, 0, 426 ),
		( 7350, 0, 0, 0 ), { Black }
                ( 7400, 0, 0, 0 )
	) ;

    { Interpolate between values by 1/50 times Diff.  Index is 1, 2, or 3 for
      R, G, or B, respectively.  The values to interpolate between are the
      elements in R_G_B of Offset - 1 and Offset. }
    function Adjust_By_Fifths( Diff, Offset, Index : integer ) : longint ;

    begin
	Result := ( ( RGBs[ Offset, Index ] - RGBs[ Offset - 1, Index ] ) * Diff ) div 50 ;
    end ;

var Diff : integer ; { Difference between actual wavelength and table wavelength }
    Dummy : integer ; { Loop control }

begin
    if( ( W < 3850 ) or ( W > 7350 ) ) then
    begin
	fillchar( RGB, sizeof( RGB ), 0 ) ;
        exit ;
    end ;
    for Dummy := 0 to High( RGBs ) do
    begin
	if( RGBs[ Dummy, 0 ] > W ) then
	begin
	    RGB.R := RGBs[ Dummy - 1, 1 ] ;
	    RGB.G := RGBs[ Dummy - 1, 2 ] ;
	    RGB.B := RGBs[ Dummy - 1, 3 ] ;
	    Diff := W - RGBs[ Dummy - 1, 0 ] ;
	    RGB.R := RGB.R + Adjust_By_Fifths( Diff, Dummy, 1 ) ;
	    RGB.G := RGB.G + Adjust_By_Fifths( Diff, Dummy, 2 ) ;
	    RGB.B := RGB.B + Adjust_By_Fifths( Diff, Dummy, 3 ) ;
	    break ;
	end ;
    end ;
end ; { Wavelength_To_RGB }



end.
