{
        Program Name : RLoggers
        Package Name : LB
        Purpose      : Remote logging class
        Institution  : Conroy & Conroy Co.
        Date Written : 22-October-2003
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2001-2003 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit RLoggers ;

interface

{$I EDEFINES.INC}

uses Loggers, { TSLogger }
     MailSlot ; { TMail_Slot }

{$IFNDEF DELPHI}
type TRemote_Logger = object( TSLogger )
                           private { Instance data... }
                               Wrapped_Logger : PSLogger ;

                           public { API... }
                              { This procedure is called to log informational messages
                                to the code supplying the logger.  Mess is the message
                                text, Level indicates the type of message, and Abort
                                indicates whether or not to abort the process which
                                created the message.  If Abort is True upon return the
                                process should be terminated gracefully as quickly as
                                possible. }
                              procedure Log_Message( Mess : PChar ; Level : integer ;
                                  var Abort : boolean ) ; virtual ;

                              { Same as Log_Message, but the message will be followed
                                by a Unest_Message call when the condition causing the
                                message goes out of scope. }
                              procedure Log_Nested_Message( Mess : PChar ;
                                  Level : integer ; var Abort : boolean ) ; virtual ;

                              { This indicates the end of the condition under which the
                                previous message with the specified level occurred.
                                If Abort is True upon return, the process should be
                                terminated gracefully as quickly as possible. }
                              procedure Unnest_Message( Level : integer ;
                                  var Abort : boolean ) ; virtual ;

                              { This function is called during a long process to allow
                                for the user to abort it or for checking Windows
                                messages, etc.  If the function returns True, the
                                process should be terminated gracefully as quickly as
                                possible. }
                              function Check_Terminate : boolean ; virtual ;

                              procedure Initialize( const Name : string ; Server : boolean ) ;
                                  virtual ;

                              procedure Terminate ; virtual ;
{$ELSE}
type TRemote_Logger = class( TLogger )
                          private { Instance data... }
                              Mail_Slot : TMail_Slot ;
                              Last_Message : string ;
                              Message_Received : boolean ; { Set to true when client receives a message }
                              Message_Source : string ;
                              Server : boolean ;
                              Wrapped_Logger : PSLogger ;

                          protected { Internal utility routines... }
                              procedure CB_Message( Sender : TObject ;
                                  const Origin, Name, Data : string ) ;

                              procedure Wait_For_Response ;

                          public { API... }
                             procedure Log_Message( Mess : PChar ; Level : integer ;
                                 var Abort : boolean ) ;
                                 override ; stdcall ;

                             procedure Log_Nested_Message( Mess : PChar ;
                                 Level : integer ; var Abort : boolean ) ;
                                 override ; stdcall ;

                             procedure Unnest_Message( Level : integer ;
                                 var Abort : boolean ) ;
                                 override ; stdcall ;

                             function Check_Terminate : boolean ;
                                 override ; stdcall ;

                              procedure Initialize( const Name : string ;
                                  _Server : boolean ; Logger : PSLogger ) ;
                                  virtual ; stdcall ;

                              procedure Terminate ; virtual ; stdcall ;
{$ENDIF}
	       end ;

{$IFDEF DELPHI}
     PRemote_Logger = TRemote_Logger ;
{$ELSE}
     PRemote_Logger = ^TRemote_Logger ;
{$ENDIF}

implementation

uses { Borland... }
     Windows, { Sleep }

     Forms ; { Application }


function Decode_Integer( const S : string ; Index : longint ) : longint ;

begin
    move( PChar( S )[ Index - 1 ], Result, sizeof( Result ) ) ;
end ;


function Encode_Integer( I : longint ) : string ;

begin
    setlength( Result, sizeof( longint ) ) ;
    move( I, PChar( Result )^, length( Result ) ) ;
end ;



{ TRemote_Logger methods... }

{ Internal utility routines... }

procedure TRemote_Logger.CB_Message( Sender : TObject ;
    const Origin, Name, Data : string ) ;

var B : boolean ;
    Command : string ;
    Dummy : integer ;
    S : string ;

begin
    if( Server ) then
    begin
        Message_Source := Origin + '\' + Name ;
        Dummy := pos( '|', Data ) ;
        Command := copy( Data, 1, Dummy - 1 ) ;
        if( Command = 'LOG_MESSAGE' ) then
        begin
            S := copy( Data, Dummy + 1 + sizeof( longint ) * 2, length( Data ) ) ;
            Dummy := Decode_Integer( Data, Dummy + 1 ) ;
            B := boolean( Decode_Integer( Data, Dummy + 1 + sizeof( longint ) ) ) ;
            Log_Message( PChar( S ), Dummy, B ) ;
            Mail_Slot.Send( Message_Source, Encode_Integer( ord( B ) ) ) ;
        end else
        if( Command = 'LOG_NESTED_MESSAGE' ) then
        begin
            S := copy( Data, Dummy + 1 + sizeof( longint ) * 2, length( Data ) ) ;
            Dummy := Decode_Integer( Data, Dummy + 1 ) ;
            B := boolean( Decode_Integer( Data, Dummy + 1 + sizeof( longint ) ) ) ;
            Log_Nested_Message( PChar( S ), Dummy, B ) ;
            Mail_Slot.Send( Message_Source, Encode_Integer( ord( B ) ) ) ;
        end else
        if( Command = 'UNNEST_MESSAGE' ) then
        begin
            B := boolean( Decode_Integer( Data, Dummy + 1 ) ) ;
            Unnest_Message( Decode_Integer( Data, Dummy + 1 ), B ) ;
            Mail_Slot.Send( Message_Source, Encode_Integer( ord( B ) ) ) ;
        end else
        if( Command = 'CHECK_TERMINATE' ) then
        begin
            Check_Terminate ;
        end ;
    end else
    begin
        Last_Message := Data ;
        Message_Received := True ;
    end ;
end ;


procedure TRemote_Logger.Wait_For_Response ;

begin
    while( not Message_Received ) do
    begin
        Sleep( 0 ) ;
        Application.ProcessMessages ;
    end ;
    Message_Received := False ;
end ;


{ API... }

procedure TRemote_Logger.Log_Message( Mess : PChar ; Level : integer ;
    var Abort : boolean ) ;

begin
    if( Server ) then
    begin
        Wrapped_Logger.Log_Message( Mess, Level, Abort ) ;
    end else
    begin
        Mail_Slot.Send( '.', 'LOG_MESSAGE|' + Encode_Integer( Level ) + Encode_Integer( ord( Abort ) ) + Mess ) ;
        Wait_For_Response ;
        Abort := boolean( Decode_Integer( Last_Message, 1 ) ) ;
    end ;
end ;


procedure TRemote_Logger.Log_Nested_Message( Mess : PChar ; Level : integer ;
    var Abort : boolean ) ;

begin
    if( Server ) then
    begin
        Wrapped_Logger.Log_Nested_Message( Mess, Level, Abort ) ;
    end else
    begin
        Mail_Slot.Send( '.', 'LOG_NESTED_MESSAGE|' + Encode_Integer( Level ) + Encode_Integer( ord( Abort ) ) + Mess ) ;
        Wait_For_Response ;
        Abort := boolean( Decode_Integer( Last_Message, 1 ) ) ;
    end ;
end ;


procedure TRemote_Logger.Unnest_Message( Level : integer ;
    var Abort : boolean ) ;

begin
    if( Server ) then
    begin
        Wrapped_Logger.Unnest_Message( Level, Abort ) ;
        Mail_Slot.Send( Message_Source, Encode_Integer( ord( Abort ) ) ) ;
    end else
    begin
        Mail_Slot.Send( '.', 'UNNEST_MESSAGE|' + Encode_Integer( Level ) ) ;
        Wait_For_Response ;
        Abort := boolean( Decode_Integer( Last_Message, 1 ) ) ;
    end ;
end ;


function TRemote_Logger.Check_Terminate : boolean ;

var B : boolean ;

begin
    if( Server ) then
    begin
        B := Wrapped_Logger.Check_Terminate ;
        Check_Terminate := B ;
        Mail_Slot.Send( Message_Source, Encode_Integer( ord( B ) ) ) ;
    end else
    begin
        Mail_Slot.Send( '.', 'CHECK_TERMINATE|' ) ;
        Wait_For_Response ;
        Check_Terminate := boolean( Decode_Integer( Last_Message, 1 ) ) ;
    end ;
end ;


procedure TRemote_Logger.Initialize( const Name : string ;
    _Server : boolean ; Logger : PSLogger ) ;

begin
    Mail_Slot := TMail_Slot.Create( Name, _Server ) ;
    Mail_Slot.Activate ;
    Wrapped_Logger := Logger ;
    Server := _Server ;
    Mail_Slot.On_Message := CB_Message ;
    if( Wrapped_Logger <> nil ) then
    begin
{        Wrapped_Logger.Attach ;}
    end ;
end ;


procedure TRemote_Logger.Terminate ;

begin
    Mail_Slot.Free ;
    Mail_Slot := nil ;
    if( Wrapped_Logger <> nil ) then
    begin
{        Wrapped_Logger.Detach ;}
    end ;
    Wrapped_Logger := nil ;
end ;


end.
