{
        Program Name : UOS_FS_Files
        Package Name : LB
        Purpose      : UOS File support
        Institution  : Conroy & Conroy Co.
        Date Written : 10-Jun-2006
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2006 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit contains classes that implement UOS files.

          TSExtent_Properties is a class that implements a collection of
        properties, each of which is an Extent Store.  The properties data
        itself is also an Extent Store.
}

unit UOSFSFil ;

interface

uses { C&C Layer 0... }
     _Common, { TObject }
     _Stores, { TStore_Address }

     { C&C Layer 1... }
     Collect, { TString_Collection }
     MStores, { PSManaged_Store }
     _UEHDefs, { TUEC }

     { C&C Layer 2... }
     ExtStore ; { TSExtent_Store }

type TSExtent_Properties = object( TObject )
                             public { Constructors and destructors... }
                                 constructor Init ;
                                 destructor Done ; virtual ;

                             private { Instance data... }
                                 _Properties : PSExtent_Store ; { Store that contains property data }
                                 Names : TString_Collection ; { Cached names of properties }
                                 Addresses : TInteger_Collection ; { Cached addresses of properties }
                                 _Stores : TCollection ;

                                 _Root : TStore_Address ;
                                 _Store : PSManaged_Store ;

                             private { Internal utility routines... }
                                 procedure Set_Error( Value : integer ) ;
                                 procedure Update ;
                                 function Valid_Index( Index : longint ) : boolean ;

                             public { API... }
                                 Last_Error : TUEC ;

                                 { General property methods... }
                                 function Count : longint ; virtual ; { Number of properties }
                                 function Name( Index : longint ) : string ;
                                     virtual ; { Name of property with index }
                                 function Add( _Name : string ) : longint ;
                                     virtual ; { Add a property }
                                 procedure Delete ; virtual ;
                                 procedure Load( S : PSManaged_Store ;
                                     R : TStore_Address ) ; virtual ;

                                 function Get_Clustersize : TStore_Address ;
                                     virtual ;
                                 function Get_Contiguous_Flag : boolean ;
                                     virtual ;
                                 function Get_Placed_Flag : boolean ; virtual ;
                                 function Get_Read_Only_Flag : boolean ;
                                     virtual ;
                                 function Get_Write_Only_Flag : boolean ;
                                     virtual ;
                                 function Get_Root : TStore_Address ; virtual ;
                                 function Get_Store : PSStore ; virtual ;
                                 procedure Set_Clustersize( Value : TStore_Address ) ;
                                     virtual ;
                                 procedure Set_Contiguous_Flag( Value : boolean ) ;
                                     virtual ;
                                 procedure Set_Placed_Flag( Value : boolean ) ;
                                     virtual ;
                                 procedure Set_Read_Only_Flag( Value : boolean ) ;
                                     virtual ;
                                 procedure Set_Root( Value : integer ) ;
                                      virtual ;
                                 procedure Set_Store( S : PSManaged_Store ) ;
                                     virtual ;
                                 procedure Set_Read_Only( Value : boolean ) ;
                                     virtual ;
                                 procedure Set_Write_Only( Value : boolean ) ;
                                     virtual ;

                                 { Individual property access... }
                                 function Read_Index( Index : longint ;
                                     var Data ; Address, Len : TStore_Address ;
                                     var UEC : TUEC ) : TStore_Address ; virtual ;
                                 function Write_Index( Index : longint ; var Data ;
                                     Address, Len : TStore_Address ;
                                     var UEC : TUEC ) : TStore_Address ; virtual ;

                                 function Get_Clustersize_Index( Index : longint ) : TStore_Address ;
                                     virtual ;
                                 function Get_Contiguous_Flag_Index( Index : longint ) : boolean ;
                                     virtual ;
                                 function Get_Placed_Flag_Index( Index : longint ) : boolean ; virtual ;
                                 function Get_Read_Only_Flag_Index( Index : longint ) : boolean ;
                                     virtual ;
                                 function Get_Write_Only_Flag_Index( Index : longint ) : boolean ;
                                     virtual ;
                                 procedure Set_Clustersize_Index( Index : longint ;
                                     Value : TStore_Address ) ; virtual ;
                                 procedure Set_Contiguous_Flag_Index( Index : longint ;
                                     Value : boolean ) ; virtual ;
                                 procedure Set_Placed_Flag_Index( Index : longint ;
                                     Value : boolean ) ; virtual ;
                                 procedure Set_Read_Only_Flag_Index( Index : longint ;
                                     Value : boolean ) ; virtual ;
                                 procedure Set_Read_Only_Index( Index : longint ;
                                     Value : boolean ) ; virtual ;
                                 procedure Set_Write_Only_Index( Index : longint ;
                                     Value : boolean ) ; virtual ;
                                 procedure Delete_Index( Index : longint ) ;
                                     virtual ;
                         end ; { TSExtent_Properties }
     PSFile_Properties = ^TSExtent_Properties ;

const File_PropertiesErr_Facility = 117 ;
const ExtentPropertiesErr_Read_Only = 1 ;
const ExtentPropertiesErr_Invalid_Index = 2 ;

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

{$IFDEF Test}
uses Heaps ; // TSHeap
{$ENDIF}


{ TSExtent_Properties methods... }

{ Constructors and destructors... }

constructor TSExtent_Properties.Init ;

begin
    inherited Init ;

    _Properties := nil ;
    Names.Init( 0, 4 ) ;
    Addresses.Init( 0, 4 ) ;
    _Stores.Init( 0, 4 ) ;
    fillchar( Last_Error, sizeof( Last_Error ), 0 ) ;
end ;


destructor TSExtent_Properties.Done ;

begin
    _Properties.Free ;
    _Properties := nil ;
    Names.Done ;
    Addresses.Done ;
    _Stores.Done ;

    inherited Done ;
end ;


{ Internal utility routines... }

procedure TSExtent_Properties.Set_Error( Value : integer ) ;

begin
    Last_Error.Facility := File_PropertiesErr_Facility ;
    Last_Error.Code := Value ;
end ;


procedure TSExtent_Properties.Update ;

var A, Ad : TStore_Address ;
    Loop : integer ;
    S : string[ 255 ] ;
    UEC : TUEC ;

begin
    if( _Store <> nil ) then
    begin
        { Write property data... }
        A := 0 ;
        _Properties.Set_Max_Storage( 0, UEC ) ;
        for Loop := 0 to Count - 1 do
        begin
            if( Names.At( Loop ) <> '' ) then
            begin
                _Properties.Write_Data( S[ 0 ], A, length( S ) + 1, UEC ) ;
                A := A + length( S ) + 1 ;
                Ad := Addresses.At( Loop ) ;
                _Properties.Write_Data( Ad, A, sizeof( Ad ), UEC ) ;
                A := A + sizeof( Ad ) ;
            end ;
        end ;
    end ;
end ;


function TSExtent_Properties.Valid_Index( Index : longint ) : boolean ;

var S : PSExtent_Store ;

begin
    if( ( Index < 0 ) or ( Index >= Names.Count ) ) then
    begin
        Valid_Index := False ;
        exit ;
    end ;
    if( Names.At( Index ) = '' ) then
    begin
        Valid_Index := False ;
        exit ;
    end ;
    if( _Stores.At( Index ) = nil ) then
    begin
        S := new( PSExtent_Store, Init ) ;
        _Stores.AtPut( Index, S ) ;
        S.Load_From_Store( _Store, Addresses.At( Index ) ) ;
    end ;
    Valid_Index := True ;
end ;


{ API... }

function TSExtent_Properties.Count : longint ;

begin
    Count := Names.Count ;
end ;


function TSExtent_Properties.Name( Index : longint ) : string ;

begin
    if( not Valid_Index( Index ) ) then
    begin
        Name := '' ;
        exit ;
    end ;
    Name := Names.At( Index ) ;
end ;


function TSExtent_Properties.Add( _Name : string ) : longint ;

var I : integer ;

begin
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentPropertiesErr_Read_Only ) ;
        Add := -1 ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    I := Names.Index_Of_Uppercase( _Name ) ;
    if( I = -1 ) then
    begin
        Names.Insert( _Name ) ;
        I := Count - 1 ;
        Addresses.Insert( 0 ) ;
        _Stores.Insert( nil ) ;
    end ;
    Add := I ;
end ;


function TSExtent_Properties.Get_Clustersize : TStore_Address ;

begin
    Get_Clustersize := _Properties.Get_Clustersize ;
    Last_Error.Facility := _Properties.Last_Error.Facility ;
    Last_Error.Code := _Properties.Last_Error.Code ;
end ;


function TSExtent_Properties.Get_Contiguous_Flag : boolean ;

begin
    Get_Contiguous_Flag := _Properties.Get_Contiguous_Flag ;
    Last_Error.Facility := _Properties.Last_Error.Facility ;
    Last_Error.Code := _Properties.Last_Error.Code ;
end ;


function TSExtent_Properties.Get_Placed_Flag : boolean ;

begin
    Get_Placed_Flag := _Properties.Get_Placed_Flag ;
    Last_Error.Facility := _Properties.Last_Error.Facility ;
    Last_Error.Code := _Properties.Last_Error.Code ;
end ;


function TSExtent_Properties.Get_Read_Only_Flag : boolean ;

begin
    Get_Read_Only_Flag := _Properties.Get_Read_Only_Flag ;
    Last_Error.Facility := _Properties.Last_Error.Facility ;
    Last_Error.Code := _Properties.Last_Error.Code ;
end ;


function TSExtent_Properties.Get_Write_Only_Flag : boolean ;

begin
    Get_Write_Only_Flag := _Properties.Get_Write_Only_Flag ;
    Last_Error.Facility := _Properties.Last_Error.Facility ;
    Last_Error.Code := _Properties.Last_Error.Code ;
end ;


function TSExtent_Properties.Get_Root : TStore_Address ;

begin
    Get_Root := _Root ;
end ;


function TSExtent_Properties.Get_Store : PSStore ;

begin
    Get_Store := _Store ;
end ;


procedure TSExtent_Properties.Set_Clustersize( Value : TStore_Address ) ;

begin
    _Properties.Set_Clustersize( Value ) ;
    Last_Error.Facility := _Properties.Last_Error.Facility ;
    Last_Error.Code := _Properties.Last_Error.Code ;
end ;


procedure TSExtent_Properties.Set_Contiguous_Flag( Value : boolean ) ;

begin
    _Properties.Set_Contiguous_Flag( Value ) ;
    Last_Error.Facility := _Properties.Last_Error.Facility ;
    Last_Error.Code := _Properties.Last_Error.Code ;
end ;


procedure TSExtent_Properties.Set_Placed_Flag( Value : boolean ) ;

begin
    _Properties.Set_Placed_Flag( Value ) ;
    Last_Error.Facility := _Properties.Last_Error.Facility ;
    Last_Error.Code := _Properties.Last_Error.Code ;
end ;


procedure TSExtent_Properties.Set_Read_Only_Flag( Value : boolean ) ;

begin
    _Properties.Set_Read_Only_Flag( Value ) ;
    Last_Error.Facility := _Properties.Last_Error.Facility ;
    Last_Error.Code := _Properties.Last_Error.Code ;
end ;


procedure TSExtent_Properties.Set_Root( Value : integer ) ;

begin
    Set_Error( 0 ) ;
    _Root := Value ;
    Load( _Store, _Root ) ;
end ;


procedure TSExtent_Properties.Set_Store( S : PSManaged_Store ) ;

var Loop : integer ;
    _S : PSExtent_Store ;

begin
    Set_Error( 0 ) ;
    _Store := S ;
    Load( _Store, _Root ) ;
    for Loop := 0 to Count - 1 do
    begin
        if( Loop < _Stores.Count ) then
        begin
            _S := PSExtent_Store( _Stores.At( Loop ) ) ;
            if( _S <> nil ) then
            begin
                _S.Set_Store( S ) ;
            end ;
        end ;
    end ;
end ;


procedure TSExtent_Properties.Set_Read_Only( Value : boolean ) ;

begin
    _Properties.Set_Read_Only( Value ) ;
    Last_Error.Facility := _Properties.Last_Error.Facility ;
    Last_Error.Code := _Properties.Last_Error.Code ;
end ;


procedure TSExtent_Properties.Set_Write_Only( Value : boolean ) ;

begin
    _Properties.Set_Write_Only( Value ) ;
    Last_Error.Facility := _Properties.Last_Error.Facility ;
    Last_Error.Code := _Properties.Last_Error.Code ;
end ;


procedure TSExtent_Properties.Delete ;

begin
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentPropertiesErr_Read_Only ) ;
        exit ;
    end ;
    _Properties.Delete ;
    Last_Error.Facility := _Properties.Last_Error.Facility ;
    Last_Error.Code := _Properties.Last_Error.Code ;
    Delete_Index( -1 ) ;
    _Root := 0 ;
end ;


procedure TSExtent_Properties.Load( S : PSManaged_Store ; R : TStore_Address ) ;

var A, Ad, SA : TStore_Address ;
    B : byte ;
    St : string[ 255 ] ;
    UEC : TUEC ;

begin
    _Store := S ;
    _Root := R ;
    if( ( _Store <> nil ) and ( R <> 0 ) ) then
    begin
        Addresses.DeleteAll ;
        Names.DeleteAll ;
        _Stores.DeleteAll ;
        A := _Properties.Max_Storage ;
        Ad := R ;
        while( A > 0 ) do
        begin
            _Properties.Read_Data( B, Ad, 1, UEC ) ;
            St[ 0 ] := chr( B ) ;
            Ad := Ad + 1 ;
            _Properties.Read_Data( St[ 1 ], Ad, B, UEC ) ;
            Names.Insert( St ) ;
            Ad := Ad + B ;
            _Properties.Read_Data( SA, Ad, sizeof( SA ), UEC ) ;
            Addresses.Insert( SA ) ;
            Ad := Ad + sizeof( SA ) ;
            A := A - B - 1 - sizeof( SA ) ;
            _Stores.Insert( nil ) ;
        end ; { while( A > 0 ) }
    end ; { if( _Store <> nil )  }
end ; { TSExtent_Properties.Load }


{ Individual property access... }

function TSExtent_Properties.Read_Index( Index : longint ;
    var Data ; Address, Len : TStore_Address ;
    var UEC : TUEC ) : TStore_Address  ;

var S : PSExtent_Store ;

begin
    if( not Valid_Index( Index ) ) then
    begin
        Read_Index := 0 ;
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    Read_Index := S.Read_Data( Data, Address, Len, UEC ) ;
    Last_Error.Facility := S.Last_Error.Facility ;
    Last_Error.Code := S.Last_Error.Code ;
end ;


function TSExtent_Properties.Write_Index( Index : longint ;
    var Data ; Address, Len : TStore_Address ;
    var UEC : TUEC ) : TStore_Address ;

var S : PSExtent_Store ;

begin
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentPropertiesErr_Read_Only ) ;
        Write_Index := 0 ;
        exit ;
    end ;
    if( not Valid_Index( Index ) ) then
    begin
        Write_Index := 0 ;
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    Write_Index := S.Write_Data( Data, Address, Len, UEC ) ;
end ;


function TSExtent_Properties.Get_Clustersize_Index( Index : longint ) : TStore_Address ;

var S : PSExtent_Store ;

begin
    if( not Valid_Index( Index ) ) then
    begin
        Get_Clustersize_Index := 0 ;
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    Get_Clustersize_Index := S.Get_Clustersize ;
end ;


function TSExtent_Properties.Get_Contiguous_Flag_Index( Index : longint ) : boolean ;

var S : PSExtent_Store ;

begin
    if( not Valid_Index( Index ) ) then
    begin
        Get_Contiguous_Flag_Index := False ;
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    Get_Contiguous_Flag_Index := S.Get_Contiguous_Flag ;
end ;


function TSExtent_Properties.Get_Placed_Flag_Index( Index : longint ) : boolean ;

var S : PSExtent_Store ;

begin
    if( not Valid_Index( Index ) ) then
    begin
        Get_Placed_Flag_Index := False ;
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    Get_Placed_Flag_Index := S.Get_Placed_Flag ;
end ;


function TSExtent_Properties.Get_Read_Only_Flag_Index( Index : longint ) : boolean ;

var S : PSExtent_Store ;

begin
    if( not Valid_Index( Index ) ) then
    begin
        Get_Read_Only_Flag_Index := True ;
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    Get_Read_Only_Flag_Index := S.Get_Read_Only_Flag ;
end ;


function TSExtent_Properties.Get_Write_Only_Flag_Index( Index : longint ) : boolean ;

var S : PSExtent_Store ;

begin
    if( not Valid_Index( Index ) ) then
    begin
        Get_Write_Only_Flag_Index := True ;
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    Get_Write_Only_Flag_Index := S.Get_Write_Only_Flag ;
end ;


procedure TSExtent_Properties.Set_Clustersize_Index( Index : longint ;
    Value : TStore_Address ) ;

var Loop : integer ;
    S : PSExtent_Store ;

begin
    if( Index = -1 ) then // Do all properties
    begin
        for Loop := 0 to Count - 1 do
        begin
            Valid_Index( Index ) ;
            S := PSExtent_Store( _Stores.At( Loop ) ) ;
            if( S <> nil ) then
            begin
                S.Set_Clustersize( Value ) ;
            end ;
        end ;
        exit ;
    end ;
    if( not Valid_Index( Index ) ) then
    begin
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    if( S <> nil ) then
    begin
        S.Set_Clustersize( Value ) ;
        Last_Error.Facility := S.Last_Error.Facility ;
        Last_Error.Code := S.Last_Error.Code ;
    end ;
end ;


procedure TSExtent_Properties.Set_Contiguous_Flag_Index( Index : longint ;
    Value : boolean ) ;

var Loop : integer ;
    S : PSExtent_Store ;

begin
    if( Index = -1 ) then // Do all properties
    begin
        for Loop := 0 to Count - 1 do
        begin
            Valid_Index( Index ) ;
            S := PSExtent_Store( _Stores.At( Loop ) ) ;
            if( S <> nil ) then
            begin
                S.Set_Contiguous_Flag( Value ) ;
            end ;
        end ;
        exit ;
    end ;
    if( not Valid_Index( Index ) ) then
    begin
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    if( S <> nil ) then
    begin
        S.Set_Contiguous_Flag( Value ) ;
        Last_Error.Facility := S.Last_Error.Facility ;
        Last_Error.Code := S.Last_Error.Code ;
    end ;
end ;


procedure TSExtent_Properties.Set_Placed_Flag_Index( Index : longint ;
    Value : boolean ) ;

var Loop : integer ;
    S : PSExtent_Store ;

begin
    if( Index = -1 ) then // Do all properties
    begin
        for Loop := 0 to Count - 1 do
        begin
            Valid_Index( Index ) ;
            S := PSExtent_Store( _Stores.At( Loop ) ) ;
            if( S <> nil ) then
            begin
                S.Set_Placed_Flag( Value ) ;
            end ;
        end ;
        exit ;
    end ;
    if( not Valid_Index( Index ) ) then
    begin
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    if( S <> nil ) then
    begin
        S.Set_Placed_Flag( Value ) ;
        Last_Error.Facility := S.Last_Error.Facility ;
        Last_Error.Code := S.Last_Error.Code ;
    end ;
end ;


procedure TSExtent_Properties.Set_Read_Only_Flag_Index( Index : longint ;
    Value : boolean ) ;

var Loop : integer ;
    S : PSExtent_Store ;

begin
    if( Index = -1 ) then // Do all properties
    begin
        for Loop := 0 to Count - 1 do
        begin
            Valid_Index( Index ) ;
            S := PSExtent_Store( _Stores.At( Loop ) ) ;
            if( S <> nil ) then
            begin
                S.Set_Read_Only_Flag( Value ) ;
            end ;
        end ;
        exit ;
    end ;
    if( not Valid_Index( Index ) ) then
    begin
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    if( S <> nil ) then
    begin
        S.Set_Read_Only_Flag( Value ) ;
        Last_Error.Facility := S.Last_Error.Facility ;
        Last_Error.Code := S.Last_Error.Code ;
    end ;
end ;


procedure TSExtent_Properties.Set_Read_Only_Index( Index : longint ;
    Value : boolean ) ;

var Loop : integer ;
    S : PSExtent_Store ;

begin
    if( Index = -1 ) then // Do all properties
    begin
        for Loop := 0 to Count - 1 do
        begin
            Valid_Index( Index ) ;
            S := PSExtent_Store( _Stores.At( Loop ) ) ;
            if( S <> nil ) then
            begin
                S.Set_Read_Only( Value ) ;
            end ;
        end ;
        exit ;
    end ;
    if( not Valid_Index( Index ) ) then
    begin
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    if( S <> nil ) then
    begin
        S.Set_Read_Only( Value ) ;
        Last_Error.Facility := S.Last_Error.Facility ;
        Last_Error.Code := S.Last_Error.Code ;
    end ;
end ;


procedure TSExtent_Properties.Set_Write_Only_Index( Index : longint ;
    Value : boolean ) ;

var Loop : integer ;
    S : PSExtent_Store ;

begin
    if( Index = -1 ) then // Do all properties
    begin
        for Loop := 0 to Count - 1 do
        begin
            Valid_Index( Index ) ;
            S := PSExtent_Store( _Stores.At( Loop ) ) ;
            if( S <> nil ) then
            begin
                S.Set_Write_Only( Value ) ;
            end ;
        end ;
        exit ;
    end ;
    if( not Valid_Index( Index ) ) then
    begin
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    if( S <> nil ) then
    begin
        S.Set_Write_Only( Value ) ;
        Last_Error.Facility := S.Last_Error.Facility ;
        Last_Error.Code := S.Last_Error.Code ;
    end ;
end ;


procedure TSExtent_Properties.Delete_Index( Index : longint ) ;

var Loop : integer ;
    S : PSExtent_Store ;

begin
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentPropertiesErr_Read_Only ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    if( Index = -1 ) then // Do all properties
    begin
        for Loop := Count - 1 downto 0 do
        begin
            Valid_Index( Index ) ;
            S := PSExtent_Store( _Stores.At( Loop ) ) ;
            if( S <> nil ) then
            begin
                S.Delete ;
                dispose( S, Done ) ;
                _Stores.AtPut( Loop, nil ) ;
                Names.Atput( Loop, '' ) ;
                Addresses.AtPut( Loop, 0 ) ;
            end ;
        end ;
        Update ;
        exit ;
    end ;
    if( not Valid_Index( Index ) ) then
    begin
        Set_Error( ExtentPropertiesErr_Invalid_Index ) ;
        exit ;
    end ;
    S := PSExtent_Store( _Stores.At( Index ) ) ;
    S.Delete ;
    dispose( S, Done ) ;
    _Stores.AtPut( Index, nil ) ;
    Names.Atput( Index, '' ) ;
    Addresses.AtPut( Index, 0 ) ;
    Update ;
end ; { TSExtent_Properties.Delete_Index }



{$IFDEF Test}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


procedure Test_Unit ;

var Managed_Store : TSHeap ;
    S, S1, S2 : string[ 255 ] ;
    Test : TSExtent_Properties ;
    UEC : TUEC ;

begin
    Managed_Store.Init ;
    Managed_Store.Initialize ;
    Test.Init ;

    if( Test.Count <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Get_Store <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Get_Root <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Get_Contiguous_Flag ) then
    begin
        Error ;
    end ;
    if( Test.Get_Read_Only_Flag ) then
    begin
        Error ;
    end ;
    if( Test.Get_Write_Only_Flag ) then
    begin
        Error ;
    end ;
    if( Test.Get_Placed_Flag ) then
    begin
        Error ;
    end ;

    Test.Set_Store( @Managed_Store ) ;
    if( Test.Get_Contiguous_Flag_Index( 0 ) ) then
    begin
        Error ;
    end ;
    if( Test.Get_Read_Only_Flag_Index( 0 ) ) then
    begin
        Error ;
    end ;
    if( Test.Get_Write_Only_Flag_Index( 0 ) ) then
    begin
        Error ;
    end ;
    if( Test.Get_Placed_Flag_Index( 0 ) ) then
    begin
        Error ;
    end ;
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;

    if( Test.Add( 'Test 1' ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 1 ) then
    begin
        Error ;
    end ;
    if( Test.Get_Contiguous_Flag_Index( 0 ) ) then
    begin
        Error ;
    end ;
    if( Test.Get_Read_Only_Flag_Index( 0 ) ) then
    begin
        Error ;
    end ;
    if( Test.Get_Write_Only_Flag_Index( 0 ) ) then
    begin
        Error ;
    end ;
    if( Test.Get_Placed_Flag_Index( 0 ) ) then
    begin
        Error ;
    end ;
    if( Test.Name( 0 ) <> 'Test 1' ) then
    begin
        Error ;
    end ;
    if( Test.Add( 'Test 2' ) <> 1 ) then
    begin
        Error ;
    end ;
    if( Test.Get_Clustersize <> Managed_Store.Min_Storage ) then
    begin
        Error ;
    end ;
    if( Test.Get_Clustersize_Index( 0 ) <> Managed_Store.Min_Storage ) then
    begin
        Error ;
    end ;
    S1 := 'Test text for Test 1' ;
    Test.Write_Index( 0, S1[ 0 ], 0, length( S1 ) + 1, UEC ) ;
    if( UEC.Code <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Read_Index( 0, S[ 0 ], 0, length( S ) + 1, UEC ) <> length( S1 ) + 1 ) then
    begin
        Error ;
    end ;
    if( S <> S1 ) then
    begin
        Error ;
    end ;
    S2 := 'Test text for Test 2' ;
    Test.Write_Index( 1, S2[ 0 ], 0, length( S2 ) + 1, UEC ) ;
    if( UEC.Code <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Read_Index( 1, S[ 0 ], 0, length( S ) + 1, UEC ) <> length( S2 ) + 1 ) then
    begin
        Error ;
    end ;
    if( S <> S2 ) then
    begin
        Error ;
    end ;

    Test.Delete_Index( 0 ) ;
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Name( 1 ) <> 'Test 2' ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 2 ) then
    begin
        Error ;
    end ;

    Test.Set_Clustersize( 256 ) ;
    if( Test.Get_Clustersize <> 256 ) then
    begin
        Error ;
    end ;
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Name( 1 ) <> 'Test 2' ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 2 ) then
    begin
        Error ;
    end ;

    Test.Set_Contiguous_Flag( True ) ;
    if( not Test.Get_Contiguous_Flag ) then
    begin
        Error ;
    end ;
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Name( 1 ) <> 'Test 2' ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 2 ) then
    begin
        Error ;
    end ;

    Test.Set_Contiguous_Flag( False ) ;
    if( Test.Get_Contiguous_Flag ) then
    begin
        Error ;
    end ;
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Name( 1 ) <> 'Test 2' ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 2 ) then
    begin
        Error ;
    end ;

    Test.Set_Placed_Flag( True ) ;
    if( not Test.Get_Placed_Flag ) then
    begin
        Error ;
    end ;
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Name( 1 ) <> 'Test 2' ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 2 ) then
    begin
        Error ;
    end ;

    Test.Set_Read_Only_Flag( True ) ;
    if( not Test.Get_Read_Only_Flag ) then
    begin
        Error ;
    end ;
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Name( 1 ) <> 'Test 2' ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 2 ) then
    begin
        Error ;
    end ;

    Test.Set_Read_Only_Flag( False ) ;
    if( Test.Get_Read_Only_Flag ) then
    begin
        Error ;
    end ;
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Name( 1 ) <> 'Test 2' ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 2 ) then
    begin
        Error ;
    end ;

    Test.Set_Read_Only( True ) ;
(*
    if( not Test.Get_Read_Only ) then
    begin
        Error ;
    end ;
*)
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Name( 1 ) <> 'Test 2' ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 2 ) then
    begin
        Error ;
    end ;

    Test.Set_Read_Only( False ) ;
(*
    if( Test.Get_Read_Only ) then
    begin
        Error ;
    end ;
*)
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Name( 1 ) <> 'Test 2' ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 2 ) then
    begin
        Error ;
    end ;

    Test.Set_Write_Only( True ) ;
(*
    if( not Test.Get_Write_Only ) then
    begin
        Error ;
    end ;
*)
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Name( 1 ) <> 'Test 2' ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 2 ) then
    begin
        Error ;
    end ;

    Test.Set_Write_Only( False ) ;
(*
    if( Test.Get_Write_Only ) then
    begin
        Error ;
    end ;
*)
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Name( 1 ) <> 'Test 2' ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 2 ) then
    begin
        Error ;
    end ;

    Test.Set_Root( Test.Get_Root ) ;
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Name( 1 ) <> 'Test 2' ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 2 ) then
    begin
        Error ;
    end ;

    Test.Set_Clustersize_Index( 1, 256 ) ;
    if( Test.Get_Clustersize_Index( 1 ) <> 256 ) then
    begin
        Error ;
    end ;
    //~~~
(*
    Test.Set_Contiguous_Flag_Index( 1, Value : boolean ) ;
    Test.Set_Placed_Flag_Index( 1, Value : boolean ) ;
    Test.Set_Read_Only_Flag_Index( 1, Value : boolean ) ;
    Test.Set_Read_Only_Index( 1, Value : boolean ) ;
    Test.Set_Write_Only_Index( 1, Value : boolean ) ;
*)

    Test.Load( @Managed_Store, Test.Get_Root ) ;
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    if( Test.Name( 1 ) <> 'Test 2' ) then
    begin
        Error ;
    end ;
    if( Test.Count <> 2 ) then
    begin
        Error ;
    end ;

    Test.Delete ;
    if( Test.Count <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Name( 0 ) <> '' ) then
    begin
        Error ;
    end ;

    Test.Done ;
    Managed_Store.Done ;
end ;
{$ENDIF}



end.

