unit _Cache ;
{
        Program Name : _Cache
        Package Name : LB
        Purpose      : Cache class definition
        Institution  : Conroy & Conroy Co.
        Date Written : 18-Jun-1994
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 1994-2005 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit defines classes for handling caches.
}

interface

{$I EDEFINES.INC}

uses { Layer 0... }
     ComInter, { TCommon_COM_Interface }
     _Common ; { TBase_Object }

type TCache_Statistics = record
			     Hits : longint ; { Total read hits }
                             Misses : longint ; { Read misses }
                             { Note: Hits + Misses = total reads }
                             Writes : longint ; { Total writes }
                             Turns : longint ; { Number of times blocks were thrown away to make room for more }
                             Replacements : longint ; { Number of write hits }
                             { Note: Write - Replacements = non-replacement writes}
                         end ;

type TSCache = object( TObject )
                   public { API... }
                       Statistics : TCache_Statistics ;

                       function Read_Cache( var Data ;
                           Address, Size : longint ) : longint ; virtual ;
                       { Reads up to Size bytes from the cache for the specified
                         address.  Returns the number of bytes actually read (if
                         0 then the data was not cached).  Cached data is copied
                         to Data. }

                       procedure Write_Cache( const Data ;
                           Address, Size : longint ) ; virtual ;
                       { Writes to cache up to Size bytes from Data. }

                       procedure Clear ; virtual ; { Clears cache }

                       procedure Initialize ; virtual ; { Initialize object }

                       procedure Terminate ; virtual ; { Terminate object }

                       procedure Set_Max_Size( Size : longint ) ; virtual ;
                       { Sets maximum cache size in bytes.  If this is more
                         than is currently cached, some cache is thrown away. }

                       function Get_Max_Size : longint ; virtual ;
                       { Gets current maximum cache size in bytes. }

                       function Get_Size : longint ; virtual ;
                       { Returns amount of currently cached data, in bytes. }

                       function Get_Stats( Index : longint ;
                           var Address : longint ) : longint ; virtual ;

                       function Lock( Address : longint ) : boolean ; virtual ;
                       { Locks the data starting at the specified address into
                         memory so that it is not thrown away.  Returns false if
                         the specified data is not in the cache. }

                       function Unlock( Address : longint ) : boolean ; virtual ;
                       { Unlocks the data starting at the specified address in
                         memory.  Returns false if the specified data is not in
                         the cache, or if it is not locked. }
     	      end ; { TSCache }
      PSCache = ^TSCache ;

type TSCache64 = object( TObject )
                     public { API... }
                         Statistics : TCache_Statistics ;

                         function Read_Cache( var Data ;
                             Address, Size : int64 ) : longint ; virtual ;
                         { Reads up to Size bytes from the cache for the specified
                           address.  Returns the number of bytes actually read (if
                           0 then the data was not cached).  Cached data is copied
                           to Data. }

                         procedure Write_Cache( const Data ;
                             Address, Size : int64 ) ; virtual ;
                         { Writes to cache up to Size bytes from Data. }

                         procedure Clear ; virtual ; { Clears cache }

                         procedure Initialize ; virtual ; { Initialize object }

                         procedure Terminate ; virtual ; { Terminate object }

                         procedure Set_Max_Size( Size : longint ) ; virtual ;
                         { Sets maximum cache size in bytes.  If this is more
                           than is currently cached, some cache is thrown away. }

                         function Get_Max_Size : longint ; virtual ;
                         { Gets current maximum cache size in bytes. }

                         function Get_Size : longint ; virtual ;
                         { Returns amount of currently cached data, in bytes. }

                         function Get_Stats( Index : longint ;
                             var Address : int64 ) : longint ; virtual ;

                         function Lock( Address : int64 ) : boolean ; virtual ;
                         { Locks the data starting at the specified address into
                           memory so that it is not thrown away.  Returns false if
                           the specified data is not in the cache. }

                         function Unlock( Address : int64 ) : boolean ; virtual ;
                         { Unlocks the data starting at the specified address in
                           memory.  Returns false if the specified data is not in
                           the cache, or if it is not locked. }
                  end ; { TSCache64 }
      PSCache64 = ^TSCache64 ;

{$IFDEF DELPHI}
type TCache = class( TDelphi_Object )
                  public { API... }
                      function Read_Cache( var Data ;
                          Address, Size : longint ) : longint ;
                          virtual ; abstract ;
                      { Reads up to Size bytes from the cache for the specified
                        address.  Returns the number of bytes actually read (if
                        0 then the data was not cached).  Cached data is copied
                        to Data. }

                      procedure Write_Cache( var Data ;
                          Address, Size : longint ) ; virtual ; abstract ;
                      { Writes to cache up to Size bytes from Data. }

                      procedure Clear ; virtual ; abstract ; { Clears cache }

                      procedure Initialize ; virtual ; abstract ; { Initialize object }

                      procedure Terminate ; virtual ; abstract ; { Terminate object }

                      procedure Set_Max_Size( Size : longint ) ;
                          virtual ; abstract ;
                      { Sets maximum cache size in bytes.  If this is more
                        than is currently cached, some cache is thrown away. }

                      function Get_Max_Size : longint ; virtual ; abstract ;
                      { Gets current maximum cache size in bytes. }

                      function Get_Size : longint ; virtual ; abstract ;
                      { Returns amount of currently cached data, in bytes. }

                      function Lock( Address : longint ) : boolean ;
                          virtual ; abstract ;
                      { Locks the data starting at the specified address into
                        memory so that it is not thrown away.  Returns false if
                        the specified data is not in the cache. }

                      function Unlock( Address : longint ) : boolean ;
                          virtual ; abstract ;
                      { Unlocks the data starting at the specified address in
                        memory.  Returns false if the specified data is not in
                        the cache, or if it is not locked. }

                      function Get_Hits : longint ; virtual ; abstract ; { Total read hits }
                      function Get_Misses : longint ; virtual ; abstract ; { Read misses }
                      function Get_Writes : longint ; virtual ; abstract ; { Total writes }
                      function Get_Turns : longint ; virtual ; abstract ; 
                      { Number of times blocks were thrown away to make room for more }
                      function Get_Replacements : longint ; virtual ; abstract ; { Number of write hits }
                      procedure Set_Hits( Value : longint ) ;
                          virtual ; abstract ; { Total read hits }
                      procedure Set_Misses( Value : longint ) ;
                          virtual ; abstract ; { Read misses }
                      procedure Set_Writes( Value : longint ) ;
                          virtual ; abstract ; { Total writes }
                      procedure Set_Turns( Value : longint ) ;
                          virtual ; abstract ; { Number of times blocks were thrown away to make room for more }
                      procedure Set_Replacements( Value : longint ) ;
                          virtual ; abstract ; { Number of write hits }
     	     end ; { TCache }

type TCOM_Cache = class( TCommon_COM_Interface )
                      public { API... }
                          function Read_Cache( var Data ;
                              Address, Size : longint ) : longint ;
                              virtual ; stdcall ; abstract ;
                          { Reads up to Size bytes from the cache for the specified
                            address.  Returns the number of bytes actually read (if
                            0 then the data was not cached).  Cached data is copied
                            to Data. }

                          procedure Write_Cache( var Data ;
                              Address, Size : longint ) ;
                              virtual ; stdcall ; abstract ;
                          { Writes to cache up to Size bytes from Data. }

                          procedure Clear ; virtual ; stdcall ; abstract ; { Clears cache }

                          procedure Set_Max_Size( Size : longint ) ;
                              virtual ; stdcall ; abstract ;
                          { Sets maximum cache size in bytes.  If this is more
                            than is currently cached, some cache is thrown away. }

                          function Get_Max_Size : longint ;
                              virtual ; stdcall ; abstract ;
                          { Gets current maximum cache size in bytes. }

                          function Get_Size : longint ;
                              virtual ; stdcall ; abstract ;
                          { Returns amount of currently cached data, in bytes. }

                          function Lock( Address : longint ) : boolean ;
                              virtual ; stdcall ; abstract ;
                          { Locks the data starting at the specified address into
                            memory so that it is not thrown away.  Returns false if
                            the specified data is not in the cache. }

                          function Unlock( Address : longint ) : boolean ;
                              virtual ; stdcall ; abstract ;
                          { Unlocks the data starting at the specified address in
                            memory.  Returns false if the specified data is not in
                            the cache, or if it is not locked. }

                          function Get_Hits : longint ;
                              virtual ; stdcall ; abstract ; { Total read hits }
                          function Get_Misses : longint ;
                              virtual ; stdcall ; abstract ; { Read misses }
                          function Get_Writes : longint ;
                              virtual ; stdcall ; abstract ; { Total writes }
                          function Get_Turns : longint ;
                              virtual ; stdcall ; abstract ; { Number of times blocks were thrown away to make room for more }
                          function Get_Replacements : longint ;
                              virtual ; stdcall ; abstract ; { Number of write hits }
                          procedure Set_Hits( Value : longint ) ;
                              virtual ; stdcall ; abstract ; { Total read hits }
                          procedure Set_Misses( Value : longint ) ;
                              virtual ; stdcall ; abstract ; { Read misses }
                          procedure Set_Writes( Value : longint ) ;
                              virtual ; stdcall ; abstract ; { Total writes }
                          procedure Set_Turns( Value : longint ) ;
                              virtual ; stdcall ; abstract ; { Number of times blocks were thrown away to make room for more }
                          procedure Set_Replacements( Value : longint ) ;
                              virtual ; stdcall ; abstract ; { Number of write hits }
                 end ; { TCOM_Cache }
{$ENDIF}

implementation

{ TSCache methods... }

{ API... }

function TSCache.Read_Cache( var Data ; Address, Size : longint ) : longint ;

begin
    Read_Cache := 0 ;
    Abstract_Error ;
end ;


procedure TSCache.Write_Cache( const Data ; Address, Size : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSCache.Clear ;

begin
    Abstract_Error ;
end ;


procedure TSCache.Initialize ;

begin
    Abstract_Error ;
end ;


procedure TSCache.Terminate ;

begin
    Abstract_Error ;
end ;


procedure TSCache.Set_Max_Size( Size : longint ) ;

begin
    Abstract_Error ;
end ;


function TSCache.Get_Max_Size : longint ;

begin
    Get_Max_Size := 0 ;
    Abstract_Error ;
end ;


function TSCache.Get_Size : longint ;

begin
    Get_Size := 0 ;
    Abstract_Error ;
end ;


function TSCache.Get_Stats( Index : longint ;
    var Address : longint ) : longint ;

begin
    Get_Stats := 0 ;
    Abstract_Error ;
end ;


function TSCache.Lock( Address : longint ) : boolean ;

begin
    Lock := False ;
    Abstract_Error ;
end ;


function TSCache.Unlock( Address : longint ) : boolean ;

begin
    Unlock := False ;
    Abstract_Error ;
end ;



{ TSCache54 methods... }

{ API... }

function TSCache64.Read_Cache( var Data ; Address, Size : int64 ) : longint ;

begin
    Read_Cache := 0 ;
    Abstract_Error ;
end ;


procedure TSCache64.Write_Cache( const Data ; Address, Size : int64 ) ;

begin
    Abstract_Error ;
end ;


procedure TSCache64.Clear ;

begin
    Abstract_Error ;
end ;


procedure TSCache64.Initialize ;

begin
    Abstract_Error ;
end ;


procedure TSCache64.Terminate ;

begin
    Abstract_Error ;
end ;


procedure TSCache64.Set_Max_Size( Size : longint ) ;

begin
    Abstract_Error ;
end ;


function TSCache64.Get_Max_Size : longint ;

begin
    Get_Max_Size := 0 ;
    Abstract_Error ;
end ;


function TSCache64.Get_Size : longint ;

begin
    Get_Size := 0 ;
    Abstract_Error ;
end ;


function TSCache64.Get_Stats( Index : longint ;
    var Address : int64 ) : longint ;

begin
    Get_Stats := 0 ;
    Abstract_Error ;
end ;


function TSCache64.Lock( Address : int64 ) : boolean ;

begin
    Lock := False ;
    Abstract_Error ;
end ;


function TSCache64.Unlock( Address : int64 ) : boolean ;

begin
    Unlock := False ;
    Abstract_Error ;
end ;



end.
