/*
              Program Name : GETCHAR
              Package Name : LB
              Purpose      : Character I/O routines
              Institution  : Conroy & Conroy Co.
              Date Written : 30-Nov-88
              Written By   : Alan Conroy
              Version      : 3.0

	      Copyright (C) 1988 by Alan Conroy.  Released to the public domain.

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          Character I/O routines.
*/

#include <dos.h>
#include <string.h>

#include "getchars.h"

/*
     This function returns a single character typed at the keyboard.  If
   Option is non-zero, control returns immediately whether or not a
   character was waiting (in which case the returned character is zero,
   if no input was waiting).
*/

char Get_Char( int Option )
{

union REGS Reg ;

    if( Option == 0 ) /* Wait for input */
    {
	Reg.w.ax = 0x0700 ;
	Reg.w.dx = 0x00FF ;
	intdos( &Reg, &Reg ) ; /* Get character */
    }
    else
    {
	Reg.w.ax = 0x0600 ;
	Reg.w.dx = 0x00FF ;
	intdos( &Reg, &Reg ) ; /* Get character */
	if( ( Reg.w.flags & 64 ) == 64 ) /* No input */
	{
	    return 0 ;
	} ;
    } ;
    return Reg.h.al ; /* Return it */
} ;



/*
	  Clear DOS type-ahead buffer.
*/

void Clear_Type_Ahead()
{

union REGS Reg ;

Loop:
    Reg.w.ax = 0x0600 ;
    Reg.w.dx = 0x00FF ;
    intdos( &Reg, &Reg ) ; /* Get next character */
    if( ( Reg.w.flags & 64 ) == 0 )
    {
	goto Loop ; /* Input present, go back for more */
    } ;
} ;


/*
          This function handles special keyboard input.
*/


char *InKey()
{

static char Command[ 6 ] ; /* Input or command line */
int Dummy ; /* Dummy loop variable */

    Command[ 0 ] = Get_Char( 0 ) ;
    if( ( Command[ 0 ] != 0 ) && ( Command[ 0 ] != 27 ) ) /* Normal character */
    {
	Command[ 1 ] = 0 ;
	return &Command ;
    } ;

    if( Command[ 0 ] == 0 ) /* IBM character */
    {
        Command[ 1 ] = Get_Char( 1 ) ;
        switch( Command[ 1 ] )
	{
	    case 0 :
			strcpy( Command, "AQ" ) ;
			break ;
	    case 30 :
			strcpy( Command, "AA" ) ;
			break ;
	    case 31 :
			strcpy( Command, "AS" ) ;
			break ;
	    case 127 :
			strcpy( Command, "A8" ) ;
			break ;
	    case 128 :
			strcpy( Command, "A9" ) ;
			break ;
	    case 129 :
			strcpy( Command, "A0" ) ;
			break ;
	    case 130 :
			strcpy( Command, "A-" ) ;
			break ;
	    case 131 :
			strcpy( Command, "A=" ) ;
			break ;
	    case 132 :
			strcpy( Command, "CKP9" ) ;
			break ;
            case 'H' :
			strcpy( Command, "UP" ) ;
			break ;
            case 'P' :
			strcpy( Command, "DOWN" ) ;
			break ;
            case 'M' :
			strcpy( Command, "RIGHT" ) ;
			break ;
            case 'K' :
			strcpy( Command, "LEFT" ) ;
			break ;
            case ';' :
			strcpy( Command, "F1" ) ;
			break ;
            case '<' :
			strcpy( Command, "F2" ) ;
			break ;
            case '=' :
			strcpy( Command, "F3" ) ;
			break ;
            case '>' :
			strcpy( Command, "F4" ) ;
			break ;
            case '?' :
			strcpy( Command, "F5" ) ;
			break ;
            case '@' :
			strcpy( Command, "F6" ) ;
			break ;
            case 'A' :
			strcpy( Command, "F7" ) ;
			break ;
            case 'B' :
			strcpy( Command, "F8" ) ;
			break ;
            case 'C' :
			strcpy( Command, "F9" ) ;
			break ;
            case 'D' :
			strcpy( Command, "F10" ) ;
			break ;
            case 'T' :
			strcpy( Command, "F11" ) ; /* Shift F1 is F11, etc */
			break ;
            case 'U' :
			strcpy( Command, "F12" ) ;
			break ;
            case 'V' :
			strcpy( Command, "F13" ) ;
			break ;
            case 'W' :
			strcpy( Command, "F14" ) ;
			break ;
            case 'X' :
			strcpy( Command, "F15" ) ;
			break ;
            case 'Y' :
			strcpy( Command, "F16" ) ;
			break ;
            case 'Z' :
			strcpy( Command, "F17" ) ;
			break ;
            case '[' :
			strcpy( Command, "F18" ) ;
			break ;
            case '\\' :
			strcpy( Command, "F19" ) ;
			break ;
            case ']' :
			strcpy( Command, "F20" ) ;
			break ;
            case 'h' :
			strcpy( Command, "E1" ) ; /* Alt F1 is E1, etc */
			break ;
            case 'i' :
			strcpy( Command, "E2" ) ;
			break ;
            case 'j' :
			strcpy( Command, "E3" ) ;
			break ;
            case 'k' :
			strcpy( Command, "E4" ) ;
			break ;
            case 'l' :
			strcpy( Command, "E5" ) ;
			break ;
            case 'm' :
			strcpy( Command, "E6" ) ;
			break ;
            case 'n' :
			strcpy( Command, "E7" ) ;
			break ;
            case 'o' :
			strcpy( Command, "E8" ) ;
			break ;
            case 'p' :
			strcpy( Command, "E9" ) ;
			break ;
            case 'q' :
			strcpy( Command, "E10" ) ;
			break ;
            case '^' :
			strcpy( Command, "E11" ) ; /* Control F1 is E11, etc */
			break ;
            case '_' :
			strcpy( Command, "E12" ) ;
			break ;
            case '`' :
			strcpy( Command, "E13" ) ;
			break ;
            case 'a' :
			strcpy( Command, "E14" ) ;
			break ;
            case 'b' :
			strcpy( Command, "E15" ) ;
			break ;
            case 'c' :
			strcpy( Command, "E16" ) ;
			break ;
            case 'd' :
			strcpy( Command, "E17" ) ;
			break ;
            case 'e' :
			strcpy( Command, "E18" ) ;
			break ;
            case 'f' :
			strcpy( Command, "E19" ) ;
			break ;
            case 'g' :
			strcpy( Command, "E20" ) ;
			break ;
            case '' :
			strcpy( Command, "BACK" ) ; /* Back tab */
			break ;
            case 'R' :
			strcpy( Command, "KP0" ) ;
			break ;
            case 'I' :
			strcpy( Command, "KP9" ) ;
			break ;
            case 'G' :
			strcpy( Command, "KP7" ) ;
			break ;
            case 's' :
			strcpy( Command, "CKP4" ) ;
			break ;
            case 't' :
			strcpy( Command, "CKP6" ) ;
			break ;
            case 'O' :
			strcpy( Command, "KP1" ) ;
			break ;
            case 'Q' :
			strcpy( Command, "KP3" ) ;
			break ;
            case 'w' :
			strcpy( Command, "CKP7" ) ;
			break ;
            case 'u' :
			strcpy( Command, "CKP1" ) ;
			break ;
            case 'v' :
			strcpy( Command, "CKP3" ) ;
			break ;
            case 'x' :
			strcpy( Command, "A1" ) ;
			break ;
            case 'y' :
			strcpy( Command, "A2" ) ;
			break ;
            case 'z' :
			strcpy( Command, "A3" ) ;
			break ;
            case '{' :
			strcpy( Command, "A4" ) ;
			break ;
            case '|' :
			strcpy( Command, "A5" ) ;
			break ;
            case '}' :
			strcpy( Command, "A6" ) ;
			break ;
            case '~' :
			strcpy( Command, "A7" ) ;
			break ;
            case '0' :
			strcpy( Command, "AB" ) ;
			break ;
            case '.' :
			strcpy( Command, "AC" ) ;
			break ;
            case ' ' :
			strcpy( Command, "AD" ) ;
			break ;
            case '' :
			strcpy( Command, "AE" ) ;
			break ;
            case '!' :
			strcpy( Command, "AF" ) ;
			break ;
            case '"' :
			strcpy( Command, "AG" ) ;
			break ;
            case '#' :
			strcpy( Command, "AH" ) ;
			break ;
            case '' :
			strcpy( Command, "AI" ) ;
			break ;
            case '$' :
			strcpy( Command, "AJ" ) ;
			break ;
            case '%' :
			strcpy( Command, "AK" ) ;
			break ;
            case '&' :
			strcpy( Command, "AL" ) ;
			break ;
            case '2' :
			strcpy( Command, "AM" ) ;
			break ;
            case '1' :
			strcpy( Command, "AN" ) ;
			break ;
            case '' :
			strcpy( Command, "AO" ) ;
			break ;
            case '' :
			strcpy( Command, "AP" ) ;
			break ;
            case '' :
			strcpy( Command, "AQ" ) ;
			break ;
            case '' :
			strcpy( Command, "AR" ) ;
			break ;
            case '' :
			strcpy( Command, "AT" ) ;
			break ;
            case '' :
			strcpy( Command, "AU" ) ;
			break ;
            case '/' :
			strcpy( Command, "AV" ) ;
			break ;
            case '' :
			strcpy( Command, "AW" ) ;
			break ;
            case '-' :
			strcpy( Command, "AX" ) ;
			break ;
            case '' :
			strcpy( Command, "AY" ) ;
			break ;
            case ',' :
			strcpy( Command, "AZ" ) ;
			break ;
        } ;
	return &Command ;
    } ;

    /* Escape character */
    Command[ 0 ] = Get_Char( 1 ) ;
    if( Command[ 0 ] == 0 ) /* Just an escape */
    {
	Command[ 0 ] = 27 ;
	Command[ 1 ] = 0 ;
	return &Command ;
    } ;

    /* Special character */
    Command[ 1 ] = Get_Char( 0 ) ; /* Get third character */
    if( Command[ 1 ] == 91 ) /* "[" */
    {
        switch( Command[ 2 ] )
	{
            case 'A' :
			strcpy( Command, "UP" ) ;
			break ;
            case 'B' :
			strcpy( Command, "DOWN" ) ;
			break ;
            case 'C' :
			strcpy( Command, "RIGHT" ) ;
			break ;
            case 'D' :
			strcpy( Command, "LEFT" ) ;
			break ;
            case '1' :
	    case '2' :
	    case '3' : 
			Command[ 2 ] = Get_Char( 1 ) ;
			Command[ 3 ] = Get_Char( 1 ) ;
                        switch( Command[ 2 ] )
			{
			    case '~' :
					Command[ 0 ] = 'E' ;
					Command[ 2 ] = 0 ;
					break ;
                            case '0' :
					if( Command[ 1 ] == '1' )
					{
					    strcpy( Command, "COMPOSE" ) ;
					}
                                        else
					{
					    strcpy( Command, "F9" ) ;
					} ;
					break ;
                            case '2' :
					strcpy( Command, "F18" ) ;
					break ;
                            case '3' :
					strcpy( Command, "F19" ) ;
					break ;
                            case '6' :
					strcpy( Command, "F14" ) ;
					break ;
                            case '7' :
					strcpy( Command, "F6" ) ;
					break ;
                            case '1' :
					if( Command[ 1 ] == '3' )
					{
					    strcpy( Command, "F17" ) ;
					}
                                        else
					{
					    strcpy( Command, "F10" ) ;
					} ;
					break ;
                            case '4' :
					if( Command[ 1 ] == '3' )
					{
					    strcpy( Command, "F20" ) ;
					}
                                        else
					{
					    strcpy( Command, "F4" ) ;
					} ;
					break ;
                            case '8' :
					if( Command[ 1 ] == '1' )
					{
					    strcpy( Command, "F7" ) ;
					}
                                        else
					{
					    strcpy( Command, "F15" ) ;
					} ;
					break ;
                            case '9' :
					if( Command[ 1 ] == '1' )
					{
					    strcpy( Command, "F8" ) ;
					}
                                        else
					{
					    strcpy( Command, "F16" ) ;
					} ;
			} ;
			break ;
            case '4' :
			strcpy( Command, "E4" ) ;
			break ;
            case '5' :
			strcpy( Command, "E5" ) ;
			break ;
            case '6' :
			strcpy( Command, "E6" ) ;
        } ;
        return &Command ;
    } ;
    if( Command[ 0 ] == 'O' )
    {
        switch( Command[ 1 ] )
	{
            case 'P' :
			strcpy( Command, "PF1" ) ;
			break ;
            case 'Q' :
			strcpy( Command, "PF2" ) ;
			break ;
            case 'R' :
			strcpy( Command, "PF3" ) ;
			break ;
            case 'S' :
			strcpy( Command, "PF4" ) ;
			break ;
            case 'M' :
			strcpy( Command, "ENTER" ) ;
			break ;
            case 'p' :
			strcpy( Command, "KP0" ) ;
			break ;
            case 'q' :
			strcpy( Command, "KP1" ) ;
			break ;
            case 'r' :
			strcpy( Command, "KP2" ) ;
			break ;
            case 's' :
			strcpy( Command, "KP3" ) ;
			break ;
            case 't' :
			strcpy( Command, "KP4" ) ;
			break ;
            case 'u' :
			strcpy( Command, "KP5" ) ;
			break ;
            case 'v' :
			strcpy( Command, "KP6" ) ;
			break ;
            case 'w' :
			strcpy( Command, "KP7" ) ;
			break ;
            case 'x' :
			strcpy( Command, "KP8" ) ;
			break ;
            case 'y' :
			strcpy( Command, "KP9" ) ;
			break ;
            case 'm' :
			strcpy( Command, "KP-" ) ;
			break ;
            case 'l' :
			strcpy( Command, "KP," ) ;
			break ;
            case 'n' :
			strcpy( Command, "KP." ) ;
        } ;
	return &Command ;
    } ;
} ;
