/*
	  TLogger is a generic logging interface.  Note that an integer
	Level is passed to most methods.  The only important thing is that
	the caller and responder have agreed on its meaning.  Generally, Level
	is used to indicate that type of message as follows:

		Value	Meaning
		-----	-------
		0	Undefined
		1	Fatal error
		2	Serious error
		3	Warning
		4	Hint
		5	Notification
*/

class TLogger 
{
	public: // API...
		/* This procedure is called to log informational messages
		   to the code supplying the logger.  Mess is the message
		   text, Level indicates the type of message, and Abort
		   indicates whether or not to abort the process which
		   created the message.  If Abort is True upon return the
		   process should be terminated gracefully as quickly as
		   possible. */
		virtual void __stdcall Log_Message( char* Mess, int Level,
		  bool &Abort ) = 0 ;

		/* Same as Log_Message, but the message will be followed
		   by a Unest_Message call when the condition causing the
		   message goes out of scope. */
		virtual void __stdcall Log_Nested_Message( char * Mess, int Level,
		  bool &Abort ) = 0 ;

		/* This indicates the end of the condition under which the
		   previous message with the specified level occurred.
		   If Abort is True upon return, the process should be
		   terminated gracefully as quickly as possible. */
		virtual void __stdcall Unnest_Message( int Level, bool &Abort ) = 0 ;

		/* This function is called during a long process to allow
		   for the user to abort it or for checking Windows
		   messages, etc.  If the function returns True, the
		   process should be terminated gracefully as quickly as
		   possible. */
		virtual bool __stdcall Check_Terminate() = 0 ;
} ; // class TLogger
