.RADIX  10
.8086

;              Program Name : OUTPUT
;              Package Name : LB
;              Purpose      : Output routines
;              Institution  : Conroy & Conroy Company.
;              Date Written : 13-Nov-86
;              Written By   : Alan Conroy
;              Version      : 9.2-10
;
;	 Copyright (C) 1986 by Alan Conroy.  Released to the public domain.
;
;        *********************************************************
;        *                                                       *
;        *        M O D I F I C A T I O N   H I S T O R Y        *
;        *                                                       *
;        *     DATE      BY             REASON                   *
;        *                                                       *
;        *********************************************************
;
;        *********************************************************
;        *                                                       *
;        *           P R O G R A M   P U R P O S E               *
;        *                                                       *
;        *********************************************************
;
;          These routines do OUTPUT for INIT.SYS.
;
; OUTR
;
;   This routine outputs a character to the screen.  No data is
; returned.  Data passed is: DL = Byte to output.  CCPOS is modified to
; reflect current output position on terminal.
;

OUTR:   PUSH  AX ; Save registers
        PUSH  BX
        PUSH  DX
        PUSHF

        PUSH  DX     ; Save character
        MOV   AH, 02H
        INT   21H    ; Output character
        POP   DX     ; Restore character
        MOV   BX, OFFSET CCPOS
        CMP   DL, 10 ; Linefeed?
        JE    Outf   ; Yes - so skip incrementing CCPOS
        CMP   DL, 8  ; Backspace?
        JNE   Outg   ; Nope
        DEC   BYTE PTR [BX] ; Decrement count
        JMP   Oute
Outg:   INC   BYTE PTR [BX] ; Increment count
Outf:   CMP   DL, 13 ; CR?
        JNE   Oute ; Nope
        MOV   BYTE PTR [BX], 0 ; Reset position
Oute:   POPF  ; Restore registers
        POP   DX
        POP   BX
        POP   AX
        RET ; All done

;
; OUTPUT
;
;   This routine outputs a string to the screen.  No data is
; returned.  Data passed is: CX = Length of string, DS:BX = Location of string
;

OUTPUT: PUSH  BX ; Save registers
        PUSH  CX
        PUSH  DX
        PUSHF

Out1:   MOV   DL, [BX]
        CALL  OUTR ; Output character
        INC   BX   ; Point to next character
Out2:   LOOP  Out1 ; Go back for next byte if any left
        POPF  ; Restore registers
        POP   DX
        POP   CX
        POP   BX
        RET ; All done

;
; STRING
;
;   This routine outputs CX number of characters as defined by AL.  If CX
; is negative it is made positive first.
;

STRING: PUSH CX     ; Save registers
        PUSH DX
        PUSHF
        CMP  CX, 0  ; None?
        JE   Str0   ; Yes - so end
        MOV  DL, AL ; Set up for output
        JG   Str1   ; Positive?
        NEG  CX     ; Make positive
Str1:   CALL OUTR   ; Output
        LOOPNE Str1 ; loop
Str0:   POPF        ; Restore registers
        POP  DX
        POP  CX
        RET

;
; TAB
;
;   This routine "tabs" to a position on the output line.  Data passed: CL =
; position to tab to.  Data returned=None.  CCPOS is maintained.
;

TAB:    PUSH AX ; Save registers
        PUSH BX
        PUSH CX
        PUSH DX
        PUSHF
        MOV  BX, OFFSET CCPOS
        MOV  AL, [BX] ; Get current position
Tab0:   CMP  AL, CL ; Already at or past desired position?
        JGE  Tab1   ; Yup so do nothing
        MOV  DL, 32
        CALL OUTR   ; Output a space
        INC  AL     ; Increment current position
        JMP  Tab0   ; Loop
Tab1:   POPF        ; Restore registers
        POP  DX
        POP  CX
        POP  BX
        POP  AX
        RET         ; All done

	END
