{$N+}
{
        Program Name : Performa
        Package Name : LB
        Purpose      : Performance monitor
        Institution  : Conroy & Conroy Co.
        Date Written : 15-Mar-2005
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (c) 2005 by Alan Conroy.  Released to the Public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit contains performance monitoring support.
}

{$I EDEFINES.INC}

unit Performa ;

interface

{$IFDEF DELPHI}
procedure Enter_Routine( Index : longint ) ;
procedure Exit_Routine( Index : longint ) ;
procedure Clear_Counts ;
function Max_Time_Index( Include_Called : boolean ) : longint ;
function Routine_Time( Routine : integer ; Include_Called : boolean ) : int64 ;
procedure Save_Statistics( const Name : string ) ;
{$ENDIF}

implementation

{$IFDEF DELPHI}
uses { Borland... }
     Windows, { GetTickCount }

     { Layer 0... }
     CommonUt, // TInteger_List
     Standard ; // TInteger64_List

var Routine_List : TInteger_List ; { Corresponds to Entry_Times.  Routine # }
    Entry_Times : TInteger64_List ; { Time index of entry into routine }

    Entry_Counts : Tinteger_List ; { Number of calls to routines }
    Total_Times : TInteger64_List ; { Total time in routine, including called code }
    Routine_Times : TInteger64_List ; { Total time in routine, not including called code }

// Internal utility routines...

procedure Extend_Lists( Index : longint ) ;

begin
    while( Total_Times.Count <= Index ) do
    begin
        Total_Times.Add( 0 ) ;
        Routine_Times.Add( 0 ) ;
        Entry_Counts.Add( 0 ) ;
    end ;
end ;


// API...

procedure Enter_Routine( Index : longint ) ;

begin
    Routine_List.Add( Index ) ; { Add routine to call stack }
    Entry_Times.Add( GetTickCount ) ; { Remember routine entry timestamp }

    { Increment call count... }
    Extend_Lists( Index ) ;
    Entry_Counts.Increment( Index ) ;
end ;


procedure Exit_Routine( Index : longint ) ;

var T : int64 ;

begin
    if( Index <> Routine_List[ Routine_List.Count - 1 ] ) then
    begin
        writeln( 'Error' ) ;
        halt ;
    end ;
    T := GetTickCount - Entry_Times[ Routine_List.Count - 1 ] ; { Amount of time spent in this routine }

    Extend_Lists( Index ) ;
    Routine_Times[ Index ] := Routine_Times[ Index ] + T ;
    Total_Times[ Index ] := Total_Times[ Index ] + T ;

    { Add this routine's time to the calling routines }
    Index := Routine_List.Count - 2 ;
    while( Index >= 0 ) do
    begin
        try
            Routine_Times[ Routine_List[ Index ] ] := Routine_Times[ Routine_List[ Index ] ] - T ;
        except
        end ;
        dec( Index ) ;
    end ;

    // Remove this routine from the call stack...
    Routine_List.Delete( Entry_Times.Count - 1 ) ;
    Entry_Times.Delete( Entry_Times.Count - 1 ) ;
end ;


procedure Clear_Counts ;

begin
    Entry_Counts.Clear ;
    Routine_Times.Clear ;
    Total_Times.Clear ;
end ;


function Max_Time_Index( Include_Called : boolean ) : longint ;

begin
    if( Include_Called ) then
    begin
        Result := Total_Times.Index_With_Max ;
    end else
    begin
        Result := Routine_Times.Index_With_Max ;
    end ;
end ;


function Routine_Time( Routine : integer ; Include_Called : boolean ) : int64 ;

begin
    if( Include_Called ) then
    begin
        Result := Total_Times[ Routine ] ;
    end else
    begin
        Result := Routine_Times[ Routine ] ;
    end ;
end ;


procedure Save_Statistics( const Name : string ) ;

var F : text ;
    I : int64 ;
    Loop : integer ;

begin
    assign( F, Name ) ;
    {$I-}
    append( F ) ;
    {$I+}
    if( IOResult <> 0 ) then
    begin
        rewrite( F ) ;
    end ;
    writeln( F, 'Routine    Count    Time    Total Time' ) ;
    writeln( F, '-------    -----    ----    ----------' ) ;
    for Loop := 0 to Entry_Counts.Count - 1 do
    begin
        if( Entry_Counts[ Loop ] > 0 ) then
        begin
            I := Routine_Times[ Loop ] ;
            if( I < 0 ) then
            begin
                I := 0 ;
            end ;
            writeln( F, Loop:7, '    ', Entry_Counts[ Loop ]:5, '    ', I:4,
                '    ', Total_Times[ Loop ]:10 ) ;
        end ;
    end ;
    close( F ) ;
end ;


initialization
    Routine_List := TInteger_List.Create ;
    Entry_Times := TInteger64_List.Create ;
    Routine_Times := TInteger64_List.Create ;
    Total_Times := TInteger64_List.Create ;
    Entry_Counts := Tinteger_List.Create ;

finalization
    Routine_List.Free ;
    Entry_Times.Free ;
    Routine_Times.Free ;
    Total_Times.Free ;
    Entry_Counts.Free ;
    Routine_List := nil ;
    Entry_Times := nil ;
    Routine_Times := nil ;
    Total_Times := nil ;
    Entry_Counts := nil ;
{$ENDIF}
end.
