{$D-}
{
              Program Name : Radix50s
              Package Name : LB
              Purpose      : Radix-50 support
              Institution  : Conroy & Conroy Company
              Original Vers: 22-Aug-86
              Date Written : 25-Jan-1990
              Written By   : Alan Conroy
              Version      : 1.0
              
              Copyright (C) 1986-2005 by Alan Conroy.  Released to the public domain.

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

}

unit Radix50s ;

interface

{$I EDEFINES.INC}

{$IFDEF DELPHI}
function Rad( Source : int64 ) : String ; { Convert Radix-50 integer to string }
function Rad50( Source : String ) : int64 ;
{$ELSE}
function Rad( Source : Integer ) : String ;
function Rad50( Source : String ) : Integer ;
{$ENDIF}

{ Convert ASCII string to a string image of Radix-50 codes. }
function ASCII_To_RAD( Source : string ) : string ;

{ Check validity of radix-50 string }
function Valid_Radix50( const S : string ) : boolean ;

implementation

type Pa = array[ 0..2 ] of Integer ;

const Radix = ' ABCDEFGHIJKLMNOPQRSTUVWXYZ$. 0123456789' ;
      Power : Pa = ( 1, 40, 1600 ) ;

{$IFDEF DELPHI}
function Rad( Source : int64 ) : string ;
{$ELSE}
function Rad( Source : Integer ) : string ;
{$ENDIF}

var X, Y  : Real ;
    Xx : string ;
    Xxx, Yyy : Integer ;

begin
    Xx := '' ;
    X := Source and $FFFF ;
    while( X < 0 ) do
    begin
        X := 65536.0 + X ;
    end ;
    for Yyy := 2 downto 0 do
    begin
        Y := Power[ Yyy ] ;
        Xxx := Trunc( Int( X / Y ) ) ;
        Xx := Xx + Copy( Radix, Xxx + 1, 1 ) ;
        X := X - Xxx * Y
    end ;
{$IFDEF DELPHI}
    X := Source shr 16 ;
    while( X < 0 ) do
    begin
        X := 65536.0 + X ;
    end ;
    for Yyy := 2 downto 0 do
    begin
        Y := Power[ Yyy ] ;
        Xxx := Trunc( Int( X / Y ) ) ;
        Xx := Xx + Copy( Radix, Xxx + 1, 1 ) ;
        X := X - Xxx * Y
    end ;
{$ENDIF}
    Rad := Xx ;
end ;


{$IFDEF DELPHI}
function Rad50( Source : string ) : Int64 ;
{$ELSE}
function Rad50( Source : string ) : Integer ;
{$ENDIF}
{ This module converts a three (3) character string into a two character
  Radix-50 value that may be used with the RAD function. }

type Pa = array[ 0..2 ] of Integer ;

var X  : Real ;
    Xx : Integer ;

begin
    while( Length( Source ) < 3 ) do
    begin
        Source := ' ' + Source ;
    end ;
    X := 0 ;
    for Xx := 0 to 2 do
    begin
        X := X + ( Pos( Source[ 3 - Xx ], Radix ) - 1 ) * Power[ Xx ] ;
    end ;
    if( X > 32767 ) then
    begin
        Xx := Trunc( X - 65536.0 )
    end else
    begin
        Xx := Trunc( X ) ;
    end ;
    Rad50 := Xx ;
end ;


function ASCII_To_RAD( Source : string ) : string ;

var Loop : integer ;
    Res : string ;
    Temp : integer ;
    Work : string[ 2 ] ;

begin
    { Setup... }
    Res := '' ;
    Loop := 1 ;
    Work := '12' ; { Force length to 2 }

    { Convert... }
    while( Loop <= length( Source ) ) do
    begin
        Temp := Rad50( copy( Source, Loop, 3 ) ) ;
        move( Temp, Work[ 1 ], 2 ) ;
        Res := Res + Work ;
        Loop := Loop + 3 ;
    end ;
    ASCII_To_RAD := Res ;
end ;


function Valid_Radix50( const S : string ) : boolean ;

var Loop : integer ;

begin
    Valid_Radix50 := True ;
    for Loop := 1 to length( S ) do
    begin
        if( pos( S[ Loop ], Radix ) = 0 ) then
        begin
            Valid_Radix50 := False ;
            exit ;
        end ;
    end ;
end ;


end.
