/*
	      Program Name : TEU
	      Package Name : LB
	      Purpose      : Text Editing Utility
	      Institution  : Conroy & Conroy
	      Original Vers: 22-Mar-89
	      Date Written : 21-Jul-1990
	      Written By   : Alan Conroy
	      Version      : 2.0

              Copyright (C) 1989-1990 by Alan Conroy.  Released to the public domain.

	      TO THE GLORY OF GOD THROUGH JESUS CHRIST

	*********************************************************
	*                                                       *
	*        M O D I F I C A T I O N   H I S T O R Y        *
	*                                                       *
	*     DATE      BY             REASON                   *
	*                                                       *
	*********************************************************

	*********************************************************
	*                                                       *
	*           P R O G R A M   P U R P O S E               *
	*                                                       *
	*********************************************************

	  Conceptual equivalent to VAX TPU.
*/

#include <errno.h>
#include <dos.h>
#include <process.h>
#include <alloc.h>
#include <string.h>
#include <dir.h>
#include <float.h>
#include <stdlib.h>
#include <stdio.h>

#include "clib.h"
#include "teu.h"
#include "edit.h"
#include "files.h"
#include "getchars.h"
#include "instr.h"
#include "smu.h"
#include "ustrings.h"

static const char Translation[ 128 ][ 6 ] =
{ "<NUL>", "<SOH>", "<STX>", "<ETX>", "", "<ENQ>", "<ACK>", "<BEL>", "<BS>",
  "", "<LF>", "<VT>", "<FF>", "<CR>", "<SO>", "<SI>", "<DLE>", "<DC1>", "<DC2>",
  "<DC3>", "", "", "<SYN>", "<ETB>", "<CAN>", "<EM>", "<SUB>", "<ESC>",
  "<FS>", "<GS>", "<RS>", "<US>", " ", "!", "\"", "#", "$", "%", "&", "\'", "(",
  ")", "*", "+", ",", "-", ".", "/", "0", "1", "2", "3", "4", "5", "6", "7",
  "8", "9", ":", ";", "<", "=", ">", "?", "@", "A", "B", "C", "D", "E", "F",
  "G", "H", "I", "J", "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T", "U",
  "V", "W", "X", "Y", "Z", "[", "\\", "]", "^", "_", "`", "a", "b", "c", "d",
  "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q", "r", "s",
  "t", "u", "v", "w", "x", "y", "z", "{", "|", "}", "~", "<DEL>"
} ;


static const char Version[ 4 ] = "2.0" ;

static int A, B ; /* Loop and dummy variables */
static struct Anchor_Type Anchor ;
static int Any_Changes ; /* True if ANY changes made to file */
static int Auto_Indent ; /* True if we must auto-indent */
static char Blank[ 81 ] ; /* 80 spaces for blanking lines */
static int Ccl ; /* If a command-line entry */
static int Changes ; /* True if changes made to the current line */
static char Character[ 256 ] ;
static char Character_Buffer ; /* Last deleted character */
static char Command[ 256 ], Command0[ 256 ] ; /* Input or command line */
static struct Line_Record *Current_Line ; /* Pointer to current line in buffer */
static short int Cursor_Column ; /* Current column of cursor */
static int Cursor_Row ; /* Current row of cursor */
static int Deleted_Character ; /* True if a deleted character */
static int Deleted_Line ; /* True if a deleted line */
static int Deleted_Word ; /* True if a deleted word */
static int Draw ; /* True if to redraw screen */
static int Dummy ; /* Dummy variable */
static int Error_Flag ; /* TRUE if error currently displayed */
static char Help_File_Name[ 256 ] ;
static struct FILE *Init_File ;
static struct FILE *Input_File ;
static char Input_Name[ 256 ], Journal_Name[ 256 ] ; /* Names of files */
static long int Iterate ; /* Number of iterations, etc */
static struct Line_Record *File_Root, *File_Last ; /* Text pointers */
static struct FILE *Journal ; /* Journal file */
static char Journal_Buffer[ 256 ] ; /* Contents of current record in journal file */
static int Journal_Offset ; /* Current offset in Journal_Buffer */
static long int Journal_Record ; /* Current record in journal file */
static int Journal_Size ;
/* Number of characters between journal file updates */
static long int Last_Line ;
static long int Last_Screen_Line ; /* Previous Screen_Line as of last Redraw */
static char Line_Buffer[ 256 ] ;
static char Logical[ 256 ] ;
static int Logical_Column, Physical_Column ; /* Editing pointers */
static char Logical_Text[ 256 ], Physical_Text[ 256 ] ; /* Current line text */
static int New_File ;
static struct Line_Record *Paste_Root ;
static short int Real_Length ; /* Actual length of display line text */
static int Recover ; /* TRUE if /RECOVER */
static long int Screen_Line ; /* Line in file corresponding to line 1 of screen */
static char Search_Text[ 256 ] ; /* Text to search for */
static int Selection ; /* True if a selectino */
static int Status ; /* True if to display status window */
static char Word_Buffer[ 256 ] ; /* Last deleted word */
static struct Window *Screen_Window, *Error_Window, *Status_Window ;

static void Flag_Changes() /* Flag any changes to line */
{
    Changes = True ;
    Logical_Column = Cursor_Column ;
} ;

static void Update_Journal()
{
    Journal_Buffer[ Journal_Offset ] = 0 ; /* Maintain string length */
    fseek( Journal, Journal_Record * 256, SEEK_SET ) ;
    fwrite( Journal_Buffer, 1, 256, Journal ) ; /* Update journal */
    rewind( Journal ) ;
} ;

static char *In_Key() /* Get next keystroke */
{

short int A, B ;
static char X[ 256 ] ;

    if( Recover ) /* /RECOVER mode */
    {
	B = Journal_Buffer[ 0 ] ; /* Get length of next keystroke */
	for( A = 0; A <= B + 1; A++ )
	{
	    X[ A ] = Journal_Buffer[ A + 1 ] ; /* Get keystroke */
	} ;
	strcpy( Journal_Buffer, &Journal_Buffer[ B + 1 ] ) ;
	if( Journal_Buffer[ 0 ] == 0 ) /* Done with this buffer */
	{
	    Journal_Record ++ ;
	    if( fread( Journal_Buffer, 1, 256, Journal ) < 256 )
	    {
		Recover = False ; /* Read it OK */
	    } ;
	} ;
    }
    else
    {
	strcpy( X, InKey() ) ; /* Get keystroke */
	if( Journal_Size > 0 ) /* If journaling is on */
	{
	    if( Journal_Offset + strlen( X ) > 254 )
	    /* Too big to fit in this string */
	    {
		Update_Journal() ;
		Journal_Offset = -1 ;
		Journal_Record ++ ; /* New record */
	    } ;
	    A = Journal_Offset ;
	    Journal_Offset ++ ;
	    Journal_Buffer[ Journal_Offset ] = strlen( X ) ;
	    /* Length for this string */
	    for( B = 0; B < strlen( X ); B++ )
	    {
		Journal_Offset ++ ;
		Journal_Buffer[ Journal_Offset ] = X[ B ] ;
		/* Copy string to buffer */
	    } ;
	    if( ( Journal_Offset & ( ~Journal_Size ) ) >
		( A & ( ~Journal_Size ) ) && A > -1 )
	    {
		Update_Journal() ;
	    } ;
	    /* More than n characters since last write */
	} ;
    } ;
    return X ; /* Return keystroke */
} ;

static void Do_Read() /* Get next input */
{

int A, B ;

    if( Recover ) /* /RECOVER mode */
    {
	B = Journal_Buffer[ 0 ] ; /* Get length of next keystroke */
	for( A = 0; A <= B + 1; A++ )
	{
	    Kb_Buffer[ A ] = Journal_Buffer[ A + 1 ] ; /* Get input */
	} ;
	strcpy( Journal_Buffer, &Journal_Buffer[ B + 1 ] ) ;
	if( Journal_Buffer[ 0 ] == 0 ) /* Done with this buffer */
	{
	    Journal_Record ++ ;
	    if( fread( Journal_Buffer, 1, 256, Journal ) < 256 )
	    {
		Recover = False ; /* Read it OK */
	    } ;
	} ;
    }
    else
    {
	Window_Read() ; /* Do the read */
	if( Journal_Size == 0 )
	{
	    return ; /* If journaling is off */
	} ;
	if( Journal_Offset + strlen( Kb_Buffer ) > 254 )
	/* Too big to fit in this string */
	{
	    Update_Journal() ;
	    Journal_Offset = -1 ;
	    Journal_Record ++ ; /* New record */
	} ;
	A = Journal_Offset ;
	Journal_Offset ++ ;
	Journal_Buffer[ Journal_Offset ] = strlen( Kb_Buffer ) ;
	/* Length for this string */
	for( B = 0; B < strlen( Kb_Buffer ); B++ )
	{
	    Journal_Offset ++ ;
	    Journal_Buffer[ Journal_Offset ] = Kb_Buffer[ B ] ;
	    /* Copy string to buffer */
	} ;
	if( ( Journal_Offset & ( ~Journal_Size ) ) > ( A & ( ~Journal_Size ) ) &&
		A > -1 )
	/* More than n characters since last write */
	{
	    Update_Journal() ;
	} ;
    } ;
} ;

static void Show_Error( char *X )
{
    Error_Flag = True ;
    Select_Window( Error_Window ) ;
    Window_Output( "\r\n" ) ;
    Window_Output( X ) ;
    Unmap_Window( Status_Window ) ; /* Unmap status window so error can show */
    Update_Window( Error_Window ) ;
    Select_Window( Screen_Window ) ;
} ;

static void End_Edit() /* Terminate TEU section */
{
    if( Get_Model == 2 )
    {
	printf( "0x1B>" ) ; /* Numeric keypad if Rainbow */
    } ;
    End_Smu() ;
} ;

static void Do_Write( char *Command, struct Line_Record *File_Root )
/* Write specified buffer to a file */
{

struct FILE *Input_File ;
struct Line_Record *T_Pointer ;
char Temp[ 256 ] ;

    Input_File = fopen( Command, "wt" ) ; /* Open input file */
    if( Input_File == NULL )
    {
	sprintf( Temp, "\a%s - creating %s", strerror( errno ), Command ) ;
	Show_Error( Temp ) ;
	return ;
    } ;
    T_Pointer = File_Root ;
    while( T_Pointer != NULL ) /* Until end of buffer */
    {
	fprintf( Input_File, "%s\n", T_Pointer->Text ) ; /* Write next line */
	T_Pointer = T_Pointer->Next_Line ; /* Follow pointer to next line */
    } ;
    fclose( Input_File ) ;
} ;


static void Include_File( char *Passed_Command, long int *Last_Line,
    struct Line_Record **File_Root, struct Line_Record **File_Last )
    /* Read a file into memory */
{

char Command[ 256 ] ;
struct FILE *Input_File ;
struct Line_Record *L_Pointer, *T_Pointer ;
char Temp[ 256 ] ;
int E_O_F ;

    if( Passed_Command == NULL )
    {
	Command[ 0 ] = 0 ;
    }
    else
    {
	strcpy( Command, Passed_Command ) ;
    } ;
    *File_Root = NULL ;
    *File_Last = NULL ;
    L_Pointer = NULL ;
    *Last_Line = 0 ;
    Input_File = fopen( Command, "rt" ) ;
    if( Input_File == NULL )
    {
	Select_Window( Error_Window ) ;
	Unmap_Window( Status_Window ) ;
	Window_Output( "\r\n" ) ; /* Clear line */
	sprintf( Temp, "\a%s - writing %s", strerror( errno ), Command ) ;
	Window_Output( Temp ) ;
	Update_Window( Error_Window ) ;
	Select_Window( Screen_Window ) ;
	return ;
    } ;
    E_O_F = False ;
    while( E_O_F == False )
    {
	if( fgets( Command, 255, Input_File ) == NULL )
	{
	    E_O_F = True ;
	    continue ;
	} ;
	Command[ strlen( Command ) - 1 ] = 0 ; /* Strip newline */
	*Last_Line = *Last_Line + 1 ;
	T_Pointer = malloc( strlen( Command ) + 9 ) ;
	if( T_Pointer != NULL )
	{
	    if( *File_Root == NULL )
	    {
		*File_Root = T_Pointer ;
	    } ;
	    T_Pointer->Next_Line = NULL ;
	    T_Pointer->Last_Line = L_Pointer ;
	    if( L_Pointer != NULL )
	    {
		L_Pointer->Next_Line = T_Pointer ;
	    } ;
	    strcpy( T_Pointer->Text, Command ) ;
	    L_Pointer = T_Pointer ;
	}
	else
	{
	    while( *File_Root != NULL ) /* Reclaim memory */
	    {
		T_Pointer = (*File_Root)->Next_Line ;
		free( *File_Root ) ;
		*File_Root = T_Pointer ;
	    } ;
	    Show_Error( "\a?Not enough memory" ) ;
	    *Last_Line = -1 ; /* Signal error */
	    return ;
	} ;
    } ;
    *File_Last = L_Pointer ;
    fclose( Input_File ) ;
} ;


static void Format( char *X )
/* This function translates physical text to logical text */
{

int A ; /* Loop variable */
char C ;
int Dummy ; /* Dummy variable */

    memset( Logical, 32, 80 ) ;
    Logical[ 0 ] = 0 ;
    for( A = 0; A < strlen( X ); A++ )
    /* For each character in physical string */
    {
	if( strlen( Logical ) > 249 )
	{
	    goto Done ;
	} ;
	C = X[ A ] ; /* Get current character */
	if( ( ( C < 31 ) && ( C != 20 ) ) || C == 127 )
	/* Special character */
	{
	    if( C == '\t' ) /* Tab */
	    {
		Dummy = ( strlen( Logical ) + 8 ) & ( ~7 ) ;
		/* New location */
		while( strlen( Logical ) < Dummy )
		/* Space to next tab position */
		{
		    strcat( Logical, " " ) ;
		} ;
	    }
	    else
	    {
		strcat( Logical, Translation[ C ] ) ;
	    } ;
	}
	else
	{
	    Logical[ strlen( Logical ) + 1 ] = 0 ;
	    Logical[ strlen( Logical ) ] = C ; /* Build string */
	} ;
    } ;
Done:
    Real_Length = strlen( Logical ) ; /* Save length */
    if( strlen( Logical ) > 80 )
    {
	Logical[ 80 ] = 0 ;
	Logical[ 79 ] = '' ; /* Diamond */
    } ;
} ;


static void Do_Update() /* Update memory with edited line - if needed */
{

struct Line_Record *T_Pointer ;

    if( Changes ) /* Changes made */
    {
	Any_Changes = True ;
	Changes = False ;
	if( Current_Line == NULL )
	{
	    return ; /* No update if at [EOB] */
	} ;
	T_Pointer = malloc( strlen( Physical_Text ) + 9 ) ;
	/* Allocate new line */
	if( T_Pointer == NULL ) /* Ran out of memory */
	{
	    Show_Error( "\aMaximum memory exceeded" ) ;
	    End_Edit() ;
	    exit( EXIT_FAILURE ) ;
	} ;
	T_Pointer->Last_Line = Current_Line->Last_Line ; /* Link in new line */
	T_Pointer->Next_Line = Current_Line->Next_Line ;
	if( T_Pointer->Last_Line != NULL ) /* Link previous line to new one */
	{
	    T_Pointer->Last_Line->Next_Line = T_Pointer ;
	} ;
	if( T_Pointer->Next_Line != NULL ) /* Link next line to new one */
	{
	    T_Pointer->Next_Line->Last_Line = T_Pointer ;
	} ;
	strcpy( T_Pointer->Text, Physical_Text ) ; /* Store text for line */
	if( File_Last == Current_Line )
	{
	    File_Last = T_Pointer ;
	} ;
	if( File_Root == Current_Line )
	{
	    File_Root = T_Pointer ;
	} ;
	free( Current_Line ) ; /* Delete old line */
	Current_Line = T_Pointer ;
    } ;
} ;


static void Redraw() /* Draw/redraw screen */
{

int A, B ;
long int Check ; /* Value of line to check */
int Save ; /* Saved attributes */
long int Temp ;
struct Line_Record *Text_Pointer ;
long int First_Line, Final_Line ;
short int First_Column, Final_Column ;
int Forced_Redraw ;

    Do_Update() ;
    Set_Window( Screen_Window, 1, 1, 0, -32 ) ;
    /* Normal video and home cursor */
    Text_Pointer = File_Root ; /* Start at file's first line */
    A = 1 ; /* First line */
    Forced_Redraw = True ; /* Default to redraw rather than scroll */
    if( !Selection )
    {
	First_Line = 0 ;
	Final_Line = 0 ;
	if( Last_Screen_Line > 0 ) /* No forced redraw */
	{
	    if( abs( Screen_Line - Last_Screen_Line ) < 24 )
	    /* Difference of one screen or less */
	    {
		if( Last_Screen_Line != Screen_Line )
		/* Movement needed */
		{
		    Forced_Redraw = False ;
		} ;
	    } ;
	} ;
    }
    else /* Selection under way */
    {
	if( Screen_Line + Cursor_Row - 1 < Anchor.Line )
	/* If we're before the selection point */
	{
	    First_Line = Screen_Line + Cursor_Row - 1 ;
	    Final_Line = Anchor.Line ;
	    First_Column = Physical_Column ;
	    Final_Column = Anchor.P_Column ;
	}
	else /* We're after the selection point */
	{
	    Final_Line = Screen_Line + Cursor_Row - 1 ;
	    First_Line = Anchor.Line ;
	    Final_Column = Physical_Column ;
	    First_Column = Anchor.P_Column ;
	} ;
    } ;
    if( Forced_Redraw )
    {
	Check = Screen_Line ;
    }
    else
    {
	Check = Last_Screen_Line ;
    } ;
    Temp = 1 ;
    while( Temp < Check ) /* Find first line of screen */
    {
	Text_Pointer = Text_Pointer->Next_Line ;
	Temp = Temp + 1 ;
    } ;
    if( !Forced_Redraw ) /* Scroll screen instead of redrawing it */
    {
	if( Last_Screen_Line > Screen_Line ) /* Scroll down needed */
	{
	    Text_Pointer = Text_Pointer->Last_Line ;
	    while( Last_Screen_Line > Screen_Line )
	    {
		Insert_Line( Screen_Window, 1 ) ;
		Format( Text_Pointer->Text ) ;
		Window_Output( Logical ) ;
		Window_Output( "\r" ) ; /* Write new line at top */
		Update_Window( Screen_Window ) ;
		Last_Screen_Line = Last_Screen_Line - 1 ;
		if( Last_Screen_Line > Screen_Line )
		{
		    Text_Pointer = Text_Pointer->Last_Line ;
		} ;
	    } ;
	    Set_Window( Screen_Window, 23, 1, 0, 0 ) ;
	    for( Dummy = 1; Dummy <= 22; Dummy++ )
	    {
		Text_Pointer = Text_Pointer->Next_Line ;
	    } ;
	}
	else /* Scroll up needed */
	{
	    for( Dummy = 1; Dummy <= 23; Dummy++ )
	    {
		Text_Pointer = Text_Pointer->Next_Line ;
		/* Go to last line currently on screen */
	    } ;
	    Set_Window( Screen_Window, 23, 1, 0, 0 ) ;
	    /* Go to bottom of screen */
	    while( Last_Screen_Line < Screen_Line )
	    {
		Window_Output( "\n" ) ; /* Scroll up one line */
		if( Text_Pointer != NULL ) /* Not at [EOB] */
		{
		    Format( Text_Pointer->Text ) ;
		    Window_Output( Logical ) ;
		    Window_Output( "\r" ) ;
		    /* Write new line at bottom */
		    Update_Window( Screen_Window ) ;
		} ;
		Last_Screen_Line = Last_Screen_Line + 1 ;
		if( Last_Screen_Line < Screen_Line )
		{
		    Text_Pointer = Text_Pointer->Next_Line ;
		} ;
	    } ; /* while() */
	} ;
    }
    else /* Redraw the screen */
    {
	Last_Screen_Line = Screen_Line ;
	while( A < 23 && Text_Pointer != NULL )
	/* Until end of screen or file */
	{
	    Save = Screen_Window->Attribute ;
	    Set_Window( Screen_Window, 0, 0, 0, -32 ) ; /* Normal video */
	    Window_Output( Blank ) ;
	    Window_Output( "\r" ) ; /* blank line first */
	    Screen_Window->Attribute = Save ; /* Restore current attributes */
	    if( Selection ) /* Selection under way */
	    {
		if( Screen_Line + A - 1 >= First_Line &&
		    Screen_Line + A - 1 <= Final_Line )
		{
		    if( Screen_Line + A - 1 == First_Line &&
			First_Line == Final_Line )
		    {
			strcpy( Character,
			    Left( Text_Pointer->Text, First_Column - 1 ) ) ;
			Format( Character ) ;
			B = strlen( Logical ) ; /* Save column */
			strcpy( Character,
			    Left( Text_Pointer->Text, Final_Column - 1 ) ) ;
			Format( Character ) ;
			Window_Output( Left( Logical, B ) ) ;
			Set_Window( Screen_Window, 0, 0, 0, -1 ) ;
			/* Reverse video */
			Window_Output( &Logical[ B ] ) ;
			B = strlen( Logical ) ;
			Set_Window( Screen_Window, 0, 0, 0, -32 ) ;
			/* Normal video */
			Format( Text_Pointer->Text ) ;
			Window_Output( &Logical[ B ] ) ;
			Window_Output( "\r\n" ) ;
		    }
		    else
		    {
			if( Screen_Line + A - 1 == First_Line )
			{
			    strcpy( Character,
				Left( Text_Pointer->Text, First_Column - 1 ) ) ;
			    Format( Character ) ;
			    B = strlen( Logical ) ; /* Save column */
			    Format( Text_Pointer->Text ) ;
			    Window_Output( Left( Logical, B ) ) ;
			    Set_Window( Screen_Window, 0, 0, 0, -1 ) ;
			    /* Reverse video */
			    Window_Output( &Logical[ B ] ) ;
			    Window_Output( "\r\n" ) ;
			}
			else
			{
			    if( Screen_Line + A - 1 == Final_Line )
			    {
				strcpy( Character,
				    Left( Text_Pointer->Text, Final_Column - 1 ) ) ;
				Format( Character ) ;
				B = strlen( Logical ) ; /* Save column */
				Format( Text_Pointer->Text ) ;
				Window_Output( Left( Logical, B ) ) ;
				Set_Window( Screen_Window, 0, 0, 0, -32 ) ;
				/* Normal video */
				Window_Output( &Logical[ B ] ) ;
				Window_Output( "\r\n" ) ;
			    }
			    else /* Entire line is selected */
			    {
				A = -A ; /* Process normally */
				Set_Window( Screen_Window, 0, 0, 0, -1 ) ;
				/* Reverse video */
			    } ;
			    A = -A ; /* Flag */
			} ;
		    } ;
		} ;
	    } ;
	    if( A < 0 )
	    {
		A = -A ;
	    }
	    else
	    {
		Format( Text_Pointer->Text ) ;
		Window_Output( Logical ) ;
		Window_Output( "\r\n" ) ; /* Display next line */
	    } ;
	    Text_Pointer = Text_Pointer->Next_Line ;
	    /* Follow link to next line */
	    A = A + 1 ; /* Count up lines */
	} ; /* while( A<23 ) */
    } ;
    Save = Screen_Window->Attribute ;
    Set_Window( Screen_Window, 0, 0, 0, -32 ) ; /* Normal video */
    Window_Output( Blank ) ; /* blank line first */
    Window_Output( "\r" ) ;
    Screen_Window->Attribute = Save ; /* Restore current attributes */
    if( Text_Pointer == NULL ) /* If at the end of the buffer */
    {
	Set_Window( Screen_Window, 0, 0, 0, -32 ) ; /* Normal video here */
	Window_Output( "[EOB]" ) ; /* Show end of buffer */
    }
    else
    {
	Format( Text_Pointer->Text ) ;
	Window_Output( Logical ) ; /* Display next line */
    } ;
    Update_Window( Screen_Window ) ;
    Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, -32 ) ;
} ;

static void Do_Help() /* Display help on screen */
{

int Dummy ;
char Input_Line[ 256 ] ;
int Line_Count ;
struct FILE *Help_File ;

    Help_File = fopen( Help_File_Name, "rt" ) ;
    if( Help_File == NULL )
    {
	Show_Error( "\aHelp not available" ) ;
	return ;
    } ;
    Clear_Window( Screen_Window ) ;
    Line_Count = 0 ;
    while( True )
    {
	if( fgets( Input_Line, 255, Help_File ) == NULL ) /* Get next line */
	{
	    Show_Error( "Press ENTER to continue..." ) ;
	    Select_Window( Error_Window ) ;
	    Do_Read() ;
	    Select_Window( Screen_Window ) ;
	    fclose( Help_File ) ;
	    Clear_Window( Screen_Window ) ; /* Erase screen and home cursor */
	    Redraw() ;
	    return ;
	} ;
	if( Input_Line[ 0 ] != '~' )
	{
	    if( Input_Line[ 0 ] == '!' )
	    {
		strcpy( Input_Line, &Input_Line[ 1 ] ) ;
	    } ;
	    Dummy = Instr( 1, Input_Line, "" ) ; /* Locate special symbol */
	    while( Dummy > 0 )
	    {
		Window_Output( Left( Input_Line, Dummy - 1 ) ) ;
		/* Show first part */
		strcpy( Input_Line, &Input_Line[ Dummy ] ) ;
		/* Extract rest */
		Set_Window( Screen_Window, 0, 0, 0, -1 ) ;
		/* Now reverse video */
		Dummy = Instr( 1, Input_Line, "" ) ;
		if( Dummy > 0 ) /* Found one */
		{
		    Window_Output( Left( Input_Line, Dummy - 1 ) ) ;
		    /* Show first part */
		    strcpy( Input_Line, &Input_Line[ Dummy ] ) ;
		    /* Extract rest */
		    Set_Window( Screen_Window, 0, 0, 0, -32 ) ;
		    /* Now normal video */
		} ;
		Dummy = Instr( 1, Input_Line, "" ) ; /* Locate another */
	    } ;
	    Window_Output( Input_Line ) ;
	    Window_Output( "\r" ) ;
	    Set_Window( Screen_Window, 0, 0, 0, -32 ) ; /* Normal text */
	    Line_Count = Line_Count + 1 ;
	    if( Line_Count == 21 )
	    {
		Line_Count = 1 ;
		Show_Error( "More..." ) ;
		Select_Window( Error_Window ) ;
		Do_Read() ;
		Select_Window( Screen_Window ) ;
		if( Kb_Buffer[ 0 ] == 3 || Kb_Buffer[ 0 ] == 25 ||
		    Kb_Buffer[ 0 ] == 26 )
		{
		    fclose( Help_File ) ;
		    Clear_Window( Screen_Window ) ;
		    /* Erase screen and home cursor */
		    Redraw() ;
		    return ;
		} ;
	    } ;
	} ;
    } ;
} ;

static void Handle_Selection( short int Cursor_Row, short int Physical_Column,
    char *Physical_Text ) /* Reverse video for selected text */
{

short int C ; /* Dummy variable */
long int First_Line, Final_Line ;
short int First_Column, Final_Column ;

    if( Selection ) /* Selection in progress */
    {
	if( Screen_Line + Cursor_Row - 1 < Anchor.Line ||
	    ( Screen_Line + Cursor_Row - 1 == Anchor.Line &&
	    Physical_Column < Anchor.P_Column ) )
	/* If we're before the selection point */
	{
	    First_Line = Screen_Line + Cursor_Row - 1 ;
	    Final_Line = Anchor.Line ;
	    First_Column = Physical_Column ;
	    Final_Column = Anchor.P_Column ;
	}
	else
	{
	    Final_Line = Screen_Line + Cursor_Row - 1 ;
	    First_Line = Anchor.Line ;
	    Final_Column = Physical_Column ;
	    First_Column = Anchor.P_Column ;
	} ;
	if( Screen_Line + Cursor_Row - 1 >= First_Line &&
	    Screen_Line + Cursor_Row - 1 <= Final_Line )
	/* Within select range */
	{
	    Set_Window( Screen_Window, Cursor_Row, 1, 0, 0 ) ;
	    if( First_Line == Final_Line )
	    {
		strcpy( Character, Left( Physical_Text, First_Column - 1 ) ) ;
		Format( Character ) ;
		C = strlen( Logical ) ;
		strcpy( Character, Left( Physical_Text, Final_Column ) ) ;
		Format( Character ) ;
		Window_Output( Left( Logical, C ) ) ;
		Set_Window( Screen_Window, 0, 0, 0, -1 ) ; /* Reverse video */
		Window_Output( &Logical[ C ] ) ;
		C = strlen( Logical ) ;
		Format( Physical_Text ) ;
		Set_Window( Screen_Window, 0, 0, 0, -32 ) ;
		Window_Output( &Logical[ C ] ) ;
	    }
 	    else
	    {
		if( Screen_Line + Cursor_Row - 1 == First_Line )
		{
		    strcpy( Character,
			Left( Physical_Text, First_Column - 1 ) ) ;
		    Format( Character ) ;
		    C = strlen( Logical ) ;
		    Window_Output( Left( Logical, C ) ) ;
		    Set_Window( Screen_Window, 0, 0, 0, -1 ) ;
		    /* Reverse video */
		    Format( Physical_Text ) ;
		    Window_Output( &Logical[ C ] ) ;
		}
		else
		{
		    if( Screen_Line + Cursor_Row - 1 == Final_Line )
		    {
			strcpy( Character,
			    Left( Physical_Text, Final_Column - 1 ) ) ;
			Format( Character ) ;
			C = strlen( Logical ) ;
			Set_Window( Screen_Window, 0, 0, 0, -1 ) ;
			/* Reverse video */
			Window_Output( Logical ) ;
			Set_Window( Screen_Window, 0, 0, 0, -32 ) ;
			/* Normal video */
			Format( Physical_Text ) ;
			Window_Output( &Logical[ C ] ) ;
		    }
		    else
		    {
			Set_Window( Screen_Window, 0, 0, 0, -1 ) ;
			/* Reverse video */
			Window_Output( Logical_Text ) ;
		    } ;
		} ;
	    } ;
	    Set_Window( Screen_Window, 0, 0, 0, -32 ) ; /* Back to normal */
	    Update_Window( Screen_Window ) ;
	} ;
    } ; /* IF Selection */
} ;

static void Adjust() /* Adjust cursor position based on tabs, EOL, etc. */
{

int C, L, P ; /* Work variables (Character, logical, physical pointers) */

    Cursor_Column = Logical_Column ;
    /* Return to proper column.  Logical_Column only changes when a vertical
     cursor movement or editing takes place. */
    if( Cursor_Column > strlen( Logical_Text ) ) /* Don't exceed line length */
    {
	Cursor_Column = strlen( Logical_Text ) + 1 ;
	Physical_Column = strlen( Physical_Text ) + 1 ;
	return ;
    } ;
    L = 1 ; /* Start at first character position */
    P = 1 ;
    while( L < Cursor_Column )
    {
	C = Physical_Text[ P - 1 ] ; /* Get next character */
	if( C == 9 ) /* Tab */
	{
	    L = ( L + 8 ) & ( ~7 ) + 1 ;
	}
	else
	{
	    if( ( C < 32 || C == 127 ) && C != 4 && C != 20 && C != 21 )
	    /* Control characters */
	    {
		while( Logical_Text[ L - 1 ] != '>' )
		{
		    L = L + 1 ;
		} ;
		L = L + 1 ; /* Skip to end of control-character */
	    }
	    else
	    {
		L = L + 1 ;
	    } ;
	} ;
	P = P + 1 ; /* Look at next character */
    } ;
    Cursor_Column = L ;
    Physical_Column = P ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
} ;

static void Cursor_Up()
{

int A ; /* Loop variable */
struct Line_Record *T_Pointer ; /* Temp line pointer */

    if( Cursor_Row + Screen_Line < 3 ) /* Top of buffer */
    {
	Show_Error( "\aBackup past top of buffer" ) ; /* Error */
	return ;
    } ;
    Do_Update() ; /* Update current line */
    Handle_Selection( Cursor_Row, 1, Physical_Text ) ;
    Cursor_Row = Cursor_Row - 1 ;
    if( Current_Line == NULL ) /* Sitting at [EOB] */
    {
	Current_Line = File_Last ;
    }
    else
    {
	Current_Line = Current_Line->Last_Line ; /* Backup */
    } ;
    strcpy( Physical_Text, Current_Line->Text ) ;
    Format( Physical_Text ) ;
    strcpy( Logical_Text, Logical ) ;
    Adjust() ;
    if( Cursor_Row < 8 && /* Scrolling time */ Screen_Line > 1 )
    {
	Screen_Line = Screen_Line - 1 ;
	Last_Screen_Line = Screen_Line ;
	Cursor_Row = 8 ;
	Set_Window( Screen_Window, 1, 1, 0, 0 ) ; /* Go to top line */
	Insert_Line( Screen_Window, 1 ) ;
	/* Insert a blank line at top (scroll down) */
	T_Pointer = Current_Line ;
	for( A = 1; A <= 7; A++ )
	{
	    T_Pointer = T_Pointer->Last_Line ; /* Find line */
	} ;
	Format( T_Pointer->Text ) ;
	Window_Output( Logical ) ; /* Write newly appearing line */
	Update_Window( Screen_Window ) ;
	Handle_Selection( 1, 81, T_Pointer->Text ) ;
    } ;
    Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
    /* Return to cursor */
} ;

static void Cursor_Down() /* Move cursor down one line */
{

int A ; /* Loop variable */
struct Line_Record *T_Pointer ; /* Temporary line pointer */

    if( Current_Line == NULL ) /* Bottom of buffer */
    {
	Show_Error( "\aAdvance past end of buffer" ) ; /* Error */
	return ;
    } ;
    Do_Update() ; /* Update current line */
    Handle_Selection( Cursor_Row, 81, Physical_Text ) ;
    Cursor_Row = Cursor_Row + 1 ;
    Current_Line = Current_Line->Next_Line ; /* Forward */
    if( Current_Line == NULL )
    {
	Physical_Text[ 0 ] = 0 ;
	strcpy( Logical, "[EOB]" ) ;
	memset( &Logical[ 5 ], 32, 75 ) ; /* Blank the line first */
	Logical[ 80 ] = 0 ;
	Cursor_Column = 1 ;
	Physical_Column = 1 ;
    }
    else
    {
	strcpy( Physical_Text, Current_Line->Text ) ;
	Format( Physical_Text ) ;
	strcpy( Logical_Text, Logical ) ;
	Adjust() ;
    } ;
    if( Cursor_Row > 16 && /* Scrolling time */ Screen_Line + 21 < Last_Line )
    {
	Screen_Line = Screen_Line + 1 ;
	Last_Screen_Line = Screen_Line ;
	Cursor_Row = 16 ;
	Set_Window( Screen_Window, 23, 1, 0, 0 ) ; /* Go to bottom of screen */
	Window_Output( "\n" ) ; /* Scroll up one line */
	T_Pointer = Current_Line ;
	for( A = 1; A <= 7; A++ )
	{
	    T_Pointer = T_Pointer->Next_Line ; /* Find line */
	} ;
	if( T_Pointer == NULL )
	{
	    Window_Output( "[EOB]" ) ;
	    for( Dummy = 6; Dummy <= 80; Dummy++ )
	    {
		Window_Output( " " ) ; /* Blank rest of line */
	    } ;
	}
	else
	{
	    Format( T_Pointer->Text ) ;
	    Window_Output( Logical ) ; /* Display newly appearing line */
	    Handle_Selection( 23, 1, T_Pointer->Text ) ;
	} ;
	Update_Window( Screen_Window ) ;
    } ;
    Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
    /* Return to cursor */
} ;

static void Cursor_Left() /* Move cursor left one character */
{

short int Ord_C ;
short int Dummy ; /* Saved cursor column */

    if( Physical_Column == 1 ) /* Going past start of line */
    {
	if( Cursor_Row + Screen_Line < 3 )
	/* Top of buffer */
	{
	    Cursor_Up() ; /* Show error */
	}
	else
	{
	    Cursor_Up() ;
	    Cursor_Column = strlen( Logical_Text ) + 1 ;
	    Physical_Column = strlen( Physical_Text ) + 1 ;
	    Logical_Column = Cursor_Column ;
	} ;
	Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
	return ;
    } ;
    Physical_Column = Physical_Column - 1 ;
    /* Physical movement is one character */
    Dummy = Cursor_Column ;
    if( Physical_Column < 81 )
    {
	Ord_C = Physical_Text[ Physical_Column - 1 ] ;
	/* Get character that we're jumping over */
	if( ( Ord_C > 30 && Ord_C != 127 ) || Ord_C == 4 || Ord_C == 20 ||
	    Ord_C == 21 ) /* Printing character */
	{
	    Cursor_Column = Cursor_Column - 1 ;
 	}
	else
	{
	    if( Ord_C != 9 ) /* Control character */
	    {
		while( Logical_Text[ Cursor_Column - 1 ] != '<' )
		{
		    Cursor_Column = Cursor_Column - 1 ;
		} ;
	    }
	    else /* Tab */
	    {
		strcpy( Character,
		    Left( Physical_Text, Physical_Column - 1 ) ) ;
		Format( Character ) ;
		Cursor_Column = Real_Length + 1 ;
	    } ;
	} ;
    } ;
    if( Dummy > 80 && Cursor_Column < 81 ) /* Cursor became visible */
    {
	strcpy( Character, Left( Physical_Text, Physical_Column - 1 ) ) ;
	Format( Character ) ;
	Cursor_Column = Real_Length + 1 ;
    } ;
    Logical_Column = Cursor_Column ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
} ;

static void Cursor_Right() /* Move cursor one character to the right */
{

short int Ord_C ;

    if( Current_Line == NULL ) /* If at [EOB] */
    {
	Cursor_Down() ; /* Display error */
	return ;
    } ;
    if( Physical_Column > strlen( Physical_Text ) ) /* Going past end of line */
    {
	if( Current_Line != NULL ) /* Not at [EOB] */
	{
	    Do_Update() ;
	    Cursor_Column = 1 ;
	    Physical_Column = 1 ;
	    Logical_Column = 1 ;
	} ;
	Cursor_Down() ;
	return ;
    } ;
    Ord_C = Physical_Text[ Physical_Column - 1 ] ;
    /* Get character that we're jumping over */
    Physical_Column = Physical_Column + 1 ;
    /* Physical movement is one character */
    if( ( Ord_C > 31 && Ord_C != 127 ) || Ord_C == 4 || Ord_C == 20 ||
	Ord_C == 21 ) /* Printing character */
    {
	Cursor_Column = Cursor_Column + 1 ;
    }
    else
    {
	if( Ord_C != 9 ) /* Control character */
	{
	    while( Logical_Text[ Cursor_Column - 1 ] != '>' &&
		Cursor_Column < 81 )
	    {
		Cursor_Column = Cursor_Column + 1 ;
	    } ;
	    if( Cursor_Column < 81 )
	    {
		Cursor_Column = Cursor_Column + 1 ;
	    } ;
	}
	else /* Tab */
	{
	    Cursor_Column = ( Cursor_Column + 7 ) & ( ~7 ) + 1 ;
	} ;
    } ;
    Logical_Column = Cursor_Column ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
} ;

static void Deletecrlf( int Explicit )
/* Delete CRLF between last line and this line.  WARNING: Should only be called
 from Delete_Char.  Use Delete_Char to delete lines. */
{

short int A, B ; /* Dummy variables */
struct Line_Record *L_Pointer, *T_Pointer ;

    if( Explicit ) /* Explicit deletion */
    {
	Deleted_Character = True ;
	Character_Buffer = 13 ;
    } ;
    T_Pointer = Current_Line ;
    if( Screen_Line == 1 ||
	( Cursor_Row > 11 && Screen_Line + 21 < Last_Line ) )
	/* If deletion of line should cause a scroll-up */
    {
	for( Dummy = Cursor_Row; Dummy <= 23; Dummy++ )
	{
	    if( T_Pointer != NULL )
	    {
		T_Pointer = T_Pointer->Next_Line ;
	    } ;
	} ;
	Cursor_Row = Cursor_Row - 1 ; /* Backup one line */
	Set_Window( Screen_Window, Cursor_Row, 1, 0, 0 ) ;
	Delete_Line( Screen_Window, 1 ) ; /* Scroll up */
	Set_Window( Screen_Window, 23, 1, 0, 0 ) ;
	A = 23 ; /* For selection routine */
	B = 1 ;
    }
    else
    {
	if( T_Pointer != NULL )
	{
	    A = 1 ;
	}
	else
	{
	    A = 2 ;
	    T_Pointer = File_Last ;
	} ;
	Delete_Line( Screen_Window, 0 ) ; /* Scroll down */
	Screen_Line = Screen_Line - 1 ;
	Last_Screen_Line = Screen_Line ;
	for( Dummy = Cursor_Row; Dummy >= A; Dummy-- )
	{
	    T_Pointer = T_Pointer->Last_Line ; /* Get newly appearing line */
	} ;
	Set_Window( Screen_Window, 1, 1, 0, 0 ) ;
	A = 1 ; /* For selection routine */
	B = 81 ;
    } ;
    if( T_Pointer == NULL ) /* Screen is not full */
    {
	memset( Logical, 32, 80 ) ; /* Blank it */
	if( Last_Line > 22 ) /* If last screen line (minus 1) is EOB */
	{
	    strcpy( Logical, "[EOB]" ) ; /* Show EOB */
	    Logical[ 5 ] = ' ' ;
	} ;
	Logical[ 80 ] = 0 ;
    }
    else
    {
	Format( T_Pointer->Text ) ; /* Else format newly appearing line */
    } ;
    Logical[ 80 ] = 0 ;
    Window_Output( Logical ) ; /* Display newly appearing line */
    Update_Window( Screen_Window ) ;
    Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
    Last_Line = Last_Line - 1 ;
    L_Pointer = Current_Line ;
    if( L_Pointer == NULL )
    {
	L_Pointer = File_Last ;
    } ;
    if( File_Last == L_Pointer ) /* Deleted last line */
    {
	File_Last = L_Pointer->Last_Line ;
    } ;
    if( File_Root == L_Pointer ) /* Deleted first line */
    {
	File_Root = L_Pointer->Last_Line ;
    } ;
    if( L_Pointer->Last_Line != NULL )
    {
	L_Pointer->Last_Line->Next_Line = L_Pointer->Next_Line ;
    } ;
    if( L_Pointer->Next_Line != NULL )
    {
	L_Pointer->Next_Line->Last_Line = L_Pointer->Last_Line ;
    } ;
    /* Relink previous and next text lines */
    T_Pointer = Current_Line ;
    if( Current_Line != NULL )
    {
	Current_Line = Current_Line->Last_Line ;
	/* Backup one text line */
    } ;
    Handle_Selection( A, B, T_Pointer->Text ) ;
    if( T_Pointer != NULL )
    {
	free( T_Pointer ) ; /* Delete line */
    } ;
} ;


static void New_Line() /* Insert a new line/create two lines from one */
{

char Character[ 256 ] ;
int Dummy ;
struct Line_Record *T_Pointer ;

    if( coreleft() < 256 )
    {
	Show_Error( "\aOut of memory" ) ; /* Error */
	return ;
    } ;
    strcpy( Character, &Physical_Text[ Physical_Column - 1 ] ) ;
    /* Save last part of line */
    strcpy( Physical_Text, Left( Physical_Text, Physical_Column - 1 ) ) ;
    Do_Update() ; /* Truncate the current line */
    Dummy = -32 ; /* Normal video */
    if( Selection && Anchor.Line < Screen_Line + Cursor_Row )
    {
	Dummy = -1 ; /* Reverse video if after beginning of selection */
    }
    Set_Window( Screen_Window, Cursor_Row, 1, 0, 0 ) ;
    /* Go to beginning of line */
    Window_Output( Blank ) ; /* Blank line */
    Window_Output( "\r" ) ;
    Set_Window( Screen_Window, 0, 0, 0, Dummy ) ; /* Set to proper video mode */
    Format( Physical_Text ) ;
    Window_Output( Logical ) ; /* Display changed/new line */
    Set_Window( Screen_Window, 0, 0, 0, -32 ) ; /* Normal video */
    strcpy( Physical_Text, Character ) ;
    Format( Physical_Text ) ;
    strcpy( Logical_Text, Logical ) ;
    Cursor_Column = 1 ;
    Physical_Column = 1 ;
    Logical_Column = 1 ;
    Last_Line = Last_Line + 1 ;
    if( Current_Line == NULL ) /* At [EOB] */
    {
	Physical_Text[ 0 ] = 0 ; /* Blank line */
	T_Pointer = malloc( 9 ) ; /* Allocate memory for blank line */
	T_Pointer->Text[ 0 ] = 0 ;
	T_Pointer->Next_Line = NULL ;
	T_Pointer->Last_Line = File_Last ;
	if( T_Pointer->Last_Line != NULL )
	{
	    T_Pointer->Last_Line->Next_Line = T_Pointer ;
	} ;
	File_Last = T_Pointer ;
	if( File_Root == NULL )
	{
	   File_Root = File_Last ;
	} ;
	strcpy( Logical, "[EOB]" ) ;
	Logical_Text[ 0 ] = 0 ;
    }
    else
    {
	Window_Output( "\r\n" ) ; /* Go down to next line */
	T_Pointer = malloc( strlen( Physical_Text ) + 9 ) ;
	strcpy( T_Pointer->Text, Physical_Text ) ;
	T_Pointer->Last_Line = Current_Line ;
	T_Pointer->Next_Line = Current_Line->Next_Line ;
	Current_Line->Next_Line = T_Pointer ;
	if( T_Pointer->Next_Line != NULL )
	{
	    T_Pointer->Next_Line->Last_Line = T_Pointer ; /* Insert new line */
	} ;
	if( File_Last == Current_Line )
	{
	    File_Last = T_Pointer ;
	} ;
	Current_Line = T_Pointer ; /* At new line */
    } ;
    if( Cursor_Row < 12 || Screen_Line + 22 > Last_Line )
    {
	Cursor_Row = Cursor_Row + 1 ; /* Scroll down */
	Set_Window( Screen_Window, Cursor_Row, 1, 0, 0 ) ;
	Insert_Line( Screen_Window, 1 ) ; /* Lose bottom line */
    }
    else
    {
	Screen_Line = Screen_Line + 1 ; /* Scroll up */
	Last_Screen_Line = Screen_Line ;
	Set_Window( Screen_Window, Cursor_Row + 1, 1, 0, 0 ) ;
	Insert_Line( Screen_Window, 0 ) ; /* Lose top line */
	Set_Window( Screen_Window, Cursor_Row, 1, 0, 0 ) ;
    } ;
    Window_Output( Logical ) ;
    Update_Window( Screen_Window ) ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
} ;

static void Delete_Char( int Explicit )
/* Delete character to left of cursor position */
{

char Character[ 256 ], Temp[ 256 ] ;
int Do_Left_Arrow ;
short int A, New_Column, Ord_C ;
struct Line_Record *L_Pointer ;

    Flag_Changes() ;
    if( Physical_Column == 1 ) /* Deleting CRLF */
    {
	if( Current_Line == NULL ) /* At [EOB] */
	{
	    if( File_Last == NULL )
	    {
		Do_Left_Arrow = True ;
	    }
	    else
	    {
		if( strlen( File_Last->Text ) > 0 )
		{
		    Do_Left_Arrow = True ;
		}
		else
		{
		    Do_Left_Arrow = False ;
		} ;
	    } ;
	    if( Do_Left_Arrow ) /* Nothing in file or last line not blank */
	    {
		Cursor_Left() ; /* Equivalent of left arrow */
		return ;
	    } ;
	    memset( Logical, 32, 80 ) ;
	    Logical[ 80 ] = 0 ;
	    Physical_Text[ 0 ] = 0 ;
	} ;
	if( Current_Line->Last_Line == NULL ) /* At first line */
	{
	    Show_Error( "\aBackup past top of buffer" ) ; /* Error */
	    return ;
	} ;
	if( strlen( Physical_Text ) +
	    strlen( Current_Line->Last_Line->Text ) > 255 )
	{
	    Show_Error( "\aLine exceeds 255 characters" ) ;
	    return ;
	} ;
	New_Column = strlen( Physical_Text ) ; /* Length of current line */
	if( Current_Line != NULL )
	{
	    Format( Current_Line->Last_Line->Text ) ;
	} ;
	Cursor_Column = strlen( Logical ) + 1 ; /* Position cursor */
	if( Current_Line != NULL )
	{
	    strcpy( Temp, Current_Line->Last_Line->Text ) ;
	    strcat( Temp, Physical_Text );
	    /* Concatenate current line and next line */
	    strcpy( Physical_Text, Temp ) ;
	} ;
	Format( Physical_Text ) ;
	for( A = 0; A < 80; A++ )
	{
	    if( Logical[ A ] == 0 )
	    {
		Logical[ A ] = 32 ;
	    } ;
	} ;
	Logical[ 80 ] = 0 ;
	strcpy( Logical_Text, Logical ) ;
	Set_Window( Screen_Window, Cursor_Row, 1, 0, 0 ) ;
	/* Position to current line */
	Deletecrlf( Explicit ) ;
	if( Current_Line == NULL )
	{
	    strcpy( Logical_Text, "[EOB]" ) ;
	    Logical_Text[ 5 ] = ' ' ;
	    Logical_Text[ 80 ] = 0 ;
	} ;
	Window_Output( "\r" ) ;
	Window_Output( Logical_Text ) ; /* Display newly appearing line */
	Update_Window( Screen_Window ) ;
	Physical_Column = strlen( Physical_Text ) - New_Column + 1 ;
	Logical_Column = Cursor_Column ;
	if( Cursor_Column > 80 )
	{
	    Cursor_Column = 80 ;
	} ;
	Adjust() ;
	return ;
    } ; /* Delete CRLF */

    Ord_C = Physical_Text[ Physical_Column - 2 ] ; /* Byte to delete */
    if( Explicit ) /* Explicit deletion */
    {
	Deleted_Character = True ;
	Character_Buffer = Ord_C ;
    } ;

    /* Delete character */
    strcpy( Temp, Left( Physical_Text, Physical_Column - 2 ) ) ;
    strcat( Temp, &Physical_Text[ Physical_Column - 1 ] ) ;
    strcpy( Physical_Text, Temp ) ;
    Physical_Column = Physical_Column - 1 ;
    if( Ord_C == 9 || Cursor_Column > 80 ) /* Tab */
    {
	strcpy( Character, Left( Physical_Text, Physical_Column - 1 ) ) ;
	Format( Character ) ;
	Cursor_Column = strlen( Logical ) + 1 ;
    }
    else
    {
	Character[ 0 ] = Ord_C ;
	Character[ 1 ] = 0 ;
	Format( Character ) ;
	Cursor_Column = Cursor_Column - strlen( Logical ) ;
    } ;
    Logical_Column = Cursor_Column ;
    Format( Physical_Text ) ;
    strcpy( Logical_Text, Logical ) ;
    if( Cursor_Column < 81 )
    {
	for( A = 0; A < 80; A++ )
	{
	    if( Logical[ A ] == 0 )
	    {
		Logical[ A ] = 32 ;
	    } ;
	} ;
	Logical[ 80 ] = 0 ; /* Make display exactly 80 characters */
	Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
	Window_Output( &Logical[ Logical_Column - 1 ] ) ;
	/* Show changed line */
	Update_Window( Screen_Window ) ;
    } ;
    Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
} ;

static void Insert_Chars( char *Character ) /* Character(s) insertion */
{

char Temp[ 256 ] ;

    Flag_Changes() ;
    if( strcmp( Character, "\r" ) == 0 ) /* Carriage Return */
    {
	New_Line() ; /* Break into two lines */
	return ;
    } ;
    if( Current_Line == NULL )
    {
	New_Line() ; /* Insert line at [EOB] */
	Cursor_Up() ; /* Stay at new line */
	Flag_Changes() ;
    } ;
    if( strlen( Physical_Text ) + strlen( Character ) > 255 )
    {
	Show_Error( "\aLine exceeds 255 characters" ) ;
	Iterate = 0 ;
	return ;
    } ;
    strcpy( Temp, Left( Physical_Text, Physical_Column - 1 ) ) ;
    strcat( Temp, Character ) ;
    strcat( Temp, &Physical_Text[ Physical_Column - 1 ] ) ;
    strcpy( Physical_Text, Temp ) ;
    Physical_Column = Physical_Column + strlen( Character ) ;
    strcpy( Character, Left( Physical_Text, Physical_Column - 1 ) ) ;
    Format( Character ) ;
    Cursor_Column = strlen( Logical ) + 1 ;
    Logical_Column = Cursor_Column ;
    Format( Physical_Text ) ;
    Set_Window( Screen_Window, Cursor_Row, 1, 0, 0 ) ;
    Window_Output( Logical ) ; /* Output updated line */
    strcpy( Logical_Text, Logical ) ;
    Update_Window( Screen_Window ) ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
} ;

static void Do_Search() /* Search for text */
{

int A ;
int Do_Redraw ; /* TRUE if redraw needs to be done */
struct Line_Record *T_Pointer ;
long int This_Line ;
char Temp[ 256 ] ;

    if( Selection )
    {
	Do_Redraw = True ; /* Must redraw if selection */
    }
    else
    {
	Do_Redraw = False ;
    } ;
    Do_Update() ;
    /* Update current line in case we move due to sucessful search */
    Select_Window( Error_Window ) ;
    Unmap_Window( Status_Window ) ;
    Window_Output( "\r\n" ) ;
    Window_Output( "Searching for: " ) ;
    Window_Output( Search_Text ) ;
    Update_Window( Error_Window ) ;
    Edit( Temp, Physical_Text, 32 ) ;
    A = Instr( Physical_Column + 1, Temp, Search_Text ) ;
    /* Search for text in current line */
    if( A > 0 ) /* Found */
    {
	Physical_Column = A ;
	strcpy( Character, Left( Physical_Text, A - 1 ) ) ;
	Format( Character ) ;
	Cursor_Column = strlen( Logical ) + 1 ;
	Logical_Column = Cursor_Column ;
	Window_Output( "\r\n" ) ; /* Clear search message and text */
	Update_Window( Error_Window ) ;
	Select_Window( Screen_Window ) ;
	Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
	if( Do_Redraw )
	{
	    Redraw() ;
	} ;
	return ;
    } ;

    /* Start search on line following cursor */
    This_Line = Screen_Line + Cursor_Row ;
    T_Pointer = Current_Line->Next_Line ;
    /* This many not be NULL even if current_line is */
    if( Current_Line == NULL )
    {
	T_Pointer = NULL ;
    } ;

    while( T_Pointer != NULL ) /* Until end of buffer */
    {
	Edit( Temp, T_Pointer->Text, 32 ) ;
	A = Instr( 1, Temp, Search_Text ) ;
	if( A > 0 ) /* Found it */
	{
	    Physical_Column = A ;
	    Current_Line = T_Pointer ;
	    strcpy( Physical_Text, Current_Line->Text ) ;
	    strcpy( Character, Left( Physical_Text, A - 1 ) ) ;
	    Format( Character ) ;
	    Cursor_Column = strlen( Logical ) + 1 ;
	    Logical_Column = Cursor_Column ;
	    if( This_Line > Screen_Line + 15 ) /* Out of range */
	    {
		Screen_Line = This_Line - 10 ;
		/* Put us at line 11 by default */
		if( Screen_Line + 21 > Last_Line )
		{
		    Screen_Line = Last_Line - 21 ; /* Don't go too far */
		} ;
		if( Screen_Line < 1 )
		{
		    Screen_Line = 1 ; /* ...in either direction */
		} ;
		Do_Redraw = True ;
		Select_Window( Error_Window ) ;
	    } ;
	    Cursor_Row = ( This_Line - Screen_Line ) + 1 ;
	    Format( Physical_Text ) ;
	    strcpy( Logical_Text, Logical ) ;
	    Window_Output( "\r\n" ) ; /* Clear search message and text */
	    Update_Window( Error_Window ) ;
	    Select_Window( Screen_Window ) ;
	    Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
	    if( Do_Redraw )
	    {
		Redraw() ;
	    } ;
	    return ;
	} ;
	T_Pointer = T_Pointer->Next_Line ;
	This_Line = This_Line + 1 ;
    } ;
    Window_Output( "\r\n" ) ; /* Clear search message and text */
    Window_Output( "\aText not found" ) ;
    Update_Window( Error_Window ) ;
    Select_Window( Screen_Window ) ;
} ;

static void Execute_Command( char *Passed_Command ) /* Execute a TEU command */
{

char Command[ 256 ], Command_Com[ 256 ] ;
/* Command and name of COMMAND.COM file */
char Directory[ 256 ] ; /* Saved current directory */
long int Temp ;
struct Line_Record *T1_Pointer, *T2_Pointer ;
struct Window_Record *Spawn_Window ;
union REGS Reg ;
char Temp_String[ 256 ] ;

    if( Passed_Command == NULL )
    {
	return ; /* Ignore blank lines */
    }
    else
    {
	strcpy( Command, Passed_Command ) ;
    } ;
    if( strlen( Command ) == 0 )
    {
	return ; /* Ignore blank lines */
    } ;
    if( Command[ 0 ] == 25 )
    {
	End_Edit() ;
	return ; /* ^Y will abort */
    } ;
    if( Command[ 0 ] == 3 )
    {
	return ; /* ^C is ignored here */
    } ;
    Do_Update() ; /* Update current line */
    if( Command[ 0 ] == 'E' ) /* EXIT command */
    {
	Select_Window( Error_Window ) ;
	Unmap_Window( Status_Window ) ;
	sprintf( Temp_String, "%s  %li line", Input_Name, Last_Line ) ;
	Window_Output( Temp_String ) ;
	if( Last_Line != 1 )
	{
	    Window_Output( "s" ) ;
	} ;
	Update_Window( Error_Window ) ;
	if( Any_Changes ) /* If changes made */
	{
	    strcpy( Command, Input_Name ) ; /* Save name for later comparison */
	    if( !New_File ) /* Editing of existing file */
	    {
		Dummy = Extension_Pos( Input_Name ) ;
		if( Dummy == 0 )
		{
		    Dummy = strlen( Input_Name ) + 1 ;
		} ;
		strcpy( Input_Name, Left( Input_Name, Dummy - 1 ) ) ;
		strcat( Input_Name, ".BAK" ) ; /* Strip extension */
		remove( Input_Name ) ; /* Delete old backup */
		if( Command != Input_Name ) /* Old file was not *.BAK */
		{
		    rename( Input_File, Input_Name ) ;
		    /* Rename previous file to *.BAK */
		} ;
	    } ;
	    Do_Write( Command, File_Root ) ; /* Write to file */
	} ;
	fclose( Journal ) ;
	remove( Journal_Name ) ; /* Close and delete journal file */
	End_Edit() ;
	exit( EXIT_SUCCESS ) ;
    } ;
    if( Command[ 0 ] == '?' || Command[ 0 ] == 'H' ) /* Help */
    {
	Do_Help() ;
	return ;
    } ;
    if( Command[ 0 ] == 'I' ) /* INCLUDE command */
    {
	Dummy = Instr( 1, Command, " " ) ;
	if( Dummy == 0 )
	{
	    Dummy = strlen( Command ) + 1 ;
	} ;
	strcpy( Command, &Command[ Dummy ] ) ;
	Include_File( Command, &Temp, &T1_Pointer, &T2_Pointer ) ;
	if( Temp < 0 )
	{
	    return ; /* Insufficient memory */
	} ;
	if( Temp > 0 ) /* Something was read in */
	{
	    Any_Changes = True ; /* Consider inclusions as changes */
	    if( File_Root == NULL ) /* Nothing in file */
	    {
		File_Root = T1_Pointer ;
		File_Last = T2_Pointer ;
		Current_Line->Last_Line = File_Last ;
	    }
	    else /* Something in file */
	    {
		if( Current_Line == NULL ) /* At [EOB] */
		{
		    File_Last->Next_Line = T1_Pointer ;
		    T1_Pointer->Last_Line = File_Last ;
		    File_Last = T2_Pointer ;
		    Current_Line->Last_Line = File_Last ;
		}
		else
		{
		    if( File_Root == Current_Line ) /* At first line */
		    {
			File_Root->Last_Line = T2_Pointer ;
			T2_Pointer->Next_Line = File_Root ;
			File_Root = T1_Pointer ;
		    }
		    else /* Insert in middle of buffer */
		    {
			Current_Line->Last_Line->Next_Line = T1_Pointer ;
			T1_Pointer->Last_Line = Current_Line->Last_Line ;
			Current_Line->Last_Line = T2_Pointer ;
			T2_Pointer->Next_Line = Current_Line ;
		    } ; /* IF File_Root=Current_Line ELSE */
		} ; /* if( Current_Line == NULL ) */
	    } ; /* if( File_Root == NULL ) */
	    Last_Line = Last_Line + Temp ;
	    Screen_Line = Screen_Line + Temp ;
	    if( Cursor_Row < 8 && Screen_Line > 1 ) /* Must adjust screen */
	    {
		Dummy = 8 - Cursor_Row ; /* Number of lines to move down */
		while( Screen_Line - Dummy < 1 )
		{
		    Dummy = Dummy - 1 ; /* Top at top */
		} ;
		Cursor_Row = Cursor_Row + Dummy ;
		Screen_Line = Screen_Line - Dummy ;
	    } ;
	    if( Cursor_Row > 16 && Last_Line > 22 ) /* Must adjust screen */
	    {
		Dummy = Cursor_Row - 16 ; /* Number of lines to move up */
		while( Screen_Line + 22 - Dummy > Last_Line )
		{
		    Dummy = Dummy - 1 ; /* [EOB] should be at bottom */
		} ;
		while( Screen_Line + Dummy + 21 > Last_Line )
		{
		    Dummy = Dummy - 1 ;
		} ;
		Cursor_Row = Cursor_Row - Dummy ;
		Screen_Line = Screen_Line + Dummy ;
	    } ;
	    if( Last_Line < 23 ) /* [EOB] should stay on screen */
	    {
		Screen_Line = 1 ;
		Cursor_Row = Cursor_Row + Temp ;
	    } ;
	    if( Screen_Line > 1 && Screen_Line + 22 > Last_Line )
	    /* Keep [EOB] at bottom of screen */
	    {
		Cursor_Row = Cursor_Row + ( Screen_Line + 21 - Last_Line ) ;
		Screen_Line = Last_Line - 21 ;
	    } ;
	    Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
	    Last_Screen_Line = 0 ; /* Force redraw */
	    Redraw() ;
	} ;
	return ;
    } ;
    if( Command[ 0 ] == 'Q' ) /* QUIT command */
    {
	if( Any_Changes )
	{
	    Clear_Cursor( 0, 0 ) ;
	    /* Disable cursor (enabled by "Command:" input) */
	    Show_Error( "Modifications will not be saved - continue quitting <NO> ?" ) ;
	    strcpy( Character, In_Key() ) ; /* Get next character */
	    if( strcmp( Character, "Y" ) != 0 && strcmp( Character, "y" ) != 0 )
	    /* Not "YES" */
	    {
		Show_Error( "" ) ;
		return ;
	    } ;
	} ;
	if( New_File )
	{
	    remove( Input_Name ) ; /* Delete new files during a quit */
	} ;
	fclose( Journal ) ;
	remove( Journal_Name ) ; /* Close and delete journal file */
	End_Edit() ;
	exit( EXIT_SUCCESS ) ;
    } ;
    if( Command[ 0 ] == 'R' )
    {
	return ; /* RESERVE command - ignored in editor */
    } ;
    if( strcmp( Left( Command, 2 ), "SE" ) == 0 ) /* SET command */
    {
	Dummy = Instr( 1, Command, " " ) ;
	if( Dummy == 0 )
	{
	    Dummy = strlen( Command ) + 1 ;
	} ;
	strcpy( Command, &Command[ Dummy ] ) ; /* Strip SET */
	if( Command[ 0 ] == 'A' ) /* SET AUTOINDENT */
	{
	    Dummy = Instr( 1, Command, " " ) ;
	    if( Dummy == 0 )
	    {
		Dummy = strlen( Command ) + 1 ;
	    } ;
	    strcpy( Command, &Command[ Dummy ] ) ; /* Strip AUTOINDENT */
	    if( strlen( Command ) == 0 )
	    {
		Auto_Indent = True ; /* Default to SET AUTOINDENT ON */
	    }
	    else
	    {
		if( strcmp( Command, "ON" ) == 0 )
		{
		    Auto_Indent = True ; /* Turn it on */
		}
		else
		{
		    if( strcmp( Command, "OFF" ) == 0 )
		    {
			Auto_Indent = False ; /* Turn it off */
		    }
		    else
		    {
			Show_Error( "\aIllegal command" ) ;
		    } ;
		} ;
	    } ;
	    return ;
	} ;
	if( Command[ 0 ] == 'J' ) /* SET JOURNAL */
	{
	    Dummy = Instr( 1, Command, " " ) ;
	    if( Dummy == 0 )
	    {
		Dummy = strlen( Command ) + 1 ;
	    } ;
	    strcpy( Command, &Command[ Dummy ] ) ; /* Strip JOURNAL */
	    if( strlen( Command ) > 0 )
	    {
		if( strcmp( Command, "OFF" ) == 0 )
		{
		    strcpy( Command, "0" ) ;
		} ;
		Dummy = atoi( Command ) ;
		if( Dummy != 0 || Temp < 0 || Temp > 64 ) /* Error */
		{
		    Show_Error( "\aInvalid parameter" ) ;
		}
		else
		{
		    if( Temp == 64 )
		    {
			Temp = 127 ;
		    } ;
		    if( Temp > 32 && Temp <= 64 )
		    {
			Temp = 63 ;
		    } ;
		    if( Temp > 16 && Temp <= 32 )
		    {
			Temp = 31 ;
		    } ;
		    if( Temp > 8 && Temp <= 16 )
		    {
			Temp = 15 ;
		    } ;
		    if( Temp > 4 && Temp <= 7 )
		    {
			Temp = 7 ;
		    } ;
		    if( Temp > 2 && Temp <= 4 )
		    {
			Temp = 3 ;
		    } ;
		    if( Temp > 0 && Temp <= 2 )
		    {
			Temp = 1 ;
		    } ;
		    Journal_Size = Temp ; /* New journaling size */
		} ;
		return ;
	    } ;
	} ;
	if( Command[ 0 ] == 'S' ) /* SET STATUS */
	{
	    Dummy = Instr( 1, Command, " " ) ;
	    if( Dummy == 0 )
	    {
		Dummy = strlen( Command ) + 1 ;
	    } ;
	    strcpy( Command, &Command[ Dummy ] ) ; /* Strip STATUS */
	    if( strlen( Command ) == 0 )
	    {
		Status = True ; /* Default to SET STATUS ON */
	    }
	    else
	    {
		if( Command == "ON" )
		{
		    Status = True ; /* Turn it on */
		}
		else
		{
		    if( strcmp( Command, "OFF" ) == 0 )
		    {
			Status = False ; /* Turn it off */
		    }
		    else
		    {
			Show_Error( "\aIllegal command" ) ;
		    } ;
		} ;
	    } ;
	    return ;
	} ;
	if( strlen( Command ) == 0 )
	{
	    Show_Error( "\aMissing operand" ) ;
	}
	else
	{
	    Show_Error( "\aIllegal command" ) ;
	} ;
	return ;
    } ;
    if( strcmp( Left( Command, 2 ), "SP" ) == 0 ) /* SPAWN command */
    {
	Spawn_Window = Create_Window( 24, 80, 4 ) ;
	Map_Window( Spawn_Window, 1, 24, 1, 80, 1, 1, 253 ) ;
	Update_Window( Spawn_Window ) ; /* Clear screen */
	getcurdir( 0, Directory ) ; /* Save current disk/directory */
	strcpy( Command_Com, getenv( "COMSPEC" ) ) ;
	if( strlen( Command_Com ) == 0 ) /* Not defined */
	{
	    strcpy( Command_Com, Filnam( "COMMAND.COM" ) ) ;
	} ;
	if( spawnl( P_WAIT, Command_Com, NULL ) == -1 ) /* Spawn */
	{
	    puts( strerror( errno ) ) ; /* Show error */
	} ;
	_fpreset() ;
	chdir( Directory ) ;
	Unmap_Window( Spawn_Window ) ;
	Delete_Window( Spawn_Window ) ;
	Update_Window( Screen_Window ) ;
	return ;
    } ;
    if( Command[ 0 ] == 'W' ) /* WRITE command */
    {
	Dummy = Instr( 1, Command, " " ) ;
	if( Dummy == 0 )
	{
	    Dummy = strlen( Command ) + 1 ;
	} ;
	strcpy( Command, &Command[ Dummy ] ) ;
	Do_Write( Command, File_Root ) ;
	return ;
    } ;
    Show_Error( "\aIllegal command" ) ;
} ;

static void Teu_Advance() /* Advance to beginning of next line */
{
    Cursor_Down() ;
    Cursor_Column = 1 ;
    Logical_Column = 1 ;
    Physical_Column = 1 ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
} ;

static void Teu_Move_Word() /* Move to beginning of next word */
{
    if( Physical_Column > strlen( Physical_Text ) )
    {
	Cursor_Down() ;
	Cursor_Column = 1 ;
	Logical_Column = 1 ;
	Physical_Column = 1 ;
    }
    else /* Move by word */
    {
	A = Physical_Column ;
	while( Physical_Column <= strlen( Physical_Text ) &&
	    Physical_Text[ Physical_Column - 1 ] != ' ' && /* not space */
	    Physical_Text[ Physical_Column - 1 ] != '\t' ) /* not tab */
	{
	    Physical_Column = Physical_Column + 1 ;
	} ;
	if( A == Physical_Column )
	{
	    Physical_Column = Physical_Column + 1 ;
	} ;
	while( Physical_Column <= strlen( Physical_Text ) &&
	    Physical_Text[ Physical_Column - 1 ] == ' ' )
	{
	    Physical_Column = Physical_Column + 1 ;
	    /* Jump to next non-space character */
	} ;
	strcpy( Character, Left( Physical_Text, Physical_Column - 1 ) ) ;
	Format( Character ) ;
	Cursor_Column = strlen( Logical ) + 1 ;
	Logical_Column = strlen( Logical ) + 1 ;
    } ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
} ;

static void Teu_Eol() /* Move to end of current line */
{
    if( Current_Line == NULL )
    {
	return ; /* Do nothing if at [EOB] */
    } ;
    if( Physical_Column > strlen( Physical_Text ) ) /* At EOL */
    {
	Cursor_Down() ; /* Goto EOL of next line */
    } ;
    Cursor_Column = 1 ;
    Logical_Column = 1 ;
    Physical_Column = 1 ;
    Cursor_Down() ;
    Cursor_Left() ;
} ;

static void Teu_Cut() /* Cut selected text and put it into the paste buffer */
{

long int First_Line, Final_Line, Root_Line ;
short int A, First_Column, Final_Column ;
struct Line_Record *L_Pointer, *T_Pointer ;
char Temp[ 256 ] ;

    if( !Selection ) /* Must have something selected in order to cut */
    {
	Show_Error( "\aNothing selected" ) ;
	return ;
    } ;
    Selection = False ; /* Cut operation ends selection */

    Do_Update() ; /* Make sure everything is updated */

    /* Determine start/end lines to cut */
    if( Screen_Line + Cursor_Row - 1 < Anchor.Line )
    /* If we're before the selection point */
    {
	First_Line = Screen_Line + Cursor_Row - 1 ;
	Final_Line = Anchor.Line ;
	First_Column = Physical_Column ;
	Final_Column = Anchor.P_Column ;
    }
    else
    {
	Final_Line = Screen_Line + Cursor_Row - 1 ;
	First_Line = Anchor.Line ;
	Final_Column = Physical_Column ;
	First_Column = Anchor.P_Column ;
    } ;

    /* Delete existing paste buffer */
    while( Paste_Root != NULL ) /* Something already in paste buffer */
    {
	T_Pointer = Paste_Root->Next_Line ;
	/* Get pointer to next line in buffer */
	free( Paste_Root ) ; /* Delete current line */
	Paste_Root = T_Pointer ; /* Move to next line */
    } ;

    if( Screen_Line + Cursor_Row - 1 == Anchor.Line )
    /* Cut within a single line */
    {
	if( Physical_Column < Anchor.P_Column )
	{
	    First_Column = Physical_Column ;
	    Final_Column = Anchor.P_Column ;
	}
	else
	{
	    First_Column = Anchor.P_Column ;
	    Final_Column = Physical_Column ;
	    Physical_Column = First_Column ;
	} ;
	strcpy( Character, &Physical_Text[ First_Column - 1 ] ) ;
	Character[ Final_Column - First_Column + 1 ] = 0 ;
	/* Get selected characters */
	Paste_Root = malloc( strlen( Character ) + 9 ) ;
	/* Create paste buffer */
	Paste_Root->Next_Line = NULL ;
	Paste_Root->Last_Line = NULL ;
	strcpy( Paste_Root->Text, Character ) ;

	strcpy( Character, Left( Physical_Text, First_Column - 1 ) ) ;
	/* Get portion of line before selected portion */
	Format( Character ) ;
	Cursor_Column = strlen( Logical ) + 1 ; /* Resulting position */
	strcpy( Character, &Physical_Text[ First_Column - 1 ] ) ;
	Character[ Final_Column - First_Column + 1 ] = 0 ;
	strcpy( Temp, Left( Physical_Text, First_Column - 1 ) ) ;
	strcat( Temp, &Physical_Text[ Final_Column - 1 ] ) ;
	/* Extract cut text from line */
	strcpy( Physical_Text, Temp ) ;
	Format( Physical_Text ) ;
	strcpy( Logical_Text, Logical ) ;
	for( A = 0; A < 80; A++ )
	{
	    if( Logical[ A ] == 0 )
	    {
		Logical[ A ] = 32 ;
	    } ;
	} ;
	Logical[ 80 ] = 0 ;
	Set_Window( Screen_Window, Cursor_Row, 1, 0, 0 ) ;
	Window_Output( Logical ) ; /* Show newly updated line */
	Update_Window( Screen_Window ) ;
	Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
	Flag_Changes() ;
	return ;
    } ;

    /* Multi-line cut operation */
    Current_Line = File_Root ; /* Start at beginning of buffer */
    Root_Line = 1 ;
    while( Root_Line < First_Line ) /* Find first line to cut */
    {
	Current_Line = Current_Line->Next_Line ; /* Follow link */
	if( Current_Line == NULL )
	{
	    return ; /* Hit end of buffer */
	} ;
	Root_Line = Root_Line + 1 ; /* Count up lines */
    } ;
    /* At this point: Current_Line points to first line of selection */

    Paste_Root = Current_Line ;
    while( Root_Line < Final_Line ) /* Find last line to cut */
    {
	if( Paste_Root != NULL ) /* Not at buffer end */
	{
	    Paste_Root = Paste_Root->Next_Line ;
	    Last_Line = Last_Line - 1 ; /* One less line in main buffer */
	} ;
	Root_Line = Root_Line + 1 ;
    } ;
    /* At this point: Current_Line points to first line of selection
		  Paste_Root points to last line of selection (NULL if at [EOB]) */

    Last_Line = Last_Line + 1 ; /* Adjust buffer line count */
    strcpy( Character, &Current_Line->Text[ First_Column - 1 ] ) ;
    /* Get portion of first line which is selected */

    T_Pointer = malloc( strlen( Character ) + 9 ) ;
    if( T_Pointer == NULL ) /* Ran out of memory */
    {
	Show_Error( "\aMaximum memory exceeded" ) ;
	End_Edit() ;
	exit( EXIT_FAILURE ) ;
    } ;
    strcpy( T_Pointer->Text, Character ) ;
    T_Pointer->Last_Line = NULL ;
    /* First line of paste buffer */

    if( Paste_Root == NULL ) /* [EOB] */
    {
	Character[ 0 ] = 0 ;
    }
    else
    {
	strcpy( Character, Left( Paste_Root->Text, Final_Column - 1 ) ) ;
    } ;
    /* Get portion of last line which is selected */

    L_Pointer = malloc( strlen( Character ) + 9 ) ;
    if( L_Pointer == NULL ) /* Ran out of memory */
    {
	Show_Error( "\aMaximum memory exceeded" ) ;
	End_Edit() ;
	exit( EXIT_FAILURE ) ;
    } ;
    strcpy( L_Pointer->Text, Character ) ;
    L_Pointer->Next_Line = NULL ;
    /* At this point: Current_Line points to first line of selection
		   Paste_Root points to last line of selection (NULL if at [EOB])
		   T_Pointer points to first line of paste buffer
		   L_Pointer points to last line of paste buffer */

    /* Now link up paste buffer */
    if( First_Line + 1 == Final_Line ) /* If only two lines */
    {
	L_Pointer->Last_Line = T_Pointer ;
	T_Pointer->Next_Line = L_Pointer ;
    }
    else
    {
	if( Paste_Root == NULL ) /* At [EOB] */
	{
	    L_Pointer->Last_Line = File_Last ;
	    File_Last->Next_Line = L_Pointer ;
	}
	else
	{
	    L_Pointer->Last_Line = Paste_Root->Last_Line ;
	    Paste_Root->Last_Line->Next_Line = L_Pointer ;
	} ;
	T_Pointer->Next_Line = Current_Line->Next_Line ;
	Current_Line->Next_Line->Last_Line = T_Pointer ;
    } ;

    /* Now remove selected lines from main buffer */
    Current_Line->Next_Line = Paste_Root ;
    if( Paste_Root == NULL ) /* At [EOB] */
    {
	File_Last = Current_Line ;
    }
    else
    {
	Paste_Root->Last_Line = Current_Line ;
    } ;
    Paste_Root = T_Pointer ; /* Make paste pointer point to paste buffer */

    /* Now adjust first and last selected lines */
    strcpy( Physical_Text, Current_Line->Text ) ;
    strcpy( Physical_Text, Left( Physical_Text, First_Column - 1 ) ) ;
    /* Adjust first line */
    Flag_Changes() ;
    Do_Update() ;
    Current_Line = Current_Line->Next_Line ; /* Go to next line */
    if( Current_Line == NULL ) /* At [EOB] */
    {
	Physical_Text[ 0 ] = 0 ;
    }
    else
    {
	strcpy( Physical_Text, &Current_Line->Text[ Final_Column - 1 ] ) ;
	/* Adjust last line */
    } ;
    Flag_Changes() ;
    Do_Update() ;

    /* Now get rid of extra line created by above process */
    Cursor_Column = 1 ;
    Physical_Column = 1 ;
    Logical_Column = 1 ;
    Delete_Char( False ) ;
    Do_Update() ;

    /* Now show updated screen */
    Cursor_Row = ( First_Line - Screen_Line ) + 1 ;
    while( Screen_Line > 1 && Screen_Line + 21 > Last_Line )
    {
	Screen_Line = Screen_Line - 1 ;
	Cursor_Row = Cursor_Row + 1 ;
    } ;

    Last_Screen_Line = 0 ; /* Force redraw */
    if( Cursor_Row < 8 || Cursor_Row > 127 ) /* Adjust screen */
    {
	Dummy = 8 - Cursor_Row ; /* Number of lines to adjust */
	Screen_Line = Screen_Line - Dummy ;
	Cursor_Row = Cursor_Row + Dummy ;
	while( Screen_Line < 1 ) /* First line can't be before line 1 */
	{
	    Screen_Line = Screen_Line + 1 ;
	    Cursor_Row = Cursor_Row - 1 ;
	} ;
    } ;
    Clear_Window( Screen_Window ) ; /* Erase screen and home cursor */
    Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
    Redraw() ; /* Redraw screen */
} ;

static void Teu_Page() /* Advance one page (22 lines) */
{
    if( Current_Line == NULL )
    {
	return ; /* Done if at EOB */
    } ;
    Do_Update() ;
    Physical_Text[ 0 ] = 0 ;

    /* Move forward 22 lines, but not past EOB */
    A = 0 ;
    for( Dummy = 1; Dummy <= 22; Dummy++ )
    {
	if( Current_Line != NULL )
	{
	    Current_Line = Current_Line->Next_Line ;
	    A = A + 1 ;
	} ;
    } ;
    Cursor_Row = Cursor_Row + A ;

    if( Cursor_Row > 11 && Last_Line > 22 ) /* Need to scroll */
    {
	Screen_Line = Screen_Line + Cursor_Row - 11 ;
	Cursor_Row = 11 ;
	if( Screen_Line + 22 > Last_Line ) /* Leave [EOB] at bottom */
	{
	    Cursor_Row = Cursor_Row + ( Screen_Line + 21 - Last_Line ) ;
	    Screen_Line = Last_Line - 21 ;
	} ;
	Redraw() ;
    } ;

    if( Current_Line != NULL )
    {
	strcpy( Physical_Text, Current_Line->Text ) ;
    } ;
    Format( Physical_Text ) ;
    strcpy( Logical_Text, Logical ) ;
    Adjust() ;
} ;

static void Teu_Half_Page() /* Advance one-half page (11 lines) */
{
    if( Current_Line == NULL )
    {
	return ; /* Done if already at EOB */
    } ;
    Do_Update() ;
    Physical_Text[ 0 ] = 0 ;

    /* Move forward 11 lines, but not past EOB */
    A = 0 ;
    for( Dummy = 1; Dummy <= 11; Dummy++ )
    {
	if( Current_Line != NULL )
	{
	    Current_Line = Current_Line->Next_Line ;
	    A = A + 1 ;
	} ;
    } ;
    Cursor_Row = Cursor_Row + A ;

    if( Cursor_Row > 11 && Last_Line > 22 ) /* Need to scroll */
    {
	Screen_Line = Screen_Line + Cursor_Row - 11 ; /* Move down 11 lines */
	Cursor_Row = 11 ; /* Fix cursor at line 11 */
	if( Screen_Line + 22 > Last_Line ) /* Leave [EOB] at bottom */
	{
	    Cursor_Row = Cursor_Row + ( Screen_Line + 21 - Last_Line ) ;
	    Screen_Line = Last_Line - 21 ;
	} ;
	Redraw() ;
    } ;

    if( Current_Line != NULL )
    {
	strcpy( Physical_Text, Current_Line->Text ) ;
    } ;
    Format( Physical_Text ) ;
    strcpy( Logical_Text, Logical ) ;
    Adjust() ;
} ;

static void Teu_Select() /* Begin selection */
{
    if( !Selection )
    {
	Selection = True ;
	Anchor.Line = Screen_Line + Cursor_Row - 1 ;
	Anchor.P_Column = Physical_Column ;
	Anchor.Column = Cursor_Column ;
    }
    else
    {
	Show_Error( "\aSelection in progress" ) ;
    } ;
} ;

static void Teu_Delete_Word() /* Delete current word */
{

char Temp[ 256 ] ;

    Deleted_Word = True ;
    if( Physical_Column > strlen( Physical_Text ) ) /* At EOL */
    {
	Cursor_Right() ;
	Delete_Char( False ) ;
	strcpy( Word_Buffer, "\r" ) ;
	return ;
    }
    else /* Delete word */
    {
	A = Physical_Column ; /* Save current position */
	while( Physical_Column <= strlen( Physical_Text ) &&
	    Physical_Text[ Physical_Column - 1 ] != ' ' && /* not space */
	    Physical_Text[ Physical_Column - 1 ] != '\t' ) /* not tab */
	{
	    Physical_Column = Physical_Column + 1 ;
	} ;
	if( A == Physical_Column )
	{
	    Physical_Column = Physical_Column + 1 ;
	} ;
	while( Physical_Column <= strlen( Physical_Text ) &&
	    Physical_Text[ Physical_Column - 1 ] == ' ' )
 	{
	    Physical_Column = Physical_Column + 1 ;
	    /* Jump to next non-space character */
	} ;
	strcpy( Word_Buffer, &Physical_Text[ A - 1 ] ) ;
	Word_Buffer[ Physical_Column - A ] = 0 ;
	/* Deleted text */
	strcpy( Temp, Left( Physical_Text, A - 1 ) ) ;
	strcat( Temp, &Physical_Text[ Physical_Column - 1 ] ) ;
	strcpy( Physical_Text, Temp ) ;
	Flag_Changes() ;
	Physical_Column = A ; /* Restore physical column */
	Format( Physical_Text ) ;
	strcpy( Logical_Text, Logical ) ;
	Set_Window( Screen_Window, Cursor_Row, 1, 0, 0 ) ; /* Show new line */
	for( A = 0; A < 80; A++ )
	{
	    if( Logical[ A ] == 0 )
	    {
		Logical[ A ] = 32 ;
	    } ;
	} ;
	Logical[ 80 ] = 0 ;
	Window_Output( Logical ) ;
	Update_Window( Screen_Window ) ;
    } ; /* if() */
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
} ;

static void Teu_Delete_Line()
{
    if( Current_Line == NULL ) /* At [EOB] */
    {
	Show_Error( "\aAdvance past end of buffer" ) ;
	return ;
    } ;
    if( Current_Line->Next_Line == NULL && strlen( Physical_Text ) > 0 &&
	Physical_Column > strlen( Physical_Text ) )
	/* If at end of last line, ignore */
    {
	Cursor_Right() ;
	return ;
    } ;
    Deleted_Line = True ;
    strcpy( Line_Buffer, &Physical_Text[ Physical_Column - 1 ] ) ;
    strcpy( Physical_Text, Left( Physical_Text, Physical_Column - 1 ) ) ;
    Format( Physical_Text ) ;
    strcpy( Logical_Text, Logical ) ;
    Flag_Changes() ;
    Cursor_Right() ; /* Go to beginning of next line */
    Delete_Char( False ) ; /* Delete CRLF */
} ;

static void Teu_Unselect() /* Unselect text */
{
    if( !Selection )
    {
	return ; /* No selection in progress */
    } ;
    Selection = False ;
    Do_Update() ;
    Redraw() ; /* Remove reverse-video */
} ;

static void Teu_Search() /* Search for text */
{
    Select_Window( Error_Window ) ;
    Unmap_Window( Status_Window ) ;
    Window_Output( "\r\n" ) ;
    Window_Output( "Text to search for: " ) ;
    Update_Window( Error_Window ) ;
    Do_Read() ;
    if( strlen( Kb_Buffer ) > 0 ) /* Something to search for */
    {
	Edit( Search_Text, Kb_Buffer, 32 ) ; /* Get search text */
	Do_Search() ;
    } ;
} ;

static void Teu_Undelete_Line() /* Undelete previously deleted line */
{
    if( Deleted_Line )
    {
	Insert_Chars( Line_Buffer ) ; /* A line to undelete */
	strcpy( Character, "\r" ) ;
	Insert_Chars( Character ) ;
    } ;
} ;

static void Teu_Insert_Line() /* Insert blank line */
{
    strcpy( Character, "\r" ) ;
    Insert_Chars( Character ) ; /* Create new line */
    Cursor_Left() ; /* Return to previous position */
} ;

static void Teu_Change_Case()
{
    if( Physical_Column > strlen( Physical_Text ) )
    /* Ignore if past end of line or at [EOB] */
    {
	return ;
    } ;
    if( Physical_Text[ Physical_Column ] >= 'A' &&
	Physical_Text[ Physical_Column ] <= 'Z' )
    {
	Physical_Text[ Physical_Column ] =
	    Physical_Text[ Physical_Column ] + 32 ;
    }
    else
    {
	if( Physical_Text[ Physical_Column ] >= 'a' &&
	    Physical_Text[ Physical_Column ] <= 'z' )
	{
	    Physical_Text[ Physical_Column ] =
		Physical_Text[ Physical_Column ] - 32 ;
	}
	else
	{
	    Cursor_Right() ; /* Skip non-alpha characters */
	    return ;
	} ;
    } ;
    Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
    Window_Output( &Physical_Text[ Physical_Column - 1 ] ) ;
    Update_Window( Screen_Window ) ;
    Flag_Changes() ;
    Cursor_Right() ;
} ;


static void Teu_Special_Insert() /* Insert a character by ASCII value */
{
    Character[ 1 ] = 0 ;
    Character[ 0 ] = Iterate ;
    Iterate = 0 ;
    Insert_Chars( Character ) ; /* Insert special character */
} ;

static void Teu_Bottom()
{
    Do_Update() ;
    if( Screen_Line + 21 < Last_Line ) /* [EOB] not on screen */
    {
	Screen_Line = Last_Line - 21 ;
	Cursor_Row = ( Last_Line - Screen_Line ) + 2 ;
	Draw = True ;
    }
    else
    {
	Draw = False ;
    } ;
    Cursor_Column = 1 ;
    Cursor_Row = ( Last_Line - Screen_Line ) + 2 ;
    Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
    Current_Line = NULL ;
    Physical_Text[ 0 ] = 0 ;
    Logical_Text[ 0 ] = 0 ;
    Logical_Column = 1 ;
    if( Selection || Draw )
    {
	Redraw() ;
    } ;
} ;

static void Teu_Top() /* Go to top of buffer */
{
    Do_Update() ;
    Cursor_Row = 1 ;
    Cursor_Column = 1 ;
    Set_Window( Screen_Window, 1, 1, 0, 0 ) ; /* Move to top */
    if( Screen_Line > 1 )
    {
	Screen_Line = 1 ;
	Draw = True ;
    }
    else
    {
	Draw = False ;
    } ;
    Current_Line = File_Root ;
    strcpy( Physical_Text, Current_Line->Text ) ;
    Format( Physical_Text ) ;
    strcpy( Logical_Text, Logical ) ;
    Logical_Column = 1 ;
    if( Selection || Draw )
    {
	Redraw() ;
    } ;
} ;

static void Teu_Paste() /* Insert text from paste buffer */
{

int Flag ; /* True if cursor should remain at [EOB] */
struct Line_Record *F_Pointer, *L_Pointer, *N_Pointer, *T_Pointer ;
char Temp[ 256 ] ;

    if( Paste_Root == NULL )
    {
	return ; /* Nothing in paste buffer */
    } ;
    if( Paste_Root->Next_Line == NULL ) /* No implicit CRLF */
    {
	if( strlen( Physical_Text ) + strlen( Paste_Root->Text ) > 255 )
	{
	    Show_Error( "\aLine exceeds 255 characters" ) ;
	    return ;
	} ;
	strcpy( Temp, Left( Physical_Text, Physical_Column - 1 ) ) ;
	strcat( Temp, Paste_Root->Text ) ;
	strcat( Temp, &Physical_Text[ Physical_Column - 1 ] ) ;
	strcpy( Physical_Text, Temp ) ;
	Physical_Column = Physical_Column + strlen( Paste_Root->Text ) ;
	Format( Physical_Text ) ;
	strcpy( Logical_Text, Logical ) ;
	strcpy( Character, Left( Physical_Text, Physical_Column - 1 ) ) ;
	Format( Character ) ;
	Cursor_Column = strlen( Logical ) + 1 ;
	Flag_Changes() ;
	Window_Output( "\r" ) ;
	Window_Output( Logical_Text ) ;
	Update_Window( Screen_Window ) ;
	Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
	return ;
    } ;
    Do_Update() ;
    T_Pointer = Paste_Root ;
    F_Pointer = NULL ;
    L_Pointer = NULL ;
    A = -1 ; /* Initialize for count of inserted lines */
    if( Current_Line == NULL ) /* At [EOB] */
    {
	A = 0 ; /* One more line */
    } ;
    while( T_Pointer != NULL ) /* Make copy of paste buffer */
    {
	N_Pointer = malloc( strlen( T_Pointer->Text ) + 9 ) ;
	/* Create next line */
	if( N_Pointer == NULL ) /* Ran out of memory */
	{
	    Show_Error( "\aMaximum memory exceeded" ) ;
	    End_Edit() ;
	    exit( EXIT_FAILURE ) ;
	} ;
	if( F_Pointer == NULL ) /* If first line */
	{
	    F_Pointer = N_Pointer ;
	} ;
	strcpy( N_Pointer->Text, T_Pointer->Text ) ; /* Copy next line */
	T_Pointer = T_Pointer->Next_Line ; /* Go to next line of past buffer */
	N_Pointer->Last_Line = L_Pointer ;
	if( L_Pointer != NULL )
	{
	    L_Pointer->Next_Line = N_Pointer ;
	} ;
	L_Pointer = N_Pointer ; /* Save this pointer for later linking */
	A = A + 1 ; /* Increment count of new lines */
    } ;
    if( Current_Line == NULL ) /* At [EOB] */
    {
	if( strlen( N_Pointer->Text ) == 0 ) /* If last line is just CRLF */
	{
	    A = A - 1 ;
	    L_Pointer = N_Pointer ;
	    N_Pointer = N_Pointer->Last_Line ;
	    free( L_Pointer ) ;
	    Flag = True ;
	}
	else
	{
	    Cursor_Row = Cursor_Row - 1 ;
	    Flag = False ;
	} ;
    } ;
    N_Pointer->Next_Line = NULL ;
    Last_Line = Last_Line + A ; /* Last line in buffer */
    Cursor_Row = Cursor_Row + A ; /* Jump down to bottom of pasted text */
    if( Cursor_Row > 16 && Last_Line > 22 ) /* Must adjust screen */
    {
	Dummy = Cursor_Row - 16 ; /* Number of lines to move up */
	while( Screen_Line + 22 - Dummy > Last_Line )
	/* [EOB] should be at bottom */
	{
	    Dummy = Dummy - 1 ;
	} ;
	while( Screen_Line + Dummy + 21 > Last_Line )
	{
	    Dummy = Dummy - 1 ;
	} ;
	Cursor_Row = Cursor_Row - Dummy ;
	Last_Screen_Line = 0 ; /* Force redraw */
	Screen_Line = Screen_Line + Dummy ;
    } ;
    /* At this point, F_Pointer is the beginning of the paste buffer copy
     and N_Pointer is the end of the paste buffer copy */
    if( File_Root == NULL ) /* Nothing in main buffer */
    {
	File_Root = F_Pointer ;
	File_Last = N_Pointer ;
	Current_Line->Last_Line = File_Last ;
    }
    else /* Something in file */
    {
	if( Current_Line == NULL ) /* At [EOB] */
	{
	    File_Last->Next_Line = F_Pointer ;
	    F_Pointer->Last_Line = File_Last ;
	    File_Last = N_Pointer ;
	    if( !Flag )
	    {
		Current_Line = File_Last ;
		strcpy( Physical_Text, Current_Line->Text ) ;
		Format( Physical_Text ) ;
		strcpy( Logical_Text, Logical ) ;
		Cursor_Column = strlen( Logical ) ;
		Logical_Column = Cursor_Column ;
		Physical_Column = strlen( Physical_Text ) ;
	    } ;
	}
	else
	{
	    if( Physical_Column + strlen( Paste_Root->Text ) > 256 )
	    {
		Show_Error( "\aLine exceeds 255 characters" ) ;
		return ;
	    } ;
	    if( strlen( N_Pointer->Text ) + strlen( Physical_Text ) + 1 -
		Physical_Column > 255 )
	    {
		Show_Error( "\aLine exceeds 255 characters" ) ;
		return ;
	    } ;
	    strcpy( Character, &Physical_Text[ Physical_Column - 1 ] ) ;
	    /* Text following cursor position */
	    strcpy( Temp, Left( Physical_Text, Physical_Column - 1 ) ) ;
	    strcat( Temp, Paste_Root->Text ) ; /* Recreate first line */
	    strcpy( Physical_Text, Temp ) ;
	    Flag_Changes() ;
	    Do_Update() ;
	    N_Pointer->Next_Line = Current_Line->Next_Line ;
	    /* Save pointer to line after pasted text */
	    N_Pointer->Next_Line->Last_Line = N_Pointer ;
	    Current_Line->Next_Line = F_Pointer->Next_Line ;
	    Current_Line->Next_Line->Last_Line = Current_Line ;
	    free( F_Pointer ) ;

	    if( Current_Line == File_Last )
	    {
		File_Last = N_Pointer ;
	    } ;
	    Current_Line = N_Pointer ; /* Go to line after paste buffer */
	    strcpy( Physical_Text, Current_Line->Text ) ; /* Get text */
	    Format( Physical_Text ) ;
	    strcpy( Logical_Text, Logical ) ;
	    Cursor_Column = strlen( Physical_Text ) + 1 ;
	    strcpy( Physical_Text, N_Pointer->Text ) ;
	    strcat( Physical_Text, Character ) ;
	    /* Recreate last line */
	    Flag_Changes() ;
	    Do_Update() ; /* Change line */
	} /* if( Current_Line == NULL ) */
    } ; /* if( File_Root == NULL ) */
    Set_Window( Screen_Window, Cursor_Row, Cursor_Column, 0, 0 ) ;
    Redraw() ;
} ;

static void Do_Change()
{

int A ; /* Dummy variable */
int Gold ;
short int Ord_C ;
char Temp[ 256 ] ;

    Gold = False ;
Prompt:
    if( Status && /* Update status window */ !Error_Flag ) /* No errors */
    {
	if( !Window_Mapped( Status_Window ) )
	{
	    Map_Window( Status_Window, 24, 1, 1, 80, 1, 1, 129 ) ;
	    /* Map window if not mapped */
	} ;
	Select_Window( Status_Window ) ;
	sprintf( Temp, "Line: %10.10li Column: %3.3li\r",
	    Screen_Line + Cursor_Row - 1, Cursor_Column ) ; /* Show position */
	Window_Output( Temp ) ;
	Update_Window( Status_Window ) ;
	Select_Window( Screen_Window ) ;
    } ;
    if( Cursor_Column > 79 )
    {
	Set_Cursor( Cursor_Row, 81 ) ;
    }
    else
    {
	Set_Cursor( Cursor_Row, Cursor_Column + 1 ) ;
    } ;
    strcpy( Character, In_Key() ) ; /* Get next character */
After_Prompt:
    if( Cursor_Column > 79 )
    {
	Clear_Cursor( Cursor_Row, 81 ) ;
    }
    else
    {
	Clear_Cursor( Cursor_Row, Cursor_Column + 1 ) ;
    } ;
    Select_Window( Error_Window ) ;
    Window_Output( "\r\n" ) ; /* Clear any previous error messages */
    Error_Flag = False ; /* No more error */
    Select_Window( Screen_Window ) ;
    Ord_C = Character[ 0 ] ;
    if( strcmp( Character, "PF2" ) == 0 || strcmp( Character, "F2" ) == 0 ||
	strcmp( Character, "F15" ) == 0 )
    /* Help */
    {
	Execute_Command( "HELP" ) ;
	goto Prompt ;
    } ;
    if( !Gold )
    {
	if( strcmp( Character, "KP0" ) == 0 ) /* Advance a line */
	{
	    Teu_Advance() ;
	    goto Prompt ;
	} ;
	if( strcmp( Character, "KP1" ) == 0 ) /* Move by word */
	{
	    Teu_Move_Word() ;
	    goto Prompt ;
	} ;
	if( strcmp( Character, "KP2" ) == 0 || strcmp( Character, "F10" ) == 0 )
	/* Go to end of line */
	{
	    Teu_Eol() ;
	    goto Prompt ;
	} ;
	if( strcmp( Character, "KP6" ) == 0 || strcmp( Character, "CKP6" ) == 0 )
	/* Cut */
	{
	    Teu_Cut() ;
	    goto Prompt ;
	} ;
	if( strcmp( Character, "KP7" ) == 0 ) /* Advance one page */
	{
	    Teu_Page() ;
	    goto Prompt ;
	} ;
	if( strcmp( Character, "KP8" ) == 0 || strcmp( Character, "F7" ) == 0 )
	/* Page (half-page) */
	{
	    Teu_Half_Page() ;
	    goto Prompt ;
	} ;
	if( strcmp( Character, "KP." ) == 0 || strcmp( Character, "F8" ) == 0 )
	/* Select */
	{
	    Teu_Select() ;
	    goto Prompt ;
	} ;
	if( strcmp( Character, "KP," ) == 0 || strcmp( Character, "F6" ) == 0 )
	/* Delete current character */
	{
	    if( Current_Line == NULL )
	    {
		goto Prompt ; /* Do nothing if at [EOB] */
	    } ;
	    Cursor_Right() ;
	    Delete_Char( True ) ;
	    goto Prompt ;
	} ;
	if( strcmp( Character, "KP-" ) == 0 || strcmp( Character, "F9" ) == 0 )
	/* Delete word */
	{
	    Teu_Delete_Word() ;
	    goto Prompt ;
	} ;
	if( strcmp( Character, "PF3" ) == 0 || strcmp( Character, "F3" ) == 0 )
	/* Search for previous text */
	{
	    Do_Search() ;
	    goto Prompt ;
	} ;
	if( strcmp( Character, "PF4" ) == 0 || strcmp( Character, "F4" ) == 0 )
	/* Delete line */
	{
	    Teu_Delete_Line() ;
	    goto Prompt ;
	} ; /* if( strcmp( Character, "PF4" ) == 0 || strcmp( Character, "F4" ) == 0 ) */
    } ;

    /* Special characters after this point are gold-X */
    if( strcmp( Character, "KP." ) == 0 || strcmp( Character, "F8" ) == 0 )
    /* Unselect */
    {
	Gold = False ;
	Teu_Unselect() ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "KP," ) == 0 || strcmp( Character, "F6" ) == 0 )
    /* Undelete last deleted character */
    {
	Gold = False ;
	if( Deleted_Character )
	{
	    Temp[ 0 ] = Character_Buffer ;
	    Temp[ 1 ] = 0 ;
	    Insert_Chars( Temp ) ; /* Restore character */
	} ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "KP-" ) == 0 || strcmp( Character, "F9" ) == 0 )
    /* Undelete a word */
    {
	Gold = False ;
	if( Deleted_Word )
	{
	    Insert_Chars( Word_Buffer ) ; /* A word to undelete */
	} ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "PF3" ) == 0 || strcmp( Character, "F3" ) == 0 )
    /* Search */
    {
	Gold = False ;
	Teu_Search() ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "PF4" ) == 0 || strcmp( Character, "F4" ) == 0 )
    /* Undelete a line */
    {
	Gold = False ;
	Teu_Undelete_Line() ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "KP0" ) == 0 ) /* Insert line */
    {
	Gold = False ;
	Teu_Insert_Line() ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "KP1" ) == 0 ) /* Change case */
    {
	Gold = False ;
	Teu_Change_Case() ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "KP3" ) == 0 ) /* Special insert */
    {
	Gold = False ;
	Teu_Special_Insert() ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "KP4" ) == 0 || strcmp( Character, "CKP4" ) == 0 )
    /* Bottom of buffer */
    {
	Gold = False ;
	Teu_Bottom() ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "KP5" ) == 0 || strcmp( Character, "F5" ) == 0 )
    /* Top of buffer */
    {
	Gold = False ;
	Teu_Top() ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "KP6" ) == 0 || strcmp( Character, "CKP6" ) == 0 )
    /* Paste */
    {
	Gold = False ;
	Teu_Paste() ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "KP7" ) == 0 || strcmp( Character, "F16" ) == 0 )
    /* Command */
    {
	Gold = False ;
	Do_Update() ;
	Select_Window( Error_Window ) ;
	Unmap_Window( Status_Window ) ;
	Window_Output( "\r\n" ) ; /* Clear any previous error messages */
	Window_Output( "Command: " ) ; /* Display prompt */
	Update_Window( Error_Window ) ;
	Do_Read() ; /* Get editing command */
	Edit( Command, Kb_Buffer, 509 ) ; /* Format it */
	Select_Window( Screen_Window ) ;
	Execute_Command( Command ) ;
	return ;
    } ;
    if( strcmp( Character, "UP" ) == 0 )
    {
	Gold = False ;
	Cursor_Up() ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "DOWN" ) == 0 )
    {
	Gold = False ;
	Cursor_Down() ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "LEFT" ) == 0 )
    {
	Gold = False ;
	Cursor_Left() ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "RIGHT" ) == 0 )
    {
	Gold = False ;
	Cursor_Right() ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "PF1" ) == 0 || strcmp( Character, "F1" ) == 0 )
    /* Gold */
    {
	Gold = True ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "ENTER" ) == 0 )
    {
	Gold = False ;
	goto Prompt ; /* Ignore ENTER */
    } ;
    if( Gold && strcmp( Character, "0" ) >= 0 && strcmp( Character, "9" ) <= 0 )
    /* GOLD number */
    {
	Gold = False ;
	Select_Window( Error_Window ) ;
	strcpy( Kb_Buffer, Character ) ;
	Window_Output( Character ) ; /* Show digit */
	while( strcmp( Character, "0" ) >= 0 && strcmp( Character, "9" ) <= 0 )
	/* Until non-numeric entered */
	{
	    Update_Window( Error_Window ) ; /* show changes */
	    strcpy( Character, In_Key() ) ;
	    if( strcmp( Character, "0" ) >= 0 && strcmp( Character, "9" ) <= 0 )
	    /* A digit */
	    {
		Window_Output( Character ) ;
		strcat( Kb_Buffer, Character ) ;/* Append digit */
	    }
	    else
	    {
		if( Character == 127 || Character == '\b' ) /* Delete */
		{
		    if( strlen( Kb_Buffer ) == 0 ) /* Nothing left in buffer */
		    {
			printf( "\a" ) ; /* Warning */
		    }
		    else
		    {
			Window_Output( "\b \b" ) ;
			/* Erase last digit */
			strcpy( Kb_Buffer,
			    Left( Kb_Buffer, strlen( Kb_Buffer ) - 1 ) ) ;
			/* And remove it from the buffer */
		    } ;
		    strcpy( Character, "0" ) ;
		    /* Fake digit so loop won't end */
		} ; /* if Delete */
	    } ;
	} ;
	Select_Window( Screen_Window ) ;
	if( strlen( Kb_Buffer ) == 0 )
	{
	    goto After_Prompt ;
	} ;
	Iterate = atoi( Kb_Buffer ) ;
	if( Iterate == 0 ) /* Some error */
	{
	    printf( "\a" ) ;
	} ;
	goto After_Prompt ;
    } ;
    if( strlen( Character ) > 1 ) /* Unknown special key */
    {
	Gold = False ;
	printf( "\a" ) ;
	goto Prompt ;
    } ;
    if( Ord_C == 3 ) /* Control-C */
    {
	Show_Error( "\aControl-C ignored" ) ;
	Iterate = 0 ;
	goto Prompt ;
    } ;
    if( strcmp( Character, "" ) == 0 ) /* Status */
    {
	Iterate = 0 ;
	Select_Window( Error_Window ) ;
	Window_Output( "\r\n" ) ;
	sprintf( Temp, "TEU V%s: File: %s, %li bytes available, %li lines",
	    Version, Input_Name, coreleft(), Last_Line ) ;
	Window_Output( Temp ) ; 
	Update_Window( Error_Window ) ;
	Select_Window( Screen_Window ) ;
	goto Prompt ;
    } ;
    if( Ord_C == 26 ) /* Control-Z */
    {
	Do_Update() ;
	Execute_Command( "EXIT" ) ;
	return ;
    } ;

    /* Anything that follows will change the current line */
    if( Ord_C == 127 || Ord_C == 8 ) /* Delete */
    {
	Delete_Char( True ) ;
	goto Prompt ;
    } ; /* Delete */

    if( Auto_Indent && strcpy( Character, "\r" ) == 0 )
    /* Auto-indent is active */
    {
	Dummy = 1 ;
	if( Physical_Text[ Dummy - 1 ] == ' ' ||
	    Physical_Text[ Dummy - 1 ] == '\t' )
	    /* If line begins with space or tab */
	{
	    while( ( Physical_Text[ Dummy - 1 ] == ' ' ||
		Physical_Text[ Dummy - 1 ] == '\t' ) &&
		Dummy <= strlen( Physical_Text ) )
	    {
		Dummy ++ ; /* Find first non-space/non-tab */
	    } ;
	} ;
	strcpy( Command0, Left( Physical_Text, Dummy - 1 ) ) ;
	/* Get space/tab portion */
    }
    else
    {
	Command0[ 0 ] = 0 ;
    } ;
    if( Iterate > 0 ) /* Iterative insert */
    {
	while( Iterate > 0 )
	{
	    if( strcmp( Character, "\r" ) != 0 )
	    /* Must insert CRs one at a time */
	    {
		if( Iterate > 255 )
		{
		    memset( Command, Character[ 0 ], 255 ) ;
		    Command[ 255 ] = 0 ;
		    Iterate = Iterate - 255 ;
		}
		else
		{
		    memset( Command, Character[ 0 ], Iterate ) ;
		    Command[ Iterate ] = 0 ;
		    Iterate = 0 ; /* Done */
		} ;
		Insert_Chars( Command ) ;
	    }
	    else
	    {
		Iterate = Iterate - 1 ;
		Insert_Chars( Character ) ; /* Insert another */
	    } ;
	} ; /* while( Iterate > 0 ) */
    }
    else
    {
	Insert_Chars( Character ) ; /* Insert a character */
    } ; /* if(  ) */
    if( Auto_Indent && strlen( Command0 ) > 0 )
    {
	Insert_Chars( Command0 ) ; /* Do auto-indent, if needed */
    } ;
    goto Prompt ;
} ;

static void Teu_Setup()
{
    strcpy( Help_File_Name, Filnam( "TEU.HLP" ) ) ;
    Any_Changes = False ; /* No changesd yet */
    Journal_Offset = -1 ;
    Journal_Record = 0 ;
    Screen_Line = 1 ; /* Line in file that corresponds to screen line 1 */
    Last_Screen_Line = 0 ; /* Force redraw on initial display */
    Cursor_Row = 1 ; /* Cursor position on screen */
    Cursor_Column = 1 ; /* Cursor column on line */
    Physical_Column = 1 ;
    Logical_Column = 1 ;
    Changes = False ;
    Deleted_Character = False ;
    Deleted_Line = False ;
    Search_Text[ 0 ] = 0 ;
    Selection = False ;
    Paste_Root = NULL ;
    Journal_Size = 63 ;
    Iterate = 0 ;
    Status = False ;
    Error_Flag = False ;
    Auto_Indent = False ;
    Init_File = fopen( Filnam( "TEU.INI" ), "rt" ) ;
    /* Find and open initialization file */
    if( Init_File != NULL ) /* If success */
    {
	while( fgets( Character, 255, Init_File ) != NULL )
	{
	    if( Character[ 0 ] == 'S' )
	    {
		Execute_Command( Character ) ;
		/* Do the command */
	    } ;
	} ;
	fclose( Init_File ) ; /* All done */
    } ;
    memset( Blank, 32, 80 ) ; /* 80 spaces - to blank lines */
    Blank[ 80 ] = 0 ;
    Init_Smu( 0, 0, 0, 0 ) ; /* Initialize for windowing */
    Screen_Window = Create_Window( 23, 80, 4 ) ; /* Screen */
    Error_Window = Create_Window( 1, 80, 4 ) ; /* Error line */
    Status_Window = Create_Window( 1, 80, 4 ) ; /* Status line */
    Map_Window( Screen_Window, 1, 23, 1, 80, 1, 1, 128 ) ;
    Map_Window( Error_Window, 24, 1, 1, 80, 1, 1, 128 ) ;
} ;

void Teu_Teu( char *Passed_Command ) /* Main routine */
{

char Command[ 256 ], Temp[ 256 ] ;

    if( Passed_Command == NULL )
    {
	Command[ 0 ] = 0 ;
    }
    else
    {
	strcpy( Command, Passed_Command ) ;
    } ;
    if( strlen( Command ) > 0 )
    {
	Ccl = True ;
    }
    else
    {
	Ccl = False ;
    } ;
    New_File = True ;
    Teu_Setup() ; /* Setup for processing */
    Select_Window( Screen_Window ) ;
    Set_Window( Screen_Window, 23, 1, 0, 0 ) ;
    if( Ccl )
    {
	goto Start ;
    } ;
Main_Prompt:
    strcpy( Command, Filnam( "TEU.LFE" ) ) ;
    /* Determine name of last file edited */
    Input_File = fopen( Command, "rt" ) ;
    if( Input_File != NULL ) /* File is there */
    {
	fgets( Command, 255, Input_File ) ;
	fclose( Input_File ) ;
	Select_Window( Error_Window ) ;
	sprintf( Temp, "Editing file: %s.  Confirm <YES> : ", Command ) ;
	Window_Output( Temp ) ;
	Update_Window( Error_Window ) ;
	Do_Read() ;
	Edit( Kb_Buffer, Kb_Buffer, -1 ) ;
	if( Kb_Buffer[ 0 ] == 'Y' || strlen( Kb_Buffer ) == 0 )
	{
	    Select_Window( Screen_Window ) ;
	    New_File = False ;
	    goto Start ;
	} ;
	Window_Output( "\r\n" ) ; /* Clear status line */
	Select_Window( Screen_Window ) ;
    } ;
    Window_Output( "TEU> " ) ; /* Prompt user */
    Update_Window( Screen_Window ) ;
    Window_Read() ; /* Get command line */
    strcpy( Command, Kb_Buffer ) ;
    Edit( Temp, Command, -1 ) ;
    if( strlen( Temp ) == 0 )
    {
	goto Main_Prompt ; /* Ignore blank lines */
    } ;
    if( Kb_Buffer[ 0 ] == 3 || Kb_Buffer[ 0 ] == 25 ||
	Kb_Buffer[ 0 ] == 26 )
    {
	End_Edit() ;
	exit( EXIT_SUCCESS ) ; /* Abort on control-C/Y/Z */
    } ;
Start:
    Edit( Input_Name, Command, 37 ) ; /* Get name of file to edit */
    A = Instr( 1, Input_Name, "/R" ) ; /* /RECOVER? */
    if( A > 0 ) /* Doing /RECOVER */
    {
	strcpy( Input_Name, Left( Input_Name, A - 1 ) ) ;
	Recover = True ;
    }
    else
    {
	Recover = False ;
    } ;
    Dummy = Extension_Pos( Command ) ; /* Find input file extension */
    if( Dummy == 0 )
    {
	Dummy = strlen( Command ) + 1 ;
    } ;
    strcpy( Journal_Name, Left( Command, Dummy - 1 ) ) ;
    strcat( Journal_Name, ".JOU" ) ; /* Build journal file name */
    if( New_File )
    {
	Input_File = fopen( "TEU.LFE", "wt" ) ;
	if( Input_File != NULL ) /* No problems */
	{
	    fprintf( Input_File, "%s\n", Input_Name ) ;
	    /* Write edited filename */
	    Update_Window( Screen_Window ) ;
	    fclose( Input_File ) ;
	} ;
    } ;
    New_File = False ; /* Assume existing file */
    Input_File = fopen( Input_Name, "rt" ) ; /* Open it */
    if( Input_File == NULL ) /* Some error */
    {
	if( errno == ENOENT || errno == ENOFILE || errno == ENOPATH )
	/* File does not exist */
	{
	    Input_File = fopen( Input_Name, "w+t" ) ;
	    /* Create file - it doesn't exist */
	    if( Input_File == NULL ) /* Error on creation */
	    {
		strcpy( Command, strerror( errno ) ) ;
		strcat( Command, " - creating " ) ;
		strcat( Command, Input_Name ) ;
		strcat( Command, "  " ) ;
		Select_Window( Error_Window ) ;
		Window_Output( Command ) ;
		Update_Window( Error_Window ) ;
		Select_Window( Screen_Window ) ;
		goto Main_Prompt ;
	    } ;
	    New_File = True ; /* No existing file */
	    Show_Error( "\aInput file does not exist" ) ;
	    Last_Line = 0 ;
	    File_Root = NULL ;
	    File_Last = NULL ;
	} ;
    }
    else /* No error */
    {
	Include_File( Input_Name, &Last_Line, &File_Root, &File_Last ) ;
	if( Last_Line < 0 ) /* Out of memory */
	{
	    End_Edit() ;
	    exit( EXIT_FAILURE ) ;
	}
    } ;
    fclose( Input_File ) ;
    if( Recover ) /* If /RECOVER */
    {
	Journal = fopen( Journal_Name, "rb" ) ; /* Open journal file */
	if( Journal == NULL )
	{
	    Select_Window( Error_Window ) ;
	    sprintf( Temp,
		"\a%s - while opening journal file: %s.  Can\'t recover.",
		strerror( errno ), Journal_Name ) ;
	    Window_Output( Temp ) ;
	    Update_Window( Error_Window ) ;
	    Select_Window( Screen_Window ) ;
	    Recover = False ;
	} ;
	if( fread( Journal_Buffer, 1, 256, Journal ) < 256 )
	{
	    Recover = False ; /* Some error on read */
	    fclose( Journal ) ;
	} ;
    } ;
    if( !Recover )
    {
	Journal = fopen( Journal_Name, "w+b" ) ; /* Create journal file */
    } ;
    Current_Line = File_Root ;
    if( Current_Line != NULL )
    {
	strcpy( Physical_Text, Current_Line->Text ) ;
    }
    else
    {
	Physical_Text[ 0 ] = 0 ;
    } ;
    Format( Physical_Text ) ;
    strcpy( Logical_Text, Logical ) ;
    Redraw() ; /* Redraw screen */

Prompt:
    Do_Change() ;
    goto Prompt ; /* Loop */
} ;
