library CEF_Util ;

uses // Borland...
     Windows, // HPalette
     Classes, // TList
     Controls, // TControl
     Extctrls, // TPanel
     Forms, // Application
     Graphics, // TBitmap
     StdCtrls, // TComboBox
     SysUtils, // inttostr

     // C&C...
     CommonUt, // Edit
     CVT, // CVTB
     HTML, // TXML_Parser
     Standard, // Extension_Pos
     _UEHDefs, // TUEC

     // CEF...
     _CEFUtil, // TCEF_Watchpoint
     Keymap_Form ; // TKey_Mapper_Dialog

{$R *.res}

type TStandard_Watchpoint = class( TCEF_Watchpoint )
                                private // Instance data...
                                    _Address : int64 ;
                                    _Access : longint ;
                                    Temp_Serialize : string ;

                                public { API... }
                                    function Get_Address : int64 ; override ;
                                    function Get_Access : longint ; override ;
                                    procedure Set_Access( Value : longint ) ;
                                        override ;

                                    function Serialize : PChar ; override ;
                                    procedure Deserialize( Image : PChar ) ;
                                        override ;
                            end ;

    TStandard_Watchpoint_Manager = class( TCEF_Watchpoint_Manager )
                                       private // Instance data...
                                           _Watchpoints : TList ;
                                           Temp_Serialize : string ;

                                       protected // Utility functions...
                                           function Watchpoints : TList ;

                                       public // API...
                                           procedure Clear ; override ;
                                           function Clear_Watchpoint( Address : int64 ;
                                               Access : longint ) : TUEC ;
                                               override ;
                                           function Create_Watchpoint( Ad : int64 ;
                                               Ac : longint ) : TCEF_Watchpoint ;
                                               override ;
                                           function Count : longint ;
                                               override ;
                                           procedure Deserialize( Source : PChar ) ;
                                               override ;
                                           function Serialize : PChar ;
                                               override ;
                                           procedure Terminate ;
                                               override ;
                                           function Watchpoint_At( Address : int64 ) : TCEF_Watchpoint ;
                                               override ;
                                   end ;

     TKey_Mapper = class( TCEF_Key_Mapper )
                       public // Constructor
                           constructor Create ;

                       private // Instance data..
                           Keys : TStringList ; // WIndows keys that can be mapped
                           _Mapping : TStringList ; // Current mappings for Keys
                           Mappings : TStringList ; // Valid mappings
                           Temp_Mapping : string ;

                       public // API...
                           procedure Add_Key( Name : PChar ) ; override ;
                           procedure Add_Mapping( Name : PChar ) ; override ;
                           procedure Clear_Keys( Mapping_Only : boolean ) ;
                               override ;
                           procedure Clear_Mappings ; override ;
                           function Query : boolean ; override ;
                           function Mapping( Name : PChar ) : PChar ; override ;
                           procedure Set_Mapping( Key, Mapping : PChar ) ;
                               override ;
                           procedure Terminate ; override ;
                   end ;

// TStandard_Watchpoint methods...

// API...

function TStandard_Watchpoint.Get_Address : int64 ;

begin
    Result := _Address ;
end ;


function TStandard_Watchpoint.Get_Access : longint ;

begin
    Result := _Access ;
end ;


procedure TStandard_Watchpoint.Set_Access( Value : longint ) ;

begin
    _Access := Value ;
end ;


function TStandard_Watchpoint.Serialize : PChar ;

begin
    Temp_Serialize := '|' + inttostr( Address ) + '|' + inttostr( Access ) ;
    Result := PChar( Temp_Serialize ) ;
end ;


procedure TStandard_Watchpoint.Deserialize( Image : PChar ) ;

var Dummy : integer ;
    S : string ;

begin
    S := string( Image ) ;
    S := copy( S, 2, length( S ) ) ;
    Dummy := pos( '|', S ) ;
    _Address := strtoint64( copy( S, 1, Dummy - 1 ) ) ;
    Access := strtoint( copy( S, Dummy + 1, length( S ) ) ) ;
end ;



// TStandard_Watchpoint_Manager methods...

// Utility functions...

function TStandard_Watchpoint_Manager.Watchpoints : TList ;

begin
    if( _Watchpoints = nil ) then
    begin
        _Watchpoints := TList.Create ;
    end ;
    Result := _Watchpoints ;
end ;


// API...

procedure TStandard_Watchpoint_Manager.Clear ;

var Loop : integer ;

begin
    for Loop := Watchpoints.Count - 1 downto 0 do
    begin
        TStandard_Watchpoint( Watchpoints[ Loop ] ).Free ;
    end ;
    Watchpoints.Clear ;
end ;


function TStandard_Watchpoint_Manager.Clear_Watchpoint( Address : int64 ;
    Access : longint ) : TUEC ;

var Index : integer ;
    Watchpoint : TStandard_Watchpoint ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
    for Index := 0 to Watchpoints.Count - 1 do
    begin
        Watchpoint := TStandard_Watchpoint( Watchpoints[ Index ] ) ;
        if( Watchpoint.Address = Address ) then
        begin
            Watchpoint.Access := Watchpoint.Access and not( Access ) ;
            if( Watchpoint.Access = 0 ) then // No access class remaining
            begin
                Watchpoint.Free ;
                Watchpoints.Delete( Index ) ;
            end ;
            exit ;
        end ;
    end ;
end ;


function TStandard_Watchpoint_Manager.Create_Watchpoint( Ad : int64 ;
    Ac : longint ) : TCEF_Watchpoint ;

var Index : integer ;
    Watchpoint : TStandard_Watchpoint ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
    if( Ac = 0 ) then // No effect
    begin
        exit ;
    end ;
    for Index := 0 to Watchpoints.Count - 1 do
    begin
        Watchpoint := TStandard_Watchpoint( Watchpoints[ Index ] ) ;
        if( Watchpoint.Address = Ad ) then
        begin
            Watchpoint.Access := Watchpoint.Access or Ac ;
            exit ;
        end ;
    end ;
    Result := TStandard_Watchpoint.Create ;
    TStandard_Watchpoint( Result )._Address := Ad ;
    Result.Access := Ac ;
    Watchpoints.Add( self ) ;
end ;


function TStandard_Watchpoint_Manager.Count : longint ;

begin
    Result := Watchpoints.Count ;
end ;


procedure TStandard_Watchpoint_Manager.Deserialize( Source : PChar ) ;

var S : string ;
    Watchpoint : TStandard_Watchpoint ;
    Watchpoint_Parser : TXML_Parser ;

begin
    Watchpoint_Parser := TXML_Parser.Create ;
    try
        Watchpoint_Parser.Set_Source( Source ) ;
        while( not Watchpoint_Parser.EOF ) do
        begin
            S := uppercase( Watchpoint_Parser.Token ) ;
            if( S = '<WATCHPOINT>' ) then
            begin
                Watchpoint := TStandard_Watchpoint.Create ;
                Watchpoint.Deserialize( Watchpoint_Parser.Get_Section( 'watchpoint' ) ) ;
                Watchpoints.Add( Watchpoint ) ;
            end else
            if( ( copy( S, 1, 1 ) = '<' ) and ( copy( S, length( S ) - 1, 2 ) <> '/>' ) ) then
            begin
                Watchpoint_Parser.Get_Section( PChar( S ) ) ;
            end ;
        end ;
    finally
        Watchpoint_Parser.Free ;
    end ;
end ;


function TStandard_Watchpoint_Manager.Serialize : PChar ;

var Loop : integer ;

begin
    Temp_Serialize := '' ;
    for Loop := 0 to Watchpoints.Count - 1 do
    begin
        Temp_Serialize := Temp_Serialize +
            '<watchpoint>' + TStandard_Watchpoint( Watchpoints[ Loop ] ).Serialize + '</watchpoint>' ;
    end ;
    Result := PChar( Temp_Serialize ) ;
end ;


procedure TStandard_Watchpoint_Manager.Terminate ;

begin
    Clear ;
    _Watchpoints.Free ;
    _Watchpoints := nil ;
    Free ;
end ;


function TStandard_Watchpoint_Manager.Watchpoint_At( Address : int64 ) : TCEF_Watchpoint ;

var Loop : integer ;

begin
    for Loop := 0 to Watchpoints.Count - 1 do
    begin
        Result := TCEF_Watchpoint( Watchpoints[ Loop ] ) ;
        if( Result.Address = Address ) then
        begin
            exit ;
        end ;
    end ;
    Result := nil ; { None found }
end ;



type TCharacter_Set = class( TCEF_Character_Set )
                          private // Instance data...
                              Character_Set : array[ 0..255 ] of TBitmap ;
                              _Invert : boolean ;
                              _Height : integer ;
                              _Width : integer ;

                          public // Property handlers...
                              function Get_Invert : boolean ; override ;

                              procedure Set_Invert( Value : boolean ) ; override ;

                          public // API...
                              procedure Clear ; override ;
                              function Draw( Handle, X, Y, Index : longint ) : boolean ;
                                  override ;
                              function Has_Glyph( Index : integer ) : boolean ;
                                  override ;
                              function Height : integer ; override ;
                              procedure Initialize ; override ;
                              procedure Load( Name : PChar ) ; override ;
                              procedure Terminate ; override ;
                              function Width( Index : longint ) : longint ; override ;
                      end ; // TCharacter_Set


// TCharacter_Set methdos...

// Property handlers...

function TCharacter_Set.Get_Invert : boolean ;

begin
    Result := _Invert ;
end ;


procedure TCharacter_Set.Set_Invert( Value : boolean ) ;

begin
    _Invert := Value ;
end ;


// API...

procedure TCharacter_Set.Clear ;

var I : integer ;

begin
    for I := 0 to 255 do
    begin
        Character_Set[ I ].Free ;
        Character_Set[ I ] := nil ;
    end ;
end ;


function TCharacter_Set.Draw( Handle, X, Y, Index : integer ) : boolean ;

    procedure Draw_Bitmap( B : TBitmap ) ;

    var OldPalette : HPalette ;
        RestorePalette : Boolean ;

    begin
        OldPalette := 0 ;
        RestorePalette := False ;

        if( B.Palette <> 0 ) then
        begin
            OldPalette := SelectPalette( Handle, B.Palette, True ) ;
            RealizePalette( Handle ) ;
            RestorePalette := True ;
        end ;
        try
            BitBlt( Handle, X, Y, B.Width, B.Height,
                B.Canvas.Handle, 0, 0, SRCCOPY ) ;
        finally
            if( RestorePalette ) then
            begin
                SelectPalette( Handle, OldPalette, True ) ;
            end ;
        end ;
    end ;

var B : TBitmap ;
    LX, LY : integer ;

begin
    Result := False ;
    if( ( Index < 0 ) or ( Index > 255 ) ) then
    begin
        exit ;
    end ;
    if( ( Index > 127 ) and ( Character_Set[ Index ] = nil ) ) then
    begin
        Index := Index - 128 ;
    end ;
    if( Character_Set[ Index ] = nil ) then
    begin
        exit ;
    end ;

    Result := True ;
    if( _Invert ) then
    begin
        B := TBitmap.Create ;
        try
            B.Assign( Character_Set[ Index ] ) ;
            for LX := 0 to B.Width - 1 do
            begin
                for LY := 0 to B.Height - 1 do
                begin
                    B.Canvas.Pixels[ LX, LY ] := not B.Canvas.Pixels[ LX, LY ] ;
                end ;
            end ;
            Draw_Bitmap( B ) ;
        finally
            B.Free ;
        end ;
    end else
    begin
        Draw_Bitmap( Character_Set[ Index ] ) ;
    end ;
end ; // TCharacter_Set.Draw


function TCharacter_Set.Has_Glyph( Index : integer ) : boolean ;

begin
    Result := False ;
    if( ( Index >= 0 ) and ( Index <= 255 ) and ( Character_Set[ Index ] <> nil ) ) then
    begin
        Result := True ;
    end ;
end ;


function TCharacter_Set.Height : integer ;

begin
    Result := _Height ;
end ;


procedure TCharacter_Set.Initialize ;

begin
    _Height := 1 ;
    _Width := 1 ;
end ;


procedure TCharacter_Set.Load( Name : PChar ) ;

var Bitmap : TBitmap ;
    Dummy : integer ;
    C, Line : integer ;
    F : textfile ;
    First : boolean ;
    N : string ;
    Rows : array[ 0..63 ] of longint ;
    S, Work : string ;
    X, Y : integer ;

begin
    Clear ;
    _Height := 0 ;
    _Width := 8 ;
    N := string( Name ) ;
    if( Extension_Pos( N ) = 0 ) then
    begin
        N := N + '.c' ;
    end ;
    assignfile( F, N ) ;
    {$I-}
    System.reset( F ) ;
    {$I+}
    if( IOResult <> 0 ) then
    begin
        exit ;
    end ;
    try
        C := 0 ; // Character to read
        First := True ;
        while( not eof( F ) ) do
        begin
            // Process next character...
            {$I-}
            readln( F, S ) ;
            {$I+}
            if( IOResult <> 0 ) then
            begin
                break ;
            end ;
            if( First ) then // First line in file...
            begin
                First := False ;
                if( Edit( S, -1 ) = 'RASTER' ) then
                begin
                    {$I-}
                    readln( F, S ) ;
                    {$I+}
                    if( IOResult <> 0 ) then
                    begin
                        break ;
                    end ;
                    try
                        _Width := strtoint( S ) ;
                    except
                    end ;

                    // Get first line of raster data
                    {$I-}
                    readln( F, S ) ;
                    {$I+}
                    if( IOResult <> 0 ) then
                    begin
                        break ;
                    end ;
                end ;
            end ;

            // Process this character
            Line := 0 ;
            fillchar( Rows, sizeof( Rows ), 0 ) ;
            while( ( length( S ) > 0 ) and ( Line < 64 ) ) do
            begin
                Dummy := pos( ';', S + ';' ) ;
                S := copy( S, 1, Dummy - 1 ) ; // Trim comments
                Dummy := pos( ',', S + ',' ) ;
                Work := copy( S, 1, Dummy - 1 ) ;
                S := copy( S, Dummy + 1, length( S ) ) ;
                Work := Edit( Work, -1 ) ;
                try
                    if( copy( Work, 1, 2 ) = '0B' ) then
                    begin
                        Rows[ Line ] := strtoint( CVTB( 2, 10, copy( Work, 3, length( Work ) ) ) ) ;
                    end else
                    if( copy( Work, 1, 2 ) = '0X' ) then
                    begin
                        Rows[ Line ] := strtoint( CVTB( 16, 10, copy( Work, 3, length( Work ) ) ) ) ;
                    end else
                    if( copy( Work, 1, 1 ) = '0' ) then
                    begin
                        Rows[ Line ] := strtoint( CVTB( 8, 10, Work ) ) ;
                    end else
                    if( copy( Work, 1, 1 ) = '$' ) then
                    begin
                        Rows[ Line ] := strtoint( CVTB( 16, 10, Work ) ) ;
                    end else
                    begin
                        Rows[ Line ] := strtoint( Work ) ;
                    end ;
                except
                end ;
                inc( Line ) ;
            end ; // while( ( length( S ) > 0 ) and ( Line < 13 ) )
            _Height := Line ;
            Bitmap := TBitmap.Create ;
            Bitmap.Height := _Height ;
            Bitmap.Width := _Width ;
            for X := 0 to _Width - 1 do
            begin
                for Y := 0 to _Height - 1 do
                begin
                    if( ( Rows[ Y ] and ( 1 shl ( 7 - X ) ) ) <> 0 ) then
                    begin
                        Bitmap.Canvas.Pixels[ X, Y ] := $FFFFFF ; // White
                    end else
                    begin
                        Bitmap.Canvas.Pixels[ X, Y ] := 0 ; // Black
                    end ;
                end ;
            end ;
            Character_Set[ C ] := Bitmap ;
            inc( C ) ;
            if( C > 255 ) then // Too many characters
            begin
                break ;
            end ;
        end ; // while( not eof( F ) )
    finally
        {$I-}
        closefile( F ) ;
        {$I+}
        IOResult ;
    end ;
end ;


procedure TCharacter_Set.Terminate ;

begin
    Clear ;
    Free ;
end ;


function TCharacter_Set.Width( Index : longint ) : longint ;

begin
    Result := _Width ;
end ;



// TKey_Mapper methods...

constructor TKey_Mapper.Create ;

begin
    inherited Create ;

    Keys := TStringList.Create ;
    _Mapping := TStringList.Create ;
    Mappings := TStringList.Create ;
end ;


procedure TKey_Mapper.Add_Key( Name : PChar ) ;

var S : string ;

begin
    S := string( Name ) ;
    if( Keys.Indexof( Name ) = -1 ) then
    begin
        Keys.Add( S ) ;
        _Mapping.Add( '' ) ;
    end ;
end ;


procedure TKey_Mapper.Add_Mapping( Name : PChar ) ;

var S : string ;

begin
    S := string( Name ) ;
    if( Mappings.Indexof( S ) = -1 ) then // Mapping not defined
    begin
        Mappings.Add( S ) ;
    end ;
end ;


procedure TKey_Mapper.Clear_Keys( Mapping_Only : boolean ) ;

var I : integer ;

begin
    if( Mapping_Only ) then
    begin
        for I := 0 to _Mapping.Count - 1 do
        begin
            _Mapping[ I ] := '' ;
        end ;
    end else
    begin
        Keys.Clear ;
        _Mapping.Clear ;
    end ;
end ;


procedure TKey_Mapper.Clear_Mappings ;

begin
    Mappings.Clear ;
end ;


function TKey_Mapper.Query : boolean ;

var CB : TComboBox ;
    Control : TControl ;
    Lab : TLabel ;
    Lab_Width : integer ;
    List : TList ;
    Loop : integer ;
    X, Y : integer ;

begin
    // Sanity check
    if( ( Keys.Count = 0 ) or ( Mappings.Count = 0 ) ) then
    begin
        Result := False ;
        exit ; // Nothing for the user to see
    end ;

    // Create dialog
    if( Key_Mapper_Dialog = nil ) then
    begin
        Key_Mapper_Dialog := TKey_Mapper_Dialog.Create( Application ) ;
    end ;

    // Clear existing dialog...
    for Loop := Key_Mapper_Dialog.ControlCount - 1 downto 0 do
    begin
        Control := Key_Mapper_Dialog.Controls[ Loop ] ;
        if( not ( Control is TPanel ) ) then
        begin
            Control.Free ;
        end ;
    end ;

    // Determine metrics...
    Lab_Width := 0 ;
    for Loop := 0 to Keys.Count - 1 do
    begin
        X := Key_Mapper_Dialog.Canvas.TextWidth( Keys[ Loop ] ) ;
        if( X > Lab_Width ) then
        begin
            Lab_Width := X ;
        end ;
    end ;
    Lab_Width := Lab_Width + 4 ;

    List := TList.Create ;
    try
        // Populate and size dialog...
        X := 8 ;
        Y := 8 ;
        for Loop := 0 to Keys.Count - 1 do
        begin
            if( Loop = Keys.Count div 2 ) then
            begin
                // Start second column...
                X := Lab_Width * 3 ;
                Y := 8 ;
            end ;

            // Create label...
            Lab := TLabel.Create( Key_Mapper_Dialog ) ;
            Lab.Parent := Key_Mapper_Dialog ;
            Lab.Caption := Keys[ Loop ] + ':' ;
            Lab.Top := Y + 4 ;
            Lab.Left := X ;

            // Create combobox...
            CB := TComboBox.Create( Key_Mapper_Dialog ) ;
            CB.Style := csDropDownList ;
            CB.Parent := Key_Mapper_Dialog ;
            CB.Top := Y ;
            CB.Left := X + Lab_Width ;
            CB.Items.Assign( Mappings ) ;
            CB.Width := Lab_Width + 2 + GetSystemMetrics( SM_CYVSCROLL ) ;
            List.Add( CB ) ;

            // Set current mapping
            if( _Mapping[ Loop ] <> '' ) then
            begin
                CB.ItemIndex := CB.Items.Indexof( _Mapping[ Loop ] ) ;
            end ;

            Y := Y + CB.Height + 4 ;
        end ;
        Key_Mapper_Dialog.ClientHeight := Y + Key_Mapper_Dialog.Panel1.Height * 2 ;
        Key_Mapper_Dialog.ClientWidth := CB.Left + CB.Width + 4 ;

        // Query user
        Result := ( Key_Mapper_Dialog.ShowModal = mrOK ) ;

        // Apply user's changes
        if( Result ) then
        begin
            for Loop := 0 to Keys.Count - 1 do
            begin
                _Mapping[ Loop ] := TComboBox( List[ Loop ] ).Text ;
            end ;
        end ;
    finally
        List.Free ;
    end ;
end ;


function TKey_Mapper.Mapping( Name : PChar ) : PChar ;

var Index : integer ;
    S : string ;

begin
    S := string( Name ) ;
    Index := Keys.Indexof( S ) ;
    if( Index = -1 ) then
    begin
        Result := nil ;
    end else
    begin
        Temp_Mapping := _Mapping[ Index ] ;
        Result := PChar( Temp_Mapping ) ;
    end ;
end ;


procedure TKey_Mapper.Set_Mapping( Key, Mapping : PChar ) ;

var _Key, __Mapping : string ;
    Index : integer ;

begin
    _Key := string( Key ) ;
    __Mapping := string( Mapping ) ;
    Index := Keys.IndexOf( _Key ) ;
    if( Index <> -1 ) then
    begin
        _Mapping[ Index ] := __Mapping ;
    end ;
end ;


procedure TKey_Mapper.Terminate ;

begin
    Keys.Free ;
    Keys := nil ;
    Mappings.Free ;
    Mappings := nil ;
    _Mapping.Free ;
    _Mapping := nil ;
    Free ;
end ;



// Unit Exported functions...

function Get_Watchpoint_Manager : TCEF_Watchpoint_Manager ; stdcall ;

begin
    Result := TStandard_Watchpoint_Manager.Create ;
end ;


function Version : longint ; stdcall ;

begin
    Result := 20 ; // V2.0
end ;


function Get_Character_Set : TCEF_Character_Set ; stdcall ;

begin
    Result := TCharacter_Set.Create ;
end ;


function Get_Key_Mapper : TCEF_Key_Mapper ; stdcall ;

begin
    Result := TKey_Mapper.Create ;
end ;


{$WARNINGS OFF}
exports Version index 0,
        Get_Watchpoint_Manager index 1,
        Get_Character_Set index 2,
        Get_Key_Mapper index 3 ;
{$WARNINGS ON}

begin
  Application.Initialize;
end.
