{
              Program Name : Structures
              Package Name : LB
              Purpose      : Data Structure object Library
              Institution  : Conroy & Conroy Co.
              Date Written : 24-Nov-1991
              Written By   : Alan Conroy
              Version      : 2.1

		TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

	  This unit is a collection of data structure objects.

          Certified objects:
                Object              Date
                ------              ----
                Double_List         15-Aug-2004
                TFast_Sparse_Matrix 2-June-2000
                TSparse_Matrix      2-June-2000
                TDList              15-Aug-2004
}

unit Structur ;

interface

{$I EDEFINES.INC}

{$IFDEF DELPHI}
    {$DEFINE DELPHI_OR_WINDOWS}
{$ENDIF}
{$IFDEF WINDOWS}
    {$DEFINE DELPHI_OR_WINDOWS}
{$ENDIF}

uses { Borland... }
{$IFNDEF DELPHI}
     Objects,
{$ENDIF}

     { C&C... }
     Common, { TBase_Object }
     Heaps, { PHeap }
     MStores, { PSManaged_Store }
     _Stores, { TStore_Address }
     TypeDefs ; { Max_Memory }

type  Array_Pointer = ^Array_Image ;
      Array_Image = array[ 0..Max_Memory ] of byte ;

      { 32-bit memory structures... }
      List_Pointer = TStore_Address ;
      List_Image = record
                       Next : TStore_Address ; { Pointer to next item on list }
                       Last : TStore_Address ;
                   end ;

      Tree_Pointer = ^Tree_Image ;
      Tree_Image_Min = record
                           Back : Tree_Pointer ;
                           Left : Tree_Pointer ;
                           Right : Tree_Pointer ;
                       end ;
      Tree_Image = record
                       Back : Tree_Pointer ;
                       Left : Tree_Pointer ;
                       Right : Tree_Pointer ;
                       D : array[ 0..Max_Memory - sizeof( Tree_Pointer ) * 3 ] of byte ; { Data area }
                   end ;
      BTree_Pointer = ^BTree_Image ;
      BTree_Image = record
                        Level : longint ; { Level 0 is data, Level 1 is lowest level of keys }
                        Active : longint ; { Number of active keys in this record (for key record) or size of record (data) }
                        Parent : BTree_Pointer ; { Pointer to parent (nil if top level) }
                        D : array[ 0..Max_Memory - 3 - sizeof( BTree_Pointer ) ] of byte ;
                        { Data area.  In a data record (Level = 0) this is the
                          image of the record's data.  In a key record (Level >
                          0) this is an array of BTree_Pointers (0 to Max_B - 1)
                          pointing to the next level records. }
                    end ;
      TSparse_Data_Header = record { Sparse array header (points to other sparse arrays) }
                                Low_Index : longint ; { Low index (element #) of subscripts with this value }
                                High_Index : longint ; { High index (element #) of subscripts with this value }
                                { For data (when a consecutive series of elements have the same value, a single
                                   header can represent all of them), or the root of a sub-array. }
                                Data : TStore_Address ; { Pointer to data at this element (data or another TSparse_Data_Header ) }
                                Next : TStore_Address ; { Next element header }
                           end ;
      TStructure = object( TBase_Object )
                       Owns_Heap : boolean ;
                       Heap : PSHeap ;

                       constructor Init ;
                       destructor Done ; virtual ;
                   end ;
      Tarray = object( TStructure ) { Abstract array object }
                   constructor Init ;
                   procedure Array_Get( Offset : TRecord_Size ; var Data ) ; virtual ;
                   procedure Array_Put( Offset : TRecord_Size ; var Data ) ; virtual ;
                   destructor Done ; virtual ;
               end ;
      Tsarray = object( Tarray ) { Standard static array object }
                    Ptr : Array_Pointer ; { Pointer to data in heap }
                    Element_Count : TRecord_Size ; { Number of elements in array }
                    Element_Size : TRecord_Size ; { Size of elements in array }
                    Storing_Objects : boolean ; { True if storing objects }

                    constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                    procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                        virtual ;
                    procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                        virtual ;
                    destructor Done ; virtual ;
                end ;
      Static_Array =	object( Tsarray ) { STATIC ARRAY object }
                          Low, High : TRecord_Size ; { Low and high bounds for array }

                          constructor Init( Lbound, Ubound : Integer ;
                              Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
      Unbounded_Static_Array =
                      object( Static_Array ) {UNBOUNDED STATIC ARRAY object}
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
{$IFNDEF DELPHI_OR_WINDOWS}
      Ts_Emsarray =
                      object( Tarray ) {Standard static EMS array object}
                          Element_Count : TRecord_Size ; {Number of elements in array}
                          Element_Size : TRecord_Size ; {Size of elements in array}
                          Storing_Objects : boolean ; {True if storing objects}
                          Current_Handle, Current_Page : Integer ;
                          {Current expanded memory handle and page}
                          Page_Frame : Integer ;
                          {Segment of the expanded memory window}
                          Handle : Integer ; {EMM handle}
                          Page_Count : Integer ; {Number of 16K pages in array}
                          Internal_Size : TRecord_Size ;
                          {Internal size of each element}

                          constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          destructor Done ; virtual ;
                      end ;
      Static_Ems_Array =
                      object( Ts_Emsarray ) {STATIC EMS ARRAY object}
                          Low, High : TRecord_Size ; {Low and high bounds for array}

                          constructor Init( Lbound, Ubound : Integer ;
                              Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
{$ENDIF}
      Tstack =	object( Tsarray ) {Abstract stack object}
                          constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Push( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                          procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
      Stack =		object( Tstack ) {STACK object}
                          Stack_Top : TRecord_Size ; {Top of stack element}

                          constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Push( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                      end ;
      Tdarray =	object( Tsarray ) {Standard dynamic array object}
                          constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                      end ;
      Dynamic_Array =	object( Tdarray ) {DYNAMIC ARRAY object}
                          Low, High : TRecord_Size ; {Low and high bounds for array}

                          constructor Init( Lbound, Ubound : Integer ;
                              Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
      Tdstack =	object( Tdarray ) {Abstract dynamic stack object}
                          constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Push( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                          procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
      Dynamic_Stack =	object( Tdstack ) {DYNAMIC STACK object}
                          Stack_Top : TRecord_Size ; {Top of stack element}

                          constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Push( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                      end ;
      Tvarray =	object( Tarray ) {Standard virtual array object}
                          Image : File ; {array image file}
                          Start : Longint ; {Starting position in file}
                          Element_Count : TRecord_Size ; {Number of elements in array}
                          Element_Size : TRecord_Size ; {Size of elements in array}
                          Storing_Objects : boolean ; {True if storing objects}

                          constructor Init( Cnt, Size, Strt : TRecord_Size ;
                              Flag : boolean ; Nam : string ) ;
                          procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          destructor Done ; virtual ;
                      end ;
      Virtual_Array =	object( Tvarray ) {VIRTUAL ARRAY object}
                          Low, High : TRecord_Size ; {Low and high bounds for array}

                          constructor Init( Lbound, Ubound : Integer ;
                              Size, Strt : TRecord_Size ; Flag : boolean ;
                              Nam : string ) ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
      Tvstack =	object( Tvarray ) {Abstract virtual stack object}
                          constructor Init( Cnt, Size, Strt : TRecord_Size ;
                              Flag : boolean ; Nam : string ) ;
                          procedure Push( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                          procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
      Virtual_Stack =	object( Tvstack ) {VIRTUAL STACK object}
                          Stack_Top : TRecord_Size ; {Top of stack element}

                          constructor Init( Cnt, Size, Strt : TRecord_Size ;
                              Flag : boolean ; Nam : string ) ;
                          procedure Push( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                      end ;
      Tlist =		object( TStructure ) {Abstract list object}
                          constructor Init ;
                          procedure List_Add( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure List_Get( Index : List_Pointer ; var Data ) ;
                              virtual ;
                          procedure List_Insert( Index : List_Pointer ;
                              var Data ) ; virtual ;
                          procedure List_Delete( Index : List_Pointer ) ;
                              virtual ;
                          function List_First : List_Pointer ; virtual ;
                          function List_Last : List_Pointer ; virtual ;
                          function List_Next( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                          destructor Done ; virtual ;
                      end ;
      Tslist =	object( Tlist ) {Standard single-linked list object}
                          Ptr : List_Pointer ; {Pointer to end of list}
                          Item_Size : TRecord_Size ; {Size of items in list}
                          Storing_Objects : boolean ; {True if storing objects}

                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Clear ;
                          procedure List_Add( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure List_Get( Index : List_Pointer ; var Data ) ;
                              virtual ;
                          procedure List_Insert( Index : List_Pointer ;
                              var Data ) ; virtual ;
                          procedure List_Delete( Index : List_Pointer ) ;
                              virtual ;
                          function List_First : List_Pointer ; virtual ;
                          function List_Last : List_Pointer ; virtual ;
                          function List_Next( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                          destructor Done ; virtual ;
                      end ;
      Single_List =	object( Tslist ) {SINGLE-LINKED LIST object}
                          Context : List_Pointer ;

                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Add( var Data ) ; virtual ;
                          procedure Get( var Data ) ; virtual ;
                          procedure Insert( var Data ) ; virtual ;
                          procedure Delete ; virtual ;
                          procedure Last( var Data ) ; virtual ;
                          procedure First( var Data ) ; virtual ;
                          function Eol : boolean ; virtual ;
                          function Bol : boolean ; virtual ;
                      end ;
      Tlstack =	object( Tslist ) {Abstract list stack object}
                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Push( var Data ) ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                          procedure List_Add( var Data ) ; virtual ;
                          procedure List_Get( Index : List_Pointer ; var Data ) ;
                              virtual ;
                          procedure List_Insert( Index : List_Pointer ;
                              var Data ) ; virtual ;
                          procedure List_Delete( Index : List_Pointer ) ;
                              virtual ;
                          function List_First : List_Pointer ; virtual ;
                          function List_Last : List_Pointer ; virtual ;
                          function List_Next( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                      end ;
      List_Stack =	object( Tlstack ) {LIST STACK object}
                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Push( var Data ) ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                      end ;
      Tdlist =	object( Tlist ) { Standard double-linked list object }
                          Ptr : List_Pointer ; { Pointer to end of list }
                          Item_Size : TRecord_Size ; { Size of items in list }
                          Storing_Objects : boolean ; { True if storing objects }

                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Clear ; virtual ;
                          procedure List_Add( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure List_Get( Index : List_Pointer ; var Data ) ;
                              virtual ;
                          procedure List_Insert( Index : List_Pointer ;
                              var Data ) ; virtual ;
                          procedure List_Delete( Index : List_Pointer ) ;
                              virtual ;
                          function List_Last : List_Pointer ; virtual ;
                          function List_First : List_Pointer ; virtual ;
                          function List_Previous( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                          function List_Next( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                          destructor Done ; virtual ;
                      end ;
      Double_List =   object( Tdlist ) { Standard double-linked list object }
                          Context : List_Pointer ; { Pointer to EOL }

                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Add( var Data ) ; virtual ;
                          procedure Clear ; virtual ;
                          procedure Put( var Data ) ; virtual ;
                          procedure Get( var Data ) ; virtual ;
                          procedure Previous( var Data ) ; virtual ;
                          procedure Insert( var Data ) ; virtual ;
                          procedure Delete ; virtual ;
                          procedure Last( var Data ) ; virtual ;
                          procedure First( var Data ) ; virtual ;
                          function EOL : boolean ; virtual ;
                          function BOL : boolean ; virtual ;
                      end ;
      Tqueue =	object( Tdlist ) {Standard queue object}
                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Add( var Data ) ; virtual ;
                          procedure Front( var Data ) ; virtual ;
                          procedure Delete( var Data ) ; virtual ;
                          procedure List_Add( var Data ) ; virtual ;
                          procedure List_Get( Index : List_Pointer ; var Data ) ;
                              virtual ;
                          procedure List_Insert( Index : List_Pointer ;
                              var Data ) ; virtual ;
                          procedure List_Delete( Index : List_Pointer ) ;
                              virtual ;
                          function List_First : List_Pointer ; virtual ;
                          function List_Last : List_Pointer ; virtual ;
                          function List_Next( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                      end ;
      Queue =		object( Tqueue ) {QUEUE object}
                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Add( var Data ) ; virtual ;
                          procedure Front( var Data ) ; virtual ;
                          procedure Delete( var Data ) ; virtual ;
                      end ;
      Sparse_Array =  object( Tslist ) { Standard sparse array object }
                          Default : List_Pointer ; { Pointer to default item }

                          constructor Init( Size : TRecord_Size ; Flag : boolean ;
                              var Def ) ;
                          procedure Get( Offset : Longint ; var Data ) ;
                              virtual ;
                          procedure Put( Offset : Longint ; var Data ) ;
                              virtual ;
                          procedure List_Add( var Data ) ; virtual ;
                          procedure List_Get( Index : List_Pointer ; var Data ) ;
                              virtual ;
                          procedure List_Insert( Index : List_Pointer ;
                              var Data ) ; virtual ;
                          procedure List_Delete( Index : List_Pointer ) ;
                              virtual ;
                          function List_First : List_Pointer ; virtual ;
                          function List_Last : List_Pointer ; virtual ;
                          function List_Next( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                      end ;

      TTree = object( TStructure ) { Abstract tree object }

                  constructor Init( Size, Key : TRecord_Size ; Flag : boolean ) ;
                  function Tree_Find( var Key ) : Tree_Pointer ; virtual ;
                  procedure Tree_Get( Index : Tree_Pointer ; var Data ) ; virtual ;
                  procedure Tree_Put( Index : Tree_Pointer ; var Data ) ; virtual ;
                  procedure Tree_Add( var Data ) ; virtual ;
                  procedure Tree_Delete( Index : Tree_Pointer ) ; virtual ;
                  procedure Tree_Prune( Index : Tree_Pointer ) ; virtual ;
                  function Tree_First : Tree_Pointer ; virtual ;
                  function Tree_Last : Tree_Pointer ; virtual ;
                  function Empty : boolean ; virtual ;
                  function Tree_Next( Index : Tree_Pointer ) : Tree_Pointer ;
                      virtual ;
                  function Tree_Previous( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  destructor Done ; virtual ;

              end ;

      STree = object( TTree ) {Standard binary tree object}

                  Ptr : Tree_Pointer ; {Pointer to root node}
                  Item_Size : TRecord_Size ; {Size of items in tree}
                  Key_Size : TRecord_Size ; {Size of key in tree items}
                  Storing_Objects : boolean ; {True if storing objects}

                  constructor Init( Size, Key : TRecord_Size ; Flag : boolean ) ;
                  function Tree_Find( var Key ) : Tree_Pointer ; virtual ;
                  procedure Tree_Get( Index : Tree_Pointer ; var Data ) ;
                      virtual ;
                  procedure Tree_Put( Index : Tree_Pointer ; var Data ) ;
                      virtual ;
                  procedure Add( var Data ) ; virtual ;
                  procedure Tree_Delete( Index : Tree_Pointer ) ;
                      virtual ;
                  procedure Tree_Prune( Index : Tree_Pointer ) ; virtual ;
                  function Tree_First : Tree_Pointer ; virtual ;
                  function Tree_Last : Tree_Pointer ; virtual ;
                  function Empty : boolean ; virtual ;
                  function Tree_Next( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  function Tree_Previous( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  destructor Done ; virtual ;
              end ;

      Tree =  object( STree ) {Standard binary tree object with context}
                  Last_Error : Integer ;
                  Context : Tree_Pointer ;

                  constructor Init( Size, Key : TRecord_Size ;
                      Flag : boolean ) ;
                  procedure Find( var Key, Data ) ; virtual ;
                  procedure Update( var Data ) ; virtual ;
                  procedure Tree_Add( var Data ) ; virtual ;
                  procedure Delete ; virtual ;
                  procedure Prune ; virtual ;
                  procedure First( var Data ) ; virtual ;
                  procedure Last( var Data ) ; virtual ;
                  procedure Next( var Data ) ; virtual ;
                  procedure Previous( var Data ) ; virtual ;
                  procedure Set_Error( E : Integer ) ;
                      virtual ;
                  function Error : Integer ; virtual ;
              end ;

      XTree = object( TTree ) { Standard binary tree object with large keys}

                  Ptr : Tree_Pointer ; {Pointer to root node}
                  Item_Size : TRecord_Size ; {Size of items in tree}
                  Storing_Objects : boolean ; {True if storing objects}

                  constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                  function Find( Key : string ) : Tree_Pointer ; virtual ;
                  procedure Tree_Get( Index : Tree_Pointer ; var Data ) ;
                      virtual ;
                  procedure Tree_Put( Index : Tree_Pointer ; var Data ) ;
                      virtual ;
                  procedure Add( Key : string ; var Data ) ; virtual ;
                  procedure Tree_Delete( Index : Tree_Pointer ) ;
                      virtual ;
                  procedure Tree_Prune( Index : Tree_Pointer ) ; virtual ;
                  function Tree_First : Tree_Pointer ; virtual ;
                  function Tree_Last : Tree_Pointer ; virtual ;
                  function Empty : boolean ; virtual ;
                  function Tree_Next( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  function Tree_Previous( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  destructor Done ; virtual ;

              end ;

      { Standard B-tree }
      TBTree = object( TTree ) { NOTE: Not finished (see Guide to OpenVMS File Applications for B-tree stuff) }

                  Ptr : BTree_Pointer ; { Pointer to root node }
                  Bucket_Size : TRecord_Size ; { Size of data bucket }
                  Item_Size : TRecord_Size ; { Size of items in tree }
                  Key_Size : TRecord_Size ; { Size of key in tree items }
                  Max_B : integer ; { Maximum number of items in a node (B=2 is binary tree) }
                  Storing_Objects : boolean ; { True if storing objects }

                  constructor Init( Size, Key, Max, Bucket : TRecord_Size ; Flag : boolean ) ;
                  function Tree_Find( var Key ) : Tree_Pointer ; virtual ;
                  procedure Tree_Get( Index : Tree_Pointer ; var Data ) ;
                      virtual ;
                  procedure Tree_Put( Index : Tree_Pointer ; var Data ) ;
                      virtual ;
                  procedure Add( var Data ) ; virtual ;
                  procedure Tree_Delete( Index : Tree_Pointer ) ;
                      virtual ;
                  procedure Tree_Prune( Index : Tree_Pointer ) ; virtual ;
                  function Tree_First : Tree_Pointer ; virtual ;
                  function Tree_Last : Tree_Pointer ; virtual ;
                  function Empty : boolean ; virtual ;
                  function Tree_Next( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  function Tree_Previous( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  destructor Done ; virtual ;
               end ;

      { TMatrix not yet implemented }
      TMatrix = object( TStructure ) { Standard multi-dimensional matrix }
                    Ptr : TStore_Address ; { Pointer to matrix header }
                    Element_Size : longint ; { Size of elements in matrix }
                    Element_Count : longint ; { Number of elements in matrix }

                    constructor Init( Size, Elements : longint ) ;
                    destructor Done ; virtual ;

                    procedure Clear ; virtual ;
                    procedure Matrix_Get( Index : string ; var Data ) ;
                        virtual ;
                    procedure Matrix_Put( Index : string ; var Data ) ;
                        virtual ;
                end ;

      { 64-bit memory structures... }
      List_Pointer64 = TStore_Address64 ;
      List_Image64 = record
                         Next : TStore_Address64 ; { Pointer to next item on list }
                         Last : TStore_Address64 ;
                     end ;

      Tree_Pointer64 = ^Tree_Image64 ;
      Tree_Image_Min64 = record
                             Back : Tree_Pointer64 ;
                             Left : Tree_Pointer64 ;
                             Right : Tree_Pointer64 ;
                         end ;
      Tree_Image64 = record
                         Back : Tree_Pointer64 ;
                         Left : Tree_Pointer64 ;
                         Right : Tree_Pointer64 ;
                         D : array[ 0..Max_Memory - sizeof( Tree_Pointer ) * 3 ] of byte ; { Data area }
                     end ;
      BTree_Pointer64 = ^BTree_Image ;
      BTree_Image64 = record
                          Level : longint ; { Level 0 is data, Level 1 is lowest level of keys }
                          Active : longint ; { Number of active keys in this record (for key record) or size of record (data) }
                          Parent : BTree_Pointer ; { Pointer to parent (nil if top level) }
                          D : array[ 0..Max_Memory - 3 - sizeof( BTree_Pointer ) ] of byte ;
                          { Data area.  In a data record (Level = 0) this is the
                            image of the record's data.  In a key record (Level >
                            0) this is an array of BTree_Pointers (0 to Max_B - 1)
                            pointing to the next level records. }
                      end ;
      TSparse_Data_Header64 = record { Sparse array header (points to other sparse arrays) }
                                Low_Index : longint ; { Low index (element #) of subscripts with this value }
                                High_Index : longint ; { High index (element #) of subscripts with this value }
                                { For data (when a consecutive series of elements have the same value, a single
                                   header can represent all of them), or the root of a sub-array. }
                                Data : TStore_Address64 ; { Pointer to data at this element (data or another TSparse_Data_Header ) }
                                Next : TStore_Address64 ; { Next element header }
                           end ;
      TStructure64 = object( TBase_Object )
                       Owns_Heap : boolean ;
                       Heap : PSManaged_Store64 ;

                       constructor Init ;
                       destructor Done ; virtual ;
                   end ;
      Tarray64 = object( TStructure64 ) { Abstract array object }
                   constructor Init ;
                   procedure Array_Get( Offset : TRecord_Size ; var Data ) ; virtual ;
                   procedure Array_Put( Offset : TRecord_Size ; var Data ) ; virtual ;
                   destructor Done ; virtual ;
               end ;
      Tsarray64 = object( Tarray64 ) { Standard static array object }
                    Ptr : Array_Pointer ; { Pointer to data in heap }
                    Element_Count : TRecord_Size ; { Number of elements in array }
                    Element_Size : TRecord_Size ; { Size of elements in array }
                    Storing_Objects : boolean ; { True if storing objects }

                    constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                    procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                        virtual ;
                    procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                        virtual ;
                    destructor Done ; virtual ;
                end ;
      Static_Array64 =	object( Tsarray64 ) { STATIC ARRAY object }
                          Low, High : TRecord_Size ; { Low and high bounds for array }

                          constructor Init( Lbound, Ubound : Integer ;
                              Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
      Unbounded_Static_Array64 =
                      object( Static_Array64 ) {UNBOUNDED STATIC ARRAY object}
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
{$IFNDEF DELPHI_OR_WINDOWS}
      Ts_Emsarray64 =
                      object( Tarray64 ) {Standard static EMS array object}
                          Element_Count : TRecord_Size ; {Number of elements in array}
                          Element_Size : TRecord_Size ; {Size of elements in array}
                          Storing_Objects : boolean ; {True if storing objects}
                          Current_Handle, Current_Page : Integer ;
                          {Current expanded memory handle and page}
                          Page_Frame : Integer ;
                          {Segment of the expanded memory window}
                          Handle : Integer ; {EMM handle}
                          Page_Count : Integer ; {Number of 16K pages in array}
                          Internal_Size : TRecord_Size ;
                          {Internal size of each element}

                          constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          destructor Done ; virtual ;
                      end ;
      Static_Ems_Array64 =
                      object( Ts_Emsarray64 ) {STATIC EMS ARRAY object}
                          Low, High : TRecord_Size ; {Low and high bounds for array}

                          constructor Init( Lbound, Ubound : Integer ;
                              Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
{$ENDIF}
      Tstack64 =      object( Tsarray64 ) {Abstract stack object}
                          constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Push( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                          procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
      Stack64 =	      object( Tstack64 ) {STACK object}
                          Stack_Top : TRecord_Size ; {Top of stack element}

                          constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Push( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                      end ;
      Tdarray64 =     object( Tsarray64 ) {Standard dynamic array object}
                          constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                      end ;
      Dynamic_Array64 =	object( Tdarray64 ) {DYNAMIC ARRAY object}
                          Low, High : TRecord_Size ; {Low and high bounds for array}

                          constructor Init( Lbound, Ubound : Integer ;
                              Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
      Tdstack64 =     object( Tdarray64 ) {Abstract dynamic stack object}
                          constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Push( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                          procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
      Dynamic_Stack64 =	object( Tdstack64 ) {DYNAMIC STACK object}
                          Stack_Top : TRecord_Size ; {Top of stack element}

                          constructor Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Push( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                      end ;
      Tvarray64 =     object( Tarray64 ) {Standard virtual array object}
                          Image : File ; {array image file}
                          Start : Longint ; {Starting position in file}
                          Element_Count : TRecord_Size ; {Number of elements in array}
                          Element_Size : TRecord_Size ; {Size of elements in array}
                          Storing_Objects : boolean ; {True if storing objects}

                          constructor Init( Cnt, Size, Strt : TRecord_Size ;
                              Flag : boolean ; Nam : string ) ;
                          procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          destructor Done ; virtual ;
                      end ;
      Virtual_Array64 =	object( Tvarray64 ) {VIRTUAL ARRAY object}
                          Low, High : TRecord_Size ; {Low and high bounds for array}

                          constructor Init( Lbound, Ubound : Integer ;
                              Size, Strt : TRecord_Size ; Flag : boolean ;
                              Nam : string ) ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
      Tvstack64 =     object( Tvarray64 ) {Abstract virtual stack object}
                          constructor Init( Cnt, Size, Strt : TRecord_Size ;
                              Flag : boolean ; Nam : string ) ;
                          procedure Push( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                          procedure Array_Get( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Array_Put( Offset : TRecord_Size ; var Data ) ;
                              virtual ;
                          procedure Get( Offset : Integer ; var Data ) ; virtual ;
                          procedure Put( Offset : Integer ; var Data ) ; virtual ;
                      end ;
      Virtual_Stack64 =	object( Tvstack64 ) {VIRTUAL STACK object}
                          Stack_Top : TRecord_Size ; {Top of stack element}

                          constructor Init( Cnt, Size, Strt : TRecord_Size ;
                              Flag : boolean ; Nam : string ) ;
                          procedure Push( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                      end ;
      Tlist64 =	      object( TStructure64 ) {Abstract list object}
                          constructor Init ;
                          procedure List_Add( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure List_Get( Index : List_Pointer ; var Data ) ;
                              virtual ;
                          procedure List_Insert( Index : List_Pointer ;
                              var Data ) ; virtual ;
                          procedure List_Delete( Index : List_Pointer ) ;
                              virtual ;
                          function List_First : List_Pointer ; virtual ;
                          function List_Last : List_Pointer ; virtual ;
                          function List_Next( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                          destructor Done ; virtual ;
                      end ;
      Tslist64 =      object( Tlist64 ) {Standard single-linked list object}
                          Ptr : List_Pointer ; {Pointer to end of list}
                          Item_Size : TRecord_Size ; {Size of items in list}
                          Storing_Objects : boolean ; {True if storing objects}

                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Clear ;
                          procedure List_Add( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure List_Get( Index : List_Pointer ; var Data ) ;
                              virtual ;
                          procedure List_Insert( Index : List_Pointer ;
                              var Data ) ; virtual ;
                          procedure List_Delete( Index : List_Pointer ) ;
                              virtual ;
                          function List_First : List_Pointer ; virtual ;
                          function List_Last : List_Pointer ; virtual ;
                          function List_Next( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                          destructor Done ; virtual ;
                      end ;
      Single_List64 = object( Tslist64 ) {SINGLE-LINKED LIST object}
                          Context : List_Pointer ;

                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Add( var Data ) ; virtual ;
                          procedure Get( var Data ) ; virtual ;
                          procedure Insert( var Data ) ; virtual ;
                          procedure Delete ; virtual ;
                          procedure Last( var Data ) ; virtual ;
                          procedure First( var Data ) ; virtual ;
                          function Eol : boolean ; virtual ;
                          function Bol : boolean ; virtual ;
                      end ;
      Tlstack64 =     object( Tslist64 ) {Abstract list stack object}
                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Push( var Data ) ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                          procedure List_Add( var Data ) ; virtual ;
                          procedure List_Get( Index : List_Pointer ; var Data ) ;
                              virtual ;
                          procedure List_Insert( Index : List_Pointer ;
                              var Data ) ; virtual ;
                          procedure List_Delete( Index : List_Pointer ) ;
                              virtual ;
                          function List_First : List_Pointer ; virtual ;
                          function List_Last : List_Pointer ; virtual ;
                          function List_Next( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                      end ;
      List_Stack64 =  object( Tlstack64 ) {LIST STACK object}
                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Push( var Data ) ; virtual ;
                          procedure Top( var Data ) ; virtual ;
                          procedure Pop( var Data ) ; virtual ;
                      end ;
      Tdlist64 =      object( Tlist64 ) { Standard double-linked list object }
                          Ptr : List_Pointer ; { Pointer to end of list }
                          Item_Size : TRecord_Size ; { Size of items in list }
                          Storing_Objects : boolean ; { True if storing objects }

                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Clear ; virtual ;
                          procedure List_Add( var Data ) ; virtual ;
                          function Empty : boolean ; virtual ;
                          procedure List_Get( Index : List_Pointer ; var Data ) ;
                              virtual ;
                          procedure List_Insert( Index : List_Pointer ;
                              var Data ) ; virtual ;
                          procedure List_Delete( Index : List_Pointer ) ;
                              virtual ;
                          function List_Last : List_Pointer ; virtual ;
                          function List_First : List_Pointer ; virtual ;
                          function List_Previous( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                          function List_Next( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                          destructor Done ; virtual ;
                      end ;
      Double_List64 =   object( Tdlist64 ) { Standard double-linked list object }
                          Context : List_Pointer ; { Pointer to EOL }

                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Add( var Data ) ; virtual ;
                          procedure Clear ; virtual ;
                          procedure Put( var Data ) ; virtual ;
                          procedure Get( var Data ) ; virtual ;
                          procedure Previous( var Data ) ; virtual ;
                          procedure Insert( var Data ) ; virtual ;
                          procedure Delete ; virtual ;
                          procedure Last( var Data ) ; virtual ;
                          procedure First( var Data ) ; virtual ;
                          function EOL : boolean ; virtual ;
                          function BOL : boolean ; virtual ;
                      end ;
      Tqueue64 =      object( Tdlist64 ) {Standard queue object}
                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Add( var Data ) ; virtual ;
                          procedure Front( var Data ) ; virtual ;
                          procedure Delete( var Data ) ; virtual ;
                          procedure List_Add( var Data ) ; virtual ;
                          procedure List_Get( Index : List_Pointer ; var Data ) ;
                              virtual ;
                          procedure List_Insert( Index : List_Pointer ;
                              var Data ) ; virtual ;
                          procedure List_Delete( Index : List_Pointer ) ;
                              virtual ;
                          function List_First : List_Pointer ; virtual ;
                          function List_Last : List_Pointer ; virtual ;
                          function List_Next( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                      end ;
      Queue64 =		object( Tqueue64 ) {QUEUE object}
                          constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                          procedure Add( var Data ) ; virtual ;
                          procedure Front( var Data ) ; virtual ;
                          procedure Delete( var Data ) ; virtual ;
                      end ;
      Sparse_Array64 =  object( Tslist64 ) { Standard sparse array object }
                          Default : List_Pointer ; { Pointer to default item }

                          constructor Init( Size : TRecord_Size ; Flag : boolean ;
                              var Def ) ;
                          procedure Get( Offset : Longint ; var Data ) ;
                              virtual ;
                          procedure Put( Offset : Longint ; var Data ) ;
                              virtual ;
                          procedure List_Add( var Data ) ; virtual ;
                          procedure List_Get( Index : List_Pointer ; var Data ) ;
                              virtual ;
                          procedure List_Insert( Index : List_Pointer ;
                              var Data ) ; virtual ;
                          procedure List_Delete( Index : List_Pointer ) ;
                              virtual ;
                          function List_First : List_Pointer ; virtual ;
                          function List_Last : List_Pointer ; virtual ;
                          function List_Next( Index : List_Pointer ) :
                              List_Pointer ; virtual ;
                      end ;

      TTree64 = object( TStructure64 ) { Abstract tree object }

                  constructor Init( Size, Key : TRecord_Size ; Flag : boolean ) ;
                  function Tree_Find( var Key ) : Tree_Pointer ; virtual ;
                  procedure Tree_Get( Index : Tree_Pointer ; var Data ) ; virtual ;
                  procedure Tree_Put( Index : Tree_Pointer ; var Data ) ; virtual ;
                  procedure Tree_Add( var Data ) ; virtual ;
                  procedure Tree_Delete( Index : Tree_Pointer ) ; virtual ;
                  procedure Tree_Prune( Index : Tree_Pointer ) ; virtual ;
                  function Tree_First : Tree_Pointer ; virtual ;
                  function Tree_Last : Tree_Pointer ; virtual ;
                  function Empty : boolean ; virtual ;
                  function Tree_Next( Index : Tree_Pointer ) : Tree_Pointer ;
                      virtual ;
                  function Tree_Previous( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  destructor Done ; virtual ;

              end ;

      STree64 = object( TTree64 ) {Standard binary tree object}

                  Ptr : Tree_Pointer ; {Pointer to root node}
                  Item_Size : TRecord_Size ; {Size of items in tree}
                  Key_Size : TRecord_Size ; {Size of key in tree items}
                  Storing_Objects : boolean ; {True if storing objects}

                  constructor Init( Size, Key : TRecord_Size ; Flag : boolean ) ;
                  function Tree_Find( var Key ) : Tree_Pointer ; virtual ;
                  procedure Tree_Get( Index : Tree_Pointer ; var Data ) ;
                      virtual ;
                  procedure Tree_Put( Index : Tree_Pointer ; var Data ) ;
                      virtual ;
                  procedure Add( var Data ) ; virtual ;
                  procedure Tree_Delete( Index : Tree_Pointer ) ;
                      virtual ;
                  procedure Tree_Prune( Index : Tree_Pointer ) ; virtual ;
                  function Tree_First : Tree_Pointer ; virtual ;
                  function Tree_Last : Tree_Pointer ; virtual ;
                  function Empty : boolean ; virtual ;
                  function Tree_Next( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  function Tree_Previous( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  destructor Done ; virtual ;
              end ;

      Tree64 = object( STree64 ) {Standard binary tree object with context}
                  Last_Error : Integer ;
                  Context : Tree_Pointer ;

                  constructor Init( Size, Key : TRecord_Size ;
                      Flag : boolean ) ;
                  procedure Find( var Key, Data ) ; virtual ;
                  procedure Update( var Data ) ; virtual ;
                  procedure Tree_Add( var Data ) ; virtual ;
                  procedure Delete ; virtual ;
                  procedure Prune ; virtual ;
                  procedure First( var Data ) ; virtual ;
                  procedure Last( var Data ) ; virtual ;
                  procedure Next( var Data ) ; virtual ;
                  procedure Previous( var Data ) ; virtual ;
                  procedure Set_Error( E : Integer ) ;
                      virtual ;
                  function Error : Integer ; virtual ;
              end ;

      XTree64 = object( TTree64 ) { Standard binary tree object with large keys}

                  Ptr : Tree_Pointer ; {Pointer to root node}
                  Item_Size : TRecord_Size ; {Size of items in tree}
                  Storing_Objects : boolean ; {True if storing objects}

                  constructor Init( Size : TRecord_Size ; Flag : boolean ) ;
                  function Find( Key : string ) : Tree_Pointer ; virtual ;
                  procedure Tree_Get( Index : Tree_Pointer ; var Data ) ;
                      virtual ;
                  procedure Tree_Put( Index : Tree_Pointer ; var Data ) ;
                      virtual ;
                  procedure Add( Key : string ; var Data ) ; virtual ;
                  procedure Tree_Delete( Index : Tree_Pointer ) ;
                      virtual ;
                  procedure Tree_Prune( Index : Tree_Pointer ) ; virtual ;
                  function Tree_First : Tree_Pointer ; virtual ;
                  function Tree_Last : Tree_Pointer ; virtual ;
                  function Empty : boolean ; virtual ;
                  function Tree_Next( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  function Tree_Previous( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  destructor Done ; virtual ;

              end ;

      { Standard B-tree }
      TBTree64 = object( TTree64 ) { NOTE: Not finished (see Guide to OpenVMS File Applications for B-tree stuff) }

                  Ptr : BTree_Pointer ; { Pointer to root node }
                  Bucket_Size : TRecord_Size ; { Size of data bucket }
                  Item_Size : TRecord_Size ; { Size of items in tree }
                  Key_Size : TRecord_Size ; { Size of key in tree items }
                  Max_B : integer ; { Maximum number of items in a node (B=2 is binary tree) }
                  Storing_Objects : boolean ; { True if storing objects }

                  constructor Init( Size, Key, Max, Bucket : TRecord_Size ; Flag : boolean ) ;
                  function Tree_Find( var Key ) : Tree_Pointer ; virtual ;
                  procedure Tree_Get( Index : Tree_Pointer ; var Data ) ;
                      virtual ;
                  procedure Tree_Put( Index : Tree_Pointer ; var Data ) ;
                      virtual ;
                  procedure Add( var Data ) ; virtual ;
                  procedure Tree_Delete( Index : Tree_Pointer ) ;
                      virtual ;
                  procedure Tree_Prune( Index : Tree_Pointer ) ; virtual ;
                  function Tree_First : Tree_Pointer ; virtual ;
                  function Tree_Last : Tree_Pointer ; virtual ;
                  function Empty : boolean ; virtual ;
                  function Tree_Next( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  function Tree_Previous( Index : Tree_Pointer ) :
                      Tree_Pointer ; virtual ;
                  destructor Done ; virtual ;
               end ;

      { TMatrix not yet implemented }
      TMatrix64 = object( TStructure64 ) { Standard multi-dimensional matrix }
                    Ptr : TStore_Address64 ; { Pointer to matrix header }
                    Element_Size : longint ; { Size of elements in matrix }
                    Element_Count : longint ; { Number of elements in matrix }

                    constructor Init( Size, Elements : longint ) ;
                    destructor Done ; virtual ;

                    procedure Clear ; virtual ;
                    procedure Matrix_Get( Index : string ; var Data ) ;
                        virtual ;
                    procedure Matrix_Put( Index : string ; var Data ) ;
                        virtual ;
                end ;

const Max_Cached_Sparse_Matrix = 4 ;

type TSparse_Matrix64 = object( TMatrix64 ) { Standard multi-dimensional sparse matrix }
                private
                    Last_Index : array[ 0..Max_Cached_Sparse_Matrix ] of TStore_Address64 ; { Last index looked up }
                    Last_PTR : array[ 0..Max_Cached_Sparse_Matrix ] of TStore_Address64 ; { Pointer to element for Last_Index }
                    Last_Previous_PTR : array[ 0..Max_Cached_Sparse_Matrix ] of TStore_Address64 ; { Pointer to previous element for Last_PTR }
                    Did_Clear : boolean ; { True if cleared Last_Index in .Clear }

                    procedure _Put( Level : integer ; Index : string ; var Data ) ;

                public
                    constructor Init( Size, Elements : longint ) ;
                    destructor Done ; virtual ;

                    procedure Clear ; virtual ;
                    function Element_Exists( Index : string ) : boolean ;
                        virtual ;
                    procedure Matrix_Get( Index : string ; var Data ) ;
                        virtual ;
                    procedure Matrix_Put( Index : string ; var Data ) ;
                        virtual ;
                    procedure Matrix_Put_Range( Index : string ;
                        High : integer ; var Data ) ; virtual ; { UNFINISHED }
                end ;
{$IFDEF DELPHI}
      { By using dynamic open arrays instead of building the element list with
        strings, use of the TFast_Sparse_Matrix is around 4.9 times faster
        than the TSparse_Matrix64. }
      TFast_Sparse_Matrix64 = object( TMatrix64 )
                                destructor Done ; virtual ;

                                procedure Clear ; virtual ;
                                function _Element_Exists( Level : integer ; Index : array of integer ) : boolean ;
                                procedure _Get_Matrix( Level : integer ; Index : array of integer ; var Data ) ;
                                procedure _Put_Matrix( Level : integer ; Index : array of integer ; var Data ) ;
                                function Element_Exists( Index : array of integer ) : boolean ;
                                    virtual ;
                                procedure Get_Matrix( Index : array of integer ; var Data ) ;
                                    virtual ;
                                procedure Put_Matrix( Index : array of integer ; var Data ) ;
                                    virtual ;
(*
                                procedure Matrix_Put_Range( Index : array of integer ;
                                    High : integer ; var Data ) ; virtual ;
*)
                            end ;
{$ENDIF}
      TGrouped_Sparse_Matrix64 = object( TMatrix64 ) { Grouped multi-dimensional sparse matrix }
                    destructor Done ; virtual ;

                    procedure Clear ; virtual ;
                    function Element_Exists( Index : string ) : boolean ;
                        virtual ;
                    procedure Matrix_Get( Index : string ; var Data ) ;
                        virtual ;
                    procedure Matrix_Put( Index : string ; var Data ) ;
                        virtual ;
                end ;

{$IFDEF DELPHI}
      TFast_Grouped_Sparse_Matrix64 = object( TMatrix64 )
                                destructor Done ; virtual ;

                                procedure Clear ; virtual ;
                                function Element_Exists( Index : array of integer ) : boolean ;
                                    virtual ;
                                procedure Get_Matrix( Index : array of integer ; var Data ) ;
                                    virtual ;
                                procedure Put_Matrix( Index : array of integer ; var Data ) ;
                                    virtual ;
                            end ;
{$ENDIF}
const BTree_Overhead64 = sizeof( longint ) * 2 + sizeof( BTree_Pointer64 ) ;

var Temp_Ptr : TStore_Address ;
var Temp_Ptr64 : TStore_Address64 ;

type TSparse_Matrix = object( TMatrix ) { Standard multi-dimensional sparse matrix }
                private
                    Last_Index : array[ 0..Max_Cached_Sparse_Matrix ] of TStore_Address ; { Last index looked up }
                    Last_PTR : array[ 0..Max_Cached_Sparse_Matrix ] of Tstore_Address ; { Pointer to element for Last_Index }
                    Last_Previous_PTR : array[ 0..Max_Cached_Sparse_Matrix ] of Tstore_Address ; { Pointer to previous element for Last_PTR }
                    Did_Clear : boolean ; { True if cleared Last_Index in .Clear }

                    procedure _Put( Level : integer ; Index : string ; var Data ) ;

                public
                    constructor Init( Size, Elements : longint ) ;
                    destructor Done ; virtual ;

                    procedure Clear ; virtual ;
                    function Element_Exists( Index : string ) : boolean ;
                        virtual ;
                    procedure Matrix_Get( Index : string ; var Data ) ;
                        virtual ;
                    procedure Matrix_Put( Index : string ; var Data ) ;
                        virtual ;
                    procedure Matrix_Put_Range( Index : string ;
                        High : integer ; var Data ) ; virtual ; { UNFINISHED }
                end ;
{$IFDEF DELPHI}
      { By using dynamic open arrays instead of building the element list with
        strings, use of the TFast_Sparse_Matrix is around 4.9 times faster
        than the TSparse_Matrix. }
      TFast_Sparse_Matrix = object( TMatrix )
                                destructor Done ; virtual ;

                                procedure Clear ; virtual ;
                                function _Element_Exists( Level : integer ; Index : array of integer ) : boolean ;
                                procedure _Get_Matrix( Level : integer ; Index : array of integer ; var Data ) ;
                                procedure _Put_Matrix( Level : integer ; Index : array of integer ; var Data ) ;
                                function Element_Exists( Index : array of integer ) : boolean ;
                                    virtual ;
                                procedure Get_Matrix( Index : array of integer ; var Data ) ;
                                    virtual ;
                                procedure Put_Matrix( Index : array of integer ; var Data ) ;
                                    virtual ;
(*
                                procedure Matrix_Put_Range( Index : array of integer ;
                                    High : integer ; var Data ) ; virtual ;
*)
                            end ;
{$ENDIF}
      TGrouped_Sparse_Matrix = object( TMatrix ) { Grouped multi-dimensional sparse matrix }
                    destructor Done ; virtual ;

                    procedure Clear ; virtual ;
                    function Element_Exists( Index : string ) : boolean ;
                        virtual ;
                    procedure Matrix_Get( Index : string ; var Data ) ;
                        virtual ;
                    procedure Matrix_Put( Index : string ; var Data ) ;
                        virtual ;
                end ;

{$IFDEF DELPHI}
      TFast_Grouped_Sparse_Matrix = object( TMatrix )
                                destructor Done ; virtual ;

                                procedure Clear ; virtual ;
                                function Element_Exists( Index : array of integer ) : boolean ;
                                    virtual ;
                                procedure Get_Matrix( Index : array of integer ; var Data ) ;
                                    virtual ;
                                procedure Put_Matrix( Index : array of integer ; var Data ) ;
                                    virtual ;
                            end ;
{$ENDIF}
const BTree_Overhead = sizeof( longint ) * 2 + sizeof( BTree_Pointer ) ;

{$IFDEF TEST}
procedure Test_Unit ;
{$ENDIF}

implementation

uses { Borland... }
{$IFNDEF DELPHI_OR_WINDOWS}
     DOS,
{$ENDIF}

     { Subroutine Library... }
     _App, { Application_Manager}
     Num1s, { Num1 }
     Standard ;


{ Unit utility functions }

function Compare( var X, Y ; Size : TRecord_Size ) : Integer ;
{Compare two memory addresses and return:
	0 = Equal
	1 = X>Y
	2 = X<Y }

var Dummy1 : TRecord_Size ;

begin
    For Dummy1 := 1 To Size do { For each byte of values to compare }
    begin
	if Array_Image( X )[ Dummy1 - 1 ] > Array_Image( Y )[ Dummy1 - 1 ] then
	begin
	    Compare := 1 ;
	    Exit
	end ;
	if Array_Image( X )[ Dummy1 - 1 ] < Array_Image( Y )[ Dummy1 - 1 ] then
	begin
	    Compare := 2 ;
	    Exit
	end
    end ;
    Compare := 0 { Otherwise they are equal }
end ;


procedure Terminate( const X : string ) ;

begin
    Application_Manager^.Fatal_Exception( X + ' template illegally referenced' ) ;
end ;


procedure Terminate1( const X, Y : string ) ;

begin
    Application_Manager^.Fatal_Exception( X + ' object cannot call ' + Y + ' routine' ) ;
end ;



{ TStructure methods... }

constructor TStructure.Init ;

var H : PSHeap ;

begin
{$IFNDEF DELPHI}
    TBase_Object.Init ;
{$ENDIF}
    new( H, Init ) ;
    Heap := H ;
    Owns_Heap := True ;
end ;


destructor TStructure.Done ;

begin
    if( ( Heap <> nil ) and Owns_Heap ) then
    begin
        dispose( Heap, Done ) ;
    end ;

{$IFNDEF DELPHI}
    inherited Done ;
{$ENDIF}
end ;


			{Abstract ARRAY object routines}

constructor Tarray.Init ;

begin
    inherited Init ; { Initialize object }
end ;

procedure Tarray.Array_Get ; {Only a template - error if called}

begin
    Terminate( 'TArray.Array_Get' ) ;
end ;

procedure Tarray.Array_Put ; { Only a template - error if called }

begin
    Terminate( 'TArray.Array_Put' ) ;
end ;

destructor Tarray.Done ;

begin
    inherited Done ;
end ;


			{Standard static ARRAY object routines}

constructor Tsarray.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ; {Initialize array}

begin
    if( Flag and ( Size <> 4 ) ) then {array contains objects, but size is not 4}
    begin
        Application_Manager^.Fatal_Exception( 'Objects in arrays must be size 4' ) ;
    end ;
    if( Size > Max_Memory - 9 ) then
    begin
        Application_Manager^.Fatal_Exception( 'Maximum size of array object data exceeded' ) ;
    end ;
    Tarray.Init ; {Initialize object}
    Element_Count := Cnt ; {Number of elements to store}
    Element_Size := Size ; {Size of elements being stored}
    Storing_Objects := Flag ;
    Getmem( Ptr, Size * Cnt ) ; {Allocate array on heap}
    Fillchar( Ptr^[ 0 ], Size * Cnt, 0 ) {Zero it}
end ;


procedure Tsarray.Array_Get( Offset : TRecord_Size ; var Data ) ;
{ Get an element from the array }

begin
    if Offset >= Element_Count then { Subscript out of range }
    begin
        Application_Manager^.Fatal_Exception( 'Subscript out of range in TSArray.Array_Get' ) ;
    end ;
    Move( Ptr^[ Offset * Element_Size ], Data, Element_Size ) {Return value}
end ;


procedure Tsarray.Array_Put( Offset : TRecord_Size ; var Data ) ;
{ Put an element in the array }

begin
    if Offset >= Element_Count then {Subscript out of range}
    begin
        Application_Manager^.Fatal_Exception( 'Subscript out of range in TSArray.Array_Put' ) ;
    end ;
    Move( Data, Ptr^[ Offset * Element_Size ], Element_Size ) {Set value}
end ;


destructor Tsarray.Done ; { Delete entire array }

Type Tsarray_Pointer = ^Tsarray ;

var Offset : Integer ;
    Object_Pointer : Pointer ;

begin
    if Storing_Objects then {if storing objects in array}
	For Offset := 0 To Element_Count - 1 do {Check each element}
	begin
	    Move( Ptr^[ Offset * Element_Size ], Object_Pointer, 4 ) ; {Get pointer}
	    if Object_Pointer <> nil then {if an object is present in element}
		Dispose( Tsarray_Pointer( Object_Pointer ), Done )
		{Dispose of object}
	end ;
    inherited Done ;
end ;


			{STATIC ARRAY object routines}

constructor Static_Array.Init( Lbound, Ubound : Integer ; Size : TRecord_Size ;
    Flag : boolean ) ; { Initialize array }

begin
    if( Lbound > Ubound ) then
    begin
        Application_Manager^.Fatal_Exception( 'Static array ' + Num1( Lbound ) + '..' + Num1( Ubound ) + ' is invalid' ) ;
    end ;
    Tsarray.Init( Ubound - Lbound + 1, Size, Flag ) ; { Initialize object }
    Low := Lbound ;
    High := Ubound
end ;


procedure Static_Array.Get( Offset : Integer ; var Data ) ;
{ Get an element from the array }

begin
    if ( Offset > High ) Or ( Offset < Low ) then { Subscript out of range }
    begin
        Application_Manager^.Fatal_Exception( 'Subscript out of range in Static_Array.Get' ) ;
    end ;
    Tsarray.Array_Get( Offset - Low, Data )
end ;


procedure Static_Array.Put( Offset : Integer ; var Data ) ;
{Put an element in the array}

begin
    if ( Offset > High ) Or ( Offset < Low ) then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in Static_Array.Put' ) ;
	Halt
    end ;
    Tsarray.Array_Put( Offset - Low, Data )
end ;


			{UNBOUNDED STATIC ARRAY object routines}

procedure Unbounded_Static_Array.Get( Offset : Integer ; var Data ) ;
{Get an element from the array}

begin
    if( Offset > High ) then
    begin
	Offset := High ;
    end ;
    if( Offset < Low ) then
    begin
	Offset := Low ;
    end ;
    inherited Array_Get( Offset - Low, Data ) ;
end ;


procedure Unbounded_Static_Array.Put( Offset : Integer ; var Data ) ;
{Put an element in the array}

begin
    if( Offset > High ) then
    begin
	Offset := High ;
    end ;
    if( Offset < Low ) then
    begin
	Offset := Low ;
    end ;
    inherited Array_Put( Offset - Low, Data )
end ;


{$IFNDEF DELPHI_OR_WINDOWS}
			{Standard static EMS ARRAY object routines}

constructor Ts_Emsarray.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
{Initialize array}

var Pages : Longint ;
    Dummy : Integer ;
    F : File ;
    Reg : Registers ;

begin
    if Flag and ( Size <> 4 ) then {array contains objects, but size is not 4}
    begin
	Writeln( 'Objects in arrays must be size 4' ) ;
	Halt
    end ;
    if Size > 16384 then
    begin
	Writeln( 'Maximum EMS array element size is 16384' ) ;
	Halt
    end ;

    {Make sure driver is present}
    Assign( F, 'EMMXXXX0' ) ;
    {$I-}
    Reset( F ) ;
    {$I+}
    if Ioresult <> 0 then
    begin
	Writeln( 'Expanded memory device not present' ) ;
	Halt
    end ;
    Close( F ) ;

    Reg.Ax := $4600 ;
    Intr( $67, Reg ) ; {Get driver version number}
    if Reg.Ah <> 0 then
    begin
	Writeln( 'Expanded memory device not responding' ) ;
	Halt
    end ;
    if Reg.Al < $30 then
    begin
	Writeln( 'Incompatible version of Expanded memory driver' ) ;
	Halt
    end ;

    Reg.Ax := $4100 ;
    Intr( $67, Reg ) ; {Get page frame address}
    if Reg.Ah <> 0 then
    begin
	Writeln( 'Expanded Memory device not responding' ) ;
	Halt
    end ;
    Page_Frame := Reg.Bx ;

    Current_Handle := Maxint ; {Nothing is mapped}
    Current_Page := Maxint ;

    Pages := Cnt ;
    Pages := Pages * Size ;
    Pages := ( Pages + 16383 ) Div 16384 ;
    {Total number of 16K pages needed (rounded up to nearest page)}
    if Pages = 0 then
    begin
	Writeln( 'Requested array size is zero pages' ) ;
	Halt
    end ;

    Page_Count := Pages - 1 ; {store the number of pages for reference}

    Reg.Ax := $4200 ;
    Intr( $67, Reg ) ; {Get EMS pages available}
    if Reg.Ah <> 0 then
    begin
	Writeln( 'Expanded memory device not responding' ) ;
	Halt
    end ;

    if Page_Count + 1 > Reg.Bx then
    begin
	Writeln( 'Insufficient Expanded Memory available for array' ) ;
	Halt
    end ;

    Tarray.Init ; {Initialize object}
    Reg.Ax := $4300 ;
    Reg.Bx := Page_Count + 1 ;
    Intr( $67, Reg ) ; {Allocate EMM space}
    if Reg.Ah <> 0 then
    begin
	Writeln( 'Expanded memory device not responding' ) ;
	Halt
    end ;
    Handle := Reg.Dx ;

    Element_Count := Cnt ;
    Element_Size := Size ;
    Internal_Size := Size ;

    {Now round up internal size to a power of 2}
    Dummy := 1 ; {Start at 2^0}
    while Dummy < Size do
    begin
	Dummy := Dummy + Dummy ; {Go to next power of 2}
	if Size < Dummy then Internal_Size := Dummy
    end ;

    {Now initialize the array:}
    For Dummy := 0 To Page_Count do
    begin
	Reg.Ax := $4400 ;
	Reg.Bx := Dummy ;
	Reg.Dx := Handle ;
	Intr( $67, Reg ) ; {Get page n}
	if Reg.Ah <> 0 then
	begin
            Application_Manager^.Fatal_Exception( 'Expanded memory device not responding' ) ;
	end ;
	Fillchar( Mem[ Page_Frame : 0 ], 16384, 0 ) {Zero the window}
    end ;
    Current_Page := Page_Count ;
    Current_Handle := Handle ;
    Storing_Objects := Flag
end ;


procedure Ts_Emsarray.Array_Get( Offset : TRecord_Size ; var Data ) ;
{Get an element from the array}

var Pn : Integer ; {16K page number}
    Po : Integer ; {offset into 16K page}
    Offs : Longint ;
    Reg : Registers ;

begin
    if Offset >= Element_Count then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in TS_EMSArray.Array_Get' ) ;
	Halt
    end ;
    Offs := Offset ;
    Offs := Offs * Internal_Size ; {Convert array offset to byte offset}
    Pn := Offs Div 16384 ; {Page}
    Po := Offs and 16383 ; {Offset}

    if ( Handle <> Current_Handle ) Or ( Pn <> Current_Page ) then
    begin
	Reg.Ax := $4400 ;
	Reg.Bx := Pn ;
	Reg.Dx := Handle ;
	Intr( $67, Reg ) ; {Get required page}
	if Reg.Ah <> 0 then
	begin
            Application_Manager^.Fatal_Exception( 'Expanded memory device not responding' ) ;
	end ;
        Current_Page := Pn ;
        Current_Handle := Handle
    end ;
    Move( Mem[ Page_Frame : Po ], Data, Element_Size ) {Get the data}
end ;


procedure Ts_Emsarray.Array_Put( Offset : TRecord_Size ; var Data ) ;
{Put an element in the array}

var Pn : Integer ; {16K page number}
    Po : Integer ; {offset into 16K page}
    Offs : Longint ;
    Reg : Registers ;

begin
    if Offset >= Element_Count then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in TS_EMSArray.Array_Put' ) ;
	Halt
    end ;
    Offs := Offset ;
    Offs := Offs * Internal_Size ; {Convert from array offset to byte offset}
    Pn := Offs Div 16384 ; {Page}
    Po := Offs and 16383 ; {Offset}

    if ( Handle <> Current_Handle ) Or ( pn <> Current_Page ) then
    begin
	Reg.Ax := $4400 ;
	Reg.Bx := Pn ;
	Reg.Dx := Handle ;
	Intr( $67, Reg ) ; {Get required page}
	if Reg.Ah <> 0 then
	begin
            Application_Manager^.Fatal_Exception( 'Expanded memory device not responding' ) ;
	end ;
	Current_Page := Pn ;
	Current_Handle := Handle ;
    end ;
    Move( Data, Mem[ Page_Frame : Po ], Element_Size ) {set the data}
end ;


destructor Ts_Emsarray.Done ; {Delete entire array}

Type Ts_Emsarray_Pointer = ^Ts_Emsarray ;

var Offset : Integer ;
    Object_Pointer : Pointer ;
    Reg : Registers ;

begin
    Reg.Ax := $4500 ;
    Reg.Dx := Handle ;
    Intr( $67, Reg ) ; { Deallocate memory }
    if Reg.Ah <> 0 then
    begin
	Writeln( 'Expanded memory device not responding' ) ;
	Halt ;
    end ;
    inherited Done ;
end ;


			{STATIC EMS ARRAY object routines}

constructor Static_Ems_Array.Init( Lbound, Ubound : Integer ; Size : TRecord_Size ;
    Flag : boolean ) ; {Initialize array}

begin
    if Lbound > Ubound then
    begin
	Writeln( 'Static EMS array ', Lbound, '..', Ubound, ' is invalid' ) ;
	Halt
    end ;
    Ts_Emsarray.Init( Ubound - Lbound + 1, Size, Flag ) ; {Initialize object}
    Low := Lbound ;
    High := Ubound
end ;


procedure Static_Ems_Array.Get( Offset : Integer ; var Data ) ;
{Get an element from the array}

begin
    if ( Offset > High ) Or ( Offset < Low ) then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in Static_EMS_Array.Get' ) ;
	Halt ;
    end ;
    Ts_Emsarray.Array_Get( Offset - Low, Data )
end ;


procedure Static_Ems_Array.Put( Offset : Integer ; var Data ) ;
{Put an element in the array}

begin
    if ( Offset > High ) Or ( Offset < Low ) then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in Static_EMS_Array.Put' ) ;
	Halt
    end ;
    Ts_Emsarray.Array_Put( Offset - Low, Data ) ;
end ;
{$ENDIF}


			{Abstract stack object routines}

constructor Tstack.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;

begin
    inherited Init( Cnt, Size, Flag ) ;
end ;


procedure Tstack.Push( var Data ) ;

begin
    Terminate( 'TSTack.Push' )
end ;


function Tstack.Empty : boolean ;

begin
    Empty := True ;
    Terminate( 'TStack.Empty' ) ;
end ;


procedure Tstack.Top( var Data ) ;

begin
    Terminate( 'TSTack.Top' )
end ;


procedure Tstack.Pop( var Data ) ;

begin
    Terminate( 'TSTack.Pop' )
end ;


procedure Tstack.Array_Get( Offset : TRecord_Size ; var Data ) ;

begin
    Terminate1( 'Stack', 'Array_Get' )
end ;


procedure Tstack.Array_Put( Offset : TRecord_Size ; var Data ) ;

begin
    Terminate1( 'Stack', 'Array_Put' )
end ;


procedure Tstack.Get( Offset : Integer ; var Data ) ;

begin
    Terminate1( 'Stack', 'Get' )
end ;


procedure Tstack.Put( Offset : Integer ; var Data ) ;

begin
    Terminate1( 'Stack', 'Put' )
end ;


			{STACK object routines}

constructor Stack.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;

begin
    if( Cnt > Max_Memory ) then
    begin
	Writeln( 'Maximum stack size is exceeded' ) ;
	Halt ;
    end ;
    Tsarray.Init( Cnt, Size, Flag ) ; {Initialize an array}
    Stack_Top := 0 {Nothing on stack}
end ;


procedure Stack.Push( var Data ) ;

begin
    if Stack_Top >= Element_Count then
    begin
	Writeln( 'Stack overflow in Stack.Push' ) ;
	Halt
    end ;
    Tsarray.Array_Put( Stack_Top, Data ) ; {Push onto top of stack}
    Stack_Top := Stack_Top + 1
end ;


function Stack.Empty : boolean ;

begin
    if Stack_Top = 0
	then
	    Empty := True
	else
	    Empty := False
end ;


procedure Stack.Top( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in Stack.Top' ) ;
	Halt
    end ;
    Tsarray.Array_Get( Stack_Top - 1, Data ) {Return data}
end ;


procedure Stack.Pop( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in Stack.Pop' ) ;
	Halt
    end ;
    Stack_Top := Stack_Top - 1 ;
    Tsarray.Array_Get( Stack_Top, Data ) {Return data}
end ;


			{Standard dynamic ARRAY object routines}

constructor Tdarray.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ; {Initialize array}

begin
    Tsarray.Init( Cnt, Size, Flag ) {Initialize object}
end ;


procedure Tdarray.Array_Get( Offset : TRecord_Size ; var Data ) ;
{Get an element from the array}

begin
    if Offset >= Element_Count {Subscript out of range}
	then
	    Fillchar( Data, Element_Size, 0 ) {Always return 0/nil/false}
	else
	    Move( Ptr^[ Offset * Element_Size ], Data, Element_Size ) {Return value}
end ;


procedure Tdarray.Array_Put( Offset : TRecord_Size ; var Data ) ;
{Put an element in the array}

var Temp : Array_Pointer ;

begin
    if Offset >= Element_Count then {Subscript out of range}
    begin
	Getmem( Temp, ( Offset + 1 ) * Element_Size ) ; {Allocate another chunk of memory}
	Fillchar( Temp^[ 0 ], ( Offset + 1 ) * Element_Size, 0 ) ; {Zero it}
	Move( Ptr^[ 0 ], Temp^[ 0 ], Element_Count * Element_Size ) ; {Copy old data}
	Freemem( Ptr, Element_Count * Element_Size ) ; {Free up old memory}
	Element_Count := Offset + 1 ; {New element count}
	Ptr := Temp {New pointer to data}
    end ;
    Move( Data, Ptr^[ Offset * Element_Size ], Element_Size ) {Set value}
end ;


			{DYNAMIC ARRAY object routines}

constructor Dynamic_Array.Init( Lbound, Ubound : Integer ; Size : TRecord_Size ;
    Flag : boolean ) ; {Initialize array}

begin
    if Lbound > Ubound then
    begin
	Writeln( 'Dynamic array ', Lbound, '..', Ubound, ' is invalid' ) ;
	Halt
    end ;
    Tdarray.Init( Ubound - Lbound + 1, Size, Flag ) ; {Initialize object}
    Low := Lbound ;
    High := Ubound
end ;


procedure Dynamic_Array.Get( Offset : Integer ; var Data ) ;
{Get an element from the array}

begin
    if Offset > High then High := Offset ;
    if Offset < Low
	then
	    Fillchar( Data, Element_Size, 0 )
	else
	    Tdarray.Array_Get( Offset - Low, Data )
end ;


procedure Dynamic_Array.Put( Offset : Integer ; var Data ) ;
{Put an element in the array}

var Temp : Array_Pointer ;

begin
    if Offset > High then High := Offset ;
    if Offset < Low then
    begin
	Getmem( Temp, ( High - Offset + 1 ) * Element_Size ) ;
	{Allocate another chunk of memory}
	Fillchar( Temp^[ 0 ], ( High - Offset + 1 ) * Element_Size, 0 ) ; {Zero it}
	Move( Ptr^[ 0 ], Temp^[ ( Low - Offset ) * Element_Size ],
	    Element_Count * Element_Size ) ; {Copy old data}
	Freemem( Ptr, Element_Count * Element_Size ) ; {Free up old memory}
	Element_Count := High - Offset + 1 ; {New element count}
	Ptr := Temp ; {New pointer to data}
	Low := Offset {Adjust low offset}
    end ;
    Tdarray.Array_Put( Offset - Low, Data )
end ;


			{Abstract dynamic stack object routines}

constructor Tdstack.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;

begin
    inherited Init( Cnt, Size, Flag ) ;
end ;


procedure Tdstack.Push( var Data ) ;

begin
    Terminate( 'TDStack.Push' )
end ;


function Tdstack.Empty : boolean ;

begin
    Empty := True ;
    Terminate( 'TDStack.Empty' ) ;
end ;


procedure Tdstack.Top( var Data ) ;

begin
    Terminate( 'TDSTack.Top' )
end ;


procedure Tdstack.Pop( var Data ) ;

begin
    Terminate( 'TDSTack.Pop' )
end ;


procedure Tdstack.Array_Get( Offset : TRecord_Size ; var Data ) ;

begin
    Terminate1( 'Dynamic stack', 'Array_Get' )
end ;


procedure Tdstack.Array_Put( Offset : TRecord_Size ; var Data ) ;

begin
    Terminate1( 'Dynamic stack', 'Array_Put' )
end ;


procedure Tdstack.Get( Offset : Integer ; var Data ) ;

begin
    Terminate1( 'Dynamic stack', 'Get' )
end ;


procedure Tdstack.Put( Offset : Integer ; var Data ) ;

begin
    Terminate1( 'Dynamic stack', 'Put' )
end ;



			{DYNAMIC STACK object routines}

constructor Dynamic_Stack.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;

begin
    if( Cnt > Max_Memory ) then
    begin
	Writeln( 'Maximum dynamic stack size is exceeded' ) ;
	Halt
    end ;
    Tdarray.Init( Cnt, Size, Flag ) ; {Initialize an array}
    Stack_Top := 0 {Nothing on stack}
end ;


procedure Dynamic_Stack.Push( var Data ) ;

begin
    if(  Stack_Top > Max_Memory ) then
    begin
	Writeln( 'Stack overflow in Dynamic_Stack.Push' ) ;
	Halt
    end ;
    Tdarray.Array_Put( Stack_Top, Data ) ; {Push onto top of stack}
    Stack_Top := Stack_Top + 1
end ;


function Dynamic_Stack.Empty : boolean ;

begin
    if Stack_Top = 0
	then
	    Empty := True
	else
	    Empty := False
end ;


procedure Dynamic_Stack.Top( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in Dynamic_Stack.Top' ) ;
	Halt
    end ;
    Tdarray.Array_Get( Stack_Top - 1, Data ) {Return data}
end ;


procedure Dynamic_Stack.Pop( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in Dynamic_Stack.Pop' ) ;
	Halt
    end ;
    Stack_Top := Stack_Top - 1 ;
    Tdarray.Array_Get( Stack_Top, Data ) {Return data}
end ;


			{Standard virtual ARRAY object routines}

constructor Tvarray.Init( Cnt, Size, Strt : TRecord_Size ; Flag : boolean ; Nam : string ) ;
{Initialize array}

var Dummy : Integer ;
    Dummy1 : TRecord_Size ;
    Ptr : Array_Pointer ; {Dummy variables}

begin
    if Flag and ( Size <> 4 ) then {array contains objects, but size is not 4}
    begin
	Writeln( 'Objects in arrays must be size 4' ) ;
	Halt
    end ;
    if( Size > Max_Memory - 9 ) then
    begin
	Writeln( 'Maximum size of array object data is exceeded' ) ;
	Halt
    end ;
    Tarray.Init ; {Initialize object}
    Element_Count := Cnt ; {Number of elements to store}
    Element_Size := Size ; {Size of elements being stored}
    Start := Strt ;
    Storing_Objects := Flag ;
    Assign( Image, Nam ) ;
    {$I-}
    Reset( Image, Element_Size ) ; {Open image file}
    {$I+}
    Dummy := Ioresult ; {Get open status}
    if Dummy = 2 then {File not found}
    begin
	{$I-}
	Rewrite( Image, Element_Size ) ; {Create it}
	{$I+}
	Dummy := Ioresult
    end ;
    if Dummy <> 0 then
    begin
	Writeln( Ert( Dummy ), ' while opening virtual array image file ', Nam ) ;
	Halt
    end ;
    if Filesize( Image ) < Start then {Need to pad the file}
    begin
	Seek( Image, Filesize( Image ) ) ; {Go to end of file}
	{$I-}
	Blockwrite( Image, Dummy, Element_Count + 1 ) ; {Fill with anything}
	{$I+}
	Dummy := Ioresult ; {Get output status}
	if Dummy <> 0 then {Error}
	begin
            Application_Manager^.Fatal_Exception( Ert( Dummy ) + ' while padding virtual array image file' ) ;
	end ;
    end ;
    Seek( Image, Start ) ; {Go to beginning position of array in file}
    Getmem( Ptr, Element_Size ) ; {Create temporary buffer to hold null image}
    Fillchar( Ptr^[ 0 ], Element_Size, 0 ) ; {Clear buffer}
    for Dummy1 := 1 to Element_Count do
    begin
        Blockwrite( Image, Ptr^[ 0 ], 1 ) ;
    end ;
    Freemem( Ptr, Element_Size ) {Now delete temporary buffer}
end ;


procedure Tvarray.Array_Get( Offset : TRecord_Size ; var Data ) ;
{Get an element from the array}

begin
    if Offset >= Element_Count then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in TVArray.Array_Get' ) ;
	Halt
    end ;
    Seek( Image, Start + Offset ) ;
    Blockread( Image, Data, 1 ) {Return value}
end ;


procedure Tvarray.Array_Put( Offset : TRecord_Size ; var Data ) ;
{ Put an element in the array }

begin
    if Offset >= Element_Count then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in TVArray.Array_Put' ) ;
	Halt ;
    end ;
    Seek( Image, Start + Offset ) ;
    Blockwrite( Image, Data, 1 ) { Write value }
end ;


destructor Tvarray.Done ; {Delete entire array}

Type Tvarray_Pointer = ^Tvarray ;

var Offset : Integer ;
    Object_Pointer : Pointer ;

begin
    if Storing_Objects then {if storing objects in array}
    begin
	Seek( Image, Start ) ; {Start at proper position in file}
	For Offset := 0 To Element_Count - 1 do {Check each element}
	begin
	    Blockread( Image, Object_Pointer, 1 ) ; {Get pointer}
	    if Object_Pointer <> nil then {if an object is present in element}
		Dispose( Tvarray_Pointer( Object_Pointer ), Done )
		{Dispose of object}
	end ;
    end ;
    inherited Done ;
end ;


			{VIRTUAL ARRAY object routines}

constructor Virtual_Array.Init( Lbound, Ubound : Integer ; Size, Strt : TRecord_Size ;
    Flag : boolean ; Nam : string ) ; {Initialize array}

begin
    if Lbound > Ubound then
    begin
	Writeln( 'virtual array ', Lbound, '..', Ubound, ' is invalid' ) ;
	Halt
    end ;
    Tvarray.Init( Ubound - Lbound + 1, Size, Strt, Flag, Nam ) ; {Initialize object}
    Low := Lbound ;
    High := Ubound
end ;


procedure Virtual_Array.Get( Offset : Integer ; var Data ) ;
{Get an element from the array}

begin
    if ( Offset > High ) Or ( Offset < Low ) then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in Virtual_Array.Get' ) ;
	Halt
    end ;
    Tvarray.Array_Get( Offset - Low, Data )
end ;


procedure Virtual_Array.Put( Offset : Integer ; var Data ) ;
{Put an element in the array}

begin
    if ( Offset > High ) Or ( Offset < Low ) then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in Virtual_Array.Put' ) ;
	Halt
    end ;
    Tvarray.Array_Put( Offset - Low, Data )
end ;


			{Abstract virtual stack object routines}

constructor Tvstack.Init( Cnt, Size, Strt : TRecord_Size ; Flag : boolean ;
    Nam : string ) ;

begin
    inherited Init( Cnt, Size, Strt, Flag, Nam ) ;
end ;


procedure Tvstack.Push( var Data ) ;

begin
    Terminate( 'TVStack.Push' )
end ;


function Tvstack.Empty : boolean ;

begin
    Empty := True ;
    Terminate( 'TVStack.Empty' ) ;
end ;


procedure Tvstack.Top( var Data ) ;

begin
    Terminate( 'TVSTack.Top' )
end ;


procedure Tvstack.Pop( var Data ) ;

begin
    Terminate( 'TVSTack.Pop' )
end ;


procedure Tvstack.Array_Get( Offset : TRecord_Size ; var Data ) ;

begin
    Terminate1( 'virtual stack', 'Array_Get' )
end ;


procedure Tvstack.Array_Put( Offset : TRecord_Size ; var Data ) ;

begin
    Terminate1( 'virtual stack', 'Array_Put' )
end ;


procedure Tvstack.Get( Offset : Integer ; var Data ) ;

begin
    Terminate1( 'virtual stack', 'Get' )
end ;


procedure Tvstack.Put( Offset : Integer ; var Data ) ;

begin
    Terminate1( 'virtual stack', 'Put' )
end ;



			{VIRTUAL STACK object routines}

constructor Virtual_Stack.Init( Cnt, Size, Strt : TRecord_Size ; Flag : boolean ;
    Nam : string ) ;

begin
    if( Cnt > Max_Memory ) then
    begin
	Writeln( 'Maximum virtual stack size is exceeded' ) ;
	Halt
    end ;
    Tvarray.Init( Cnt, Size, Strt, Flag, Nam ) ; {Initialize an array}
    Stack_Top := 0 {Nothing on stack}
end ;


procedure Virtual_Stack.Push( var Data ) ;

begin
    if( Stack_Top > Max_Memory ) then
    begin
	Writeln( 'Stack overflow in Virtual_Stack.Push' ) ;
	Halt
    end ;
    Tvarray.Array_Put( Stack_Top, Data ) ; {Push onto top of stack}
    Stack_Top := Stack_Top + 1
end ;


function Virtual_Stack.Empty : boolean ;

begin
    if Stack_Top = 0
	then
	    Empty := True
	else
	    Empty := False
end ;


procedure Virtual_Stack.Top( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in Virtual_Stack.Top' ) ;
	Halt
    end ;
    Tvarray.Array_Get( Stack_Top - 1, Data ) {Return data}
end ;


procedure Virtual_Stack.Pop( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in Virtual_Stack.Pop' ) ;
	Halt
    end ;
    Stack_Top := Stack_Top - 1 ;
    Tvarray.Array_Get( Stack_Top, Data ) {Return data}
end ;


			{Abstract single-linked list object routines}

constructor Tlist.Init ;

begin
    inherited Init ;
end ;


function Tlist.Empty : boolean ;

begin
    Empty := True ;
    Terminate( 'TList.Empty' ) ;
end ;


function Tlist.List_Last : List_Pointer ;

begin
    List_last := 0 ;
    Terminate( 'TList.List_Last' )
end ;


function Tlist.List_First : List_Pointer ;

begin
    List_First := 0 ;
    Terminate( 'TList.List_First' )
end ;


function Tlist.List_Next( Index : List_Pointer ) : List_Pointer ;

begin
    List_Next := 0 ;
    Terminate( 'TList.List_Next' ) ;
end ;


procedure Tlist.List_Add( var Data ) ;

begin
    Terminate( 'TList.List_Add' )
end ;


procedure Tlist.List_Get( Index : List_Pointer ; var Data ) ;

begin
    Terminate( 'TList.List_Get' )
end ;


procedure Tlist.List_Insert( Index : List_Pointer ; var Data ) ;

begin
    Terminate( 'TList.List_Insert' )
end ;


procedure Tlist.List_Delete( Index : List_Pointer ) ;

begin
    Terminate( 'TList.List_delete' )
end ;


destructor Tlist.Done ;

begin
    inherited Done ;
end ;


			{Standard single-linked list object routines}

constructor Tslist.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    Ptr := 0 ; {List is empty}
    Item_Size := Size ; {Get size of items}
    Storing_Objects := Flag ; {True if storing objects}
    if( Size > Max_Memory ) then
    begin
	Writeln( 'Maximum size of list object data is exceeded' ) ;
	Halt
    end ;
    if Storing_Objects and ( Size <> 4 ) then
    begin
	Writeln( 'Objects in lists must be size 4' ) ;
	Halt
    end
end ;


procedure TSList.Clear ;

type Tslist_Pointer = ^Tslist ;

var Tmp, Object_Pointer, Work : List_Pointer ;
    Image : List_Image ;

begin
    if( Ptr = 0 ) then
    begin
        Exit ; { Done if list is empty }
    end ;
    Heap^.Read_Heap( Ptr, sizeof( Image ), Image ) ;
    Tmp := Image.Next ; { Get first item in list }
    while( Tmp <> 0 ) do { Until deleted up to last item }
    begin
        Heap^.Read_Heap( Tmp, sizeof( Image ), Image ) ;
	if Storing_Objects then
	begin
	    Object_Pointer := Image.Last ; { Get object pointer }
	    Dispose( Tslist_Pointer( Object_Pointer ), Done ) ;
	end ;
	Work := Image.Next ; { Save link to next item }
        Heap^.Deallocate( Tmp, Item_Size + sizeof( Image ) ) ; { Delete from heap }
	if( Tmp = Ptr ) { if just deleted last item }
	    then
		Tmp := 0 { Done }
	    else
		Tmp := Work ; {Point to next item}
    end ;
    PTR := 0 ;
end ;


function Tslist.Empty : boolean ;

begin
    if Ptr = 0 {Nothing in list}
	then
	    Empty := True
	else
	    Empty := False
end ;


procedure Tslist.List_Add( var Data ) ;

var P, Tmp : List_Pointer ;
    Image : List_Image ;

begin
    if Ptr = 0 then
    begin
        Ptr := Heap^.Allocate( Item_Size + sizeof( List_Image ) ) ; { Allocate first item }
        fillchar( Image, sizeof( Image ), 0 ) ;
        Image.Next := Ptr ; { Only item=last item, which points to first item }
        Heap^.Write_Heap( Ptr, sizeof( Image ), Image ) ;
    end else
    begin
        Heap^.Read_Heap( Ptr, sizeof( Image ), Image ) ;
        Tmp := Image.Next ; { Save pointer to first item on list }
        Image.Next := Heap^.Allocate( Item_Size + sizeof( Image ) ) ; {Allocate a new item}
        Heap^.Write_Heap( Ptr, sizeof( Image ), Image ) ;
        P := Image.Next ;
        Heap^.Read_Heap( P, sizeof( Image ), Image ) ;
        Image.Next := Tmp ; { Point last item to first }
        Heap^.Write_Heap( P, sizeof( Image ), Image ) ;
        Ptr := P ; { Move to end of list }
    end ;
    Heap^.Write_Heap( Image.Last + sizeof( List_Image ), Item_Size, Data ) ;
    { Write data to item }
end ;


procedure Tslist.List_Get( Index : List_Pointer ; var Data ) ;

var Image : List_Image ;

begin
    Heap^.Read_Heap( Index, sizeof( Image ), Image ) ;
    Heap^.Read_Heap( Image.Last + sizeof( Image ), Item_Size, Data ) ; { Return data }
end ;


procedure TSList.List_Insert( Index : List_Pointer ; var Data ) ;
{ Insert after Index^ }

var P, Tmp : List_Pointer ;
    Image : List_Image ;

begin
    Tmp := Heap^.Allocate( Item_Size + sizeof( Image ) ) ; { Get the memory }
    Heap^.Write_Heap( Tmp + sizeof( Image ), Item_Size, Data ) ; { Copy data to item }
    if( Index = 0 ) then { Inserting at beginning of list }
    begin
	if( Ptr = 0 ) then { Previously empty }
	begin
	    Ptr := Tmp ;
	    Image.Next := Ptr ; { Point to oneself }
            Heap^.Write_Heap( Tmp, sizeof( Image ), Image ) ;
	end else
	begin
            Heap^.Read_Heap( Ptr, sizeof( Image ), Image ) ;
            P := Image.Next ;
            Image.Next := Tmp ;
            Heap^.Write_Heap( Ptr, sizeof( Image ), Image ) ; { Last item points to it now }
            Image.Next := P ;
            Heap^.Write_Heap( Tmp, sizeof( Image ), Image ) ; { Link to first item on list }
	end
    end else
    begin
        Heap^.Read_Heap( Index, sizeof( Image ), Image ) ;
        P := Image.Next ;
        Image.Next := Tmp ;
        Heap^.Write_Heap( Index, sizeof( Image ), Image ) ;
	Image.Next := P ; { Link item into list }
        Heap^.Write_Heap( Tmp, sizeof( Image ), Image ) ;
	if( Index = Ptr ) then
        begin
            Ptr := Tmp ; { Inserting at end of list }
        end ;
    end
end ;


procedure Tslist.List_Delete( Index : List_Pointer ) ;

Type Tslist_Pointer = ^Tslist ;

var Tmp, Work : List_Pointer ;
    Tmp_Image, Work_Image : List_Image ;

begin
    Tmp := Index ;
    Heap^.Read_Heap( Tmp, sizeof( Tmp_Image ), Tmp_Image ) ;
    Work := Tmp ;
    Heap^.Read_Heap( Work, sizeof( Work_Image ), Work_Image ) ;
    while( Work_Image.Next <> Tmp ) do
    begin
        Work := Work_Image.Next ; { Follow links to previous item }
        Heap^.Read_Heap( Work, sizeof( Work_Image ), Work_Image ) ;
    end ;
    { At this point, Tmp^ is the item to delete and Work^ is the previous item}
    Work_Image.Next := Tmp_Image.Next ; { Remove item from linked list }
    Heap^.Write_Heap( Work, sizeof( Work_Image ), Work_Image ) ;
    if Storing_Objects then
    begin
        Dispose( Tslist_Pointer( Tmp_Image.Last ), Done ) ;
    end ;
    if Work = Tmp
	then
	    Ptr := 0 { Deleted only item in list }
	else
	    if Ptr = Tmp then
            begin
                Ptr := Work ; { Deleted last item on list }
            end ;
    Heap^.Deallocate( Tmp, Item_Size + sizeof( Tmp_Image ) ) ; { Delete item from heap }
end ;


function Tslist.List_Last : List_Pointer ;

begin
    List_Last := Ptr ;
end ;


function Tslist.List_First : List_Pointer ;

var Image : List_Image ;

begin
    if( Ptr = 0 ) then
    begin
        List_First := 0 ; { No first if list is null }
    end else
    begin
        Heap^.Read_Heap( Ptr, sizeof( Image ), Image ) ;
	List_First := Image.Next ;
    end ;
end ;


function Tslist.List_Next( Index : List_Pointer ) : List_Pointer ;

var Image : List_Image ;

begin
    if( ( Index = Ptr ) or ( Ptr = 0 ) ) then { List is null or at end of list }
    begin
	List_Next := 0 ;
    end else
    begin
        Heap^.Read_Heap( Index, sizeof( Image ), Image ) ;
	List_Next := Image.Next ; { Return pointer to next item }
    end ;
end ;


destructor Tslist.Done ;

begin
    Clear ;
    inherited Done ;
end ;


			{SINGLE-LINKED LIST object routines}

constructor Single_List.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    Tslist.Init( Size, Flag ) ; {Initialize list}
    Context := 0 ; {No current context}
end ;


procedure Single_List.Add( var Data ) ;

begin
    Tslist.List_Add( Data ) ;
    Context := Ptr { Context is now last item in list }
end ;


procedure Single_List.Get( var Data ) ; { Return current item }

begin
    if( Context = 0 ) { Context = BOL }
	then
	    Context := Tslist.List_First { Get first item }
	else
	    Context := Tslist.List_Next( Context ) ; { Go to next item }
    if( Context = 0 ) then
    begin
	Writeln( 'end of list in Single_List.Get' ) ;
	Halt ;
    end ;
    Tslist.List_Get( Context, Data ) ;
end ;


procedure Single_List.Insert( var Data ) ;

begin
    Tslist.List_Insert( Context, Data ) ; { Insert item }
    if( Context = 0 )
	then
	    Context := Tslist.List_First
	else
	    Context := Tslist.List_Next( Context ) ; { Point to new item }
end ;


procedure Single_List.Delete ;

begin
    if( Context = 0 ) then { No current context }
    begin
	Writeln( 'No context in Single_List.Delete' ) ;
	Halt ;
    end ;
    Tslist.List_Delete( Context ) ; { Delete current item }
    Context := 0 ; { Reset to beginning of list }
end ;


procedure Single_List.Last( var Data ) ;

begin
    Context := Ptr ; { Move context to first item }
    Tslist.List_Get( Context, Data ) ; { Return item }
end ;


procedure Single_List.First( var Data ) ;

begin
    Context := Tslist.List_First ; { Move context to first item }
    Tslist.List_Get( Context, Data ) ; { Return item }
end ;


function Single_List.Eol : boolean ;

begin
    if Context = Ptr { At end of list? }
	then
	    Eol := True
	else
	    Eol := False ;
end ;


function Single_List.Bol : boolean ;

var Image : List_Image ;

begin
    if( Context = 0 ) then
    begin
        BOL := True ;
    end else
    begin
        Heap^.Read_Heap( Ptr, sizeof( Image), Image ) ;
        BOL := ( Context = Image.Next ) ;
    end ;
end ;


			{Abstract list stack object routines}

constructor Tlstack.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    inherited Init( Size, Flag ) ;
end ;


procedure Tlstack.Push( var Data ) ;

begin
    Terminate( 'TLSTack.Push' )
end ;


procedure Tlstack.Top( var Data ) ;

begin
    Terminate( 'TLSTack.Top' )
end ;


procedure Tlstack.Pop( var Data ) ;

begin
    Terminate( 'TLSTack.Pop' )
end ;


procedure Tlstack.List_Add( var Data ) ;

begin
    Terminate1( 'List Stack', 'List_Add' )
end ;


procedure Tlstack.List_Get( Index : List_Pointer ; var Data ) ;

begin
    Terminate1( 'List Stack', 'List_Get' )
end ;


procedure Tlstack.List_Insert( Index : List_Pointer ; var Data ) ;

begin
    Terminate1( 'List Stack', 'List_Insert' )
end ;


function Tlstack.List_Next( Index : List_Pointer ) : List_Pointer ;

begin
    List_Next := 0 ;
    Terminate1( 'List Stack', 'List_Next' ) ;
end ;


procedure Tlstack.List_Delete( Index : List_Pointer ) ;

begin
    Terminate1( 'List Stack', 'List_Delete' )
end ;


function Tlstack.List_First : List_Pointer ;

begin
    List_First := 0 ;
    Terminate1( 'List Stack', 'List_First' ) ;
end ;


function Tlstack.List_Last : List_Pointer ;

begin
    List_Last := 0 ;
    Terminate1( 'List Stack', 'List_Last' ) ;
end ;


			{LIST STACK object routines}

constructor List_Stack.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    Tslist.Init( Size, Flag )
end ;


procedure List_Stack.Push( var Data ) ;

begin
    Tslist.List_Insert( 0, Data ) {Push onto top of stack}
end ;


procedure List_Stack.Top( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in List_Stack.Top' ) ;
	Halt ;
    end ;
    Tslist.List_Get( Tslist.List_First, Data ) ; { Get item from top of stack }
end ;


procedure List_Stack.Pop( var Data ) ;

var Tmp : List_Pointer ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in List_Stack.Pop' ) ;
	Halt
    end ;
    Tmp := Tslist.List_First ; {Get first item pointer}
    Tslist.List_Get( Tmp, Data ) ; {Return data}
    Tslist.List_Delete( Tmp ) {Delete item at top of stack}
end ;


			{Standard double-linked list object routines}

constructor Tdlist.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    Ptr := 0 ; { List is empty }
    Item_Size := Size ; { Get size of items }
    Storing_Objects := Flag ; { True if storing objects }
    if( Size > Max_Memory ) then
    begin { UNTESTED }
	Writeln( 'Maximum size of list object data is exceeded' ) ;
	Halt
    end ;
    if Storing_Objects and ( Size <> 4 ) then
    begin { UNTESTED }
	Writeln( 'Objects in lists must be size 4' ) ;
	Halt ;
    end ;
end ;


procedure TDList.Clear ;

Type Tdlist_Pointer = ^Tdlist ;

var Tmp, Object_Pointer, Work : List_Pointer ;
    Image : List_Image ;

begin
    if( Ptr = 0 ) then
    begin
        Exit ; {Done if list is empty}
    end ;
    Heap^.Read_Heap( Ptr, sizeof( Image ), Image ) ; { Get first item in list }
    Tmp := Image.Next ; {Get first item in list}
    while( Tmp <> 0 ) do { Until deleted up to last item }
    begin
        Heap^.Read_Heap( Tmp, sizeof( Image ), Image ) ;
	if Storing_Objects then
	begin { UNTESTED }
	    Object_Pointer := Image.Last ; { Get object pointer }
	    Dispose( Tdlist_Pointer( Object_Pointer ), Done ) ;
	end ;
	Work := Image.Next ; { Save link to next item }
        Heap^.Deallocate( Tmp, Item_Size + sizeof( Image ) ) ; { Delete from heap }
	if( Tmp = Ptr ) then { if just deleted last item }
        begin
	    Tmp := 0 ; { Done }
        end else
        begin
	    Tmp := Work ; { Point to next item }
        end ;
    end ;
    PTR := 0 ;
end ; { while( Tmp <> 0 ) }


function Tdlist.Empty : boolean ;

begin
    if Ptr = 0 { Nothing in list }
	then
	    Empty := True
	else
	    Empty := False ;
end ;


procedure Tdlist.List_Add( var Data ) ;

var Tmp : List_Pointer ;
    Image, Temp : List_Image ;

begin
    if( Ptr = 0 ) then
    begin
        Ptr := Heap^.Allocate( Item_Size + sizeof( Image ) ) ; { Allocate first item }
        Image.Last := Ptr ; { Only item=first item, which points to last item }
        Image.Next := Ptr ; { Only item=last item, which points to first item }
        Heap^.Write_Heap( Ptr, sizeof( Image ), Image ) ;
    end else
    begin
        Heap^.Read_Heap( Ptr, sizeof( Image ), Image ) ;
        Tmp := Image.Next ; { Save pointer to first item on list }
        Image.Next := Heap^.Allocate( Item_Size + sizeof( Image ) ) ; { Allocate a new item }
        Heap^.Write_Heap( Ptr, sizeof( Image ), Image ) ;
        Temp.Next := Tmp ; { Point last item forward to first }
        Temp.Last := Ptr ; { Point last item back to previous last }
        Heap^.Write_Heap( Image.Next, sizeof( Temp ), Temp ) ;
        Heap^.Read_Heap( Tmp, sizeof( Temp ), Temp ) ;
        Temp.Last := Image.Next ; { Point first item back to last }
        Heap^.Write_Heap( Tmp, sizeof( Temp ), Temp ) ;
        Ptr := Image.Next ; { Move to end of list }
    end ;
    Heap^.Write_Heap( Ptr + sizeof( Image ), Item_Size, Data ) ; { Write data to item }
end ;


procedure Tdlist.List_Get( Index : List_Pointer ; var Data ) ;

begin
    Heap^.Read_Heap( Index + sizeof( List_Image ), Item_Size, Data ) ;
end ;


procedure TDList.List_Insert( Index : List_Pointer ; var Data ) ;
{ Insert after Index }

var Tmp : List_Pointer ;
    Image, P, Temp : List_Image ;

begin
    Tmp := Heap^.Allocate( Item_Size + sizeof( Image ) ) ; { Get the memory }
    Heap^.Write_Heap( Tmp + sizeof( Image ), Item_Size, Data ) ; { Copy data to item }
    if( Index = 0 ) then { Inserting at beginning of list }
    begin
	if( Ptr = 0 ) then { Previously empty }
	begin
	    Ptr := Tmp ;
	    Image.Last := Ptr ;
	    Image.Next := Ptr ; { Point to oneself }
            Heap^.Write_Heap( Ptr, sizeof( Image ), Image ) ;
	end else
	begin
            Heap^.Read_Heap( Ptr, sizeof( P ), P ) ;
	    Image.Next := P.Next ; { Link to first item on list }
	    Image.Last := Ptr ;
            Heap^.Write_Heap( Tmp, sizeof( Image ), Image ) ;
            Heap^.Read_Heap( P.Next, sizeof( Temp ), Temp ) ;
	    Temp.Last := Tmp ;
            Heap^.Write_Heap( P.Next, sizeof( Temp ), Temp ) ;
	    P.Next := Tmp ; { Last item points to it now }
            Heap^.Write_Heap( Ptr, sizeof( P ), P ) ;
	end ;
    end else
    begin
        Heap^.Read_Heap( Index, sizeof( P ), P ) ;
	Image.Next := P.Next ; { Link item into list }
	Image.Last := Index ;
        Heap^.Write_Heap( Tmp, sizeof( Image ), Image ) ;
        Heap^.Read_Heap( P.Next, sizeof( Temp ), Temp ) ;
	Temp.Last := Tmp ;
        Heap^.Write_Heap( P.Next, sizeof( Temp ), Temp ) ;
        Heap^.Read_Heap( Index, sizeof( P ), P ) ;
	P.Next := Tmp ;
        Heap^.Write_Heap( Index, sizeof( P ), P ) ;
	if( Index = Ptr ) then
        begin
            Ptr := Tmp ; { Inserting at end of list }
        end ;
    end ;
end ; { Tdlist.List_Insert }


procedure TDList.List_Delete( Index : List_Pointer ) ;

var Point : ^Tdlist ;
    Image, Temp : List_Image ;

begin
    Heap^.Read_Heap( Index, sizeof( Image ), Image ) ;
    Heap^.Read_Heap( Image.Next, sizeof( Temp ), Temp ) ;
    Temp.Last := Image.Last ;
    Heap^.Write_Heap( Image.Next, sizeof( Temp ), Temp ) ;
    Heap^.Read_Heap( Image.Last, sizeof( Temp ), Temp ) ;
    Temp.Next := Image.Next ; { Remove item from linked list }
    Heap^.Write_Heap( Image.Last, sizeof( Temp ), Temp ) ;
    if Storing_Objects then
    begin { UNTESTED }
        Heap^.Read_Heap( Index + sizeof( Image ), 4, Point ) ; { Get object pointer }
	Dispose( Point, Done ) ; { Delete object }
    end ;
    if( Image.Next = Index ) then
    begin
        Ptr := 0 { Deleted only item in list }
    end else
    begin
        if( Ptr = Index ) then
        begin
            Ptr := Image.Last ; { Deleted last item on list }
        end ;
    end ;
    Heap^.Deallocate( Index, Item_Size + sizeof( Image ) ) ; { Delete item from heap }
end ;


function TDList.List_Last : List_Pointer ;

begin
    List_Last := Ptr ;
end ;


function TDList.List_First : List_Pointer ;

var Image : List_Image ;

begin
    if( Ptr = 0 ) then
    begin
        List_First := 0 ; { No first if list is null }
    end else
    begin
        Heap^.Read_Heap( Ptr, sizeof( Image ), Image ) ;
	List_First := Image.Next ;
    end ;
end ;


function TDList.List_Previous( Index : List_Pointer ) : List_Pointer ;

var Image : List_Image ;

begin
    if( Ptr = 0 ) then { List is null }
    begin
	List_Previous := 0
    end else
    begin
        Heap^.Read_Heap( Ptr, sizeof( Image ), Image ) ;
        if( Index = Image.Next ) then { At beginning of list }
        begin
            List_Previous := 0 ;
        end else
        begin
            Heap^.Read_Heap( Index, sizeof( Image ), Image ) ;
            List_Previous := Image.Last ; { Return pointer to last item }
        end ;
    end ;
end ;


function TDList.List_Next( Index : List_Pointer ) : List_Pointer ;

var Image : List_Image ;

begin
    if( ( Index = Ptr ) or ( Ptr = 0 ) ) then { List is null or at end of list }
    begin
        List_Next := 0
    end else
    begin
        Heap^.Read_Heap( Index, sizeof( Image ), Image ) ;
	List_Next := Image.Next ; { Return pointer to next item }
    end ;
end ;


destructor TDList.Done ;

begin
    Clear ;
    inherited Done ;
end ;


{ Double_List methods... }

constructor Double_List.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    TDList.Init( Size, Flag ) ; { Initialize list }
    Context := 0 ; { No current context }
end ;


procedure Double_List.Add( var Data ) ;

begin
    List_Add( Data ) ;
    Context := Ptr ; { Context is now last item in list }
end ;


procedure Double_List.Clear ;

begin
    inherited Clear ;

    Context := 0 ; { No more context }
end ;


procedure Double_List.Put( var Data ) ; { Update current item }

begin
    if( Context = 0 ) then
    begin { UNTESTED }
	Writeln( 'End of list in Double_List.Get' ) ;
	Halt ;
    end ;
    Heap^.Write_Heap( Context + sizeof( List_Image ), Item_Size, Data ) ;
end ;


procedure Double_List.Get( var Data ) ; { Return current item }

begin
    if( Context = 0 ) then { Context = BOL }
    begin
	Context := List_First ; { Get first item }
    end else
    begin
	Context := List_Next( Context ) ; { Go to next item }
    end ;
    if( Context = 0 ) then
    begin
	Writeln( 'End of list in Double_List.Get' ) ;
	Halt ;
    end ;
    List_Get( Context, Data ) ;
end ;


procedure Double_List.Previous( var Data ) ; { Get previous item on list }

begin
    if( Context <> 0 ) then
    begin
        Context := List_Previous( Context ) ;
    end ;
    { Go to previous item }
    if( Context = 0 ) then
    begin { UNTESTED }
	Writeln( 'Beginning of list in Double_List.Get' ) ;
	Halt ;
    end ;
    List_Get( Context, Data ) ;
end ;


procedure Double_List.Insert( var Data ) ;

begin
    List_Insert( Context, Data ) ; { Insert item }
    if( Context = 0 ) then
    begin
	Context := List_First ;
    end else
    begin
	Context := List_Next( Context ) ; { Point to new item }
    end ;
end ;


procedure Double_List.Delete ;

begin
    if Context = 0 then { No current context }
    begin
	Writeln( 'No context in Double_List.Delete' ) ;
	Halt ;
    end ;
    List_Delete( Context ) ; { Delete current item }
    Context := 0 ; { Reset to beginning of list }
end ;


procedure Double_List.Last( var Data ) ;

begin
    if( Ptr = 0 ) then { List is empty }
    begin
	Writeln( 'Empty list in Double_List.Last' ) ;
	Halt ;
    end ;
    Context := List_Last ; { Move context to last item }
    List_Get( Context, Data ) { Return item }
end ;


procedure Double_List.First( var Data ) ;

begin
    if( Ptr = 0 ) then { List is empty }
    begin
	Writeln( 'Empty list in Double_List.First' ) ;
	Halt ;
    end ;
    Context := Tdlist.List_First ; { Move context to first item }
    List_Get( Context, Data ) ; { Return item }
end ;


function Double_List.Eol : boolean ;

begin
    if Context = Ptr { At end of list? }
	then
	    Eol := True
	else
	    Eol := False ;
end ;


function Double_List.Bol : boolean ;

var Image : List_Image ;

begin
    if ( Context = 0 ) then
    begin
        BOL := True
    end else
    begin
        Heap^.Read_Heap( Context, sizeof( Image ), Image ) ;
        BOL := ( Context = Image.Next ) ; { At beginning of list? }
    end ;
end ;


			{Standard queue object routines}

constructor Tqueue.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    inherited Init( Size, Flag ) ;
end ;


procedure Tqueue.Add( var Data ) ;

begin
    Terminate( 'TQueue.Add' )
end ;


procedure Tqueue.Front( var Data ) ;

begin
    Terminate( 'TQueue.Front' )
end ;


procedure Tqueue.Delete( var Data ) ;

begin
    Terminate( 'TQueue.Delete' )
end ;


procedure Tqueue.List_Add( var Data ) ;

begin
    Terminate1( 'Queue', 'List_Add' )
end ;


procedure Tqueue.List_Get( Index : List_Pointer ; var Data ) ;

begin
    Terminate1( 'Queue', 'List_Get' )
end ;


procedure Tqueue.List_Insert( Index : List_Pointer ; var Data ) ;

begin
    Terminate1( 'Queue', 'List_Insert' )
end ;


function Tqueue.List_Next( Index : List_Pointer ) : List_Pointer ;

begin
    List_Next := 0 ;
    Terminate1( 'Queue', 'List_Next' ) ;
end ;


procedure Tqueue.List_Delete( Index : List_Pointer ) ;

begin
    Terminate1( 'Queue', 'List_Delete' )
end ;


function Tqueue.List_First : List_Pointer ;

begin
    List_First := 0 ;
    Terminate1( 'Queue', 'List_First' ) ;
end ;


function Tqueue.List_Last : List_Pointer ;

begin
    List_Last := 0 ;
    Terminate1( 'Queue', 'List_Last' ) ;
end ;


			{QUEUE object routines}

constructor Queue.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    Tdlist.Init( Size, Flag )
end ;


procedure Queue.Add( var Data ) ;

begin
    Tdlist.List_Add( Data )
end ;


procedure Queue.Front( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Queue is empty in Queue.Front' ) ;
	Halt
    end ;
    Tdlist.List_Get( Tdlist.List_First, Data ) {Get front of queue}
end ;


procedure Queue.Delete( var Data ) ;

var Tmp : List_Pointer ;

begin
    if Empty then
    begin
	Writeln( 'Queue is empty in Queue.Delete' ) ;
	Halt
    end ;
    Tmp := Tdlist.List_First ;
    Tdlist.List_Get( Tmp, Data ) ; {Return item from queue}
    Tdlist.List_Delete( Tmp ) {...and delete from queue}
end ;


			{Standard sparse array object routines}

constructor Sparse_Array.Init( Size : TRecord_Size ; Flag : boolean ; var Def ) ;

begin
    if Flag and ( Size <> 4 ) then {array contains objects, but size is not 4}
    begin
	Writeln( 'Objects in arrays must be size 4' ) ;
	Halt ;
    end ;
    if( Size > Max_Memory ) then
    begin
	Writeln( 'Maximum size of array object data is exceeded' ) ;
	Halt ;
    end ;
    Ptr := 0 ; {array starts out empty}
    Item_Size := Size ;
    Storing_Objects := Flag ;
    Default := Heap^.Allocate( Size ) ;
    Heap^.Write_Heap( Default, Size, Def ) ; { Write default value for later }
end ;


procedure Sparse_Array.List_Add( var Data ) ;

begin
    Terminate1( 'Sparse array', 'List_Add' )
end ;


procedure Sparse_Array.List_Get( Index : List_Pointer ; var Data ) ;

begin
    Terminate1( 'Sparse array', 'List_Get' )
end ;


procedure Sparse_Array.List_Insert( Index : List_Pointer ; var Data ) ;

begin
    Terminate1( 'Sparse array', 'List_Insert' )
end ;


function Sparse_Array.List_Next( Index : List_Pointer ) : List_Pointer ;

begin
    List_Next := 0 ;
    Terminate1( 'Sparse array', 'List_Next' ) ;
end ;


procedure Sparse_Array.List_Delete( Index : List_Pointer ) ;

begin
    Terminate1( 'Sparse array', 'List_Delete' )
end ;


function Sparse_Array.List_First : List_Pointer ;

begin
    List_First := 0 ;
    Terminate1( 'Sparse array', 'List_First' ) ;
end ;


function Sparse_Array.List_Last : List_Pointer ;

begin
    List_Last := 0 ;
    Terminate1( 'Sparse array', 'List_Last' ) ;
end ;


procedure Sparse_Array.Get( Offset : Longint ; var Data ) ;

var Current : List_Pointer ;
    Image : List_Image ;

begin
    if Ptr <> 0 then { Something in array }
    begin
        Heap^.Read_Heap( Ptr, sizeof( List_Pointer ), Image ) ;
	Current := Image.Next ; { Get first element in array }
        Heap^.Read_Heap( Current, sizeof( Image ), Image ) ;
	while( ( Image.Last < Offset ) and ( Current <> Ptr ) ) do
        begin
	    Current := Image.Next ; { Find item }
            Heap^.Read_Heap( Current, sizeof( Image ), Image ) ;
        end ;
	if Longint( Image.Last ) = Offset then
	begin
            Heap^.Read_Heap( Current + sizeof( Image ), Item_Size, Data ) ; { Return item }
	    Exit ;
	end ;
    end ;
    Heap^.Read_Heap( Default, Item_Size, Data ) ; { Return default }
end ;


procedure Sparse_Array.Put( Offset : Longint ; var Data ) ;

var Current, Previous : List_Pointer ;
    Image : List_Image ;
    Tmp : List_Pointer ;

begin
    Current := 0 ;
    Previous := Ptr ;
    if( Ptr <> 0 ) then { Something in array }
    begin
        Heap^.Read_Heap( Ptr, sizeof( Image ), Image ) ;
	Current := Image.Next ; { Get first element in array }
	Previous := 0 ;
        Heap^.Read_Heap( Current, sizeof( Image ), Image ) ;
	while( ( Image.Last < Offset ) and ( Current <> Ptr ) ) do
	begin
	    Previous := Current ; { Save current location }
	    Current := Image.Next ; { Go to next item }
            Heap^.Read_Heap( Current, sizeof( Image ), Image ) ;
	end ;
	if Image.Last = Offset then { Element already exists }
	begin
            Heap^.Write_Heap( Current + sizeof( Image ), Item_Size, Data ) ; { Set item }
	    Exit ;
	end ;
	if( Current = Ptr ) then { Jump to next one if end of list }
	begin
            Heap^.Read_Heap( Ptr, sizeof( Image ), Image ) ;
	    Current := Image.Next ;
	    Previous := Ptr ;
	end ;
    end ; { if ( Ptr <> 0 ) }

    { At this point, the specified offset does not exist }
    Tmp := Heap^.Allocate( Item_Size + sizeof( Image ) ) ; { Allocate memory for element }
    Image.Last := List_Pointer( Offset ) ; { Set element number }
    Heap^.Write_Heap( Tmp + sizeof( Image ), Item_Size, Data ) ; { Set data }
    if Ptr = 0 then { Nothing in array }
    begin
	Image.Next := Tmp ; { Set pointer forward to self }
	Image.Last := List_Pointer( Offset ) ; { Set element number field }
	Ptr := Tmp ; { Last item in list is only item }
        Heap^.Write_Heap( Tmp, sizeof( Image ), Image ) ;
	Exit ;
    end ;
    Image.Next := Current ;
    Heap^.Write_Heap( Tmp, sizeof( Image ), Image ) ;
    if( Previous = 0 ) then
    begin
        Heap^.Read_Heap( Ptr, sizeof( Image ), Image ) ;
        Image.Next := Tmp ; { Insert item into beginning of list }
        Heap^.Write_Heap( Ptr, sizeof( Image ), Image ) ;
    end else
    begin
        Heap^.Read_Heap( Previous, sizeof( Image ), Image ) ;
        Image.Next := Tmp ; { Insert item into list }
        Heap^.Write_Heap( Previous, sizeof( Image ), Image ) ;
        if Previous = Ptr then Ptr := Tmp { Item is last on list }
    end ;
end ; { Sparse_Array.Put }


				{Abstract tree object routines}

constructor Ttree.Init ;

begin
    inherited Init ;
end ;


function Ttree.Tree_Find( var Key ) : Tree_Pointer ;

begin
    Tree_Find := nil ;
    Terminate( 'TTree.Tree_Find' ) ;
end ;


procedure Ttree.Tree_Get( Index : Tree_Pointer ; var Data ) ;

begin
    Terminate( 'TTree.Tree_Get' ) ;
end ;


procedure Ttree.Tree_Put( Index : Tree_Pointer ; var Data ) ;

begin
    Terminate( 'TTree.Tree_Put' ) ;
end ;


procedure Ttree.Tree_Add( var Data ) ;

begin
    Terminate( 'TTree.Add' )
end ;


procedure Ttree.Tree_Delete( Index : Tree_Pointer ) ;

begin
    Terminate( 'TTree.Tree_Delete' )
end ;


procedure Ttree.Tree_Prune( Index : Tree_Pointer ) ;

begin
    Terminate( 'TTree.Tree_Prune' )
end ;


function Ttree.Tree_First : Tree_Pointer ;

begin
    Tree_First := nil ;
    Terminate( 'TTree.Tree_First' ) ;
end ;


function Ttree.Tree_Last : Tree_Pointer ;

begin
    Tree_Last := nil ;
    Terminate( 'TTree.Tree_Last' ) ;
end ;


function Ttree.Empty : boolean ;

begin
    Empty := True ;
    Terminate( 'TTree.Empty' ) ;
end ;


function Ttree.Tree_Next( Index : Tree_Pointer ) : Tree_Pointer ;

begin
    Tree_Next := nil ;
    Terminate( 'TTree.Tree_Next' ) ;
end ;


function Ttree.Tree_Previous( Index : Tree_Pointer ) : Tree_Pointer ;

begin
    Tree_Previous := nil ;
    Terminate( 'TTree.Tree_Previous' ) ;
end ;


destructor Ttree.Done ;

begin
    inherited Done ;
end ;


			{Standard binary tree object routines}

constructor Stree.Init( Size, Key : TRecord_Size ; Flag : boolean ) ;

begin
    if( Flag and ( Size <> 4 ) ) then {Tree contains objects, but size is not 4}
    begin
	Writeln( 'Objects in trees must be size 4' ) ;
	Halt
    end ;
    if Key > Size then
    begin
	Writeln( 'Key size cannot exceed total tree object data size' ) ;
	Halt
    end ;
    if( Size > Max_Memory - 21 ) then
    begin
	Writeln( 'Maximum size of tree object data is exceeded' ) ;
	Halt
    end ;
    Ptr := nil ; {Tree is empty}
    Item_Size := Size ;
    Storing_Objects := Flag ;
    Key_Size := Key
end ;


procedure Stree.Tree_Get( Index : Tree_Pointer ; var Data ) ;

begin
    Move( Index^.D[ 0 ], Data, Item_Size ) {Return data from node}
end ;


procedure Stree.Tree_Put( Index : Tree_Pointer ; var Data ) ;

begin
    Move( Data, Index^.D[ 0 ], Item_Size ) {Return data from node}
end ;


function Stree.Tree_Find( var Key ) : Tree_Pointer ;

label Loop ;

var Current : Tree_Pointer ;
    Dummy : Integer ;

begin
    Current := Ptr ; {Start at root}
Loop :
    if Current <> nil then {Not yet at end}
    begin
	Dummy := Compare( Current^.D[ 0 ], Key, Key_Size ) ;
        if Dummy = 0 then {Found item in tree}
        begin
	    Tree_Find := Current ;
            Exit
        end ;
        if Dummy = 1
	    then
		Current := Current^.Left
	    else
		Current := Current^.Right ; {Follow link}
        goto Loop
    end ;
    Tree_Find := nil {Item not found}
end ;


procedure Stree.Add( var Data ) ;

label Loop ;

var Current, Previous : Tree_Pointer ;
    Dummy : Integer ;

begin
    if Ptr = nil then {Tree is empty}
    begin
	Getmem( Ptr, Item_Size + 12 ) ; {Allocate memory for node}
	Ptr^.Back := nil ;
	Ptr^.Left := nil ;
	Ptr^.Right := nil ;
	Move( Data, Ptr^.D[ 0 ], Item_Size ) ;
	Exit
    end ;
    Previous := nil ;
    Current := Ptr ; {Start at root}
Loop :
    if Current <> nil then {Not yet at end}
    begin
	Dummy := Compare( Current^.D[ 0 ], Data, Key_Size ) ;
        if Dummy = 0 then {Found item in tree}
        begin
            Application_Manager^.Fatal_Exception( 'Duplicate items encountered in tree' ) ;
        end ;
        Previous := Current ; {Save current pointer as previous}
        if Dummy = 1
	    then
		Current := Current^.Left
	    else
		Current := Current^.Right ; {Follow link}
        goto Loop
    end ;
    Getmem( Current, Item_Size + 12 ) ; {Create node}
    if Compare( Previous^.D[ 0 ], Data, Key_Size ) = 1
	then
	    Previous^.Left := Current
	else
	    Previous^.Right := Current ;
    Current^.Left := nil ;
    Current^.Right := nil ;
    Current^.Back := Previous ;
    Move( Data, Current^.D[ 0 ], Item_Size )
end ;

procedure Stree.Tree_Delete( Index : Tree_Pointer ) ;

var Tmp : Tree_Pointer ;

begin
    if ( Index^.Right <> nil ) and ( Index^.Left = nil ) then
    {Right branch but no left branch}
    begin
	Index^.Right^.Back := Index^.Back ;
	if Ptr = Index
	    then
		Ptr := Index^.Right {New root node}
	    else
		if Index^.Back^.Left = Index
		    then
			Index^.Back^.Left := Index^.Right
		    else
			Index^.Back^.Right := Index^.Right ;
	Freemem( Index, Item_Size + 12 ) ; {Delete node}
	Exit
    end ;
    if ( Index^.Left = nil ) and ( Index^.Right = nil ) then {Terminal node}
    begin
	if Ptr = Index
	    then
		Ptr := nil {Tree is empty}
	    else
		if Index^.Back^.Left = Index
		    then
			Index^.Back^.Left := nil
		    else
			Index^.Back^.Right := nil ;
	Freemem( Index, Item_Size + 12 ) ; {Delete node}
	Exit
    end ;
    {Bring up left branch}
    Index^.Left^.Back := Index^.Back ;
    if Ptr = Index
	then
	    Ptr := Index^.Left {New root node}
	else
	    if Index^.Back^.Left = Index
		then
		    Index^.Back^.Left := Index^.Left
		else
		    Index^.Back^.Right := Index^.Left ;
    if Index^.Right <> nil then {A right branch to handle}
    begin
	Tmp := Index^.Left ; {Start at replacement node}
	while Tmp^.Right <> nil do Tmp := Tmp^.Right ; {Go all the way right}
	Tmp^.Right := Index^.Right ;
	Tmp^.Right^.Back := Tmp
    end ;
    Freemem( Index, Item_Size + 12 ) {Actually delete node}
end ;


procedure Stree.Tree_Prune( Index : Tree_Pointer ) ;

begin
    while Index^.Left <> nil do Tree_Delete( Index^.Left ) ; {Delete left branch}
    while Index^.Right <> nil do Tree_Delete( Index^.Right ) ; {Delete right branch}
    Tree_Delete( Index ) {and delete sub-tree root node}
end ;


function Stree.Tree_First : Tree_Pointer ;

var Tmp : Tree_Pointer ;

begin
    if Ptr = nil then {Tree is empty}
    begin
	Tree_First := nil ;
	Exit
    end ;
    Tmp := Ptr ; {Start at root node}
    while Tmp^.Left <> nil do Tmp := Tmp^.Left ; {Go all the way left}
    Tree_First := Tmp {Return result}
end ;


function Stree.Tree_Last : Tree_Pointer ;

var Tmp : Tree_Pointer ;

begin
    if Ptr = nil then {Tree is empty}
    begin
	Tree_Last := nil ;
	Exit
    end ;
    Tmp := Ptr ;
    while Tmp^.Right <> nil do Tmp := Tmp^.Right ;
    Tree_Last := Tmp {Return result}
end ;


function Stree.Empty : boolean ;

begin
    if Ptr = nil
	then
	    Empty := True
	else
	    Empty := False
end ;


function Stree.Tree_Next( Index : Tree_Pointer ) : Tree_Pointer ;

label Loop ;

var Previous : Tree_Pointer ;

begin
    if( Ptr = nil ) then
    begin
	Tree_Next := nil ;
    end else
    begin
        if Index^.Right = nil then
        begin
Loop :
            if( Index^.Back = nil ) then {Cannot go further up tree}
            begin
                Tree_Next := nil ; {No more items in tree}
                Exit ;
            end ;
            Previous := Index ; {Save previous position}
            Index := Index^.Back ; {Backup one level}
            if Index^.Left = Previous then {Came from left branch}
            begin
                Tree_Next := Index ;
                Exit ;
            end else goto Loop {Backup again}
        end ;
        Index := Index^.Right ; {Go down right branch}
        while Index^.Left <> nil do Index := Index^.Left ; {Go left}
        Tree_Next := Index {Return result}
    end ;
end ;


function Stree.Tree_Previous( Index : Tree_Pointer ) : Tree_Pointer ;

label Loop ;

var Previous : Tree_Pointer ;

begin
    if( Ptr = nil ) then { Nothing in tree }
    begin
	Tree_Previous := nil ;
    end else
    begin
        if Index^.Left = nil then
        begin
Loop :
            if( Index^.Back = nil ) then {Cannot go further up tree}
            begin
                Tree_Previous := nil ; {No more items in tree}
                Exit ;
            end ;
            Previous := Index ; {Save current location in tree}
            Index := Index^.Back ; {Backup one level}
            if Index^.Right = Previous then {Came from right branch}
            begin
                Tree_Previous := Index ;
                Exit ;
            end else goto Loop
        end ;
        Index := Index^.Left ;
        while Index^.Right <> nil do Index := Index^.Right ;
        Tree_Previous := Index ; {Return new position}
    end ;
end ;


destructor Stree.Done ;

begin
    Tree_Prune( Ptr ) ; {Delete tree by pruning from root node}
    inherited Done ;
end ;


			{BINARY TREE object routines}

constructor Tree.Init( Size, Key : TRecord_Size ; Flag : boolean ) ;

begin
    Stree.Init( Size, Key, Flag ) ; {Initialize tree}
    Context := nil ; {No current context}
    Last_Error := 0 ; {No errors}
end ;


function Tree.Error ; {Return error code of last error}

begin
    Error := Last_Error ;
    Last_Error := 0 ;
end ;


procedure Tree.Set_Error( E : integer ) ; {Set an error code}

begin
    if Last_Error <> 0 then
    begin
	writeln( 'Tree error ', E ) ;
	halt ;
    end ;
    Last_Error := E ;
end ;


procedure Tree.Find( var Key, Data ) ;
{Find node with specified key (Key) and load return it in Data}

begin
    Context := Stree.Tree_Find( Key ) ; {Locate specified key}
    if Context = nil then
    begin
	Set_Error( - 155 ) ; {Key not found}
	exit ;
    end ;
    Stree.Tree_Get( Context, Data ) ; {Get item from tree}
end ;


procedure Tree.Update( var Data ) ;

begin
    if Context = nil then
    begin
	Set_Error( - 6 ) ; {No current node}
	exit ;
    end ;
    if Compare( Context^.D[ 0 ], Data, Key_Size ) <> 0 then {Key change}
    begin
	Set_Error( - 138 ) ;
	exit ;
    end ;
    Stree.Tree_Put( Context, Data ) ;
end ;


procedure Tree.Tree_Add( var Data ) ;

begin
    Stree.Add( Data ) ;
    Context := Stree.Tree_Find( Data ) ;
end ;


procedure Tree.Delete ;

begin
    if Context = nil then
    begin
	Writeln( 'No context in Tree.Delete' ) ;
	Halt
    end ;
    Stree.Tree_Delete( Context ) ;
    Context := nil {No more context}
end ;

procedure Tree.Prune ;

begin
    if Context = nil then
    begin
	Writeln( 'No context in Tree.Prune' ) ;
	Halt
    end ;
    Stree.Tree_Prune( Context ) ;
    Context := nil {No more context}
end ;


procedure Tree.First( var Data ) ;

begin
    Context := Stree.Tree_First ;
    if Context = nil then
    begin
	Set_Error( - 11 ) ; {Tree is empty}
	exit ;
    end ;
    Stree.Tree_Get( Context, Data ) ;
end ;


procedure Tree.Last( var Data ) ;

begin
    Context := Stree.Tree_Last ;
    if Context = nil then
    begin
	Writeln( 'Tree is empty in Tree.Last' ) ;
	Halt ;
    end ;
    Stree.Tree_Get( Context, Data ) ;
end ;


procedure Tree.Next( var Data ) ;
{Return node which follows current context node.}

begin
    Context := Stree.Tree_Next( Context ) ;
    if Context = nil then
    begin
	Set_Error( - 11 ) ; { End of tree }
	exit ;
    end ;
    Stree.Tree_Get( Context, Data ) ;
end ;


procedure Tree.Previous( var Data ) ;

begin
    Context := Stree.Tree_Previous( Context ) ;
    if Context = nil then
    begin
	Writeln( 'end of tree in Tree.Previous' ) ;
	Halt
    end ;
    Stree.Tree_Get( Context, Data )
end ;


		{ Standard binary tree object routines with large keys }

constructor XTree.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    if Flag and ( Size <> 4 ) then {Tree contains objects, but size is not 4}
    begin
	Writeln( 'Objects in trees must be size 4' ) ;
	Halt ;
    end ;
    if( Size > Max_Memory - 21 ) then
    begin
	Writeln( 'Maximum size of tree object data is exceeded' ) ;
	Halt ;
    end ;
    Ptr := nil ; {Tree is empty}
    Item_Size := Size ;
    Storing_Objects := Flag ;
end ;


procedure XTree.Tree_Get( Index : Tree_Pointer ; var Data ) ;

begin
    Move( Index^.D[ 4 ], Data, Item_Size ) {Return data from node}
end ;


procedure XTree.Tree_Put( Index : Tree_Pointer ; var Data ) ;

begin
    Move( Data, Index^.D[ 4 ], Item_Size ) {Return data from node}
end ;


function XTree.Find( Key : string ) : Tree_Pointer ;

label Loop ;

var Current : Tree_Pointer ;
    K : PSmall_String ;

begin
    Current := Ptr ; { Start at root }
Loop:
    if( Current <> nil ) then { Not yet at end }
    begin
	move( Current^.D[ 0 ], K, 4 ) ;
        if( Key = K^ ) then { Found item in tree }
        begin
	    Find := Current ;
            Exit
        end ;
        if( Key < K^ ) then
        begin
            Current := Current^.Left ;
        end else
        begin
	    Current := Current^.Right ; { Follow link }
        end ;
        goto Loop ;
    end ;
    Find := nil ; { Item not found }
end ; { XTree.Find }


procedure XTree.Add( Key : string ; var Data ) ;

label Loop ;

var Current, Previous : Tree_Pointer ;
    K : PSmall_String ;

begin
    if( Ptr = nil ) then { Tree is empty }
    begin
	Getmem( Ptr, Item_Size + 16 ) ; { Allocate memory for node }
	getmem( K, length( Key ) + 1 ) ;
	K^ := Key ;
	move( K, Ptr^.D[ 0 ], 4 ) ;
	Ptr^.Back := nil ;
	Ptr^.Left := nil ;
	Ptr^.Right := nil ;
	Move( Data, Ptr^.D[ 4 ], Item_Size ) ;
	Exit ;
    end ;
    Current := Ptr ; { Start at root }
    Previous := nil ;
Loop :
    if( Current <> nil ) then { Not yet at end }
    begin
	move( Current^.D[ 0 ], K, 4 ) ;
        if( Key = K^ ) then { Found item in tree }
        begin
            Application_Manager^.Fatal_Exception( 'Duplicate items encountered in tree' ) ;
        end ;
        Previous := Current ; {Save current pointer as previous}
        if( Key < K^ ) then
	begin
	    Current := Current^.Left ;
	end else
	begin
	    Current := Current^.Right ; {Follow link}
	end ;
        goto Loop ;
    end ;
    Getmem( Current, Item_Size + 16 ) ; {Create node}
    getmem( K, length( Key ) + 1 ) ;
    K^ := Key ;
    move( K, Current^.D[ 0 ], 4 ) ;
    move( Previous^.D[ 0 ], K, 4 ) ;
    if( K^ > Key ) then
    begin
	Previous^.Left := Current ;
    end else
    begin
	Previous^.Right := Current ;
    end ;
    Current^.Left := nil ;
    Current^.Right := nil ;
    Current^.Back := Previous ;
    Move( Data, Current^.D[ 4 ], Item_Size ) ;
end ;


procedure XTree.Tree_Delete( Index : Tree_Pointer ) ;

var Tmp : Tree_Pointer ;
    K : PSmall_String ;
    Work : Tree_Image_Min ;

begin
    Heap.Read_Heap( longint( Index ), sizeof( Work ), Work ) ;
    if( ( Work.Right <> nil ) and ( Work.Left = nil ) ) then
    { Right branch but no left branch }
    begin
	Index^.Right^.Back := Index^.Back ;
	if( Ptr = Index )
	    then
		Ptr := Work.Right { New root node }
	    else
		if Index^.Back^.Left = Index
		    then
			Index^.Back^.Left := Work.Right
		    else
			Index^.Back^.Right := Work.Right ;
	move( Index^.D[ 0 ], K, 4 ) ;
	Freemem( K, length( K^ ) + 1 ) ;
	Freemem( Index, Item_Size + 16 ) ; { Delete node }
	Exit ;
    end ;
    if ( Work.Left = nil ) and ( Work.Right = nil ) then {Terminal node}
    begin
	if( Ptr = Index )
	    then
		Ptr := nil {Tree is empty}
	    else
		if Index^.Back^.Left = Index
		    then
			Index^.Back^.Left := nil
		    else
			Index^.Back^.Right := nil ;
	move( Index^.D[ 0 ], K, 4 ) ;
	Freemem( K, length( K^ ) + 1 ) ;
	Freemem( Index, Item_Size + 16 ) ; {Delete node}
	Exit ;
    end ;
    {Bring up left branch}
    Index^.Left^.Back := Index^.Back ;
    if( Ptr = Index )
	then
	    Ptr := Work.Left { New root node }
	else
	    if Index^.Back^.Left = Index
		then
		    Index^.Back^.Left := Work.Left
		else
		    Index^.Back^.Right := Work.Left ;
    if( Work.Right <> nil ) then { A right branch to handle }
    begin
	Tmp := Index^.Left ; {Start at replacement node}
	while( Tmp^.Right <> nil ) do
        begin
            Tmp := Tmp^.Right ; {Go all the way right}
        end ;
	Tmp^.Right := Index^.Right ;
	Tmp^.Right^.Back := Tmp ;
    end ;
    move( Index^.D[ 0 ], K, 4 ) ;
    Heap.Deallocate( longint( K ), length( K^ ) + 1 ) ;
    Heap.Deallocate( longint( Index ), Item_Size + 16 ) ; { Actually delete node }
end ; { XTree.Tree_Delete }


procedure XTree.Tree_Prune( Index : Tree_Pointer ) ;

begin
    while( Index^.Left <> nil ) do
    begin
        Tree_Delete( Index^.Left ) ; {Delete left branch}
    end ;
    while Index^.Right <> nil do
    begin
        Tree_Delete( Index^.Right ) ; {Delete right branch}
    end ;
    Tree_Delete( Index ) ; {and delete sub-tree root node}
end ;


function XTree.Tree_First : Tree_Pointer ;

var Tmp : Tree_Pointer ;

begin
    if( Ptr = nil ) then {Tree is empty}
    begin
	Tree_First := nil ;
	Exit ;
    end ;
    Tmp := Ptr ; {Start at root node}
    while( Tmp^.Left <> nil ) do
    begin
        Tmp := Tmp^.Left ; {Go all the way left}
    end ;
    Tree_First := Tmp {Return result}
end ;


function XTree.Tree_Last : Tree_Pointer ;

var Tmp : Tree_Pointer ;

begin
    if( Ptr = nil ) then {Tree is empty}
    begin
	Tree_Last := nil ;
	Exit ;
    end ;
    Tmp := Ptr ;
    while( Tmp^.Right <> nil ) do Tmp := Tmp^.Right ;
    Tree_Last := Tmp ; {Return result}
end ;


function XTree.Empty : boolean ;

begin
    if( Ptr = nil )
	then
	    Empty := True
	else
	    Empty := False ;
end ;


function XTree.Tree_Next( Index : Tree_Pointer ) : Tree_Pointer ;

label Loop ;

var Previous : Tree_Pointer ;

begin
    if( Ptr = nil ) then
    begin
	Tree_Next := nil
    end else
    begin
        if Index^.Right = nil then
        begin
Loop:
            if( Index^.Back = nil ) then {Cannot go further up tree}
            begin
                Tree_Next := nil ; {No more items in tree}
                Exit ;
            end ;
            Previous := Index ; {Save previous position}
            Index := Index^.Back ; {Backup one level}
            if Index^.Left = Previous then {Came from left branch}
            begin
                Tree_Next := Index ;
                Exit ;
            end else goto Loop {Backup again}
        end ;
        Index := Index^.Right ; {Go down right branch}
        while Index^.Left <> nil do Index := Index^.Left ; {Go left}
        Tree_Next := Index {Return result}
    end ;
end ;


function XTree.Tree_Previous( Index : Tree_Pointer ) : Tree_Pointer ;

label Loop ;

var Previous : Tree_Pointer ;

begin
    if( Ptr = nil ) then { Nothing in tree }
    begin
	Tree_Previous := nil ;
    end else
    begin
        if Index^.Left = nil then
        begin
Loop:
            if( Index^.Back = nil ) then {Cannot go further up tree}
            begin
                Tree_Previous := nil ; {No more items in tree}
                Exit ;
            end ;
            Previous := Index ; {Save current location in tree}
            Index := Index^.Back ; {Backup one level}
            if Index^.Right = Previous then {Came from right branch}
            begin
                Tree_Previous := Index ;
                Exit ;
            end else goto Loop
        end ;
        Index := Index^.Left ;
        while Index^.Right <> nil do Index := Index^.Right ;
        Tree_Previous := Index {Return new position}
    end ;
end ;


destructor XTree.Done ;

begin
    Tree_Prune( Ptr ) ; {Delete tree by pruning from root node}
    inherited Done ;
end ;


{ TBTree methods... }

constructor TBTree.Init( Size, Key, Max, Bucket : TRecord_Size ; Flag : boolean ) ;

begin
    if( Flag and ( Size <> 4 ) ) then { Tree contains objects, but size is not 4 }
    begin
	Writeln( 'Objects in trees must be size 4' ) ;
	Halt ;
    end ;
    if( Size > Max_Memory - 3 ) then
    begin
	Writeln( 'Maximum size of B-tree object data is exceeded' ) ;
	Halt ;
    end ;
    if( Bucket * Size + 4 > Max_Memory - 3 ) then
    begin
	Writeln( 'Maximum size of B-tree bucket is exceeded' ) ;
	Halt ;
    end ;
    Item_Size := Size ;
    Key_Size := Key ;
    Storing_Objects := Flag ;
    Max_B := Max ;
    Bucket_Size := Bucket ;
    Ptr := nil ;
end ; { TBTree.Init }


function TBTree.Tree_Find( var Key ) : Tree_Pointer ;

    function _Find( Index : BTree_Pointer ) : BTree_Pointer ;

    var Active : integer ;

    begin
        _Find := nil ; { Assume failure }
        for Active := 0 to Index^.Active - 1 do
        begin
            {~~~}
        end ;
    end ;

begin
    if( Ptr = nil ) then { Nothing in tree }
    begin
        Tree_Find := nil ;
        exit ;
    end ;

    Tree_Find := Tree_Pointer( _Find( Ptr ) ) ;
end ;


procedure TBTree.Tree_Get( Index : Tree_Pointer ; var Data ) ;

var _Index : BTree_Pointer ;

begin
    _Index := BTree_Pointer( PChar( Index ) - BTree_Overhead ) ;
    if( ( _Index^.Level <> 0 ) or ( _Index.Active < 0 ) ) then
    begin
	Writeln( 'Illegal record context' ) ;
	Halt ;
    end ;
    move( _Index^.D, Data, _Index^.Active ) ;
end ;


procedure TBTree.Tree_Put( Index : Tree_Pointer ; var Data ) ;

var _Index : BTree_Pointer ;

begin
    _Index := BTree_Pointer( PChar( Index ) - BTree_Overhead ) ;
    if( ( _Index^.Level <> 0 ) or ( _Index.Active < 0 ) ) then
    begin
	Writeln( 'Illegal record context' ) ;
	Halt ;
    end ;
    move( Data, _Index^.D, _Index^.Active ) ;
end ;


procedure TBTree.Add( var Data ) ;

var Index : BTree_Pointer ;

begin
    if( Ptr = nil ) then { First record }
    begin
        { Create data record... }
        Getmem( Ptr, BTree_Overhead + Item_Size ) ; { Allocate array on heap }
        Ptr^.Level := 0 ;
        Ptr^.Active := Item_Size ;
        move( Data, Ptr^.D, Item_Size ) ;
        Getmem( Ptr^.Parent, BTree_Overhead + ( Max_B * sizeof( BTree_Pointer ) ) ) ;
        Index := Ptr ;
        Ptr := Ptr^.Parent ;

        { Create parent key record... }
        Ptr^.Level := 1 ;
        Ptr^.Active := 1 ;
        Ptr.Parent := nil ;
        fillchar( Ptr^.D, Max_B * sizeof( BTree_Pointer ), 0 ) ;
        PPointer_Array( @Ptr^.D )[ 0 ] := Index ;
        exit ;
    end ; { if( Ptr = nil ) }

    { Find place to insert record... }
    {~~~}
end ; { TBTree.Add }


procedure TBTree.Tree_Delete( Index : Tree_Pointer ) ;

var Active : integer ;
    Count : integer ;
    Found : boolean ;
    _Index, I : BTree_Pointer ;

begin
    { Adjust pointer to beginning of record header }
    _Index := BTree_Pointer( PChar( Index ) - BTree_Overhead ) ;

    { Validate valid header information }
    if( ( _Index^.Level <> 0 ) or ( _Index.Active < 0 ) ) then
    begin
	Writeln( 'Illegal record context' ) ;
	Halt ;
    end ;

    { Delete the record... }
    _Index.Active := -_Index.Active ; { Mark as deleted }
    freemem( _Index, BTree_Overhead - _Index^.Active ) ;
    I := _Index ;
    _Index := PPointer_Array( @_Index^.Parent^.D )[ 0 ] ; { Move to key record... }

    { Remove pointer to deleted record from key record... }
    Count := 0 ;
    Found := False ;
    for Active := 0 to _Index^.Active - 1 do
    begin
        if( I = PPointer_Array( @_Index^.D )[ Active ] ) then
        begin
            PPointer_Array( @_Index^.D )[ Active ] := nil ;
            Found := True ;
        end else
        begin
            if( PPointer_Array( @_Index^.D )[ Active ] <> nil ) then
            begin
                inc( Count ) ;
            end ;
        end ;
        if( Found and ( Count > 0 ) ) then { Early out }
        begin
            exit ;
        end ;
    end ;
    if( Count = 0 ) then { Deleted our last record }
    begin
        Tree_Prune( Tree_Pointer( _Index ) ) ;
    end ;
end ; { TBTree.Tree_Delete }


procedure TBTree.Tree_Prune( Index : Tree_Pointer ) ;

var Active : integer ;
    Count : integer ;
    Found : boolean ;
    I : BTree_Pointer ;
    _Index : BTree_Pointer ;

label Delete_Key ;

begin
    if( Ptr = nil ) then { Nothing in tree }
    begin
        exit ;
    end ;
    _Index := BTree_Pointer( Index ) ;
    if( _Index^.Level = 0 ) then
    begin
        Tree_Delete( Index ) ;
        exit ;
    end ;

    { Delete all sub-levels }
    for Active := 0 to _Index^.Active - 1 do
    begin
        I := PPointer_Array( @_Index^.D )[ Active ] ;
        if( I <> nil ) then
        begin
	    if( _Index^.Level = 1 ) then { Deleting a data record }
            begin
                I.Active := -I.Active ; { Mark as deleted }
                freemem( I, BTree_Overhead - I^.Active ) ;
            end else
            begin
                Tree_Prune( Tree_Pointer( I ) ) ; { Prune sub-tree }
            end ;
        end ;
    end ;

Delete_Key:
    { Now delete ourself... }
    _Index^.Active := 0 ;
    I := _Index ;
    _Index := _Index^.Parent ;
    freemem( I, BTree_Overhead + ( Max_B * sizeof( BTree_Pointer ) ) ) ;

    { Remove ourself from our parent key record... }
    if( _Index = nil ) then { Deleted entire tree }
    begin
        Ptr := nil ;
        exit ;
    end ;
    Count := 0 ;
    Found := False ;
    for Active := 0 to _Index^.Active - 1 do
    begin
        if( I = PPointer_Array( @_Index^.D )[ Active ] ) then
        begin
            PPointer_Array( @_Index^.D )[ Active ] := nil ;
            Found := True ;
        end else
        begin
            if( PPointer_Array( @_Index^.D )[ Active ] <> nil ) then
            begin
                inc( Count ) ;
            end ;
        end ;
        if( Found and ( Count > 0 ) ) then { Early out }
        begin
            exit ;
        end ;
    end ;
    if( Count = 0 ) then { Deleted our last record }
    begin
        goto Delete_Key ; { Go delete ourself }
    end ;
end ; { TBTree.Tree_Prune }


function TBTree.Tree_First : Tree_Pointer ;

var Active : integer ;
    I, Image : BTree_Pointer ;

begin
    if( Ptr = nil ) then { Nothing in tree }
    begin
        Tree_First := nil ;
        exit ;
    end ;
    Image := Ptr ;
    while( Image^.Level <> 0 ) do { Go down to data level }
    begin
        for Active := 0 to Image^.Active - 1 do { Search for a valid record }
        begin
            I := PPointer_Array( @Image^.D )[ Active ] ;
            if( I <> nil ) then
            begin
                Image := I ;
                break ;
            end ;
        end ;
    end ;
    Tree_First := Tree_Pointer( PPointer_Array( @Image^.D )[ 0 ] ) ;
end ;


function TBTree.Tree_Last : Tree_Pointer ;

var Active : integer ;
    I, Image : BTree_Pointer ;

begin
    if( Ptr = nil ) then { Nothing in tree }
    begin
        Tree_Last := nil ;
        exit ;
    end ;
    Image := Ptr ;
    while( Image^.Level <> 0 ) do { Go down to data level }
    begin
        for Active := Image^.Active - 1 downto 0 do { Search for a valid record }
        begin
            I := PPointer_Array( @Image^.D )[ Active ] ;
            if( I <> nil ) then
            begin
                Image := I ;
                break ;
            end ;
        end ;
    end ;
    Tree_Last := Tree_Pointer( PPointer_Array( @Image^.D )[ 0 ] ) ;
end ;


function TBTree.Empty : boolean ;

begin
    Empty := ( Ptr = nil ) ;
end ;


function TBTree.Tree_Next( Index : Tree_Pointer ) : Tree_Pointer ;

    function _Next( Key, Data : BTree_Pointer ) : BTree_Pointer ;

    var Active : integer ;
        Found : boolean ;

    begin
        if( Key = nil ) then
        begin
            _Next := nil ;
            exit ;
        end ;
        Found := False ;
        for Active := 0 to Key^.Active - 1 do
        begin
            if( Found and ( PPointer_Array( @Key^.D )[ Active ] <> nil ) ) then
            begin
                _Next := PPointer_Array( @Key^.D )[ Active ] ;
                exit ;
            end ;
            if( Data = PPointer_Array( @Key^.D )[ Active ] ) then
            begin
                Found := True ;
            end ;
        end ;

        { Data was last record at this level... }
        _Next := _Next( Key^.Parent, Key ) ;
    end ;

var Active : integer ;
    _Index, I : BTree_Pointer ;

begin { TBTree.Tree_Next }
    { Adjust pointer to beginning of record header }
    _Index := BTree_Pointer( PChar( Index ) - BTree_Overhead ) ;

    { Validate valid header information }
    if( ( _Index^.Level <> 0 ) or ( _Index.Active < 0 ) ) then
    begin
	Writeln( 'Illegal record context' ) ;
	Halt ;
    end ;

    _Index := _Next( _Index^.Parent, _Index ) ;
    if( _Index <> nil ) then
    begin
        while( _Index^.Level > 0 ) do
        begin
            for Active := 0 to _Index^.Active - 1 do
            begin
                I := PPointer_Array( @_Index^.D )[ Active ] ;
                if( I <> nil ) then
                begin
                    _Index := I ;
                    break ;
                end ;
            end ;
        end ;
    end ;
    Tree_Next := Tree_Pointer( PChar( _Index ) + BTree_Overhead ) ;
end ; { TBTree.Tree_Next }


function TBTree.Tree_Previous( Index : Tree_Pointer ) : Tree_Pointer ;

    function _Next( Key, Data : BTree_Pointer ) : BTree_Pointer ;

    var Active : integer ;
        Found : boolean ;

    begin
        if( Key = nil ) then
        begin
            _Next := nil ;
            exit ;
        end ;
        Found := False ;
        for Active := Key^.Active - 1 downto 0 do
        begin
            if( Found and ( PPointer_Array( @Key^.D )[ Active ] <> nil ) ) then
            begin
                _Next := PPointer_Array( @Key^.D )[ Active ] ;
                exit ;
            end ;
            if( Data = PPointer_Array( @Key^.D )[ Active ] ) then
            begin
                Found := True ;
            end ;
        end ;

        { Data was last record at this level... }
        _Next := _Next( Key^.Parent, Key ) ;
    end ;

var Active : integer ;
    _Index, I : BTree_Pointer ;

begin { TBTree.Tree_Previous }
    { Adjust pointer to beginning of record header }
    _Index := BTree_Pointer( PChar( Index ) - BTree_Overhead ) ;

    { Validate valid header information }
    if( ( _Index^.Level <> 0 ) or ( _Index.Active < 0 ) ) then
    begin
	Writeln( 'Illegal record context' ) ;
	Halt ;
    end ;

    _Index := _Next( _Index^.Parent, _Index ) ;
    if( _Index <> nil ) then
    begin
        while( _Index^.Level > 0 ) do
        begin
            for Active := _Index^.Active - 1 downto 0 do
            begin
                I := PPointer_Array( @_Index^.D )[ Active ] ;
                if( I <> nil ) then
                begin
                    _Index := I ;
                    break ;
                end ;
            end ;
        end ;
    end ;
    Tree_Previous := Tree_Pointer( PChar( _Index ) + BTree_Overhead ) ;
end ; { TBTree.Tree_Previous }


destructor TBTree.Done ;

begin
    Tree_Prune( Tree_Pointer( Ptr ) ) ; { Delete tree by pruning from root node }
    inherited Done ;
end ;



{ TMatrix methods... }

constructor TMatrix.Init( Size, Elements : longint ) ;

begin
    inherited Init ;

    Ptr := 0 ;
    Element_Size := Size ;
    Element_Count := Elements ;
end ;


destructor TMatrix.Done ;

begin
    Clear ;
    inherited Done ;
end ;


procedure TMatrix.Clear ;

begin
    if( Ptr <> 0 ) then
    begin
        { Release matrix storage }
    end ;
end ;


procedure TMatrix.Matrix_Get( Index : string ; var Data ) ;

begin
end ;


procedure TMatrix.Matrix_Put( Index : string ; var Data ) ;

begin
end ;



{ TSparse_Matrix methods... }

constructor TSparse_Matrix.Init( Size, Elements : longint ) ;

begin
    inherited Init( Size, Elements ) ;

    fillchar( Last_Index, sizeof( Last_Index ), 255 ) ;
    Did_Clear := False ;
end ;


destructor TSparse_Matrix.Done ;

begin
    Ptr := 0 ; { So our ancestor doesn't try to release it. }
    inherited Done ;
end ;


procedure TSparse_Matrix.Clear ;

var Cleared : boolean ;
    Element : TSparse_Data_Header ;
    Saved : TStore_Address ;

begin
    Cleared := False ;
    if( not Did_Clear ) then { Only clear cache once during a clear operation }
    begin
        Did_Clear := True ;
        Cleared := True ;
        fillchar( Last_Index, sizeof( Last_Index ), 255 ) ;
    end ;
    while( Ptr <> 0 ) do
    begin
        Heap^.Read_Heap( Ptr, sizeof( Element ), Element ) ;
        if( Element_Count > 1 ) then { Clear sub-array }
        begin
            dec( Element_Count ) ;
            Saved := Ptr ;
            Ptr := Element.Data ;
            Clear ;
            Ptr := Saved ;
            inc( Element_Count ) ;
        end else
        begin
            if( Element_Size > sizeof( Element.Data ) ) then
            begin
                Heap^.Deallocate( Element.Data, Element_Size ) ;
            end ;
        end ;

        { Now deallocate this element... }
        Heap^.Deallocate( Ptr, sizeof( Element ) ) ;
        Ptr := Element.Next ;
    end ; { while( Ptr <> nil ) }
    if( Cleared ) then
    begin
        Did_Clear := False ; { Allow next clear to work }
    end ;
end ; { TSparse_Matrix.Clear }


function TSparse_Matrix.Element_Exists( Index : string ) : boolean ;

var Dummy : integer ;
    Element : TSparse_Data_Header ;
    P : integer ;
    This_Element_Ptr : TStore_Address ;
    This_Index : longint ;
    Saved_Ptr : Tstore_Address ;

begin
    { Setup... }
    Element_Exists := False ; { Assume failure }
    P := pos( ',', Index + ',' ) ;
    val( copy( Index, 1, P - 1 ), This_Index, Dummy ) ;
    Index := copy( Index, P + 1, length( Index ) ) ;
    This_Element_Ptr := Ptr ;
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        exit ;
    end ;
    Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

    { Now find this element... }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        This_Element_Ptr := Element.Next ;
        Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;
    if(
        ( This_Index < Element.Low_Index )
        or
        ( This_Index > Element.High_Index )
      ) then { Element doesn't exist }
    begin
        exit ;
    end ;
    if( length( Index ) = 0 ) then { Getting data, rather than another subscript }
    begin
        if( Element_Count > 1 ) then { Incomplete subscript }
        begin
            exit ;
        end ;
        Element_Exists := True ;
        exit ;
    end else { Getting another level of subscripting }
    begin
        { Save our state and then point ourselves at the root of the next level }
        Saved_Ptr := Ptr ;
        Ptr := Element.Data ;
        dec( Element_Count ) ;

        { Now get the data from the subarray }
        Element_Exists := Element_Exists( Index ) ;

        { Restore our state }
        Ptr := Saved_Ptr ;
        inc( Element_Count ) ;
    end ;
end ; { TSparse_Matrix.Element_Exists }


procedure TSparse_Matrix.Matrix_Get( Index : string ; var Data ) ;

var Dummy : integer ;
    Element : TSparse_Data_Header ;
    P : integer ;
    This_Element_Ptr : TStore_Address ;
    This_Index : longint ;
    Saved_Ptr : TStore_Address ;

begin
    { Setup... }
    P := pos( ',', Index + ',' ) ;
    val( copy( Index, 1, P - 1 ), This_Index, Dummy ) ;
    Index := copy( Index, P + 1, length( Index ) ) ;
    This_Element_Ptr := Ptr ;
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        fillchar( Data, Element_Size, 0 ) ;
        exit ;
    end ;
    Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

    { Now find this element... }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        Temp_Ptr := This_Element_Ptr ;
        This_Element_Ptr := Element.Next ;
        Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;
    if(
        ( This_Index < Element.Low_Index )
        or
        ( This_Index > Element.High_Index )
      ) then { Element doesn't exist }
    begin
        fillchar( Data, Element_Size, 0 ) ;
        exit ;
    end ;
    if( length( Index ) = 0 ) then { Getting data, rather than another subscript }
    begin
        if( Element_Size <= sizeof( Element.Data ) ) then
        begin
            move( Element.Data, Data, Element_Size ) ;
        end else
        begin
            Heap^.Read_Heap( Element.Data, Element_Size, Data ) ;
        end ;
    end else { Getting another level of subscripting }
    begin
        { Save our state and then point ourselves at the root of the next level }
        Saved_Ptr := Ptr ;
        Ptr := Element.Data ;

        { Now get the data from the subarray }
        Matrix_Get( Index, Data ) ;

        { Restore our state }
        Ptr := Saved_Ptr ;
    end ;
end ; { TSparse_Matrix.Matrix_Get }


procedure TSparse_Matrix._Put( Level : integer ; Index : string ; var Data ) ;

var Buffer : PChar ;
    Dummy : integer ;
    Element, Element1 : TSparse_Data_Header ;
    Old_Ptr : TStore_Address ;
    P : integer ;
    This_Element_Ptr, Previous_Ptr, Next_Ptr : TStore_Address ;
    This_Index : longint ;
    Saved_Next : TStore_Address ;
    Saved_Ptr : TStore_Address ;

    procedure Set_Data ;

    begin
        if( Element_Size <= sizeof( Element.Data ) ) then { If data fits within the element }
        begin
            move( Data, Element.Data, Element_Size ) ;
        end else
        begin
            Element.Data := Heap^.Allocate( Element_Size ) ;
            Heap^.Write_Heap( Element.Data, Element_Size, Data ) ;
        end ;
    end ;

begin { TSparse_Matrix._Put }
    { Setup... }
    P := pos( ',', Index + ',' ) ;
    val( copy( Index, 1, P - 1 ), This_Index, Dummy ) ;
    Index := copy( Index, P + 1, length( Index ) ) ;
    This_Element_Ptr := Ptr ;

    { Handle empty matrix }
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
        Ptr := This_Element_Ptr ;
        Element.Low_Index := This_Index ;
        Element.High_Index := This_Index ;
        Element.Next := 0 ;
        if( length( Index ) = 0 ) then
        begin
            Set_Data ;
        end else { Write next level }
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ; { Save our state }
            Ptr := 0 ;
            _Put( Level + 1, Index, Data ) ;
            Element.Data := Ptr ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end ;
        Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
        exit ;
    end ; { if( This_Element_Ptr = 0 ) }

    { Prepare to search structure... }
    Previous_Ptr := 0 ;
    if(
        ( Level <= Max_Cached_Sparse_Matrix )
        and
        ( Last_Index[ Level ] > -1 )
        and
        ( Last_Index[ Level ] <= This_Index )
      ) then
    begin
        This_Element_Ptr := Last_Ptr[ Level ] ;
        Previous_Ptr := Last_Previous_Ptr[ Level ] ;
    end ;
    Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

    { Now find this element }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        Previous_Ptr := This_Element_Ptr ;
        This_Element_Ptr := Element.Next ;
        Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;

    { Handle cache }
    if( Level <= Max_Cached_Sparse_Matrix ) then
    begin
        if( Last_Index[ Level ] <> Element.Low_Index ) then
        begin
            fillchar( Last_Index[ Level + 1 ], ( Max_Cached_Sparse_Matrix - Level ) * sizeof( TStore_Address ), 255 ) ;
        end ;
        Last_Index[ Level ] := Element.Low_Index ;
        Last_Ptr[ Level ] := This_Element_Ptr ;
        Last_Previous_Ptr[ Level ] := Previous_Ptr ;
    end ;

    { Perform the put operation... }
    { where "Element" is the element found above,
      if Element.Low_Index > This_Index then insert before Element (condition 1)
          if Element.Low_Index = This_Index + 1 and same value as Element then extend Element range (condition 1A)
          else create new element record (condition 1B)
      else
      if Element.High_Index >= This_Index then change within Element (condition 2)
          if same as existing value, do nothing (condition 2A)
          else
          if Element.Low_Index = Element.High_Index then change existing single item (condition 2B)
          else
          if Element.Low_Index = This_Index then change first element and split rest into new record (condition 2C)
          else
          if Element.High_Index = This_Index then insert new item after Element (condition 2D)
          else change element within range of Element (condition 2E)
      else New element record (condition 3)
    }
    if( Element.Low_Index > This_Index ) then { Inserting new element prior to item - condition 1 }
    begin
        if( Element.Low_Index = This_Index + 1 ) then { Just before next item }
        begin
            if( length( Index ) = 0 ) then { A data reference }
            begin
                getmem( Buffer, Element_Size ) ;
                if( Element_Size <= sizeof( Element.Data ) ) then
                begin
                    move( Element.Data, Buffer^, Element_Size ) ;
                end else
                begin
                    Heap^.Read_Heap( Element.Data, Element_Size, Buffer^ ) ;
                end ;
                if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { Same value as next item - condition 1A }
                begin
                    { If inserting the same value, just prior to another element
                      then adjust index on existing element. }
                    dec( Element.Low_Index ) ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                    freemem( Buffer, Element_Size ) ;
                    if( Level <= Max_Cached_Sparse_Matrix ) then
                    begin
                        Last_Index[ Level ] := Element.Low_Index ;
                    end ;
                    exit ;
                end ;
                freemem( Buffer, Element_Size ) ;
            end ; { if( length( Index ) <> 0 ) }
        end ; { if( Element.Low_Index = This_Index + 1 ) }

        { New element prior to This_Element_Ptr - condition 1B }

        { First create the new header... }
        Element.Next := This_Element_Ptr ;
        This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
        Element.Low_Index := This_Index ;
        Element.High_Index := This_Index ;
        if( Level <= Max_Cached_Sparse_Matrix ) then
        begin
            Last_Index[ Level ] := Element.Low_Index ;
            Last_Ptr[ Level ] := This_Element_Ptr ;
            Last_Previous_Ptr[ Level ] := Previous_Ptr ;
        end ;
        if( length( Index ) = 0 ) then { A data reference }
        begin
            Set_Data ;
        end else
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ; { Save our state }
            Ptr := 0 ;
            _Put( Level + 1, Index, Data ) ;
            Element.Data := Ptr ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end ;
        Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

        { Now insert into list... }
        if( Previous_Ptr = 0 ) then { New first element }
        begin
            { Insert at beginning of list }
            Ptr := This_Element_Ptr ;
        end else
        begin
            { Insert after Previous_Ptr... }
            Heap^.Read_Heap( Previous_Ptr, sizeof( Element ), Element ) ;
            Element.Next := This_Element_Ptr ;
            Heap^.Write_Heap( Previous_Ptr, sizeof( Element ), Element ) ;
(*
            if( Level <= Max_Cached_Sparse_Matrix ) then
            begin
                fillchar( Last_Index[ Level ], ( Max_Cached_Sparse_Matrix - Level + 1 ) * sizeof( TStore_Address ), 255 ) ;
            end ;
*)
        end ;
        exit ;
    end ; { if( Element.Low_Index > This_Index ) }
    
    if( Element.High_Index >= This_Index ) then { Changing an existing value - condition 2 }
    begin
        if( length( Index ) > 0 ) then { More subscripts to process }
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ;
            Ptr := Element.Data ; { Save our state }
            Old_Ptr := Element.Data ;
            dec( Element_Count ) ;
            _Put( Level + 1, Index, Data ) ;
            if( Ptr <> Old_Ptr ) then { Put changed root pointer }
            begin
                Element.Data := Ptr ;
                Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
            end ;

            { Restore our saved state... }
            Ptr := Saved_Ptr ;
            inc( Element_Count ) ;
        end else
        begin
            { Changing an already-existing value... }
            getmem( Buffer, Element_Size ) ;
            if( Element_Size <= sizeof( Element.Data ) ) then
            begin
                move( Element.Data, Buffer^, Element_Size ) ;
            end else
            begin
                Heap^.Read_Heap( Element.Data, Element_Size, Buffer^ ) ;
            end ;
            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { Same value as existing item }
            begin
                freemem( Buffer, Element_Size ) ; { Nothing to do - cleanup and exit - condition 2A }
                exit ;
            end ;

            { Changing a single item - condition 2B }
            if( Element.Low_Index = Element.High_Index ) then
            begin
                if( Element_Size <= sizeof( Element.Data ) ) then
                begin
                    move( Data, Element.Data, Element_Size ) ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                end else
                begin
                    Heap^.Write_Heap( Element.Data, Element_Size, Data ) ;
                end ;

                { See if we can combine with the next item }
                if( Element.Next <> 0 ) then { There is a next index }
                begin
                    Heap^.Read_Heap( Element.Next, sizeof( Element1 ), Element1 ) ;
                    if( Element1.Low_Index = Element.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read_Heap( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Saved_Next := Element.Next ;
                            Element.Next := Element1.Next ;
                            Element.High_Index := Element1.High_Index ;
                            Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                            Heap^.Deallocate( Saved_Next, sizeof( Element1 ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element1.Data, Element_Size ) ;
                            end ;
                        end ;
                    end ;
                end ; { if( Element.Next <> 0 ) }

                { Now see if we can combine with the previous item }
                if( Previous_Ptr <> 0 ) then { There was a previous index }
                begin
                    Heap^.Read_Heap( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                    if( Element.Low_Index = Element1.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read_Heap( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Element1.Next := Element.Next ;
                            Element1.High_Index := Element.High_Index ;
                            Heap^.Write_Heap( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                            Heap^.Deallocate( This_Element_Ptr, sizeof( Element ) ) ;
                            fillchar( Last_Index[ Level ], ( Max_Cached_Sparse_Matrix - Level + 1 ) * sizeof( TStore_Address ), 255 ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element.Data, Element_Size ) ;
                            end ;
                        end ; { if( Compare_Memory( Data, Buffer^, Element_Size ) ) }
                    end ; { if( Element.Low_Index = Element1.High_Index + 1 ) }
                end ; { if( Previous_Ptr <> 0 ) }
            end else

            { Need to split... - condition 2C }
            if( Element.Low_Index = This_Index ) then
            begin
                { Put new data at current location and rest of existing at new place }
                Next_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                Element.Low_Index := This_Index + 1 ;
                Heap^.Write_Heap( Next_Ptr, sizeof( Element ), Element ) ;
                Element.Next := Next_Ptr ;
                Element.Low_Index := This_Index ;
                Element.High_Index := This_Index ;
                Set_Data ;
                Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

                { Now see if we can combine with the previous item }
                if( Previous_Ptr <> 0 ) then { There was a previous index }
                begin
                    Heap^.Read_Heap( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                    if( Element.Low_Index = Element1.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read_Heap( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Element1.Next := Element.Next ;
                            Element1.High_Index := Element.High_Index ;
                            Heap^.Write_Heap( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                            fillchar( Last_Index[ Level ], ( Max_Cached_Sparse_Matrix - Level + 1 ) * sizeof( TStore_Address ), 255 ) ;
                            Heap^.Deallocate( This_Element_Ptr, sizeof( Element ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element.Data, Element_Size ) ;
                            end ;
                        end ; { if( Compare_Memory( Data, Buffer^, Element_Size ) ) }
                    end ; { if( Element.Low_Index = Element1.High_Index + 1 ) }
                end ; { if( Previous_Ptr <> 0 ) }
            end else
            begin
                if( Element.High_Index = This_Index ) then { condition 2D  }
                begin
                    { Insert new element just after the existing one... }
                    Next_Ptr := Element.Next ;
                    Element.High_Index := This_Index - 1 ;
                    Element.Next := Heap^.Allocate( sizeof( Element ) ) ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                    This_Element_Ptr := Element.Next ;
                    Element.Next := Next_Ptr ;
                    Element.Low_Index := This_Index ;
                    Element.High_Index := This_Index ;
                    Set_Data ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

                    { See if we can combine with the next item }
                    if( Element.Next <> 0 ) then { There is a next index }
                    begin
                        Heap^.Read_Heap( Element.Next, sizeof( Element1 ), Element1 ) ;
                        if( Element1.Low_Index = Element.High_Index + 1 ) then { Immediately follows }
                        begin
                            if( Element_Size <= sizeof( Element.Data ) ) then
                            begin
                                move( Element1.Data, Buffer^, Element_Size ) ;
                            end else
                            begin
                                Heap^.Read_Heap( Element1.Data, Element_Size, Buffer^ ) ;
                            end ;
                            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                            begin
                                Saved_Next := Element.Next ;
                                Element.Next := Element1.Next ;
                                Element.High_Index := Element1.High_Index ;
                                Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                                fillchar( Last_Index[ Level ], ( Max_Cached_Sparse_Matrix - Level + 1 ) * sizeof( TStore_Address ), 255 ) ;
                                Heap^.Deallocate( Saved_Next, sizeof( Element1 ) ) ;
                                if( Element_Size > sizeof( Element.Data ) ) then
                                begin
                                    Heap^.Deallocate( Element1.Data, Element_Size ) ;
                                end ;
                            end ;
                        end ;
                    end ;
                end else
                begin
                    { New data is in the middle of an existing range... condition 2E }

                    { Copy high part of range to new range element header... }
                    Next_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                    Element.Low_Index := This_Index + 1 ;
                    if( Element_Size > sizeof( Element.Data ) ) then
                    begin
                        { Copy the data to a new location... }
                        getmem( Buffer, Element_Size ) ;
                        Heap^.Read_Heap( Element.Data, Element_Size, Buffer^ ) ;
                        Element.Data := Heap^.Allocate( Element_Size ) ;
                        Heap^.Write_Heap( Element.Data, Element_Size, Buffer^ ) ;
                        freemem( Buffer, Element_Size ) ;
                    end ;
                    Heap^.Write_Heap( Next_Ptr, sizeof( Element ), Element ) ;

                    { Now create the new element... }
                    Previous_Ptr := This_Element_Ptr ;
                    This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                    Element.Low_Index := This_Index ;
                    Element.High_Index := This_Index ;
                    Element.Next := Next_Ptr ;
                    Set_Data ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                    if( Level <= Max_Cached_Sparse_Matrix ) then
                    begin
                        Last_Index[ Level ] := Element.Low_Index ;
                        Last_Ptr[ Level ] := This_Element_Ptr ;
                        Last_Previous_Ptr[ Level ] := Previous_Ptr ;
                    end ;

                    { Now update existing low part of range... }
                    Heap^.Read_Heap( Previous_Ptr, sizeof( Element ), Element ) ;
                    Element.High_Index := This_Index - 1 ;
                    Element.Next := This_Element_Ptr ;
                    Heap^.Write_Heap( Previous_Ptr, sizeof( Element ), Element ) ;
                end ;
            end ; { if( Element.Low_Index = This_Index ) }
            freemem( Buffer, Element_Size ) ;
        end ; { if( length( Index ) <> 0 ) }
        exit ;
    end ; { if( Element.High_Index >= This_Index ) }

    { Only other possibility is that this is a new element at the end of the
      list. - condition 3 }

    if( length( Index ) = 0 ) then { Data }
    begin
        { Now see if we can combine with the previous item }
        if( This_Index = Element.High_Index + 1 ) then { Immediately follows }
        begin
            getmem( Buffer, Element_Size ) ;
            if( Element_Size <= sizeof( Element.Data ) ) then
            begin
                move( Element.Data, Buffer^, Element_Size ) ;
            end else
            begin
                Heap^.Read_Heap( Element.Data, Element_Size, Buffer^ ) ;
            end ;
            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
            begin
                Element.High_Index := This_Index ;
                Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                freemem( Buffer, Element_Size ) ;
                exit ;
            end ; { if( Compare_Memory( Data, Buffer^, Element_Size ) ) }
            freemem( Buffer, Element_Size ) ;
        end ; { if( Element.Low_Index = Element1.High_Index + 1 ) }
    end ; { if( length( Index ) = 0 ) }

    Element.Next := Heap^.Allocate( sizeof( Element ) ) ;
    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
    This_Element_Ptr := Element.Next ;
    Element.Next := 0 ; { End of list }
    Element.Low_Index := This_Index ;
    Element.High_Index := This_Index ;
    if( length( Index ) = 0 ) then { Data }
    begin
        Set_Data ;
    end else
    begin
        { Drill down to next level... }
        Saved_Ptr := Ptr ;
        Ptr := 0 ; { Save our state }
        dec( Element_Count ) ;
        _Put( Level + 1, Index, Data ) ;
        inc( Element_Count ) ;
        Element.Data := Ptr ;

        { Restore our saved state }
        Ptr := Saved_Ptr ;
    end ;
    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
end ; { TSparse_Matrix._Put }


procedure TSparse_Matrix.Matrix_Put( Index : string ; var Data ) ;

begin
    _Put( 0, Index, Data ) ;
end ;


procedure TSparse_Matrix.Matrix_Put_Range( Index : string ; High : integer ;
    var Data ) ; { UNFINISHED }

var Buffer : PChar ;
    Dummy : integer ;
    Element, Element1 : TSparse_Data_Header ;
    Old_Ptr : TStore_Address ;
    P : integer ;
    This_Element_Ptr, Previous_Ptr, Next_Ptr : TStore_Address ;
    This_Index : longint ;
    Saved_Next : TStore_Address ;
    Saved_Ptr : TStore_Address ;

    procedure Set_Data ;

    begin
        if( Element_Size <= sizeof( Element.Data ) ) then { If data fits within the element }
        begin
            move( Data, Element.Data, Element_Size ) ;
        end else
        begin
            Element.Data := Heap^.Allocate( Element_Size ) ;
            Heap^.Write_Heap( Element.Data, Element_Size, Data ) ;
        end ;
    end ;

begin { TSparse_Matrix.Matrix_Put_Range }
    { Setup... }
    P := pos( ',', Index + ',' ) ;
    val( copy( Index, 1, P - 1 ), This_Index, Dummy ) ;
    Index := copy( Index, P + 1, length( Index ) ) ;
    This_Element_Ptr := Ptr ;

    { Handle empty matrix }
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
        Ptr := This_Element_Ptr ;
        Element.Low_Index := This_Index ;
        Element.High_Index := High ;
        Element.Next := 0 ;
        if( length( Index ) = 0 ) then
        begin
            Set_Data ;
        end else { Write next level }
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ; { Save our state }
            Ptr := 0 ;
            dec( Element_Count ) ;
            Matrix_Put_Range( Index, High, Data ) ;
            inc( Element_Count ) ;
            Element.Data := Ptr ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end ;
        Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
        exit ;
    end ; { if( This_Element_Ptr = 0 ) }

    { Prepare to search structure... }
    Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
    Previous_Ptr := 0 ;

    { Now find this element... }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        Previous_Ptr := This_Element_Ptr ;
        This_Element_Ptr := Element.Next ;
        Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;
{ TODO: following to be converted to handle range of elements }
    if( Element.Low_Index > This_Index ) then { Inserting new element prior to item }
    begin
        if( Element.Low_Index = This_Index + 1 ) then { Just before next item }
        begin
            if( length( Index ) = 0 ) then { A data reference }
            begin
                getmem( Buffer, Element_Size ) ;
                if( Element_Size <= sizeof( Element.Data ) ) then
                begin
                    move( Element.Data, Buffer^, Element_Size ) ;
                end else
                begin
                    Heap^.Read_Heap( Element.Data, Element_Size, Buffer^ ) ;
                end ;
                if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { Same value as next item }
                begin
                    { If inserting the same value, just prior to another element
                      then adjust index on existing element. }
                    dec( Element.Low_Index ) ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                    freemem( Buffer, Element_Size ) ;
                    exit ;
                end ;
                freemem( Buffer, Element_Size ) ;
            end ; { if( length( Index ) <> 0 ) }
        end ; { if( Element.Low_Index = This_Index + 1 ) }

        { New element prior to This_Element_Ptr... }

        { First create the new header... }
        Element.Next := This_Element_Ptr ;
        This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
        Element.Low_Index := This_Index ;
        Element.High_Index := This_Index ;
        if( length( Index ) = 0 ) then { A data reference }
        begin
            Set_Data ;
        end else
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ; { Save our state }
            Ptr := 0 ;
            dec( Element_Count ) ;
            Matrix_Put_Range( Index, High, Data ) ;
            inc( Element_Count ) ;
            Element.Data := Ptr ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end ;
        Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

        { Now insert into list... }
        if( Previous_Ptr = 0 ) then { New first element }
        begin
            { Insert at beginning of list }
            Ptr := This_Element_Ptr ;
        end else
        begin
            { Insert after Previous_Ptr... }
            Heap^.Read_Heap( Previous_Ptr, sizeof( Element ), Element ) ;
            Element.Next := This_Element_Ptr ;
            Heap^.Write_Heap( Previous_Ptr, sizeof( Element ), Element ) ;
        end ;
        exit ;
    end ; { if( Element.Low_Index > This_Index ) }
    if( Element.High_Index >= This_Index ) then { Changing an existing value }
    begin
        if( length( Index ) <> 0 ) then { More subscripts to process}
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ;
            Ptr := Element.Data ; { Save our state }
            Old_Ptr := Element.Data ;
            dec( Element_Count ) ;
            Matrix_Put_Range( Index, High, Data ) ;
            inc( Element_Count ) ;
            if( Ptr <> Old_Ptr ) then { Put changed root pointer }
            begin
                Element.Data := Ptr ;
                Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
            end ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end else
        begin
            { Changing an already-existing value... }
            getmem( Buffer, Element_Size ) ;
            if( Element_Size <= sizeof( Element.Data ) ) then
            begin
                move( Element.Data, Buffer^, Element_Size ) ;
            end else
            begin
                Heap^.Read_Heap( Element.Data, Element_Size, Buffer^ ) ;
            end ;
            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { Same value as existing item }
            begin
                freemem( Buffer, Element_Size ) ; { Nothing to do - cleanup and exit }
                exit ;
            end ;

            { Changing a single item }
            if( Element.Low_Index = Element.High_Index ) then
            begin
                if( Element_Size <= sizeof( Element.Data ) ) then
                begin
                    move( Data, Element.Data, Element_Size ) ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                end else
                begin
                    Heap^.Write_Heap( Element.Data, Element_Size, Data ) ;
                end ;

                { See if we can combine with the next item }
                if( Element.Next <> 0 ) then { There is a next index }
                begin
                    Heap^.Read_Heap( Element.Next, sizeof( Element1 ), Element1 ) ;
                    if( Element1.Low_Index = Element.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read_Heap( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Saved_Next := Element.Next ;
                            Element.Next := Element1.Next ;
                            Element.High_Index := Element1.High_Index ;
                            Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                            Heap^.Deallocate( Saved_Next, sizeof( Element1 ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element1.Data, Element_Size ) ;
                            end ;
                        end ;
                    end ;
                end ;

                { Now see if we can combine with the previous item }
                if( Previous_Ptr <> 0 ) then { There was a previous index }
                begin
                    Heap^.Read_Heap( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                    if( Element.Low_Index = Element1.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read_Heap( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Element1.Next := Element.Next ;
                            Element1.High_Index := Element.High_Index ;
                            Heap^.Write_Heap( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                            Heap^.Deallocate( This_Element_Ptr, sizeof( Element ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element.Data, Element_Size ) ;
                            end ;
                        end ;
                    end ;
                end ;
            end else

            { Need to split... }
            if( Element.Low_Index = This_Index ) then
            begin
                { Put new data at current location and rest of existing at new place }
                Next_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                Element.Low_Index := This_Index + 1 ;
                Heap^.Write_Heap( Next_Ptr, sizeof( Element ), Element ) ;
                Element.Next := Next_Ptr ;
                Element.Low_Index := This_Index ;
                Element.High_Index := This_Index ;
                Set_Data ;
                Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

                { Now see if we can combine with the previous item }
                if( Previous_Ptr <> 0 ) then { There was a previous index }
                begin
                    Heap^.Read_Heap( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                    if( Element.Low_Index = Element1.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read_Heap( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Element1.Next := Element.Next ;
                            Element1.High_Index := Element.High_Index ;
                            Heap^.Write_Heap( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                            Heap^.Deallocate( This_Element_Ptr, sizeof( Element ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element.Data, Element_Size ) ;
                            end ;
                        end ; { if( Compare_Memory( Data, Buffer^, Element_Size ) ) }
                    end ; { if( Element.Low_Index = Element1.High_Index + 1 ) }
                end ; { if( Previous_Ptr <> 0 ) }
            end else
            begin
                if( Element.High_Index = This_Index ) then
                begin
                    { Insert new element just after the existing one... }
                    Next_Ptr := Element.Next ;
                    Element.High_Index := This_Index - 1 ;
                    Element.Next := Heap^.Allocate( sizeof( Element ) ) ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                    This_Element_Ptr := Element.Next ;
                    Element.Next := Next_Ptr ;
                    Element.Low_Index := This_Index ;
                    Element.High_Index := This_Index ;
                    Set_Data ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

                    { See if we can combine with the next item }
                    if( Element.Next <> 0 ) then { There is a next index }
                    begin
                        Heap^.Read_Heap( Element.Next, sizeof( Element1 ), Element1 ) ;
                        if( Element1.Low_Index = Element.High_Index + 1 ) then { Immediately follows }
                        begin
                            if( Element_Size <= sizeof( Element.Data ) ) then
                            begin
                                move( Element1.Data, Buffer^, Element_Size ) ;
                            end else
                            begin
                                Heap^.Read_Heap( Element1.Data, Element_Size, Buffer^ ) ;
                            end ;
                            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                            begin
                                Saved_Next := Element.Next ;
                                Element.Next := Element1.Next ;
                                Element.High_Index := Element1.High_Index ;
                                Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                                Heap^.Deallocate( Saved_Next, sizeof( Element1 ) ) ;
                                if( Element_Size > sizeof( Element.Data ) ) then
                                begin
                                    Heap^.Deallocate( Element1.Data, Element_Size ) ;
                                end ;
                            end ;
                        end ;
                    end ;
                end else
                begin
                    { New data is in the middle of an existing range... }

                    { Copy high part of range to new range element header... }
                    Next_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                    Element.Low_Index := This_Index + 1 ;
                    if( Element_Size > sizeof( Element.Data ) ) then
                    begin
                        { Copy the data to a new location... }
                        getmem( Buffer, Element_Size ) ;
                        Heap^.Read_Heap( Element.Data, Element_Size, Buffer^ ) ;
                        Element.Data := Heap^.Allocate( Element_Size ) ;
                        Heap^.Write_Heap( Element.Data, Element_Size, Buffer^ ) ;
                        freemem( Buffer, Element_Size ) ;
                    end ;
                    Heap^.Write_Heap( Next_Ptr, sizeof( Element ), Element ) ;

                    { Now create the new element... }
                    Previous_Ptr := This_Element_Ptr ;
                    This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                    Element.Low_Index := This_Index ;
                    Element.High_Index := This_Index ;
                    Element.Next := Next_Ptr ;
                    Set_Data ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

                    { Now update existing low part of range... }
                    Heap^.Read_Heap( Previous_Ptr, sizeof( Element ), Element ) ;
                    Element.High_Index := This_Index - 1 ;
                    Element.Next := This_Element_Ptr ;
                    Heap^.Write_Heap( Previous_Ptr, sizeof( Element ), Element ) ;
                end ;
            end ; { if( Element.Low_Index = This_Index ) }
            freemem( Buffer, Element_Size ) ;
        end ;
        exit ;
    end ; { if( Element.High_Index >= This_Index ) }

    { Only other possibility is that this is a new element at the end of the
      list. }
    Element.Next := Heap^.Allocate( sizeof( Element ) ) ;
    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
    This_Element_Ptr := Element.Next ;
    Element.Next := 0 ; { End of list }
    Element.Low_Index := This_Index ;
    Element.High_Index := This_Index ;
    if( length( Index ) = 0 ) then { Data }
    begin
        Set_Data ;
    end else
    begin
        { Drill down to next level... }
        Saved_Ptr := Ptr ;
        Ptr := 0 ; { Save our state }
        dec( Element_Count ) ;
        Matrix_Put_Range( Index, High, Data ) ;
        inc( Element_Count ) ;
        Element.Data := Ptr ;

        { Restore our saved state }
        Ptr := Saved_Ptr ;
    end ;
    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
end ; { TSparse_Matrix.Matrix_Put_Range }



{$IFDEF DELPHI}
{ TFast_Sparse_Matrix methods... }

destructor TFast_Sparse_Matrix.Done ;

begin
    Ptr := 0 ; { So our ancestor doesn't try to release it. }

    inherited Done ;
end ;


procedure TFast_Sparse_Matrix.Clear ;

var Element : TSparse_Data_Header ;
    Saved : TStore_Address ;

begin
    while( Ptr <> 0 ) do
    begin
        Heap^.Read_Heap( Ptr, sizeof( Element ), Element ) ;
        if( Element_Count > 1 ) then { Clear sub-array }
        begin
            dec( Element_Count ) ;
            Saved := Ptr ;
            Ptr := Element.Data ;
            Clear ;
            Ptr := Saved ;
            inc( Element_Count ) ;
        end else
        begin
            if( Element_Size > sizeof( Element.Data ) ) then
            begin
                Heap^.Deallocate( Element.Data, Element_Size ) ;
            end ;
        end ;

        { Now deallocate this element... }
        Heap^.Deallocate( Ptr, sizeof( Element ) ) ;
        Ptr := Element.Next ;
    end ; { while( Ptr <> nil ) }
end ;


function TFast_Sparse_Matrix._Element_Exists( Level : integer ;
    Index : array of integer ) : boolean ;

var Element : TSparse_Data_Header ;
    This_Element_Ptr : TStore_Address ;
    This_Index : longint ;
    Saved_Ptr : TStore_Address ;

begin
    { Setup... }
    _Element_Exists := False ; { Assume failure }
    This_Index := Index[ Level ] ;
    This_Element_Ptr := Ptr ;
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        exit ;
    end ;
    Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

    { Now find this element... }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        This_Element_Ptr := Element.Next ;
        Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;
    if(
        ( This_Index < Element.Low_Index )
        or
        ( This_Index > Element.High_Index )
      ) then { Element doesn't exist }
    begin
        exit ;
    end ;
    if( length( Index ) = 1 ) then { Getting data, rather than another subscript }
    begin
        if( This_Index <= Element.High_Index ) then
        begin
            _Element_Exists := True ;
        end ;
    end else { Getting another level of subscripting }
    begin
        { Save our state and then point ourselves at the root of the next
          level... }
        Saved_Ptr := Ptr ;
        Ptr := Element.Data ;

        { Now get the data from the subarray }
        _Element_Exists := _Element_Exists( Level + 1, Index ) ;

        { Restore our state }
        Ptr := Saved_Ptr ;
    end ;
end ; { TFast_Sparse_Matrix._Element_Exists }


function TFast_Sparse_Matrix.Element_Exists( Index : array of integer ) : boolean ;

begin
    Element_Exists := _Element_Exists( 0, Index ) ;
end ; { TFast_Sparse_Matrix.Element_Exists }


procedure TFast_Sparse_Matrix._Get_Matrix( Level : integer ;
    Index : array of integer ; var Data ) ;

var Element : TSparse_Data_Header ;
    This_Element_Ptr : TStore_Address ;
    This_Index : longint ;
    Saved_Ptr : TStore_Address ;

begin
    { Setup... }
    This_Index := Index[ Level ] ;
    This_Element_Ptr := Ptr ;
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        fillchar( Data, Element_Size, 0 ) ;
        exit ;
    end ;
    Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

    { Now find this element... }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        This_Element_Ptr := Element.Next ;
        Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;
    if(
        ( This_Index < Element.Low_Index )
        or
        ( This_Index > Element.High_Index )
      ) then { Element doesn't exist }
    begin
        fillchar( Data, Element_Size, 0 ) ;
        exit ;
    end ;
    if( length( Index ) = 1 ) then { Getting data, rather than another subscript }
    begin
        if( This_Index <= Element.High_Index ) then
        begin
            if( Element_Size <= sizeof( Element.Data ) ) then
            begin
                move( Element.Data, Data, Element_Size ) ;
            end else
            begin
                Heap^.Read_Heap( Element.Data, Element_Size, Data ) ;
            end ;
        end else
        begin
            fillchar( Data, Element_Size, 0 ) ;
            exit ;
        end ;
    end else { Getting another level of subscripting }
    begin
        { Save our state and then point ourselves at the root of the next
          level... }
        Saved_Ptr := Ptr ;
        Ptr := Element.Data ;

        { Now get the data from the subarray }
        _Get_Matrix( Level, Index, Data ) ;

        { Restore our state }
        Ptr := Saved_Ptr ;
    end ;
end ; { TFast_Sparse_Matrix._Get_Matrix }


procedure TFast_Sparse_Matrix.Get_Matrix( Index : array of integer ; var Data ) ;

begin
    _Get_Matrix( 0, Index, Data ) ;
end ;


procedure TFast_Sparse_Matrix._Put_Matrix( Level : integer ;
    Index : array of integer ; var Data ) ;

var Buffer : PChar ;
    Element, Element1 : TSparse_Data_Header ;
    Old_Ptr : TStore_Address ;
    This_Element_Ptr, Previous_Ptr, Next_Ptr : TStore_Address ;
    This_Index : longint ;
    Saved_Next : TStore_Address ;
    Saved_Ptr : TStore_Address ;

    procedure Set_Data ;

    begin
        if( Element_Size <= sizeof( Element.Data ) ) then { If data fits within the element }
        begin
            move( Data, Element.Data, Element_Size ) ;
        end else
        begin
            Element.Data := Heap^.Allocate( Element_Size ) ;
            Heap^.Write_Heap( Element.Data, Element_Size, Data ) ;
        end ;
    end ;

begin
    { Setup... }
    This_Index := Index[ Level ] ;
    This_Element_Ptr := Ptr ;

    { Handle empty matrix }
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
        Ptr := This_Element_Ptr ;
        Element.Low_Index := This_Index ;
        Element.High_Index := This_Index ;
        Element.Next := 0 ;
        if( length( Index ) = 1 ) then
        begin
            Set_Data ;
        end else { Write next level }
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ; { Save our state }
            Ptr := 0 ;
            _Put_Matrix( Level + 1, Index, Data ) ;
            Element.Data := Ptr ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end ;
        Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
        exit ;
    end ; { if( This_Element_Ptr = 0 ) }

    { Prepare to search structure... }
    Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
    Previous_Ptr := 0 ;

    { Now find this element... }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        Previous_Ptr := This_Element_Ptr ;
        This_Element_Ptr := Element.Next ;
        Heap^.Read_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;
    if( Element.Low_Index > This_Index ) then { Inserting new element prior to item }
    begin
        if( Element.Low_Index = This_Index + 1 ) then { Just before next item }
        begin
            if( length( Index ) = 1 ) then { A data reference }
            begin
                getmem( Buffer, Element_Size ) ;
                if( Element_Size <= sizeof( Element.Data ) ) then
                begin
                    move( Element.Data, Buffer^, Element_Size ) ;
                end else
                begin
                    Heap^.Read_Heap( Element.Data, Element_Size, Buffer^ ) ;
                end ;
                if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { Same value as next item }
                begin
                    { If inserting the same value, just prior to another element
                      then adjust index on existing element. }
                    dec( Element.Low_Index ) ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                    freemem( Buffer, Element_Size ) ;
                    exit ;
                end ;
                freemem( Buffer, Element_Size ) ;
            end ; { if( length( Index ) <> 1 ) }
        end ; { if( Element.Low_Index = This_Index + 1 ) }

        { New element prior to This_Element_Ptr... }

        { First create the new header... }
        Element.Next := This_Element_Ptr ;
        This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
        Element.Low_Index := This_Index ;
        Element.High_Index := This_Index ;
        if( length( Index ) = 1 ) then { A data reference }
        begin
            Set_Data ;
        end else
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ; { Save our state }
            Ptr := 0 ;
            _Put_Matrix( Level + 1, Index, Data ) ;
            Element.Data := Ptr ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end ;
        Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

        { Now insert into list... }
        if( Previous_Ptr = 0 ) then { New first element }
        begin
            { Insert at beginning of list }
            Ptr := This_Element_Ptr ;
        end else
        begin
            { Insert after Previous_Ptr... }
            Heap^.Read_Heap( Previous_Ptr, sizeof( Element ), Element ) ;
            Element.Next := This_Element_Ptr ;
            Heap^.Write_Heap( Previous_Ptr, sizeof( Element ), Element ) ;
        end ;
        exit ;
    end ; { if( Element.Low_Index > This_Index ) }
    if( Element.High_Index >= This_Index ) then { Changing an existing value }
    begin
        if( length( Index ) <> 1 ) then { More subscripts to process}
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ;
            Ptr := Element.Data ; { Save our state }
            Old_Ptr := Element.Data ;
            _Put_Matrix( Level + 1, Index, Data ) ;
            if( Ptr <> Old_Ptr ) then { Put changed root pointer }
            begin
                Element.Data := Ptr ;
                Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
            end ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end else
        begin
            { Changing an already-existing value... }
            getmem( Buffer, Element_Size ) ;
            if( Element_Size <= sizeof( Element.Data ) ) then
            begin
                move( Element.Data, Buffer^, Element_Size ) ;
            end else
            begin
                Heap^.Read_Heap( Element.Data, Element_Size, Buffer^ ) ;
            end ;
            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { Same value as existing item }
            begin
                freemem( Buffer, Element_Size ) ; { Nothing to do - cleanup and exit }
                exit ;
            end ;

            { Changing a single item }
            if( Element.Low_Index = Element.High_Index ) then
            begin
                if( Element_Size <= sizeof( Element.Data ) ) then
                begin
                    move( Data, Element.Data, Element_Size ) ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                end else
                begin
                    Heap^.Write_Heap( Element.Data, Element_Size, Data ) ;
                end ;

                { See if we can combine with the next item }
                if( Element.Next <> 0 ) then { There is a next index }
                begin
                    Heap^.Read_Heap( Element.Next, sizeof( Element1 ), Element1 ) ;
                    if( Element1.Low_Index = Element.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read_Heap( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Saved_Next := Element.Next ;
                            Element.Next := Element1.Next ;
                            Element.High_Index := Element1.High_Index ;
                            Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                            Heap^.Deallocate( Saved_Next, sizeof( Element1 ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element1.Data, Element_Size ) ;
                            end ;
                        end ;
                    end ;
                end ;

                { Now see if we can combine with the previous item }
                if( Previous_Ptr <> 0 ) then { There was a previous index }
                begin
                    Heap^.Read_Heap( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                    if( Element.Low_Index = Element1.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read_Heap( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Element1.Next := Element.Next ;
                            Element1.High_Index := Element.High_Index ;
                            Heap^.Write_Heap( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                            Heap^.Deallocate( This_Element_Ptr, sizeof( Element ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element.Data, Element_Size ) ;
                            end ;
                        end ;
                    end ;
                end ;
            end else

            { Need to split... }
            if( Element.Low_Index = This_Index ) then
            begin
                { Put new data at current location and rest of existing at new place }
                Next_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                Element.Low_Index := This_Index + 1 ;
                Heap^.Write_Heap( Next_Ptr, sizeof( Element ), Element ) ;
                Element.Next := Next_Ptr ;
                Element.Low_Index := This_Index ;
                Element.High_Index := This_Index ;
                Set_Data ;
                Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

                { Now see if we can combine with the previous item }
                if( Previous_Ptr <> 0 ) then { There was a previous index }
                begin
                    Heap^.Read_Heap( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                    if( Element.Low_Index = Element1.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read_Heap( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Element1.Next := Element.Next ;
                            Element1.High_Index := Element.High_Index ;
                            Heap^.Write_Heap( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                            Heap^.Deallocate( This_Element_Ptr, sizeof( Element ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element.Data, Element_Size ) ;
                            end ;
                        end ;
                    end ;
                end ;
            end else
            begin
                if( Element.High_Index = This_Index ) then
                begin
                    { Insert new element just after the existing one... }
                    Next_Ptr := Element.Next ;
                    Element.High_Index := This_Index - 1 ;
                    Element.Next := Heap^.Allocate( sizeof( Element ) ) ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                    This_Element_Ptr := Element.Next ;
                    Element.Next := Next_Ptr ;
                    Element.Low_Index := This_Index ;
                    Element.High_Index := This_Index ;
                    Set_Data ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

                    { See if we can combine with the next item }
                    if( Element.Next <> 0 ) then { There is a next index }
                    begin
                        Heap^.Read_Heap( Element.Next, sizeof( Element1 ), Element1 ) ;
                        if( Element1.Low_Index = Element.High_Index + 1 ) then { Immediately follows }
                        begin
                            if( Element_Size <= sizeof( Element.Data ) ) then
                            begin
                                move( Element1.Data, Buffer^, Element_Size ) ;
                            end else
                            begin
                                Heap^.Read_Heap( Element1.Data, Element_Size, Buffer^ ) ;
                            end ;
                            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                            begin
                                Saved_Next := Element.Next ;
                                Element.Next := Element1.Next ;
                                Element.High_Index := Element1.High_Index ;
                                Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
                                Heap^.Deallocate( Saved_Next, sizeof( Element1 ) ) ;
                                if( Element_Size > sizeof( Element.Data ) ) then
                                begin
                                    Heap^.Deallocate( Element1.Data, Element_Size ) ;
                                end ;
                            end ;
                        end ;
                    end ;
                end else
                begin
                    { New data is in the middle of an existing range... }

                    { Copy high part of range to new range element header... }
                    Next_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                    Element.Low_Index := This_Index + 1 ;
                    if( Element_Size > sizeof( Element.Data ) ) then
                    begin
                        { Copy the data to a new location... }
                        getmem( Buffer, Element_Size ) ;
                        Heap^.Read_Heap( Element.Data, Element_Size, Buffer^ ) ;
                        Element.Data := Heap^.Allocate( Element_Size ) ;
                        Heap^.Write_Heap( Element.Data, Element_Size, Buffer^ ) ;
                        freemem( Buffer, Element_Size ) ;
                    end ;
                    Heap^.Write_Heap( Next_Ptr, sizeof( Element ), Element ) ;

                    { Now create the new element... }
                    Previous_Ptr := This_Element_Ptr ;
                    This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                    Element.Low_Index := This_Index ;
                    Element.High_Index := This_Index ;
                    Element.Next := Next_Ptr ;
                    Set_Data ;
                    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;

                    { Now update existing low part of range... }
                    Heap^.Read_Heap( Previous_Ptr, sizeof( Element ), Element ) ;
                    Element.High_Index := This_Index - 1 ;
                    Element.Next := This_Element_Ptr ;
                    Heap^.Write_Heap( Previous_Ptr, sizeof( Element ), Element ) ;
                end ;
            end ; { if( Element.Low_Index = This_Index ) }
            freemem( Buffer, Element_Size ) ;
        end ;
        exit ;
    end ; { if( Element.High_Index >= This_Index ) }

    { Only other possibility is that this is a new element at the end of the
      list. }
    Element.Next := Heap^.Allocate( sizeof( Element ) ) ;
    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
    This_Element_Ptr := Element.Next ;
    Element.Next := 0 ; { End of list }
    Element.Low_Index := This_Index ;
    Element.High_Index := This_Index ;
    if( length( Index ) = 1 ) then { Data }
    begin
        Set_Data ;
    end else
    begin
        { Drill down to next level... }
        Saved_Ptr := Ptr ;
        Ptr := 0 ; { Save our state }
        _Put_Matrix( Level + 1, Index, Data ) ;
        Element.Data := Ptr ;

        { Restore our saved state }
        Ptr := Saved_Ptr ;
    end ;
    Heap^.Write_Heap( This_Element_Ptr, sizeof( Element ), Element ) ;
end ; { TFast_Sparse_Matrix._Put_Matrix }


procedure TFast_Sparse_Matrix.Put_Matrix( Index : array of integer ; var Data ) ;

begin
    _Put_Matrix( 0, Index, Data ) ;
end ;
{$ENDIF}


{ TGrouped_Sparse_Matrix methods... }

destructor TGrouped_Sparse_Matrix.Done ;

begin
    Ptr := 0 ; { So our ancestor doesn't try to release it. }

    inherited Done ;
end ;


procedure TGrouped_Sparse_Matrix.Clear ;

begin
end ;


function TGrouped_Sparse_Matrix.Element_Exists( Index : string ) : boolean ;

begin
    Element_Exists := False ;
end ;


procedure TGrouped_Sparse_Matrix.Matrix_Get( Index : string ; var Data ) ;

begin
end ;


procedure TGrouped_Sparse_Matrix.Matrix_Put( Index : string ; var Data ) ;

begin
end ;



{ 64-bit-memory structures... }

{ TStructure64 methods... }

constructor TStructure64.Init ;

var H : PSManaged_Store64 ;

begin
{$IFNDEF DELPHI}
    TBase_Object.Init ;
{$ENDIF}
    new( H, Init ) ;
    Heap := H ;
    Owns_Heap := True ;
end ;


destructor TStructure64.Done ;

begin
    if( ( Heap <> nil ) and Owns_Heap ) then
    begin
        dispose( Heap, Done ) ;
    end ;

{$IFNDEF DELPHI}
    inherited Done ;
{$ENDIF}
end ;


			{Abstract ARRAY object routines}

constructor TArray64.Init ;

begin
    inherited Init ; { Initialize object }
end ;

procedure TArray64.Array_Get ; {Only a template - error if called}

begin
    Terminate( 'TArray64.Array_Get' ) ;
end ;

procedure TArray64.Array_Put ; { Only a template - error if called }

begin
    Terminate( 'TArray64.Array_Put' ) ;
end ;

destructor TArray64.Done ;

begin
    inherited Done ;
end ;


			{Standard static ARRAY object routines}

constructor TsArray64.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ; {Initialize array}

begin
    if( Flag and ( Size <> 4 ) ) then {array contains objects, but size is not 4}
    begin
	Writeln( 'Objects in arrays must be size 4' ) ;
	Halt
    end ;
    if( Size > Max_Memory - 9 ) then
    begin
	Writeln( 'Maximum size of array object data exceeded' ) ;
	Halt
    end ;
    TArray64.Init ; {Initialize object}
    Element_Count := Cnt ; {Number of elements to store}
    Element_Size := Size ; {Size of elements being stored}
    Storing_Objects := Flag ;
    Getmem( Ptr, Size * Cnt ) ; {Allocate array on heap}
    Fillchar( Ptr^[ 0 ], Size * Cnt, 0 ) {Zero it}
end ;


procedure TsArray64.Array_Get( Offset : TRecord_Size ; var Data ) ;
{Get an element from the array}

begin
    if Offset >= Element_Count then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in TSArray64.Array_Get' ) ;
	Halt
    end ;
    Move( Ptr^[ Offset * Element_Size ], Data, Element_Size ) {Return value}
end ;


procedure TsArray64.Array_Put( Offset : TRecord_Size ; var Data ) ;
{Put an element in the array}

begin
    if Offset >= Element_Count then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in TSArray64.Array_Put' ) ;
	Halt
    end ;
    Move( Data, Ptr^[ Offset * Element_Size ], Element_Size ) {Set value}
end ;


destructor TsArray64.Done ; {Delete entire array}

Type Tsarray_Pointer = ^Tsarray ;

var Offset : Integer ;
    Object_Pointer : Pointer ;

begin
    if Storing_Objects then {if storing objects in array}
	For Offset := 0 To Element_Count - 1 do {Check each element}
	begin
	    Move( Ptr^[ Offset * Element_Size ], Object_Pointer, 4 ) ; {Get pointer}
	    if Object_Pointer <> nil then {if an object is present in element}
		Dispose( Tsarray_Pointer( Object_Pointer ), Done )
		{Dispose of object}
	end ;
    inherited Done ;
end ;


			{STATIC ARRAY object routines}

constructor Static_Array64.Init( Lbound, Ubound : Integer ; Size : TRecord_Size ;
    Flag : boolean ) ; {Initialize array}

begin
    if( Lbound > Ubound ) then
    begin
	Writeln( 'Static array ', Lbound, '..', Ubound, ' is invalid' ) ;
	Halt
    end ;
    TsArray64.Init( Ubound - Lbound + 1, Size, Flag ) ; {Initialize object}
    Low := Lbound ;
    High := Ubound
end ;


procedure Static_Array64.Get( Offset : Integer ; var Data ) ;
{Get an element from the array}

begin
    if ( Offset > High ) Or ( Offset < Low ) then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in Static_Array64.Get' ) ;
	Halt
    end ;
    TsArray64.Array_Get( Offset - Low, Data )
end ;


procedure Static_Array64.Put( Offset : Integer ; var Data ) ;
{Put an element in the array}

begin
    if ( Offset > High ) Or ( Offset < Low ) then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in Static_Array64.Put' ) ;
	Halt
    end ;
    TsArray64.Array_Put( Offset - Low, Data )
end ;


			{UNBOUNDED STATIC ARRAY object routines}

procedure Unbounded_Static_Array64.Get( Offset : Integer ; var Data ) ;
{Get an element from the array}

begin
    if( Offset > High ) then
    begin
	Offset := High ;
    end ;
    if( Offset < Low ) then
    begin
	Offset := Low ;
    end ;
    inherited Array_Get( Offset - Low, Data ) ;
end ;


procedure Unbounded_Static_Array64.Put( Offset : Integer ; var Data ) ;
{Put an element in the array}

begin
    if( Offset > High ) then
    begin
	Offset := High ;
    end ;
    if( Offset < Low ) then
    begin
	Offset := Low ;
    end ;
    inherited Array_Put( Offset - Low, Data )
end ;


{$IFNDEF DELPHI_OR_WINDOWS}
			{Standard static EMS ARRAY object routines}

constructor Ts_EmsArray64.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;
{Initialize array}

var Pages : Longint ;
    Dummy : Integer ;
    F : File ;
    Reg : Registers ;

begin
    if Flag and ( Size <> 4 ) then {array contains objects, but size is not 4}
    begin
	Writeln( 'Objects in arrays must be size 4' ) ;
	Halt
    end ;
    if Size > 16384 then
    begin
	Writeln( 'Maximum EMS array element size is 16384' ) ;
	Halt
    end ;

    {Make sure driver is present}
    Assign( F, 'EMMXXXX0' ) ;
    {$I-}
    Reset( F ) ;
    {$I+}
    if Ioresult <> 0 then
    begin
	Writeln( 'Expanded memory device not present' ) ;
	Halt
    end ;
    Close( F ) ;

    Reg.Ax := $4600 ;
    Intr( $67, Reg ) ; {Get driver version number}
    if Reg.Ah <> 0 then
    begin
	Writeln( 'Expanded memory device not responding' ) ;
	Halt
    end ;
    if Reg.Al < $30 then
    begin
	Writeln( 'Incompatible version of Expanded memory driver' ) ;
	Halt
    end ;

    Reg.Ax := $4100 ;
    Intr( $67, Reg ) ; {Get page frame address}
    if Reg.Ah <> 0 then
    begin
	Writeln( 'Expanded Memory device not responding' ) ;
	Halt
    end ;
    Page_Frame := Reg.Bx ;

    Current_Handle := Maxint ; {Nothing is mapped}
    Current_Page := Maxint ;

    Pages := Cnt ;
    Pages := Pages * Size ;
    Pages := ( Pages + 16383 ) Div 16384 ;
    {Total number of 16K pages needed (rounded up to nearest page)}
    if Pages = 0 then
    begin
	Writeln( 'Requested array size is zero pages' ) ;
	Halt
    end ;

    Page_Count := Pages - 1 ; {store the number of pages for reference}

    Reg.Ax := $4200 ;
    Intr( $67, Reg ) ; {Get EMS pages available}
    if Reg.Ah <> 0 then
    begin
	Writeln( 'Expanded memory device not responding' ) ;
	Halt
    end ;

    if Page_Count + 1 > Reg.Bx then
    begin
	Writeln( 'Insufficient Expanded Memory available for array' ) ;
	Halt
    end ;

    TArray64.Init ; {Initialize object}
    Reg.Ax := $4300 ;
    Reg.Bx := Page_Count + 1 ;
    Intr( $67, Reg ) ; {Allocate EMM space}
    if Reg.Ah <> 0 then
    begin
	Writeln( 'Expanded memory device not responding' ) ;
	Halt
    end ;
    Handle := Reg.Dx ;

    Element_Count := Cnt ;
    Element_Size := Size ;
    Internal_Size := Size ;

    {Now round up internal size to a power of 2}
    Dummy := 1 ; {Start at 2^0}
    while Dummy < Size do
    begin
	Dummy := Dummy + Dummy ; {Go to next power of 2}
	if Size < Dummy then Internal_Size := Dummy
    end ;

    {Now initialize the array:}
    For Dummy := 0 To Page_Count do
    begin
	Reg.Ax := $4400 ;
	Reg.Bx := Dummy ;
	Reg.Dx := Handle ;
	Intr( $67, Reg ) ; {Get page n}
	if Reg.Ah <> 0 then
	begin
            Application_Manager^.Fatal_Exception( 'Expanded memory device not responding' ) ;
	end ;
	Fillchar( Mem[ Page_Frame : 0 ], 16384, 0 ) ; { Zero the window }
    end ;
    Current_Page := Page_Count ;
    Current_Handle := Handle ;
    Storing_Objects := Flag
end ;


procedure Ts_EmsArray64.Array_Get( Offset : TRecord_Size ; var Data ) ;
{Get an element from the array}

var Pn : Integer ; {16K page number}
    Po : Integer ; {offset into 16K page}
    Offs : Longint ;
    Reg : Registers ;

begin
    if Offset >= Element_Count then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in TS_EMSArray64.Array_Get' ) ;
	Halt
    end ;
    Offs := Offset ;
    Offs := Offs * Internal_Size ; {Convert array offset to byte offset}
    Pn := Offs Div 16384 ; {Page}
    Po := Offs and 16383 ; {Offset}

    if ( Handle <> Current_Handle ) Or ( Pn <> Current_Page ) then
    begin
	Reg.Ax := $4400 ;
	Reg.Bx := Pn ;
	Reg.Dx := Handle ;
	Intr( $67, Reg ) ; {Get required page}
	if Reg.Ah <> 0 then
	begin
            Application_Manager^.Fatal_Exception( 'Expanded memory device not responding' ) ;
	end ;
        Current_Page := Pn ;
        Current_Handle := Handle
    end ;
    Move( Mem[ Page_Frame : Po ], Data, Element_Size ) {Get the data}
end ;


procedure Ts_EmsArray64.Array_Put( Offset : TRecord_Size ; var Data ) ;
{Put an element in the array}

var Pn : Integer ; {16K page number}
    Po : Integer ; {offset into 16K page}
    Offs : Longint ;
    Reg : Registers ;

begin
    if Offset >= Element_Count then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in TS_EMSArray64.Array_Put' ) ;
	Halt
    end ;
    Offs := Offset ;
    Offs := Offs * Internal_Size ; {Convert from array offset to byte offset}
    Pn := Offs Div 16384 ; {Page}
    Po := Offs and 16383 ; {Offset}

    if ( Handle <> Current_Handle ) Or ( pn <> Current_Page ) then
    begin
	Reg.Ax := $4400 ;
	Reg.Bx := Pn ;
	Reg.Dx := Handle ;
	Intr( $67, Reg ) ; {Get required page}
	if Reg.Ah <> 0 then
	begin
            Application_Manager^.Fatal_Exception( 'Expanded memory device not responding' ) ;
	end ;
	Current_Page := Pn ;
	Current_Handle := Handle ;
    end ;
    Move( Data, Mem[ Page_Frame : Po ], Element_Size ) {set the data}
end ;


destructor Ts_EmsArray64.Done ; {Delete entire array}

Type Ts_Emsarray_Pointer = ^Ts_Emsarray ;

var Offset : Integer ;
    Object_Pointer : Pointer ;
    Reg : Registers ;

begin
    Reg.Ax := $4500 ;
    Reg.Dx := Handle ;
    Intr( $67, Reg ) ; { Deallocate memory }
    if Reg.Ah <> 0 then
    begin
	Writeln( 'Expanded memory device not responding' ) ;
	Halt ;
    end ;
    inherited Done ;
end ;


			{STATIC EMS ARRAY object routines}

constructor Static_Ems_Array64.Init( Lbound, Ubound : Integer ; Size : TRecord_Size ;
    Flag : boolean ) ; {Initialize array}

begin
    if Lbound > Ubound then
    begin
	Writeln( 'Static EMS array ', Lbound, '..', Ubound, ' is invalid' ) ;
	Halt
    end ;
    Ts_EmsArray64.Init( Ubound - Lbound + 1, Size, Flag ) ; {Initialize object}
    Low := Lbound ;
    High := Ubound
end ;


procedure Static_Ems_Array64.Get( Offset : Integer ; var Data ) ;
{Get an element from the array}

begin
    if ( Offset > High ) Or ( Offset < Low ) then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in Static_EMS_Array64.Get' ) ;
	Halt ;
    end ;
    Ts_EmsArray64.Array_Get( Offset - Low, Data )
end ;


procedure Static_Ems_Array64.Put( Offset : Integer ; var Data ) ;
{Put an element in the array}

begin
    if ( Offset > High ) Or ( Offset < Low ) then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in Static_EMS_Array64.Put' ) ;
	Halt
    end ;
    Ts_EmsArray64.Array_Put( Offset - Low, Data ) ;
end ;
{$ENDIF}


			{Abstract stack object routines}

constructor TStack64.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;

begin
    inherited Init( Cnt, Size, Flag ) ;
end ;


procedure TStack64.Push( var Data ) ;

begin
    Terminate( 'TStack64.Push' )
end ;


function TStack64.Empty : boolean ;

begin
    Empty := True ;
    Terminate( 'TStack64.Empty' ) ;
end ;


procedure TStack64.Top( var Data ) ;

begin
    Terminate( 'TStack64.Top' )
end ;


procedure TStack64.Pop( var Data ) ;

begin
    Terminate( 'TStack64.Pop' )
end ;


procedure TStack64.Array_Get( Offset : TRecord_Size ; var Data ) ;

begin
    Terminate1( 'Stack', 'Array_Get' )
end ;


procedure TStack64.Array_Put( Offset : TRecord_Size ; var Data ) ;

begin
    Terminate1( 'Stack', 'Array_Put' )
end ;


procedure TStack64.Get( Offset : Integer ; var Data ) ;

begin
    Terminate1( 'Stack', 'Get' )
end ;


procedure TStack64.Put( Offset : Integer ; var Data ) ;

begin
    Terminate1( 'Stack', 'Put' )
end ;


			{STACK object routines}

constructor Stack64.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;

begin
    if( Cnt > Max_Memory ) then
    begin
	Writeln( 'Maximum stack size is exceeded' ) ;
	Halt ;
    end ;
    TsArray64.Init( Cnt, Size, Flag ) ; {Initialize an array}
    Stack_Top := 0 {Nothing on stack}
end ;


procedure Stack64.Push( var Data ) ;

begin
    if Stack_Top >= Element_Count then
    begin
	Writeln( 'Stack overflow in Stack64.Push' ) ;
	Halt
    end ;
    TsArray64.Array_Put( Stack_Top, Data ) ; {Push onto top of stack}
    Stack_Top := Stack_Top + 1
end ;


function Stack64.Empty : boolean ;

begin
    if Stack_Top = 0
	then
	    Empty := True
	else
	    Empty := False
end ;


procedure Stack64.Top( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in Stack64.Top' ) ;
	Halt
    end ;
    TsArray64.Array_Get( Stack_Top - 1, Data ) {Return data}
end ;


procedure Stack64.Pop( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in Stack64.Pop' ) ;
	Halt
    end ;
    Stack_Top := Stack_Top - 1 ;
    TsArray64.Array_Get( Stack_Top, Data ) {Return data}
end ;


			{Standard dynamic ARRAY object routines}

constructor TdArray64.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ; {Initialize array}

begin
    TsArray64.Init( Cnt, Size, Flag ) {Initialize object}
end ;


procedure TdArray64.Array_Get( Offset : TRecord_Size ; var Data ) ;
{Get an element from the array}

begin
    if Offset >= Element_Count {Subscript out of range}
	then
	    Fillchar( Data, Element_Size, 0 ) {Always return 0/nil/false}
	else
	    Move( Ptr^[ Offset * Element_Size ], Data, Element_Size ) {Return value}
end ;


procedure TdArray64.Array_Put( Offset : TRecord_Size ; var Data ) ;
{Put an element in the array}

var Temp : Array_Pointer ;

begin
    if Offset >= Element_Count then {Subscript out of range}
    begin
	Getmem( Temp, ( Offset + 1 ) * Element_Size ) ; {Allocate another chunk of memory}
	Fillchar( Temp^[ 0 ], ( Offset + 1 ) * Element_Size, 0 ) ; {Zero it}
	Move( Ptr^[ 0 ], Temp^[ 0 ], Element_Count * Element_Size ) ; {Copy old data}
	Freemem( Ptr, Element_Count * Element_Size ) ; {Free up old memory}
	Element_Count := Offset + 1 ; {New element count}
	Ptr := Temp {New pointer to data}
    end ;
    Move( Data, Ptr^[ Offset * Element_Size ], Element_Size ) {Set value}
end ;


			{DYNAMIC ARRAY object routines}

constructor Dynamic_Array64.Init( Lbound, Ubound : Integer ; Size : TRecord_Size ;
    Flag : boolean ) ; {Initialize array}

begin
    if Lbound > Ubound then
    begin
	Writeln( 'Dynamic array ', Lbound, '..', Ubound, ' is invalid' ) ;
	Halt
    end ;
    TdArray64.Init( Ubound - Lbound + 1, Size, Flag ) ; {Initialize object}
    Low := Lbound ;
    High := Ubound
end ;


procedure Dynamic_Array64.Get( Offset : Integer ; var Data ) ;
{Get an element from the array}

begin
    if Offset > High then High := Offset ;
    if Offset < Low
	then
	    Fillchar( Data, Element_Size, 0 )
	else
	    TdArray64.Array_Get( Offset - Low, Data )
end ;


procedure Dynamic_Array64.Put( Offset : Integer ; var Data ) ;
{Put an element in the array}

var Temp : Array_Pointer ;

begin
    if Offset > High then High := Offset ;
    if Offset < Low then
    begin
	Getmem( Temp, ( High - Offset + 1 ) * Element_Size ) ;
	{Allocate another chunk of memory}
	Fillchar( Temp^[ 0 ], ( High - Offset + 1 ) * Element_Size, 0 ) ; {Zero it}
	Move( Ptr^[ 0 ], Temp^[ ( Low - Offset ) * Element_Size ],
	    Element_Count * Element_Size ) ; {Copy old data}
	Freemem( Ptr, Element_Count * Element_Size ) ; {Free up old memory}
	Element_Count := High - Offset + 1 ; {New element count}
	Ptr := Temp ; {New pointer to data}
	Low := Offset {Adjust low offset}
    end ;
    TdArray64.Array_Put( Offset - Low, Data )
end ;


			{Abstract dynamic stack object routines}

constructor TdStack64.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;

begin
    inherited Init( Cnt, Size, Flag ) ;
end ;


procedure TdStack64.Push( var Data ) ;

begin
    Terminate( 'TDStack64.Push' )
end ;


function TdStack64.Empty : boolean ;

begin
    Empty := True ;
    Terminate( 'TDStack64.Empty' ) ;
end ;


procedure TdStack64.Top( var Data ) ;

begin
    Terminate( 'TDStack64.Top' )
end ;


procedure TdStack64.Pop( var Data ) ;

begin
    Terminate( 'TDStack64.Pop' )
end ;


procedure TdStack64.Array_Get( Offset : TRecord_Size ; var Data ) ;

begin
    Terminate1( 'Dynamic stack', 'Array_Get' )
end ;


procedure TdStack64.Array_Put( Offset : TRecord_Size ; var Data ) ;

begin
    Terminate1( 'Dynamic stack', 'Array_Put' )
end ;


procedure TdStack64.Get( Offset : Integer ; var Data ) ;

begin
    Terminate1( 'Dynamic stack', 'Get' )
end ;


procedure TdStack64.Put( Offset : Integer ; var Data ) ;

begin
    Terminate1( 'Dynamic stack', 'Put' )
end ;



			{DYNAMIC STACK object routines}

constructor Dynamic_Stack64.Init( Cnt, Size : TRecord_Size ; Flag : boolean ) ;

begin
    if( Cnt > Max_Memory ) then
    begin
	Writeln( 'Maximum dynamic stack size is exceeded' ) ;
	Halt
    end ;
    TdArray64.Init( Cnt, Size, Flag ) ; {Initialize an array}
    Stack_Top := 0 {Nothing on stack}
end ;


procedure Dynamic_Stack64.Push( var Data ) ;

begin
    if(  Stack_Top > Max_Memory ) then
    begin
	Writeln( 'Stack overflow in Dynamic_Stack64.Push' ) ;
	Halt
    end ;
    TdArray64.Array_Put( Stack_Top, Data ) ; {Push onto top of stack}
    Stack_Top := Stack_Top + 1
end ;


function Dynamic_Stack64.Empty : boolean ;

begin
    if Stack_Top = 0
	then
	    Empty := True
	else
	    Empty := False
end ;


procedure Dynamic_Stack64.Top( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in Dynamic_Stack64.Top' ) ;
	Halt
    end ;
    TdArray64.Array_Get( Stack_Top - 1, Data ) {Return data}
end ;


procedure Dynamic_Stack64.Pop( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in Dynamic_Stack64.Pop' ) ;
	Halt
    end ;
    Stack_Top := Stack_Top - 1 ;
    TdArray64.Array_Get( Stack_Top, Data ) {Return data}
end ;


			{Standard virtual ARRAY object routines}

constructor TvArray64.Init( Cnt, Size, Strt : TRecord_Size ; Flag : boolean ; Nam : string ) ;
{Initialize array}

var Dummy : Integer ;
    Dummy1 : TRecord_Size ;
    Ptr : Array_Pointer ; {Dummy variables}

begin
    if Flag and ( Size <> 4 ) then {array contains objects, but size is not 4}
    begin
	Writeln( 'Objects in arrays must be size 4' ) ;
	Halt
    end ;
    if( Size > Max_Memory - 9 ) then
    begin
	Writeln( 'Maximum size of array object data is exceeded' ) ;
	Halt
    end ;
    TArray64.Init ; {Initialize object}
    Element_Count := Cnt ; {Number of elements to store}
    Element_Size := Size ; {Size of elements being stored}
    Start := Strt ;
    Storing_Objects := Flag ;
    Assign( Image, Nam ) ;
    {$I-}
    Reset( Image, Element_Size ) ; {Open image file}
    {$I+}
    Dummy := Ioresult ; {Get open status}
    if Dummy = 2 then {File not found}
    begin
	{$I-}
	Rewrite( Image, Element_Size ) ; {Create it}
	{$I+}
	Dummy := Ioresult
    end ;
    if Dummy <> 0 then
    begin
	Writeln( Ert( Dummy ), ' while opening virtual array image file ', Nam ) ;
	Halt
    end ;
    if Filesize( Image ) < Start then {Need to pad the file}
    begin
	Seek( Image, Filesize( Image ) ) ; {Go to end of file}
	{$I-}
	Blockwrite( Image, Dummy, Element_Count + 1 ) ; {Fill with anything}
	{$I+}
	Dummy := Ioresult ; {Get output status}
	if Dummy <> 0 then {Error}
	begin
            Application_Manager^.Fatal_Exception( Ert( Dummy ) + ' while padding virtual array image file' ) ;
	end ;
    end ;
    Seek( Image, Start ) ; {Go to beginning position of array in file}
    Getmem( Ptr, Element_Size ) ; {Create temporary buffer to hold null image}
    Fillchar( Ptr^[ 0 ], Element_Size, 0 ) ; {Clear buffer}
    For Dummy1 := 1 To Element_Count do
    begin
        Blockwrite( Image, Ptr^[ 0 ], 1 ) ;
    end ;
    Freemem( Ptr, Element_Size ) {Now delete temporary buffer}
end ;


procedure TvArray64.Array_Get( Offset : TRecord_Size ; var Data ) ;
{Get an element from the array}

begin
    if Offset >= Element_Count then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in TVArray64.Array_Get' ) ;
	Halt
    end ;
    Seek( Image, Start + Offset ) ;
    Blockread( Image, Data, 1 ) {Return value}
end ;


procedure TvArray64.Array_Put( Offset : TRecord_Size ; var Data ) ;
{Put an element in the array}

begin
    if Offset >= Element_Count then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in TVArray64.Array_Put' ) ;
	Halt ;
    end ;
    Seek( Image, Start + Offset ) ;
    Blockwrite( Image, Data, 1 ) ; { Write value }
end ;


destructor TvArray64.Done ; {Delete entire array}

Type Tvarray_Pointer = ^Tvarray ;

var Offset : Integer ;
    Object_Pointer : Pointer ;

begin
    if Storing_Objects then {if storing objects in array}
    begin
	Seek( Image, Start ) ; {Start at proper position in file}
	For Offset := 0 To Element_Count - 1 do {Check each element}
	begin
	    Blockread( Image, Object_Pointer, 1 ) ; {Get pointer}
	    if Object_Pointer <> nil then {if an object is present in element}
		Dispose( Tvarray_Pointer( Object_Pointer ), Done )
		{Dispose of object}
	end ;
    end ;
    inherited Done ;
end ;


			{VIRTUAL ARRAY object routines}

constructor Virtual_Array64.Init( Lbound, Ubound : Integer ; Size, Strt : TRecord_Size ;
    Flag : boolean ; Nam : string ) ; {Initialize array}

begin
    if Lbound > Ubound then
    begin
	Writeln( 'virtual array ', Lbound, '..', Ubound, ' is invalid' ) ;
	Halt
    end ;
    TvArray64.Init( Ubound - Lbound + 1, Size, Strt, Flag, Nam ) ; {Initialize object}
    Low := Lbound ;
    High := Ubound
end ;


procedure Virtual_Array64.Get( Offset : Integer ; var Data ) ;
{Get an element from the array}

begin
    if ( Offset > High ) Or ( Offset < Low ) then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in Virtual_Array64.Get' ) ;
	Halt
    end ;
    TvArray64.Array_Get( Offset - Low, Data )
end ;


procedure Virtual_Array64.Put( Offset : Integer ; var Data ) ;
{Put an element in the array}

begin
    if ( Offset > High ) Or ( Offset < Low ) then {Subscript out of range}
    begin
	Writeln( 'Subscript out of range in Virtual_Array64.Put' ) ;
	Halt
    end ;
    TvArray64.Array_Put( Offset - Low, Data )
end ;


			{Abstract virtual stack object routines}

constructor TvStack64.Init( Cnt, Size, Strt : TRecord_Size ; Flag : boolean ;
    Nam : string ) ;

begin
    inherited Init( Cnt, Size, Strt, Flag, Nam ) ;
end ;


procedure TvStack64.Push( var Data ) ;

begin
    Terminate( 'TVStack64.Push' )
end ;


function TvStack64.Empty : boolean ;

begin
    Empty := True ;
    Terminate( 'TVStack64.Empty' ) ;
end ;


procedure TvStack64.Top( var Data ) ;

begin
    Terminate( 'TVStack64.Top' )
end ;


procedure TvStack64.Pop( var Data ) ;

begin
    Terminate( 'TVStack64.Pop' )
end ;


procedure TvStack64.Array_Get( Offset : TRecord_Size ; var Data ) ;

begin
    Terminate1( 'virtual stack', 'Array_Get' )
end ;


procedure TvStack64.Array_Put( Offset : TRecord_Size ; var Data ) ;

begin
    Terminate1( 'virtual stack', 'Array_Put' )
end ;


procedure TvStack64.Get( Offset : Integer ; var Data ) ;

begin
    Terminate1( 'virtual stack', 'Get' )
end ;


procedure TvStack64.Put( Offset : Integer ; var Data ) ;

begin
    Terminate1( 'virtual stack', 'Put' )
end ;



			{VIRTUAL STACK object routines}

constructor Virtual_Stack64.Init( Cnt, Size, Strt : TRecord_Size ; Flag : boolean ;
    Nam : string ) ;

begin
    if( Cnt > Max_Memory ) then
    begin
	Writeln( 'Maximum virtual stack size is exceeded' ) ;
	Halt
    end ;
    TvArray64.Init( Cnt, Size, Strt, Flag, Nam ) ; {Initialize an array}
    Stack_Top := 0 {Nothing on stack}
end ;


procedure Virtual_Stack64.Push( var Data ) ;

begin
    if( Stack_Top > Max_Memory ) then
    begin
	Writeln( 'Stack overflow in Virtual_Stack64.Push' ) ;
	Halt
    end ;
    TvArray64.Array_Put( Stack_Top, Data ) ; {Push onto top of stack}
    Stack_Top := Stack_Top + 1
end ;


function Virtual_Stack64.Empty : boolean ;

begin
    if Stack_Top = 0
	then
	    Empty := True
	else
	    Empty := False
end ;


procedure Virtual_Stack64.Top( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in Virtual_Stack64.Top' ) ;
	Halt
    end ;
    TvArray64.Array_Get( Stack_Top - 1, Data ) {Return data}
end ;


procedure Virtual_Stack64.Pop( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in Virtual_Stack64.Pop' ) ;
	Halt
    end ;
    Stack_Top := Stack_Top - 1 ;
    TvArray64.Array_Get( Stack_Top, Data ) {Return data}
end ;


			{Abstract single-linked list object routines}

constructor TList64.Init ;

begin
    inherited Init ;
end ;


function TList64.Empty : boolean ;

begin
    Empty := True ;
    Terminate( 'TList64.Empty' ) ;
end ;


function TList64.List_Last : List_Pointer ;

begin
    List_last := 0 ;
    Terminate( 'TList64.List_Last' )
end ;


function TList64.List_First : List_Pointer ;

begin
    List_First := 0 ;
    Terminate( 'TList64.List_First' )
end ;


function TList64.List_Next( Index : List_Pointer ) : List_Pointer ;

begin
    List_Next := 0 ;
    Terminate( 'TList64.List_Next' ) ;
end ;


procedure TList64.List_Add( var Data ) ;

begin
    Terminate( 'TList64.List_Add' )
end ;


procedure TList64.List_Get( Index : List_Pointer ; var Data ) ;

begin
    Terminate( 'TList64.List_Get' )
end ;


procedure TList64.List_Insert( Index : List_Pointer ; var Data ) ;

begin
    Terminate( 'TList64.List_Insert' )
end ;


procedure TList64.List_Delete( Index : List_Pointer ) ;

begin
    Terminate( 'TList64.List_delete' )
end ;


destructor TList64.Done ;

begin
    inherited Done ;
end ;


			{Standard single-linked list object routines}

constructor TsList64.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    Ptr := 0 ; {List is empty}
    Item_Size := Size ; {Get size of items}
    Storing_Objects := Flag ; {True if storing objects}
    if( Size > Max_Memory ) then
    begin
	Writeln( 'Maximum size of list object data is exceeded' ) ;
	Halt
    end ;
    if Storing_Objects and ( Size <> 4 ) then
    begin
	Writeln( 'Objects in lists must be size 4' ) ;
	Halt
    end
end ;


procedure TSList64.Clear ;

type Tslist_Pointer = ^Tslist ;

var Tmp, Object_Pointer, Work : List_Pointer ;
    Image : List_Image ;

begin
    if( Ptr = 0 ) then
    begin
        Exit ; { Done if list is empty }
    end ;
    Heap^.Read( Ptr, sizeof( Image ), Image ) ;
    Tmp := Image.Next ; { Get first item in list }
    while( Tmp <> 0 ) do { Until deleted up to last item }
    begin
        Heap^.Read( Tmp, sizeof( Image ), Image ) ;
	if Storing_Objects then
	begin
	    Object_Pointer := Image.Last ; { Get object pointer }
	    Dispose( Tslist_Pointer( Object_Pointer ), Done ) ;
	end ;
	Work := Image.Next ; { Save link to next item }
        Heap^.Deallocate( Tmp, Item_Size + sizeof( Image ) ) ; { Delete from heap }
	if( Tmp = Ptr ) { if just deleted last item }
	    then
		Tmp := 0 { Done }
	    else
		Tmp := Work ; {Point to next item}
    end ;
    PTR := 0 ;
end ;


function TsList64.Empty : boolean ;

begin
    if Ptr = 0 {Nothing in list}
	then
	    Empty := True
	else
	    Empty := False
end ;


procedure TsList64.List_Add( var Data ) ;

var P, Tmp : List_Pointer ;
    Image : List_Image ;

begin
    if Ptr = 0 then
    begin
        Ptr := Heap^.Allocate( Item_Size + sizeof( List_Image ) ) ; { Allocate first item }
        fillchar( Image, sizeof( Image ), 0 ) ;
        Image.Next := Ptr ; { Only item=last item, which points to first item }
        Heap^.Write( Ptr, sizeof( Image ), Image ) ;
    end else
    begin
        Heap^.Read( Ptr, sizeof( Image ), Image ) ;
        Tmp := Image.Next ; { Save pointer to first item on list }
        Image.Next := Heap^.Allocate( Item_Size + sizeof( Image ) ) ; {Allocate a new item}
        Heap^.Write( Ptr, sizeof( Image ), Image ) ;
        P := Image.Next ;
        Heap^.Read( P, sizeof( Image ), Image ) ;
        Image.Next := Tmp ; { Point last item to first }
        Heap^.Write( P, sizeof( Image ), Image ) ;
        Ptr := P ; { Move to end of list }
    end ;
    Heap^.Write( Image.Last + sizeof( List_Image ), Item_Size, Data ) ;
    { Write data to item }
end ;


procedure TsList64.List_Get( Index : List_Pointer ; var Data ) ;

var Image : List_Image ;

begin
    Heap^.Read( Index, sizeof( Image ), Image ) ;
    Heap^.Read( Image.Last + sizeof( Image ), Item_Size, Data ) ; { Return data }
end ;


procedure TSList64.List_Insert( Index : List_Pointer ; var Data ) ;
{ Insert after Index^ }

var P, Tmp : List_Pointer ;
    Image : List_Image ;

begin
    Tmp := Heap^.Allocate( Item_Size + sizeof( Image ) ) ; { Get the memory }
    Heap^.Write( Tmp + sizeof( Image ), Item_Size, Data ) ; { Copy data to item }
    if( Index = 0 ) then { Inserting at beginning of list }
    begin
	if( Ptr = 0 ) then { Previously empty }
	begin
	    Ptr := Tmp ;
	    Image.Next := Ptr ; { Point to oneself }
            Heap^.Write( Tmp, sizeof( Image ), Image ) ;
	end else
	begin
            Heap^.Read( Ptr, sizeof( Image ), Image ) ;
            P := Image.Next ;
            Image.Next := Tmp ;
            Heap^.Write( Ptr, sizeof( Image ), Image ) ; { Last item points to it now }
            Image.Next := P ;
            Heap^.Write( Tmp, sizeof( Image ), Image ) ; { Link to first item on list }
	end
    end else
    begin
        Heap^.Read( Index, sizeof( Image ), Image ) ;
        P := Image.Next ;
        Image.Next := Tmp ;
        Heap^.Write( Index, sizeof( Image ), Image ) ;
	Image.Next := P ; { Link item into list }
        Heap^.Write( Tmp, sizeof( Image ), Image ) ;
	if( Index = Ptr ) then
        begin
            Ptr := Tmp ; { Inserting at end of list }
        end ;
    end
end ;


procedure TsList64.List_Delete( Index : List_Pointer ) ;

Type Tslist_Pointer = ^Tslist ;

var Tmp, Work : List_Pointer ;
    Tmp_Image, Work_Image : List_Image ;

begin
    Tmp := Index ;
    Heap^.Read( Tmp, sizeof( Tmp_Image ), Tmp_Image ) ;
    Work := Tmp ;
    Heap^.Read( Work, sizeof( Work_Image ), Work_Image ) ;
    while( Work_Image.Next <> Tmp ) do
    begin
        Work := Work_Image.Next ; { Follow links to previous item }
        Heap^.Read( Work, sizeof( Work_Image ), Work_Image ) ;
    end ;
    { At this point, Tmp^ is the item to delete and Work^ is the previous item}
    Work_Image.Next := Tmp_Image.Next ; { Remove item from linked list }
    Heap^.Write( Work, sizeof( Work_Image ), Work_Image ) ;
    if Storing_Objects then
    begin
        Dispose( Tslist_Pointer( Tmp_Image.Last ), Done ) ;
    end ;
    if Work = Tmp
	then
	    Ptr := 0 { Deleted only item in list }
	else
	    if Ptr = Tmp then
            begin
                Ptr := Work ; { Deleted last item on list }
            end ;
    Heap^.Deallocate( Tmp, Item_Size + sizeof( Tmp_Image ) ) ; { Delete item from heap }
end ;


function TsList64.List_Last : List_Pointer ;

begin
    List_Last := Ptr ;
end ;


function TsList64.List_First : List_Pointer ;

var Image : List_Image ;

begin
    if( Ptr = 0 ) then
    begin
        List_First := 0 ; { No first if list is null }
    end else
    begin
        Heap^.Read( Ptr, sizeof( Image ), Image ) ;
	List_First := Image.Next ;
    end ;
end ;


function TsList64.List_Next( Index : List_Pointer ) : List_Pointer ;

var Image : List_Image ;

begin
    if( ( Index = Ptr ) or ( Ptr = 0 ) ) then { List is null or at end of list }
    begin
	List_Next := 0 ;
    end else
    begin
        Heap^.Read( Index, sizeof( Image ), Image ) ;
	List_Next := Image.Next ; { Return pointer to next item }
    end ;
end ;


destructor TsList64.Done ;

begin
    Clear ;
    inherited Done ;
end ;


			{SINGLE-LINKED LIST object routines}

constructor Single_List64.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    TsList64.Init( Size, Flag ) ; {Initialize list}
    Context := 0 ; {No current context}
end ;


procedure Single_List64.Add( var Data ) ;

begin
    TsList64.List_Add( Data ) ;
    Context := Ptr { Context is now last item in list }
end ;


procedure Single_List64.Get( var Data ) ; { Return current item }

begin
    if( Context = 0 ) { Context = BOL }
	then
	    Context := TsList64.List_First { Get first item }
	else
	    Context := TsList64.List_Next( Context ) ; { Go to next item }
    if( Context = 0 ) then
    begin
	Writeln( 'end of list in Single_List64.Get' ) ;
	Halt ;
    end ;
    TsList64.List_Get( Context, Data ) ;
end ;


procedure Single_List64.Insert( var Data ) ;

begin
    TsList64.List_Insert( Context, Data ) ; { Insert item }
    if( Context = 0 )
	then
	    Context := TsList64.List_First
	else
	    Context := TsList64.List_Next( Context ) ; { Point to new item }
end ;


procedure Single_List64.Delete ;

begin
    if( Context = 0 ) then { No current context }
    begin
	Writeln( 'No context in Single_List64.Delete' ) ;
	Halt ;
    end ;
    TsList64.List_Delete( Context ) ; { Delete current item }
    Context := 0 ; { Reset to beginning of list }
end ;


procedure Single_List64.Last( var Data ) ;

begin
    Context := Ptr ; { Move context to first item }
    TsList64.List_Get( Context, Data ) ; { Return item }
end ;


procedure Single_List64.First( var Data ) ;

begin
    Context := TsList64.List_First ; { Move context to first item }
    TsList64.List_Get( Context, Data ) ; { Return item }
end ;


function Single_List64.Eol : boolean ;

begin
    if Context = Ptr { At end of list? }
	then
	    Eol := True
	else
	    Eol := False ;
end ;


function Single_List64.Bol : boolean ;

var Image : List_Image ;

begin
    if( Context = 0 ) then
    begin
        BOL := True ;
    end else
    begin
        Heap^.Read( Ptr, sizeof( Image), Image ) ;
        BOL := ( Context = Image.Next ) ;
    end ;
end ;


			{Abstract list stack object routines}

constructor TlStack64.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    inherited Init( Size, Flag ) ;
end ;


procedure TlStack64.Push( var Data ) ;

begin
    Terminate( 'TLStack64.Push' )
end ;


procedure TlStack64.Top( var Data ) ;

begin
    Terminate( 'TLStack64.Top' )
end ;


procedure TlStack64.Pop( var Data ) ;

begin
    Terminate( 'TLStack64.Pop' )
end ;


procedure TlStack64.List_Add( var Data ) ;

begin
    Terminate1( 'List Stack', 'List_Add' )
end ;


procedure TlStack64.List_Get( Index : List_Pointer ; var Data ) ;

begin
    Terminate1( 'List Stack', 'List_Get' )
end ;


procedure TlStack64.List_Insert( Index : List_Pointer ; var Data ) ;

begin
    Terminate1( 'List Stack', 'List_Insert' )
end ;


function TlStack64.List_Next( Index : List_Pointer ) : List_Pointer ;

begin
    List_Next := 0 ;
    Terminate1( 'List Stack', 'List_Next' ) ;
end ;


procedure TlStack64.List_Delete( Index : List_Pointer ) ;

begin
    Terminate1( 'List Stack', 'List_Delete' )
end ;


function TlStack64.List_First : List_Pointer ;

begin
    List_First := 0 ;
    Terminate1( 'List Stack', 'List_First' ) ;
end ;


function TlStack64.List_Last : List_Pointer ;

begin
    List_Last := 0 ;
    Terminate1( 'List Stack', 'List_Last' ) ;
end ;


			{LIST STACK object routines}

constructor List_Stack64.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    TsList64.Init( Size, Flag )
end ;


procedure List_Stack64.Push( var Data ) ;

begin
    TsList64.List_Insert( 0, Data ) {Push onto top of stack}
end ;


procedure List_Stack64.Top( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in List_Stack64.Top' ) ;
	Halt ;
    end ;
    TsList64.List_Get( TsList64.List_First, Data ) ; { Get item from top of stack }
end ;


procedure List_Stack64.Pop( var Data ) ;

var Tmp : List_Pointer ;

begin
    if Empty then
    begin
	Writeln( 'Stack is empty in List_Stack64.Pop' ) ;
	Halt
    end ;
    Tmp := TsList64.List_First ; {Get first item pointer}
    TsList64.List_Get( Tmp, Data ) ; {Return data}
    TsList64.List_Delete( Tmp ) {Delete item at top of stack}
end ;


			{Standard double-linked list object routines}

constructor TdList64.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    Ptr := 0 ; { List is empty }
    Item_Size := Size ; { Get size of items }
    Storing_Objects := Flag ; { True if storing objects }
    if( Size > Max_Memory ) then
    begin { UNTESTED }
	Writeln( 'Maximum size of list object data is exceeded' ) ;
	Halt
    end ;
    if Storing_Objects and ( Size <> 4 ) then
    begin { UNTESTED }
	Writeln( 'Objects in lists must be size 4' ) ;
	Halt ;
    end ;
end ;


procedure TDList64.Clear ;

Type Tdlist_Pointer = ^Tdlist ;

var Tmp, Object_Pointer, Work : List_Pointer ;
    Image : List_Image ;

begin
    if( Ptr = 0 ) then
    begin
        Exit ; {Done if list is empty}
    end ;
    Heap^.Read( Ptr, sizeof( Image ), Image ) ; { Get first item in list }
    Tmp := Image.Next ; {Get first item in list}
    while( Tmp <> 0 ) do { Until deleted up to last item }
    begin
        Heap^.Read( Tmp, sizeof( Image ), Image ) ;
	if Storing_Objects then
	begin { UNTESTED }
	    Object_Pointer := Image.Last ; { Get object pointer }
	    Dispose( Tdlist_Pointer( Object_Pointer ), Done ) ;
	end ;
	Work := Image.Next ; { Save link to next item }
        Heap^.Deallocate( Tmp, Item_Size + sizeof( Image ) ) ; { Delete from heap }
	if( Tmp = Ptr ) then { if just deleted last item }
        begin
	    Tmp := 0 ; { Done }
        end else
        begin
	    Tmp := Work ; { Point to next item }
        end ;
    end ;
    PTR := 0 ;
end ; { while( Tmp <> 0 ) }


function TdList64.Empty : boolean ;

begin
    if Ptr = 0 { Nothing in list }
	then
	    Empty := True
	else
	    Empty := False ;
end ;


procedure TdList64.List_Add( var Data ) ;

var Tmp : List_Pointer ;
    Image, Temp : List_Image ;

begin
    if( Ptr = 0 ) then
    begin
        Ptr := Heap^.Allocate( Item_Size + sizeof( Image ) ) ; { Allocate first item }
        Image.Last := Ptr ; { Only item=first item, which points to last item }
        Image.Next := Ptr ; { Only item=last item, which points to first item }
        Heap^.Write( Ptr, sizeof( Image ), Image ) ;
    end else
    begin
        Heap^.Read( Ptr, sizeof( Image ), Image ) ;
        Tmp := Image.Next ; { Save pointer to first item on list }
        Image.Next := Heap^.Allocate( Item_Size + sizeof( Image ) ) ; { Allocate a new item }
        Heap^.Write( Ptr, sizeof( Image ), Image ) ;
        Temp.Next := Tmp ; { Point last item forward to first }
        Temp.Last := Ptr ; { Point last item back to previous last }
        Heap^.Write( Image.Next, sizeof( Temp ), Temp ) ;
        Heap^.Read( Tmp, sizeof( Temp ), Temp ) ;
        Temp.Last := Image.Next ; { Point first item back to last }
        Heap^.Write( Tmp, sizeof( Temp ), Temp ) ;
        Ptr := Image.Next ; { Move to end of list }
    end ;
    Heap^.Write( Ptr + sizeof( Image ), Item_Size, Data ) ; { Write data to item }
end ;


procedure TdList64.List_Get( Index : List_Pointer ; var Data ) ;

begin
    Heap^.Read( Index + sizeof( List_Image ), Item_Size, Data ) ;
end ;


procedure TDList64.List_Insert( Index : List_Pointer ; var Data ) ;
{ Insert after Index }

var Tmp : List_Pointer ;
    Image, P, Temp : List_Image ;

begin
    Tmp := Heap^.Allocate( Item_Size + sizeof( Image ) ) ; { Get the memory }
    Heap^.Write( Tmp + sizeof( Image ), Item_Size, Data ) ; { Copy data to item }
    if( Index = 0 ) then { Inserting at beginning of list }
    begin
	if( Ptr = 0 ) then { Previously empty }
	begin
	    Ptr := Tmp ;
	    Image.Last := Ptr ;
	    Image.Next := Ptr ; { Point to oneself }
            Heap^.Write( Ptr, sizeof( Image ), Image ) ;
	end else
	begin
            Heap^.Read( Ptr, sizeof( P ), P ) ;
	    Image.Next := P.Next ; { Link to first item on list }
	    Image.Last := Ptr ;
            Heap^.Write( Tmp, sizeof( Image ), Image ) ;
            Heap^.Read( P.Next, sizeof( Temp ), Temp ) ;
	    Temp.Last := Tmp ;
            Heap^.Write( P.Next, sizeof( Temp ), Temp ) ;
	    P.Next := Tmp ; { Last item points to it now }
            Heap^.Write( Ptr, sizeof( P ), P ) ;
	end ;
    end else
    begin
        Heap^.Read( Index, sizeof( P ), P ) ;
	Image.Next := P.Next ; { Link item into list }
	Image.Last := Index ;
        Heap^.Write( Tmp, sizeof( Image ), Image ) ;
        Heap^.Read( P.Next, sizeof( Temp ), Temp ) ;
	Temp.Last := Tmp ;
        Heap^.Write( P.Next, sizeof( Temp ), Temp ) ;
        Heap^.Read( Index, sizeof( P ), P ) ;
	P.Next := Tmp ;
        Heap^.Write( Index, sizeof( P ), P ) ;
	if( Index = Ptr ) then
        begin
            Ptr := Tmp ; { Inserting at end of list }
        end ;
    end ;
end ; { TdList64.List_Insert }


procedure TDList64.List_Delete( Index : List_Pointer ) ;

var Point : ^Tdlist ;
    Image, Temp : List_Image ;

begin
    Heap^.Read( Index, sizeof( Image ), Image ) ;
    Heap^.Read( Image.Next, sizeof( Temp ), Temp ) ;
    Temp.Last := Image.Last ;
    Heap^.Write( Image.Next, sizeof( Temp ), Temp ) ;
    Heap^.Read( Image.Last, sizeof( Temp ), Temp ) ;
    Temp.Next := Image.Next ; { Remove item from linked list }
    Heap^.Write( Image.Last, sizeof( Temp ), Temp ) ;
    if Storing_Objects then
    begin { UNTESTED }
        Heap^.Read( Index + sizeof( Image ), 4, Point ) ; { Get object pointer }
	Dispose( Point, Done ) ; { Delete object }
    end ;
    if( Image.Next = Index ) then
    begin
        Ptr := 0 { Deleted only item in list }
    end else
    begin
        if( Ptr = Index ) then
        begin
            Ptr := Image.Last ; { Deleted last item on list }
        end ;
    end ;
    Heap^.Deallocate( Index, Item_Size + sizeof( Image ) ) ; { Delete item from heap }
end ;


function TDList64.List_Last : List_Pointer ;

begin
    List_Last := Ptr ;
end ;


function TDList64.List_First : List_Pointer ;

var Image : List_Image ;

begin
    if( Ptr = 0 ) then
    begin
        List_First := 0 ; { No first if list is null }
    end else
    begin
        Heap^.Read( Ptr, sizeof( Image ), Image ) ;
	List_First := Image.Next ;
    end ;
end ;


function TDList64.List_Previous( Index : List_Pointer ) : List_Pointer ;

var Image : List_Image ;

begin
    if( Ptr = 0 ) then { List is null }
    begin
	List_Previous := 0
    end else
    begin
        Heap^.Read( Ptr, sizeof( Image ), Image ) ;
        if( Index = Image.Next ) then { At beginning of list }
        begin
            List_Previous := 0 ;
        end else
        begin
            Heap^.Read( Index, sizeof( Image ), Image ) ;
            List_Previous := Image.Last ; { Return pointer to last item }
        end ;
    end ;
end ;


function TDList64.List_Next( Index : List_Pointer ) : List_Pointer ;

var Image : List_Image ;

begin
    if( ( Index = Ptr ) or ( Ptr = 0 ) ) then { List is null or at end of list }
    begin
        List_Next := 0
    end else
    begin
        Heap^.Read( Index, sizeof( Image ), Image ) ;
	List_Next := Image.Next ; { Return pointer to next item }
    end ;
end ;


destructor TDList64.Done ;

begin
    Clear ;
    inherited Done ;
end ;


{ Double_List methods... }

constructor Double_List64.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    TDList64.Init( Size, Flag ) ; { Initialize list }
    Context := 0 ; { No current context }
end ;


procedure Double_List64.Add( var Data ) ;

begin
    List_Add( Data ) ;
    Context := Ptr ; { Context is now last item in list }
end ;


procedure Double_List64.Clear ;

begin
    inherited Clear ;

    Context := 0 ; { No more context }
end ;


procedure Double_List64.Put( var Data ) ; { Update current item }

begin
    if( Context = 0 ) then
    begin { UNTESTED }
	Writeln( 'End of list in Double_List64.Get' ) ;
	Halt ;
    end ;
    Heap^.Write( Context + sizeof( List_Image ), Item_Size, Data ) ;
end ;


procedure Double_List64.Get( var Data ) ; { Return current item }

begin
    if( Context = 0 ) then { Context = BOL }
    begin
	Context := List_First ; { Get first item }
    end else
    begin
	Context := List_Next( Context ) ; { Go to next item }
    end ;
    if( Context = 0 ) then
    begin
	Writeln( 'End of list in Double_List64.Get' ) ;
	Halt ;
    end ;
    List_Get( Context, Data ) ;
end ;


procedure Double_List64.Previous( var Data ) ; { Get previous item on list }

begin
    if( Context <> 0 ) then
    begin
        Context := List_Previous( Context ) ;
    end ;
    { Go to previous item }
    if( Context = 0 ) then
    begin { UNTESTED }
	Writeln( 'Beginning of list in Double_List64.Get' ) ;
	Halt ;
    end ;
    List_Get( Context, Data ) ;
end ;


procedure Double_List64.Insert( var Data ) ;

begin
    List_Insert( Context, Data ) ; { Insert item }
    if( Context = 0 ) then
    begin
	Context := List_First ;
    end else
    begin
	Context := List_Next( Context ) ; { Point to new item }
    end ;
end ;


procedure Double_List64.Delete ;

begin
    if Context = 0 then { No current context }
    begin
	Writeln( 'No context in Double_List64.Delete' ) ;
	Halt ;
    end ;
    List_Delete( Context ) ; { Delete current item }
    Context := 0 ; { Reset to beginning of list }
end ;


procedure Double_List64.Last( var Data ) ;

begin
    if( Ptr = 0 ) then { List is empty }
    begin
	Writeln( 'Empty list in Double_List64.Last' ) ;
	Halt ;
    end ;
    Context := List_Last ; { Move context to last item }
    List_Get( Context, Data ) { Return item }
end ;


procedure Double_List64.First( var Data ) ;

begin
    if( Ptr = 0 ) then { List is empty }
    begin
	Writeln( 'Empty list in Double_List64.First' ) ;
	Halt ;
    end ;
    Context := TdList64.List_First ; { Move context to first item }
    List_Get( Context, Data ) ; { Return item }
end ;


function Double_List64.Eol : boolean ;

begin
    if Context = Ptr { At end of list? }
	then
	    Eol := True
	else
	    Eol := False ;
end ;


function Double_List64.Bol : boolean ;

var Image : List_Image ;

begin
    if ( Context = 0 ) then
    begin
        BOL := True
    end else
    begin
        Heap^.Read( Context, sizeof( Image ), Image ) ;
        BOL := ( Context = Image.Next ) ; { At beginning of list? }
    end ;
end ;


			{Standard queue object routines}

constructor TQueue64.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    inherited Init( Size, Flag ) ;
end ;


procedure TQueue64.Add( var Data ) ;

begin
    Terminate( 'TQueue64.Add' )
end ;


procedure TQueue64.Front( var Data ) ;

begin
    Terminate( 'TQueue64.Front' )
end ;


procedure TQueue64.Delete( var Data ) ;

begin
    Terminate( 'TQueue64.Delete' )
end ;


procedure TQueue64.List_Add( var Data ) ;

begin
    Terminate1( 'Queue', 'List_Add' )
end ;


procedure TQueue64.List_Get( Index : List_Pointer ; var Data ) ;

begin
    Terminate1( 'Queue', 'List_Get' )
end ;


procedure TQueue64.List_Insert( Index : List_Pointer ; var Data ) ;

begin
    Terminate1( 'Queue', 'List_Insert' )
end ;


function TQueue64.List_Next( Index : List_Pointer ) : List_Pointer ;

begin
    List_Next := 0 ;
    Terminate1( 'Queue', 'List_Next' ) ;
end ;


procedure TQueue64.List_Delete( Index : List_Pointer ) ;

begin
    Terminate1( 'Queue', 'List_Delete' )
end ;


function TQueue64.List_First : List_Pointer ;

begin
    List_First := 0 ;
    Terminate1( 'Queue', 'List_First' ) ;
end ;


function TQueue64.List_Last : List_Pointer ;

begin
    List_Last := 0 ;
    Terminate1( 'Queue', 'List_Last' ) ;
end ;


			{QUEUE object routines}

constructor Queue64.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    Tdlist64.Init( Size, Flag )
end ;


procedure Queue64.Add( var Data ) ;

begin
    Tdlist64.List_Add( Data )
end ;


procedure Queue64.Front( var Data ) ;

begin
    if Empty then
    begin
	Writeln( 'Queue is empty in Queue64.Front' ) ;
	Halt
    end ;
    Tdlist64.List_Get( Tdlist64.List_First, Data ) {Get front of queue}
end ;


procedure Queue64.Delete( var Data ) ;

var Tmp : List_Pointer ;

begin
    if Empty then
    begin
	Writeln( 'Queue is empty in Queue64.Delete' ) ;
	Halt
    end ;
    Tmp := Tdlist64.List_First ;
    Tdlist64.List_Get( Tmp, Data ) ; {Return item from queue}
    Tdlist64.List_Delete( Tmp ) {...and delete from queue}
end ;


			{Standard sparse array object routines}

constructor Sparse_Array64.Init( Size : TRecord_Size ; Flag : boolean ; var Def ) ;

begin
    if Flag and ( Size <> 4 ) then {array contains objects, but size is not 4}
    begin
	Writeln( 'Objects in arrays must be size 4' ) ;
	Halt ;
    end ;
    if( Size > Max_Memory ) then
    begin
	Writeln( 'Maximum size of array object data is exceeded' ) ;
	Halt ;
    end ;
    Ptr := 0 ; {array starts out empty}
    Item_Size := Size ;
    Storing_Objects := Flag ;
    Default := Heap^.Allocate( Size ) ;
    Heap^.Write( Default, Size, Def ) ; { Write default value for later }
end ;


procedure Sparse_Array64.List_Add( var Data ) ;

begin
    Terminate1( 'Sparse array', 'List_Add' )
end ;


procedure Sparse_Array64.List_Get( Index : List_Pointer ; var Data ) ;

begin
    Terminate1( 'Sparse array', 'List_Get' )
end ;


procedure Sparse_Array64.List_Insert( Index : List_Pointer ; var Data ) ;

begin
    Terminate1( 'Sparse array', 'List_Insert' )
end ;


function Sparse_Array64.List_Next( Index : List_Pointer ) : List_Pointer ;

begin
    List_Next := 0 ;
    Terminate1( 'Sparse array', 'List_Next' ) ;
end ;


procedure Sparse_Array64.List_Delete( Index : List_Pointer ) ;

begin
    Terminate1( 'Sparse array', 'List_Delete' )
end ;


function Sparse_Array64.List_First : List_Pointer ;

begin
    List_First := 0 ;
    Terminate1( 'Sparse array', 'List_First' ) ;
end ;


function Sparse_Array64.List_Last : List_Pointer ;

begin
    List_Last := 0 ;
    Terminate1( 'Sparse array', 'List_Last' ) ;
end ;


procedure Sparse_Array64.Get( Offset : Longint ; var Data ) ;

var Current : List_Pointer ;
    Image : List_Image ;

begin
    if Ptr <> 0 then { Something in array }
    begin
        Heap^.Read( Ptr, sizeof( List_Pointer ), Image ) ;
	Current := Image.Next ; { Get first element in array }
        Heap^.Read( Current, sizeof( Image ), Image ) ;
	while( ( Image.Last < Offset ) and ( Current <> Ptr ) ) do
        begin
	    Current := Image.Next ; { Find item }
            Heap^.Read( Current, sizeof( Image ), Image ) ;
        end ;
	if Longint( Image.Last ) = Offset then
	begin
            Heap^.Read( Current + sizeof( Image ), Item_Size, Data ) ; { Return item }
	    Exit ;
	end ;
    end ;
    Heap^.Read( Default, Item_Size, Data ) ; { Return default }
end ;


procedure Sparse_Array64.Put( Offset : Longint ; var Data ) ;

var Current, Previous : List_Pointer ;
    Image : List_Image ;
    Tmp : List_Pointer ;

begin
    Current := 0 ;
    Previous := Ptr ;
    if( Ptr <> 0 ) then { Something in array }
    begin
        Heap^.Read( Ptr, sizeof( Image ), Image ) ;
	Current := Image.Next ; { Get first element in array }
	Previous := 0 ;
        Heap^.Read( Current, sizeof( Image ), Image ) ;
	while( ( Image.Last < Offset ) and ( Current <> Ptr ) ) do
	begin
	    Previous := Current ; { Save current location }
	    Current := Image.Next ; { Go to next item }
            Heap^.Read( Current, sizeof( Image ), Image ) ;
	end ;
	if Image.Last = Offset then { Element already exists }
	begin
            Heap^.Write( Current + sizeof( Image ), Item_Size, Data ) ; { Set item }
	    Exit ;
	end ;
	if( Current = Ptr ) then { Jump to next one if end of list }
	begin
            Heap^.Read( Ptr, sizeof( Image ), Image ) ;
	    Current := Image.Next ;
	    Previous := Ptr ;
	end ;
    end ; { if ( Ptr <> 0 ) }

    { At this point, the specified offset does not exist }
    Tmp := Heap^.Allocate( Item_Size + sizeof( Image ) ) ; { Allocate memory for element }
    Image.Last := List_Pointer( Offset ) ; { Set element number }
    Heap^.Write( Tmp + sizeof( Image ), Item_Size, Data ) ; { Set data }
    if Ptr = 0 then { Nothing in array }
    begin
	Image.Next := Tmp ; { Set pointer forward to self }
	Image.Last := List_Pointer( Offset ) ; { Set element number field }
	Ptr := Tmp ; { Last item in list is only item }
        Heap^.Write( Tmp, sizeof( Image ), Image ) ;
	Exit ;
    end ;
    Image.Next := Current ;
    Heap^.Write( Tmp, sizeof( Image ), Image ) ;
    if( Previous = 0 ) then
    begin
        Heap^.Read( Ptr, sizeof( Image ), Image ) ;
        Image.Next := Tmp ; { Insert item into beginning of list }
        Heap^.Write( Ptr, sizeof( Image ), Image ) ;
    end else
    begin
        Heap^.Read( Previous, sizeof( Image ), Image ) ;
        Image.Next := Tmp ; { Insert item into list }
        Heap^.Write( Previous, sizeof( Image ), Image ) ;
        if Previous = Ptr then Ptr := Tmp { Item is last on list }
    end ;
end ; { Sparse_Array64.Put }


				{Abstract tree object routines}

constructor TTree64.Init ;

begin
    inherited Init ;
end ;


function TTree64.Tree_Find( var Key ) : Tree_Pointer ;

begin
    Tree_Find := nil ;
    Terminate( 'TTree64.Tree_Find' ) ;
end ;


procedure TTree64.Tree_Get( Index : Tree_Pointer ; var Data ) ;

begin
    Terminate( 'TTree64.Tree_Get' ) ;
end ;


procedure TTree64.Tree_Put( Index : Tree_Pointer ; var Data ) ;

begin
    Terminate( 'TTree64.Tree_Put' ) ;
end ;


procedure TTree64.Tree_Add( var Data ) ;

begin
    Terminate( 'TTree64.Add' )
end ;


procedure TTree64.Tree_Delete( Index : Tree_Pointer ) ;

begin
    Terminate( 'TTree64.Tree_Delete' )
end ;


procedure TTree64.Tree_Prune( Index : Tree_Pointer ) ;

begin
    Terminate( 'TTree64.Tree_Prune' )
end ;


function TTree64.Tree_First : Tree_Pointer ;

begin
    Tree_First := nil ;
    Terminate( 'TTree64.Tree_First' ) ;
end ;


function TTree64.Tree_Last : Tree_Pointer ;

begin
    Tree_Last := nil ;
    Terminate( 'TTree64.Tree_Last' ) ;
end ;


function TTree64.Empty : boolean ;

begin
    Empty := True ;
    Terminate( 'TTree64.Empty' ) ;
end ;


function TTree64.Tree_Next( Index : Tree_Pointer ) : Tree_Pointer ;

begin
    Tree_Next := nil ;
    Terminate( 'TTree64.Tree_Next' ) ;
end ;


function TTree64.Tree_Previous( Index : Tree_Pointer ) : Tree_Pointer ;

begin
    Tree_Previous := nil ;
    Terminate( 'TTree64.Tree_Previous' ) ;
end ;


destructor TTree64.Done ;

begin
    inherited Done ;
end ;


			{Standard binary tree object routines}

constructor STree64.Init( Size, Key : TRecord_Size ; Flag : boolean ) ;

begin
    if( Flag and ( Size <> 4 ) ) then {Tree contains objects, but size is not 4}
    begin
	Writeln( 'Objects in trees must be size 4' ) ;
	Halt
    end ;
    if Key > Size then
    begin
	Writeln( 'Key size cannot exceed total tree object data size' ) ;
	Halt
    end ;
    if( Size > Max_Memory - 21 ) then
    begin
	Writeln( 'Maximum size of tree object data is exceeded' ) ;
	Halt
    end ;
    Ptr := nil ; {Tree is empty}
    Item_Size := Size ;
    Storing_Objects := Flag ;
    Key_Size := Key
end ;


procedure STree64.Tree_Get( Index : Tree_Pointer ; var Data ) ;

begin
    Move( Index^.D[ 0 ], Data, Item_Size ) {Return data from node}
end ;


procedure STree64.Tree_Put( Index : Tree_Pointer ; var Data ) ;

begin
    Move( Data, Index^.D[ 0 ], Item_Size ) {Return data from node}
end ;


function STree64.Tree_Find( var Key ) : Tree_Pointer ;

label Loop ;

var Current : Tree_Pointer ;
    Dummy : Integer ;

begin
    Current := Ptr ; {Start at root}
Loop :
    if Current <> nil then {Not yet at end}
    begin
	Dummy := Compare( Current^.D[ 0 ], Key, Key_Size ) ;
        if Dummy = 0 then {Found item in tree}
        begin
	    Tree_Find := Current ;
            Exit
        end ;
        if Dummy = 1
	    then
		Current := Current^.Left
	    else
		Current := Current^.Right ; {Follow link}
        goto Loop
    end ;
    Tree_Find := nil {Item not found}
end ;


procedure STree64.Add( var Data ) ;

label Loop ;

var Current, Previous : Tree_Pointer ;
    Dummy : Integer ;

begin
    if Ptr = nil then {Tree is empty}
    begin
	Getmem( Ptr, Item_Size + 12 ) ; {Allocate memory for node}
	Ptr^.Back := nil ;
	Ptr^.Left := nil ;
	Ptr^.Right := nil ;
	Move( Data, Ptr^.D[ 0 ], Item_Size ) ;
	Exit
    end ;
    Previous := nil ;
    Current := Ptr ; {Start at root}
Loop :
    if Current <> nil then {Not yet at end}
    begin
	Dummy := Compare( Current^.D[ 0 ], Data, Key_Size ) ;
        if Dummy = 0 then {Found item in tree}
        begin
            Application_Manager^.Fatal_Exception( 'Duplicate items encountered in tree' ) ;
        end ;
        Previous := Current ; {Save current pointer as previous}
        if Dummy = 1
	    then
		Current := Current^.Left
	    else
		Current := Current^.Right ; {Follow link}
        goto Loop ;
    end ;
    Getmem( Current, Item_Size + 12 ) ; {Create node}
    if Compare( Previous^.D[ 0 ], Data, Key_Size ) = 1
	then
	    Previous^.Left := Current
	else
	    Previous^.Right := Current ;
    Current^.Left := nil ;
    Current^.Right := nil ;
    Current^.Back := Previous ;
    Move( Data, Current^.D[ 0 ], Item_Size )
end ;

procedure STree64.Tree_Delete( Index : Tree_Pointer ) ;

var Tmp : Tree_Pointer ;

begin
    if ( Index^.Right <> nil ) and ( Index^.Left = nil ) then
    {Right branch but no left branch}
    begin
	Index^.Right^.Back := Index^.Back ;
	if Ptr = Index
	    then
		Ptr := Index^.Right {New root node}
	    else
		if Index^.Back^.Left = Index
		    then
			Index^.Back^.Left := Index^.Right
		    else
			Index^.Back^.Right := Index^.Right ;
	Freemem( Index, Item_Size + 12 ) ; {Delete node}
	Exit
    end ;
    if ( Index^.Left = nil ) and ( Index^.Right = nil ) then {Terminal node}
    begin
	if Ptr = Index
	    then
		Ptr := nil {Tree is empty}
	    else
		if Index^.Back^.Left = Index
		    then
			Index^.Back^.Left := nil
		    else
			Index^.Back^.Right := nil ;
	Freemem( Index, Item_Size + 12 ) ; {Delete node}
	Exit
    end ;
    {Bring up left branch}
    Index^.Left^.Back := Index^.Back ;
    if Ptr = Index
	then
	    Ptr := Index^.Left {New root node}
	else
	    if Index^.Back^.Left = Index
		then
		    Index^.Back^.Left := Index^.Left
		else
		    Index^.Back^.Right := Index^.Left ;
    if Index^.Right <> nil then {A right branch to handle}
    begin
	Tmp := Index^.Left ; {Start at replacement node}
	while Tmp^.Right <> nil do Tmp := Tmp^.Right ; {Go all the way right}
	Tmp^.Right := Index^.Right ;
	Tmp^.Right^.Back := Tmp
    end ;
    Freemem( Index, Item_Size + 12 ) {Actually delete node}
end ;


procedure STree64.Tree_Prune( Index : Tree_Pointer ) ;

begin
    while Index^.Left <> nil do Tree_Delete( Index^.Left ) ; {Delete left branch}
    while Index^.Right <> nil do Tree_Delete( Index^.Right ) ; {Delete right branch}
    Tree_Delete( Index ) {and delete sub-tree root node}
end ;


function STree64.Tree_First : Tree_Pointer ;

var Tmp : Tree_Pointer ;

begin
    if Ptr = nil then {Tree is empty}
    begin
	Tree_First := nil ;
	Exit
    end ;
    Tmp := Ptr ; {Start at root node}
    while Tmp^.Left <> nil do Tmp := Tmp^.Left ; {Go all the way left}
    Tree_First := Tmp {Return result}
end ;


function STree64.Tree_Last : Tree_Pointer ;

var Tmp : Tree_Pointer ;

begin
    if Ptr = nil then {Tree is empty}
    begin
	Tree_Last := nil ;
	Exit
    end ;
    Tmp := Ptr ;
    while Tmp^.Right <> nil do Tmp := Tmp^.Right ;
    Tree_Last := Tmp {Return result}
end ;


function STree64.Empty : boolean ;

begin
    if Ptr = nil
	then
	    Empty := True
	else
	    Empty := False
end ;


function STree64.Tree_Next( Index : Tree_Pointer ) : Tree_Pointer ;

label Loop ;

var Previous : Tree_Pointer ;

begin
    if( Ptr = nil ) then
    begin
	Tree_Next := nil ;
    end else
    begin
        if Index^.Right = nil then
        begin
Loop :
            if( Index^.Back = nil ) then {Cannot go further up tree}
            begin
                Tree_Next := nil ; {No more items in tree}
                Exit ;
            end ;
            Previous := Index ; {Save previous position}
            Index := Index^.Back ; {Backup one level}
            if Index^.Left = Previous then {Came from left branch}
            begin
                Tree_Next := Index ;
                Exit ;
            end else goto Loop {Backup again}
        end ;
        Index := Index^.Right ; {Go down right branch}
        while Index^.Left <> nil do Index := Index^.Left ; {Go left}
        Tree_Next := Index {Return result}
    end ;
end ;


function STree64.Tree_Previous( Index : Tree_Pointer ) : Tree_Pointer ;

label Loop ;

var Previous : Tree_Pointer ;

begin
    if( Ptr = nil ) then { Nothing in tree }
    begin
	Tree_Previous := nil ;
    end else
    begin
        if Index^.Left = nil then
        begin
Loop :
            if( Index^.Back = nil ) then {Cannot go further up tree}
            begin
                Tree_Previous := nil ; {No more items in tree}
                Exit ;
            end ;
            Previous := Index ; {Save current location in tree}
            Index := Index^.Back ; {Backup one level}
            if Index^.Right = Previous then {Came from right branch}
            begin
                Tree_Previous := Index ;
                Exit ;
            end else goto Loop
        end ;
        Index := Index^.Left ;
        while Index^.Right <> nil do Index := Index^.Right ;
        Tree_Previous := Index ; {Return new position}
    end ;
end ;


destructor STree64.Done ;

begin
    Tree_Prune( Ptr ) ; {Delete tree by pruning from root node}
    inherited Done ;
end ;


			{BINARY TREE object routines}

constructor Tree64.Init( Size, Key : TRecord_Size ; Flag : boolean ) ;

begin
    STree64.Init( Size, Key, Flag ) ; {Initialize tree}
    Context := nil ; {No current context}
    Last_Error := 0 ; {No errors}
end ;


function Tree64.Error ; {Return error code of last error}

begin
    Error := Last_Error ;
    Last_Error := 0 ;
end ;


procedure Tree64.Set_Error( E : integer ) ; {Set an error code}

begin
    if Last_Error <> 0 then
    begin
	writeln( 'Tree error ', E ) ;
	halt ;
    end ;
    Last_Error := E ;
end ;


procedure Tree64.Find( var Key, Data ) ;
{Find node with specified key (Key) and load return it in Data}

begin
    Context := STree64.Tree_Find( Key ) ; {Locate specified key}
    if Context = nil then
    begin
	Set_Error( - 155 ) ; {Key not found}
	exit ;
    end ;
    STree64.Tree_Get( Context, Data ) ; {Get item from tree}
end ;


procedure Tree64.Update( var Data ) ;

begin
    if Context = nil then
    begin
	Set_Error( - 6 ) ; {No current node}
	exit ;
    end ;
    if Compare( Context^.D[ 0 ], Data, Key_Size ) <> 0 then {Key change}
    begin
	Set_Error( - 138 ) ;
	exit ;
    end ;
    STree64.Tree_Put( Context, Data ) ;
end ;


procedure Tree64.Tree_Add( var Data ) ;

begin
    STree64.Add( Data ) ;
    Context := STree64.Tree_Find( Data ) ;
end ;


procedure Tree64.Delete ;

begin
    if Context = nil then
    begin
	Writeln( 'No context in Tree64.Delete' ) ;
	Halt
    end ;
    STree64.Tree_Delete( Context ) ;
    Context := nil {No more context}
end ;

procedure Tree64.Prune ;

begin
    if Context = nil then
    begin
	Writeln( 'No context in Tree64.Prune' ) ;
	Halt
    end ;
    STree64.Tree_Prune( Context ) ;
    Context := nil {No more context}
end ;


procedure Tree64.First( var Data ) ;

begin
    Context := STree64.Tree_First ;
    if Context = nil then
    begin
	Set_Error( - 11 ) ; {Tree is empty}
	exit ;
    end ;
    STree64.Tree_Get( Context, Data ) ;
end ;


procedure Tree64.Last( var Data ) ;

begin
    Context := STree64.Tree_Last ;
    if Context = nil then
    begin
        Application_Manager^.Fatal_Exception( 'Tree is empty in Tree64.Last' ) ;
    end ;
    STree64.Tree_Get( Context, Data ) ;
end ;


procedure Tree64.Next( var Data ) ;
{Return node which follows current context node.}

begin
    Context := STree64.Tree_Next( Context ) ;
    if Context = nil then
    begin
	Set_Error( - 11 ) ; { End of tree }
	exit ;
    end ;
    STree64.Tree_Get( Context, Data ) ;
end ;


procedure Tree64.Previous( var Data ) ;

begin
    Context := STree64.Tree_Previous( Context ) ;
    if Context = nil then
    begin
        Application_Manager^.Fatal_Exception( 'end of tree in Tree64.Previous' ) ;
    end ;
    STree64.Tree_Get( Context, Data )
end ;


		{ Standard binary tree object routines with large keys }

constructor XTree64.Init( Size : TRecord_Size ; Flag : boolean ) ;

begin
    if Flag and ( Size <> 4 ) then {Tree contains objects, but size is not 4}
    begin
        Application_Manager^.Fatal_Exception( 'Objects in trees must be size 4' ) ;
    end ;
    if( Size > Max_Memory - 21 ) then
    begin
        Application_Manager^.Fatal_Exception( 'Maximum size of tree object data is exceeded' ) ;
    end ;
    Ptr := nil ; {Tree is empty}
    Item_Size := Size ;
    Storing_Objects := Flag ;
end ;


procedure XTree64.Tree_Get( Index : Tree_Pointer ; var Data ) ;

begin
    Move( Index^.D[ 4 ], Data, Item_Size ) {Return data from node}
end ;


procedure XTree64.Tree_Put( Index : Tree_Pointer ; var Data ) ;

begin
    Move( Data, Index^.D[ 4 ], Item_Size ) {Return data from node}
end ;


function XTree64.Find( Key : string ) : Tree_Pointer ;

label Loop ;

var Current : Tree_Pointer ;
    K : PSmall_String ;

begin
    Current := Ptr ; { Start at root }
Loop:
    if( Current <> nil ) then { Not yet at end }
    begin
	move( Current^.D[ 0 ], K, 4 ) ;
        if( Key = K^ ) then { Found item in tree }
        begin
	    Find := Current ;
            Exit
        end ;
        if( Key < K^ ) then
        begin
            Current := Current^.Left ;
        end else
        begin
	    Current := Current^.Right ; { Follow link }
        end ;
        goto Loop ;
    end ;
    Find := nil ; { Item not found }
end ; { XTree64.Find }


procedure XTree64.Add( Key : string ; var Data ) ;

label Loop ;

var Current, Previous : Tree_Pointer ;
    K : PSmall_String ;

begin
    if( Ptr = nil ) then { Tree is empty }
    begin
	Getmem( Ptr, Item_Size + 16 ) ; { Allocate memory for node }
	getmem( K, length( Key ) + 1 ) ;
	K^ := Key ;
	move( K, Ptr^.D[ 0 ], 4 ) ;
	Ptr^.Back := nil ;
	Ptr^.Left := nil ;
	Ptr^.Right := nil ;
	Move( Data, Ptr^.D[ 4 ], Item_Size ) ;
	Exit ;
    end ;
    Current := Ptr ; { Start at root }
    Previous := nil ;
Loop :
    if( Current <> nil ) then { Not yet at end }
    begin
	move( Current^.D[ 0 ], K, 4 ) ;
        if( Key = K^ ) then { Found item in tree }
        begin
            Application_Manager^.Fatal_Exception( 'Duplicate items encountered in tree' ) ;
        end ;
        Previous := Current ; {Save current pointer as previous}
        if( Key < K^ ) then
	begin
	    Current := Current^.Left ;
	end else
	begin
	    Current := Current^.Right ; {Follow link}
	end ;
        goto Loop ;
    end ;
    Getmem( Current, Item_Size + 16 ) ; {Create node}
    getmem( K, length( Key ) + 1 ) ;
    K^ := Key ;
    move( K, Current^.D[ 0 ], 4 ) ;
    move( Previous^.D[ 0 ], K, 4 ) ;
    if( K^ > Key ) then
    begin
	Previous^.Left := Current ;
    end else
    begin
	Previous^.Right := Current ;
    end ;
    Current^.Left := nil ;
    Current^.Right := nil ;
    Current^.Back := Previous ;
    Move( Data, Current^.D[ 4 ], Item_Size ) ;
end ;


procedure XTree64.Tree_Delete( Index : Tree_Pointer ) ;

var Tmp : Tree_Pointer ;
    K : PSmall_String ;
    Work : Tree_Image_Min ;

begin
    Heap.Read( longint( Index ), sizeof( Work ), Work ) ;
    if( ( Work.Right <> nil ) and ( Work.Left = nil ) ) then
    { Right branch but no left branch }
    begin
	Index^.Right^.Back := Index^.Back ;
	if( Ptr = Index )
	    then
		Ptr := Work.Right { New root node }
	    else
		if Index^.Back^.Left = Index
		    then
			Index^.Back^.Left := Work.Right
		    else
			Index^.Back^.Right := Work.Right ;
	move( Index^.D[ 0 ], K, 4 ) ;
	Freemem( K, length( K^ ) + 1 ) ;
	Freemem( Index, Item_Size + 16 ) ; { Delete node }
	Exit ;
    end ;
    if ( Work.Left = nil ) and ( Work.Right = nil ) then {Terminal node}
    begin
	if( Ptr = Index )
	    then
		Ptr := nil {Tree is empty}
	    else
		if Index^.Back^.Left = Index
		    then
			Index^.Back^.Left := nil
		    else
			Index^.Back^.Right := nil ;
	move( Index^.D[ 0 ], K, 4 ) ;
	Freemem( K, length( K^ ) + 1 ) ;
	Freemem( Index, Item_Size + 16 ) ; {Delete node}
	Exit ;
    end ;
    {Bring up left branch}
    Index^.Left^.Back := Index^.Back ;
    if( Ptr = Index )
	then
	    Ptr := Work.Left { New root node }
	else
	    if Index^.Back^.Left = Index
		then
		    Index^.Back^.Left := Work.Left
		else
		    Index^.Back^.Right := Work.Left ;
    if( Work.Right <> nil ) then { A right branch to handle }
    begin
	Tmp := Index^.Left ; {Start at replacement node}
	while( Tmp^.Right <> nil ) do
        begin
            Tmp := Tmp^.Right ; {Go all the way right}
        end ;
	Tmp^.Right := Index^.Right ;
	Tmp^.Right^.Back := Tmp ;
    end ;
    move( Index^.D[ 0 ], K, 4 ) ;
    Heap.Deallocate( longint( K ), length( K^ ) + 1 ) ;
    Heap.Deallocate( longint( Index ), Item_Size + 16 ) ; { Actually delete node }
end ; { XTree64.Tree_Delete }


procedure XTree64.Tree_Prune( Index : Tree_Pointer ) ;

begin
    while( Index^.Left <> nil ) do
    begin
        Tree_Delete( Index^.Left ) ; {Delete left branch}
    end ;
    while Index^.Right <> nil do
    begin
        Tree_Delete( Index^.Right ) ; {Delete right branch}
    end ;
    Tree_Delete( Index ) ; {and delete sub-tree root node}
end ;


function XTree64.Tree_First : Tree_Pointer ;

var Tmp : Tree_Pointer ;

begin
    if( Ptr = nil ) then {Tree is empty}
    begin
	Tree_First := nil ;
	Exit ;
    end ;
    Tmp := Ptr ; {Start at root node}
    while( Tmp^.Left <> nil ) do
    begin
        Tmp := Tmp^.Left ; {Go all the way left}
    end ;
    Tree_First := Tmp {Return result}
end ;


function XTree64.Tree_Last : Tree_Pointer ;

var Tmp : Tree_Pointer ;

begin
    if( Ptr = nil ) then {Tree is empty}
    begin
	Tree_Last := nil ;
	Exit ;
    end ;
    Tmp := Ptr ;
    while( Tmp^.Right <> nil ) do Tmp := Tmp^.Right ;
    Tree_Last := Tmp ; {Return result}
end ;


function XTree64.Empty : boolean ;

begin
    if( Ptr = nil )
	then
	    Empty := True
	else
	    Empty := False ;
end ;


function XTree64.Tree_Next( Index : Tree_Pointer ) : Tree_Pointer ;

label Loop ;

var Previous : Tree_Pointer ;

begin
    if( Ptr = nil ) then
    begin
	Tree_Next := nil
    end else
    begin
        if Index^.Right = nil then
        begin
Loop:
            if( Index^.Back = nil ) then {Cannot go further up tree}
            begin
                Tree_Next := nil ; {No more items in tree}
                Exit ;
            end ;
            Previous := Index ; {Save previous position}
            Index := Index^.Back ; {Backup one level}
            if Index^.Left = Previous then {Came from left branch}
            begin
                Tree_Next := Index ;
                Exit ;
            end else goto Loop {Backup again}
        end ;
        Index := Index^.Right ; {Go down right branch}
        while Index^.Left <> nil do Index := Index^.Left ; {Go left}
        Tree_Next := Index {Return result}
    end ;
end ;


function XTree64.Tree_Previous( Index : Tree_Pointer ) : Tree_Pointer ;

label Loop ;

var Previous : Tree_Pointer ;

begin
    if( Ptr = nil ) then { Nothing in tree }
    begin
	Tree_Previous := nil ;
    end else
    begin
        if Index^.Left = nil then
        begin
Loop:
            if( Index^.Back = nil ) then {Cannot go further up tree}
            begin
                Tree_Previous := nil ; {No more items in tree}
                Exit ;
            end ;
            Previous := Index ; {Save current location in tree}
            Index := Index^.Back ; {Backup one level}
            if Index^.Right = Previous then {Came from right branch}
            begin
                Tree_Previous := Index ;
                Exit ;
            end else goto Loop
        end ;
        Index := Index^.Left ;
        while Index^.Right <> nil do Index := Index^.Right ;
        Tree_Previous := Index {Return new position}
    end ;
end ;


destructor XTree64.Done ;

begin
    Tree_Prune( Ptr ) ; {Delete tree by pruning from root node}
    inherited Done ;
end ;


{ TBTree methods... }

constructor TBTree64.Init( Size, Key, Max, Bucket : TRecord_Size ; Flag : boolean ) ;

begin
    if( Flag and ( Size <> 4 ) ) then { Tree contains objects, but size is not 4 }
    begin
        Application_Manager^.Fatal_Exception( 'Objects in trees must be size 4' ) ;
    end ;
    if( Size > Max_Memory - 3 ) then
    begin
        Application_Manager^.Fatal_Exception( 'Maximum size of B-tree object data is exceeded' ) ;
    end ;
    if( Bucket * Size + 4 > Max_Memory - 3 ) then
    begin
        Application_Manager^.Fatal_Exception( 'Maximum size of B-tree bucket is exceeded' ) ;
    end ;
    Item_Size := Size ;
    Key_Size := Key ;
    Storing_Objects := Flag ;
    Max_B := Max ;
    Bucket_Size := Bucket ;
    Ptr := nil ;
end ; { TBTree64.Init }


function TBTree64.Tree_Find( var Key ) : Tree_Pointer ;

    function _Find( Index : BTree_Pointer ) : BTree_Pointer ;

    var Active : integer ;

    begin
        _Find := nil ; { Assume failure }
        for Active := 0 to Index^.Active - 1 do
        begin
            {~~~}
        end ;
    end ;

begin
    if( Ptr = nil ) then { Nothing in tree }
    begin
        Tree_Find := nil ;
        exit ;
    end ;

    Tree_Find := Tree_Pointer( _Find( Ptr ) ) ;
end ;


procedure TBTree64.Tree_Get( Index : Tree_Pointer ; var Data ) ;

var _Index : BTree_Pointer ;

begin
    _Index := BTree_Pointer( PChar( Index ) - BTree_Overhead ) ;
    if( ( _Index^.Level <> 0 ) or ( _Index.Active < 0 ) ) then
    begin
        Application_Manager^.Fatal_Exception( 'Illegal record context' ) ;
    end ;
    move( _Index^.D, Data, _Index^.Active ) ;
end ;


procedure TBTree64.Tree_Put( Index : Tree_Pointer ; var Data ) ;

var _Index : BTree_Pointer ;

begin
    _Index := BTree_Pointer( PChar( Index ) - BTree_Overhead ) ;
    if( ( _Index^.Level <> 0 ) or ( _Index.Active < 0 ) ) then
    begin
        Application_Manager^.Fatal_Exception( 'Illegal record context' ) ;
    end ;
    move( Data, _Index^.D, _Index^.Active ) ;
end ;


procedure TBTree64.Add( var Data ) ;

var Index : BTree_Pointer ;

begin
    if( Ptr = nil ) then { First record }
    begin
        { Create data record... }
        Getmem( Ptr, BTree_Overhead + Item_Size ) ; { Allocate array on heap }
        Ptr^.Level := 0 ;
        Ptr^.Active := Item_Size ;
        move( Data, Ptr^.D, Item_Size ) ;
        Getmem( Ptr^.Parent, BTree_Overhead + ( Max_B * sizeof( BTree_Pointer ) ) ) ;
        Index := Ptr ;
        Ptr := Ptr^.Parent ;

        { Create parent key record... }
        Ptr^.Level := 1 ;
        Ptr^.Active := 1 ;
        Ptr.Parent := nil ;
        fillchar( Ptr^.D, Max_B * sizeof( BTree_Pointer ), 0 ) ;
        PPointer_Array( @Ptr^.D )[ 0 ] := Index ;
        exit ;
    end ; { if( Ptr = nil ) }

    { Find place to insert record... }
    {~~~}
end ; { TBTree64.Add }


procedure TBTree64.Tree_Delete( Index : Tree_Pointer ) ;

var Active : integer ;
    Count : integer ;
    Found : boolean ;
    _Index, I : BTree_Pointer ;

begin
    { Adjust pointer to beginning of record header }
    _Index := BTree_Pointer( PChar( Index ) - BTree_Overhead ) ;

    { Validate valid header information }
    if( ( _Index^.Level <> 0 ) or ( _Index.Active < 0 ) ) then
    begin
        Application_Manager^.Fatal_Exception( 'Illegal record context' ) ;
    end ;

    { Delete the record... }
    _Index.Active := -_Index.Active ; { Mark as deleted }
    freemem( _Index, BTree_Overhead - _Index^.Active ) ;
    I := _Index ;
    _Index := PPointer_Array( @_Index^.Parent^.D )[ 0 ] ; { Move to key record... }

    { Remove pointer to deleted record from key record... }
    Count := 0 ;
    Found := False ;
    for Active := 0 to _Index^.Active - 1 do
    begin
        if( I = PPointer_Array( @_Index^.D )[ Active ] ) then
        begin
            PPointer_Array( @_Index^.D )[ Active ] := nil ;
            Found := True ;
        end else
        begin
            if( PPointer_Array( @_Index^.D )[ Active ] <> nil ) then
            begin
                inc( Count ) ;
            end ;
        end ;
        if( Found and ( Count > 0 ) ) then { Early out }
        begin
            exit ;
        end ;
    end ;
    if( Count = 0 ) then { Deleted our last record }
    begin
        Tree_Prune( Tree_Pointer( _Index ) ) ;
    end ;
end ; { TBTree64.Tree_Delete }


procedure TBTree64.Tree_Prune( Index : Tree_Pointer ) ;

var Active : integer ;
    Count : integer ;
    Found : boolean ;
    I : BTree_Pointer ;
    _Index : BTree_Pointer ;

label Delete_Key ;

begin
    if( Ptr = nil ) then { Nothing in tree }
    begin
        exit ;
    end ;
    _Index := BTree_Pointer( Index ) ;
    if( _Index^.Level = 0 ) then
    begin
        Tree_Delete( Index ) ;
        exit ;
    end ;

    { Delete all sub-levels }
    for Active := 0 to _Index^.Active - 1 do
    begin
        I := PPointer_Array( @_Index^.D )[ Active ] ;
        if( I <> nil ) then
        begin
	    if( _Index^.Level = 1 ) then { Deleting a data record }
            begin
                I.Active := -I.Active ; { Mark as deleted }
                freemem( I, BTree_Overhead - I^.Active ) ;
            end else
            begin
                Tree_Prune( Tree_Pointer( I ) ) ; { Prune sub-tree }
            end ;
        end ;
    end ;

Delete_Key:
    { Now delete ourself... }
    _Index^.Active := 0 ;
    I := _Index ;
    _Index := _Index^.Parent ;
    freemem( I, BTree_Overhead + ( Max_B * sizeof( BTree_Pointer ) ) ) ;

    { Remove ourself from our parent key record... }
    if( _Index = nil ) then { Deleted entire tree }
    begin
        Ptr := nil ;
        exit ;
    end ;
    Count := 0 ;
    Found := False ;
    for Active := 0 to _Index^.Active - 1 do
    begin
        if( I = PPointer_Array( @_Index^.D )[ Active ] ) then
        begin
            PPointer_Array( @_Index^.D )[ Active ] := nil ;
            Found := True ;
        end else
        begin
            if( PPointer_Array( @_Index^.D )[ Active ] <> nil ) then
            begin
                inc( Count ) ;
            end ;
        end ;
        if( Found and ( Count > 0 ) ) then { Early out }
        begin
            exit ;
        end ;
    end ;
    if( Count = 0 ) then { Deleted our last record }
    begin
        goto Delete_Key ; { Go delete ourself }
    end ;
end ; { TBTree64.Tree_Prune }


function TBTree64.Tree_First : Tree_Pointer ;

var Active : integer ;
    I, Image : BTree_Pointer ;

begin
    if( Ptr = nil ) then { Nothing in tree }
    begin
        Tree_First := nil ;
        exit ;
    end ;
    Image := Ptr ;
    while( Image^.Level <> 0 ) do { Go down to data level }
    begin
        for Active := 0 to Image^.Active - 1 do { Search for a valid record }
        begin
            I := PPointer_Array( @Image^.D )[ Active ] ;
            if( I <> nil ) then
            begin
                Image := I ;
                break ;
            end ;
        end ;
    end ;
    Tree_First := Tree_Pointer( PPointer_Array( @Image^.D )[ 0 ] ) ;
end ;


function TBTree64.Tree_Last : Tree_Pointer ;

var Active : integer ;
    I, Image : BTree_Pointer ;

begin
    if( Ptr = nil ) then { Nothing in tree }
    begin
        Tree_Last := nil ;
        exit ;
    end ;
    Image := Ptr ;
    while( Image^.Level <> 0 ) do { Go down to data level }
    begin
        for Active := Image^.Active - 1 downto 0 do { Search for a valid record }
        begin
            I := PPointer_Array( @Image^.D )[ Active ] ;
            if( I <> nil ) then
            begin
                Image := I ;
                break ;
            end ;
        end ;
    end ;
    Tree_Last := Tree_Pointer( PPointer_Array( @Image^.D )[ 0 ] ) ;
end ;


function TBTree64.Empty : boolean ;

begin
    Empty := ( Ptr = nil ) ;
end ;


function TBTree64.Tree_Next( Index : Tree_Pointer ) : Tree_Pointer ;

    function _Next( Key, Data : BTree_Pointer ) : BTree_Pointer ;

    var Active : integer ;
        Found : boolean ;

    begin
        if( Key = nil ) then
        begin
            _Next := nil ;
            exit ;
        end ;
        Found := False ;
        for Active := 0 to Key^.Active - 1 do
        begin
            if( Found and ( PPointer_Array( @Key^.D )[ Active ] <> nil ) ) then
            begin
                _Next := PPointer_Array( @Key^.D )[ Active ] ;
                exit ;
            end ;
            if( Data = PPointer_Array( @Key^.D )[ Active ] ) then
            begin
                Found := True ;
            end ;
        end ;

        { Data was last record at this level... }
        _Next := _Next( Key^.Parent, Key ) ;
    end ;

var Active : integer ;
    _Index, I : BTree_Pointer ;

begin { TBTree64.Tree_Next }
    { Adjust pointer to beginning of record header }
    _Index := BTree_Pointer( PChar( Index ) - BTree_Overhead ) ;

    { Validate valid header information }
    if( ( _Index^.Level <> 0 ) or ( _Index.Active < 0 ) ) then
    begin
        Application_Manager^.Fatal_Exception( 'Illegal record context' ) ;
    end ;

    _Index := _Next( _Index^.Parent, _Index ) ;
    if( _Index <> nil ) then
    begin
        while( _Index^.Level > 0 ) do
        begin
            for Active := 0 to _Index^.Active - 1 do
            begin
                I := PPointer_Array( @_Index^.D )[ Active ] ;
                if( I <> nil ) then
                begin
                    _Index := I ;
                    break ;
                end ;
            end ;
        end ;
    end ;
    Tree_Next := Tree_Pointer( PChar( _Index ) + BTree_Overhead ) ;
end ; { TBTree64.Tree_Next }


function TBTree64.Tree_Previous( Index : Tree_Pointer ) : Tree_Pointer ;

    function _Next( Key, Data : BTree_Pointer ) : BTree_Pointer ;

    var Active : integer ;
        Found : boolean ;

    begin
        if( Key = nil ) then
        begin
            _Next := nil ;
            exit ;
        end ;
        Found := False ;
        for Active := Key^.Active - 1 downto 0 do
        begin
            if( Found and ( PPointer_Array( @Key^.D )[ Active ] <> nil ) ) then
            begin
                _Next := PPointer_Array( @Key^.D )[ Active ] ;
                exit ;
            end ;
            if( Data = PPointer_Array( @Key^.D )[ Active ] ) then
            begin
                Found := True ;
            end ;
        end ;

        { Data was last record at this level... }
        _Next := _Next( Key^.Parent, Key ) ;
    end ;

var Active : integer ;
    _Index, I : BTree_Pointer ;

begin { TBTree64.Tree_Previous }
    { Adjust pointer to beginning of record header }
    _Index := BTree_Pointer( PChar( Index ) - BTree_Overhead ) ;

    { Validate valid header information }
    if( ( _Index^.Level <> 0 ) or ( _Index.Active < 0 ) ) then
    begin
        Application_Manager^.Fatal_Exception( 'Illegal record context' ) ;
    end ;

    _Index := _Next( _Index^.Parent, _Index ) ;
    if( _Index <> nil ) then
    begin
        while( _Index^.Level > 0 ) do
        begin
            for Active := _Index^.Active - 1 downto 0 do
            begin
                I := PPointer_Array( @_Index^.D )[ Active ] ;
                if( I <> nil ) then
                begin
                    _Index := I ;
                    break ;
                end ;
            end ;
        end ;
    end ;
    Tree_Previous := Tree_Pointer( PChar( _Index ) + BTree_Overhead ) ;
end ; { TBTree64.Tree_Previous }


destructor TBTree64.Done ;

begin
    Tree_Prune( Tree_Pointer( Ptr ) ) ; { Delete tree by pruning from root node }
    inherited Done ;
end ;



{ TMatrix methods... }

constructor TMatrix64.Init( Size, Elements : longint ) ;

begin
    inherited Init ;

    Ptr := 0 ;
    Element_Size := Size ;
    Element_Count := Elements ;
end ;


destructor TMatrix64.Done ;

begin
    Clear ;
    inherited Done ;
end ;


procedure TMatrix64.Clear ;

begin
    if( Ptr <> 0 ) then
    begin
        { Release matrix storage }
    end ;
end ;


procedure TMatrix64.Matrix_Get( Index : string ; var Data ) ;

begin
end ;


procedure TMatrix64.Matrix_Put( Index : string ; var Data ) ;

begin
end ;



{ TSparse_Matrix methods... }

constructor TSparse_Matrix64.Init( Size, Elements : longint ) ;

begin
    inherited Init( Size, Elements ) ;

    fillchar( Last_Index, sizeof( Last_Index ), 255 ) ;
    Did_Clear := False ;
end ;


destructor TSparse_Matrix64.Done ;

begin
    Ptr := 0 ; { So our ancestor doesn't try to release it. }
    inherited Done ;
end ;


procedure TSparse_Matrix64.Clear ;

var Cleared : boolean ;
    Element : TSparse_Data_Header ;
    Saved : TStore_Address64 ;

begin
    Cleared := False ;
    if( not Did_Clear ) then { Only clear cache once during a clear operation }
    begin
        Did_Clear := True ;
        Cleared := True ;
        fillchar( Last_Index, sizeof( Last_Index ), 255 ) ;
    end ;
    while( Ptr <> 0 ) do
    begin
        Heap^.Read( Ptr, sizeof( Element ), Element ) ;
        if( Element_Count > 1 ) then { Clear sub-array }
        begin
            dec( Element_Count ) ;
            Saved := Ptr ;
            Ptr := Element.Data ;
            Clear ;
            Ptr := Saved ;
            inc( Element_Count ) ;
        end else
        begin
            if( Element_Size > sizeof( Element.Data ) ) then
            begin
                Heap^.Deallocate( Element.Data, Element_Size ) ;
            end ;
        end ;

        { Now deallocate this element... }
        Heap^.Deallocate( Ptr, sizeof( Element ) ) ;
        Ptr := Element.Next ;
    end ; { while( Ptr <> nil ) }
    if( Cleared ) then
    begin
        Did_Clear := False ; { Allow next clear to work }
    end ;
end ; { TSparse_Matrix64.Clear }


function TSparse_Matrix64.Element_Exists( Index : string ) : boolean ;

var Dummy : integer ;
    Element : TSparse_Data_Header ;
    P : integer ;
    This_Element_Ptr : TStore_Address64 ;
    This_Index : longint ;
    Saved_Ptr : TStore_Address64 ;

begin
    { Setup... }
    Element_Exists := False ; { Assume failure }
    P := pos( ',', Index + ',' ) ;
    val( copy( Index, 1, P - 1 ), This_Index, Dummy ) ;
    Index := copy( Index, P + 1, length( Index ) ) ;
    This_Element_Ptr := Ptr ;
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        exit ;
    end ;
    Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;

    { Now find this element... }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        This_Element_Ptr := Element.Next ;
        Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;
    if(
        ( This_Index < Element.Low_Index )
        or
        ( This_Index > Element.High_Index )
      ) then { Element doesn't exist }
    begin
        exit ;
    end ;
    if( length( Index ) = 0 ) then { Getting data, rather than another subscript }
    begin
        if( Element_Count > 1 ) then { Incomplete subscript }
        begin
            exit ;
        end ;
        Element_Exists := True ;
        exit ;
    end else { Getting another level of subscripting }
    begin
        { Save our state and then point ourselves at the root of the next level }
        Saved_Ptr := Ptr ;
        Ptr := Element.Data ;
        dec( Element_Count ) ;

        { Now get the data from the subarray }
        Element_Exists := Element_Exists( Index ) ;

        { Restore our state }
        Ptr := Saved_Ptr ;
        inc( Element_Count ) ;
    end ;
end ; { TSparse_Matrix64.Element_Exists }


procedure TSparse_Matrix64.Matrix_Get( Index : string ; var Data ) ;

var Dummy : integer ;
    Element : TSparse_Data_Header ;
    P : integer ;
    This_Element_Ptr : TStore_Address64 ;
    This_Index : longint ;
    Saved_Ptr : TStore_Address64 ;

begin
    { Setup... }
    P := pos( ',', Index + ',' ) ;
    val( copy( Index, 1, P - 1 ), This_Index, Dummy ) ;
    Index := copy( Index, P + 1, length( Index ) ) ;
    This_Element_Ptr := Ptr ;
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        fillchar( Data, Element_Size, 0 ) ;
        exit ;
    end ;
    Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;

    { Now find this element... }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        Temp_Ptr := This_Element_Ptr ;
        This_Element_Ptr := Element.Next ;
        Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;
    if(
        ( This_Index < Element.Low_Index )
        or
        ( This_Index > Element.High_Index )
      ) then { Element doesn't exist }
    begin
        fillchar( Data, Element_Size, 0 ) ;
        exit ;
    end ;
    if( length( Index ) = 0 ) then { Getting data, rather than another subscript }
    begin
        if( Element_Size <= sizeof( Element.Data ) ) then
        begin
            move( Element.Data, Data, Element_Size ) ;
        end else
        begin
            Heap^.Read( Element.Data, Element_Size, Data ) ;
        end ;
    end else { Getting another level of subscripting }
    begin
        { Save our state and then point ourselves at the root of the next level }
        Saved_Ptr := Ptr ;
        Ptr := Element.Data ;

        { Now get the data from the subarray }
        Matrix_Get( Index, Data ) ;

        { Restore our state }
        Ptr := Saved_Ptr ;
    end ;
end ; { TSparse_Matrix64.Matrix_Get }


procedure TSparse_Matrix64._Put( Level : integer ; Index : string ; var Data ) ;

var Buffer : PChar ;
    Dummy : integer ;
    Element, Element1 : TSparse_Data_Header ;
    Old_Ptr : TStore_Address64 ;
    P : integer ;
    This_Element_Ptr, Previous_Ptr, Next_Ptr : TStore_Address64 ;
    This_Index : longint ;
    Saved_Next : TStore_Address64 ;
    Saved_Ptr : TStore_Address64 ;

    procedure Set_Data ;

    begin
        if( Element_Size <= sizeof( Element.Data ) ) then { If data fits within the element }
        begin
            move( Data, Element.Data, Element_Size ) ;
        end else
        begin
            Element.Data := Heap^.Allocate( Element_Size ) ;
            Heap^.Write( Element.Data, Element_Size, Data ) ;
        end ;
    end ;

begin { TSparse_Matrix64._Put }
    { Setup... }
    P := pos( ',', Index + ',' ) ;
    val( copy( Index, 1, P - 1 ), This_Index, Dummy ) ;
    Index := copy( Index, P + 1, length( Index ) ) ;
    This_Element_Ptr := Ptr ;

    { Handle empty matrix }
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
        Ptr := This_Element_Ptr ;
        Element.Low_Index := This_Index ;
        Element.High_Index := This_Index ;
        Element.Next := 0 ;
        if( length( Index ) = 0 ) then
        begin
            Set_Data ;
        end else { Write next level }
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ; { Save our state }
            Ptr := 0 ;
            _Put( Level + 1, Index, Data ) ;
            Element.Data := Ptr ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end ;
        Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
        exit ;
    end ; { if( This_Element_Ptr = 0 ) }

    { Prepare to search structure... }
    Previous_Ptr := 0 ;
    if(
        ( Level <= Max_Cached_Sparse_Matrix )
        and
        ( Last_Index[ Level ] > -1 )
        and
        ( Last_Index[ Level ] <= This_Index )
      ) then
    begin
        This_Element_Ptr := Last_Ptr[ Level ] ;
        Previous_Ptr := Last_Previous_Ptr[ Level ] ;
    end ;
    Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;

    { Now find this element }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        Previous_Ptr := This_Element_Ptr ;
        This_Element_Ptr := Element.Next ;
        Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;

    { Handle cache }
    if( Level <= Max_Cached_Sparse_Matrix ) then
    begin
        if( Last_Index[ Level ] <> Element.Low_Index ) then
        begin
            fillchar( Last_Index[ Level + 1 ], ( Max_Cached_Sparse_Matrix - Level ) * sizeof( TStore_Address64 ), 255 ) ;
        end ;
        Last_Index[ Level ] := Element.Low_Index ;
        Last_Ptr[ Level ] := This_Element_Ptr ;
        Last_Previous_Ptr[ Level ] := Previous_Ptr ;
    end ;

    { Perform the put operation... }
    { where "Element" is the element found above,
      if Element.Low_Index > This_Index then insert before Element (condition 1)
          if Element.Low_Index = This_Index + 1 and same value as Element then extend Element range (condition 1A)
          else create new element record (condition 1B)
      else
      if Element.High_Index >= This_Index then change within Element (condition 2)
          if same as existing value, do nothing (condition 2A)
          else
          if Element.Low_Index = Element.High_Index then change existing single item (condition 2B)
          else
          if Element.Low_Index = This_Index then change first element and split rest into new record (condition 2C)
          else
          if Element.High_Index = This_Index then insert new item after Element (condition 2D)
          else change element within range of Element (condition 2E)
      else New element record (condition 3)
    }
    if( Element.Low_Index > This_Index ) then { Inserting new element prior to item - condition 1 }
    begin
        if( Element.Low_Index = This_Index + 1 ) then { Just before next item }
        begin
            if( length( Index ) = 0 ) then { A data reference }
            begin
                getmem( Buffer, Element_Size ) ;
                if( Element_Size <= sizeof( Element.Data ) ) then
                begin
                    move( Element.Data, Buffer^, Element_Size ) ;
                end else
                begin
                    Heap^.Read( Element.Data, Element_Size, Buffer^ ) ;
                end ;
                if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { Same value as next item - condition 1A }
                begin
                    { If inserting the same value, just prior to another element
                      then adjust index on existing element. }
                    dec( Element.Low_Index ) ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                    freemem( Buffer, Element_Size ) ;
                    if( Level <= Max_Cached_Sparse_Matrix ) then
                    begin
                        Last_Index[ Level ] := Element.Low_Index ;
                    end ;
                    exit ;
                end ;
                freemem( Buffer, Element_Size ) ;
            end ; { if( length( Index ) <> 0 ) }
        end ; { if( Element.Low_Index = This_Index + 1 ) }

        { New element prior to This_Element_Ptr - condition 1B }

        { First create the new header... }
        Element.Next := This_Element_Ptr ;
        This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
        Element.Low_Index := This_Index ;
        Element.High_Index := This_Index ;
        if( Level <= Max_Cached_Sparse_Matrix ) then
        begin
            Last_Index[ Level ] := Element.Low_Index ;
            Last_Ptr[ Level ] := This_Element_Ptr ;
            Last_Previous_Ptr[ Level ] := Previous_Ptr ;
        end ;
        if( length( Index ) = 0 ) then { A data reference }
        begin
            Set_Data ;
        end else
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ; { Save our state }
            Ptr := 0 ;
            _Put( Level + 1, Index, Data ) ;
            Element.Data := Ptr ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end ;
        Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;

        { Now insert into list... }
        if( Previous_Ptr = 0 ) then { New first element }
        begin
            { Insert at beginning of list }
            Ptr := This_Element_Ptr ;
        end else
        begin
            { Insert after Previous_Ptr... }
            Heap^.Read( Previous_Ptr, sizeof( Element ), Element ) ;
            Element.Next := This_Element_Ptr ;
            Heap^.Write( Previous_Ptr, sizeof( Element ), Element ) ;
(*
            if( Level <= Max_Cached_Sparse_Matrix ) then
            begin
                fillchar( Last_Index[ Level ], ( Max_Cached_Sparse_Matrix - Level + 1 ) * sizeof( TStore_Address64 ), 255 ) ;
            end ;
*)
        end ;
        exit ;
    end ; { if( Element.Low_Index > This_Index ) }
    
    if( Element.High_Index >= This_Index ) then { Changing an existing value - condition 2 }
    begin
        if( length( Index ) > 0 ) then { More subscripts to process }
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ;
            Ptr := Element.Data ; { Save our state }
            Old_Ptr := Element.Data ;
            dec( Element_Count ) ;
            _Put( Level + 1, Index, Data ) ;
            if( Ptr <> Old_Ptr ) then { Put changed root pointer }
            begin
                Element.Data := Ptr ;
                Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
            end ;

            { Restore our saved state... }
            Ptr := Saved_Ptr ;
            inc( Element_Count ) ;
        end else
        begin
            { Changing an already-existing value... }
            getmem( Buffer, Element_Size ) ;
            if( Element_Size <= sizeof( Element.Data ) ) then
            begin
                move( Element.Data, Buffer^, Element_Size ) ;
            end else
            begin
                Heap^.Read( Element.Data, Element_Size, Buffer^ ) ;
            end ;
            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { Same value as existing item }
            begin
                freemem( Buffer, Element_Size ) ; { Nothing to do - cleanup and exit - condition 2A }
                exit ;
            end ;

            { Changing a single item - condition 2B }
            if( Element.Low_Index = Element.High_Index ) then
            begin
                if( Element_Size <= sizeof( Element.Data ) ) then
                begin
                    move( Data, Element.Data, Element_Size ) ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                end else
                begin
                    Heap^.Write( Element.Data, Element_Size, Data ) ;
                end ;

                { See if we can combine with the next item }
                if( Element.Next <> 0 ) then { There is a next index }
                begin
                    Heap^.Read( Element.Next, sizeof( Element1 ), Element1 ) ;
                    if( Element1.Low_Index = Element.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Saved_Next := Element.Next ;
                            Element.Next := Element1.Next ;
                            Element.High_Index := Element1.High_Index ;
                            Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                            Heap^.Deallocate( Saved_Next, sizeof( Element1 ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element1.Data, Element_Size ) ;
                            end ;
                        end ;
                    end ;
                end ; { if( Element.Next <> 0 ) }

                { Now see if we can combine with the previous item }
                if( Previous_Ptr <> 0 ) then { There was a previous index }
                begin
                    Heap^.Read( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                    if( Element.Low_Index = Element1.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Element1.Next := Element.Next ;
                            Element1.High_Index := Element.High_Index ;
                            Heap^.Write( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                            Heap^.Deallocate( This_Element_Ptr, sizeof( Element ) ) ;
                            fillchar( Last_Index[ Level ], ( Max_Cached_Sparse_Matrix - Level + 1 ) * sizeof( TStore_Address64 ), 255 ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element.Data, Element_Size ) ;
                            end ;
                        end ; { if( Compare_Memory( Data, Buffer^, Element_Size ) ) }
                    end ; { if( Element.Low_Index = Element1.High_Index + 1 ) }
                end ; { if( Previous_Ptr <> 0 ) }
            end else

            { Need to split... - condition 2C }
            if( Element.Low_Index = This_Index ) then
            begin
                { Put new data at current location and rest of existing at new place }
                Next_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                Element.Low_Index := This_Index + 1 ;
                Heap^.Write( Next_Ptr, sizeof( Element ), Element ) ;
                Element.Next := Next_Ptr ;
                Element.Low_Index := This_Index ;
                Element.High_Index := This_Index ;
                Set_Data ;
                Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;

                { Now see if we can combine with the previous item }
                if( Previous_Ptr <> 0 ) then { There was a previous index }
                begin
                    Heap^.Read( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                    if( Element.Low_Index = Element1.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Element1.Next := Element.Next ;
                            Element1.High_Index := Element.High_Index ;
                            Heap^.Write( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                            fillchar( Last_Index[ Level ], ( Max_Cached_Sparse_Matrix - Level + 1 ) * sizeof( TStore_Address64 ), 255 ) ;
                            Heap^.Deallocate( This_Element_Ptr, sizeof( Element ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element.Data, Element_Size ) ;
                            end ;
                        end ; { if( Compare_Memory( Data, Buffer^, Element_Size ) ) }
                    end ; { if( Element.Low_Index = Element1.High_Index + 1 ) }
                end ; { if( Previous_Ptr <> 0 ) }
            end else
            begin
                if( Element.High_Index = This_Index ) then { condition 2D  }
                begin
                    { Insert new element just after the existing one... }
                    Next_Ptr := Element.Next ;
                    Element.High_Index := This_Index - 1 ;
                    Element.Next := Heap^.Allocate( sizeof( Element ) ) ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                    This_Element_Ptr := Element.Next ;
                    Element.Next := Next_Ptr ;
                    Element.Low_Index := This_Index ;
                    Element.High_Index := This_Index ;
                    Set_Data ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;

                    { See if we can combine with the next item }
                    if( Element.Next <> 0 ) then { There is a next index }
                    begin
                        Heap^.Read( Element.Next, sizeof( Element1 ), Element1 ) ;
                        if( Element1.Low_Index = Element.High_Index + 1 ) then { Immediately follows }
                        begin
                            if( Element_Size <= sizeof( Element.Data ) ) then
                            begin
                                move( Element1.Data, Buffer^, Element_Size ) ;
                            end else
                            begin
                                Heap^.Read( Element1.Data, Element_Size, Buffer^ ) ;
                            end ;
                            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                            begin
                                Saved_Next := Element.Next ;
                                Element.Next := Element1.Next ;
                                Element.High_Index := Element1.High_Index ;
                                Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                                fillchar( Last_Index[ Level ], ( Max_Cached_Sparse_Matrix - Level + 1 ) * sizeof( TStore_Address64 ), 255 ) ;
                                Heap^.Deallocate( Saved_Next, sizeof( Element1 ) ) ;
                                if( Element_Size > sizeof( Element.Data ) ) then
                                begin
                                    Heap^.Deallocate( Element1.Data, Element_Size ) ;
                                end ;
                            end ;
                        end ;
                    end ;
                end else
                begin
                    { New data is in the middle of an existing range... condition 2E }

                    { Copy high part of range to new range element header... }
                    Next_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                    Element.Low_Index := This_Index + 1 ;
                    if( Element_Size > sizeof( Element.Data ) ) then
                    begin
                        { Copy the data to a new location... }
                        getmem( Buffer, Element_Size ) ;
                        Heap^.Read( Element.Data, Element_Size, Buffer^ ) ;
                        Element.Data := Heap^.Allocate( Element_Size ) ;
                        Heap^.Write( Element.Data, Element_Size, Buffer^ ) ;
                        freemem( Buffer, Element_Size ) ;
                    end ;
                    Heap^.Write( Next_Ptr, sizeof( Element ), Element ) ;

                    { Now create the new element... }
                    Previous_Ptr := This_Element_Ptr ;
                    This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                    Element.Low_Index := This_Index ;
                    Element.High_Index := This_Index ;
                    Element.Next := Next_Ptr ;
                    Set_Data ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                    if( Level <= Max_Cached_Sparse_Matrix ) then
                    begin
                        Last_Index[ Level ] := Element.Low_Index ;
                        Last_Ptr[ Level ] := This_Element_Ptr ;
                        Last_Previous_Ptr[ Level ] := Previous_Ptr ;
                    end ;

                    { Now update existing low part of range... }
                    Heap^.Read( Previous_Ptr, sizeof( Element ), Element ) ;
                    Element.High_Index := This_Index - 1 ;
                    Element.Next := This_Element_Ptr ;
                    Heap^.Write( Previous_Ptr, sizeof( Element ), Element ) ;
                end ;
            end ; { if( Element.Low_Index = This_Index ) }
            freemem( Buffer, Element_Size ) ;
        end ; { if( length( Index ) <> 0 ) }
        exit ;
    end ; { if( Element.High_Index >= This_Index ) }

    { Only other possibility is that this is a new element at the end of the
      list. - condition 3 }

    if( length( Index ) = 0 ) then { Data }
    begin
        { Now see if we can combine with the previous item }
        if( This_Index = Element.High_Index + 1 ) then { Immediately follows }
        begin
            getmem( Buffer, Element_Size ) ;
            if( Element_Size <= sizeof( Element.Data ) ) then
            begin
                move( Element.Data, Buffer^, Element_Size ) ;
            end else
            begin
                Heap^.Read( Element.Data, Element_Size, Buffer^ ) ;
            end ;
            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
            begin
                Element.High_Index := This_Index ;
                Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                freemem( Buffer, Element_Size ) ;
                exit ;
            end ; { if( Compare_Memory( Data, Buffer^, Element_Size ) ) }
            freemem( Buffer, Element_Size ) ;
        end ; { if( Element.Low_Index = Element1.High_Index + 1 ) }
    end ; { if( length( Index ) = 0 ) }

    Element.Next := Heap^.Allocate( sizeof( Element ) ) ;
    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
    This_Element_Ptr := Element.Next ;
    Element.Next := 0 ; { End of list }
    Element.Low_Index := This_Index ;
    Element.High_Index := This_Index ;
    if( length( Index ) = 0 ) then { Data }
    begin
        Set_Data ;
    end else
    begin
        { Drill down to next level... }
        Saved_Ptr := Ptr ;
        Ptr := 0 ; { Save our state }
        dec( Element_Count ) ;
        _Put( Level + 1, Index, Data ) ;
        inc( Element_Count ) ;
        Element.Data := Ptr ;

        { Restore our saved state }
        Ptr := Saved_Ptr ;
    end ;
    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
end ; { TSparse_Matrix64._Put }


procedure TSparse_Matrix64.Matrix_Put( Index : string ; var Data ) ;

begin
    _Put( 0, Index, Data ) ;
end ;


procedure TSparse_Matrix64.Matrix_Put_Range( Index : string ; High : integer ;
    var Data ) ; { UNFINISHED }

var Buffer : PChar ;
    Dummy : integer ;
    Element, Element1 : TSparse_Data_Header ;
    Old_Ptr : TStore_Address64 ;
    P : integer ;
    This_Element_Ptr, Previous_Ptr, Next_Ptr : TStore_Address64 ;
    This_Index : longint ;
    Saved_Next : TStore_Address64 ;
    Saved_Ptr : TStore_Address64 ;

    procedure Set_Data ;

    begin
        if( Element_Size <= sizeof( Element.Data ) ) then { If data fits within the element }
        begin
            move( Data, Element.Data, Element_Size ) ;
        end else
        begin
            Element.Data := Heap^.Allocate( Element_Size ) ;
            Heap^.Write( Element.Data, Element_Size, Data ) ;
        end ;
    end ;

begin { TSparse_Matrix64.Matrix_Put_Range }
    { Setup... }
    P := pos( ',', Index + ',' ) ;
    val( copy( Index, 1, P - 1 ), This_Index, Dummy ) ;
    Index := copy( Index, P + 1, length( Index ) ) ;
    This_Element_Ptr := Ptr ;

    { Handle empty matrix }
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
        Ptr := This_Element_Ptr ;
        Element.Low_Index := This_Index ;
        Element.High_Index := High ;
        Element.Next := 0 ;
        if( length( Index ) = 0 ) then
        begin
            Set_Data ;
        end else { Write next level }
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ; { Save our state }
            Ptr := 0 ;
            dec( Element_Count ) ;
            Matrix_Put_Range( Index, High, Data ) ;
            inc( Element_Count ) ;
            Element.Data := Ptr ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end ;
        Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
        exit ;
    end ; { if( This_Element_Ptr = 0 ) }

    { Prepare to search structure... }
    Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;
    Previous_Ptr := 0 ;

    { Now find this element... }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        Previous_Ptr := This_Element_Ptr ;
        This_Element_Ptr := Element.Next ;
        Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;
{ TODO: following to be converted to handle range of elements }
    if( Element.Low_Index > This_Index ) then { Inserting new element prior to item }
    begin
        if( Element.Low_Index = This_Index + 1 ) then { Just before next item }
        begin
            if( length( Index ) = 0 ) then { A data reference }
            begin
                getmem( Buffer, Element_Size ) ;
                if( Element_Size <= sizeof( Element.Data ) ) then
                begin
                    move( Element.Data, Buffer^, Element_Size ) ;
                end else
                begin
                    Heap^.Read( Element.Data, Element_Size, Buffer^ ) ;
                end ;
                if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { Same value as next item }
                begin
                    { If inserting the same value, just prior to another element
                      then adjust index on existing element. }
                    dec( Element.Low_Index ) ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                    freemem( Buffer, Element_Size ) ;
                    exit ;
                end ;
                freemem( Buffer, Element_Size ) ;
            end ; { if( length( Index ) <> 0 ) }
        end ; { if( Element.Low_Index = This_Index + 1 ) }

        { New element prior to This_Element_Ptr... }

        { First create the new header... }
        Element.Next := This_Element_Ptr ;
        This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
        Element.Low_Index := This_Index ;
        Element.High_Index := This_Index ;
        if( length( Index ) = 0 ) then { A data reference }
        begin
            Set_Data ;
        end else
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ; { Save our state }
            Ptr := 0 ;
            dec( Element_Count ) ;
            Matrix_Put_Range( Index, High, Data ) ;
            inc( Element_Count ) ;
            Element.Data := Ptr ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end ;
        Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;

        { Now insert into list... }
        if( Previous_Ptr = 0 ) then { New first element }
        begin
            { Insert at beginning of list }
            Ptr := This_Element_Ptr ;
        end else
        begin
            { Insert after Previous_Ptr... }
            Heap^.Read( Previous_Ptr, sizeof( Element ), Element ) ;
            Element.Next := This_Element_Ptr ;
            Heap^.Write( Previous_Ptr, sizeof( Element ), Element ) ;
        end ;
        exit ;
    end ; { if( Element.Low_Index > This_Index ) }
    if( Element.High_Index >= This_Index ) then { Changing an existing value }
    begin
        if( length( Index ) <> 0 ) then { More subscripts to process}
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ;
            Ptr := Element.Data ; { Save our state }
            Old_Ptr := Element.Data ;
            dec( Element_Count ) ;
            Matrix_Put_Range( Index, High, Data ) ;
            inc( Element_Count ) ;
            if( Ptr <> Old_Ptr ) then { Put changed root pointer }
            begin
                Element.Data := Ptr ;
                Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
            end ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end else
        begin
            { Changing an already-existing value... }
            getmem( Buffer, Element_Size ) ;
            if( Element_Size <= sizeof( Element.Data ) ) then
            begin
                move( Element.Data, Buffer^, Element_Size ) ;
            end else
            begin
                Heap^.Read( Element.Data, Element_Size, Buffer^ ) ;
            end ;
            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { Same value as existing item }
            begin
                freemem( Buffer, Element_Size ) ; { Nothing to do - cleanup and exit }
                exit ;
            end ;

            { Changing a single item }
            if( Element.Low_Index = Element.High_Index ) then
            begin
                if( Element_Size <= sizeof( Element.Data ) ) then
                begin
                    move( Data, Element.Data, Element_Size ) ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                end else
                begin
                    Heap^.Write( Element.Data, Element_Size, Data ) ;
                end ;

                { See if we can combine with the next item }
                if( Element.Next <> 0 ) then { There is a next index }
                begin
                    Heap^.Read( Element.Next, sizeof( Element1 ), Element1 ) ;
                    if( Element1.Low_Index = Element.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Saved_Next := Element.Next ;
                            Element.Next := Element1.Next ;
                            Element.High_Index := Element1.High_Index ;
                            Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                            Heap^.Deallocate( Saved_Next, sizeof( Element1 ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element1.Data, Element_Size ) ;
                            end ;
                        end ;
                    end ;
                end ;

                { Now see if we can combine with the previous item }
                if( Previous_Ptr <> 0 ) then { There was a previous index }
                begin
                    Heap^.Read( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                    if( Element.Low_Index = Element1.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Element1.Next := Element.Next ;
                            Element1.High_Index := Element.High_Index ;
                            Heap^.Write( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                            Heap^.Deallocate( This_Element_Ptr, sizeof( Element ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element.Data, Element_Size ) ;
                            end ;
                        end ;
                    end ;
                end ;
            end else

            { Need to split... }
            if( Element.Low_Index = This_Index ) then
            begin
                { Put new data at current location and rest of existing at new place }
                Next_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                Element.Low_Index := This_Index + 1 ;
                Heap^.Write( Next_Ptr, sizeof( Element ), Element ) ;
                Element.Next := Next_Ptr ;
                Element.Low_Index := This_Index ;
                Element.High_Index := This_Index ;
                Set_Data ;
                Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;

                { Now see if we can combine with the previous item }
                if( Previous_Ptr <> 0 ) then { There was a previous index }
                begin
                    Heap^.Read( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                    if( Element.Low_Index = Element1.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Element1.Next := Element.Next ;
                            Element1.High_Index := Element.High_Index ;
                            Heap^.Write( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                            Heap^.Deallocate( This_Element_Ptr, sizeof( Element ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element.Data, Element_Size ) ;
                            end ;
                        end ; { if( Compare_Memory( Data, Buffer^, Element_Size ) ) }
                    end ; { if( Element.Low_Index = Element1.High_Index + 1 ) }
                end ; { if( Previous_Ptr <> 0 ) }
            end else
            begin
                if( Element.High_Index = This_Index ) then
                begin
                    { Insert new element just after the existing one... }
                    Next_Ptr := Element.Next ;
                    Element.High_Index := This_Index - 1 ;
                    Element.Next := Heap^.Allocate( sizeof( Element ) ) ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                    This_Element_Ptr := Element.Next ;
                    Element.Next := Next_Ptr ;
                    Element.Low_Index := This_Index ;
                    Element.High_Index := This_Index ;
                    Set_Data ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;

                    { See if we can combine with the next item }
                    if( Element.Next <> 0 ) then { There is a next index }
                    begin
                        Heap^.Read( Element.Next, sizeof( Element1 ), Element1 ) ;
                        if( Element1.Low_Index = Element.High_Index + 1 ) then { Immediately follows }
                        begin
                            if( Element_Size <= sizeof( Element.Data ) ) then
                            begin
                                move( Element1.Data, Buffer^, Element_Size ) ;
                            end else
                            begin
                                Heap^.Read( Element1.Data, Element_Size, Buffer^ ) ;
                            end ;
                            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                            begin
                                Saved_Next := Element.Next ;
                                Element.Next := Element1.Next ;
                                Element.High_Index := Element1.High_Index ;
                                Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                                Heap^.Deallocate( Saved_Next, sizeof( Element1 ) ) ;
                                if( Element_Size > sizeof( Element.Data ) ) then
                                begin
                                    Heap^.Deallocate( Element1.Data, Element_Size ) ;
                                end ;
                            end ;
                        end ;
                    end ;
                end else
                begin
                    { New data is in the middle of an existing range... }

                    { Copy high part of range to new range element header... }
                    Next_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                    Element.Low_Index := This_Index + 1 ;
                    if( Element_Size > sizeof( Element.Data ) ) then
                    begin
                        { Copy the data to a new location... }
                        getmem( Buffer, Element_Size ) ;
                        Heap^.Read( Element.Data, Element_Size, Buffer^ ) ;
                        Element.Data := Heap^.Allocate( Element_Size ) ;
                        Heap^.Write( Element.Data, Element_Size, Buffer^ ) ;
                        freemem( Buffer, Element_Size ) ;
                    end ;
                    Heap^.Write( Next_Ptr, sizeof( Element ), Element ) ;

                    { Now create the new element... }
                    Previous_Ptr := This_Element_Ptr ;
                    This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                    Element.Low_Index := This_Index ;
                    Element.High_Index := This_Index ;
                    Element.Next := Next_Ptr ;
                    Set_Data ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;

                    { Now update existing low part of range... }
                    Heap^.Read( Previous_Ptr, sizeof( Element ), Element ) ;
                    Element.High_Index := This_Index - 1 ;
                    Element.Next := This_Element_Ptr ;
                    Heap^.Write( Previous_Ptr, sizeof( Element ), Element ) ;
                end ;
            end ; { if( Element.Low_Index = This_Index ) }
            freemem( Buffer, Element_Size ) ;
        end ;
        exit ;
    end ; { if( Element.High_Index >= This_Index ) }

    { Only other possibility is that this is a new element at the end of the
      list. }
    Element.Next := Heap^.Allocate( sizeof( Element ) ) ;
    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
    This_Element_Ptr := Element.Next ;
    Element.Next := 0 ; { End of list }
    Element.Low_Index := This_Index ;
    Element.High_Index := This_Index ;
    if( length( Index ) = 0 ) then { Data }
    begin
        Set_Data ;
    end else
    begin
        { Drill down to next level... }
        Saved_Ptr := Ptr ;
        Ptr := 0 ; { Save our state }
        dec( Element_Count ) ;
        Matrix_Put_Range( Index, High, Data ) ;
        inc( Element_Count ) ;
        Element.Data := Ptr ;

        { Restore our saved state }
        Ptr := Saved_Ptr ;
    end ;
    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
end ; { TSparse_Matrix64.Matrix_Put_Range }



{$IFDEF DELPHI}
{ TFast_Sparse_Matrix methods... }

destructor TFast_Sparse_Matrix64.Done ;

begin
    Ptr := 0 ; { So our ancestor doesn't try to release it. }

    inherited Done ;
end ;


procedure TFast_Sparse_Matrix64.Clear ;

var Element : TSparse_Data_Header ;
    Saved : TStore_Address64 ;

begin
    while( Ptr <> 0 ) do
    begin
        Heap^.Read( Ptr, sizeof( Element ), Element ) ;
        if( Element_Count > 1 ) then { Clear sub-array }
        begin
            dec( Element_Count ) ;
            Saved := Ptr ;
            Ptr := Element.Data ;
            Clear ;
            Ptr := Saved ;
            inc( Element_Count ) ;
        end else
        begin
            if( Element_Size > sizeof( Element.Data ) ) then
            begin
                Heap^.Deallocate( Element.Data, Element_Size ) ;
            end ;
        end ;

        { Now deallocate this element... }
        Heap^.Deallocate( Ptr, sizeof( Element ) ) ;
        Ptr := Element.Next ;
    end ; { while( Ptr <> nil ) }
end ;


function TFast_Sparse_Matrix64._Element_Exists( Level : integer ;
    Index : array of integer ) : boolean ;

var Element : TSparse_Data_Header ;
    This_Element_Ptr : TStore_Address64 ;
    This_Index : longint ;
    Saved_Ptr : TStore_Address64 ;

begin
    { Setup... }
    _Element_Exists := False ; { Assume failure }
    This_Index := Index[ Level ] ;
    This_Element_Ptr := Ptr ;
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        exit ;
    end ;
    Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;

    { Now find this element... }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        This_Element_Ptr := Element.Next ;
        Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;
    if(
        ( This_Index < Element.Low_Index )
        or
        ( This_Index > Element.High_Index )
      ) then { Element doesn't exist }
    begin
        exit ;
    end ;
    if( length( Index ) = 1 ) then { Getting data, rather than another subscript }
    begin
        if( This_Index <= Element.High_Index ) then
        begin
            _Element_Exists := True ;
        end ;
    end else { Getting another level of subscripting }
    begin
        { Save our state and then point ourselves at the root of the next
          level... }
        Saved_Ptr := Ptr ;
        Ptr := Element.Data ;

        { Now get the data from the subarray }
        _Element_Exists := _Element_Exists( Level + 1, Index ) ;

        { Restore our state }
        Ptr := Saved_Ptr ;
    end ;
end ; { TFast_Sparse_Matrix64._Element_Exists }


function TFast_Sparse_Matrix64.Element_Exists( Index : array of integer ) : boolean ;

begin
    Element_Exists := _Element_Exists( 0, Index ) ;
end ; { TFast_Sparse_Matrix64.Element_Exists }


procedure TFast_Sparse_Matrix64._Get_Matrix( Level : integer ;
    Index : array of integer ; var Data ) ;

var Element : TSparse_Data_Header ;
    This_Element_Ptr : TStore_Address64 ;
    This_Index : longint ;
    Saved_Ptr : TStore_Address64 ;

begin
    { Setup... }
    This_Index := Index[ Level ] ;
    This_Element_Ptr := Ptr ;
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        fillchar( Data, Element_Size, 0 ) ;
        exit ;
    end ;
    Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;

    { Now find this element... }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        This_Element_Ptr := Element.Next ;
        Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;
    if(
        ( This_Index < Element.Low_Index )
        or
        ( This_Index > Element.High_Index )
      ) then { Element doesn't exist }
    begin
        fillchar( Data, Element_Size, 0 ) ;
        exit ;
    end ;
    if( length( Index ) = 1 ) then { Getting data, rather than another subscript }
    begin
        if( This_Index <= Element.High_Index ) then
        begin
            if( Element_Size <= sizeof( Element.Data ) ) then
            begin
                move( Element.Data, Data, Element_Size ) ;
            end else
            begin
                Heap^.Read( Element.Data, Element_Size, Data ) ;
            end ;
        end else
        begin
            fillchar( Data, Element_Size, 0 ) ;
            exit ;
        end ;
    end else { Getting another level of subscripting }
    begin
        { Save our state and then point ourselves at the root of the next
          level... }
        Saved_Ptr := Ptr ;
        Ptr := Element.Data ;

        { Now get the data from the subarray }
        _Get_Matrix( Level, Index, Data ) ;

        { Restore our state }
        Ptr := Saved_Ptr ;
    end ;
end ; { TFast_Sparse_Matrix64._Get_Matrix }


procedure TFast_Sparse_Matrix64.Get_Matrix( Index : array of integer ; var Data ) ;

begin
    _Get_Matrix( 0, Index, Data ) ;
end ;


procedure TFast_Sparse_Matrix64._Put_Matrix( Level : integer ;
    Index : array of integer ; var Data ) ;

var Buffer : PChar ;
    Element, Element1 : TSparse_Data_Header ;
    Old_Ptr : TStore_Address64 ;
    This_Element_Ptr, Previous_Ptr, Next_Ptr : TStore_Address64 ;
    This_Index : longint ;
    Saved_Next : TStore_Address64 ;
    Saved_Ptr : TStore_Address64 ;

    procedure Set_Data ;

    begin
        if( Element_Size <= sizeof( Element.Data ) ) then { If data fits within the element }
        begin
            move( Data, Element.Data, Element_Size ) ;
        end else
        begin
            Element.Data := Heap^.Allocate( Element_Size ) ;
            Heap^.Write( Element.Data, Element_Size, Data ) ;
        end ;
    end ;

begin
    { Setup... }
    This_Index := Index[ Level ] ;
    This_Element_Ptr := Ptr ;

    { Handle empty matrix }
    if( This_Element_Ptr = 0 ) then { Nothing in matrix }
    begin
        This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
        Ptr := This_Element_Ptr ;
        Element.Low_Index := This_Index ;
        Element.High_Index := This_Index ;
        Element.Next := 0 ;
        if( length( Index ) = 1 ) then
        begin
            Set_Data ;
        end else { Write next level }
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ; { Save our state }
            Ptr := 0 ;
            _Put_Matrix( Level + 1, Index, Data ) ;
            Element.Data := Ptr ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end ;
        Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
        exit ;
    end ; { if( This_Element_Ptr = 0 ) }

    { Prepare to search structure... }
    Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;
    Previous_Ptr := 0 ;

    { Now find this element... }
    while( ( Element.High_Index < This_Index ) and ( Element.Next <> 0 ) ) do
    begin
        Previous_Ptr := This_Element_Ptr ;
        This_Element_Ptr := Element.Next ;
        Heap^.Read( This_Element_Ptr, sizeof( Element ), Element ) ;
    end ;
    if( Element.Low_Index > This_Index ) then { Inserting new element prior to item }
    begin
        if( Element.Low_Index = This_Index + 1 ) then { Just before next item }
        begin
            if( length( Index ) = 1 ) then { A data reference }
            begin
                getmem( Buffer, Element_Size ) ;
                if( Element_Size <= sizeof( Element.Data ) ) then
                begin
                    move( Element.Data, Buffer^, Element_Size ) ;
                end else
                begin
                    Heap^.Read( Element.Data, Element_Size, Buffer^ ) ;
                end ;
                if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { Same value as next item }
                begin
                    { If inserting the same value, just prior to another element
                      then adjust index on existing element. }
                    dec( Element.Low_Index ) ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                    freemem( Buffer, Element_Size ) ;
                    exit ;
                end ;
                freemem( Buffer, Element_Size ) ;
            end ; { if( length( Index ) <> 1 ) }
        end ; { if( Element.Low_Index = This_Index + 1 ) }

        { New element prior to This_Element_Ptr... }

        { First create the new header... }
        Element.Next := This_Element_Ptr ;
        This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
        Element.Low_Index := This_Index ;
        Element.High_Index := This_Index ;
        if( length( Index ) = 1 ) then { A data reference }
        begin
            Set_Data ;
        end else
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ; { Save our state }
            Ptr := 0 ;
            _Put_Matrix( Level + 1, Index, Data ) ;
            Element.Data := Ptr ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end ;
        Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;

        { Now insert into list... }
        if( Previous_Ptr = 0 ) then { New first element }
        begin
            { Insert at beginning of list }
            Ptr := This_Element_Ptr ;
        end else
        begin
            { Insert after Previous_Ptr... }
            Heap^.Read( Previous_Ptr, sizeof( Element ), Element ) ;
            Element.Next := This_Element_Ptr ;
            Heap^.Write( Previous_Ptr, sizeof( Element ), Element ) ;
        end ;
        exit ;
    end ; { if( Element.Low_Index > This_Index ) }
    if( Element.High_Index >= This_Index ) then { Changing an existing value }
    begin
        if( length( Index ) <> 1 ) then { More subscripts to process}
        begin
            { Drill down to next level... }
            Saved_Ptr := Ptr ;
            Ptr := Element.Data ; { Save our state }
            Old_Ptr := Element.Data ;
            _Put_Matrix( Level + 1, Index, Data ) ;
            if( Ptr <> Old_Ptr ) then { Put changed root pointer }
            begin
                Element.Data := Ptr ;
                Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
            end ;

            { Restore our saved state }
            Ptr := Saved_Ptr ;
        end else
        begin
            { Changing an already-existing value... }
            getmem( Buffer, Element_Size ) ;
            if( Element_Size <= sizeof( Element.Data ) ) then
            begin
                move( Element.Data, Buffer^, Element_Size ) ;
            end else
            begin
                Heap^.Read( Element.Data, Element_Size, Buffer^ ) ;
            end ;
            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { Same value as existing item }
            begin
                freemem( Buffer, Element_Size ) ; { Nothing to do - cleanup and exit }
                exit ;
            end ;

            { Changing a single item }
            if( Element.Low_Index = Element.High_Index ) then
            begin
                if( Element_Size <= sizeof( Element.Data ) ) then
                begin
                    move( Data, Element.Data, Element_Size ) ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                end else
                begin
                    Heap^.Write( Element.Data, Element_Size, Data ) ;
                end ;

                { See if we can combine with the next item }
                if( Element.Next <> 0 ) then { There is a next index }
                begin
                    Heap^.Read( Element.Next, sizeof( Element1 ), Element1 ) ;
                    if( Element1.Low_Index = Element.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Saved_Next := Element.Next ;
                            Element.Next := Element1.Next ;
                            Element.High_Index := Element1.High_Index ;
                            Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                            Heap^.Deallocate( Saved_Next, sizeof( Element1 ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element1.Data, Element_Size ) ;
                            end ;
                        end ;
                    end ;
                end ;

                { Now see if we can combine with the previous item }
                if( Previous_Ptr <> 0 ) then { There was a previous index }
                begin
                    Heap^.Read( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                    if( Element.Low_Index = Element1.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Element1.Next := Element.Next ;
                            Element1.High_Index := Element.High_Index ;
                            Heap^.Write( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                            Heap^.Deallocate( This_Element_Ptr, sizeof( Element ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element.Data, Element_Size ) ;
                            end ;
                        end ;
                    end ;
                end ;
            end else

            { Need to split... }
            if( Element.Low_Index = This_Index ) then
            begin
                { Put new data at current location and rest of existing at new place }
                Next_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                Element.Low_Index := This_Index + 1 ;
                Heap^.Write( Next_Ptr, sizeof( Element ), Element ) ;
                Element.Next := Next_Ptr ;
                Element.Low_Index := This_Index ;
                Element.High_Index := This_Index ;
                Set_Data ;
                Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;

                { Now see if we can combine with the previous item }
                if( Previous_Ptr <> 0 ) then { There was a previous index }
                begin
                    Heap^.Read( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                    if( Element.Low_Index = Element1.High_Index + 1 ) then { Immediately follows }
                    begin
                        if( Element_Size <= sizeof( Element.Data ) ) then
                        begin
                            move( Element1.Data, Buffer^, Element_Size ) ;
                        end else
                        begin
                            Heap^.Read( Element1.Data, Element_Size, Buffer^ ) ;
                        end ;
                        if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                        begin
                            Element1.Next := Element.Next ;
                            Element1.High_Index := Element.High_Index ;
                            Heap^.Write( Previous_Ptr, sizeof( Element1 ), Element1 ) ;
                            Heap^.Deallocate( This_Element_Ptr, sizeof( Element ) ) ;
                            if( Element_Size > sizeof( Element.Data ) ) then
                            begin
                                Heap^.Deallocate( Element.Data, Element_Size ) ;
                            end ;
                        end ;
                    end ;
                end ;
            end else
            begin
                if( Element.High_Index = This_Index ) then
                begin
                    { Insert new element just after the existing one... }
                    Next_Ptr := Element.Next ;
                    Element.High_Index := This_Index - 1 ;
                    Element.Next := Heap^.Allocate( sizeof( Element ) ) ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                    This_Element_Ptr := Element.Next ;
                    Element.Next := Next_Ptr ;
                    Element.Low_Index := This_Index ;
                    Element.High_Index := This_Index ;
                    Set_Data ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;

                    { See if we can combine with the next item }
                    if( Element.Next <> 0 ) then { There is a next index }
                    begin
                        Heap^.Read( Element.Next, sizeof( Element1 ), Element1 ) ;
                        if( Element1.Low_Index = Element.High_Index + 1 ) then { Immediately follows }
                        begin
                            if( Element_Size <= sizeof( Element.Data ) ) then
                            begin
                                move( Element1.Data, Buffer^, Element_Size ) ;
                            end else
                            begin
                                Heap^.Read( Element1.Data, Element_Size, Buffer^ ) ;
                            end ;
                            if( Compare_Memory( Data, Buffer^, Element_Size ) ) then { A match }
                            begin
                                Saved_Next := Element.Next ;
                                Element.Next := Element1.Next ;
                                Element.High_Index := Element1.High_Index ;
                                Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
                                Heap^.Deallocate( Saved_Next, sizeof( Element1 ) ) ;
                                if( Element_Size > sizeof( Element.Data ) ) then
                                begin
                                    Heap^.Deallocate( Element1.Data, Element_Size ) ;
                                end ;
                            end ;
                        end ;
                    end ;
                end else
                begin
                    { New data is in the middle of an existing range... }

                    { Copy high part of range to new range element header... }
                    Next_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                    Element.Low_Index := This_Index + 1 ;
                    if( Element_Size > sizeof( Element.Data ) ) then
                    begin
                        { Copy the data to a new location... }
                        getmem( Buffer, Element_Size ) ;
                        Heap^.Read( Element.Data, Element_Size, Buffer^ ) ;
                        Element.Data := Heap^.Allocate( Element_Size ) ;
                        Heap^.Write( Element.Data, Element_Size, Buffer^ ) ;
                        freemem( Buffer, Element_Size ) ;
                    end ;
                    Heap^.Write( Next_Ptr, sizeof( Element ), Element ) ;

                    { Now create the new element... }
                    Previous_Ptr := This_Element_Ptr ;
                    This_Element_Ptr := Heap^.Allocate( sizeof( Element ) ) ;
                    Element.Low_Index := This_Index ;
                    Element.High_Index := This_Index ;
                    Element.Next := Next_Ptr ;
                    Set_Data ;
                    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;

                    { Now update existing low part of range... }
                    Heap^.Read( Previous_Ptr, sizeof( Element ), Element ) ;
                    Element.High_Index := This_Index - 1 ;
                    Element.Next := This_Element_Ptr ;
                    Heap^.Write( Previous_Ptr, sizeof( Element ), Element ) ;
                end ;
            end ; { if( Element.Low_Index = This_Index ) }
            freemem( Buffer, Element_Size ) ;
        end ;
        exit ;
    end ; { if( Element.High_Index >= This_Index ) }

    { Only other possibility is that this is a new element at the end of the
      list. }
    Element.Next := Heap^.Allocate( sizeof( Element ) ) ;
    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
    This_Element_Ptr := Element.Next ;
    Element.Next := 0 ; { End of list }
    Element.Low_Index := This_Index ;
    Element.High_Index := This_Index ;
    if( length( Index ) = 1 ) then { Data }
    begin
        Set_Data ;
    end else
    begin
        { Drill down to next level... }
        Saved_Ptr := Ptr ;
        Ptr := 0 ; { Save our state }
        _Put_Matrix( Level + 1, Index, Data ) ;
        Element.Data := Ptr ;

        { Restore our saved state }
        Ptr := Saved_Ptr ;
    end ;
    Heap^.Write( This_Element_Ptr, sizeof( Element ), Element ) ;
end ; { TFast_Sparse_Matrix64._Put_Matrix }


procedure TFast_Sparse_Matrix64.Put_Matrix( Index : array of integer ; var Data ) ;

begin
    _Put_Matrix( 0, Index, Data ) ;
end ;
{$ENDIF}


{ TGrouped_Sparse_Matrix methods... }

destructor TGrouped_Sparse_Matrix64.Done ;

begin
    Ptr := 0 ; { So our ancestor doesn't try to release it. }

    inherited Done ;
end ;


procedure TGrouped_Sparse_Matrix64.Clear ;

begin
end ;


function TGrouped_Sparse_Matrix64.Element_Exists( Index : string ) : boolean ;

begin
    Element_Exists := False ;
end ;


procedure TGrouped_Sparse_Matrix64.Matrix_Get( Index : string ; var Data ) ;

begin
end ;


procedure TGrouped_Sparse_Matrix64.Matrix_Put( Index : string ; var Data ) ;

begin
end ;



{$IFDEF DELPHI}

{ TFast_Grouped_Sparse_Matrix64 methods... }

destructor TFast_Grouped_Sparse_Matrix64.Done ;

begin
    Ptr := 0 ; { So our ancestor doesn't try to release it. }

    inherited Done ;
end ;


procedure TFast_Grouped_Sparse_Matrix64.Clear ;

begin
end ;


function TFast_Grouped_Sparse_Matrix64.Element_Exists( Index : array of integer ) : boolean ;

begin
    Element_Exists := False ;
end ;


procedure TFast_Grouped_Sparse_Matrix64.Get_Matrix( Index : array of integer ; var Data ) ;

begin
end ;


procedure TFast_Grouped_Sparse_Matrix64.Put_Matrix( Index : array of integer ; var Data ) ;

begin
end ;



{ TFast_Grouped_Sparse_Matrix methods... }

destructor TFast_Grouped_Sparse_Matrix.Done ;

begin
    Ptr := 0 ; { So our ancestor doesn't try to release it. }

    inherited Done ;
end ;


procedure TFast_Grouped_Sparse_Matrix.Clear ;

begin
end ;


function TFast_Grouped_Sparse_Matrix.Element_Exists( Index : array of integer ) : boolean ;

begin
    Element_Exists := False ;
end ;


procedure TFast_Grouped_Sparse_Matrix.Get_Matrix( Index : array of integer ; var Data ) ;

begin
end ;


procedure TFast_Grouped_Sparse_Matrix.Put_Matrix( Index : array of integer ; var Data ) ;

begin
end ;
{$ENDIF}


{$IFDEF TEST}
procedure Show_Error( const S : string ) ;

begin
    writeln( S ) ;
    halt ;
end ;


procedure Test_Double_List ; { Tests TDList and Double_List }

var I : integer ;
    Test : Double_List ;

begin
    Test.Init( sizeof( integer ), False ) ;
    Test.Heap := new( PSHeap, Init ) ;

    if( not Test.Empty ) then
    begin
        Show_Error( 'TDList Empty failed' ) ;
    end ;
    if( Test.EOL = False ) then
    begin
        Show_Error( 'Double_List EOL failed' ) ;
    end ;
    if( Test.BOL = False ) then
    begin
        Show_Error( 'Double_List EOL failed' ) ;
    end ;
    Test.Clear ;
    I := 1 ;
    Test.Add( I ) ;
    if( Test.Empty ) then
    begin
        Show_Error( 'TDList Empty failed' ) ;
    end ;
    if( Test.EOL = False ) then
    begin
        Show_Error( 'Double_List EOL failed' ) ;
    end ;
    if( Test.BOL = False ) then
    begin
        Show_Error( 'Double_List EOL failed' ) ;
    end ;
    Test.First( I ) ;
    if( I <> 1 ) then
    begin
        Show_Error( 'Double_List First failed' ) ;
    end ;
    Test.Last( I ) ;
    if( I <> 1 ) then
    begin
        Show_Error( 'Double_List Last failed' ) ;
    end ;
    Test.Delete ;
    if( not Test.Empty ) then
    begin
        Show_Error( 'TDList Empty failed' ) ;
    end ;
    if( Test.EOL = False ) then
    begin
        Show_Error( 'Double_List EOL failed' ) ;
    end ;
    if( Test.BOL = False ) then
    begin
        Show_Error( 'Double_List EOL failed' ) ;
    end ;
    I := 1 ;
    Test.Add( I ) ;
    I := 2 ;
    Test.Add( I ) ;
    Test.First( I ) ;
    if( I <> 1 ) then
    begin
        Show_Error( 'Double_List First failed' ) ;
    end ;
    Test.Last( I ) ;
    if( I <> 2 ) then
    begin
        Show_Error( 'Double_List Last failed' ) ;
    end ;
    Test.Previous( I ) ;
    if( I <> 1 ) then
    begin
        Show_Error( 'Double_List Previous failed' ) ;
    end ;
    Test.Delete ;
    Test.First( I ) ;
    if( I <> 2 ) then
    begin
        Show_Error( 'Double_List Delete failed' ) ;
    end ;
    I := 4 ;
    Test.Insert( I ) ;
    Test.First( I ) ;
    I := 3 ;
    Test.Insert( I ) ;
    Test.First( I ) ;
    if( I <> 2 ) then
    begin
        Show_Error( 'Double_List Insert failed' ) ;
    end ;
    Test.Get( I ) ;
    if( I <> 3 ) then
    begin
        Show_Error( 'Double_List Insert failed' ) ;
    end ;
    Test.Get( I ) ;
    if( I <> 4 ) then
    begin
        Show_Error( 'Double_List Insert failed' ) ;
    end ;
    if( Test.BOL ) then
    begin
        Show_Error( 'Double_List BOL failed' ) ;
    end ;
    Test.First( I ) ;
    if( I <> 2 ) then
    begin
        Show_Error( 'Double_List Get failed' ) ;
    end ;
    if( Test.EOL ) then
    begin
        Show_Error( 'Double_List EOL failed' ) ;
    end ;
    Test.Delete ;
    Test.Get( I ) ;
    if( I <> 3 ) then
    begin
        Show_Error( 'Double_List Get failed' ) ;
    end ;
    I := 0 ;
    Test.Insert( I ) ;
    Test.Clear ;
    if( Test.List_First <> 0 ) then
    begin
        Show_Error( 'TDList List_First failed' ) ;
    end ;
    if( Test.List_Previous( Test.Ptr ) <> 0 ) then
    begin
        Show_Error( 'TDList List_Previous failed' ) ;
    end ;
    if( Test.List_Next( Test.Ptr ) <> 0 ) then
    begin
        Show_Error( 'TDList List_Next failed' ) ;
    end ;
    I := 0 ;
    Test.Insert( I ) ;
    if( Test.List_Previous( Test.Ptr ) <> 0 ) then
    begin
        Show_Error( 'TDList List_Previous failed' ) ;
    end ;
    I := 1 ;
    Test.Insert( I ) ;
    Test.Delete ;
    I := 2 ;
    Test.Insert( I ) ;
    Test.First( I ) ;
    if( I <> 2 ) then
    begin
        Show_Error( 'Double_List First failed' ) ;
    end ;
    I := 3 ;
    Test.Put( I ) ;
    Test.First( I ) ;
    if( I <> 3 ) then
    begin
        Show_Error( 'Double_List First failed' ) ;
    end ;
    Test.Put( I ) ;

    Test.Clear ;
    I := 0 ;
    Test.Insert( I ) ;
    I := 1 ;
    Test.Insert( I ) ;
    Test.First( I ) ;
    I := 2 ;
    Test.Insert( I ) ;
    Test.First( I ) ;
    if( I <> 0 ) then
    begin
        Show_Error( 'Double_List linkage failure' ) ;
    end ;
    Test.Get( I ) ;
    if( I <> 2 ) then
    begin
        Show_Error( 'Double_List linkage failure' ) ;
    end ;
    Test.Get( I ) ;
    if( I <> 1 ) then
    begin
        Show_Error( 'Double_List linkage failure' ) ;
    end ;
    Test.Previous( I ) ;
    if( I <> 2 ) then
    begin
        Show_Error( 'Double_List linkage failure' ) ;
    end ;
    Test.Previous( I ) ;
    if( I <> 0 ) then
    begin
        Show_Error( 'Double_List linkage failure' ) ;
    end ;

    Test.Done ;
end ; { Test_Double_List }


procedure Test_Unit ;

begin
    Test_Double_List ;
end ;
{$ENDIF}

end.

